/*
 * BanRecord.C - member functions for the BanRecord class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "BanRecord.h"
#include "Channel.h"
#define CLASS "BanRecord"

BanRecord::BanRecord (Channel * chn, char * desc) 
: IgnoreRecord ( ) {
   FUNCTION;
   char * host, * expi, * add, * last, * usr;
   int flgs = 0;
   
   chan = (chn == G->chanset) ? 0 : chn;
   host = G->Split(&desc);
   expi = G->Split(&desc);
   if (*expi == '+') {
      flgs |= PERM;
      expi++;
   }
   add = G->Split(&desc);
   if (*add=='*') {
      flgs |= STICKY;
      add++;
   }
   last = G->Split(&desc);
   usr = G->Split(&desc);
   if (*desc) {
      KeyedList_Cons( *(chan ? chan->Bans() : G->global_bans), host, 
		     KeyedList<UHOSTLEN>::APPEND | KeyedList<UHOSTLEN>::NODUPS);
      expires = atoi(expi);
      added = atoi(add);
      lastactive = atoi(last);
      flags = flgs;
      strncpy(user, usr, HANDLEN);
      user[HANDLEN] = 0;
      msg = (char *)my_malloc(strlen(desc)+1,G->memtype_ban);
      strcpy(msg, desc);
      if (!chan)
	G->gban_total++;
   } else {
      G->LOG_MISC.Put(G->Lang(0x277, "s", chan ? chan->Key() : "global bans"));
      DELETE ( this );
   }
}

BanRecord::BanRecord ( Channel * c, const char * h, const char * u,
		      const char * m, time_t e, int f)
: IgnoreRecord () {
   FUNCTION;
   expires = e;
   chan = (c == G->chanset) ? 0 : c;
   added = G->now;
   lastactive = 0;
   flags = f;
   KeyedList_Cons(*(chan ? chan->Bans() : G->global_bans), h,
		  KeyedList<UHOSTLEN>::APPEND | KeyedList<UHOSTLEN>::NODUPS);
   strncpy(user, u, HANDLEN);
   user[HANDLEN] = 0;
   msg = (char *)my_malloc(strlen(m)+1,G->memtype_ban);
   strcpy(msg, m);
   if (!chan)
     G->gban_total++;
}

BanRecord::BanRecord ( Channel * ch ) : IgnoreRecord () {
   FUNCTION;
   if ((chan = ch)) 
     chan->bans = this;
   else 
     G->global_bans = this;
}

BanRecord::~BanRecord ( void ) {
   FUNCTION;
   if (key[0]) {
      if (chan)
	chan->Bans()->Remove(this);
      else {
	 G->global_bans->Remove(this);
	 G->gban_total--;
      }
   } else 
     while (next) {
       DELETE (next);
     }
   if (key[0] && !G->noshare && G->shareout)
     /* distribute chan bans differently */
     G->shareout(chan,"B%s %s\n", chan ? chan->Key() : "*", key);
   if (msg) {
      my_free(msg, G->memtype_ban);
      msg = 0;
   };
   key[0] = 0;
}

int BanRecord::ExpMem ( void ) {
   FUNCTION;
   int tot = sizeof(BanRecord);
   
   if (!key[0] && !chan) {
      Channel *chan = G->chanset;
      BanRecord * b = this;
      
      context;
      while ((b = b->Next()))
	tot += b->ExpMem();
      context;
      if (chan) 
	while ((chan = chan->Next())) {
	   tot += sizeof(BanRecord);
	   for (b = chan->Bans(); (b = b->Next()); ) 
	     tot += b->ExpMem();
	}
      context;
   } else if (key[0])
     tot += strlen(msg) + 1;
   return tot;
}

int BanRecord::Write ( FILE * f ) {
   FUNCTION;
   return (fprintf(f,"-%s %s%lu %s%lu %lu %s %s\n", key, 
		   (flags & PERM) ? "+" : "", expires,
		   (flags & STICKY) ? "*" : "", added, lastactive,
		   user[0] ? user : G->botnetnick,
		   msg[0] ? msg : "requested") != EOF);
}
