/*
 * BindEntry.C - member functions for BindEntry class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "BindEntry.h"
#include "BindTable.h"
#include "Socket.h"
#include "ModuleEntry.h"
#include "ScriptContext.h"
#define CLASS "BindEntry"

BindEntry::BindEntry ( BindEntry & mask, const char * proc, const char * fl,
		      BindReturn (*fc) (void **, const char *), ModuleEntry * m,
		      BindMask * o, ScriptContext * sc)
: KeyedList<MISCLEN> ( mask, proc, 0, m ) {
   FUNCTION;
   owner = o;
   script = sc;
   Reset(fl, fc);
}

BindEntry::~BindEntry ( void ) {
   FUNCTION;
}

void BindEntry::Reset (const char * fl, BindReturn (*fc) (void **, const char *) ) {
   FUNCTION;
   flags.SetMatch ( FlagRecord::GLOBAL | FlagRecord::CHANNEL | FlagRecord::BOT);
   flags.BreakDown(fl);
   F = fc;
}

int BindEntry::ExpMem ( void ) {
   FUNCTION;
   return (key[0] ? sizeof(BindEntry) : 0) 
     + (next ? Next()->ExpMem () : 0);
}

void BindEntry::Dump (ScriptContext * sc) {
   FUNCTION;
   BindEntry * p = this;
   const char * q;
   
   while ((p = p->Next()))
     sc->AppendArray(1, &(q = p->key));
}

int BindEntry::Display ( BindTable * table, char * mask, Socket * idx, 
			int showall ) {   
   FUNCTION;
   char flg[100];
   
   flags.Build(flg);
   if (showall || (key[0] != '*' || !strchr(key, ':'))) {
      idx->Send(G->Lang(0x14B, "s s s 4scs", table->Key(), 7 - strlen(table->Key()),
			flg, 8 - strlen(flg),mask, 20 - strlen(mask), hits,
			key, F ? 'C' : 'S', mod ? mod->Key() : ""));
      return 1;
   }
   return 0;
}

BindReturn BindEntry::Check ( const char * match, FlagRecord * fr, 
			 void ** args, int flg) {	 
   FUNCTION;
   if (!(flg & BindTable::ATTR) || ((flg & BindTable::BUILTINS) &&
				    flags.Ok(*fr))
       || (!(flg & BindTable::BUILTINS) && (flags == *fr))) {
      hits++;
      return F ? F(args, match) : owner->Owner()->ToScript(this, args, match);
   }
   return NOMATCH;
}

BindReturn BindEntry::Call (const char * mask, int l, ... ) {
   FUNCTION;
   BindReturn ret = NOMATCH;
   va_list val;
   
   va_start(val, l);
   if (script) 
     ret = script->CallBind(this, mask, l, val);
   va_end(val);
   return ret;
}
