/*
 * BindMask.C - member functions for BindMask class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "BindMask.h"
#include "BindTable.h"
#define CLASS "BindMask"

BindMask::~BindMask ( void ) {
   FUNCTION;
}

int BindMask::ExpMem ( void ) {
   FUNCTION;
   return (key[0] ? sizeof(BindMask) : 0)
     + first.ExpMem() + (next ? Next()->ExpMem() : 0);
}

int BindMask::RemBind ( const char * proc ) {
   FUNCTION;
   BindEntry * be;
   
   if ((be = first.Find ( proc ))) {
      first.Remove( be );
      DELETE(be);
      return 1;
   }
   return 0;
}

BindEntry * BindMask::AddBind ( const char * flags, const char * proc,
			       ScriptContext * sc,
			       BindReturn (* func) (void **, const char *), int fl,
			       ModuleEntry * m) {
   FUNCTION;
   BindEntry * be = first.Find ( proc );
   
   if (be)
     be->Reset ( flags, func );
   else {
      if (!fl && (be = first.Next())) {
	 first.Remove(be);
	 delete be;
      }
      be = NEW ( BindEntry ( first, proc, flags, func, m, this, sc ) );
   }
   return be;
}

int BindMask::Display ( BindTable * table, Socket * idx, int showall ) {
   FUNCTION;
   BindEntry * be = &first;
   int fnd = 0;
   
   while ((be = be->Next()))
     fnd += be->Display(table, key, idx, showall);
   return fnd + (next ? Next()->Display(table, idx, showall) : 0);
}

BindReturn BindMask::Check ( const char * match, FlagRecord * fr, 
			void ** args, int flags ) { 	    
   FUNCTION;
   BindEntry * be = &first;
   BindReturn ret = NOMATCH;
   
   while ((be = be->Next())) {
      ret = be->Check( match, fr, args, flags);
      if (flags & BindTable::ALTER_ARGS) {
	 args[0] = (void *)ret;
	 if (!ret || !((char *)ret)[0]) 
	   return ret;
	 match = (const char *)ret;
	 if (!G->WildMatchPer((unsigned char *)key, (unsigned char *)match))
	   return ret;
      } 
      if (!(flags & BindTable::STACKABLE)) 
	return ret;
   }
   return ret;
}
