/*
 * BindTable.C - member functions for BindTable class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "BindTable.h"
#include "Socket.h"
#include "ScriptContext.h"
#include "BoundCommand.h"
#define CLASS "BindTable"

int BindTable::ExpMem ( void ) {
   FUNCTION;
   BindTable * ptr = this;
   int tot = sizeof(BindTable);
   
   while ((ptr = ptr->Next()))
     tot += sizeof(BindTable) + ptr->first.ExpMem();
   return tot;
}

BindTable::~BindTable ( ) {
   FUNCTION;
   if (key[0]) 
     G->bind_list->Remove(this);
   else
     while (next)
       DELETE(next);
}

void BindTable::Dump (ScriptContext * sc) {
   FUNCTION;
   BindTable * p = this;
   const char * q;
   
   while ((p = p->Next())) 
     sc->AppendArray(1, &(q = p->key));
}

void BindTable::AddBinds ( command_binds * cc, int count, ModuleEntry * m) {
   FUNCTION;
   char p[512];
   BindEntry * be;
      
   context;
   while (count) {
      count --;
      G->SPrintF(p, 511, "*%s:%s", key,
		 cc[count].funcname ? cc[count].funcname : cc[count].name);
      be = AddBind(cc[count].flags,cc[count].name,p,0,cc[count].func, m);
      if (flags & CALLBACK) 
	NEW ( BoundCommand ( p, this, cc[count].func, m ) );
   }
}

/* bring the default msg/dcc/fil commands into the Tcl interpreter */
void BindTable::RemBinds (command_binds * cc, int count) {
   FUNCTION;
   char p[512];
   
   while (count) {
      count --;
      G->SPrintF(p, 511, "*%s:%s", key,
		 cc[count].funcname ? cc[count].funcname : cc[count].name);
      RemBind (cc[count].name,p);
      if (flags & CALLBACK) {
	 BoundCommand * bc = (BoundCommand *)G->command_list->Find(p);
	 if (bc)
	   DELETE (bc);
      }
   }
}

int BindTable::RemBind ( const char * cmd, const char * proc ) {
   FUNCTION;
   BindMask * ma;
   int r = 0;
   
   if ((ma = first.Find ( cmd ))) {
      r = ma->RemBind ( proc );
      if (!ma->First()) {
	 first.Remove( ma );
	 DELETE(ma);
      }
   }
   return r;
}

BindEntry * BindTable::AddBind ( const char * fgs, const char * cmd, 
				const char * proc, ScriptContext * sc,
				BindReturn (* func) (void **, const char *), 
				ModuleEntry * m) {
   FUNCTION;
   BindMask * ma;
   
   if (!(ma = first.Find ( cmd ))) 
      ma = NEW( BindMask ( first, cmd, this ));
   return ma->AddBind ( fgs, proc, sc, 
		       
		       func, flags & STACKABLE, m );
}

void BindTable::Display ( Socket * idx, char * xtra ) {
   FUNCTION;
   char * name = G->Split(&xtra);
   int showall = 0, fnd = 0;
   BindTable * ptr, * kind = 0;
   
   if (!strcasecmp(xtra, "all") || !strcasecmp(name, "all"))
     showall = 1;
   kind = Find(name);
   idx->Send(G->Lang(0x149, ""));
   idx->Send(G->Lang(0x14A, ""));
   for (ptr = kind ? kind : Next(); ptr; ptr = kind ? 0 : ptr->Next()) 
     fnd += ptr->First()->Display(ptr, idx, showall);
   if (!fnd) 
     idx->Send(G->Lang(kind ? 0x147 : 0x148, "s", name));
}

BindReturn BindTable::Check ( const char * match, FlagRecord * fr,
			     void ** args ) {
   FUNCTION;
   BindMask * bm = &first, *lbm = 0;
   int ok = 0, cnt = 0;
   BindReturn ret;
   
   while ((bm = bm->Next())) {
      switch (flags & 0x03) {
	case PARTIAL:
	 ok = !strncasecmp(match, bm->Key(), strlen(match));
	 break;
	case EXACT:
	 ok = !strcasecmp(match, bm->Key());
	 break;
	case MASK:
	 ok = G->WildMatchPer((unsigned char *)bm->Key(), (unsigned char *) match);
      }
      if (ok) {
	 cnt++;
	 if (flags & STACKABLE) {
	    ret = bm->Check ( match, fr, args, flags );
	    if ((flags & ALTER_ARGS) && (!ret || !((char *)ret)[0]))
	      return ret;
	    if ((flags & WANTRET) && (ret == EXEC_LOG))
	      return ret;
	 } else {
	    lbm = bm;
	    if (!((flags & 0x03) == PARTIAL) || 
		!strcasecmp(match, bm->Key())) {
	       cnt = 1;
	       break;
	    }
	 }
      }
   }
   if (flags & ALTER_ARGS)
     return EXECUTED;
   if (!cnt)
     return NOMATCH;
   if (flags & STACKABLE)
     return EXECUTED;
   if (cnt > 1)
     return AMBIGUOUS;
   first.Remove(lbm);
   first.Add(lbm);
   return lbm->Check( match, fr, args, flags);
}

void BindTable::Cleanup ( ModuleEntry * m ) {
   FUNCTION;
   BindTable * ptr, *pback = this;
   BindMask * pbm, *pbmnext;
   
   while ((ptr = pback->Next())) {
      if (ptr->Module() == m) {
	 DELETE ( ptr );
      } else {
	 pback = ptr;
	 pbmnext = ptr->first.Next();
	 while ((pbm = pbmnext)) {
	    pbmnext = pbm->Next();
	    pbm->Cleanup(m);
	 }
      }
   }
}

int BindTable::FromScript (BoundCommand *, ScriptContext *, int, ScriptArgument **) {
   FUNCTION;
   return SCRIPT_ERROR;
}

BindReturn BindTable::ToScript (BindEntry *, void **, const char *) {
   FUNCTION;
   return NOMATCH;
}
