/*
 * BindTable.h - class definition for BindTable class
 */

#ifndef _KEYEDLIST_
#include "KeyedList.h"
#endif

#ifndef _BINDMASK_
#include "BindMask.h"
#endif

#ifndef _BINDTABLE_
#define _BINDTABLE_

class BindTable : public KeyedList<MISCLEN> {
  protected:
   BindMask first;
   int flags;
   friend class BindEntry;
   
  public:
   enum {
      PARTIAL = 0, EXACT = 1, MASK = 2, ATTR = 4, STACKABLE = 8,
	BUILTINS = 16, WANTRET = 32, ALTER_ARGS = 64, CALLBACK = 128,
	SCRIPT_FUNC = 256
	};
   inline BindTable ( ) : KeyedList<MISCLEN> ( ) { }
   inline BindTable ( const char * nme, int flg, ModuleEntry * m = 0)
     : KeyedList<MISCLEN> (*G->bind_list,nme,KeyedList<MISCLEN>::NODUPS,m) {
	flags = flg;
     }
   ~BindTable ();

   virtual int FromScript (BoundCommand *, ScriptContext *, int, ScriptArgument **);
   virtual BindReturn ToScript (BindEntry *,void **, const char *);

   void AddBinds ( command_binds *, int, ModuleEntry * );
   void RemBinds ( command_binds *, int );
   BindEntry * AddBind ( const char *, const char *, const char *,
			ScriptContext *, BindReturn (*) (void **, const char *),
			ModuleEntry *);
   int RemBind ( const char *, const char * );
   BindReturn Check ( const char *, FlagRecord *, void **);
   void Dump (ScriptContext *);
   int ExpMem ( void );
   void Cleanup ( ModuleEntry * m );
   void Display (Socket *, char *);
   BindReturn TriggerBind (char *, char *, char *);
   
   inline BindMask * First ( void ) { return &first; }
   inline BindTable * Find ( const char * x) {
      return (BindTable *)KeyedList<MISCLEN>::Find(x);
   }
   inline int Flags ( void ) { return flags; }
   MEMORY_TYPE(G->memtype_bind);
   DEF_NEXT(BindTable);
};

#endif
