/*
 * Channel.h - stubs for channel class in core
 */

#ifndef _KEYEDLIST_
#include "KeyedList.h"
#endif

#ifndef _CHANNEL_
#define _CHANNEL_

/* only this part is needed by most modules, lets hide the rest */
class Channel : public KeyedList<CHANLEN> {
  protected:
   BanRecord *bans;      /* temporary channel bans */
   int status;

   friend class BanRecord;
   friend class UserRecord;
  public:
   enum { SECRET = 1, MEOP = 2, ACTIVE = 4, PEND = 8 };
   Channel ( Channel & );
   inline Channel ( void ) : KeyedList<CHANLEN> () { }
   Channel ( const char * );
   ~Channel ( void );
   
   virtual Channel * NewChannel ( const char * ) = 0;
   inline Channel * Find ( const char * n ) {
      return this ? (Channel *)KeyedList<CHANLEN>::Find(n) : 0;
   }
   inline Channel * Next ( void ) { return (Channel *)next; }
     
   inline int Secret ( void ) { return status & SECRET; };
   inline int Active ( void ) { return status & ACTIVE; };
   inline int Pending ( void ) { return status & PEND; };
   inline int Meop ( void ) { return status & MEOP; };
   inline int SetMeop ( void ) { return status |= MEOP; };
   inline int UnsetMeop ( void ) { return status &= ~MEOP; };
   
   inline void SetActive ( void ) { status |= ACTIVE; }
   inline void SetPending ( void ) { status |= PEND; }
   
   inline void UnsetActive ( void ) { status &= ~ACTIVE; }
   inline void UnsetPending ( void ) { status &= ~PEND; }
   
   inline BanRecord * Bans ( void ) { return bans; }
   MEMORY_TYPE(G->memtype_chanset);
};
#endif
