/*
 * DccChat.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "DccChat.h"
#include "FlagRecord.h"
#include "HelpReference.h"
#include "BindTable.h"
#include "UserEntryLaston.h"
#define CLASS "DccChat"

DccChat::DccChat ( DccChatPass * old ) : DccChatPass(*old) {
   FUNCTION;
   SetType(G->DCC_CHAT);
   timeout_val = 0;
   type_flags = DCT_CHAT | DCT_MASTER | DCT_SHOWWHO | DCT_VALIDIDX
     | DCT_SIMUL | DCT_CANBOOT;
   con_flags = 0;
   msgs_per_sec = 0;
   con_chan[0] = 0;
   channel = -2;
}

DccChat::DccChat ( DccChat & old ) : DccChatPass(old) {
   FUNCTION;
   type = G->DCC_CHAT;
   type_flags = DCT_CHAT | DCT_MASTER | DCT_SHOWWHO | DCT_VALIDIDX
     | DCT_SIMUL | DCT_CANBOOT;
   timeout_val = old.timeout_val;
   con_flags = old.con_flags;
   msgs_per_sec = old.msgs_per_sec;
   strcpy(con_chan, old.con_chan);
   channel = old.channel;
}
  
DccChat::DccChat ( int fd ) : DccChatPass(fd, G->DCC_CHAT,
					  DCT_CHAT | DCT_MASTER | DCT_SHOWWHO
					  | DCT_VALIDIDX | DCT_SIMUL | 
					  DCT_CANBOOT) {
   FUNCTION;
   con_flags = G->conmask;
   msgs_per_sec = 0;
   con_chan[0] = 0;
   channel = 0;
}
 
DccChat::~DccChat ( void ) {
   FUNCTION;
}

void DccChat::Eof (void) {
   FUNCTION;
   context;
   con_flags.Clear();
   Close();
   G->LOG_MISC.Put(G->Lang( 0x14E, "ssd", nick, host, remoteport));
   CheckChof(G->Lang( 0x153, ""));
   context;
}

/* check for tcl-bound dcc command, return 1 if found */
/* dcc: proc-name <handle> <sock> <args...> */
inline int DccChat::CheckDcc (char * cmd, char * args) {
   FUNCTION;
   FlagRecord fr (user, FlagRecord::GLOBAL | FlagRecord::CHANNEL, con_chan);
   void * argv[2] = { this, args};
   BindReturn result;
   
   context;
   if ((result = G->bt_dcc->Check(cmd, &fr, argv)) == EXEC_BRK)
     return 1;			/* quit */
   else if (result == AMBIGUOUS)
     Send(G->Lang( 0x1E9, ""));
   else if (result == NOMATCH)
     Send(G->Lang( 0x1EA, ""));
   else if (result == EXEC_LOG)
     G->LOG_CMDS.Put(G->Lang( 0x154, "sss", nick, cmd, args));
   return 0;
}

void DccChat::Activity (char * buf,int i) {
   FUNCTION;
   int nathan = 0, doron = 0, fixed = 0;
   char *v, *d;
   Socket * ptr = G->socklist;
   void * args[3];
   
   context;
   StripTelnet(buf, &i);
   if (buf[0] && (buf[0] != '.') && DetectDccFlood())
     return;
   TimeNow();
   if (buf[0]) {
      args[0] = buf;
      args[1] = this;
      G->bt_filt->Check(buf, 0, args);
      strcpy(buf, (char *)args[0]);
   }
   if (buf[0]) {
      /* check for beeps and cancel annoying ones */
      v = buf;
      d = buf;
      while (*v)
	 switch (*v) {
	 case 7:		/* beep - no more than 3 */
	    nathan++;
	    if (nathan > 3)
	      v++;
	    else
	      *d++ = *v++;
	    break;
	 case 8:		/* backspace - for lame telnet's :) */
	    if (d > buf) {
	       d--;
	    }
	    v++;
	    break;
	 case 27:		/* ESC - ansi code? */
	    if (*v++ != '\033')
	      doron = 1;
	    else if (*v++ != '[')
	      doron = 1;
	    else {
	       doron = 2;
	       while (*v) {
		  if (*v == 'm') {
		     doron = 0;
		     break;
		  }
		  if ((*v != ';') && ((*v < '0') || (*v > '9')))
		    break;
		  v++;
		  doron++;
	       }
	    }
	    /* if it's valid, append a return-to-normal code at the end */
	    if (!doron) {
	       *d++ = *v++;
	       fixed = 1;
	    } else
	      v += doron;
	    break;
	 case '\r':		/* weird pseudo-linefeed */
	    v++;
	    break;
	 default:
	    *d++ = *v++;
	 }
      if (fixed)
	strcpy(d, "\033[0m");
      else 
	*d = 0;
      if (buf[0]) {		/* nothing to say - maybe paginG->.. */
	 if (!strncasecmp(buf, "/me ", 4))
	   buf[0] = '.';
	 if (!strncasecmp(buf, "\001ACTION ", 8)) {
	    if (*(--d) == '\001')
	      *d = 0;
	    buf += 4;
	    strncpy(buf, ".me", 3);
	 }
	 if ((buf[0] == '.') || (channel < 0)) {
	    if (buf[0] == '.')
	      buf++;
	    v = G->Split(&buf);
	    if (CheckDcc(v,buf)) {
	       CheckChof(buf);
	       Send(G->Lang( 0x155, ""));
	       G->LOG_MISC.Put(G->Lang( 0x156, "ss", nick, host));
	       FlushLines();
	       if ((key != STDIN) || G->backgrd) 
		 Close();
	       else {
		  Send(G->Lang( 0x157, ""));
		  Chatter();
	       }
	       return;
	    }
	 } else if (buf[0] == ',') {
	    FlagRecord fr;
	    fixed = 0x158;
	    if ((buf[1] == 'm') && (buf[2] == 'e') && (buf[3] == ' ')) {
	       fixed = 0x29F;
	       buf += 2;
	    }
	    while ((ptr = ptr->Next()))
		if ((ptr->TypeFlags() & DCT_MASTER) 
		    && ((ptr->TypeFlags() & DCT_CHAT) ||
			(((DccChat *)ptr)->Channel() >= 0)) &&
		    ((ptr != this) || HasStatus(STAT_ECHO))
		    && fr.Set(ptr->User())->GlobMaster())
		ptr->Send(G->Lang( fixed, "ss", nick, buf + 1));
 	 } else if (buf[0] == '\'') {
	    fixed = 0;
	    if ((buf[1] == 'm') && (buf[2] == 'e') && 
		((buf[3] == ' ') || (buf[3] == '\'') || (buf[3] == ',')))
	      fixed = 1;
	    d = G->Lang( 0x159 + fixed, "ss", nick, buf + 1 + (2 * fixed));
	    while ((ptr = ptr->Next()))
	      if ((ptr->TypeFlags() & DCT_CHAT) && 
		  ((ptr != this) || HasStatus(STAT_ECHO)))
		ptr->Send(d);
	 } else {
	    if (Away())
	      SetAway(0);
	    if (HasStatus(STAT_ECHO))
	      Chanout(channel, G->Lang( 0x10F, "ss", nick, buf));
	    else
	      ChanoutBut(channel, G->Lang( 0x10F, "ss", nick, buf));
	    args[0] = nick;
	    args[1] = &channel;
	    args[2] = buf;
	    G->bt_chat->Check(buf, 0, args);
	 }
      }
   }
   if (HasStatus(STAT_PAGE))
       FlushLines();
}

void DccChat::Display2 (Socket * where) {
   FUNCTION;
   char buf[6];
   
   buf[0] = HasStatus(STAT_CHAT) ? 'C' : 'c';
   buf[1] = HasStatus(STAT_PARTY) ? 'P' : 'p';
   buf[2] = HasStatus(STAT_TELNET) ? 'T' : 't';
   buf[3] = HasStatus(STAT_ECHO) ? 'E' : 'e';
   buf[4] = HasStatus(STAT_PAGE) ? 'P' : 'p';
   buf[5]=0;
   where->Send(G->Lang( 0x15B, "ssd", nick, buf, channel));
}

/* show motd to dcc chatter */
void DccChat::ShowMotd (void) {
   FUNCTION;
   FILE *vv;
   char s[1024];
   FlagRecord fr ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   
   fr.Set(user,con_chan);
   if ((vv = fopen(G->motdfile, "r"))) {
      if (!is_file(G->motdfile)) {
	 fclose(vv);
	 Send(G->Lang( 0x134, ""));
      } else {
	 Send(G->Lang( 0x131, ""));
	 G->HelpSubst(NULL, NULL, 0, HS_ISDCC | 
		      ((strip_flags & STRIP_HILITE) ? 0 : HS_HILITE), 0);
	 while (!feof(vv)) {
	    fgets(s, 120, vv);
	    if (!feof(vv)) {
	       if (s[strlen(s) - 1] == '\n')
		 s[strlen(s) - 1] = 0;
	       if (!s[0])
		 strcpy(s, " ");
	       G->HelpSubst(s, Nick(), &fr, 0, G->botnetnick);
	       if (s[0])
		 Send(G->Lang( 0x132, "s", s));
	    }
	 }
	 fclose(vv);
	 Send(G->Lang( 0x133, ""));
      }
   }
}

/* print to all on this channel but one */
void DccChat::ChanoutBut(int chan, char * s) {
   FUNCTION;
   int l = strlen(s);
   DccChat * ptr = (DccChat *)G->socklist;
   
   s[l++] = '\n';
   s[l] = 0;
   while ((ptr = ptr->Next()))
     if ((ptr != this) && (ptr->Channel() == chan)) 
       ptr->Dump(s, l);
}

void DccChat::Chanout(int chan, char * s) {
   FUNCTION;
   int l = strlen(s);
   DccChat * ptr = (DccChat *)G->socklist;
   
   s[l++] = '\n';
   s[l] = 0;
   while ((ptr = ptr->Next()))
     if ((ptr->Channel() == chan)) 
       ptr->Dump(s, l);
}

void DccChat::Chatter (void) {
   FUNCTION;
   int i, j;
   DccChat * ptr;
   
   Send(G->Lang( 0x19E, ""));
   ShowMotd();
   Send(G->Lang( 0x19F, ""));
   Send(G->Lang( 0x1A0, ""));
   i = channel;
   channel = 234567;
   j = key;
   strcpy(con_chan,"***");
   CheckChon();
   if ((ptr = (DccChat *)G->socklist->Find(j,G->DCC_CHAT))) {
      /* tcl script may have taken control */
      if (!strcmp(ptr->con_chan,"***"))
	strcpy(ptr->con_chan,"*");
      if (i == -2)
	i = 0;
      if (ptr->channel == 234567) 
	ptr->channel = i;
      ptr->CheckLChjn();
   }
}

void DccChat::SetAway (const char * s) {
   FUNCTION;
   void * args[3];
   
   if (!s || !s[0]) {
      context;
      if (!away) {
	 Send(G->Lang( 0x1A1, ""));
	 return;
      } else {
	 if (channel >= 0) 
	   Chanout(channel,G->Lang( 0x10C, "s",G->Lang(0x11A, "s", nick)));
	 Send(G->Lang( 0x1A2, ""));
	 my_free(away,G->memtype_socket);
	 away = 0;
      }
   } else {
      SetAwayRaw(s);
      if (channel >= 0) 
	Chanout(channel, G->Lang( 0x10C, "s",G->Lang(0x119, "ss", nick, s)));
      Send(G->Lang( 0x1A3, ""));
   }
   args[0] = G->botnetnick;
   args[1] = &key;
   args[2] = (void *)s;
   G->bt_away->Check(G->botnetnick, 0, args);
}

int DccChat::DetectDccFlood ( void ) {
   FUNCTION;
   if (G->dcc_flood_thr) {
      if (timeval != G->now) {
	 TimeNow();
	 msgs_per_sec = 0;
      } else {
	 msgs_per_sec++;
	 if (msgs_per_sec > G->dcc_flood_thr) {
	    Send(G->Lang( 0x1A6, ""));
	    if (type_flags & DCT_CHAT) 
	      CheckChof(G->Lang( 0x184, "s", nick));
	    if ((key != STDIN) || G->backgrd) 
	      Close();
	    else {
	       Send(G->Lang( 0x157, ""));
	       Chatter();
	    }
	    return 1;		/* <- flood */
	 }
      }
   }
   return 0;
}

void DccChat::CheckChon ( void ) {
   FUNCTION;
   FlagRecord fr (user, FlagRecord::GLOBAL | FlagRecord::CHANNEL);
   void * argv = this;
   
   context;
   TouchLaston(user,"partyline",G->now);
   if ((channel >= 0) && (channel < MAX_CHANS)) 
     CheckLChjn();
   G->bt_chon->Check(nick, &fr, &argv);
}

void DccChat::CheckChof (const char * reason) {
   FUNCTION;
   FlagRecord fr (user, FlagRecord::GLOBAL | FlagRecord::CHANNEL);
   int fl = reason && reason[0];
   void * argv[2];
   
   context;
   TouchLaston(user,"partyline",G->now);
   argv[0] = this;
   if ((channel >= 0) && (channel < MAX_CHANS)) 
     CheckLChpt(reason);
   argv[1] = fl ? (void *)reason : "*";
   G->bt_chof->Check(nick, &fr, argv);
}

void DccChat::CheckLChjn ( void ) {
   FUNCTION;
   FlagRecord fr (user);
   char s[32];
   void * args = this;
   
   Chanout(channel,G->Lang(0x10C,"s",G->Lang(channel?0x113:0x112,"s",nick)));
   strcpy(s, G->ToBase10(channel));
   G->bt_lchjn->Check( s, &fr, &args);
}

void DccChat::CheckLChpt ( const char * reason ) {
   FUNCTION;
   FlagRecord fr (user);
   char s[32];
   void * args[2] = { this, (void *)reason};
   
   Chanout(channel, G->Lang( 0x10C, "s", G->Lang( channel ? 0x117 : 0x116,
						"ss", nick, reason)));
   strcpy(s, G->ToBase10(channel));
   G->bt_lchpt->Check ( s, &fr, args);
}

int DccChat::ExpMem ( void ) {
   FUNCTION;
   return DccChatPass::ExpMem () - sizeof(DccChatPass) + sizeof(DccChat);
}
