/*
 * DccChatPass.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "DccChatPass.h"
#include "DccChat.h"
#include "MessageQueue.h"
#include "UserRecord.h"
#include <ctype.h>
#define CLASS "DccChatPass"

DccChatPass::DccChatPass ( Socket & old, const char * type )
: Socket ( old, type, 0, &G->password_timeout ) {
   FUNCTION;
   away = 0;
   buffer = NEW( MessageQueue () );
   strip_flags = 0;
   current_lines = 0;
   line_count = 0;
   max_line = 0;
}

DccChatPass::DccChatPass ( int fd, const char * type, int nt )
: Socket ( FD_, type, nt, 0, fd, G->Lang( 0x1F2, ""), 0) {
   FUNCTION;
   away = 0;
   buffer = NEW( MessageQueue () );
   strip_flags = STRIP_ALL;
   current_lines = 0;
   line_count = 0;
   max_line = 0;
   TimeNow();
   status = STAT_ECHO | STAT_TELNET;
}

DccChatPass::DccChatPass ( struct in_addr in, int pt ) 
: Socket ( G->DCC_CHAT_PASS, CONNECTING, in, pt, 0, &G->password_timeout ) {
   FUNCTION;
   away = 0;
   buffer = NEW( MessageQueue () );
   strip_flags = 0;
   current_lines = 0;
   line_count = 0;
   max_line = 0;
   TimeNow();
   status = STAT_ECHO;
}

DccChatPass::DccChatPass ( DccChatPass & old ) 
: Socket ( old, G->DCC_CHAT_PASS, 0, &G->password_timeout ) {
   FUNCTION;
   away = old.away;
   old.away = 0;
   buffer = old.buffer;
   old.buffer = 0;
   strip_flags = old.strip_flags;
   current_lines = old.current_lines;
   line_count = old.line_count;
   max_line = old.max_line;
}

DccChatPass::~DccChatPass ( void ) {
   FUNCTION;
   DELETE ( buffer );
   if (away) 
     my_free(away,G->memtype_socket);
}

void DccChatPass::Activity (char * buf,int len) {
   FUNCTION;
   FlagRecord atr (user);
   DccChat * newone;
      
   if (!len)
     return;
   StripTelnet(buf, &len);
   if (HasStatus(STAT_TELNET))
     Printf("\377\374\001\n");	/* turn echo back on */
   if (user->PassMatch(buf)) {
      context;
      SetAwayRaw(0);
      newone = NEW ( DccChat ( this ) );
      newone->ClearStatus(STAT_CHAT);
      newone->ConsoleFlags(atr.GlobMaster() ? G->conmask.Value() : 0);
      if (HasStatus(STAT_TELNET))
	Printf("\377\374\001\n");	/* turn echo back on */
      newone->Chatter();
   } else {
      Send(G->Lang( 0x14C, ""));
      G->LOG_MISC.Put(G->Lang( 0x14D, "ssd", nick, host, remoteport));
      if (away) {	/* su from a dumb user */
	 SetUser(G->userlist->Find(away));
	 SetNick(away);
	 SetAwayRaw(0);
	 newone = NEW ( DccChat ( this ) );
	 newone->CheckChon();
      } else 
	Close();
   }
}

void DccChatPass::Eof(void) {
   FUNCTION;
   G->LOG_MISC.Put(G->Lang( 0x14E, "ssd", nick, host, remoteport));
   Close();
}

void DccChatPass::Timeout (void) {
   FUNCTION;
   DccChat * newone;
   
   Send(G->Lang( 0x109, ""));
   G->LOG_MISC.Put(G->Lang( 0x14F, "ssd", nick, host, remoteport));
   if (away) {
      SetUser(G->userlist->Find(away));
      SetNick(away);
      SetAwayRaw(0);
      newone = NEW ( DccChat ( this ) );
      newone->CheckChon();
   } else if ((key != STDIN) || G->backgrd)
	Close();
   else {
      Printf("\n%s\377\373\001\n", G->Lang( 0x10A, ""));
      TimeNow();
   }
}

void DccChatPass::Display2 (Socket * where) {
   FUNCTION;
   where->Send(G->Lang( 0x150, "sd", nick,G->now - timeval));
}

int DccChatPass::ExpMem (void) {
   FUNCTION;
   return sizeof(DccChatPass) + (away ? strlen(away) + 1 : 0) + buffer->ExpMem();
}

int DccChatPass::ProcessCodes (int flags, int status, char * text, char * desto) {
   FUNCTION;
   int len = 0, codes = 0;
   char * toadd, tbuf[12];
   
   while (*text && (len < 511)) {
      toadd = 0;
      if (*text == 033) {
	 text++;
	 if ((*text != 2) && (*text != 0x16) && (*text != 0x1F)
	     && (*text != 033)) {
	    if ((*text == '0') || (*text == 'A')) {
	       codes &= ~STRIP_COLOR;
	       toadd = "\033[0m";
	       text++;
	    } else if ((*text >= '1') && (*text <= '9')) { 
	       if (status & STAT_TELNET) {
		  G->SPrintF(tbuf, 11, "\033[3%cm", (*text) - 1);
		  toadd = tbuf;
		  codes |= STRIP_COLOR;
	       }
	       text++;
	    } else if ((*text >= 'B') && (*text <= 'I')) {
	       if (status & STAT_TELNET) {
		  G->SPrintF(tbuf, 11, "\033[4%cm", (*text) - 'B' + '0');
		  toadd = tbuf;
		  codes |= STRIP_COLOR;
	       }
	       text++;
	    }
	    if (toadd && (flags & STRIP_COLOR)) 
	      toadd = NULL;
	    else if (!toadd && (flags & STRIP_ANSI)) {
	       if (*text == '[') {
		  text++;
		  while ((*text == ';') || isdigit(*text)) 
		    text++;
		  if (*text)
		    text++;		/* also kill the following char */
	       }
	       continue;
	    }
	 }
      }
      if (!toadd) 
	switch (*text) {
	  case 2:			/* Bold text */
	   if (flags & STRIP_BOLD) {
	      text++;
	      continue;
	   } else if (status & STAT_TELNET) {
	      text++;
	      if (codes & STRIP_BOLD) {
		 codes &= ~STRIP_BOLD;
		 toadd = "\033[0m";
	      } else {
		 codes |= STRIP_BOLD;
		 toadd = "\033[1m";
	      }
	   }
	   break;
	  case 3:			/* mIRC colors? */
	   if (flags & STRIP_COLOR) {
	      if (isdigit(text[1])) {	/* Is the first char a number? */
		 text += 2;	/* Skip over the ^C and the first digit */
		 if (isdigit(*text))
		   text++;		/* Is this a double digit number? */
		 if (*text == ',') {	/* Do we have a background color next? */
		    if (isdigit(text[1]))
		      text += 2;	/* Skip over the first background digit */
		    if (isdigit(*text))
		      text++;	/* Is it a double digit? */
		 }
	      } else
		text++;
	      continue;
	   } 
	   break;
	  case 7:
	   if (flags & STRIP_BELLS) {
	      text++;
	      continue;
	   }
	   break;
	  case 0x16:		/* Reverse video */
	   if (flags & STRIP_REV) {
	      text++;
	      continue;
	   } else if (status & STAT_TELNET) {
	      text++;
	      if (codes & STRIP_REV) {
		 codes &= ~STRIP_REV;
		 toadd = "\033[0m";
	      } else {
		 codes |= STRIP_REV;
		 toadd = "\033[7m";
	      }
	   }
	   break;
	  case 0x1f:		/* Underlined text */
	   if (flags & STRIP_UNDER) {
	      text++;
	      continue;
	   } else if (status & STAT_TELNET) {
	      text++;
	      if (codes & STRIP_REV) {
		 codes &= ~STRIP_REV;
		 toadd = "\033[0m";
	      } else {
		 codes |= STRIP_REV;
		 toadd = "\033[4m";
	      }
	   }
	   break;
	  case 033:
	   text++;
	   if (status & STAT_TELNET) {
	      if (codes & STRIP_REV) {
		 codes &= ~STRIP_REV;
		 toadd = "\033[0m";
	      } else {
		 codes |= STRIP_REV;
		 toadd = "\033[7m";
	      }
	   } else 
	     toadd = "\002\026";
	   break;
	  case '\n':
	   if (status & STAT_TELNET) {
	      if (codes) 
		toadd = "\033[0m\r\n";
	      else
		toadd = "\r\n";
	      text++;
	   } else if (codes) {
	      toadd = "\033[0m\n";
	      text++;
	   }
	}
      if (toadd) {
	 while (*toadd && (len < 511))
	   desto[len++] = *toadd++;
      } else
	desto[len++] = *text++;		/* Move on to the next char */
   }
   desto[len] = 0;
   return len;
}

int DccChatPass::Dump(char * buf, int l) {
   FUNCTION;
   char line[512];
   
   l = ProcessCodes(strip_flags,status,buf,line);
   if (!HasStatus(STAT_PAGE))
     l = Write(line, l);
   else if (current_lines > G->max_sendq) {
      DELETE ( buffer );
      buffer = NEW ( MessageQueue ( ) ); /* safety */
      ClearStatus(STAT_PAGE);
      DoBoot(G->botnetnick, G->Lang( 0x151, ""));
      l = -1;
   } else if (line_count < max_line) {
      line_count++;
      l = Write(line, l);
   } else {
      current_lines++;
      NEW ( MessageQueue ( *buffer, line, l ) );
   }
   return l;
}

void DccChatPass::FlushLines ( void ) {
   FUNCTION;
   char * s;
   
   if (buffer && (buffer->Flush(max_line,this) == max_line)) {
      s = G->Lang(HasStatus(STAT_TELNET) ? 0x1A4 : 0x1A5, "");
      Write(s, strlen(s));
   }
   line_count = 0;
}

/* handle someone being booted from dcc chat */
void DccChatPass::DoBoot (const char * by, const char * reason) {
   FUNCTION;
   int files = (type_flags & DCT_CHAT);
   
   Send(G->Lang( 0x181, ""));
   Send(G->Lang( files ? 0x182 : 0x183, "ss", by, reason[0] ? reason : by));
   if (files) {
      DccChat * dcc = (DccChat *)this;
      dcc->CheckChof(G->Lang( 0x1A7, "ss", by, reason[0] ? reason : by));
      if ((key != STDIN) || G->backgrd)
	Close();
      else {
	 Send(G->Lang( 0x157, ""));
	 dcc->Chatter();
      }
   } else
     Close();
}
