/*
 * FlagRecord.c
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Socket.h"
#include "UserRecord.h"
#include "FlagRecord.h"
#include "ChannelRecord.h"
#include "Channel.h"
#include <ctype.h>
#define CLASS "FlagRecord"

FlagRecord & FlagRecord::operator += ( FlagRecord & pls ) {
   FUNCTION;
   if (match & GLOBAL) {
      global |= pls.global;
      udef_global |= pls.udef_global;
   }
   if (match & CHANNEL) {
      chan |= pls.chan;
      udef_chan |= pls.udef_chan;
   }
   if (match & BOT) 
     bot |= pls.bot;
   return *this;
}

FlagRecord & FlagRecord::operator -= ( FlagRecord & mns ) {
   FUNCTION;
   if (match & GLOBAL) {
      global &= ~mns.global;
      udef_global &= ~mns.udef_global;
   }
   if (match & CHANNEL) {
      chan &= ~mns.chan;
      udef_chan &= ~mns.udef_chan;
   }
   if (match & BOT) 
     bot &= ~mns.bot;
   return *this;
}

void FlagRecord::BreakDown (const char * string, FlagRecord * minus) {
   FUNCTION;
   FlagRecord * which = this;
   int mode = 0; /* 0 = glob, 1 = chan, 2 = bot */
   int flags = match;
   
   if (!(flags & GLOBAL)) {
      if (flags & BOT)
	mode = 2;
      else if (flags & CHANNEL)
	mode = 1;
      else
	return; /* we dont actually want any..huh?*/
   }
   match = udef_chan = chan = global = udef_global = bot = 0;
   if (minus) 
     minus->match = minus->udef_chan = minus->chan = minus->global =
     minus->udef_global = minus->bot = 0;
   match = OR; /* befault binding type OR */
   
   while (*string) {
      switch (*string) {
       case '+':
	 which = this;
	 break;
       case '-':
	 which = minus ? minus : this;
	 break;
       case '|':
       case '&':
	 if (!mode) 
	   match = (*string == '|') ? OR : AND;
	 which = this;
	 mode++;
	 if ((mode == 2) && !(flags & (CHANNEL | BOT)))
	   string = "";
	 else if (mode == 3) 
	   mode = 1;
	 break;
       default:
	 if ((*string >= 'a') && (*string <= 'z')) {
	    switch (mode) {
	     case 0:
	       which->global |= 1 << (*string - 'a');
	       break;
	     case 1:
	       which->chan |= 1 << (*string - 'a');
	      break;
	     case 2:
	       which->bot |= 1 << (*string - 'a');
	    }
	 } else if ((*string >= 'A') && (*string <= 'Z')) {
	    switch (mode) {
	     case 0:
	       which->udef_global |= 1 << (*string - 'A');
	       break;
	     case 1:
	       which->udef_chan |= 1 << (*string - 'A');
	      break;
	    }
	 } else if ((*string >= '0') && (*string <= '9')) {
	    switch (mode) {
	       /* map 0->9 to A->K for glob/chan so they are not lost */
	     case 0:
	       which->udef_global |= 1 << (*string - '0');
	       break;
	     case 1:
	       which->udef_chan |= 1 << (*string - '0');
	      break;
	     case 2:
	       which->bot |= BOT_FLAG0 << (*string - '0');
	       break;
	    }
	 }
      }
      string++;
   }
   for (which = this; which; which = (which == this ? minus : 0)) {
      which->global &= USER_VALID;
      which->udef_global &= 0x03ffffff;
      which->chan &= CHAN_VALID;
      which->udef_chan &= 0x03ffffff;
      which->bot &= BOT_VALID;
   }
   match |= flags;
   if (minus) {
      minus->match |= flags;
      if (!(match & (AND | OR)))
	match |= OR;
   }
}	 

static int flag2str(char * string, int bot, int udef) {
   FUNCTION;
   char x = 'a', *old = string;
   while (bot && (x <= 'z')) {
      if (bot & 1) 
	*string ++ = x;
      x++;
      bot = bot >> 1;
   }
   x = 'A';
   while (udef && (x <= 'Z')) {
      if (udef & 1) 
	*string ++ = x;
      udef = udef >> 1;
      x++;
   }  
   if (string == old)
     *string++ = '-';
   return string - old;
}

static int bot2str(char * string, int bot) {
   FUNCTION;
   char x = 'a',*old = string;
   while (x < 'v') {
      if (bot & 1)
	*string ++ = x;
      x++;
      bot >>= 1;
   }
   x = '0';
   while (x <= '9') {
      if (bot & 1)
	*string ++ = x;
      x++;
      bot >>= 1;
   }
   return string -old;
}

void FlagRecord::Build (char * string, FlagRecord * minus) {
   FUNCTION;
   char * old = string;
   if (match & GLOBAL) {
      if (minus && (global || udef_global)) 
	*string ++ = '+';
      string += flag2str(string, global, udef_global);
      if (minus && (minus->global || minus->udef_global)) {
	 *string ++ = '-';
	 string += flag2str(string, minus->global, minus->udef_global);
      }
   } else if (match & BOT) {
      if (minus && bot)
	 *string ++ = '+';
      string += bot2str(string,bot);
      if (minus && minus->bot) {
	 *string ++ = '-';
	 string += bot2str(string,minus->bot);
      }
   }
   if (match & CHANNEL) {
      if (match & (GLOBAL | BOT)) 
	*string ++ = (match & AND) ? '&' : '|';
      if (minus && (chan || udef_chan))
	*string ++ = '+';
      string += flag2str(string, chan, udef_chan);
      if (minus && (minus->chan || minus->udef_chan)) {
	 *string ++ = '-';
	 string += flag2str(string, minus->global, minus->udef_chan);
      }
   } 
   if ((match & (BOT | CHANNEL)) == (BOT | CHANNEL)) {
      *string ++ = (match & AND) ? '&' : '|';
      if (minus && bot)
	 *string ++ = '+';
      string += bot2str(string,bot);
      if (minus && minus->bot) {
	 *string ++ = '-';
	 string += bot2str(string,minus->bot);
      }
   }
   if (string == old) {
      *string ++ = '-';
      *string = 0;
   }
   *string = 0;
}

int FlagRecord::Ok (FlagRecord & have) {
   FUNCTION;
   if (match & AND) 
     return *this == have;
   else if (match & OR) {
      int hav = have.global;
      
      /* exception 1 - global +d/+k cant use -|-, unless they are +p */
      if (!chan && !global && !udef_global && !udef_chan) {
	 if (!List::G->allow_dk_cmds) {
	    if (have.GlobParty())
	      return 1;
	    if (have.AnyKick())
	      return 0; /* +k cant use -|- commands */
	    if (have.AnyValidDeop())
	      return 0; /* neither can +d's */
	 }
	 return 1;
      }
      /* the +n/+m checks arent needed anymore since +n/+m
       * automatically add lower flags */
      if (!List::G->require_p && (have.GlobOp() || have.ChanOwner()))
	hav |= USER_PARTY;
      if (hav & global)
	return 1;
      if (have.chan & chan)
	return 1;
      if (have.udef_global & udef_global)
	return 1;
      if (have.udef_chan & udef_chan)
	return 1;
      return 0;
   }
   return 0; /* fr0k3 binding, dont pass it */
}

int FlagRecord::operator == (FlagRecord & have) {
   FUNCTION;
   if (match & AND) {
      if (match & GLOBAL) {
	 if ((global & have.global) != global)
	   return 0;
	 if ((udef_global & have.udef_global) != udef_global)
	   return 0;
      }
      if (match & BOT) 
	if ((bot & have.bot) != bot)
	  return 0;
      if (match & CHANNEL) {
	 if ((chan & have.chan) != chan)
	   return 0;
	 if ((udef_chan & have.udef_chan) != udef_chan)
	   return 0;
      }
      return 1;
   } else if (match & OR) {
      if (!chan && !global && !udef_chan && !udef_global && !bot)
	return 1;
      if (match & GLOBAL) {
	 if (have.global & global)
	   return 1;
	 if (have.udef_global & udef_global)
	   return 1;
      }
      if (match & BOT) 
	if (have.bot & bot)
	  return 1;
      if (match & CHANNEL) {
	 if (have.chan & chan)
	   return 1;
	 if (have.udef_chan & udef_chan)
	   return 1;
      }
      return 0;
   }
   return 0; /* fr0k3 binding, dont pass it */
}

/* some flags are mutually exclusive -- this roots them out */
void FlagRecord::Sanity(void) {
   FUNCTION;
   if (match & GLOBAL) {
      if ((global & USER_BOT) 
	  && (global & (USER_PARTY | USER_MASTER | USER_COMMON | USER_OWNER)))
	global &= ~(USER_PARTY | USER_MASTER | USER_COMMON | USER_OWNER);
      if ((global & (USER_OP | USER_HELPER)) && (global & USER_DEOP))
	global &= ~(USER_OP | USER_HELPER | USER_DEOP);
      if (global & USER_OP)
	global |= USER_HELPER;
      if ((global & USER_VOICE) && (global & USER_QUIET))
	global &= ~(USER_VOICE | USER_QUIET);
      /* can't be owner without also being master */
      if (global & USER_OWNER)
	global |= USER_MASTER;
      /* master implies botmaster, op & friend */
      if (global & USER_MASTER)
	global |= USER_BOTMAST|USER_OP|USER_FRIEND|USER_JANITOR;
      /* can't be botnet master without party-line access */
      if (global & USER_BOTMAST)
	global |= USER_PARTY;
      /* janitors can use the file area */
      if (global & USER_JANITOR)
	global |= USER_XFER;
   }
   if (match & CHANNEL) {
      if ((chan & (USER_OP | USER_HELPER)) && (chan & USER_DEOP))
	chan &= ~(USER_OP | USER_HELPER | USER_DEOP);
      if (chan & USER_OP)
	chan |= USER_HELPER;
      if ((chan & USER_VOICE) && (chan & USER_QUIET))
	chan &= ~(USER_VOICE | USER_QUIET);
      /* can't be channel owner without also being channel master */
      if (chan & USER_OWNER)
	chan |= USER_MASTER;
      /* master implies friend & op */
      if (chan & USER_MASTER)
	chan |= USER_OP|USER_FRIEND;
      /* can't be +s on chan unless you're a bot */
      if (!(global & USER_BOT))
	chan &= ~BOT_AGGRESSIVE;
   }
   if (match & BOT) {
      bot = (bot & BOT_VALID);
      if ((bot & BOT_HUB) && (bot & BOT_ALT))
	bot &= ~BOT_ALT;
      if (bot & BOT_REJECT) {
	 if (bot & (BOT_AGGRESSIVE | BOT_PASSIVE))
	   bot &= ~(BOT_AGGRESSIVE | BOT_PASSIVE | BOT_REJECT);
	 if (bot & BOT_HUB)
	   bot &= ~(BOT_HUB | BOT_REJECT);
	 if (bot & BOT_ALT)
	   bot &= ~(BOT_ALT | BOT_REJECT);
	 if (bot & BOT_CROSS)
	   bot &= ~(BOT_CROSS | BOT_REJECT);
      }
      if ((bot & BOT_CROSS) && (bot & (BOT_AGGRESSIVE | BOT_PASSIVE)))
	bot &= ~(BOT_CROSS | BOT_PASSIVE | BOT_AGGRESSIVE);
      if (!(bot & (BOT_AGGRESSIVE | BOT_PASSIVE)))
	bot &= ~(BOT_GLOBAL|BOT_SHAREFLAGS);
   }
}

void FlagRecord::ReaffirmOwners(void) {
   FUNCTION;
   char *p, *q, s[121];
   UserRecord * u;
   FlagRecord goober;
   
   /* make sure default owners are +n */
   if (List::G->owner[0]) {
      q = List::G->owner;
      p = strchr(q, ',');
      while (p) {
	 strncpy(s,q,p-q);
	 s[p-q] = 0;
	 List::G->RMSpace(s);
	 u = List::G->userlist->Find(s);
	 if (u) {
	    goober.Set(u);
	    goober.SetGlobOwner();
	    goober.Sanity();
	    u->Set(goober);
	 }
	 q = p+1;
	 p = strchr(q,',');
      }
      strcpy(s,q);
      List::G->RMSpace(s);
      u = List::G->userlist->Find(s);
      if (u) {
	 goober.Set(u);
	 goober.SetGlobOwner();
	 goober.Sanity();
	 u->Set(goober);
      }
   }
}

FlagRecord * FlagRecord::Get (UserRecord * u, const char * chname) {
   FUNCTION;
   ChannelRecord * cr;
   int oldflags = match;
   char buffer[100];
   Channel * ch;
   
   if (oldflags & GLOBAL) {
      u->flags = global;
      u->flags_udef = udef_global;
      if (!List::G->noshare && List::G->shareout && !(u->flags & USER_UNSHARED)) {
	 match = GLOBAL;
	 Build(buffer);
	 List::G->shareout(NULL,"a%s %s\n",u->Key(),buffer);
      }
   } 
   if ((oldflags & BOT) && (u->flags & USER_BOT)) {
      UserEntryBotfl * ueb = (UserEntryBotfl *)u->GetEntry(List::G->entrytype_botfl);
      ueb->Set(bot);
   }
   /* dont share bot attrs */
   if ((oldflags & CHANNEL) && chname) {
      for (cr = u->chanrec;cr;cr = cr->Next())
	if (!strcasecmp(chname,cr->Key()))
	    break;
      ch = List::G->chanset->Find(chname);
      if (!cr && ch) 
	cr = NEW ( ChannelRecord ( u, chname ) );
      if (cr && ch) {
	 cr->flags = chan;
	 cr->flags_udef = udef_chan;
	 if (!List::G->noshare && List::G->shareout && 
	     !(u->flags & USER_UNSHARED)) {
	    match = CHANNEL;
	    Build(buffer);
	    List::G->shareout(ch,"a%s %s %s\n",u->Key(),buffer,chname);
	 }
      }
   }
   match = oldflags;
   return this;
}

FlagRecord * FlagRecord::Set (UserRecord * u, const char * chname) {
   FUNCTION;
   ChannelRecord * cr;

   if (!u) {
      global = udef_global = chan = udef_chan = bot = 0;
      return this;
   }
   if (match & GLOBAL) {
      global = u->flags;
      udef_global = u->flags_udef;
   } else 
     global = udef_global = 0;
   if (match & BOT) {
      UserEntryBotfl * ueb = (UserEntryBotfl *)u->FindEntry(List::G->entrytype_botfl);
      bot = ueb ? ueb->Flags() : 0;
   } else
     bot = 0;
   if (match & CHANNEL) {
      if (match & ANYWHERE) {
	 chan = u->flags;
	 udef_chan = u->flags_udef;
	 for (cr = u->chanrec; cr; cr = cr->Next())
	   if (List::G->chanset->Find(cr->Key())) {
	      chan |= cr->flags;
	      udef_chan |= cr->flags_udef;
	   }
      } else {
	 if (chname)
	   for (cr = u->chanrec; cr; cr = cr->Next())
	     if (!strcasecmp(chname,cr->Key()))
	       break;
	 if (chname && cr) {
	    chan = cr->flags;
	    udef_chan = cr->flags_udef;
	 } else 
	   chan = udef_chan =0;
      }
   }
   return this;
}
