/*
 * FlagRecord.h - class definition
 */

#ifndef _FLAGRECORD_
#define _FLAGRECORD_

#ifndef _MAIN_
#include "main.h"
#endif

class FlagRecord {
   int match;
   int global;
   int udef_global;
   int chan;
   int udef_chan;
   int bot;
   friend class UserRecord;
   
/* userflags:                              */
/*   ???????? x?vut??q ponm?kj? h?f?dcba   */
/* + user defined A-Z                      */
/*   unused letters: eglrwyz               */
/* botflags soon to be:                    */
/*   ?9876543 210??sr? p???l??i h?f??c?a   */
/*   unused letters: cdefgjkmnoqtuvwxyz    */
/* chanflags:                              */
/*   ???????? ??vu?s?q ?onm?k?? h?f?d??a   */
/* + user defined A-Z                      */
/*   unused letters: bceghjlprtwxyz        */
   enum {
      USER_CHANMASK = 0x00017569, /* common channel & global flags */
	USER_VALID  = 0x00B9f6Af, /* valid user flags */
	CHAN_VALID  = 0x003574A9, /* valid channel flags */
	BOT_VALID   = 0x7fe689A5, /* valid bot flags */
	
	/* user flags */
        USER_AUTOOP   = 0x00000001,  /* a   auto-op - future */
        USER_BOT      = 0x00000002,  /* b   user is a bot */
        USER_COMMON   = 0x00000004,  /* c   user is actually a public irc site */
        USER_DEOP     = 0x00000008,  /* d   user is global de-op */
        USER_FRIEND   = 0x00000020,  /* f   user is global friend*/
        USER_HELPER   = 0x00000080,  /* h   user is global helper */
        USER_JANITOR  = 0x00000200,  /* j   user has file area master */
        USER_KICK     = 0x00000400,  /* k   user is global auto-kick */
        USER_MASTER   = 0x00001000,  /* m   user has full bot access */
        USER_OWNER    = 0x00002000,  /* n   user is the bot owner */
        USER_OP       = 0x00004000,  /* o   user is +o on all channels */
        USER_PARTY    = 0x00008000,  /* p   user has party line access */
        USER_QUIET    = 0x00010000,  /* q   never let 'em get a voice */
        USER_BOTMAST  = 0x00080000,  /* t   user is botnet master */
        USER_UNSHARED = 0x00100000,  /* u   not shared with sharebots */
        USER_VOICE    = 0x00200000,  /* v   auto-voice on join */
        USER_XFER     = 0x00800000,  /* x   user has file area access */

	/* flags specifically for bots */
        BOT_ALT        = 0x00000001,  /* a   auto-link here if all +h's fail */
        BOT_BOT        = 0x00000002,  /* b   sanity bot flag */
	BOT_CROSS      = 0x00000004,  /* c   crosslink to this bot */
        BOT_SHAREFLAGS = 0x00000020,  /* f   share global flags */
        BOT_GLOBAL     = 0x00000040,  /* g   share all channels */
        BOT_HUB        = 0x00000080,  /* h   auto-link to ONE of these bots */
        BOT_ISOLATE    = 0x00000100,  /* i   isolate the partyline to this bot */
        BOT_LEAF       = 0x00000800,  /* l   may not link other bots */
        BOT_PASSIVE    = 0x00008000,  /* p   share passively with this bot */
        BOT_REJECT     = 0x00020000,  /* r   automatically reject anywhere */
        BOT_AGGRESSIVE = 0x00040000,  /* s   bot shares user files */
        BOT_FLAG0      = 0x00200000,  /* 0   user-defined flag #0 */
        BOT_FLAG1      = 0x00400000,  /* 1   user-defined flag #1 */
        BOT_FLAG2      = 0x00800000,  /* 2   user-defined flag #2 */
        BOT_FLAG3      = 0x01000000,  /* 3   user-defined flag #3 */
        BOT_FLAG4      = 0x02000000,  /* 4   user-defined flag #4 */
        BOT_FLAG5      = 0x04000000,  /* 5   user-defined flag #5 */
        BOT_FLAG6      = 0x08000000,  /* 6   user-defined flag #6 */
        BOT_FLAG7      = 0x10000000,  /* 7   user-defined flag #7 */
        BOT_FLAG8      = 0x20000000,  /* 8   user-defined flag #8 */
        BOT_FLAG9      = 0x40000000,  /* 9   user-defined flag #9 */
   };
  public:
   enum {
      GLOBAL      = 0x00000001,
        BOT       = 0x00000002,
        CHANNEL   = 0x00000004,
        OR        = 0x40000000,
        AND       = 0x20000000,
        ANYWHERE  = 0x10000000,
        ALL       = 0x0fffffff,
   };
   inline FlagRecord (int m = GLOBAL) {
      match = m;
      Clear();
   }
   inline FlagRecord ( UserRecord * u, int m = GLOBAL, const char * c = 0) {
      match = m;
      Set (u, c);
   }
   
   inline int AnyFlags (void) { return global | udef_global | chan | udef_chan
	| bot; }
   inline void operator ^= (FlagRecord & other) {
      match = other.match ^ (AND | OR);
   }
   
   inline int ChanOp(void) { return chan & USER_OP; }
   inline int ChanHelper(void) { return chan & USER_HELPER; }
   inline int ChanDeop(void) { return chan & USER_DEOP; }
   inline int ChanMaster(void) { return chan & USER_MASTER; }
   inline int ChanOwner(void) { return chan & USER_OWNER; }
   inline int ChanAutoop(void) { return chan & USER_AUTOOP; }
   inline int ChanKick(void) { return chan & USER_KICK; }
   inline int ChanVoice(void) { return chan & USER_VOICE; }
   inline int ChanQuiet(void) { return chan & USER_QUIET; }
   inline int ChanFriend(void) { return chan & USER_FRIEND; }
   inline int ChanUdef ( int x ) { return udef_chan & x; } 
   inline int ChanShare ( void ) { return chan & BOT_AGGRESSIVE; }
   inline int ChanBot(void) { return (global & USER_BOT) && AnyFriend(); }
   
   inline int SetChanShare ( void ) { return chan |= BOT_AGGRESSIVE; }
   inline int SetChanDeop ( void ) { return chan |= USER_DEOP; }
   inline void SetChan ( int x ) { chan = x; }
   
   inline int ClearChanShare ( void ) { return chan &= ~BOT_AGGRESSIVE; }
   inline int ClearChanOp ( void ) { return chan &= ~USER_OP; }
   inline void ClearChan ( void ) { chan = 0; udef_chan = 0; }
   
   inline void ChanMaskShare ( void ) { chan &= ~BOT_AGGRESSIVE; }
   inline void ChanMaskOwner ( void ) { chan &= ~USER_OWNER; }
   inline void ChanMaskMaster ( void ) { chan &= ~USER_MASTER; }

   inline int Chan ( void ) { return chan; }
   inline int ChanUdef ( void ) { return udef_chan; }
   
   inline int ChanCheckables ( void ) { 
      return chan & (USER_OWNER | USER_MASTER | USER_BOTMAST | USER_OP);
   }
   
   inline int GlobOp(void) { return global & USER_OP; }
   inline int GlobDeop(void) { return global & USER_DEOP; }
   inline int GlobMaster(void) { return global & USER_MASTER; }
   inline int GlobBot(void) { return global & USER_BOT; }
   inline int GlobOwner(void) { return global & USER_OWNER; }
   inline int GlobAutoop(void) { return global & USER_AUTOOP; }
   inline int GlobKick(void) { return global & USER_KICK; }
   inline int GlobVoice(void) { return global & USER_VOICE; }
   inline int GlobQuiet(void) { return global & USER_QUIET; }
   inline int GlobFriend(void) { return global & USER_FRIEND; }
   inline int GlobBotmast(void) { return global & USER_BOTMAST; }
   inline int GlobParty(void) { return global & USER_PARTY; }
   inline int GlobUnshared(void) { return global & USER_UNSHARED; }
   inline int GlobXfer(void) { return global & USER_XFER; }
   inline int GlobHelper(void) { return global & USER_HELPER; }
   inline int GlobUdef ( int x ) { return udef_chan & x; }
   inline int GlobJanitor ( void ) { return global & USER_JANITOR; }
   inline int GlobCommon ( void ) { return global & USER_COMMON; }
   
   inline int SetGlobBot(void) { return global |= USER_BOT; }
   inline void SetGlobOwner ( void ) { global |= USER_OWNER; }
   inline void SetGlobParty ( void ) { global |= USER_PARTY; }
   inline void SetGlobal ( int x ) { global = x; }
   
   inline int ClearGlobBot(void) { return global &= ~USER_BOT; }
   inline void ClearGlobal ( void ) { global = 0; udef_global = 0; }
   
   inline void MaskButBot ( void ) { global &= USER_BOT; udef_global = 0; }
   inline void MaskBot ( void ) { global &= ~USER_BOT; }
   inline void MaskNonOwner ( void ) { global &=
	~(USER_OWNER|USER_MASTER|USER_BOTMAST|USER_UNSHARED);
   }
   inline void MaskNonMaster ( void ) { global &= (USER_PARTY|USER_XFER); }
   
   inline int Global ( void ) { return global; }
   inline int Checkables ( void ) { 
      return global & (USER_OWNER | USER_MASTER | USER_BOTMAST | USER_OP);
   }
      
   inline int AnyOp(void) { return (chan | global) & USER_OP; }
   inline int AnyDeop(void) { return (chan | global) & USER_DEOP; }
   inline int AnyMaster(void) { return (chan | global) & USER_MASTER; }
   inline int AnyOwner(void) { return (chan | global) & USER_OWNER; }
   inline int AnyAutoop(void) { return (chan | global) & USER_AUTOOP; }
   inline int AnyKick(void) { return (chan | global) & USER_KICK; }
   inline int AnyVoice(void) { return (chan | global) & USER_VOICE; }
   inline int AnyQuiet(void) { return (chan | global) & USER_QUIET; }
   inline int AnyFriend(void) { return (chan | global) & USER_FRIEND; }
   inline int AnyUdef ( int x ) { return (udef_chan | udef_global) & x; }
   inline int AnyValidOp (void) { return (chan & USER_OP) ||
	((global & USER_OP) && !(chan & USER_DEOP)); }
   inline int AnyValidDeop (void) { return (chan & USER_DEOP) ||
	((global & USER_DEOP) && !(chan & USER_OP)); }
   inline int AnyValidVoice (void) { return (chan & USER_VOICE) ||
	((global & USER_VOICE) && !(chan & USER_QUIET)); }
   inline int AnyValidQuiet (void) { return (chan & USER_QUIET) ||
	((global & USER_QUIET) && !(chan & USER_VOICE)); }
   inline int AnyValidHelper(void) { return (chan & USER_HELPER) ||
	((global & USER_HELPER) && !(chan & USER_DEOP)); }
   
   inline int BotAlt ( void ) { return bot & BOT_ALT; }
   inline int BotBot ( void ) { return bot & BOT_BOT; }
   inline int BotCross ( void ) { return bot & BOT_CROSS; }
   inline int BotShareflags ( void ) { return bot & BOT_SHAREFLAGS; }
   inline int BotGlobal ( void ) { return bot & BOT_GLOBAL; }
   inline int BotHub ( void ) { return bot & BOT_HUB; }
   inline int BotIsolate ( void ) { return bot & BOT_ISOLATE; }
   inline int BotLeaf ( void ) { return bot & BOT_LEAF; }
   inline int BotPassive ( void ) { return bot & BOT_PASSIVE; }
   inline int BotReject ( void ) { return bot & BOT_REJECT; }
   inline int BotAggressive ( void ) { return bot & BOT_AGGRESSIVE; }
   inline int BotShare ( void ) { return bot & (BOT_PASSIVE | BOT_AGGRESSIVE); }
   inline int BotUdef ( int x ) { return bot & x; }
   inline int Bot ( void ) { return bot; }

   inline void SetBot ( int x ) { bot = x; }
   
   inline int ClearButShare ( void ) { return bot &= (BOT_PASSIVE | BOT_AGGRESSIVE); }
   
   inline void KillOwnerBotFlags ( void ) { bot &= ~
	(BOT_PASSIVE|BOT_AGGRESSIVE|BOT_GLOBAL|BOT_SHAREFLAGS|BOT_CROSS);
   }
   inline void BotFlagsOnly ( void ) { chan &= BOT_AGGRESSIVE; }
   
   inline void SetMatch ( int x ) { match = x; }
   inline int Match ( void ) { return match; }
   inline void Clear ( void ) { global = udef_global = bot = chan = udef_chan = 0; }
   FlagRecord * Get (UserRecord *, const char * = 0);
   FlagRecord * Set (UserRecord *, const char * = 0);
   FlagRecord & operator += ( FlagRecord &);
   FlagRecord & operator -= ( FlagRecord &);
   void BreakDown (const char *, FlagRecord * f = 0);
   void Build (char *, FlagRecord * x = 0);
   int operator == ( FlagRecord & );
   int Ok ( FlagRecord & );
   void Sanity(void);
   
   static void ReaffirmOwners (void);
};
#endif
