/*
 * HelpReference.C - member functions for the HelpRefence class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "HelpReference.h"
#include "DccChatPass.h"
#define CLASS "HelpKeys"

HelpKeys::~HelpKeys ( void ) {
   FUNCTION;
}

#undef CLASS
#define CLASS "HelpReference"

HelpReference::HelpReference ( const char * name , ModuleEntry * m )
: KeyedList<MISCLEN> ( *G->help_list, name, KeyedList<MISCLEN>::NODUPS |
		      KeyedList<MISCLEN>::APPEND, m ) {
   FUNCTION;
   char s[DIRLEN+1];
   
   G->SPrintF(s, DIRLEN, "%smsg/%s", G->helpdir, name);
   ScanFile(s,0);
   G->SPrintF(s, DIRLEN, "%s%s", G->helpdir, name);
   ScanFile(s,1);
   G->SPrintF(s, DIRLEN, "%sset/%s", G->helpdir, name);
   ScanFile(s,2);
}

HelpReference::~HelpReference ( void ) {
   FUNCTION;
   if (key[0]) {
      G->help_list->Remove(this);
      next = 0;
   }
}

int HelpReference::ExpMem ( void ) {
   FUNCTION;
   int tot = sizeof(HelpReference);
   HelpReference * ptr = this;
   HelpKeys * hk;
   
   while ((ptr = ptr->Next())) {
      HelpEntry * he = &ptr->first;
      
      tot += sizeof(HelpReference);
      while ((he = he->Next())) {
	 tot += sizeof(HelpEntry);
	 hk = &he->keylist;
	 while ((hk = (HelpKeys *)hk->next))
	   tot += sizeof(HelpKeys);
      }
   }
   return tot;
}

void HelpReference::ScanFile ( char * filename, int type ) {
   FUNCTION;
   FILE * f = NULL;
   char s[HELP_BUF_LEN + 1], *p, *q;
   HelpEntry * he = 0;
   
   if ((f = fopen(filename,"r"))) {
      while (!feof(f)) {
	 fgets(s, HELP_BUF_LEN, f);
	 if (!feof(f)) {
	    p = s;
	    while ((q = strstr(p,"%{"))) {
	       q += 2;
	       if (!strncmp("help=", q, 5)) {
		  q += 5;
		  if ((p = strchr(q,'}'))) {
		     *p = 0;
		     he = NEW ( HelpEntry ( this, q,  type ) );
		     p++;
		  } else 
		    p = "";
	       } else if (he && !strncmp("keys=", q, 5)) {
		  q += 5;
		  if ((p = strchr(q,'}'))) {
		     char * r;
		     
		     *p = 0;
		     while (q) {
			if ((r = strchr(q, ',')))
			  *r = 0;
			NEW ( HelpKeys ( he, q ) );
			q = r ? r + 1 : 0;
		     }
		     p++;
		  } else 
		    p = "";
	       } else if ((p = strchr(q,'}'))) 
		    p++;
	       else 
		 p = "";
	    }
	 }
      }
      fclose(f);
   }
}

void HelpReference::Reload ( void ) {
   FUNCTION;
   HelpReference * current = this, * next = current->Next();
   int entries=0, count=0;

   while ((current = next)) {
      next = current->Next();
      entries++;
   }
   current = this;
   next = current->Next();
   while ((current = next)) {
      next = current->Next();
      G->help_list->Remove(current);
      NEW ( HelpReference( current->Key() ));
      DELETE ( current );
      count++;
      if (count == entries)
	break;
   }
}  

void HelpReference::Debug (Socket * idx) {
   FUNCTION;
   HelpReference * current = this;
   
   while ((current = current->Next())) {
      HelpEntry * item = current->First();
      
      idx->Send(G->Lang( 0x136, "s", current->Key()));
      while ((item = item->Next())) {
	 HelpKeys * k = &item->keylist;
	 
	 idx->Send(G->Lang( 0x135, "ss", item->Key(), 
			   (item->Type() == 0) ? "msg/" : 
			   (item->Type() == 1) ? "" : "set/"));
	 while ((k = (HelpKeys*)k->next)) 
	   idx->Send(G->Lang(0x146, "s", k->Key()));
      }
   } 
}

FILE * HelpReference::Resolve ( int dcc, char * file) {
   FUNCTION;
   char s[DIRLEN+1], *p;
   FILE * f;
   HelpReference * current = this;
   
   /* somewhere here goes the eventual substituation */
   if (!(dcc & TEXT))
     while ((current = current->Next())) {
	HelpEntry * item = current->First()->Find(file);
	if (item) {
	   if (!item->Type() && !dcc) {
	      G->SPrintF(s, DIRLEN, "%smsg/%s", G->helpdir, current->Key());
	      if ((f = fopen(s,"r")))
		return f;
	   } else if (dcc && item->Type()) {
	      if (item->Type() == 1) 
		G->SPrintF(s, DIRLEN, "%s%s", G->helpdir, current->Key());
	      else
		G->SPrintF(s, DIRLEN, "%sset/%s", G->helpdir, current->Key());
	      if ((f = fopen(s,"r")))
		return f;
	   }
	}
     }
   for (p = s + G->SPrintF(s, DIRLEN, "%s%s", G->helpdir, dcc ? "" : "msg/");
	*file && (p < s + DIRLEN); file++, p++) {
      switch(*file) {
       case ' ':
       case '.':
	 *p = '/';
	 break;
       case '-':
	 *p = '-';
	 break;
       case '+':
	 *p = 'P';
	 break;
       default:
	 *p = *file;
      }
   }
   *p = 0;
   if (!is_file(s)) {
      strcat(s, "/");
      strcat(s, file);
      if (!is_file(s)) 
	return NULL;
   }
   return fopen(s, "r");
}

void HelpReference::Tell (DccChatPass * idx, char * file, 
			  FlagRecord & flags, int fl) {
   FUNCTION;
   char s[HELP_BUF_LEN+1];
   int lines = 0;
   FILE *f = Resolve(DCC | fl, file);
   
   if (f) {
      G->HelpSubst(0, 0, 0, HS_ISDCC | 
		   ((idx->StripFlags() & STRIP_HILITE) ? 0 : HS_HILITE), 0);
      while (!feof(f)) {
	 fgets(s, HELP_BUF_LEN, f);
	 if (!feof(f)) {
	    if (s[strlen(s) - 1] == '\n')
	      s[strlen(s) - 1] = 0;
	    if (!s[0])
	      strcpy(s, " ");
	    G->HelpSubst(s, idx->Nick(), &flags, 0, file);
	    if (s[0]) {
	       idx->Send(s);
	       lines++;
	    }
	 }
      }
      fclose(f);
   }
   if (!lines && !(fl & TEXT))
      idx->Send(G->Lang( 0x10B, ""));
}

void HelpReference::Search (DccChatPass * idx, char * word, FlagRecord &) {
   FUNCTION;
   HelpReference * current = this;
   HelpEntry * he;
   int count = 0;
   
   idx->Send(G->Lang(0x298, "s", word));
   /* somewhere here goes the eventual substituation */
   while ((current = current->Next())) {
      he = &current->first; 
      while ((he = he->Next())) 
	if (he->keylist.Find(word)) {
	   count++;
	   idx->Send(G->Lang(0x29B,"s", he->Key()));
	}
   }
   idx->Send(G->Lang(count ? 0x29A : 0x299, "d", count));
   return;
}
