/*
 * HookList.C - class members for HookList class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "HookList.h"
#define CLASS "Hook"

Hook::Hook ( HookList * l, void (*f) (void), ModuleEntry * m)
: List (l->top) {
   FUNCTION;
   func = f;
   mod = m;
}

Hook::~Hook ( void ) {
   FUNCTION;
}

int Hook::ExpMem ( void ) {
   FUNCTION;
   int tot = 0;
   Hook * ptr = this;
   while ((ptr = ptr->Next()))
     tot += sizeof(Hook);
   return tot;
}

#undef CLASS
#define CLASS "HookList"

HookList::~HookList ( void ) {
   FUNCTION;
}

int HookList::ExpMem ( void ) {
   FUNCTION;
   HookList * ptr = this;
   int tot = sizeof(HookList);
   
   while ((ptr = ptr->Next()))
     tot += sizeof(HookList) + ptr->top.ExpMem();
   return tot;
}  

void HookList::Add ( void (*f) (void), ModuleEntry * m ) {
   FUNCTION;
   Hook * ptr = &top;
   
   while ((ptr = ptr->Next()))
     if (ptr->Func() == f)
       break;
   if (!ptr)
     NEW ( Hook ( this, f, m ) );
}

void HookList::Remove ( void (*f) (void) ) {
   FUNCTION;
   Hook * ptr = &top;
   
   while ((ptr = ptr->Next()))
     if (ptr->Func() == f) {
	DELETE ( ptr );
	break;
     } 
}

void HookList::Cleanup ( ModuleEntry * m ) {
   FUNCTION;
   HookList * ptr, *pback = this;
   Hook * hook, *hback;
   
   while ((ptr = pback->Next())) {
      if (ptr->Module() == m) {
	 List::Remove( ptr );
	 DELETE ( ptr );
      } else {
	 hback = &ptr->top;
	 while ((hook = hback->Next())) {
	    if (hook->Module() == m) {
	       ptr->top.Remove(hook);
	       DELETE ( hook );
	    } else 
	      hback = hook;
	 }
	 pback = ptr;
      }
   }
}

void HookList::Dump ( Socket * idx ) {
   FUNCTION;
   HookList * ptr = this;
   Hook * hook;
   
   while ((ptr = ptr->Next())) {
      idx->Send(G->Lang(0x291, "s", ptr->key));
      hook = &ptr->top;
      while ((hook = hook->Next()))
	hook->Dump(idx);
   }
}	         

void HookList::Call ( void ) {
   FUNCTION;
   Hook * ptr = &top;
   
   while ((ptr = ptr->Next()))
     ptr->Call();
}

