/*
 * HostnameCache.C - member functions
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "HostnameCache.h" 
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <setjmp.h>
#define CLASS "HostnameCache"

extern jmp_buf alarmret;

HostnameCache::HostnameCache ( const char * str ) 
: KeyedList<UHOSTLEN> ( *G->host_cache, str, KeyedList<UHOSTLEN>::NODUPS ) {
   FUNCTION;
   struct hostent * he;
   
   if (!setjmp(alarmret)) {
      alarm(G->resolve_timeout);
      he = gethostbyname(str);
      alarm(0);
   } else
     he = 0;
   age = G->now;
   if (he) 
     address.s_addr = *(int *)he->h_addr_list[0];
   else
     address.s_addr = 0;
}

HostnameCache::~HostnameCache ( void ) {
   FUNCTION;
}

HostnameCache::HostnameCache ( struct in_addr * in_ptr ) 
: KeyedList<UHOSTLEN> ( ) {
   FUNCTION;
   const char * name;
   struct hostent * he;
   
   if (!setjmp(alarmret)) {
      alarm(G->resolve_timeout);
      he = gethostbyaddr((char *)in_ptr, sizeof(struct in_addr), AF_INET);
      alarm(0);
   } else
     he = 0;
   
   age = G->now;
   address.s_addr = in_ptr->s_addr;
   name = he ? he->h_name : inet_ntoa(*in_ptr);
   KeyedList_Cons(*G->host_cache, name, KeyedList<UHOSTLEN>::NODUPS);
}

struct in_addr * HostnameCache::Find ( const char * what ) {
   FUNCTION;
   HostnameCache * which = (HostnameCache *)KeyedList<UHOSTLEN>::Find(what);
   
   if (!which) 
     which = NEW ( HostnameCache ( what ) );
   if (which->address.s_addr)
     return & which->address;
   return 0;
}

const char * HostnameCache::Find ( struct in_addr * which) {
   FUNCTION;
   HostnameCache * ptr = this;
   
   while ((ptr = (HostnameCache *)ptr->Next()))
     if (which->s_addr == ptr->address.s_addr) 
       break;
   if (!ptr) 
     ptr = NEW ( HostnameCache ( which ) );
   if (ptr->key[0])
     return ptr->key;
   return 0;
}

void HostnameCache::Flush ( void ) {
   FUNCTION;
   HostnameCache * ptr, * optr = this;
   int count = 0;
   
   while ((ptr = (HostnameCache *)optr->Next()) 
	  && ((ptr->age - G->now) < (G->hostcache_age * 60))
	  && (++count < G->hostcache_count))
     optr = ptr;
   if (ptr) {
      optr->next = 0;
      DELETE ( ptr );
   }
}

int HostnameCache::ExpMem ( void ) {
   FUNCTION;
   HostnameCache * ptr = this;
   int tot = sizeof(HostnameCache);
   
   while ((ptr = (HostnameCache *)ptr->Next()))
     tot += sizeof(HostnameCache);
   return tot;
}
