/*
 * IgnoreRecord.C - member functions for the IgnoreRecord class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "IgnoreRecord.h"
#include "Socket.h"
#include "ModuleEntry.h"
#ifdef HAVE_SERVER_H
#include "server.h"
#include "ModuleEntryServer.h"
#endif
#define CLASS "IgnoreRecord"

IgnoreRecord::IgnoreRecord ( char * desc )
: KeyedList<UHOSTLEN> ( ) {
   FUNCTION;
   char * expi, * usr, * addd, * host;
   int flgs = 0;

   host = G->Split(&desc);
   expi = G->Split(&desc);
   if (*expi == '+') {
      flgs |= IgnoreRecord::PERM;
      expi++;
   }
   usr = G->Split(&desc);
   addd = G->Split(&desc);
   if (*addd) {
      KeyedList_Cons(*G->ignores, host, KeyedList<UHOSTLEN>::NODUPS |
		     KeyedList<UHOSTLEN>::APPEND);
      expires = atoi(expi);
      added = atoi(addd);
      flags = flgs;
      strncpy(user, usr, HANDLEN);
      user[HANDLEN] = 0;
      msg = (char *)my_malloc(strlen(desc)+1,G->memtype_ignore);
      strcpy(msg,desc);
   } else
     G->LOG_MISC.Put(G->Lang( 0x278, ""));
}

IgnoreRecord::IgnoreRecord ( const char * ign, const char * from,
			    const char * note, time_t expire_time ) 
: KeyedList<UHOSTLEN> ( *G->ignores, ign, KeyedList<UHOSTLEN>::NODUPS 
		       | KeyedList<UHOSTLEN>::APPEND) {
   FUNCTION;
   expires = expire_time;
   added = G->now;
   flags = expire_time ? 0 : PERM;
   strncpy(user, from, HANDLEN);

   msg = (char *)my_malloc(strlen(note) + 1, G->memtype_ignore);
   strcpy(msg,note);
   if (!G->noshare && G->shareout)
      G->shareout(NULL,"i%s %lu %c %s %s\n", ign, expire_time - G->now,
	       (flags & PERM) ? 'p' : '-', from, note);
}

IgnoreRecord::~IgnoreRecord ( void ) {
   FUNCTION;
   if (key[0]) {
      if (!G->noshare && G->shareout)
	G->shareout(NULL,"I%s\n", key);
      G->ignores->Remove(this);
   }
   if (msg)
     my_free(msg, G->memtype_ignore);
}

int IgnoreRecord::ExpMem ( void ) {
   FUNCTION;
   IgnoreRecord * ptr = this;
   int tot = sizeof(IgnoreRecord);
   
   while ((ptr = ptr->Next())) 
     tot += sizeof(IgnoreRecord) 
       + (ptr->msg ? strlen(ptr->msg) + 1 : 0);
   return tot;
}

IgnoreRecord * IgnoreRecord::Find ( char * val ) {
   FUNCTION;
   int which = atoi(val);
   IgnoreRecord * ptr = this;
   
   if (!which) 
     ptr = (IgnoreRecord *)KeyedList<UHOSTLEN>::Find(val);
   else {
      while ((ptr = ptr->Next()) && --which);
      if (ptr)
	strncpyz ( val, ptr->Key(), UHOSTLEN );
   }
   return ptr;
}

/* check for expired timed-ignores */
void IgnoreRecord::Expire ( void ) {
   FUNCTION;
   IgnoreRecord * ptr = this, *next;
#ifdef HAVE_SERVER_H
   ModuleEntryServer * me;
#endif

   while (ptr && (next = ptr->Next())) {
      if (!(next->flags & PERM) && (G->now >= next->expires)) {
	 G->LOG_MISC.Put(G->Lang( 0x205, "s", next->key));
#ifdef HAVE_SERVER_H
	 if ((me = (ModuleEntryServer *)
	      G->module_list->Find("server", &G->ServerVer)))
	   me->SendUnSilence(Q_SERVER, next->key);
#endif
	 DELETE ( next );
      } else 
	ptr = next;
   }
}

void IgnoreRecord::Display ( Socket * idx, const char * match ) {
   FUNCTION;
   IgnoreRecord * ptr = this;
   int k = 0;
   char s[81], *expire;
   
   idx->Send(G->Lang( next ? 0x204 : 0x203, ""));
   while ((ptr = ptr->Next())) {
      k++;
      if (!match[0] || G->WildMatch(match, ptr->key)
	  || G->WildMatch(match, ptr->msg) || G->WildMatch(match, ptr->user)) {
	 if (ptr->flags & PERM)
	   expire = G->Lang( 0x271, "");
	 else {
	    G->Days(ptr->expires, G->now, s);
	    expire = G->Lang( 0x272, "s", s);
	 }
	 idx->Send(G->Lang( (k >= 0) ? 0x273 : 0x274, "ss3", 
			ptr->key, expire, k));
	 idx->Send(G->Lang( (ptr->msg && ptr->msg[0]) ? 0x275 : 0x276,
			"ss", ptr->user, ptr->msg));
	 if (ptr->added) {
	    G->DaysAgo(G->now, ptr->added, s);
	    idx->Send(G->Lang( 0x270, "s", s));
	 }
      }
   }
}

int IgnoreRecord::Write ( FILE * f ) {
   FUNCTION;
   IgnoreRecord * ptr = this;
   
   if (next && (fprintf(f, "*i\n") == EOF))
     return 0;
   while ((ptr = ptr->Next()))
     if (fprintf(f,"-%s %s%lu %s %lu %s\n", ptr->key,
		 (ptr->flags & PERM) ? "+" : "", ptr->expires,
		 ptr->user, ptr->added, safe_str(ptr->msg)) == EOF)
       return 0;
   return 1;
}

