/*
 * List.C - member functions for class List
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "List.h"
#include "Socket.h"
#define CLASS "List"

ModuleEntryEggdrop * List::G = 0;

List::~List ( void ) {
   if (next)
     delete (next);
}

void List::Add(List * which, int append) {
   FUNCTION;
   List * ptr = this;
   
   which->next = next;
   if (which->next && append) {
      which->next = 0;
      while (ptr->next)
	ptr = ptr->next;
   }
   ptr->next = which;
}

int List::Remove(List * which) {
   FUNCTION;
   List * ptr = next, * old = this;
   
   while (ptr && (ptr != which)) {
      old = ptr;
      ptr = ptr->next;
   }
   if (ptr) {
      old->next = ptr->next;
      ptr->next = 0;
   }
   return ptr != 0;
}

int List::Count ( void ) {
   FUNCTION;
   List * ptr = this;
   int count = 0;
   
   while ((ptr = ptr->next))
     count++;
   return count;
}

void List::Flush(void) {
   FUNCTION;
   List * ptr, * nxt = next;
   
   while ((ptr = nxt)) {
      nxt = ptr->next;
      ptr->next = 0;
      DELETE ( ptr );
   }
}

void List::Cleanup (ModuleEntry * m ) {
   FUNCTION;
   List * ptr, *bptr = this;

   while ((ptr = bptr->next)) {
      if (ptr->mod == m) {
	 bptr->next = ptr->next;
	 ptr->next = 0;
	 DELETE ( ptr );
      } else
	bptr = ptr;
   }
}

#ifdef EBUG_MEM
Debugger * Debugger::full_list = 0;
Debugger * Debugger::last = 0;

Debugger::~Debugger ( void ) {
   if (G && G->Stdout && (G->term_z == 2)) {
      if (prev) 
	G->Stdout->Printf("<-%s::%s->%s::%s\n", clss, func, prev->clss,prev->func);
      else
      	G->Stdout->Printf("<-%s::%s\n", clss, func);
   }
   if (last != this) 
     G->Fatal("Jumping to far back up debug stack!");
   else if (prev && (prev->next != this))
     G->Fatal("Debug stack died tracing back!");
   else if (prev) {
      prev->next = 0;
      last = prev;
   } else 
     full_list = last = 0;
   next = 0;
}

Debugger::Debugger ( const char * cl,const char * fn, const char * mn, 
		    const char * fl, int ln ) {
   clss = cl;
   func = fn;
   modl = mn;
   file = fl;
   line = ln;
   if (last && (!last->prev || (last->prev->next == last))) {
      last->next = this;
      prev = last;
      last = this;
   } else if (last) {
	 G->Fatal("Debugger stack died!");
   } else {
      full_list = last = this;
      prev = 0;
   }
   next = 0;
   if (G && G->Stdout && (G->term_z == 2)) 
      G->Stdout->Printf("%s::%s (%s,%s,%d)\n", clss, func, modl, file, line);
}

void Debugger::DumpStack(Socket * desto) {
   if (next && (((Debugger *)next)->prev != this))
     desto->Printf("Error linking!\n");
   else {
      desto->Printf("%s::%s (%s,%s,%d)\n", clss, func, modl, file, line);
      if (next)
	((Debugger *)next)->DumpStack(desto);
   }
}

int Debugger::ExpMem ( void ) {
   return 0;
}
#endif
