/*
 * ListeningSocket.C - class members for ListeningSocket class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ListeningSocket.h"
#include "HostnameCache.h"
#include <arpa/inet.h>
#include <sys/socket.h>
#define CLASS "ListeningSocket"

ListeningSocket::ListeningSocket (const char * ntype, struct in_addr * addr,
				  int port )
: Socket ( LISTENING, ntype, 0, 0, 0 ) {
   FUNCTION;
   int parm = 1;
   localport = port;
   address.s_addr = addr ? addr->s_addr
     : G->natip[0] ? inet_addr(G->natip) : G->GetMyIp()->s_addr;
   setsockopt(key, SOL_SOCKET, SO_REUSEADDR, (void *) &parm, sizeof(int));
}
   
int ListeningSocket::Enable ( void ) {
   FUNCTION;
   struct sockaddr_in sain;
#if __GLIBC__ >= 2
   unsigned
#endif
     int addrlen = sizeof(struct sockaddr_in);

   memset((char *) &sain, 0, sizeof(struct sockaddr_in));
   sain.sin_family = AF_INET;
   sain.sin_port = htons(localport);
   sain.sin_addr.s_addr = address.s_addr;
   if ((bind(key, (struct sockaddr *)&sain,
	     sizeof(struct sockaddr_in)) >= 0) &&
       (getsockname(key, (struct sockaddr *) &sain, &addrlen) >= 0) &&
       (listen(key, 1) >= 0)) {
      localport = htons(sain.sin_port);
      return 1;
   } 
   return 0;
}

Socket * ListeningSocket::Answer ( void ) {
   FUNCTION;
   struct sockaddr_in sain;
#if __GLIBC__ >= 2
   unsigned
#endif
     int sain_size = sizeof(struct sockaddr_in);
   int fd;
   Socket * sock;
   const char * name;
   
   memset((char *) &sain, 0, sizeof(struct sockaddr_in));
   fd = accept(key,(struct sockaddr *)&sain, &sain_size);
   if ((fd >= 0) && (sock = Create ( fd ))) {
      LogModes Socklog ( LogModes::SOCKET );
      
      FD_SET(fd,&G->fds_read);
      sock->localport = localport;
      sock->remoteport = htons(sain.sin_port);
      sock->address.s_addr = htonl(sain.sin_addr.s_addr);
      if ((name = G->host_cache->Find(&sain.sin_addr))) {
	 strncpy(sock->host,name,UHOSTLEN);
	 sock->host[UHOSTLEN] = 0;
      } else 
	strcpy(sock->host, inet_ntoa(sain.sin_addr));
      Socklog.Put(G->Lang(0x003, "Xddd", sock->address.s_addr, 
			  sock->remoteport,sock->localport,sock->Key()));
      return sock;
   }
   return NULL;   
}
