/*
 * LogFile.C - member functions for the LogFile class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "LogFile.h"
#include "Socket.h"
#define CLASS "LogFile"

LogFile::LogFile( const char * file, const char * chan, LogModes m)
: KeyedList<UHOSTLEN> ( *G->log_list, file, KeyedList<UHOSTLEN>::APPEND |
		       KeyedList<UHOSTLEN>::NODUPS) {
   FUNCTION;
   strncpy(chname, chan, UHOSTLEN);
   chname[UHOSTLEN] = 0;
   flags = m;
}

LogFile::~LogFile ( void ) {
   FUNCTION;
   if (!key[0]) {
      while ( next )
	DELETE ( next );
   } else {
      if (f)
	fclose(f);
      G->log_list->Remove(this);
   }
}

int LogFile::ExpMem ( void ) {
   FUNCTION;
   int tot = sizeof(LogFile);
   LogFile * ptr = this;
   while ((ptr = ptr->Next()))
     tot += sizeof(LogFile);
   return tot;
}

void LogFile::Close ( void ) {
   FUNCTION;
   if (f)
     fclose(f);
   f = 0;
   if (next)
     Next()->Close();
}

void LogFile::Backup ( void ) {
   FUNCTION;
   char s[DIRLEN+1];
   if (key[0]) {
      if (f)
	fclose(f);
      f = 0;
      G->SPrintF(s, DIRLEN, "%s.yesterday", key);
      unlink(s);
      rename(key, s);
   }
   if (next)
     Next()->Backup();
}

void LogFile::Display ( Socket * idx ) {
   FUNCTION;
   if (key[0]) 
     idx->Send(G->Lang(0x211, "ssss", key, chname, flags.Mask(),flags.Names()));
   if (next)
     Next()->Display(idx);
}
      
void LogFile::Output ( char * data ) {
   FUNCTION;
   char s1[DIRLEN+1], ct[81];
   time_t tt = G->now;
   
   if (!f) {
      if (G->keep_all_logs) {
	 strcpy(ct, ctime(&tt));
	 ct[7] = 0;
	 ct[24] = 0;
	 G->SPrintF(s1, DIRLEN, "%s.%c%c%s%s", key,
		    ct[8] == ' ' ? '0' : ct[8], ct[9], ct + 4, ct + 20);
	    f = fopen(s1, "a+");
      } else
	f = fopen(key, "a+");
   }
   if (f)
     fputs(data, f);
}

void LogFile::Flush ( void ) {
   FUNCTION;
   if (f)
     fflush(f);
   if (next)
     Next()->Flush();
}
