/*
 * LogModes.C - class members for LogModes class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "DccChat.h"
#include "LogFile.h"
#include "Channel.h"
#define CLASS "LogModes"

unsigned int LogModes::Set (const char * s) {
   FUNCTION;
   value = 0;
   
   while (*s)
     Set(*s++);
   return value;
}

unsigned int LogModes::Set ( char s ) {
   FUNCTION;
   switch (s) {
     case 'm':
     case 'M':
      value |= MSGS;
      break;
     case 'p':
     case 'P':
      value |= PUBLIC;
      break;
     case 'j':
     case 'J':
	value |= JOIN;
      break;
     case 'k':
     case 'K':
      value |= MODES;
      break;
     case 'c':
     case 'C':
      value |= CMDS;
      break;
     case 'o':
     case 'O':
	value |= MISC;
      break;
     case 'b':
     case 'B':
      value |= BOTS;
      break;
     case 'r':
     case 'R':
      value |= List::G->use_console_r ? RAW : NONE;
      break;
     case 'w':
     case 'W':
      value |= WALL;
      break;
     case 'x':
     case 'X':
      value |= FILES;
      break;
     case 's':
      value |= SERV;
      break;
     case 'S':
      value |= SOCKET;
      break;
     case 'v':
     case 'V':
      value |= List::G->debug_output ? SRVOUT : NONE;
      break;
     case 't':
      value |= List::G->debug_output ? BOTNET : NONE;
      break;
     case 'T':
      value |= List::G->debug_output ? BOTNET_SEND : NONE;
      break;
     case 'h':
     case 'H':
      value |= List::G->debug_output ? BOTSHARE : NONE;
      break;
     case '1':
      value |= LEV1;
      break;
     case '2':
      value |= LEV2;
      break;
     case '3':
      value |= LEV3;
      break;
     case '4':
      value |= LEV4;
      break;
     case '5':
      value |= LEV5;
      break;
     case '6':
      value |= LEV6;
      break;
     case '7':
      value |= LEV7;
      break;
     case '8':
      value |= LEV8;
      break;
     case '*':
      value |= ALL;
      break;
   }
   return value;
}

const char * LogModes::Mask ( void ) {
   FUNCTION;
   static char s[20];
   char *p = s;
   if (value & MSGS)
      *p++ = 'm';
   if (value & PUBLIC)
      *p++ = 'p';
   if (value & JOIN)
      *p++ = 'j';
   if (value & MODES)
      *p++ = 'k';
   if (value & CMDS)
      *p++ = 'c';
   if (value & MISC)
      *p++ = 'o';
   if (value & BOTS)
      *p++ = 'b';
   if ((value & RAW) && List::G->use_console_r)
      *p++ = 'r';
   if (value & FILES)
      *p++ = 'x';
   if (value & SERV)
      *p++ = 's';
   if (value & WALL)
      *p++ = 'w';
   if ((value & SRVOUT) && List::G->debug_output)
     *p++ = 'v';
   if ((value & BOTNET) && List::G->debug_output)
     *p++ = 't';
   if ((value & BOTNET_SEND) && List::G->debug_output)
     *p++ = 'T';
   if ((value & BOTSHARE) && List::G->debug_output)
     *p++ = 'h';
   if (value & SOCKET)
      *p++ = 'S';
   if (value & LEV1)
      *p++ = '1';
   if (value & LEV2)
      *p++ = '2';
   if (value & LEV3)
      *p++ = '3';
   if (value & LEV4)
      *p++ = '4';
   if (value & LEV5)
      *p++ = '5';
   if (value & LEV6)
      *p++ = '6';
   if (value & LEV7)
      *p++ = '7';
   if (value & LEV8)
      *p++ = '8';
   if (p == s)
     *p++ = '-';
   *p = 0;
   return s;
}

const char * LogModes::Names ( void ) {
   FUNCTION;
   static char s[161];
   int i = 0;
   
   s[0] = 0;
   if (value & MSGS) 
     i += List::G->SPrintF(s, 160, "%s, ", List::G->Lang( 0x1AF, ""));
   if (value & PUBLIC)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B0, ""));
   if (value & JOIN)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B1, ""));
   if (value & MODES)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B2, ""));
   if (value & CMDS)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B3, ""));
   if (value & MISC)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B4, ""));
   if (value & BOTS)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B5, ""));
   if ((value & RAW) && List::G->use_console_r) 
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B6, ""));
   if (value & FILES) 
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B7, ""));
   if (value & SERV) 
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B8, ""));
   if (value & WALL)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1BA, ""));
   if ((value & SRVOUT) && List::G->debug_output)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1BB, ""));
   if ((value & BOTNET) && List::G->debug_output)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1BC, ""));
   if ((value & BOTNET_SEND) && List::G->debug_output)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1B9, ""));
   if ((value & BOTSHARE) && List::G->debug_output)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1BD, ""));
   if (value & SOCKET)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1BE, ""));
   if (value & LEV1) 
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1BF, ""));
   if (value & LEV2)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C0, ""));
   if (value & LEV3)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C1, ""));
   if (value & LEV4)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C2, ""));
   if (value & LEV5)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C3, ""));
   if (value & LEV6)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C4, ""));
   if (value & LEV7) 
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C5, ""));
   if (value & LEV8)
     i += List::G->SPrintF(s + i, 160 - i, "%s, ", List::G->Lang( 0x1C6, ""));
   if (i)
     s[i-2] = 0;
   else
     strcpy(s, List::G->Lang( 0x1C7, ""));
   return s;
}

void LogModes::Put ( Channel * chan, const char * format, ... )  {
   va_list va;
   char buf[768];
   
   va_start(va, format);
   
   /* format log entry at offset 8, then i can prepend the timestamp */
#ifdef HAVE_VSNPRINTF
   if (vsnprintf(buf, 500, format, va) < 0)
     buf[500] = 0;
#else
   vsprintf(buf, format, va);
#endif   
   Put(buf, chan);
}

void LogModes::Put ( const char * out, Channel * chan)  {
   char s[510], *s1;
   time_t tt = List::G->now;
   DccChat * ptr = (DccChat *)List::G->socklist;
   LogFile * lf = List::G->log_list;

   if (out[0] && List::G->shtime) {
      s1 = ctime(&tt);
      s[0] = '[';
      s[1] = s1[11];
      s[2] = s1[12];
      s[3] = s1[13];
      s[4] = s1[14];
      s[5] = s1[15];
      s[6] = ']';
      s[7] = ' ';
      s[8] = 0;
      s1 = s + 8;
   } else
     s1 = s;
   List::G->SPrintF(s1, 500, "%s\n", out);
   if (!List::G->use_stderr) 
     while ((lf = lf->Next())) 
       if ((lf->Mask().value & value) && (!chan ||
					  (lf->Channel()[0] == '*') ||
					  !strcasecmp(chan->Key(), lf->Channel()))) 
	 lf->Output(s);
   if (!List::G->backgrd && !List::G->con_chan && !List::G->term_z)
     List::G->Stdout->Send(s);
   while ((ptr = ptr->Next()))
     if ((ptr->ConsoleFlags().value & value) &&
	 (!chan || (ptr->ConsoleChan()[0] == '*') ||
	  !strcasecmp(chan->Key(), ptr->ConsoleChan())))
       ptr->Send(s);
   if ((value & MISC) && List::G->use_stderr)
     List::G->Stderr->Send(s1);
}
