/*
 * MemType.C - member functions for MemType class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#define MEM_C
#include "Socket.h"
#include <stdlib.h>
#if !defined(malloc) && !defined(FREEBSD) && defined(HAVE_MALLOC_H)
#include <malloc.h>
#endif
#include <stdio.h>
#include <string.h>

#define MEMTABLE_INIT_SIZE 32 /* initial size of each memory table */
#define MEMTABLE_MEMTYPE  128 /* initial size of memtype memory table */
#define CLASS "MemType"

static int total_memory;

int MemType::ExpMem ( void ) {
   FUNCTION;
   MemType * temp = this;
   int size = (sizeof(struct memtype_internal_struct) * temp->TableSize());
   
   while ((temp = temp->Next()))
     size += sizeof(MemType) +
     (sizeof(struct memtype_internal_struct) * temp->TableSize());
   return size;
}

MemType::MemType ( void ) : KeyedList<MISCLEN> () {
   FUNCTION;
#ifdef EBUG_MEM
   int size = MEMTABLE_MEMTYPE * sizeof(struct memtype_internal_struct);
#endif
   context;
   next = 0;
   SetKey ("memtype");
   G->memtype_list = this;
   ptr = (List **)&G->memtype_list;
#ifdef EBUG_MEM
   total_memory = expmem = size;
   lasttable = 1;
   tablesize = MEMTABLE_MEMTYPE;
   data = (struct memtype_internal_struct *)malloc(size);
   G->SPrintF(data[0].where, 59, "%s/%s:%d", MODULE_NAME, "mem.c", __LINE__);
   data[0].size = size;
   data[0].ptr = data;
#else
   total_memory = expmem = 0;
   tablesize = 0;
   data = 0;
   lasttable = 0;
#endif
   G->memtype_misc = NEW ( MemType ( "misc", 0 ) );
}

MemType::MemType ( const char * name, List ** p, ModuleEntry * m)
: KeyedList<MISCLEN> ( *G->memtype_list, name, 0, m) {
   FUNCTION;
   expmem = 0;
   lasttable = 0;
   ptr = p;
#ifdef EBUG_MEM
   tablesize = MEMTABLE_INIT_SIZE;
   data = (struct memtype_internal_struct *)
     my_malloc(sizeof(struct memtype_internal_struct) *
			    MEMTABLE_INIT_SIZE, G->memtype_list);
#else
   tablesize = 0;
   data = 0;
#endif
}

MemType::~MemType ( void ) {
   FUNCTION;
   G->memtype_list->Remove(this);
   next = 0;
#ifdef EBUG_MEM
   if (expmem)
     G->LOG_MISC.Put(G->Lang( 0x224, "s", key));
   context;
   G->memtype_list->Free(data);
#endif
}

void MemType::Free ( void * ptr ) {
   FUNCTION;
   char * where = G->context_stack[G->context_ptr];
#ifdef EBUG_MEM
   int i, size = 0;
#else
   int size, *tptr = (int *)ptr;
#endif
#ifdef DEBUG_MEM_PARANOID
   MemType * type;

   for (type = G->memtype_list; type; type = type->Next())
     if (this == type)
       break;
   if (!type) {
      G->LOG_MISC.Put(G->Lang( 0x21F, "s", where));
      exit (1);
   } 
#endif
#ifdef EBUG_MEM
   for (i = 0; (i < lasttable) && (data[i].ptr != ptr); i++);
   if (i >= lasttable) {
      G->LOG_MISC.Put(G->Lang( 0x220, "s", where));
#ifdef DEBUG_MEM_PARANOID
      exit (1);
#endif
   } else {
      lasttable --;
      size = data[i].size;
      expmem -= size;
      if (i < lasttable) {
	 data[i].size = data[lasttable].size;
	 data[i].ptr = data[lasttable].ptr;
	 strcpy(data[i].where, data[lasttable].where);
      }
      total_memory -= size;
   }
#else
   size = *--tptr;
   expmem -= size;
   total_memory -= size;
#endif
   if (!ptr) {
      G->LOG_MISC.Put(G->Lang( 0x221, "s", where));
#ifdef DEBUG_MEM_PARANOID
      exit (1);
#endif
   }
#ifdef EBUG_MEM
   free(ptr);
#else
   free(tptr);
#endif
}

void * MemType::Malloc (size_t size, int) {
   FUNCTION;
   char * where = G->context_stack[G->context_ptr];
#ifdef EBUG_MEM
   void * buffer;
#else
   int * buffer;
#endif
#ifdef DEBUG_MEM_PARANOID
   MemType * type;
   
   for (type = memtype_list; type; type = type->next)
     if (this == type)
       break;
   if (!type) {
      G->LOG_MISC.Put(G->Lang( 0x222, "sd", where, size));
      exit (1);
   } 
#endif
#ifdef EBUG_MEM
   buffer = malloc ( size );
#else
   buffer = (int *)malloc ( size + sizeof(int) + 1 );
#endif
   if (!buffer) {
      G->LOG_MISC.Put(G->Lang( 0x223, "sd", where, size));
#ifdef DEBUG_MEM_PARANOID
      exit (1);
#endif
   }
#ifdef EBUG_MEM
   if (lasttable >= tablesize) {
      if (this != G->memtype_list) {
	 int size = (tablesize * 2) * sizeof(struct memtype_internal_struct);
	 struct memtype_internal_struct * newone;
	 
	 CONTEXT(key);
	 newone = (struct memtype_internal_struct *)my_malloc(size,G->memtype_list);
	 memcpy(newone, data, size / 2);
	 my_free(data, G->memtype_list);
	 data = newone;
	 tablesize *= 2;
      } else 
	lasttable--; /* yes this will cause 'leaks' */
   }
   strncpy(data[lasttable].where, where, 59);
   data[lasttable].where[59] = 0;
   data[lasttable].size = size;
   data[lasttable].ptr = buffer;
   lasttable ++;
#else
   *buffer = size;
   buffer++;
#endif
   expmem += size;
   total_memory += size;
   memset(buffer, 0, size);
   return buffer;
}

void MemType::Status ( Socket * sock ) {
   FUNCTION;
   int expmem = 0, countmem = 0;
#ifdef EBUG_MEM
   int tallymem = 0, i;
#endif
   MemType * work;
   
   for (work = this; work; work = work->Next()) {
      countmem += work->expmem;
      expmem += work->ptr ? (*work->ptr)->ExpMem () : work->expmem;
#ifdef EBUG_MEM
      for (i = 0; i < work->lasttable; i++)
	tallymem += work->data[i].size;
#endif
   }
#ifdef EBUG_MEM
   sock->Send(G->Lang( 0x225, "dd", total_memory, countmem));
   sock->Send(G->Lang( 0x226, "dd", expmem, tallymem));
#else
   sock->Send(G->Lang( 0x227, "ddd", total_memory, countmem, expmem));
#endif
}

void MemType::Debug ( Socket * sock ) {
#ifdef EBUG_MEM
   int tallymem =0, i;
#endif
   int expmem = 0, thisexp = 0, countmem = 0;
   MemType * work;
   
   for (work = this; work; work = work->Next()) {
      int tally = 0;
     
       thisexp = work->ptr ? (*work->ptr)->ExpMem() : work->expmem;
#ifdef EBUG_MEM
      for (i = 0; i < work->lasttable; i++)
	tally += work->data[i].size;
#endif
      sock->Send(G->Lang( 0x228, "s 77744", work->key,
		      12 - strlen(work->key), work->expmem, thisexp, 
		      tally, work->lasttable, work->tablesize));
      expmem += work->expmem;
      countmem += thisexp;
#ifdef EBUG_MEM
      tallymem += tally;
#endif
      if ((work->expmem != thisexp)
#ifdef EBUG_MEM
	  || (work->expmem != tally) || (thisexp != tally)
#endif
	  ) {
	 sock->Send(G->Lang( 0x229, ""));
#ifdef EBUG_MEM
	 for (i = 0; i < work->lasttable; i++) 
	   sock->Send(G->Lang( 0x22A, "s 5X", work->data[i].where,
			   60 - strlen(work->data[i].where),
			   work->data[i].size, work->data[i].ptr));
#endif
      }
   }
#ifdef EBUG_MEM
   sock->Send(G->Lang( 0x22B, "777", expmem, countmem, tallymem));
#else
   sock->Send(G->Lang( 0x22B, "777", expmem, countmem, 0));
#endif
   sock->Send(G->Lang( 0x22C, "7", total_memory));
}

void MemType::Dump (Socket * sock, char * which) {
   FUNCTION;
   MemType * work = (MemType *)Find(which);
   
   if (!work) 
     sock->Send(G->Lang(0x295, ""));
   else {
      int thisexp = work->ptr ? (*work->ptr)->ExpMem() : work->expmem;
      int tally = 0;
      
#ifdef EBUG_MEM
      int i;

      for (i = 0; i < work->lasttable; i++)
	tally += work->data[i].size;
#endif
      sock->Send(G->Lang( 0x228, "s 66644", work->key,
			 12 - strlen(work->key), work->expmem, thisexp, 
			 tally, work->lasttable, work->tablesize));
#ifdef EBUG_MEM
      for (i = 0; i < work->lasttable; i++) 
	sock->Send(G->Lang( 0x22A, "s 5X", work->data[i].where,
			   60 - strlen(work->data[i].where),
			   work->data[i].size, work->data[i].ptr));
#endif
   }
}

int MemType::ExpectedMemory ( void ) {
   FUNCTION;
   MemType * ptr;
   int expmem = 0;
   
   for (ptr = this; ptr; ptr = ptr->Next())
     expmem += (ptr->ptr ? (*ptr->ptr)->ExpMem() : ptr->expmem);
   return expmem;
}
