/*
 * ModuleEntry.C - class members for ModuleEntry class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryEggdrop.h"
#include "Socket.h"
#include "Dependancy.h"
#include "StaticMods.h"
#include "BindTable.h"
#include "HookList.h"
#include "HelpReference.h"
#include "UserEntry.h"
#include "ModuleLanguage.h"
#include "LanguageEntry.h"
#include "Command.h"
#include "Variable.h"
#define CLASS "ModuleEntry"

#ifdef HAVE_UNAME
#include <sys/utsname.h>
#endif
#ifndef STATIC_ONLY 
#ifdef HPUX_HACKS
#include <dl.h>
#else
#ifdef OSF1_HACKS
#include <loader.h>
#else
#if DLOPEN_1
char *dlerror();
void *dlopen (const char *, int);
int dlclose (void *);
void * dlsym (void *,char *);
#define DLFLAGS 1
#else
#include <dlfcn.h>
#ifndef RTLD_GLOBAL
#define RTLD_GLOBAL 0
#endif
#ifndef RTLD_NOW
#define RTLD_NOW 1
#endif
#ifdef RTLD_LAZY
#define DLFLAGS RTLD_LAZY|RTLD_GLOBAL
#else
#define DLFLAGS RTLD_NOW|RTLD_GLOBAL
#endif
#endif /* DLOPEN_1 */
#endif /* OSF1_HACKS */
#endif /* HPUX_HACKS */

#endif /* STATIC_ONLY */

int ModuleEntry::major_number = 0;
static char v[VERSION_LEN + 1], fv[PATCH_LEN + 1];

ModuleEntry::ModuleEntry ( ModuleEntry & o, const char * name )
: KeyedList<MISCLEN> ( o, name, 0 ) {
   G = (ModuleEntryEggdrop *)this;
   strcpy(version = v, name);
   full_version = fv;
}

#ifdef HPUX_HACKS
ModuleEntry::ModuleEntry ( const char * name, shl_t f )
#else
ModuleEntry::ModuleEntry ( const char * name, void * f )
#endif
: KeyedList<MISCLEN> ( *G->module_list, name, KeyedList<MISCLEN>::APPEND ) {
   FUNCTION;
   hand = f;
   language = G->langtab_list->Module(name);
   stat = (f == 0);
}

ModuleEntry::ModuleEntry (const char * name, Version * v,
			  const struct patches * patches, ModuleEntry * o)
: KeyedList<MISCLEN> (), ver ( v ) {
   FUNCTION;
   char verwork[VERSION_LEN + 1], fullver[PATCH_LEN + 1];

   context;
   if (!v || (v->Major() == major_number)) {
      hand = o ? o->hand : 0;
      stat = o ? o->stat : 0;
      language = G->langtab_list->Module(name);
      KeyedList_Cons( *G->module_list, name, KeyedList<MISCLEN>::NODUPS
		     | KeyedList<MISCLEN>::APPEND);
      strcpy(verwork, key);
      Patches(v,fullver,verwork,patches);
      version = (char *)my_malloc(strlen(verwork)+1,G->memtype_modules);
      strcpy(version,verwork);
      full_version = (char *)my_malloc(strlen(fullver)+1,G->memtype_modules);
      strcpy(full_version,fullver);
   }
}

ModuleEntry::~ModuleEntry ( void ) {
   FUNCTION;
   Dependancy *d = G->dep_list, *n = d->Next();

   if (version)
     my_free(version, G->memtype_modules);
   if (full_version)
     my_free(full_version, G->memtype_modules);
   context;
   while ((d = n)) {
      n = d->Next();
      if (d->Needing() == this) 
	DELETE ( d );
   }
   context;
   G->module_list->Remove(this);
}

int ModuleEntry::ExpMem ( void ) {
   FUNCTION;
   int tot = 0;
   
   if (!key[0]) {
      ModuleEntry * ptr = this;
      
      if ((ptr = ptr->Next())) /* skip eggdrop (static) module */
	while ((ptr = ptr->Next()))
	  tot += ptr->ExpMem ();
   } else {
      if (version)
	tot += strlen(version) + 1;
      if (full_version)
	tot += strlen(full_version) + 1;
   }
   return tot;
}

const char * ModuleEntry::CheckDepend (const char * name) {
   FUNCTION;
   FILE *f;
   int l;
   char *modname;
   char workbuf[DIRLEN + 1], * buf;

   if (G->moddir[0] != '/') {
      if (!getcwd(workbuf, DIRLEN))
	 return G->Lang( 0x0F7, "" );
      l = strlen(workbuf);
      G->SPrintF(workbuf + l, DIRLEN - l, "/%s%s.so.dep", G->moddir, name);
   } else
      G->SPrintF(workbuf, DIRLEN, "%s%s.so.dep", G->moddir, name);

   if ((f = fopen(workbuf, "r"))) {
      while (!feof(f)) {
	 fgets(workbuf, sizeof(workbuf) - 1, f);
	 if (feof(f))
	   break;
	 l = strlen(workbuf);
	 if (workbuf[l] == '\n')
	   workbuf[l] = 0;
	 buf = workbuf;
	 modname = G->Split(&buf);
	 G->RMSpace(modname);
	 if (modname[0]) {
	    Load(modname);
	    /* I think this is sufficient for the moment */
	    if (!Find(modname, &G->AnyVer)) {
	       fclose(f);
	       return G->Lang( 0x0CE, "s", modname );
	    }
	 }
      }
      fclose(f);
   }
   return 0;
}

const char * ModuleEntry::Load (const char * name) {
   FUNCTION;
   ModuleEntry * p;
   char *e;
   char * (*f)(void *, ModuleEntry *) = 0;
   void * args;
#ifndef STATIC_ONLY
   int l;
   char workbuf[DIRLEN + 1];
# ifdef HPUX_HACKS
   shl_t hand;
# else
#  ifdef OSF1_HACKS
   ldr_module_t hand;
#  else
   void *hand = 0;
#  endif
# endif
#else
   void *hand = 0;
#endif
   StaticMods * sl;
   
   CONTEXT(name);
   if (Find(name, &G->AnyVer))
      return G->Lang( 0x0F5, "");
   if ((sl = (StaticMods *)G->statics->Find(name))) {
      f = (char *(*)(void *,ModuleEntry *))sl->Func();
      hand = 0;
#ifndef STATIC_ONLY
   } else if (G->module_list->CheckDepend(name)) 
	return G->module_list->CheckDepend(name);
   else {
      if (G->moddir[0] != '/') {
	 if (!getcwd(workbuf, DIRLEN))
	   return G->Lang( 0x0F7, "");
	 l = strlen(workbuf);
	 G->SPrintF(workbuf + l, DIRLEN - l, "/%s%s.so", G->moddir, name);
      } else
	G->SPrintF(workbuf, DIRLEN, "%s%s.so", G->moddir, name);
# ifdef HPUX_HACKS
      if (!(hand = shl_load(workbuf, BIND_IMMEDIATE, 0L)))
	return G->Lang( 0x0F6, "");
# else
#  ifdef OSF1_HACKS
      if ((hand = (Tcl_PackageInitProc *) load (workbuf, LDR_NOFLAGS))
	  == LDR_NULL_MODULE)
	return G->Lang( 0x0F6, "");
#  else
      if (!(hand = dlopen(workbuf, DLFLAGS))) 
	return dlerror(); 
#  endif
# endif
      G->SPrintF(workbuf, DIRLEN, "%s_start", name);
# ifdef HPUX_HACKS
      if (shl_findsym(&hand, workbuf, (short) TYPE_PROCEDURE, (void *)&f))
	f = NULL;
# else
#  ifdef OSF1_HACKS
      f = ldr_lookup_package(hand, workbuf);
#  else
      f = (char * (*)(void *,ModuleEntry *))dlsym(hand, workbuf);
#  endif
# endif
      if (!f) {		/* some OS's need the _ */
	 G->SPrintF(workbuf, DIRLEN, "_%s_start", name);
# ifdef HPUX_HACKS
	 if (shl_findsym(&hand, workbuf, (short) TYPE_PROCEDURE, (void *)&f))
	   f = NULL;
# else
#  ifdef OSF1_HACKS
	 f = ldr_lookup_package(hand, workbuf);
#  else
	 f = (char * (*)(void *,ModuleEntry *))dlsym(hand, workbuf);
#  endif
# endif
	 if (!f) {
# ifdef HPUX_HACKS
	    shl_unload(hand);
# else
#  ifdef OSF1_HACKS
#  else
	    dlclose(hand);
#  endif
# endif
	 }
      }
#endif
   }
   if (!f)
     return G->Lang( 0x0F8, "");
#ifdef HPUX_HACKS
   if (!(p = NEW ( ModuleEntry ( name, (shl_t)hand ) ) ) ) 
#else
   if (!(p = NEW ( ModuleEntry ( name, (void *)hand ) ) ) ) 
#endif
     return G->Lang( 0x0FE, "");
   if ((e = f(G,p))) {
      CONTEXT(e);
      DELETE ( G->module_list->Find(name) );
      return e;
   }
   args = p = G->module_list->Find(name);
   G->bt_load->Check(name, 0, &args);
   G->LOG_MISC.Put(G->Lang( 0x0CF, "s", name));
   context;
   return NULL;
}

const char * ModuleEntry::Unload (const char * name) {
   FUNCTION;
   ModuleEntry * p = (ModuleEntry *)G->module_list->Find(name);

   context;
   if (p) {
      Dependancy *d = G->dep_list;
      void * args = p;
      
      while ((d = d->Next()))
	if (d->Needed() == p) 
	  return G->Lang( 0x0F9, "");
      G->bt_unld->Check(name, 0, &args);
      args = p->hand;
      DELETE ( p );
      G->hook_list->Cleanup(p);
      G->memtype_list->Cleanup(p);
      G->help_list->Cleanup(p);
      G->entrytype_list->Cleanup(p);
      G->bind_list->Cleanup(p);
      G->command_list->Cleanup(p);
      G->variable_list->Cleanup(p);
      G->LOG_MISC.Put(G->Lang( 0x0CB, "s", p->Key()));
#ifndef STATIC_ONLY
# ifdef HPUX_HACKS
      shl_unload((shl_t)args);
# else
#  ifdef OSF1_HACKS
#  else
      dlclose((char *)name);
#  endif
# endif
#endif /* STATIC */
      return 0;
   }
   return G->Lang( 0x0FB, "");
}

ModuleEntry * ModuleEntry::Find (const char * name, Version * ver) {
   FUNCTION;
   ModuleEntry * p = (ModuleEntry *)KeyedList<MISCLEN>::Find ( name );
   
   if (!p || !(*ver <= *p->Vers()))
     p = 0;
   return p;
}

int ModuleEntry::Rename (const char * newname) {
   FUNCTION;
   ModuleEntry * p = (ModuleEntry *)G->module_list->KeyedList<MISCLEN>::Find(newname);
   
   if (!p) 
     SetKey(newname);
   return p == 0;
}
 
ModuleEntry * ModuleEntry::Depend (const char * name2, Version * ver) {
   FUNCTION;
   ModuleEntry *p;

   context;
   if (ver->Major() == major_number) {
      if (!(p = G->module_list->Find(name2, ver))) 
	if (G->module_list->Load(name2) 
	    || !(p = G->module_list->Find(name2, ver)))
	  return 0;
      if (p && this) {
	 NEW ( Dependancy ( this, p, ver ) );
	 context;
	 return p;
      }
   }
   return 0;
}

void ModuleEntry::Report (Socket * idx,int details,char * which) {
   FUNCTION;
   ModuleEntry *p = this;
   Dependancy *d;
   
   if (!which && details)
     idx->Send(G->Lang( 0x120, ""));
   while ((p = p->Next()))
     if (!which || !strcasecmp(which,p->Key())) {
	idx->Send(G->Lang( 0x121, "ss", p->Key(), p->VersionStr()));
	if (details) {
	   if (details > 1) {
	      idx->Send(G->Lang( 0x122, "s", p->FullVersion()));
	      d = G->dep_list;
	      while ((d = d->Next()))
		if (d->Needing() == p) 
		  idx->Send(G->Lang( 0x123, "sddd", d->Needed()->Key(),
				 d->Major(), d->Minor(), d->Fix()));
	   }
	   p->Report(idx, details);
	   if (which)
	     return;
	}
     }
   if (which)
     idx->Send(G->Lang( 0x0FB, ""));
}

/* make a cycle of 16 buffers, prevents consecutive calls overlapping */
static char WORKAREA[16][512];
char * ModuleEntry::Lang (int idx, char * args, ... ) {
   FUNCTION;
   LanguageEntry * lt;
   static int which = 0;
   char * work_area = WORKAREA[which = (which + 1) & 15];
   const char * read;
   
   if (language && (lt = (LanguageEntry *)language->Entry(idx)->Find(idx))) {
      int len = 0, argc = 0;
      struct {
	 int i;
	 char c;
	 char * s;
	 void * p;
	 unsigned long l;
	 double f;
      } argv[10];
      char * towrite, sub[81];
#ifdef HAVE_UNAME
      struct utsname uname_info;
#endif	   
      
      if (args) {
	 va_list valist;
	 va_start(valist,args);
	 for (argc = 0; args[argc] && argc < 10; argc++) 
	   switch (args[argc]) {
	     case 's':
	      argv[argc].s = va_arg(valist,char *);
	      break;
	     case 'c':
	      argv[argc].c = va_arg(valist,int);
	      break;
	     case ' ':
	     case 'u':
	     case 'd':
	     case 'D':
	     case '2':
	     case '3':
	     case '4':
	     case '5':
	     case '6':
	     case '7':
	     case '8':
	     case '9':
	      argv[argc].i = va_arg(valist,int);
	      break;
	     case 'X':
	     case 'x':
	      argv[argc].p = va_arg(valist,void *);
	      break;
	     case 'l':
	      argv[argc].l = va_arg(valist,long);
	      break;
	     case 'f':
	     case 'F':
	      argv[argc].f = va_arg(valist,double);
	      break;
	   }
	 va_end(valist);
      }
      for (read = lt->Text(); *read && (len < 510); read++) {
	 if (*read == '%') {
	    read++;
	    towrite = NULL;
	    switch (*read) {
	      case 'V':
	       towrite = G->version;
	       break;
	      case 'A':
	       towrite = G->admin;
	       break;
	      case 'T':
	       len += strftime(work_area+len, 510 - len,"%H:%M %Z",
			       localtime(&G->now));
	       break;
	      case 'u':
	       len += strftime(work_area + len, 510 - len, 
			       "%a %b %d %Y at %H:%M:%S %Z",
			       localtime(&G->online_since));
	       break;
	      case 'U':
#ifdef HAVE_UNAME
	       if (!uname(&uname_info)) {
		  G->SPrintF(sub, 80, "%s %s",uname_info.sysname,
			     uname_info.release);
		  towrite = sub;
	       } else
#endif
		 towrite = "*UNKNOWN*";
	      case 'B':
	       towrite = G->botnetnick;
	       break;
	      case 'O':
	       towrite = G->origbotname;
	       break;
	      case 'b':
	       towrite = "\033\002";
	       break;
	      case '_':
	       towrite = "\033\037";
	       break;
	      case 'v':
	       towrite = "\033\022";
	       break;
	      case 'f':
	       towrite = "\033\033";
	       break;
	      case 'C':
	       G->SPrintF(sub, 80, "\033%c", *++read);
	       towrite = sub;
	       break;
	      default:
	       if ((*read >= '0') && (*read <= '9')) {
		  int idx = *read - '0', i = 0;
		  
		  if (idx < argc) {
		     switch(args[idx]) {
		       case 's':
			towrite = argv[idx].s;
			break;
		       case 'c':
			work_area[len++] = argv[idx].c;
			break;
		       case 'd':
			if (argv[idx].i < 0) {
			   sub[0] = '-';
			   strcpy(sub,G->ToBase10(-argv[idx].i));
			   towrite = sub;
			   break;
			}
		       case 'u':
			towrite = G->ToBase10(argv[idx].i);
			break;
		       case 'D':
			towrite = G->ToBase64(argv[idx].i);
			break;
		       case '9':
		       case '8':
		       case '7':
		       case '6': 
		       case '5':
		       case '4':
		       case '3':
		       case '2': {
			  int j = argv[idx].i;
			  
			  sub[i = args[idx] - '0'] = 0;
			  if (!j)
			    sub[--i] = '0';
			  while (i && j) {
			     sub[--i] = '0' + j % 10;
			     j /= 10;
			  }
			  while (i)
			    sub[--i] = ' ';
			  towrite = sub;
		       }
			break;
		       case 'X': {
			  long l = (long)argv[idx].p;
			  i = sizeof(long) * 2 + 1;
			  sub[i--] = 0;
			  while (i) {
			     sub[i--] = '0' + (l & 0xf) + 
			       (((l & 0xf) > 9) ? 'A' - '9' - 1 : 0);
			     l = l >> 4;
			  }
			  towrite = sub + i + 1;
		       }
			break;
		       case 'x': {
			  long l = (long)argv[idx].p;
			  i = sizeof(long) * 2;
			  sub[i] = 0;
			  while (l) {
			     sub[--i] = '0' + (l & 0xf) + 
			       (((l & 0xf) > 9) ? 'A' - '9' - 1 : 0);
			     l = l >> 4;
			  }
			  towrite = sub + i;
		       }
			break;
		       case 'l':
			towrite = G->ToBase10(argv[idx].l);
			break;
		       case 'f':
			sprintf(sub,"%f",argv[idx].f);
			towrite = sub;
			break;
		       case 'F':
			sprintf(sub,"%.1f",argv[idx].f);
			towrite = sub;
			break;
		       case ' ': {
			  i = argv[idx].i;
			  while ((i > 0) && (len < 510)) {
			     work_area[len++] = ' ';
			     i--;
			  }
			  break;
		       }
		     }
		  }
	       } else
		 work_area[len++] = *read;
	    }
	    if (towrite) 
	      while (*towrite && (len < 510)) 
		work_area[len++] = *towrite++;
	 } else
	   work_area[len++] = *read;
      }
      work_area[len] = 0;
      /* process */
      return work_area;
   } 
   sprintf(work_area,"MSG:%s[%03x]",language ? language->Key() : key,idx);
   return work_area;
}

void ModuleEntry::Report ( Socket *, int) {
   FUNCTION;
}


void ModuleEntry::Patches (Version * numver, char * version, char * ver,
			   const struct patches * patches) {
   FUNCTION;
   char patchname[20][16];
   int highest[20], lastpatch = 0, i, j, k;
   
   for (i = 0; i < 20; i++) {
      patchname[i][0] = 0;
      highest[i] = 0;
   }
   if (patches)
     for (i = 0; patches[i].name; i++)
       numver->AddPatch();
   lastpatch = G->SPrintF (version, PATCH_LEN, "v%d.%d.%d %2%2%2%2 ",
			   numver->Major(), numver->Minor(), numver->Fix(),
			   numver->Major(), numver->Minor(), numver->Fix(),
			   numver->Patch());
   if (patches)
     for (i = 0; patches[i].name; i++) {
	k = (patches[i].number < 1) ? 1 : patches[i].number;
	for (j = 0; j < 20; j++)
	  if (!highest[j]) {
	     strncpy(patchname[j],patches[i].name,15); 
	     patchname[j][15]=0; 
	     highest[j] = k; 
	     break; 
	  } else if (!strcasecmp(patches[i].name,patchname[j])) {
	     if (k > highest[j])
	       highest[j] = k;
	     break; 
	  } 
	lastpatch += G->SPrintF(version + lastpatch,
				PATCH_LEN-lastpatch,"%s%d ",patches[i].name,k);
     }
   lastpatch = strlen(ver);
   lastpatch += G->SPrintF(ver + lastpatch, VERSION_LEN - lastpatch,
			   " v%d.%d.%d", numver->Major(), numver->Minor(),
			   numver->Fix());
   for (i = 0; (i < 20) && highest[i] && 
	((lastpatch + strlen(patchname[i]) + 3) < VERSION_LEN); i++)
     lastpatch += G->SPrintF (ver + lastpatch, VERSION_LEN - lastpatch,
			      "+%s%d", patchname[i], highest[i]);
}
