/*
 * ModuleEntry.h - class definition 
 */

#ifndef _KEYEDLIST_
#include "KeyedList.h"
#endif

#ifndef _VERSION_
#include "Version.h"
#endif

#ifdef HPUX_HACKS
#include <dl.h>
#endif

#ifndef _MODULEENTRY_
#define _MODULEENTRY_

class ModuleEntry : public KeyedList<MISCLEN> {
  protected:
   Version ver;
   char * version;
   char * full_version;
#ifdef HPUX_HACKS
   shl_t hand;
#else
   void * hand;           /* module handle */
#endif
   ModuleLanguage * language;
   int stat;
   
   static int major_number;
   friend class ModuleLanguage;
   friend int main ( int, char ** );
  public:
   inline ModuleEntry ( void );
   ModuleEntry ( ModuleEntry &, const char * );
#ifdef HPUX_HACKS
   ModuleEntry ( const char *, shl_t );
#else
   ModuleEntry ( const char *, void * );
#endif
   ModuleEntry ( const char *, Version * v = 0, const struct patches * p = 0,
		ModuleEntry * m = 0);
   ~ModuleEntry ( void );
   
   int Rename (const char *);
   ModuleEntry * Find ( const char *, Version *);
   ModuleEntry * Depend ( const char *, Version *);
   
   int ExpMem ( void );
   const char * CheckDepend ( const char * );
   const char * Load ( const char * );
   const char * Unload ( const char * );
   void Report (Socket *,int, char *);
   char * Lang (int, char *, ...);
   virtual void Report ( Socket *, int );
   void Patches (Version *, char *, char *, const struct patches *);
   
   inline int IsStatic ( void ) { return stat; }
   inline ModuleLanguage * Language ( void ) { return language; }
   inline Version * Vers ( void ) { return &ver; }
   inline const char * VersionStr ( void ) { return version; }
   inline const char * FullVersion ( void ) { return full_version; }
   inline ModuleEntry * Find ( const char * x ) {
      return (ModuleEntry *)KeyedList<MISCLEN>::Find(x);
   }
   DEF_NEXT(ModuleEntry);
   inline void * operator new ( size_t , int);
   inline void operator delete ( void * );
};

#ifndef MAKING_MODS
inline ModuleEntry::ModuleEntry ( void ) 
: KeyedList<MISCLEN> () {
   major_number = 2; 
}
#endif
#endif
