/*
 * ModuleEntryEggdrop.C - core functions
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryEggdrop.h"
#include "DccChat.h"
#include "LogFile.h"
#include "UserRecord.h"
#include "HelpReference.h"
#include "ModuleLanguage.h"
#include "ModuleEntry.h"
#include "HostnameCache.h"
#include "FlagRecord.h"
#include "BindTable.h"
#include "Channel.h"
#include "HookList.h"
#include "ScriptContext.h"
#include <arpa/inet.h>
#include <fcntl.h>
#include <ctype.h>
#ifdef HAVE_UNAME
#include <sys/utsname.h>
#endif
#define CLASS "ModuleEntryEggdrop"

#ifndef HAVE_VSNPRINTF
static int vsnprintf ( char * buf, size_t len, const char * format, va_list valist ) {
   char xxx[8192]; /* if this is not enough, woi! */
   int outlen;
   
   outlen = vsprintf(xxx, format, valist);
   if (outlen > len) {
      strncpy(buf, xxx, len);
      outlen = -1;
   }
   return outlen;
}
#endif

#ifndef HAVE_MEMSET
static void * memset ( void * buf, int c, size_t s) {
   FUNCTION;
   while (s)
     ((char *)buf)[--s] = c;
   return buf;
}
#endif

#if !HAVE_STRCASECMP
#define upcase(c) (((c)>='a' && (c)<='z') ? (c)-'a'+'A' : (c))
/* unixware has no strcasecmp() without linking in a hefty library */
static int strcasecmp (char * s1, char * s2) {
   FUNCTION;
   while (*s1 && *s2 && (upcase(*s1) == upcase(*s2))) {
      s1++;
      s2++;
   }
   return upcase(*s1) - upcase(*s2);
}
#endif

ModuleEntryEggdrop::ModuleEntryEggdrop ( ModuleEntry & o ) 
: ModuleEntry (o,"eggdrop") {
   int i;
   
   Log = Stdout = Stderr = 0;
   G = this;
   module_list = &o;
   /* initialise context list */
   for (i = 0; i < 16; i++)
     context;
   vsnprintf = ::vsnprintf;
   memset = ::memset;
   strcasecmp = ::strcasecmp;
   AnyVer.Set( 2, -1 );
   ServerVer.Set(2);
   IrcVer.Set(2);
   LOG_CMDS = LogModes::CMDS;
   LOG_MISC = LogModes::MISC;
   LOG_SOCKET = LogModes::SOCKET;
   conmask = LogModes::MODES | LogModes::MISC | LogModes::CMDS;
   DCC_CHAT_PASS = "DCC_CHAT_PASS";
   DCC_CHAT = "DCC_CHAT";
   DCC_CHAT_LOGIN = "DCC_CHAT_LOGIN";
   DCC_CHAT_NEWUSER = "DCC_CHAT_NEWUSER";
   DCC_CHAT_NEWPASS = "DCC_CHAT_NEWPASS";
   TELNET_PORT = "TELNET_PORT";
   IDENT_WAIT = "IDENT_WAIT";
   IDENT_LOOKUP = "IDENT_LOOKUP";
   SPECIAL = "SPECIAL";
   entrytype_info = entrytype_pass = entrytype_hosts = entrytype_laston = 0;
   entrytype_laston = entrytype_botfl = entrytype_comment = 0;
   dcc_total = gban_total =  make_userfile = default_flags = 0;
   con_chan = term_z = cache_hit = cache_miss = dcc_total = do_restart = 0;
   use_console_r = reserved_port = debug_output = noshare = 0;
   protect_readonly = allow_new_telnets = keep_all_logs = 0;
   backgrd = share_greet = require_p = protect_telnet = 1;
   check_chanlist = 0;
   allow_dk_cmds = shtime = use_stderr = 1 ;
   port_range = dcc_flood_thr = 3;
   flood_telnet_thr = 5;
   ident_timeout = resolve_timeout = 10;
   connect_timeout = 15;
   flood_telnet_time = 30;
   max_dcc = hostcache_count = 50;
   ignore_time = password_timeout = 60;
   firewallport = 178;
   hostcache_age = 240;
   max_sendq = 1000;
   strcpy(moddir, "modules/");
   long_version[0] = 0;
   strcpy(tempdir,"/tmp/");
   notify_new[0] = 0;
   strcpy(helpdir, "help/");
   natip[0] = 0;
   origbotname[0] = 0;
   botnetnick[0] = 0;
   botuser[0] = 0;
   admin[0] = 0;
   strcpy(userfile, "userfile");
   myip[0] = 0;
   strcpy(motdfile, "motd");
   whois_fields[0] = 0;
   owner[0] = 0;
   pid_file[0] = 0;
   firewall[0] = 0;
   strcpy(configfile,"egg.config");
   savedconfigfile[0] = 0;
   module_encrypt = 0;
   shareout = 0;
};

ModuleEntryEggdrop::~ModuleEntryEggdrop ( void ) {
   FUNCTION;
   Fatal ("CORE MODULE DELETED!!");
}

/* thank you ircu :) */
static char tobase64array[64] = {
     'A','B','C','D','E','F','G','H','I','J','K','L','M',
       'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
       'a','b','c','d','e','f','g','h','i','j','k','l','m',
       'n','o','p','q','r','s','t','u','v','w','x','y','z',
       '0','1','2','3','4','5','6','7','8','9',
       '[',']'
};

char * ModuleEntryEggdrop::ToBase64 (unsigned int val) {
   FUNCTION;
   static char buf_base64[12];
   int i = 11;
   
   buf_base64[11] = 0;
   if (!val) {
      buf_base64[10] = 'A';
      return buf_base64+10;
   }
   while (val) {
      i--;
      buf_base64[i] = tobase64array[val & 0x3f];
      val = val >> 6;
   }
   return buf_base64+i;
}

char * ModuleEntryEggdrop::ToBase10 (unsigned long val) {
   FUNCTION;
   static char buf_base10[32];
   int i = 31;
   
   buf_base10[31] = 0;
   if (!val) {
      buf_base10[30] = '0';
      return buf_base10+30;
   }
   while (val) {
      i--;
      buf_base10[i] = '0' + (val % 10);
      val /= 10;
   }
   return buf_base10+i;
}

static char base64to[256] = {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0,52,53,54,55,56,57,58,59,60,61, 0, 0,
     0, 0, 0, 0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,
     15,16,17,18,19,20,21,22,23,24,25,62, 0,63, 0, 0, 0,26,27,28,
     29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,
     49,50,51, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

int ModuleEntryEggdrop::Base64To ( char * buf ) {
   FUNCTION;
   int i = 0;
   while (*buf) {
      i = i << 6;
      i += base64to[(int)*buf];
      buf++;
   }
   return i;
}

void ModuleEntryEggdrop::MemCpy (char * dest, char * src, int len) {
   FUNCTION;
   while (len--)
      *dest++ = *src++;
}

struct in_addr * ModuleEntryEggdrop::GetMyIp(void) {
   FUNCTION;
   char s[121];
   static struct in_addr in;
   struct in_addr * ptr = 0;

#ifdef HPUX_HACKS
   if (myip[0] && (myip[strlen(myip) - 1] >= '0') &&
       (myip[strlen(myip) - 1] <= '9'))
#else
   if (myip[0] && (myip[strlen(myip) - 1] >= '0') &&
       (myip[strlen(myip) - 1] <= '9')  && !inet_aton(myip, &in))
#endif
     return &in;
   if (myip[0])
     ptr = host_cache->Find(myip);
   if (!ptr) {
      gethostname(s, 120);
      ptr = host_cache->Find(s);
   }
   if (!ptr)
     Fatal(Lang(0x130, ""));
   return ptr;
}

/***** MISC FUNCTIONS *****/
int ModuleEntryEggdrop::StrCpy(char * a, const char * b) {
   FUNCTION;
   const char * c = b;
   while (*b) 
     *a++ = *b++;
   *a = *b;
   return b-c;
}

/* split first word off of rest and put it in first */
void ModuleEntryEggdrop::SplitC (char * first, char * rest, char divider) {
   FUNCTION;
   char *p;
   
   if (!(p = strchr(rest, divider))) {
      if ((first != rest) && first)
	first[0] = 0;
   } else {
      *p = 0;
      if (first)
	strcpy(first, rest);
      if (first != rest)
	strcpy(rest, p + 1);
   }
}

char * ModuleEntryEggdrop::SplitNick ( char ** blah ) {
   FUNCTION;
   char * p = strchr(*blah,'!'), *q = *blah;
   if (p) {
      *p = 0;
      *blah = p + 1;
      return q;
   }
   return "";
}

char * ModuleEntryEggdrop::Split ( char ** rest ) {
   FUNCTION;
   register char * o, * r;
   
   if (!rest) 
     return *rest = "";
   for (o = *rest; *o == ' '; o++);
   for (r = o; *o && (*o != ' '); o++);
   if (*o)
     *o++ = 0;
   *rest = o;
   return r;
}

/* convert "abc!user@a.b.host" into "*!user@*.b.host"
   or "abc!user@1.2.3.4" into "*!user@1.2.3.*"  */
void ModuleEntryEggdrop::MaskHost (const char * s, char * nw) {
   FUNCTION;
   const char * e = 0, * f, * p, * q;
   int i;
   
   *nw++ = '*';
   *nw++ = '!';
   p = (q = strchr(s, '!')) ? q + 1 : s;
   /* strip of any nick, if a username is found, use last 8 chars */
   if ((q = strchr(p, '@'))) {
      if ((q - p) > 9) {
	 nw[0] = '*';
	 p = q - 7;
	 i = 1;
      } else 
	i = 0;
      while (*p != '@') {
	 if (*p < 32)
	   nw[i++] = '?';
	 else
	   nw[i++] = *p++;
      }
      nw[i++] = '@';
      q++;
   } else {
      nw[0] = '*';
      nw[1] = '@';
      i = 2;
      q = s;
   }
   nw += i;
   /* now q points to the hostname, i point to where to put the mask */
   if (!(p = strchr(q, '.')) || !(e = strchr(p + 1, '.')))
     /* TLD or 2 part host */
     strcpy(nw, q);
   else {
      for (f = e; *f; f++);
      f--;
      if ((*f >= '0') && (*f <= '9')) {  /* numeric IP address */
	 while (*f != '.')
	   f--;
	 strncpy(nw, q, f - q);
	 nw += (f - q);
	 strcpy(nw, ".*");
      } else {
	 const char * x = strchr(e + 1, '.');
	 
	 if (!x)
	   x = p;
	 else if (strchr(x + 1, '.'))
	   x = e;
	 else if (strlen(x) == 3)
	   x = p;
	 else
	   x = e;
	 *nw++ = '*';
	 strcpy(nw, x);
      }
   }
}

/* copy a file from one place to another (possibly erasing old copy) */
/* returns 0 if OK, 1 if can't open original file, 2 if can't open new */
/* file, 3 if original file isn't normal, 4 if ran out of disk space */
int ModuleEntryEggdrop::CopyFile (char * oldpath, char * newpath) {
   FUNCTION;
   int fi, fo, x;
   char buf[512];
   struct stat st;
   fi = open(oldpath, O_RDONLY, 0);
   if (fi < 0)
      return 1;
   fstat(fi, &st);
   if (!(st.st_mode & S_IFREG))
      return 3;
   if ((fo = creat(newpath, (int) (st.st_mode & 0777))) < 0) {
      close(fi);
      return 2;
   }
   for (x = 1; x > 0;) {
      x = read(fi, buf, 512);
      if (x > 0) {
	 if (write(fo, buf, x) < x) {	/* couldn't write */
	    close(fo);
	    close(fi);
	    unlink(newpath);
	    return 4;
	 }
      }
   }
   close(fo);
   close(fi);
   return 0;
}

int ModuleEntryEggdrop::MoveFile (char * oldpath, char * newpath) {
   FUNCTION;
   int x = CopyFile(oldpath, newpath);
   if (!x)
      unlink(oldpath);
   return x;
}

/* convert an interval (in seconds) to one of: */
/* "19 days ago", "1 day ago", "18:12" */
void ModuleEntryEggdrop::DaysAgo (time_t now, time_t then, char * out) {
   FUNCTION;
   if (now - then > 86400) {
      int days = (now - then) / 86400;
      strncpy(out, Lang( (days == 1) ? 0x13F : 0x13E, "d", days), 40);
      out[40] = 0;
   } else {
      strncpy(out, ctime(&then) + 11, 6);
      out[5] = 0;
   }
}

/* convert an interval (in seconds) to one of: */
/* "in 19 days", "in 1 day", "at 18:12" */
void ModuleEntryEggdrop::Days (time_t now, time_t then, char * out) {
   FUNCTION;
   if (now - then > 86400) {
      int days = (now - then) / 86400;
      strncpy(out, Lang( (days == 1) ? 0x13D : 0x13C, "d", days), 80);
   } else {
      char s[81];
      strcpy(s, ctime(&now));
      s[16] = 0;
      strncpy(out, Lang( 0x13B, "s", s+11),80);
   }
   out[80] = 0;
}

/* convert an interval (in seconds) to one of: */
/* "for 19 days", "for 1 day", "for 09:10" */
void ModuleEntryEggdrop::DaysDur (time_t now, time_t then, char * out) {
   FUNCTION;
   int hrs, mins;
   if (now - then > 86400) {
      int days = (now - then) / 86400;
      strncpy(out, Lang( (days == 1) ? 0x13A : 0x139, "d", days), 80);
   } else {
      now -= then;
      hrs = (now / 3600);
      mins = (now % 3600) / 60;
      strncpy(out, Lang( 0x138, "22", hrs, mins),80);
   }
   out[80] = 0;
}


/********** STRING SUBSTITUTION **********/

void ModuleEntryEggdrop::SubstAddCol (char * s, int newcols, int size) {
   FUNCTION;
   static int cols = 0;
   static int colsofar = 0;
   static int colwidth = 0;
   static char *colstr = NULL;
   char * p;
   int l;
   
   if (!s) { 
      cols = 0;
      colwidth = 0;
      if (colstr)
	my_free(colstr, memtype_misc);
      colsofar = 0;
   } else if (size) {
      if (size < 30) 
	size = 30;
      if (colstr) {
	 strcpy(s, colstr);
	 my_free(colstr, memtype_misc);
	 colstr = NULL;
      } else
	s[0] = 0;
      cols = newcols;
      colsofar = 0;
      if (newcols)  {
	 colstr = (char *)my_malloc(6, memtype_misc);
	 strcpy(colstr, "     ");
	 colwidth = (size - 10) / newcols;
      } else 
	colwidth = 0;
   } else if (cols && s[0]) {
      l = (strlen(s) / colwidth) + 1;
      if ((colsofar + l) > cols) {
	 colsofar = l;
	 l *= colwidth;
	 p = (char *)my_malloc(l +  6, memtype_misc);
	 SPrintF(p, l + 5, "     %s% ", s, l - strlen(s));
	 strcpy(s, colstr);
      } else {
	 colsofar += l;
	 p = (char *)my_malloc((colsofar * colwidth) + 6, memtype_misc);
	 l *= colwidth;
	 SPrintF(p, (colsofar * colwidth) + 5, "%s%s% ", colstr, s, l - strlen(s));
	 s[0] = 0;
      }
      if (colstr)
	my_free(colstr, memtype_misc);
      colstr = p;
   }
}

/* substitute %x codes in help files */
/* %B = bot nickname */
/* %V = version */
/* %C = list of channels i monitor */
/* %E = eggdrop banner */
/* %A = admin line */
/* %T = current time ("14:15") */
/* %N = user's nickname */
/* %U = display system name if possible */
/* %{+xy}     require flags to read this section */
/* %{-}       turn of required flag matching only */
/* %{center}  center this line */
/* %{cols=N}  start of columnated section (indented) */
/* %{help=TOPIC} start a section for a particular command */
/* %{end}     end of section */
#define HELP_BOLD  1
#define HELP_REV   2
#define HELP_UNDER 4
#define HELP_FLASH 8

void ModuleEntryEggdrop::HelpSubst (char * s, const char * nick, 
				    FlagRecord * flags, int isdcc, 
				    const char * topic) {
   FUNCTION;
   char xx[HELP_BUF_LEN+1], sub[161], *current, *q, chr, *writeidx,
     *readidx;
   const char *towrite;
   Channel *chan;
   int i, j, center = 0;
   static int blind = 0, help_flags = 0;
#ifdef HAVE_UNAME
   struct utsname uname_info;
#endif
	  
   if (s == NULL) {
      /* used to reset substitutions */
      blind = 0;
      SubstAddCol(0, 0, 0);
      help_flags = isdcc;
      return;
   }
   strncpy(xx,s,HELP_BUF_LEN);
   xx[HELP_BUF_LEN]=0;
   readidx = xx;
   writeidx = s;
   current = strchr(readidx, '%');
   while (current) {
      /* are we about to copy a chuck to the end of the buffer? 
       * if so return */
      if ((writeidx + (current - readidx)) >= (s + HELP_BUF_LEN)) {
	 strncpy(writeidx,readidx,(s + HELP_BUF_LEN) - writeidx);
	 s[HELP_BUF_LEN] = 0;
	 return;
      }      
      chr = *(current + 1);
      *current = 0;
      if (!blind)    
	writeidx += StrCpy(writeidx, readidx);
      towrite = NULL;
      switch (chr) {
	case 'b':
	 if (help_flags & HS_HILITE) 
	   towrite = "\033\002";
	 break;
	case 'v':
	 if (help_flags & HS_HILITE)
	   towrite = "\033\026";
	 break;
	case '_':
	 if (help_flags & HS_HILITE) 
	   towrite = "\033\037";
	 break;
	case 'f':
	 if (help_flags & HS_HILITE)
	   towrite = "\033\033";
	 break;
	case 'U':
#ifdef HAVE_UNAME
	 if (!uname(&uname_info)) {
	    SPrintF(sub, 160, "%s %s",uname_info.sysname, uname_info.release);
	    towrite = sub;
	 } else
#endif
	   towrite = Lang( 0x137, "");
	 break;
	case 'B':
	 towrite = ((help_flags & HS_ISDCC) ? botnetnick : origbotname);
	 break;
	case 'V':
	 towrite = version;
	 break;
	case 'E':
	 towrite = full_version;
	 break;
	case 'A':
	 towrite = admin;
	 break;
	case 'T':
	 strcpy(sub, ctime(&now));
	 sub[16] = 0;
	 towrite = sub+11;
	 break;
	case 'N':
	 towrite = strchr(nick,':');
	 if (towrite)
	   towrite++;
	 else
	   towrite = nick;
	 break;
	case 'C':
	 if (!blind) 
	   for (chan = chanset; chan; chan = chan->Next()) {
	      if ((strlen(chan->Key()) + writeidx + 2) >=
		  (s + HELP_BUF_LEN)) {
		 strncpy(writeidx,chan->Key(),(s + HELP_BUF_LEN) - writeidx);
		 s[HELP_BUF_LEN] = 0;
		 return;
	      }
	      writeidx += StrCpy(writeidx,chan->Key());
	      if (chan->Next()) {
		 *writeidx++ = ',';
		 *writeidx++ = ' ';
	      }
	   }
	 break;
	case '{':
	 q = current;
	 current++;
	 while ((*current != '}') && (*current))
	   current++;
	 if (*current) {
	    *current = 0;
	    current--;
	    q += 2;
	    /* now q is the string and p is where the rest of the fcn expects */
	    if (!strncmp(q, "help=",5)) {
	       if (topic && strcasecmp(q+5,topic))
		 blind |= 2;
	       else
		 blind &= ~2;
	    } else if (!(blind & 2)) {
	       if (q[0] == '+') {
		  FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
		  fr.BreakDown(q+1);
		  if (!fr.Ok(*flags))
		    blind |= 1;
		  else 
		    blind &= ~1;
	       } else if (q[0] == '-') {
		  blind &= ~1;
	       } else if ((!strncasecmp(q, "fg=", 3) || 
			  !strncasecmp(q, "bg=",3)) &&
			  (help_flags & HS_HILITE)) {
		  if (!strcasecmp(q + 3, "black"))
		    i = '1';
		  else if (!strcasecmp(q + 3, "red"))
		    i = '2';
		  else if (!strcasecmp(q + 3, "green"))
		    i = '3';
		  else if (!strcasecmp(q + 3, "yellow"))
		    i = '4';
		  else if (!strcasecmp(q + 3, "blue"))
		    i = '5';
		  else if (!strcasecmp(q + 3, "magenta"))
		    i = '6';
		  else if (!strcasecmp(q + 3, "cyan"))
		    i = '7';
		  else if (!strcasecmp(q + 3, "white"))
		    i = '8';
		  else if (!strcasecmp(q + 3, "default"))
		    i = '0';
		  else
		    break;
		  SPrintF(sub, 10, "\033%c", i + ((q[0] == 'b') ? 'A' - '0' : 0));
		  towrite = sub;
	       } else if (!strcasecmp(q, "end")) {
		  blind &= ~1;
		  SubstAddCol(sub, 0, 1);
		  towrite = sub;
	       } else if (!strcasecmp(q, "center"))
		    center = 1;
	       else if (!strncmp(q, "cols=", 5)) {
		  char * r = strchr(q + 5, '/');
		  SubstAddCol(sub, atoi(q + 5), r ? atoi(r + 1) : 70);
		  towrite = sub;
	       }
	    }
	 } else
	   current = q;		/* no } so ignore */
	 break;
	default:
	 if (!blind) {
	    *writeidx++ = chr;
	    if (writeidx >= (s + HELP_BUF_LEN)) {
	       *writeidx = 0;
	       return;
	    }
	 }
      }
      if (towrite && !blind) {
	 if ((writeidx + strlen(towrite)) >= (s + HELP_BUF_LEN)) {
	    strncpy(writeidx,towrite,(s + HELP_BUF_LEN) - writeidx);
	    s[HELP_BUF_LEN] = 0;
	    return;
	 }
	 writeidx += StrCpy(writeidx,towrite);
      }
      if (chr) {
	 readidx = current + 2;
	 current = strchr(readidx,'%');
      } else {
	 readidx = current + 1;
	 current = NULL;
      }
   }
   if (!blind) {
      i = strlen(readidx);
      if (i && ((writeidx + i) >= (s + HELP_BUF_LEN))) {
			strncpy(writeidx,readidx,(s + HELP_BUF_LEN) - writeidx);
			s[HELP_BUF_LEN] = 0;
			return;
      }      
      strcpy(writeidx,readidx);
   } else
     *writeidx = 0;
   if (!(blind & 2) && center) {
      strcpy(xx, s);
      i = 35 - (strlen(xx) / 2);
      if (i > 0) {
	 s[0] = 0;
	 for (j = 0; j < i; j++)
	   s[j] = ' ';
	 strcpy(s+i, xx);
      }
   }
   if (!(blind & 2))
     SubstAddCol(s, 0, 0);
}

int ModuleEntryEggdrop::LoadLanguage (const char * par) {
   FUNCTION;
   FILE	*FLANG;
   char	langfile[DIRLEN], *lread, ltext[512], lbuf[512], *ctmp, *ctmp1;
   int	lidx, lline = 0, lskip;
   ModuleLanguage * mlang = language;
   
   context;
   if (par[0] == '.' && par[0] == '/') {
      strncpy(langfile, par, DIRLEN);
      langfile[DIRLEN] = 0;
   } else
     SPrintF(langfile, DIRLEN, "./language/%s.lang", par);
   if (!(FLANG = fopen(langfile, "r"))) 
     return 0;
   else {
      lskip = 0;
      while ((lread = fgets(lbuf, 511, FLANG))) {
	 lline++;
	 if ((lbuf[0] != '#') && !lskip) {
	    if (sscanf(lbuf, "%s", ltext) != EOF) {
	       if ((lbuf[0] == 'M') && (lbuf[1] == ',')) {
		  RMSpace(lbuf+2);
		  if (!(mlang = (ModuleLanguage *)langtab_list->Find(lbuf + 2)))
		    mlang = NEW ( ModuleLanguage ( lbuf + 2 ) );
	       } else if (sscanf(lbuf, "0x%x,%500c", &lidx, ltext) != 2) {
		  LOG_MISC.Put(0, "Malformed text line in %s at %d.", 
			       langfile, lline);
	       } else if (mlang) {
		  *strchr( ltext, '\n') = 0;
		  while (ltext[strlen(ltext) - 1] == '\\') {
		     ltext[strlen(ltext) - 1] = 0;
		     if (fgets(lbuf, 511, FLANG)) {
			lline++;
			*(strchr(lbuf, '\n') ) = 0;
			if (strlen(lbuf) + strlen(ltext) > 511) {
			   LOG_MISC.Put(0, 
					"Language: Message 0x%lx in %s at line %d too lon",
					lidx, langfile, lline);
			   lskip = 1;
			} else 
			  strcpy(strchr(ltext, 0), lbuf);
		     } 
		  }
		  
		  /* We gotta fix \n's here as, being arguments to */
		  /* sprintf(), they won't get translated */
		  for (ctmp = ltext, ctmp1 = ltext; *ctmp1; ctmp++, ctmp1++)
		    if(*ctmp1 == '\\' && *(ctmp1+1) == 'n') {
		       *ctmp = '\n';
		       ctmp1++;
		    } else 
		    *ctmp = *ctmp1;
		  *ctmp = '\0';
		  NEW ( LanguageEntry ( mlang, lidx, ltext ) );
	       }
	    }
	 } else if (lskip && 
		    (strlen(lbuf)==1  || *(strchr(lbuf, '\n')-1) != '\\'))
	      lskip = 0;
      }
      fclose(FLANG);
   }
   return 1;
}

/* this next section of code comes from eggdrops original match.c by:
 *   Chris Fuller  (aka Fred1@IRC & Fwitz@IRC)
 *     crf@cfox.bchs.uh.edu
 * I've substituted in all the defines, makes for neater reading :)
 */

/* removed the tolower stuff, it's exactly the same as ctype.h */

/*========================================================================*
 * EGGDROP:   wild_match_per(char *m, char *n)                            *
 *                                                                        *
 * Features:  Forward, case-insensitive, ?, *, %, ~(optional)             *
 * Best use:  Generic string matching, such as in IrcII-esque bindings    *
 *========================================================================*/
int ModuleEntryEggdrop::WildMatchPer (register unsigned char * m,
		    register unsigned char * n) {
   FUNCTION;
   unsigned char *ma = m, *lsm = 0, *lsn = 0, *lpm = 0, *lpn = 0;
   int match = 1, saved = 0, space;
   register unsigned int sofar = 0;

   /* take care of null strings (should never match) */
   if ((m == 0) || (n == 0) || (!*n))
      return 0;
   /* (!*m) test used to be here, too, but I got rid of it.  After all,
    * If (!*n) was false, there must be a character in the name (the
    * second string), so if the mask is empty it is a non-match.  Since
    * the algorithm handles this correctly without testing for it here
    * and this shouldn't be called with null masks anyway, it should be
    * a bit faster this way */

   while (*n) {
      /* Used to test for (!*m) here, but this scheme seems to work better */
      if (*m == '~') {	/* Match >=1 space       */
	 for (space = 1; (*(++m) == '~') || (*m == ' '); space++);
	 for (sofar += space; *n == ' '; n++, space--);
	 if (space <= 0)
	    continue;		/* Had enough spaces!    */
      } else { /* Do the fallback       */ 
	 switch (*m) {
	 case 0:
	    while ((--m > ma) && (*m == '?'));
	    if ((m > ma) ? ((*m == '*') && (m[-1] != '\\')) : (*m == '*'))
	       return (match + saved + sofar);	/* nonquoted * = match   */
	    break;
	 case '%':
	    while (*(++m) == '%');	/* Zap redundant %s      */
	    if (*m != '*') {	/* Don't both if next=*  */
	       if (*n != ' ') {	/* '*' can't match ' ' */
		  lpm = m;
		  lpn = n;	/* Save % fallback spot  */
		  saved += sofar;
		  sofar = 0;	/* And save tally count  */
	       }
	       continue;	/* Done with %           */
	    }
	    /* FALL THROUGH */
	 case '*':
	    while ((*(++m) == '*') || (*m == '%'));
	    lsm = m;
	    lsn = n;
	    lpm = 0;		/* Save * fallback spot  */
	    match += (saved + sofar);	/* Save tally count      */
	    saved = sofar = 0;
	    continue;		/* Done with *           */
	 case '?':
	    m++;
	    n++;
	    continue;		/* Match one char        */
	 case '\\':
	    m++;		/* Handle quoting        */
	 }

	 if (tolower(*m) == tolower(*n)) {	/* If matching           */
	    m++;
	    n++;
	    sofar++;
	    continue;		/* Tally the match       */
	 }
      }
      if (lpm) {		/* Try to fallback on %  */
	 n = ++lpn;
	 m = lpm;
	 sofar = 0;		/* Restore position      */
	 if ((*n | 32) == 32)
	    lpm = 0;		/* Can't match 0 or ' '  */
	 continue;		/* Next char, please     */
      }
      if (lsm) {		/* Try to fallback on *  */
	 n = ++lsn;
	 m = lsm;		/* Restore position      */
	 /* Used to test for (!*n) here but it wasn't necessary so it's gone */
	 saved = sofar = 0;
	 continue;		/* Next char, please     */
      }
      return 0;		/* No fallbacks=No match */
   }
   while ((*m == '*') || (*m == '%'))
      m++;			/* Zap leftover %s & *s  */
   return (*m) ? 0 : (match + saved + sofar);	/* End of both = match   */
}

/*=======================================================================*
 * EGGDROP:   wild_match(char *ma, char *na)                             *
 *                                                                       *
 * Features:  Backwards, case-insensitive, ?, *                          *
 * Best use:  Matching of hostmasks (since they are likely to begin with *
 *             a * rather than end with one).                            *
 *=======================================================================*/

int ModuleEntryEggdrop::_WildMatch (register unsigned char * m, register unsigned char * n) {
   FUNCTION;
   unsigned char *ma = m, *na = n, *lsm = 0, *lsn = 0;
   int match = 1;
   register int sofar = 0;

   /* take care of null strings (should never match) */
   if ((ma == 0) || (na == 0) || (!*ma) || (!*na))
      return 0;
   /* find the end of each string */
   while (*(++m));
   m--;
   while (*(++n));
   n--;

   while (n >= na) {
      if ((m <= ma) || (m[-1] != '\\')) {	/* Only look if no quote */
	 switch (*m) {
	 case '*':		/* Matches anything      */
	    while ((--m >= ma) && ((*m == '*') || (*m == '%')));
	    if ((m >= ma) && (*m == '\\'))
	       m++;		/* Keep quoted wildcard! */
	    lsm = m;
	    lsn = n;
	    match += sofar;
	    sofar = 0;		/* Update fallback pos   */
	    continue;		/* Next char, please     */
	 case '?':
	    m--;
	    n--;
	    continue;		/* ? always matches      */
	 }
	 sofar &= 0x7FFF;	/* Remember not quoted   */
      } else
	 sofar |= 0x8000;	/* Remember quoted       */
      if (tolower(*m) == tolower(*n)) {		/* If matching char      */
	 m--;
	 n--;
	 sofar++;		/* Tally the match       */
	 if (sofar & 0x8000)
	    m--;		/* Skip the quote char   */
	 continue;		/* Next char, please     */
      }
      if (lsm) {		/* To to fallback on *   */
	 n = --lsn;
	 m = lsm;
	 if (n < na)
	    lsm = 0;		/* Rewind to saved pos   */
	 sofar = 0;
	 continue;		/* Next char, please     */
      }
      return 0;		/* No fallback=No match  */
   }
   while ((m >= ma) && ((*m == '*') || (*m == '%')))
      m--;			/* Zap leftover %s & *s  */
   return (m >= ma) ? 0 : ((match + sofar) & 0x7FFF);
}

/* end of match.c code */

int ModuleEntryEggdrop::StripModes (const char * s) {
   FUNCTION;
   int res = 0;
   for (;*s; s++)
     switch (tolower(*s)) {
	case 'b':
	 res |= STRIP_BOLD;
	 break;
	case 'c':
	 res |= STRIP_COLOR;
	 break;
	case 'r':
	 res |= STRIP_REV;
	 break;
	case 'u':
	 res |= STRIP_UNDER;
	 break;
	case 'a':
	 res |= STRIP_ANSI;
	 break;
	case 'g':
	 res |= STRIP_BELLS;
	 break;
	case 'h':
	 res |= STRIP_HILITE;
	 break;
	case '*':
	 res |= STRIP_ALL;
	 break;
      }
   return res;
}

char * ModuleEntryEggdrop::StripMaskType(int x) {
   FUNCTION;
   static char s[20];
   char *p = s;
   
   if (x & STRIP_BOLD)
     *p++ = 'b';
   if (x & STRIP_COLOR)
     *p++ = 'c';
   if (x & STRIP_REV)
     *p++ = 'r';
   if (x & STRIP_UNDER)
     *p++ = 'u';
   if (x & STRIP_ANSI)
     *p++ = 'a';
   if (x & STRIP_BELLS)
     *p++ = 'g';
   if (x & STRIP_HILITE)
     *p++ = 'h';
   if (p == s) 
     *p++ = '-';
   *p = 0;
   return s;
}

/* make a password, 10-15 random letters and digits */
void ModuleEntryEggdrop::MakePass (char * s) {
   FUNCTION;
   int i = 10 + (random() % 6), j;
   
   for (j = 0; j < i; j++) {
      if (!(random() % 3))
	s[j] = '0' + (random() % 10);
      else
	s[j] = 'a' + (random() % 26);
   }
   s[i] = 0;
}

Socket * ModuleEntryEggdrop::FindValidIdx (int z) {
   FUNCTION;
   Socket * j = socklist;
   
   while ((j = j->Next()))
     if ((j->FD() == z) && (j->TypeFlags() & DCT_VALIDIDX))
       break;
   return j;
}

/* this part of add_note is more relevant to the botnet than
 * to the notes file */
int ModuleEntryEggdrop::AddNote (const char * to, const char * from, 
				 char * msg, Socket * idx, int echo) {
   FUNCTION;
   int status, sock;
   const char * fromreal, * toreal;
   UserRecord * u;
   Socket * ptr = socklist, * away;
   void * args [5];
   FlagRecord fr;
   BindReturn ret;
   
   if (strlen(msg) > 450)
     msg[450] = 0;
   fromreal = (fromreal = strchr(from, ':')) ? fromreal + 1 : from;
   toreal = (toreal = strchr(to, ':')) ? toreal + 1 : to;
   /* don't process if there's a note binding for it */
   args[0] = (void *)fromreal;
   args[1] = (void *)toreal;
   args[2] = msg;
   args[3] = (void *)to;
   args[4] = idx;
   if (bt_note->Check (from, 0, args) == EXEC_LOG) {
      if (idx && echo)
	idx->Send(Lang( 0x1AC, "ss", to, msg));
      return NOTE_SCRIPT;
   }
   if (!(u = userlist->Find(toreal))) {
      if (idx)
	 idx->Send(Lang( 0x100,""));
      return NOTE_ERROR;
   }
   if (fr.Set(u)->GlobBot()) {
      if (idx)
	 idx->Send(Lang( 0x206, ""));
      return NOTE_ERROR;
   }
   status = NOTE_STORED;
   away = 0;
   /* online right now? */
   sock = (to == toreal) ? -1 : atoi(to);
   while ((ptr = ptr->Next()))
     if ((ptr->TypeFlags() & DCT_GETNOTES) && ((sock == -1) || (sock == ptr->FD())) && 
	 !strcasecmp(ptr->Nick(), toreal)) {
	int aok = 1;
	if (ptr->Type() == DCC_CHAT) {
	   DccChat * dcc = (DccChat *)ptr;
	   
	   if (dcc->Away() && (idx || !echo)) {
	      /* only check away if it's not from a bot */
	      aok = 0;
	      if (idx)
		idx->Send(Lang( 0x207, "ss", dcc->Nick(), dcc->Away()));
	      if (!away)
		away = ptr;
	      status = NOTE_AWAY;
	   }
	}
	if (aok) {
	   char * p, work[512];
	   const char *fr = fromreal;
	   int l = 0;
	   
	   while ((*msg == '<') || (*msg == '>')) {
	      p = Split(&msg);
	      if (*p == '<')
		l += SPrintF(work+l, 512 - l, "via %s, ", p + 1);
	      else if (*from == '@')
		fr = p + 1;
	   }  
	   ptr->Send(Lang(((idx || (echo < 0)) || !strcasecmp(fromreal, botnetnick))
			  ? 0x1AD : 0x1AE, "sssc", fr, l ? work : "", msg, 7));
	   if (idx && echo)
	     idx->Send(Lang( 0x1AC, "ss", to, msg));
	   return NOTE_OK;
	}
     }
   if (!idx && echo)
      return NOTE_OK;		/* error msg from a tandembot: don't store */
   /* call store note here */
   args[0] = (void *)fromreal;
   args[1] = (void *)toreal;
   args[2] = msg;
   args[3] = (void *)to;
   args[4] = idx;
   ret = bt_strnote->Check(toreal, 0, args);
   if (ret == EXEC_LOG)
     return NOTE_FWD;
   if (ret == EXECUTED) {
      if (status == NOTE_AWAY) 
	/* user is away in all sessions -- just notify the user that a */
	/* message arrived and was stored.  (only oldest session is notified.) */
	away->Send(Lang( 0x208, ""));
      return status;
   }
   return NOTE_ERROR;
}

void ModuleEntryEggdrop::Fatal (const char * s) {
   FUNCTION;
   void * args = (void *)s;
   char * o;
   Socket * ptr;
   
   o = Lang( 0x10C, "s", s);
   LOG_MISC.Put(Lang( 0x1CA, "s", s));
   log_list->Flush();
   bt_die->Check("*", NULL, &args);
   sleep(2);
   while ((ptr = socklist->Next())) {
      if ((ptr->Type() == DCC_CHAT) && (((DccChat *)ptr)->Channel() >= 0))
	ptr->Send(o);
      delete ptr;
   }
   delete socklist;
   unlink(pid_file);
   exit(1);
}
 
/* here is as good a place as any to shove this :) */
int ModuleEntryEggdrop::SPrintF (char * buf, int maxlen, char * format, ... ) {
   FUNCTION;
   char *s;
   int c = 0,i;
   
   va_list va;
   va_start(va,format);
   
   while (*format && (c < maxlen)) {
      if (*format == '%') {
	 format++;
	 
	 switch (*format) {
	  case 's':
	    s = va_arg(va,char *);
	    break;
	  case 'd':
	    i = va_arg(va,unsigned int);
	    s = ToBase10(i);
	    break;
	   case 'l':
	    i = va_arg(va,unsigned long);
	    s = ToBase10(i);
	    break;
	  case 'D':
	    i = va_arg(va,unsigned int);
	    s = ToBase64(i);
	    break;
	  case '%':
	    buf[c++] = *format++;
	    continue;
	   case ' ':
	    for (i = va_arg(va, unsigned int); i && (c < maxlen); i--)
	      buf[c++] = ' ';
	    continue;
	  case 'c':
	    buf[c++] = (char)va_arg(va,int);
	    format++;
	    continue;
	   case '9':
	   case '8':
	   case '7':
	   case '6':
	   case '5':
	   case '4':
	   case '3':
	   case '2': {
	      int p = *format - '0';
	      
	      i = va_arg(va,unsigned int);
	      if ((c + p) < maxlen) {
		 while ( ((--p) >= 0) ) {
		    buf[c + p] = '0' + i % 10;
		    i = i / 10;
		 }
		 c += *format - '0';
	      }
	      format++;
	      continue;
	   }
	   default:
	    continue;
	 }
	 if (s)
	   while (*s && (c< maxlen))
	     buf[c++] = *s++;
	 format++;
      } else
	buf[c++] = *format++;
   }
   va_end(va);
   buf[c] = 0;
   return c;
}

/* remove space characters from beginning and end of string */
/* (more efficent by Fred1) */
void ModuleEntryEggdrop::RMSpace (char * s) {
#define whitespace(c) ( ((c)==32) || ((c)==9) || ((c)==13) || ((c)==10) )
   char *p;
   /* wipe end of string */
   for (p = s + strlen(s) - 1; ((whitespace(*p)) && (p >= s)); p--);
   if (p != s + strlen(s) - 1)
      *(p + 1) = 0;
   for (p = s; ((whitespace(*p)) && (*p)); p++);
   if (p != s)
      strcpy(s, p);
}
 
void ModuleEntryEggdrop::Chanprog(void) {
   FUNCTION;
   FILE *f;
   char s[DIRLEN+1];
   
   admin[0] = 0;
   helpdir[0] = 0;
   tempdir[0] = 0;
   
   DELETE ( log_list );
   log_list = NEW( LogFile () );
   conmask = 0;
   /* turn off read-only variables (make them write-able) for rehash */
   protect_readonly = 0;
   /* now read it */
   context;
   if (scripts->ParseFile(configfile) == SCRIPT_ERROR)
      Fatal(Lang( 0x214, "s", scripts->AsString()));   
   /* We should be safe now */
   hook_rehash->Call();
   context;
   protect_readonly = 1;
   if (!userfile[0])
     Fatal(Lang( 0x215, ""));
   if (!(int) getuid())
      /* perhaps you should make it run something innocent here ;) */
      /* like rm -rf /etc :) */
     printf("\n\n%s\n", Lang( 0x213, ""));
   if (!userlist->ReadUserfile(userfile, NULL)) {
      if (!make_userfile)
	Fatal(Lang( 0x216, ""));
      printf("\n\n%s\n", Lang( 0x217, ""));
      if (module_list->Find("server", &ServerVer))
	printf("%s\n", Lang( 0x218, "s", origbotname));
      printf("%s\n\n", Lang( 0x219, ""));
   } else if (make_userfile)
	Fatal(Lang( 0x21A, ""));
   context;
   if (helpdir[0])
      if (helpdir[strlen(helpdir) - 1] != '/')
	 strcat(helpdir, "/");
   if (tempdir[0])
      if (tempdir[strlen(tempdir) - 1] != '/')
	 strcat(tempdir, "/");
   if (!botnetnick[0]) {
      strncpy(botnetnick, origbotname, HANDLEN);
      botnetnick[HANDLEN] = 0;
   }
   if (!botnetnick[0])
     Fatal(Lang( 0x21B, ""));
   context;
   SPrintF(s,  DIRLEN, "%s.test.file", tempdir);
   if (!(f = fopen(s, "w")))
     Fatal(Lang( 0x21C, "s", s));
   fclose(f);
   unlink(s);
   context;
   FlagRecord::ReaffirmOwners();
}

/* reload the user file from disk */
void ModuleEntryEggdrop::Reload (void) {
   FUNCTION;
   FILE *f;
   
   if (!(f = fopen(userfile, "r"))) 
     LOG_MISC.Put(Lang( 0x21D, ""));
   else {
      fclose(f);
      noshare = 1;
      DELETE (userlist);
      noshare = 0;
      userlist = NEW ( UserRecord () );
      if (!userlist->ReadUserfile(userfile, NULL))
	Fatal(Lang( 0x21E, ""));
      context;
      FlagRecord::ReaffirmOwners();
      hook_read_userfile->Call();
   }
}

void ModuleEntryEggdrop::Rehash (void) {
   FUNCTION;
   noshare = 1;
   DELETE (userlist);
   noshare = 0;
   userlist = NEW ( UserRecord () );
   hook_pre_rehash->Call();
   Chanprog();
}

int ModuleEntryEggdrop::LoadConfig ( DccChat * x ) {
   if (savedconfigfile[0]) {
      if (scripts->ParseFile(savedconfigfile) == SCRIPT_ERROR)
	LOG_MISC.Put(Lang( 0x2A7, "s", scripts->AsString()));   
      else if (x)
	x->Send(Lang(0x028,""));
   } else 
     LOG_MISC.Put(Lang(0x2A5,""));
   return 1;
}

int ModuleEntryEggdrop::SaveConfig (DccChat * x) {
   FILE * f;
   char s[DIRLEN + 1];
   
   if (!savedconfigfile[0]) 
     LOG_MISC.Put(Lang(0x2A5,""));
   else {
      G->SPrintF(s, DIRLEN, "%s~new", G->savedconfigfile);
      if (!(f = fopen(s, "w")))
	G->LOG_MISC.Put(G->Lang( 0x2A6, ""));
      else {
	 chmod(s, 0600);           /* make it -rw------- */
	 if ((scripts->WriteVariables(f) == EOF) || (fflush(f) == EOF)) {
	    fclose(f);
	    G->LOG_MISC.Put(G->Lang( 0x2A6, ""));
	 } else {
	    fclose(f);
	    unlink(G->savedconfigfile);
	    rename(s, G->savedconfigfile);
	    if (x)
	      x->Send(Lang(0x028,""));
	 }
      }
   }
   return 1;
}


