/* 
 * ScriptContext.h - A context to describe the current scripting interface state
 */

#ifndef _KEYEDLIST_
#include "KeyedList.h"
#endif

#ifndef _SCRIPTCONTEXT_
#define _SCRIPTCONTEXT_

class StringVariable;
class IntegerVariable;
class CoupletVariable;
class MiscVariable;

class ScriptContext : public KeyedList<MISCLEN> {
  protected:
  public:
   ScriptContext ( void );
   inline ScriptContext ( const char * kn, ModuleEntry * m = 0 ) 
     : KeyedList<MISCLEN> (*G->scripts, kn, KeyedList<MISCLEN>::NODUPS, m ) {
     }
   ~ScriptContext ( void );

   MEMORY_TYPE(G->memtype_script);
   int ExpMem ( void );
   
   virtual void SetString ( const char *);
   virtual void SetInteger ( int );
   virtual void SetArray ( int, const char ** );
   virtual void AppendString ( const char * );
   virtual void AppendInteger ( int );
   virtual void AppendArray ( int, const char ** );
   
   virtual void AddStringVar ( StringVariable *);
   virtual void RemStringVar ( StringVariable *);
   virtual void AddIntVar ( IntegerVariable * );
   virtual void RemIntVar ( IntegerVariable * );
   virtual void AddCoupletVar ( CoupletVariable * );
   virtual void RemCoupletVar ( CoupletVariable * );
   virtual void AddMiscVar ( MiscVariable * );
   virtual void RemMiscVar ( MiscVariable * );
   
   virtual void AddCommand ( Command * );
   virtual void RemCommand ( Command * );

   virtual BindReturn CallBind ( BindEntry *, const char *, int, va_list );
   virtual int Evaluate ( const char *, int, ... );
   virtual const char * AsString ( void );
   virtual int AsInteger ( void );
   virtual void Clear ( void );
   
   void CleanupModule ( ModuleEntry * );
   void SetMeUp ( void );
   int ParseFile ( const char * );
   void SetVar ( Socket *, const char *, const char * );
   void DumpVar ( Socket * , const char *);
   void DumpVars ( Socket * );
   int WriteVariables ( FILE * f );
		      
   inline ScriptContext * Next ( void ) { return (ScriptContext *)next; }
};
#endif
