/*
 * Socket.h - class definition for the Socket class
 */

#ifndef _INTLIST_
#include "IntList.h"
#endif

#ifndef _SOCKET_
#define _SOCKET_

class Socket : public IntList {
  protected:
   static const char * SOCKET_NONSOCK;
   const char * type;
   int flags;
   int type_flags;
   long status; 
   struct in_addr address;
   int localport;
   int remoteport;
   UserRecord *user;
   char nick[HANDLEN + 1];
   char host[UHOSTLEN + 1];
   time_t timeval;
   time_t * timeout_val;
   int inbuflen;
   int outbuflen;
   int outbufoff;
   int outbufsize;
   char * outbuf;
   char inbuf[514];
   friend int SocketHandle ( void );
   friend class ListeningSocket;
   friend int main ( int, char ** );
  public:
   enum {
      FDMASK = 0x00007FFF,
	CONNECTING = 0x00800000, /* connection attempt */
	LISTENING = 0x01000000, /* listening port */
	STRONG = 0x02000000, /* don't report success until sure */
	LOST = 0x04000000, /* lost in one form or another */
	RAW = 0x08000000, /* do not buffer input */
	STALE = 0x10000000, /* hasnt received any data recently */
	FD_ = 0x20000000,
	NONSOCK = 0x40000000, /* used for file i/o on debug */
   };
   inline Socket ( ) { SOCKET_NONSOCK = "nonsock"; }
   Socket ( int, const char *, int, time_t *, int, const char * = NULL, int = 0 );
   Socket ( const char *, int , struct in_addr, int, int, time_t *);
   Socket ( Socket &, const char *, int, time_t * );
   virtual ~Socket ( void );
   MEMORY_TYPE(G->memtype_socket);
   int ExpMem ( void );
   static int Handle ( void );
   
   void Close ( );
   int Init ( int, int);
   int Scan ( int );
   int Write ( char *, int );
   void Send ( char * );
   void Printf ( char * format, ... );
   void Display (Socket *);
   void StripTelnet(char *,int *);
   void TellVerboseStatus ( int );
   
   char GetIcon ( void );
   void DumpLots ( char *, char * );
   Socket * Find ( int, const char * );
   Socket * Find ( const char *, int);
   void SendFirewall ( struct in_addr *, int );
   
   virtual void Eof (void);
   virtual void Activity (char *, int);
   virtual void Timeout (void);
   virtual void Display2 (Socket *);
   virtual int Dump ( char *, int );
   
   inline Socket * Next ( void ) { return (Socket *)next; }
   inline int TypeFlags ( void ) { return type_flags; }
   inline const char * Type ( void ) { return type; }
   inline void SetType ( const char * ntype ) { type = ntype; }
   inline const char * Nick ( void ) { return nick; }
   inline const int FD ( void ) { return key; }
   inline const char * Host ( void ) { return host; }
   inline const int Status ( void ) { return status; }
   inline void SetStatus ( int fl ) { status |= fl; }
   inline void SetStatus2 ( int fl ) { status = fl; }
   inline void ClearStatus ( int fl ) { status &= ~fl; }
   inline const int HasStatus ( int fl ) { return status & fl; }
   inline UserRecord * User ( void ) { return this ? user : 0; }
   inline const time_t Timeval ( void ) { return timeval; }
   inline time_t Age ( void ) { return G->now - timeval; }
   inline void TimeNow ( void ) { timeval = G->now; }
   inline Socket * Find ( int v ) { return (Socket *)IntList::Find(v); }
   inline void SetUser ( UserRecord * u ) { user = u; }
   inline void SetNick ( const char * c ) {
      strncpy(nick, c, HANDLEN);
      nick[HANDLEN] = 0;
   }
   inline int RemotePort ( void ) { return remoteport; }
   inline int LocalPort ( void ) { return localport; }
   inline struct in_addr Address ( void ) { return address; }
   inline void SetAddress (unsigned long IP) { address.s_addr = IP; }
   inline int Flags ( void ) { return flags; }
   inline void AddFlags ( int f ) { flags |= f; }
   inline void SetHost ( const char * x ) {
      strncpy(host, x, UHOSTLEN);
      host[UHOSTLEN] = 0;
   }
   inline void CheckTimeout (void) {
      if (timeout_val && ((G->now - timeval) > *timeout_val))
	Timeout();
   }
   inline void StealThings ( Socket & other ) {
      inbuflen = other.inbuflen;
      outbuflen = other.outbuflen;
      outbufoff = other.outbufoff;
      outbufsize = other.outbufsize;
      outbuf = other.outbuf;
      other.outbuf = 0;
      memcpy(inbuf, other.inbuf, 514);
      other.flags |= NONSOCK;
      other.key = 0;
      other.Close();
   }
};

#endif
