/*
 * UserEntry.C - member functions for the UserEntry class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "UserEntry.h"
#include "UserEntryUnknown.h"
#include "UserRecord.h"
#define CLASS "UserEntry"

UserEntry::UserEntry ( UserRecord * u, UserEntry * t ) 
: List ( u->entries ) {
   FUNCTION;
   type = t;
   owner = u;
} 

UserEntry::UserEntry ( const char * name, ModuleEntry * m )
: List (*G->entrytype_list, APPEND, m) {
   FUNCTION;
   UserRecord * u = G->userlist;
   UserEntryUnknown * ueu;
   UserEntry * ue = G->entrytype_list;
   
   while ((ue = ue->Next()))
     if ((ue != this) && ue->type && !strcasecmp((char *)ue->type,name)) {
	DELETE (ue);
	break;
     }
   type = (UserEntry *)name;
   owner = 0;
   while ((u = u->Next())) 
     if ((ueu = (UserEntryUnknown *)u->GetEntry(G->entrytype_unknown)))
       ueu->Unpack(this);
}

UserEntry::~UserEntry ( void ) {
   FUNCTION;
   if (owner) {
      if (type)
	owner->entries.Remove(this);
   } else {
      UserRecord * u = G->userlist;
      UserEntry * ue;
      
      while ((u = u->Next())) {
	 if ((ue = u->FindEntry(this)))
	   Pack(ue);
      }
      G->entrytype_list->Remove(this);
   }
}

int UserEntry::ExpMem ( void ) {
   FUNCTION;
   int tot = sizeof(UserEntry);
   
   if (!owner && !type) {
      UserEntry * ptr = this;
      UserRecord * u = G->userlist;
      
      while ((ptr = ptr->Next()))
	tot += ptr->ExpMem();
      while ((u = u->Next())) {
	 ptr = u->Entries();
	 while ((ptr = ptr->Next()))
	   tot += ptr->ExpMem();
      }
   }
   return tot;
}

void UserEntry::Unpack ( UserRecord *, MessageQueue * ) {
   FUNCTION;
}

void UserEntry::Pack ( UserEntry * ) {
   FUNCTION;
}

UserEntry * UserEntry::Create ( UserRecord * ) {
   FUNCTION;
   return 0;
}

void UserEntry::GotShare ( char *, Socket * ) {
   FUNCTION;
}

void UserEntry::Display ( Socket * ) {
   FUNCTION;
}

void UserEntry::PruneUser ( Socket * ) {
   FUNCTION;
}

int UserEntry::Write ( FILE * , Socket *) {
   FUNCTION;
   return 0;
}

int UserEntry::ScriptGet ( ScriptContext *, int, ScriptArgument ** ) {
   FUNCTION;
   return SCRIPT_ERROR;
}

int UserEntry::ScriptSet ( ScriptContext *, int, ScriptArgument ** ) {
   FUNCTION;
   return SCRIPT_ERROR;
}

int UserEntry::MatchThis ( const char * ) {
   FUNCTION;
   return 0;
}

UserEntry * UserEntry::Find ( UserEntry * et ) {
   FUNCTION;
   UserEntry * e = this, * t = 0;
   
   if (this) 
     while ((t = e->Next())) {
	if (t->type == et) {
	   e->next = t->next;
	   t->next = next;
	   next = t;
	   break;
	}
	e = t;
     }
   return t;
}

UserEntry * UserEntry::Find ( const char * c ) {
   FUNCTION;
   UserEntry * e = this;
   
   if (this && !owner)
     while ((e = e->Next()))
       if (!strcasecmp(c, (char *)e->type))
	 return e;
   return 0;
}
