/*
 * UserEntryHosts.C - member functions for the UserEntryHosts class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "UserEntryHosts.h"
#include "UserRecord.h"
#include "Socket.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "UserEntryUnknown.h"
#define CLASS "UserEntryHosts"

UserEntryHosts::~UserEntryHosts ( void ) {
   FUNCTION;
}

void UserEntryHosts::Unpack ( UserRecord * u, MessageQueue * e ) {
   FUNCTION;
   UserEntryHosts * ueh = (UserEntryHosts *)u->GetEntry(this);
   
   while ((e = e->Next()))
     NEW ( UserEntryHostsList ( ueh->list, e->Data(), 
			       KeyedList<UHOSTLEN>::APPEND) );
}

void UserEntryHosts::Pack ( UserEntry * ue ) {
   FUNCTION;
   KeyedList<UHOSTLEN> * l = &((UserEntryHosts *)ue)->list;
   UserEntryUnknown * ueu = 
     (UserEntryUnknown *)ue->Owner()->GetEntry(G->entrytype_unknown);
   
   while ((l = l->Next()))
     ueu->NewLine((char *)type, l->Key());
}

UserEntry * UserEntryHosts::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryHosts ( u, this ) );
}

void UserEntryHosts::PruneUser (Socket *) {
   FUNCTION;
   DELETE ( this );
}

int UserEntryHosts::Write ( FILE * f, Socket * ) {
   FUNCTION;
   KeyedList<UHOSTLEN> * ptr = &list;
   
   while ((ptr = ptr->Next()))
     if (fprintf(f,"=%s %s\n",type->Name(),ptr->Key()) == EOF)
       return 0;
   return 1;
}

int UserEntryHosts::ScriptGet (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   KeyedList<UHOSTLEN> * ptr = &list;
   const char * list;
   
   BADARGS(3, 3, G->Lang(0x26B, ""));
   while ((ptr = ptr->Next()))
     sc->AppendArray(1, &(list = ptr->Key()));
   return SCRIPT_OK;
}

int UserEntryHosts::ScriptSet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BADARGS(3, 4, G->Lang(0x26C, ""));
   if (argc == 4) 
     owner->AddHost(argv[3]->AsString());
   else 
     list.Flush();
   return SCRIPT_OK;
}

int UserEntryHosts::ExpMem ( void ) {
   FUNCTION;
   UserEntryHostsList * ptr = &list;
   int tot = sizeof(UserEntryHosts);
   
   while ((ptr = (UserEntryHostsList *)ptr->Next()))
     tot += sizeof(UserEntryHostsList);
   return tot;
}

void UserEntryHosts::Display ( Socket * idx ) {
   FUNCTION;
   char s[512];
   KeyedList<UHOSTLEN> * ptr = &list;
   int i, ii;
   
   ii = i = G->SPrintF(s, 511, "  %s ", G->Lang(0x26A, ""));
   while ((ptr = ptr->Next()))
     if (i + strlen(ptr->Key()) + 2 > 65) {
	idx->Send(s);
	i = G->SPrintF(s, 511, "  % %s, ", ii, ptr->Key());
     } else 
     i += G->SPrintF(s + i, 511 - i, "%s, ", ptr->Key());
   if (i > ii)
     idx->Send(s);
}

void UserEntryHosts::Add (const char * buf) {
   FUNCTION;
   char host[UHOSTLEN + 1], *p;
   KeyedList<UHOSTLEN> * ptr, * t;
   
   context;
   if (!buf || !strcasecmp(buf,"none")) 
     list.Flush();
   else {
      strncpy(host, buf, UHOSTLEN);
      host[UHOSTLEN] = 0;
      
      /* can't have ,'s in hostmasks */
      while ((p = strchr(host, ',')))
	*p = '?';
      /* fred1: check for redundant hostmasks with */
      /* controversial "superpenis" algorithm ;) */
      /* I'm surprised Raistlin hasn't gotten involved in this controversy */
      ptr = list.Next();
      while ((t = ptr)) {
	 ptr = t->Next();
	 if (G->WildMatch(host,t->Key()))  {
	    list.Remove(t);
	    DELETE(t);
	 }
      }
      NEW ( UserEntryHostsList(list, host, KeyedList<UHOSTLEN>::APPEND));
   }
   
}

int UserEntryHosts::FindHost ( const char * host, int old ) {
   FUNCTION;
   int ret = old, v;
   KeyedList<UHOSTLEN> * ptr = &list;
   
   while ((ptr = ptr->Next()))
     if ((v = G->WildMatch(ptr->Key(), host)) > ret)
       ret = v;
   return ret;
}

int UserEntryHosts::MatchThis ( const char * mask ) {
   FUNCTION;
   UserEntryHostsList * ptr = &list;
   
   while ((ptr = (UserEntryHostsList *)ptr->Next()))
     if (G->WildMatch(mask, ptr->Key()))
       break;
   return ptr != 0;
}
