/*
 * UserEntryLaston.C - member functions for the UserEntryLaston class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */
#include "UserEntryLaston.h"
#include "Socket.h"
#include "UserRecord.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "ChannelRecord.h"
#include "UserEntryUnknown.h"
#define CLASS "UserEntryLaston"

UserEntryLaston::~UserEntryLaston ( void ) {
   FUNCTION;
}

UserEntry * UserEntryLaston::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryLaston ( u, this ) );
}

void UserEntryLaston::Unpack ( UserRecord * u, MessageQueue * e ) {
   FUNCTION;
   UserEntryLaston * ue = (UserEntryLaston *)u->GetEntry(this);
   MessageQueue * x = e->Next();
   
   if (x) {
      char * q = strchr(x->Data(),' ');
      
      if (q) 
	q++;
      else
	q = "???";
      ue->Set(q, atoi(x->Data()));
   }
}

void UserEntryLaston::Pack ( UserEntry * ue ) {
   FUNCTION;
   UserEntryLaston * uel = (UserEntryLaston *)ue;
   UserEntryUnknown * ueu = 
     (UserEntryUnknown *)ue->Owner()->GetEntry(G->entrytype_unknown);
   char number[512];
   
   G->SPrintF(number, 511, "%lu %s", uel->When(), uel->Where());
   ueu->NewLine((char *)type, number);
}

int UserEntryLaston::Write ( FILE * f, Socket * idx ) {
   FUNCTION;
   if (!idx) 
     return (fprintf(f,"=LASTON %lu %s\n", when, safe_str(string)) != EOF);
   return 1;
}

int UserEntryLaston::ScriptGet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   ChannelRecord * cr;
   
   BADARGS(3, 4, G->Lang(0x265, ""));
   if (argc == 4) {
      for (cr = owner->ChanRecs(); cr; cr = cr->Next())
	if (!strcasecmp(cr->Key(), argv[3]->AsString())) {
	   sc->SetInteger(cr->Laston());
	   break;
	}
      if (!cr)
	sc->SetString("0");
   } else {
      const char * list[2] = {G->ToBase10(when), string};
      
      sc->SetArray(2, list);
   }
   return SCRIPT_OK;
}

int UserEntryLaston::ScriptSet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   ChannelRecord * cr;
   
   BADARGS(5, 6, G->Lang(0x266, ""));
   if (argc == 5) {
      char buf[512];
      
      strncpyz(buf, argv[4]->AsString(), 511);
      Set(buf, argv[3]->AsInteger());
   } else 
     for (cr = owner->ChanRecs(); cr; cr = cr->Next())
       if (!strcasecmp(cr->Key(), argv[4]->AsString()))
	 cr->SetLaston(argv[3]->AsInteger());
   return SCRIPT_OK;
}

int UserEntryLaston::ExpMem ( void ) {
   FUNCTION;
   return UserEntryString::ExpMem() + sizeof(UserEntryLaston)
     - sizeof(UserEntryString);
}   

void UserEntryLaston::Display (Socket *) {
   FUNCTION;
}
