/*
 * UserEntryString.C - member functions for the UserEntryString class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "UserEntryString.h"
#include "Socket.h"
#include "UserRecord.h"
#include "FlagRecord.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "UserEntryUnknown.h"
#define CLASS "UserEntryString"

UserEntryString::~UserEntryString ( void ) {
   FUNCTION;
   if (string)
     my_free(string, G->memtype_userent);
}

void UserEntryString::Unpack ( UserRecord * u, MessageQueue * e ) {
   FUNCTION;
   UserEntryString * ue = (UserEntryString *)u->GetEntry(this);
   MessageQueue * x = e->Next();
   
   if (x) 
     ue->Set((char *)x->Data());
}

void UserEntryString::Pack ( UserEntry * ue ) {
   FUNCTION;
   UserEntryUnknown * ueu = 
     (UserEntryUnknown *)ue->Owner()->GetEntry(G->entrytype_unknown);
   
   ueu->NewLine((char *)type, ((UserEntryString *)ue)->Value());
}

UserEntry * UserEntryString::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryString ( u, this ) );
}

int UserEntryString::Write ( FILE * f, Socket * idx ) {
   FUNCTION;
   if (!idx)
     return fprintf(f,"=%s %s\n",type->Name(), safe_str(string)) != EOF;
   return 1;
}

int UserEntryString::ScriptGet (ScriptContext * sc, int, ScriptArgument **) {
   FUNCTION;
   sc->SetString(string);
   return SCRIPT_OK;
}

int UserEntryString::ScriptSet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char buf[512];
   
   BADARGS(4,4, G->Lang(0x261, ""));
   strncpyz(buf, argv[3]->AsString(),511);
   Set(buf);
   return SCRIPT_OK;
}

int UserEntryString::ExpMem ( void ) {
   FUNCTION;
   context;
   return sizeof(UserEntryString) + (string ? strlen(string) + 1 : 0);
}

void UserEntryString::Display (Socket * idx) {
   FUNCTION;
   context;
   idx->Send(G->Lang(0x262, "ss", type->Name(), string));
}

void UserEntryString::Set ( char * buf ) {
   FUNCTION;
   unsigned char * p = (unsigned char *)buf, buf2[161], *q = buf2;
   
   if (string)
     my_free(string, G->memtype_userent);
   if (buf) {
      while (*p && ((q - buf2) < 160)) {
	 if ((*p < 32) || ((*p > 126) && (*p < 224)))
	   p++;
	 else
	   *q++ = *p++;
      }
      *q = 0;
      string = (char *)my_malloc(q - buf2 + 1,G->memtype_userent);
      strcpy(string,(char *)buf2);
   } else 
     string = 0;
   context;
}

int UserEntryString::MatchThis ( const char * x ) {
   FUNCTION;
   return G->WildMatch(x, string);
}
