/*
 * UserEntryXtra.C - member functions for the UserEntryXtra class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "UserEntryXtra.h"
#include "UserRecord.h"
#include "Socket.h"
#include "FlagRecord.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "UserEntryUnknown.h"
#define CLASS "XtraKeys"

XtraKeys::~XtraKeys ( void ) {
   FUNCTION;
   if (data)
     my_free(data, G->memtype_userent);
}

void XtraKeys::Set ( const char * k ) {
   FUNCTION;
   int l = strlen(k);
   
   if (data)
     my_free (data, G->memtype_userent);
   if (l > 459)
     l = 459;
   data = (char *)my_malloc(l + 1, G->memtype_userent);
   strncpy(data, k, l);
   data[l] = 0;
}

int XtraKeys::ExpMem ( void ) {
   FUNCTION;
   XtraKeys * ptr = this;
   int tot = 0;
   
   while ((ptr = ptr->Next()))
     tot += sizeof(XtraKeys) + (ptr->data ? strlen(ptr->data) + 1 : 0);
   return tot;
}

#undef CLASS
#define CLASS "UserEntryXtra"

UserEntryXtra::UserEntryXtra ( UserRecord * u, UserEntry * uet ) 
: UserEntry (u, uet) {
   FUNCTION;
}

UserEntryXtra::~UserEntryXtra ( void ) {
   FUNCTION;
}

UserEntry * UserEntryXtra::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryXtra ( u, this ) );
}

void UserEntryXtra::Unpack ( UserRecord * u, MessageQueue * e ) {
   FUNCTION;
   UserEntryXtra * ue = (UserEntryXtra *)u->GetEntry(this);
   char buf[512];
   
   while ((e = e->Next())) {
      strncpy(buf, e->Data(), 511);
      buf[511] = 0;
      ue->GotShare(buf, 0);
   }
}

void UserEntryXtra::Pack ( UserEntry * ue ) {
   FUNCTION;
   XtraKeys * xk = &((UserEntryXtra *)ue)->list;
   UserEntryUnknown * ueu = 
     (UserEntryUnknown *)ue->Owner()->GetEntry(G->entrytype_unknown);
   char buf[512];
     
   while ((xk = xk->Next())) {
      G->SPrintF(buf, 511, "%s %s", xk->Key(), xk->Data());
      ueu->NewLine((char *)type, buf);
   }
}

int UserEntryXtra::ExpMem ( void ) {
   FUNCTION;
   return sizeof(UserEntryXtra) + list.ExpMem ( );
}

void UserEntryXtra::GotShare (char * buf, Socket *) {
   FUNCTION;
   char * x = strchr(buf,' ');
   
   if (x)
     *x++ = 0;
   else
     x = "";
   NEW ( XtraKeys ( list, buf, x ) );
}

int UserEntryXtra::Write (FILE * f, Socket * idx) {
   FUNCTION;
   XtraKeys * x = &list;

   if (!idx) 
     while ((x = x->Next()))
       if (fprintf(f,"=%s %s %s\n",type->Name(),x->Key(),x->Data()) == EOF)
	 return 0;
   return 1;
}

int UserEntryXtra::ScriptGet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   XtraKeys * x = &list;
   
   BADARGS(3, 4, G->Lang(0x269, ""));
   if (argc == 4) {
      if ((x = list.Find(argv[3]->AsString())))
	sc->SetString(x->Data());
   } else 
     while ((x = x->Next())) {
	const char * list[2];
	
	list[0] = x->Key();
	list[1] = x->Data();
	sc->AppendArray(2, list);
     }
   return SCRIPT_OK;
}

int UserEntryXtra::ScriptSet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   XtraKeys * x;
   
   BADARGS(4, 5, G->Lang(0x267, ""));
   
   if ((x = list.Find(argv[3]->AsString())))
     x->Set(safe_str(argv[4]->AsString()));
   else 
     NEW ( XtraKeys ( list, argv[3]->AsString(), argv[4]->AsString() ) );
   return SCRIPT_OK;
}

void UserEntryXtra::Display (Socket * idx) {
   FUNCTION;
   char buf[121], *p, *t = buf;
   XtraKeys * x;
   
   context;
   strncpyz(buf, G->whois_fields, 120);
   while ((p = G->Split(&t)) && p[0]) {
      x = &list;
      while ((x = x->Next())) 
	if (!strcasecmp(p, x->Key()))
	  idx->Send(G->Lang(0x268, "ss", x->Key(), x->Data()));
   }
   context;
}

void UserEntryXtra::SetKey (const char * key, const char * data) {
   FUNCTION;
   XtraKeys * x = GetKey(key);
   FlagRecord fr (owner);
   
   x->Set(safe_str(data));
   if (!G->noshare && G->shareout && !fr.GlobBot() && !fr.GlobUnshared())
     G->shareout(NULL,"c%s %s %s %s\n", owner->Handle(), type->Name(), x->Key(),
		 x->Data());
}
