/*
 * UserRecord.C - member functions for the UserRecord class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "DccChat.h"
#include "UserRecord.h"
#include "BindTable.h"
#include "BanRecord.h"
#include "UserEntryXtra.h"
#include "UserEntryHosts.h"
#include "UserEntryPass.h"
#include "UserEntryLaston.h"
#include "UserEntryUnknown.h"
#include "HookList.h"
#include "ModuleEntryEncryption.h"
#include <sys/stat.h>
#ifdef HAVE_BOTNET_H
#include "Bot.h"
#endif
#include "Channel.h"
#include "ChannelRecord.h"
#define CLASS "UserRecord"

static int noxtra = 0;

UserRecord::UserRecord (UserRecord * list, const char * newhand, 
		        char * newhost, FlagRecord & newflags )
: KeyedList<HANDLEN> ( *list, newhand, KeyedList<HANDLEN>::NODUPS |
			KeyedList<HANDLEN>::APPEND ), entries ( this ) {
   FUNCTION;
   UserEntryXtra * xe;
   UserEntryHosts * ueh;
   int oldshare = G->noshare;
   char x[100], *p;
   
   G->noshare = 1;
   flags = newflags.Global();
   flags_udef = newflags.udef_global;
   chanrec = NEW ( ChannelRecord ( ) );
   if (!noxtra && (xe = (UserEntryXtra *)GetEntry(G->entrytype_xtra)))
      xe->SetKey( "created", G->ToBase10(G->now) );
   /* strip out commas -- they're illegal */
   if (newhost && newhost[0]) 
     while ((p = strchr(newhost, ',')))
       *p = '?';
   else
     newhost = "none";
   if ((ueh = ((UserEntryHosts *)GetEntry(G->entrytype_hosts))))
     ueh->Add(newhost);
   G->noshare = oldshare;
   if (list == G->userlist) {
      if (G->check_chanlist && newhost && newhost[0])
	G->check_chanlist(newhost, NULL, CHANLIST_DELHOST);
      if (!G->noshare && (key[0] != '*') && !newflags.GlobUnshared()
	  && G->shareout) {
	 newflags.Build(x);
	 G->shareout(NULL,"n%s %s %s\n", key, newhost, x);
      }
      G->lastuser = this;
   }
}

UserRecord::~UserRecord ( void ) {
   FUNCTION;
   Channel * cst = G->chanset;
   Socket * ptr = G->socklist;
   FlagRecord fr (this);
   
   if (this == G->userlist) {
      Socket * ptr = G->socklist;
      
      while ((ptr = ptr->Next()))
	ptr->SetUser(0);
      if (G->check_chanlist)
	G->check_chanlist(NULL,NULL,CHANLIST_CLEAR);
      DELETE ( G->global_bans );
      NEW ( BanRecord () );
      DELETE ( G->ignores );
      G->ignores = NEW ( IgnoreRecord () );
      if (cst)
	while ((cst = cst->Next())) 
	  if (cst->bans) {
	     DELETE ( cst->bans );
	     NEW ( BanRecord ( cst ) );
	  }
      while (next)
	DELETE ( next );
   } else {
      if (G->userlist) {
	 G->userlist->Remove(this);
	 if (!G->noshare && (key[0] != '*') && !fr.GlobUnshared() && G->shareout)
	   G->shareout(NULL,"k%s\n", key);
	 while ((ptr = ptr->Next()))
	   if (ptr->User() == this)
	     ptr->SetUser(0);
	 if (G->check_chanlist)
	   G->check_chanlist(NULL, this, CHANLIST_DELHAND);
      }
      DELETE ( chanrec );
   }
   chanrec = NULL;
   entries.Flush();
   G->lastuser = NULL;
}

int UserRecord::ExpMem ( void ) {
   FUNCTION;
   UserRecord * ptr = this;
   int tot = sizeof(UserRecord);
   while ((ptr = ptr->Next()))
     tot += sizeof(UserRecord);
   return tot;
}

UserRecord * UserRecord::Find ( const char * handle ) {
   FUNCTION;
   UserRecord * ret = 0;
   Socket * ptr;
   
   if (handle && handle[0] && (handle[0] != '*')) {
      if (this == G->userlist) {
	 if (((ret = G->lastuser) && !strcasecmp(G->lastuser->key, handle))
	     || ((ptr = G->socklist->Find(handle, 0)) && (ret = ptr->User()))
	     || (G->check_chanlist && 
		 (ret = G->check_chanlist(handle, NULL,CHANLIST_HAND))))
	   G->cache_hit++;
	 else {
	    ret = 0;
	    G->cache_miss++;
	 }
      }
      if (!ret) 
	ret = (UserRecord *)KeyedList<HANDLEN>::Find(handle);
   }
   return ret == G->userlist ? 0 : ret;
}

/* find CLOSEST host match */
/* (if "*!*@*" and "*!*@*clemson.edu" both match, use the latter!) */
/* 26feb: CHECK THE CHANLIST FIRST to possibly avoid needless search */
UserRecord * UserRecord::FindHost ( const char * host ) {
   FUNCTION;
   UserRecord * ret = 0, * user = 0;
   UserEntryHosts * q;
   int cnt = 0, i = 0;
   
   if (host && host[0]) {
      ret = G->check_chanlist ? 
	   G->check_chanlist(host, NULL, CHANLIST_GETHOST) : 0;
      if (ret) 
	 G->cache_hit++;
      else {
	 ret = user = G->userlist;
	 while ((user = user->Next()))
	   if ((q = (UserEntryHosts *)user->FindEntry(G->entrytype_hosts)) 
	       && ((i = q->FindHost(host, cnt)) > cnt)) {
	      ret = user;
	      cnt = i;
	   }	   
	 if (G->check_chanlist)
	   G->check_chanlist(host, ret, CHANLIST_SETHOST);
	 if (ret != G->userlist)
	   G->lastuser = ret;
      }
   }
   return ret ? ret : G->userlist;
}

UserRecord * UserRecord::FindExactHost ( const char * host ) {
   FUNCTION;
   UserRecord * ret = this;
   UserEntryHosts * q;
   
   while ((ret = ret->Next()))
     if ((q = (UserEntryHosts *)(ret->FindEntry(G->entrytype_hosts)))
	 && q->Find(host))
       break;
   return ret;
}

/* try: pass_match_by_host("-",host)
   will return 1 if no password is set for that host   */
int UserRecord::PassMatch ( const char * pass ) {
   FUNCTION;
   char newp[32], pbuf[16];
   UserEntryPass * cmp = (UserEntryPass *)FindEntry(G->entrytype_pass);
   FlagRecord fr (this);
   
   if (!cmp || !cmp->Value())
     return 1;
   if (pass && pass[0] && (pass[0] != '-')) {
      if (fr.GlobBot()) {
	 if (!strcmp(cmp->Value(), pass))
	   return 1;
      } else {
	 strncpyz(pbuf, pass, 15);
	 G->module_encrypt->EncryptPass(pbuf, newp);
	 if (!strcmp(cmp->Value(), newp))
	   return 1;
      }
   }
   return 0;
}

int UserRecord::ChangeHandle (char * newh) {
   FUNCTION;
   Socket * ptr = (Socket *)G->socklist;
   char s[HANDLEN+1];
   void * args[2];
   FlagRecord fr (this);
   
   /* nothing that will confuse the userfile */
   if (!newh[1] && strchr("+*:=.-#&",newh[0]))
     return 0;
   args[0] = key;
   args[1] = newh;
   G->bt_lnkch->Check(key, 0, args);
   if (!G->noshare && G->shareout && !fr.GlobUnshared ())
     G->shareout(NULL,"h%s %s\n", key, newh);
   strcpy(s, key);
   SetKey(newh);
   while ((ptr = ptr->Next()))
     if (!strcasecmp(ptr->Nick(), s) && !(ptr->TypeFlags() & DCT_BOT)) {
	ptr->SetNick(newh);
	if ((ptr->TypeFlags() & DCT_CHAT) && (((DccChat *)ptr)->Channel() >= 0)) 
	  DccChat::Chanout(((DccChat *)ptr)->Channel(),
			   G->Lang( 0x10C, "s",
				   G->Lang( 0x11B, "ss", s, newh)));
     }
   return 1;
}

int UserRecord::DelHost (const char * host) {
   FUNCTION;
   UserEntryHosts * hosts = (UserEntryHosts *)GetEntry(G->entrytype_hosts);
   int i = hosts ? hosts->Del(host) : 0;
   FlagRecord fr (this);
   
   context;
   if (!G->noshare && i && !fr.GlobUnshared() && G->shareout)
     G->shareout(NULL,"-%s %s\n", key, host); 
   if (G->check_chanlist)
     G->check_chanlist(host, this, CHANLIST_DELHAND);
   return i;
}

void UserRecord::AddHost ( const char * host ) {
   FUNCTION;
   UserEntryHosts * hosts = (UserEntryHosts *)GetEntry(G->entrytype_hosts);
   FlagRecord fr (this);
		  
   if (hosts)
     hosts->Add(host);
   if (!G->noshare && G->shareout && !fr.GlobUnshared()) {
      if (fr.GlobBot())
	G->shareout(NULL,"#%s %s\n", key, host);
      else
	G->shareout(NULL,"+%s %s\n", key, host);
   }
   if (G->check_chanlist)
     G->check_chanlist(host, this, CHANLIST_DELHOST);
}

void UserRecord::Display (Socket * idx) {
   FUNCTION;
   char s[81], s1[81], where[11];
   ChannelRecord *ch = chanrec;
   UserEntry * ue = &entries;
   UserEntryLaston * li = (UserEntryLaston *)GetEntry(G->entrytype_laston);
   UserEntryPass * pa = (UserEntryPass *)GetEntry(G->entrytype_pass);
   FlagRecord fr (this);
   time_t t = li->When();
   
   context;
   
   fr.Build(s);
   if (li) {
      if (!li->When())
	strcpy(s1, G->Lang( 0x279, ""));
      else {
	 strcpy(s1, ctime(&t));
	 if ((G->now - t) > 86400) {
	    s1[0] = s1[9];
	    s1[1] = s1[10];
	    s1[2] = s1[4];
	    s1[3] = s1[5];
	    s1[4] = s1[6];
	    s1[5] = 0;
	 } else {
	    s1[16] = 0;
	    strcpy(s1, &s1[11]);
	 }
      }
      context;
      strncpy(where, li->Where() ? li->Where() : G->Lang( 0x27A, ""), 10);
      where[10] = 0;
   } else
     where[0] = 0;
   idx->Send(G->Lang( 0x27B, "s ss ss ", key, HANDLEN-strlen(key),
		  G->Lang( (pa && pa->Value()) ? 0x27C : 0x27D, ""), s, 
		  15 - strlen(s), s1, where, 10-strlen(where)));
   /* channel flags? */
   context;
   while ((ch = ch->Next())) {
      fr.SetMatch(FlagRecord::CHANNEL | FlagRecord::GLOBAL);
      fr.Set(idx->User(),ch->Key());
      if (fr.AnyValidOp()) {
	 if (!ch->Laston())
	   strcpy(s1, G->Lang( 0x279, ""));
	 else {
	    strcpy(s1, ctime(&(t = ch->Laston())));
	    if ((G->now - ch->Laston()) > 86400) {
	       s1[0] = s1[9];
	       s1[1] = s1[10];
	       s1[2] = s1[4];
	       s1[3] = s1[5];
	       s1[4] = s1[6];
	       s1[5] = 0;
	    } else {
	       s1[16] = 0;
	       strcpy(s1, &s1[11]);
	    }
	 }
	 fr.SetMatch(FlagRecord::CHANNEL);
	 fr.Set(this, ch->Key());
	 fr.Build(s);
	 idx->Send(G->Lang( 0x27E, " s s s", HANDLEN - 9, ch->Key(),
			18 - strlen(ch->Key()), s, 15 - strlen(s), s1));
	 if (ch->Info())
	   idx->Send(G->Lang( 0x27F, "s", ch->Info()));
      }
   }
   /* user-defined extra fields */
   context;
   while ((ue = ue->Next()))
     ue->Display(idx);
}

/* match string: wildcard to match nickname or hostmasks */
/*               +attr to find all with attr */
void UserRecord::DisplayMatch ( DccChat * idx, char * mtch, int start, 
			       int limit, char * chname ) {
   FUNCTION;
   UserRecord * u = this;
   int cnt, nomns = 0, flags = 0;
   UserEntry * ue;
   FlagRecord user,pls,mns;
   
   context;
   idx->Send(G->Lang( 0x281, "s", mtch));
   cnt = 0;
   idx->Send(G->Lang( 0x280, " ", HANDLEN - 6));
   if (start > 1)
     idx->Send(G->Lang( 0x283, "d", start - 1));
   if (strchr("+-&|",*mtch)) {
      user.SetMatch(FlagRecord::GLOBAL | FlagRecord::BOT | FlagRecord::CHANNEL);
      pls.SetMatch(FlagRecord::GLOBAL | FlagRecord::BOT | FlagRecord::CHANNEL);
      pls.BreakDown(mtch,&mns);
      mns ^= pls;
      if (!mns.AnyFlags()) {
	 nomns = 1;
	 if (!pls.AnyFlags()) {
	    /* happy now BB you weenie :P */
	    idx->Send(G->Lang( 0x284, ""));
	    return;
	 }
      }
      if (!chname || !chname[0])
	chname = idx->ConsoleChan();
      flags = 1;
   }
   while ((u = u->Next()))
     if (flags) {
	user.Set(u,chname);
	if (pls == user) {
	   if (nomns || !(mns == user)) {
	      cnt++;
	      if ((cnt <= limit) && (cnt >= start))
		u->Display(idx);
	      if (cnt == limit + 1)
		idx->Send(G->Lang( 0x285, "d", limit));
	   }
	}
     } else if (G->WildMatch(mtch, u->Handle())) {
	cnt++;
	if ((cnt <= limit) && (cnt >= start))
	  u->Display(idx);
	if (cnt == limit + 1)
	  idx->Send(G->Lang( 0x285, "d", limit));
     } else {	 
	ue = u->Entries();
	while ((ue = ue->Next())) {
	   if (ue->MatchThis(mtch)) {
	      cnt++;
	      if ((cnt <= limit) && (cnt >= start)) 
		u->Display(idx);
	      if (cnt == limit + 1)
		idx->Send(G->Lang( 0x285, "d", limit));
	      break;
	   }
	}
     }
   idx->Send(G->Lang( cnt == 1 ? 0x286 : 0x287, "d", cnt));
}

int UserRecord::Write (FILE * f, Socket * idx, int f1) {
   FUNCTION;
   char s[181];
   ChannelRecord *ch = chanrec;
   Channel *cst;
   FlagRecord fr (this);
   UserEntry * e = &entries;
   int ok = 1;
   
   if (idx && !f1 && !fr.GlobUnshared() && !fr.GlobBot()) 
     fr.MaskButBot();
   fr.Build(s);
   if (fprintf(f, "%s %s\n", key, s) == EOF)
     return 0;
   fr.SetMatch(FlagRecord::CHANNEL);
   while ((ch = ch->Next())) {
      if ((cst = G->chanset->Find(ch->Key()))) {
	 if (idx) 
	   fr.Set(idx->User(),ch->Key());
	 if (!idx || fr.ChanShare()) {
	    fr.chan = ch->flags;
	    fr.udef_chan = ch->flags_udef;
	    fr.Build(s);
	    if (fprintf(f, "!%s %lu %s %s\n", ch->Key(), ch->Laston(),
			s, ((!idx || G->share_greet) && ch->Info()) ? ch->Info()
			: "" ) == EOF)
	      return 0;
	 }
      }
   }
   if (!idx || !(fr.GlobUnshared() || fr.GlobBot()))
     while ((e = e->Next()) && ok)
       ok = e->Write(f, idx);
   return ok;
}

/* rewrite the entire user file */
void UserRecord::WriteUserfile ( Socket * ) {
   FUNCTION;
   FILE *f;
   char s[DIRLEN+1], s1[81];
   time_t tt;
   UserRecord *u = this;
   int ok;
   
   context;
   /* also write the channel file at the same time */
   if (!next)
     return;
   G->SPrintF(s, DIRLEN, "%s~new", G->userfile);
   if (!(f = fopen(s, "w"))) 
     G->LOG_MISC.Put(G->Lang( 0x26D, ""));
   else {
      chmod(s, 0600);		/* make it -rw------- */
      G->LOG_MISC.Put(G->Lang( 0x26F, ""));
      tt = G->now;
      strcpy(s1, ctime(&tt));
      fprintf(f, "#5v: %s -- %s -- written %s", G->VersionStr(), G->botnetnick, s1);
      context;
      ok = 1;
      while ((ok && (u = u->Next())))
	ok = u->Write(f, 0);
      context;
      if (ok)
	ok = G->ignores->Write(f);
      if (!ok || fflush (f)) {
	 G->LOG_MISC.Put(G->Lang( 0x26E, "s", strerror(ferror(f))));
	 fclose(f);
      } else {
	 fclose(f);
	 context;
	 G->hook_write_userfile->Call();
	 context;
	 unlink(G->userfile);
	 G->SPrintF(s, DIRLEN, "%s~new", G->userfile);
	 rename(s, G->userfile);
      }
   }
}

/*
 * tagged lines in the user file:
 * *b global bals
 * *i global ignores
 * :#chan channel bans
 * -entries in each
 * <handle> begin user entry
 * =KEY INFO - info on each
 * !channel lines
 */
int UserRecord::ReadUserfile (const char * file, Socket * idx) {
   FUNCTION;
   char buf[512], lasthand[512], *attr, code,*s,*chname,*st,*fl,ignored[512];
   FILE *f;
   UserRecord *u = 0;
   Channel * cst = NULL;
   FlagRecord fr, f2(FlagRecord::BOT);
   ChannelRecord * cr;
   Socket * ptr;
   UserEntryUnknown * ueu;
     
   context;
   if (this == G->userlist) {
      if (G->check_chanlist)
	G->check_chanlist(NULL,NULL,CHANLIST_CLEAR);
      G->lastuser = NULL;
      DELETE ( G->global_bans );
      NEW ( BanRecord () );
      DELETE ( G->ignores );
      G->ignores = NEW ( IgnoreRecord () );
   }
   lasthand[0] = 0;
   ignored[0] = 0;
   if (!(f = fopen(file, "r")))
     return 0;
   G->noshare = noxtra = 1;
   context;
   /* read opening comment */
   fgets(buf, 511, f);
   if (buf[1] < '5') {
      fclose(f);
      if (G->module_list->Load("update"))
	G->Fatal(G->Lang( 0x288, "c", buf[1]));
      if (!(f = fopen(file, "r")))
	return 0;
      fgets(buf, 511, f);
   } else if (buf[1] > '5')
     G->Fatal(G->Lang( 0x289, "c", buf[1]));
   G->gban_total = 0;
   while (!feof(f)) {
      s = buf;
      fgets(s, 511, f);
      if (!feof(f)) {
	 if ((s[0] != '#') && (s[0] != ';') && s[0]) {
	    G->RMSpace(s);
	    code = s[0];
	    s++;
	    switch (code) {
	      case '-':
	       if ((lasthand[0] == '#') || (lasthand[0] == '+'))
		 NEW ( BanRecord ( cst, s ) );
	       else if (lasthand[0] == '*') {
		  if (lasthand[1] == 'i') 
		    NEW ( IgnoreRecord ( s ) );
		  else 
		    NEW ( BanRecord ( 0, s ) );
	       }
	       break;
	      case '!':
	       /* ! #chan laston flags [info] */
	       if (u) {
		  chname = G->Split(& s);	
		  fr.SetMatch(FlagRecord::CHANNEL);
		  if (idx) 
		    fr.Set(idx->User(),chname);
		  if (!idx || fr.ChanShare()) {
		     st = G->Split(&s);
		     fl = G->Split(&s);
		     fr.BreakDown(fl);
		     if (G->chanset->Find(chname)) {
			if (!(cr = u->chanrec->Find(chname)))
			  cr = NEW ( ChannelRecord (u, chname) );
			cr->SetLaston( atoi(st) );
			cr->flags = fr.Chan();
			cr->flags_udef = fr.ChanUdef();
			cr->SetInfo(s);
		     }
		  }
	       }
	       break;
	      case ':':
	       /* channel-specific bans */
	       strcpy(lasthand, s);
	       if (!G->chanset->Find(lasthand)) {
		  strcat(ignored, lasthand);
		  strcat(ignored, " ");
		  lasthand[0] = 0;
		  u = 0;
	       } else {
		  /* Remove all bans for this channel to avoid dupes */
		  /* NOTE only remove bans for when getting a userfile
		   * from another bot & that channel is shared BY THAT BOT */
		  cst = G->chanset->Find(lasthand);
		  if (idx) {
		     fr.SetMatch(FlagRecord::CHANNEL);
		     fr.Set(idx->User(), lasthand);
		  }
		  if (!idx || fr.ChanShare()) {
		     DELETE ( cst->bans );
		     NEW ( BanRecord ( cst ) );
		  } else {
		     /* otherwise ignore any bans for this channel */
		     cst = NULL;
		     lasthand[0] = 0;
		  }
	       }
	       break;
	      case '=':
	       /* new format storage */
	       context;
	       if (u) {
		  ueu = (UserEntryUnknown *)u->GetEntry(G->entrytype_unknown);
		  attr = G->Split(&s);
		  ueu->NewLine(attr,s);
	       }
	       break;
	      case '*':
	       u = NULL;
	       if ((s[0] == 'b') || (s[0] == 'i')) {
		  lasthand[0] = '*';
		  lasthand[1] = s[0];
	       } else 
		 lasthand[0] = 0;
	       break;
	      default:
	       s--;
	       st = G->Split(&s);
	       attr = G->Split(&s);
	       if (!attr[0]) {
		  G->LOG_MISC.Put(G->Lang( 0x28A, "s", st));
		  lasthand[0] = 0;
	       } else {
		  u = Find(st);
		  fr.SetMatch ( FlagRecord::GLOBAL);
		  fr.Set(u);
		  if (u && !fr.GlobUnshared()) {
		     G->LOG_MISC.Put(G->Lang( 0x28B, "s", st));
		     lasthand[0] = 0;
		     u = NULL;
		  } else if (u) {
		     lasthand[0] = 0;
		     u = NULL;
		  } else {
		     fr.BreakDown(attr);
		     cst = NULL;
		     if (strlen(st) > HANDLEN)
		       st[HANDLEN] = 0;
		     strcpy(lasthand, st);
		     if (idx && f2.Set(idx->User())->BotShareflags())
		       fr.SetGlobal(G->default_flags);
		     fr.Sanity();
		     u = NEW ( UserRecord ( this, st, 0, fr));
		     ptr = G->socklist;
		     while ((ptr = ptr->Next()))
		       if (!strcasecmp(st,ptr->Nick()))
			 ptr->SetUser(u);
		  }
	       }
	    }
	 }
      }
   }
   context;
   fclose(f);
   if (ignored[0]) 
     G->LOG_MISC.Put(G->Lang( 0x28C, "s", ignored));
   G->LOG_MISC.Put(G->Lang( 0x28D, ""));
   context;
   u = this;
   while ((u = u->Next())) {
      CONTEXT(u->Handle());
      ueu = (UserEntryUnknown *)u->GetEntry(G->entrytype_unknown);
      ueu->Unpack( 0 );
   }
   G->noshare = noxtra = 0;
   context;
   /* process the user data *now* */
   return 1;
}

/* after messing with someone's user flags, make sure the dcc-chat flags
   are set correctly */
int UserRecord::CheckDccAttrs( unsigned int oatr ) {
   FUNCTION;
   int stat, own = 0;;
   char *p, *q, s[121];
   Socket * ptr = G->socklist; 
   FlagRecord ofr, fr (this);
  
   ofr.SetGlobal(oatr);
   /* make sure default owners are +n */
   if (G->owner[0]) {
      for (q = G->owner; !own && (p = strchr(q, ',')); q = p + 1) {
	 strncpy(s, q, p-q);
	 s[p-q] = 0;
	 G->RMSpace(s);
	 if (!strcasecmp(key,s))
	   own = 1;
      }
      if (!own) {
	 strcpy(s,q);
	 G->RMSpace(s);
	 if (!strcasecmp(key,s))
	   own = 1;
      }
      if (own) {
	 fr.SetGlobOwner();
	 fr.Sanity();
	 Set(fr);
      }
   }
   while ((ptr = ptr->Next())) {
      if ((ptr->TypeFlags() & DCT_MASTER) && (this == ptr->User())) {
	 stat = ptr->Status();	
	 if (ptr->Type() == G->DCC_CHAT) {
	    DccChat * dcc = (DccChat *)ptr;
	    
	    if (fr.Checkables() != ofr.Checkables())
	      dcc->CheckLChjn();
	    if (ofr.GlobMaster() && !fr.GlobMaster()) {
	       FlagRecord cfr (this,FlagRecord::GLOBAL |FlagRecord::ANYWHERE);
	       
	       dcc->ConsoleFlags().Remove ( LogModes::MISC | LogModes::CMDS |
					   LogModes::FILES | LogModes::LEVS |
					   LogModes::WALL );
	       if (!cfr.ChanMaster())
		 dcc->ConsoleFlags().Add ( LogModes::MISC | LogModes::CMDS);
	       dcc->Send(G->Lang( 0x079, ""));
	       dcc->Send(G->Lang( 0x07A, ""));
	    } else if (!ofr.GlobMaster() && fr.GlobMaster()) {
	       dcc->ConsoleFlags() += G->conmask;
	       dcc->Send(G->Lang( 0x079, ""));
	       dcc->Send(G->Lang( 0x07B, ""));
	    }
	    if (!ofr.GlobBotmast() && fr.GlobBotmast()) {
	       dcc->Send(G->Lang( 0x07C, ""));
	       dcc->Send(G->Lang( 0x07D, ""));
	    } else if (ofr.GlobBotmast() && !fr.GlobBotmast()) {
	       dcc->Send(G->Lang( 0x07C, ""));
	       dcc->Send(G->Lang( 0x07E, ""));
	    }
	    if (!ofr.GlobOwner() && fr.GlobOwner()) {
	       dcc->Send(G->Lang( 0x07F, ""));
	       dcc->Send(G->Lang( 0x080, ""));
	    } else if (ofr.GlobOwner() && !fr.GlobOwner()) {
	       dcc->Send(G->Lang( 0x07F, ""));
	       dcc->Send(G->Lang( 0x081, ""));
	    }
	 }
	 if ((stat & STAT_PARTY) && fr.GlobOp())
	   stat &= ~STAT_PARTY;
	 if (!(stat & STAT_PARTY) && !fr.GlobOp())
	   stat |= STAT_PARTY;
	 if ((stat & STAT_CHAT) && !fr.GlobParty() && 
	     (!fr.GlobOp() || G->require_p))
	   stat &= ~STAT_CHAT;
	 if ((ptr->TypeFlags() & DCT_FILES) && !(stat & STAT_CHAT) &&
	     (fr.GlobParty() || (fr.GlobOp() && !G->require_p)))
	   stat |= STAT_CHAT;
	 ptr->ClearStatus (-1);
	 ptr->SetStatus(stat);
	 /* check if they no longer have access to wherever they are */
	 /* DON'T kick someone off the party line just cuz they lost +p */
	 /* (pinvite script removes +p after 5 mins automatically) */
	 if ((ptr->TypeFlags() & DCT_FILES) && !fr.GlobXfer()) {
	    ptr->Send(G->Lang( 0x082, ""));
	    ptr->Send(G->Lang( 0x083, ""));
	    G->LOG_MISC.Put(G->Lang( 0x084, "ss", ptr->Nick(), ptr->Host()));
	    if (ptr->HasStatus(STAT_CHAT))
	      /* simulate the user typing .quit */
	      ptr->Activity(".quit", 5);
	    else {
	       ptr->Close();
	       continue;
	    }
	 }
      }
#ifdef HAVE_BOTNET_H
      if ((ptr->TypeFlags() & DCT_BOT) && (this == ptr->User())) {
	 FlagRecord bfr (this, FlagRecord::BOT);
	 if (ptr->HasStatus(Bot::LEAF) && !bfr.BotLeaf())
	   ptr->ClearStatus(Bot::LEAF | Bot::WARNED);
	 if (!ptr->HasStatus(Bot::LEAF) && bfr.BotLeaf())
	   ptr->SetStatus(Bot::LEAF);
      }
#endif
   }
   return flags;
}

int UserRecord::CheckDccChanAttrs( const char *chname, int ochatr ) {
   FUNCTION;
   Channel *chan;
   DccChat * ptr = (DccChat *)G->socklist;
   FlagRecord fr (this, FlagRecord::GLOBAL | FlagRecord::CHANNEL),
     ofr (FlagRecord::CHANNEL);
   
   ofr.SetChan(ochatr);
   while ((ptr = ptr->Next())) 
     if (this == ptr->User()) {
	if (ofr.ChanCheckables() != fr.ChanCheckables())
	  ptr->CheckLChjn();
	if (ofr.ChanMaster() && !fr.ChanMaster()) {
	   if (!fr.GlobMaster())
	     ptr->ConsoleFlags().Remove(LogModes::MISC | LogModes::CMDS);
	   ptr->Send(G->Lang( 0x079, ""));
	   ptr->Send(G->Lang( 0x085, "s", chname));
	} else if (!ofr.ChanMaster() && fr.ChanMaster()) { 
	   ptr->ConsoleFlags() += G->conmask;
	   if (!fr.GlobMaster())
	     ptr->ConsoleFlags().Remove ( LogModes::MISC | LogModes::CMDS |
					 LogModes::FILES | LogModes::LEVS |
					 LogModes::WALL | LogModes::SRVOUT |
					 LogModes::RAW);
	   ptr->Send(G->Lang( 0x079, ""));
	   ptr->Send(G->Lang( 0x086, "s", chname));
	}
	if (ofr.ChanOwner() && !fr.ChanOwner()) { 
	   ptr->Send(G->Lang( 0x07F, ""));
	   ptr->Send(G->Lang( 0x087, "s", chname));
	} else if (!ofr.ChanOwner() && fr.ChanOwner()) { 
	   ptr->Send(G->Lang( 0x07F, ""));
	   ptr->Send(G->Lang( 0x088, "s", chname));
	}
	if (ofr.ChanOp() != fr.ChanOp()) {
	   for (chan = G->chanset; chan; chan = chan->Next())
	     if (ofr.Set(this,chan->Key())->GlobOp())
	       break;
	   if (!chan)
	     chan = G->chanset;
	   ptr->SetConsoleChan(chan ? chan->Key() : "*");
	}
     }
   return fr.Chan();
}

void UserRecord::Backup(void) {
   FUNCTION;
   char s[DIRLEN + 1];
   
   G->LOG_MISC.Put(G->Lang( 0x1E7, ""));
   G->SPrintF(s, DIRLEN, "%s~bak", G->userfile);
   G->CopyFile(G->userfile, s);
}
