/* 
 * Version.h - class definition
 */

#ifndef _MAIN_
#include "main.h"
#endif

#ifndef _VERSION_
#define _VERSION_

class Version {
  protected:
   char major;
   char minor;
   char fix;
   char patch;
   
  public:
   inline Version ( void ) { major = minor = fix = patch = 0; }
   inline Version ( char a, char b, char c = 0, char d = 0) {
      Set(a,b,c,d);
   }
   inline Version ( int val ) { Set ( val ); }
   inline Version ( const char * str ) { Set ( atoi ( str ) ); }
   inline Version ( Version * x ) {
      if (x) {
	 major = x->major;
	 minor = x->minor;
	 fix = x->fix;
	 patch = x->patch;
      }
   }
   inline void Set ( int val ) {
      major = val / 1000000;
      minor = (val / 10000) % 100;
      fix = (val / 100) % 100;
      patch = val % 100;
   }
   inline Version * Set ( char a, char b = 0, char c = 0, char d = 0) {
      major = a;
      minor = b;
      fix = c;
      patch = d;
      return this;
   }
   inline void AddPatch ( void ) { patch++; }
   inline int AsInt ( void ) { 
      return (major * 1000000) + (minor * 10000) + (fix * 100) + patch;
   }
   inline void AsNumber ( char *, int);
   inline void AsString ( char *, int);
   inline int Major ( void ) { return major; }
   inline int Minor ( void ) { return minor; }
   inline int Fix ( void ) { return fix; }
   inline int Patch ( void ) { return patch; }
   inline operator <= (Version & other ) {
      return (major == other.major) 
	&& ((minor == other.minor) || (minor == -1))
	  && (fix <= other.fix);
   }
};

#ifndef _MODULEENTRYEGGDROP_
#include "ModuleEntryEggdrop.h"
#endif

inline void Version::AsNumber ( char * buf, int len ) {
   List::G->SPrintF(buf, len, "%2%2%2%2", major, minor, fix, patch);
}
inline void Version::AsString ( char * buf, int len) {
   List::G->SPrintF(buf, len, "%d.%d.%d.%d", major, minor, fix, patch);
}
#endif
