/* 
 * Main.c -- handles:
 * changing nicknames when the desired nick is in use
 * flood detection
 * signal handling
 * command line arguments
 */
/*
 * 
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author (Darrin Smith) can be reached at:  beldin@au.eggdrop.net
 * a more general (and more likely to be answered) mailing list is
 * avaliable at eggdrop@eggdrop.net.
 */

#include "DccChat.h"
#include "UserRecord.h"
#include "BindTable.h"
#include "ModuleLanguage.h"
#include "MessageQueue.h"
#include "LogFile.h"
#include "HelpReference.h"
#include "BanRecord.h"
#include "UserEntryInfo.h"
#include "UserEntryPass.h"
#include "UserEntryHosts.h"
#include "UserEntryComment.h"
#include "UserEntryXtra.h"
#include "UserEntryLaston.h"
#include "UserEntryBotfl.h"
#include "UserEntryUnknown.h"
#include "BindTableCAB.h"
#include "BindTableDcc.h"
#include "BindTableChon.h"
#include "BindTableChof.h"
#include "BindTableLChjn.h"
#include "BindTableLChpt.h"
#include "BindTableString.h"
#include "BindTableAway.h"
#include "BindTableNote.h"
#include "BindTableListen.h"
#include "BindTableFilt.h"
#include "BindTableLNkch.h"
#include "BindTableTime.h"
#include "HookList.h"
#include "HostnameCache.h"
#include "ChannelRecord.h"
#include "StaticMods.h"
#include "Dependancy.h"
#include "ModuleEntry.h"
#include "SpecialSocket.h"
#include "ScriptContext.h"
#include "Channel.h"
#include "UserEntryPass.h"
#include "UserEntryComment.h"
#include "UserEntryHosts.h"
#include "MessageQueue.h"
#include "Dependancy.h"
#include "ScriptArgument.h"
#include "TelnetPort.h"
#include "StringVariable.h"
#include "IntegerVariable.h"
#include "CoupletVariable.h"
#include "Command.h"
#include "Firewall.h"
#include "patch.h"

#ifdef HAVE_IRC_H
#include "Channel3.h"
#endif
#ifdef HAVE_BOTNET_H
#include "Bot.h"
#endif

#include <time.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <sys/stat.h>
#include <setjmp.h>
#ifdef HAVE_UNAME
#include <sys/utsname.h>
#endif
#ifdef STOP_UAC			/* osf/1 complains a lot */
#include <sys/sysinfo.h>
#define UAC_NOPRINT    0x00000001	/* Don't report unaligned fixups */
#endif
#define CLASS ""

#ifndef _POSIX_SOURCE
/* solaris needs this */
#define _POSIX_SOURCE 1
#endif

static int notify_users_at = 0, switch_logfiles_at = 300, enable_simul = 1;
static int die_on_sighup = 0, die_on_sigterm = 0, nested_debug = 0;
static char console_user[HANDLEN+1] = "HQ", num_version[10];
static struct ModuleEntryEggdrop * G = 0;
jmp_buf alarmret;

static void write_debug(void) {
   Socket * sock;
   char s[80];
   int x;

   G->context_ptr &= 15;
   if (nested_debug) {
      x = creat("DEBUG.DEBUG", 0644);
      if ((sock = NEW ( Socket(Socket::NONSOCK, 0, 0, 0, x, 0, 0) ))) {
	 strcpy(s, ctime(&G->now));
	 sock->Printf("Debug (%s) written %s", G->VersionStr(), s);
	 sock->Printf("Full Version: %s\n", G->FullVersion());
	 sock->Printf("Context: mod/file/line\n");
	 for (x = 0; x < 16; x++) {
	    G->context_ptr = (G->context_ptr + 1) & 15;
	    sock->Printf("        %s\n", G->context_stack[G->context_ptr]);
	 }
	 delete sock;
      }
      exit (1); /* dont even try & tell people about, that may
		 * have caused the fault last time */
   } else
      nested_debug = 1;
   G->LOG_MISC.Put(G->Lang( 0x1D6, "s", G->context_stack[G->context_ptr]));
   x = creat("DEBUG", 0644);
   if (!(sock = new (0) Socket(Socket::NONSOCK,0,0,0,x,0,0) ))
     G->LOG_MISC.Put("* Failed to write DEBUG");
   else {
      strcpy(s, ctime(&G->now));
      sock->Printf("Debug (%s) written %s", G->VersionStr(), s);
      sock->Printf("Full Version: %s\n", G->FullVersion());
#ifdef STATIC_ONLY
      sock->Printf("ONLY STATICALLY LINKED\n");
#endif
      sock->Printf("Compile flags: %s\n", CCFLAGS);
      sock->Printf("Link flags   : %s\n", LDFLAGS);
      sock->Printf("Strip flags  : %s\n", STRIPFLAGS);
      sock->Printf("Context: mod/file/line\n");
      for (x = 0; x < 16; x++) {
	 G->context_ptr = (G->context_ptr + 1) & 15;
	 sock->Printf("        %s\n", G->context_stack[G->context_ptr]);
      }
      G->socklist->Display(sock);
      sock->Printf("\n");
      G->memtype_list->Debug(sock);
#ifdef EBUG_MEM
      Debugger::full_list->DumpStack(sock);
#endif
      delete sock;
      G->LOG_MISC.Put("* Wrote DEBUG");
   }
}

static BindReturn cmd_addlog (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   idx->Send( G->Lang(par[0] ? 0x031 : 0x032, ""));
   if (par[0]) 
      G->LOG_MISC.Put(G->Lang(0x033, "ss", idx->Nick(), par));
   return EXECUTED;
}

static BindReturn cmd_away (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * par = (char *)args[1];
   
   context;
   if ((mask[0] == 'a') && (strlen(par) > 60))
     par[60] = 0;
   idx->SetAway((mask[0] == 'a') ? par : 0);
   return EXECUTED;
}

static BindReturn cmd_banner (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0], * ptr = G->socklist;
   char * par = (char *)args[1], * s;
   
   if (!par[0]) 
     idx->Send( G->Lang(0x077, ""));
   else {
      s = G->Lang(0x078, "css", 7, idx->Nick(), par);
      while ((ptr = ptr->Next())) 
	if (ptr->TypeFlags() & DCT_CHAT)
	  ptr->Send(s);
   }
   return EXECUTED;
}

static BindReturn cmd_binds (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   G->LOG_CMDS.Put(G->Lang(0x076, "ss", idx->Nick(), par));
   G->bind_list->Display(idx,par);
   return EXECUTED;
}

static BindReturn cmd_boot (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0], * ptr = NULL;
   int ok = 0, langv = 0;
   char * who = 0, *par = (char *)args[1];
   FlagRecord fr, fr2(idx->User());
   
   if (!par[0]) 
     langv = 0x03E;
   else {
      who = G->Split(&par);
      ptr = G->socklist;
      while ((ptr = ptr->Next()))
	if (!strcasecmp(ptr->Nick(), who) && !ok &&
	    (ptr->TypeFlags() & DCT_CANBOOT)) {
	   fr.Set(ptr->User());
	   if (fr.GlobOwner() && strcasecmp(idx->Nick(), who)) 
	     langv = 0x03F;
	   else if (ptr->User() && fr.GlobMaster() && !fr2.GlobMaster())
	     langv = 0x040;
	   else {
	      langv =(ptr->TypeFlags() & DCT_FILES) ? 0x041 : 0x042;
	      G->LOG_CMDS.Put(G->Lang(0x044, "sss",idx->Nick(), who, par));
	      ((DccChat *)ptr)->DoBoot(idx->Nick(), par);
	      ok = 1;
	   }
	}
      if (!ok) 
	langv = 0x043;
   }   
   if (langv)
     idx->Send( G->Lang(langv, "s", ptr ? who : 0));
   return EXECUTED;
}

static BindReturn cmd_chat (void ** args, const char *) {
   FUNCTION;
   int newchan = -1, langv = 0;
   DccChat * idx = (DccChat *)args[0], *ptr = (DccChat *)G->socklist;
   char * par = (char *)args[1];
   void * argv[3] = { &newchan, par, 0 };
   
   if (!strcasecmp(par, "off")) {
      /* turn chat off */
      if (idx->Channel() < 0)
	langv = 0x094;
      else {
	 par = G->Lang(0x095, "");
	 idx->CheckLChpt(par);
	 langv = 0x096;
	 context;
      }
      idx->SetChannel(-1);
   } else {
      if (par[0] == '*') {
	 if (!par[1])
	   newchan = 0;
	 else if (((par[1] < '0') || (par[1] > '9'))) {
	    if (G->bt_name->Check("", 0, argv) != EXEC_LOG)
	      langv = 0x097;
	 } else
	   newchan = GLOBAL_CHANS + atoi(par + 1);
	 if (!langv && ((newchan < GLOBAL_CHANS) || 
			(newchan >= (GLOBAL_CHANS * 2)))) 
	   langv = 0x098;
      } else {
	 if (!strcasecmp(par, "on"))
	   newchan = 0;
	 else if (((par[0] < '0') || (par[0] > '9')) && par[0]) {
	   if (G->bt_name->Check("", 0, argv) != EXEC_LOG)
	     langv = 0x097;
	 } else
	   newchan = atoi(par);
	 if (!langv && ((newchan < 0) || (newchan >= (GLOBAL_CHANS * 2)))) 
	   langv = 0x099;
      }
      if (!langv) {
	 /* if coming back from being off the party line, make sure they're
         not away */
	 if ((idx->Channel() < 0) && idx->Away())
	   idx->SetAway(0);
	 if (idx->Channel() == newchan) 
	   langv = newchan ? 0x09B : 0x09A;
	 else {
	    while ((ptr = ptr->Next()))
	      if ((ptr->Channel() == newchan) && 
		  !strcasecmp(ptr->Nick(), idx->Nick()))
		break;
	    if (ptr) 
	      langv = 0x296;
	    else {
	       idx->SetChannel(newchan);
	       langv = newchan ? 0x09D : 0x09C;
	       idx->CheckLChjn();
	    }
	 }
      }
   }
   if (langv)
     idx->Send( G->Lang(langv, "s", par));
   return EXECUTED;
}

static BindReturn cmd_chattr (void ** args, const char *) {
   FUNCTION;
   char * hand = NULL, * chg = NULL, work[1024], *par = (char *)args[1];
   Channel *chan = NULL;
   FlagRecord pls, mns, user;
   int fl = -1, of = 0, ocf = 0, langv = 0;
   DccChat * idx = (DccChat *)args[0];
   UserRecord* u2 = NULL, *u = idx->User();
   
   if (!par[0]) 
     langv = 0x089;
   else if (((hand = G->Split(&par))[0] == '*') || 
	    !(u2 = G->userlist->Find(hand)))
     langv = 0x100;
   else {
      if ((par[0] != '#') && (par[0] != '&')) {
	 chg = G->Split(&par);
	 if (!par[0] && strpbrk(chg,"&|")) 
	   par = idx->ConsoleChan();
	 else if (par[0] && !strpbrk(chg,"&|"))
	   fl = ~FlagRecord::GLOBAL;
      }
      if (!(chan = G->chanset->Find(par)) && par[0] && (par[0] != '*'))
	langv = 0x08A;
      else {
	 user.SetMatch (FlagRecord::GLOBAL);
	 if (chan)
	   user.SetMatch (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
	 user.Set(u, chan ? chan->Key() : 0);
	 if (!chan && !user.GlobBotmast())
	   langv = 0x08B;
	 else if (chan && !user.AnyMaster())
	   langv = 0x08C;
	 else {
	    user.SetMatch(user.Match() & fl);
	    if (chg) {
	       pls.SetMatch(user.Match());
	       pls.BreakDown(chg, &mns);
	       /* no-one can change these flags on-the-fly */
	       pls.MaskBot();
	       mns.MaskBot();
	       if (chan) {
		  pls.ChanMaskShare();
		  mns.ChanMaskShare();
	       }
	       if (!user.GlobOwner()) {
		  pls.MaskNonOwner();
		  mns.MaskNonOwner();
		  if (chan) {
		     pls.ChanMaskOwner();
		     mns.ChanMaskOwner();
		  }
		  if (!user.GlobMaster()) {
		     pls.MaskNonMaster();
		     mns.MaskNonMaster();
		     if (!user.GlobBotmast()) {
			pls.ClearGlobal();
			mns.ClearGlobal();
		     }
		  }
	       }
	       if (chan && !user.AnyOwner()) {
		  pls.ChanMaskMaster();
		  mns.ChanMaskMaster();
		  if (!user.AnyMaster()) {
		     pls.ClearChan();
		     mns.ClearChan();
		  }
	       }	   
	       user.Set(u2,par);
	       if (user.Match() & FlagRecord::GLOBAL) 
		 of = user.Global();
	       if (chan) 
		 ocf = user.Chan();
	       user += pls;
	       user -= mns;
	       user.Sanity();
	       u2->Set(user,par);
	    }
	    G->LOG_CMDS.Put(G->Lang(chan ? 0x091 : (chg ? 0x092 : 0x093),
			      "ssss", idx->Nick(), hand, chg, 
			      chan ? chan->Key() : ""));
	    if (user.Match() & FlagRecord::GLOBAL) {
	       int tsock = idx->FD();
	       
	       user.SetMatch(FlagRecord::GLOBAL);
	       if (chg)
		 u2->CheckDccAttrs(of);
	       /* check_dcc_attrs may move idx's */
	       idx = (DccChat *)G->socklist->Find(tsock);
	       user.Set(u2);
	       user.Build(work);
	       langv = (work[0] == '-') ? 0x08E : 0x08D;
	    }
	    if (chan) {
	       user.SetMatch(FlagRecord::CHANNEL);
	       user.Set(u2,par);
	       user.ClearChanShare();
	       if (chg)
		 u2->CheckDccChanAttrs(chan->Key(), ocf);
	       user.Build(work);
	       langv = (work[0] == '-') ? 0x090 : 0x08F;
#ifdef HAVE_IRC_H
	       if (G->module_list->Find("irc", &irc_ver))
		 ((Channel3 *)chan)->Recheck(0);
#endif
	    }
	    par = work;
	 }
      }
   }
   if (langv) 
     idx->Send( G->Lang(langv, "sss", hand, chan ? chan->Key() : "",
			   par));
   return EXECUTED;
}

static BindReturn cmd_chnick (void ** args, const char * mask) {
   FUNCTION;
   const char * hand;
   char * newhand = "", *par = (char *)args[1], lognick[HANDLEN + 1];
   Socket * idx = (Socket *)args[0];
   int langv = 0;
   unsigned int i;
   UserRecord * u2;
   FlagRecord atr (idx->User(), FlagRecord::GLOBAL | FlagRecord::CHANNEL
		   | FlagRecord::ANYWHERE);
   FlagRecord atr2 (FlagRecord::GLOBAL | FlagRecord::BOT);
   
   strcpy(lognick, idx->Nick());
   if (mask[0] == 'c') 
     hand = G->Split(&par);
   else 
     hand = idx->Nick();
   if (!hand[0] || !par[0]) 
     langv = (mask[0] == 'c') ? 0x04C : 0x054;
   else {
      newhand = G->Split(&par);
      if (strlen(newhand) > HANDLEN)
	newhand[HANDLEN] = 0;
      for (i = 0; i < strlen(newhand); i++)
	if ((newhand[i] <= 32) || (newhand[i] >= 127) || (newhand[i] == '@'))
	  newhand[i] = '?';
      if (strchr("-,+*=:!.@#;$", newhand[0]))
	langv = 0x04D;
      else if (G->userlist->Find(newhand) && strcasecmp(hand, newhand))
	langv = 0x04E;
      else if (!strcasecmp(newhand, G->origbotname) || 
	       !strcasecmp(newhand, G->botnetnick))
	langv = 0x04F;
      else {
	 u2 = G->userlist->Find(hand);
	 if (mask[0] == 'c') {
	    atr2.Set(u2);
	    if (atr.GlobBotmast() && !atr.AnyMaster() && !atr2.GlobBot())
	      langv = 0x050;
	    else if (atr2.BotShare() && !atr.GlobOwner())
	      langv = 0x051;
	    else if (atr2.GlobOwner() && !atr.GlobOwner() 
		     && strcasecmp(idx->Nick(), hand))
	      langv = 0x052;
	 }
	 if (!langv) {
	   if (u2->ChangeHandle(newhand)) {
	      G->LOG_CMDS.Put(G->Lang(mask[0] == 'c' ? 0x053 : 0x055,
				 "sss", lognick, hand, newhand));
	      langv = 0x11C;
	   } else 
	   langv = 0x104;
         }
      }
   }
   if (langv) 
     idx->Send( G->Lang(langv, "cs", newhand[0], newhand));
   return EXECUTED;
}

static BindReturn cmd_chpass (void ** args, const char *) {
   FUNCTION;
   char * handle, * newp, *par = (char *)args[1];
   Socket * idx = (Socket *)args[0];
   int l, langv = 0;
   UserRecord * u;
   FlagRecord atr (idx->User()), atr2 (FlagRecord::BOT | FlagRecord::GLOBAL);

   if (!par[0])
     langv = 0x056;
   else {
      handle = G->Split(&par);
      if (!(u = G->userlist->Find(handle)))
	langv = 0x100;
      else if (!(atr2.Set(u)->GlobBot()) && atr.GlobBotmast() 
	       && !atr.GlobMaster())
	langv = 0x057;
      else if (atr2.BotShare() && !atr.GlobOwner())
	langv = 0x058;
      else if (atr2.GlobOwner() && !atr.GlobOwner()
	       && strcasecmp(handle, idx->Nick()))
	langv = 0x059;
      else if (!par[0]) {
	 G->LOG_CMDS.Put(G->Lang(0x05D, "ss", idx->Nick(), handle));
	 SetPassword(u, NULL);
	 langv = 0x05A;
      } else {
	 l = strlen(newp = G->Split(&par));
	 if (l > 16)
	   newp[16] = 0;
	 if (l < 6) 
	   langv = 0x05B;
	 else {
	    SetPassword(u, newp);
	    G->LOG_CMDS.Put(G->Lang(0x05D, "ss", idx->Nick(), handle));
	    langv = 0x05C;
	 }
      }
   }
   if (langv)
     idx->Send( G->Lang(langv, ""));
   return EXECUTED;
}

static BindReturn cmd_comment (void ** args, const char *) {
   FUNCTION;
   char * handle, *par = (char *)args[1];
   UserRecord * u1;
   Socket * idx = (Socket *)args[0];
   int langv = 0;
   UserEntryComment * uec;
   FlagRecord fr;
   
   if (!par[0])
     langv = 0x05E;
   else {
      handle = G->Split(&par);
      if (!(u1 = G->userlist->Find(handle)))
	langv = 0x100;
      else if (fr.Set(u1)->GlobOwner() && !fr.Set(idx->User())->GlobOwner()
	       && strcasecmp(handle, idx->Nick()))
	langv = 0x05F;
      else {
	 G->LOG_CMDS.Put(G->Lang(0x062, "sss", idx->Nick(), handle, par));
	 if (!strcasecmp(par, "none")) {
	    langv = 0x060;
	    par = NULL;
	 } else
	   langv = 0x061;
	 uec = (UserEntryComment *)u1->GetEntry(G->entrytype_comment);
	 uec->Set(par);
      }
   }
   if (langv)
     idx->Send( G->Lang(langv, ""));
   return EXECUTED;
}
   
static BindReturn cmd_console (void ** args, const char *) {
   FUNCTION;
   char * nick, s[2], s1[512], *par = (char *)args[1];
   int ok = 0, pls, langv = 0;
   Socket * ptr = G->socklist;
   DccChat * idx = (DccChat *)args[0], * dest = idx;
   FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
   LogModes md;
   
   if (!par[0]) 
     langv = 0x045;
   else {
      fr.Set(idx->User(),idx->ConsoleChan());
      strcpy(s1, par);
      nick = G->Split(&par);
      if (nick[0] && !strchr("#&+-*",nick[0]) && fr.GlobMaster()) {
	 while ((ptr = ptr->Next())) 
	   if (!strcasecmp(nick, ptr->Nick())
	       && (ptr->TypeFlags() & DCT_CHAT) && !ok) {
	      ok = 1;
	      dest = (DccChat *)ptr;
	   }
	 if (!ok) {
	    idx->Send( G->Lang(0x043, ""));
	    return EXECUTED;
	 }
	 nick[0] = 0;
      }
      if (!nick[0])
	nick = G->Split(&par);
      if (strchr("#&*",nick[0])) {
	 if ((nick[0] != '*') && !G->chanset->Find(nick)) {
	    idx->Send( G->Lang(0x046, "s", nick));
	    return EXECUTED;
	 }
	 fr.Set(idx->User(),nick);
	 if (!fr.AnyValidOp()) {
	    idx->Send( G->Lang(0x047, "s", nick));
	    return EXECUTED;
	 }
	 dest->SetConsoleChan(nick);
	 nick[0] = 0;
	 if ((dest == idx) && !fr.AnyMaster())
	    dest->ConsoleFlags().Remove(LogModes::MISC | LogModes::CMDS 
					| LogModes::WALL);
      }
      if (!nick[0])
	nick = G->Split(&par);
      pls = 1;
      if (nick[0]) {
	 if ((nick[0] != '+') && (nick[0] != '-'))
	   dest->ConsoleFlags().Clear();
	 for (; *nick; nick++) {
	    if (*nick == '+')
	      pls = 1;
	    else if (*nick == '-')
	      pls = 0;
	    else {
	       s[0] = *nick;
	       s[1] = 0;
	       md.Set(s);
	       if ((dest == idx) && !fr.GlobMaster() && pls) 
		 md.Remove ( LogModes::FILES | LogModes::LEVS 
			    | (fr.ChanMaster() ? 0 : LogModes::MISC |
			       LogModes::CMDS | LogModes::WALL) );
	       if (!fr.GlobOwner() && pls) 
		 md.Remove ( LogModes::RAW | LogModes::SRVOUT | 
			    LogModes::BOTNET | LogModes::BOTSHARE );
	       if (!fr.GlobBotmast() && pls) 
		 md.Remove ( LogModes::BOTS );
	       if (pls)
		 dest->ConsoleFlags() += md;
	       else
		 dest->ConsoleFlags() -= md;
	    }
	 }
      }
      G->LOG_CMDS.Put(G->Lang(0x048, "ss", idx->Nick(), s1));
      if (dest == idx) 
	langv = 0x049;
      else {
	 langv = 0x04A;
	 dest->Send( G->Lang(0x04B, "ssss", idx->Nick(),
			  dest->ConsoleChan(),
			  dest->ConsoleFlags().Mask(),
			  dest->ConsoleFlags().Names()));
      }
   }
   if (langv) 
     idx->Send( G->Lang(langv, "ssss", dest->Nick(), dest->ConsoleChan(),
		     dest->ConsoleFlags().Mask(),
		     dest->ConsoleFlags().Names()));
   return EXECUTED;
}

static BindReturn cmd_dccstat (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   G->LOG_CMDS.Put(G->Lang(0x03D, "s", idx->Nick()));
   G->socklist->Display(idx);
   return EXECUTED;
}

static BindReturn cmd_debug (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   G->LOG_CMDS.Put(G->Lang(0x070, "ss", idx->Nick(), par));
   if (!strcasecmp(par,"help"))
     G->help_list->Debug(idx);
   else if (!strcasecmp(par, "write")) {
      write_debug();
      nested_debug = 0;
   } else if (par[0] == 'm') 
	G->memtype_list->Dump(idx, par+1);
   else 
     G->memtype_list->Debug(idx);
   return EXECUTED;
}

static BindReturn cmd_die (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char *par = args[1] ? (char *)args[1] : "";
   
   G->LOG_CMDS.Put(G->Lang(0x06D, "ss", idx->Nick(), par));
   G->userlist->WriteUserfile(NULL);
   G->LOG_MISC.Put(G->Lang(par[0] ? 0x06E : 0x06F, "ss", idx->Nick(), par));
   G->Fatal(par[0] ? par : idx->Nick());
   return EXECUTED;
}

static BindReturn cmd_echo (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   int langv = 0x09E;
   
   if (!par[0]) 
     par = G->Lang( idx->HasStatus(STAT_ECHO) ? 0x11D : 0x11E, "");
   else if (!strcasecmp(par, "on")) {
      langv = 0x09F;
      idx->SetStatus(STAT_ECHO);
   } else if (!strcasecmp(par, "off")) {
      langv = 0x0A0;
      idx->ClearStatus(STAT_ECHO);
   } else
     langv = 0x0A1;
   idx->Send( G->Lang(langv, "s", par));
   return EXECUTED;
}
   
static BindReturn cmd_fixcodes (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   int langv = 0;
   
   if (idx->HasStatus(STAT_ECHO)) {
      idx->SetStatus(STAT_TELNET);
      idx->ClearStatus(STAT_ECHO);
      langv = 0x0B5;
   } else if (idx->HasStatus(STAT_TELNET)) {
      idx->SetStatus(STAT_ECHO);
      idx->ClearStatus(STAT_TELNET);	
      langv = 0x0B7;
   }
   if (langv) {
      idx->Send( G->Lang(langv, ""));
      G->LOG_CMDS.Put(G->Lang(langv + 1, "s", idx->Nick()));
   }
   return EXECUTED;
}

static BindReturn cmd_help (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * par = (char *)args[1];
   FlagRecord fr (idx->User(), FlagRecord::GLOBAL | FlagRecord::CHANNEL,
		  idx->ConsoleChan());
   int search = strlen(mask) > 4;
   
   context;
   G->RMSpace(par); /* maintain some sanity */
   G->LOG_CMDS.Put(G->Lang(search ? 0x297 : 0x030, "ss", idx->Nick(), par));
   if (search)
     G->help_list->Search(idx, par, fr);
   else
     G->help_list->Tell(idx, par[0] ? par : (fr.AnyValidOp() ? "help" : 
					     "helpparty"), fr, 0);
   return EXECUTED;
}

static BindReturn cmd_hooks (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   context;
   G->LOG_CMDS.Put(G->Lang(0x290, "s", idx->Nick()));
   G->hook_list->Dump(idx);
   return EXECUTED;
}
   
static BindReturn cmd_ignores (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   G->LOG_CMDS.Put(G->Lang(0x0D9, "ss", idx->Nick(), par));
   G->ignores->Display(idx, par);
   return EXECUTED;
}
   
static BindReturn cmd_languagedump (void ** args, const char *) {
   FUNCTION;
   char	* t, * par = (char *)args[1];
   Socket * idx = (Socket *)args[0];
   int idx2;
   ModuleLanguage * mlang;
   LanguageEntry * le;
   
   context;
   G->LOG_CMDS.Put(0, "#%s# ldump %s", idx->Nick(), par);
   t = G->Split(&par);
   if (t[0]) {
      if (!(mlang = (ModuleLanguage *)G->langtab_list->Find(t))) 
	idx->Printf("Unknown module.\n");
      else {
	 if (par[0]) {
	    if (strlen(par) > 2 && par[0] == '0' && par[1] == 'x')
	      sscanf(par, "%x", &idx2);
	    else
	      idx2 = (int) strtol(par, (char **)NULL, 10);
	    if ((le = (LanguageEntry *)mlang->Entry(idx2)->Find(idx2))) 
	      le->Dump(0,idx);
	    else
	      idx->Printf("No such language entry: %s [0x%x]\n", t, idx2);
	 } else {
	    idx->Printf("Language dump for module: %s\n",t);
	    mlang->Dump(idx);
	 }
      }
      return EXECUTED;
   }
   idx->Printf("MODULE     LANGIDX TEXT\n");
   for (mlang = G->langtab_list; mlang; mlang = mlang->Next())
      mlang->Dump (idx);
   return EXECUTED;
}

static BindReturn cmd_languagestatus (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   int ltexts = 0, maxdepth = 0, used = 0, empty = 0, i, c, mlc = 0;
   ModuleLanguage * mlang;
   char work[512];
   
   for (mlang = G->langtab_list; mlang; mlang = mlang->Next()) {
      mlc++;
      for (i = 0; i < 32; i++) {
	 c = mlang->Entry(i)->Count();
	 if (c > maxdepth)
	   maxdepth = c;
	 if (c)
	   used++;
	 else
	   empty++;
	 G->SPrintF(work + i * 4, 4, "%2, ", c);
	 ltexts += c;
      }
      idx->Printf("Module %s: %s\n", mlang->Key(), work);
   }
   context;
   idx->Printf("language code report: %d modules\n", mlc);
   idx->Printf("   Table size: %d bytes\n", G->memtype_language->ExpMem());
   idx->Printf("   Text messages: %d\n", ltexts);
   idx->Printf("   %d used, %d unused, maxdepth %d, avg %f\n",
		 used, empty, maxdepth, used ? (float)ltexts/used : 0.0);
   return EXECUTED;
}

static BindReturn cmd_loadconfig (void ** args, const char * x) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   
   if (!G->savedconfigfile[0]) 
     idx->Send(G->Lang(0x02A5, ""));
   else if (*x == 'l')
     G->LoadConfig(idx);
   else
     G->SaveConfig(idx);
   return EXECUTED;
}

static BindReturn cmd_loadlanguage (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   if (!par || !par[0]) 
     idx->Printf("Usage: language <language>\n");
   else {
      G->LOG_CMDS.Put(0, "#%s# language %s", idx->Nick(), par);
      if (!G->LoadLanguage(par))
	idx->Printf("Can't load language file: '%s'\n", par);
   }
   return EXECUTED;
}

static BindReturn cmd_loadmod (void ** args, const char * mask) {
   FUNCTION;
   const char *p = NULL;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   int langv = (mask[0] == 'u') ? 0x0C9 : 0x0CD;
   
   context;
   if (par[0]) {
      if ((mask[0] == 'u') && (p = G->module_list->Unload(par)))
	langv++;
      else if ((mask[0] == 'l') && (p = G->module_list->Load(par)))
	langv++;
      else {
	 langv += 2;
	 G->LOG_CMDS.Put(G->Lang(langv + 1, "ss", idx->Nick(), par));
      }
   }
   idx->Send( G->Lang(langv, "ss", par, p));
   context;
   return EXECUTED;
}

static BindReturn cmd_match (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int start = 1, limit = 20;
   char * par = (char *)args[1], * s, * s1, * chname;
   
   if (!par[0]) 
     idx->Send( G->Lang(0x038, ""));
   else {
      G->LOG_CMDS.Put(G->Lang(0x039, "ss", idx->Nick(), par));
      s = G->Split(&par);
      chname = ((par[0] == '#') || (par[0] == '+') || (par[0] == '&')) ?
	G->Split(&par) : "";
      if (atoi(par) > 0) {
	 s1 = G->Split(&par);
	 if (atoi(par) > 0) {
	    start = atoi(s1);
	    limit = atoi(par);
	 } else
	   limit = atoi(s1);
      }
      G->userlist->DisplayMatch(idx, s, start, limit, chname);
   }
   return EXECUTED;
}

static BindReturn cmd_me (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * par = (char *)args[1];
   void * argv[3];
   int t;
   
   context;
   if (idx->Channel() < 0) 
     idx->Send( G->Lang(0x025, ""));
   else if (!par[0])
     idx->Send( G->Lang(0x026, ""));
   else {
      if (idx->Away())
	idx->SetAway(0);
      if (idx->HasStatus(STAT_ECHO))
	idx->Chanout (idx->Channel(),G->Lang(0x110,"ss",idx->Nick(),par));
      else
	idx->ChanoutBut(idx->Channel(),G->Lang(0x110,"ss",idx->Nick(),par));
      argv[0] = (void *)idx->Nick();
      t = idx->Channel();
      argv[1] = &t;
      argv[2] = par;
      G->bt_act->Check(par, 0, argv);
   }
   return EXECUTED;
}

static BindReturn cmd_mns_host (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   const char * handle = NULL;
   char *host = NULL, *par = (char *)args[1];
   UserRecord * u2, *u = idx->User();
   int langv = 0x0ED;
   FlagRecord fr (FlagRecord::CHANNEL | FlagRecord::ANYWHERE | FlagRecord::GLOBAL);
   FlagRecord fr2 (FlagRecord::GLOBAL | FlagRecord::BOT);
   
   if (par[0]) {
      host = G->Split(&par);
      if (par[0]) {
	 handle = host;
	 host = G->Split(&par);
	 u2 = G->userlist->Find(handle);
      } else {
	 handle = idx->Nick();
	 u2 = u;
      }
      if (!u2 || !u)
	langv = 0x100;
      else {
	 fr.Set(u);
	 fr2.Set(u2);
	 if ((u != u2) && !fr2.GlobBot() && !fr.AnyMaster() && fr.GlobBotmast())
	   langv = 0x0EE;
	 else if (fr2.GlobBot() && fr2.BotShare() && !fr.GlobOwner())
	   langv = 0x0EF;
	 else if (fr2.GlobOwner() && !fr.GlobOwner() && (u2 != u))
	   langv = 0x0F0;
	 else if ((u != u2) && !fr.GlobBotmast() && !fr.AnyMaster())
	   langv = 0x29E;
	 else if (u2->DelHost(host)) {
	    G->LOG_CMDS.Put(G->Lang(0x0F2, "sss", idx->Nick(), handle, host));
	    langv = 0x0F1;
	 } else
	   langv = 0x104;
	 }
   }
   idx->Send( G->Lang(langv, "ss", handle, host));
   return EXECUTED;
}

static BindReturn cmd_mns_ignore (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char par[UHOSTLEN+1];
   int langv = 0x0D5;
   IgnoreRecord * igrec;
   
   context;
   strncpyz(par, (char *)args[1], UHOSTLEN);
   if (par[0] && (igrec = G->ignores->Find(par))) {
      DELETE ( igrec );
      G->LOG_CMDS.Put(G->Lang(0x0D8, "ss", idx->Nick(), par));
      langv = 0x0D6;
   } else if (par[0])
	langv = 0x0D7;
   idx->Send( G->Lang(langv, "s", par));
   return EXECUTED;
}

static BindReturn cmd_mns_user (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   UserRecord * u2 = G->userlist->Find(par), *u = idx->User();
   int langv = 0x104;
   FlagRecord fr (u), fr2(u2,FlagRecord::GLOBAL | FlagRecord::BOT);
   
   if (!par[0])
     langv = 0x0E0;
   else if (!u2 || !u)
     langv = 0x100;
   else if (fr2.GlobOwner() && !fr.GlobOwner())
     langv = 0x0E1;
   else if (fr2.GlobBot() && fr2.BotShare() && !fr.GlobOwner())
     langv = 0x0E2;
   else if (!fr2.GlobBot() && fr.GlobBotmast() && !fr.GlobMaster())
     langv = 0x0E3;
   else {
      DELETE ( u2 );
      G->LOG_CMDS.Put(G->Lang(0x0E5, "ss", idx->Nick(), par));
      langv = 0x0E4;
   }
   idx->Send( G->Lang(langv, "s", par));
   return EXECUTED;
}

static BindReturn cmd_module (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   context;
   G->LOG_CMDS.Put(G->Lang(0x0C8, "ss", idx->Nick(), par));
   if (par[0] == '*')
     G->module_list->Report(idx,0,NULL);
   else
     G->module_list->Report(idx,2,par[0] ? par : 0);
   return EXECUTED;
}

static BindReturn cmd_motd (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * par = (char *)args[1];
   
   context;
   if (par[0]) 
     idx->Send( G->Lang(0x027, ""));
   else {
      G->LOG_CMDS.Put(G->Lang(0x028, "s", idx->Nick()));
      idx->ShowMotd();
   }
   return EXECUTED;
}

static BindReturn cmd_newpass (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char *newp = NULL, *par = (char *)args[1];
   int langv = 0;
   
   context;
   if (!par[0])
     langv = 0x02A;
   else {
      newp = G->Split(&par);
      if (strlen(newp) > 16)
	newp[16] = 0;
      if (strlen(newp) < 6)
	langv = 0x02B;
      else {
	 SetPassword(idx->User(),newp);
	 G->LOG_CMDS.Put(G->Lang(0x02D, "s", idx->Nick()));
	 langv = 0x02C;
      }
   }
   idx->Send( G->Lang(langv, "s", newp));
   return EXECUTED;
}

static BindReturn cmd_note (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   int echo;
   char * par = (char *)args[1], handle[512], *p;

   context;
   if (!par[0])
     idx->Send( G->Lang(0x029, ""));
   else {
      p = G->Split(&par);
      while ((*par == ' ') || (*par == '<') || (*par == '>'))
	par++; /* these are now illegal *starting* notes characters */
      echo = idx->HasStatus(STAT_ECHO);
      G->SplitC(handle,p,',');
      while (handle[0]) {
	 G->RMSpace(handle);
	 G->AddNote(handle, idx->Nick(), par, idx, echo);
	 G->SplitC(handle,p,',');
      }
      G->RMSpace(p);
      G->AddNote(p, idx->Nick(), par, idx, echo);
   }
   return EXECUTED;
}

static BindReturn cmd_page (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int a = 0, langv = 0;
   char * par = (char *)args[1];

   if (!par[0]) 
     langv = idx->HasStatus(STAT_PAGE) ? 0x0B9 : 0x0BA;
   else if (!strcasecmp(par, "off")) {
      idx->ClearStatus(STAT_PAGE);
      idx->SetMaxLines(0x7ffffff);	/* flush_lines needs this */
      while (idx->Buffer()->Next())
	idx->FlushLines();
      langv = 0x0BB;
      G->LOG_CMDS.Put(G->Lang(0x0BE, "s", idx->Nick()));
   } else if ((a = atoi(par))) {
      idx->SetStatus(STAT_PAGE);
      idx->SetMaxLines(a);
      idx->ClearCount ();
      G->LOG_CMDS.Put(G->Lang(0x0BF, "sd", idx->Nick(), a));
      langv = 0x0BC;
   } else
     langv = 0x0BD;
   idx->Send( G->Lang(langv, "d", idx->MaxLines()));
   return EXECUTED;
}

static BindReturn cmd_pls_host (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1], * handle, * host = NULL;
   UserRecord * u2, *u = idx->User();
   int langv = 0x0E6;
   FlagRecord fr (FlagRecord::CHANNEL | FlagRecord::ANYWHERE | FlagRecord::GLOBAL);
   FlagRecord fr2 (FlagRecord::GLOBAL | FlagRecord::BOT);
   UserEntryHosts * ueh;
   
   context;
   handle = G->Split(&par);
   if (par[0]) {
      host = G->Split(&par);
      if (!(u2 = G->userlist->Find(handle)))
	langv = 0x100;
      else {
	 ueh = (UserEntryHosts *)u2->GetEntry(G->entrytype_hosts);
	 if (ueh->Find(host))
	   langv = 0x0E7;
	 else {
	    fr.Set(u);
	    fr2.Set(u2);
	    if (!fr2.GlobBot() && fr.GlobBotmast() && !fr.AnyMaster())
	      langv = 0x0E8;
	    else if (fr2.GlobBot() && fr2.BotShare() && !fr.GlobOwner())
	      langv = 0x0E9;
	    else if (fr2.GlobOwner() && !fr.GlobOwner() &&
		     strcasecmp(handle, idx->Nick()))
	      langv = 0x0EA;
	    else {
	       G->LOG_CMDS.Put(G->Lang(0x0EC, "sss", idx->Nick(), handle, host));
	       u2->AddHost(host);
	       langv = 0x0EB;
	    }
	 }
      }
   }
   idx->Send( G->Lang (langv, "ss", handle, host));
   return EXECUTED;
}

static BindReturn cmd_pls_ignore (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1], *who = NULL, t[UHOSTLEN + 1];
   int langv = 0x0D1;
     
   context;
   if (par[0]) {
      who = G->Split(&par);
      if (strlen(par) > 65)
	par[65] =0 ;
      if (G->ignores->Match(who))
	langv =0x0D2;
      else {
	 if (!par[0])
	   par = G->Lang(0x29D, "");
	 if (!strchr(who, '!')) {
	    if (!strchr(who, '@'))
	      G->SPrintF(t, UHOSTLEN, "%s!*@*", who);
	    else 
	      G->SPrintF(t, UHOSTLEN, "*!%s", who);
	 } else if (!strchr(who, '@')) 
	      G->SPrintF(t, UHOSTLEN, "%s@*", who);
	 else
	   strncpyz(t, who, UHOSTLEN);
	 langv = 0x0D3;
	 NEW ( IgnoreRecord ( who, idx->Nick(), par, 0 ) );
	 G->LOG_CMDS.Put(G->Lang(0x0D4, "sss", idx->Nick(), who, par));
      }
   }
   idx->Send( G->Lang(langv, "ss", who, par));
   return EXECUTED;
}

static BindReturn cmd_pls_user (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1], *handle = "", *host = NULL;
   int langv = 0x0DA;
   
   context;
   if (par[0]) {
      handle = G->Split(&par);
      host = G->Split(&par);
      if (strlen(handle) > HANDLEN)
	handle[HANDLEN] = 0;		/* max len = XX */
      if (G->userlist->Find(handle))
	langv = 0x0DB;
      else if (strchr("-,+*=:!.@#;$", handle[0]) != NULL) 
	langv = 0x0DC;
      else {
	 FlagRecord fr;
	 
	 G->LOG_CMDS.Put(G->Lang(0x0DF, "sss", idx->Nick(), handle, host));
	 fr.SetGlobal(G->default_flags);
	 fr.Sanity();
	 NEW ( UserRecord ( G->userlist, handle, host, fr) );
	 langv = host[0] ? 0x0DD : 0x0DE;
      }
   }
   idx->Send( G->Lang(langv, "ssc", handle, host, handle[0]));
   return EXECUTED;
}

static BindReturn cmd_quit (void **, const char *) {
   FUNCTION;
   return EXEC_BRK;
}

static BindReturn cmd_restart (void ** args, const char * mask) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   int langv = (mask[2] == 's') ? 0x2A1 : 0x066;
   
   G->LOG_CMDS.Put(G->Lang(langv, "s", idx->Nick()));
   if (G->make_userfile) {
      G->LOG_MISC.Put(G->Lang(0x06A, ""));
      G->make_userfile = 0;
   }
   G->userlist->WriteUserfile(0);
   G->LOG_MISC.Put(G->Lang(langv + 2, ""));
   G->do_restart = (langv == 0x2A1) ? (idx->FD() + 1) : -2;
   idx->Send(G->Lang(langv + 1, ""));
   return EXECUTED;
}

static BindReturn cmd_rehelp (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   context;
   G->LOG_CMDS.Put(G->Lang(0x02E, "s", idx->Nick()));
   idx->Send( G->Lang(0x02F, ""));
   G->help_list->Reload();
   return EXECUTED;
}
   
static BindReturn cmd_reload (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   G->LOG_CMDS.Put(G->Lang(0x06B, "s", idx->Nick()));
   idx->Send( G->Lang(0x06C, ""));
   G->Reload();
   return EXECUTED;
}

static BindReturn cmd_save (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   G->LOG_CMDS.Put(G->Lang(0x074, "s", idx->Nick()));
   idx->Send( G->Lang(0x075, ""));
   G->userlist->WriteUserfile(NULL);
   return EXECUTED;
}

/* perform a 'set' command */
static BindReturn cmd_set (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * msg = (char *)args[1], * var;
   
   G->LOG_CMDS.Put(G->Lang(0x0C3, "ss", idx->Nick(), msg));
   var = G->Split(&msg);
   if (!var[0]) 
     G->scripts->DumpVars(idx);
   else if (!msg[0])
     G->scripts->DumpVar(idx, var);
   else {
      FlagRecord fr(idx->User());
      
      if (!fr.GlobOwner())
	return NOMATCH;
      G->scripts->SetVar(idx, var, msg);
   }
   return EXECUTED;
}

static BindReturn cmd_simul (void ** args, const char *) {
   FUNCTION;
   char * nick, *par = (char *)args[1];
   Socket * idx = (Socket *)args[0], * ptr = G->socklist;
   
   nick = G->Split(&par);
   if (!par[0]) 
     idx->Send( G->Lang(0x071, ""));
   else {
      while ((ptr = ptr->Next()))
	if (!strcasecmp(nick, ptr->Nick()) &&
	    (ptr->TypeFlags() & DCT_SIMUL)) {
	   G->LOG_CMDS.Put(G->Lang(0x073, "sss", idx->Nick(), nick, par));
	   ptr->Activity(par,strlen(par));
	   break;
	}
      if (!ptr)
	idx->Send( G->Lang(0x072, ""));
   }
   return EXECUTED;
}

static BindReturn cmd_status (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   FlagRecord fr (idx->User());
   
   if (!strcasecmp(par, "all")) {
      if (!fr.GlobMaster())
	idx->Send( G->Lang(0x03A, ""));
      else {
	 G->LOG_CMDS.Put(G->Lang(0x03B, "s", idx->Nick()));
	 idx->TellVerboseStatus(1);
      }
   } else {
      G->LOG_CMDS.Put(G->Lang(0x03C, "s", idx->Nick()));
      idx->TellVerboseStatus(0);
   }
   return EXECUTED;
}

static char *stripmaskname(int x) {
   FUNCTION;
   static char s[161];
   int i = 0;
   
   s[i] = 0;
   if (x & STRIP_BOLD)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A2, ""));
   if (x & STRIP_COLOR)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A3, ""));
   if (x & STRIP_REV)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A4, ""));
   if (x & STRIP_UNDER)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A5, ""));
   if (x & STRIP_ANSI)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A6, ""));
   if (x & STRIP_BELLS)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A7, ""));
   if (x & STRIP_HILITE)
     i += G->SPrintF(s + i, 160 - i, "%s, ", G->Lang(0x0A8, ""));
   if (!i)
     strcpy(s, G->Lang(0x0A9, ""));
   else 
     s[i-2] = 0;
   return s;
}

static BindReturn cmd_strip (void ** args, const char *) {
   FUNCTION;
   char * nick, *changes, s[2], *par = (char *)args[1];
   int pls, md, langv = 0x294;
   DccChat * idx = (DccChat *)args[0], * dest = idx;
   FlagRecord fr (idx->User());
   
   if (par[0]) {
      nick = G->Split(&par);
      if ((nick[0] != '+') && (nick[0] != '-') && fr.GlobMaster()) {
	 dest = (DccChat *)G->socklist;
	 while ((dest = (DccChat *)dest->Next()))
	   if (!strcasecmp(nick, dest->Nick()) && (dest->TypeFlags() & DCT_CHAT)) 
	     break;
	 if (!dest) {
	    idx->Send( G->Lang(0x072, ""));
	    return EXECUTED;
	 }
	 changes = par;
      } else { 
	 changes = nick;
	 nick = "";
	 dest = idx;
      }
      if ((changes[0] != '+') && (changes[0] != '-'))
	dest->SetStripFlags(0);
      s[1] = 0;
      for (pls = 1; *changes; changes++) {
	 switch (*changes) {
	   case '+':
	    pls = 1;
	    break;
	   case '-':
	    pls = 0;
	    break;
	   default:
	    s[0] = *changes;
	    md = G->StripModes(s);
	    if (pls == 1)
	      dest->AddStripFlags(md);
	    else
	      dest->DelStripFlags(md);
	 }
      }
      G->LOG_CMDS.Put(G->Lang(0x0AD, "sss", idx->Nick(), nick, changes));
      if (dest == idx) 
	langv = 0x0AA;
      else {
	 langv = 0x0AB;
	 dest->Send(G->Lang(0x0AC, "sss", idx->Nick(),
			    G->StripMaskType(dest->StripFlags()),
			    stripmaskname(dest->StripFlags())));
      }
   }
   idx->Send( G->Lang(langv, "sss", G->StripMaskType(dest->StripFlags()),
		      stripmaskname(dest->StripFlags()),dest->Nick()));
   return EXECUTED;
}
   
static BindReturn cmd_su (void ** args, const char *) {
   FUNCTION;
   char * par = (char *)args[1];
   DccChat * idx = (DccChat *)args[0];
   DccChatPass * out = idx;
   int langv = 0;
   UserRecord * u = G->userlist->Find(par);
   FlagRecord atr (idx->User());
   FlagRecord fr (u, FlagRecord::GLOBAL | FlagRecord::CHANNEL 
		  | FlagRecord::ANYWHERE | FlagRecord::BOT);

   context;
   if (!par[0]) 
     langv = 0x0AE;
   else if (!u) 
     langv = 0x100;
   else if (fr.GlobBot())
     langv = 0x0AF;
   else if (u->PassMatch("-"))
     langv = 0x0B0;
   else if (!fr.GlobParty() && (G->require_p || !fr.AnyOp()) 
	    && !atr.GlobBotmast())
     langv = 0x0FF;
   else  {
      G->LOG_CMDS.Put(G->Lang(0x0B2, "ss", idx->Nick(), par));
      if (!atr.GlobOwner() || fr.GlobOwner()) {
	 idx->CheckChof(G->Lang(0x0B1, ""));
	 context;
	 /* store the old nick in the away section, for weenies who can't get
	  * their password right ;) */
	 idx->SetAwayRaw(idx->Nick());
	 idx->SetUser(u);
	 idx->SetNick(u->Key());
	 idx->Send( G->Lang(0x0B3, "ss", par, 
			 idx->HasStatus(STAT_TELNET) ? "\377\373\001":""));
	 out = NEW ( DccChatPass ( *idx ) );
      } else if (atr.GlobOwner()) {
	 idx->CheckChof(G->Lang(0x0B1, ""));
	 context;
	 idx->Send( G->Lang(0x0B4, "s", par));
	 if (atr.GlobMaster())
	   idx->ConsoleFlags( G->conmask );
	 idx->SetUser(u);
	 idx->SetNick(u->Key());
	 idx->Chatter();
	 context;
      }
   }
   if (langv)
     out->Send( G->Lang(langv, "s", par));
   return EXECUTED;
}

static BindReturn cmd_who (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0], * ptr = G->socklist;
   DccChat * dcc = (DccChat *)idx;
   char * par = (char *)args[1], s[512], *t, *u, C;
   int k = 511, ok = 0;
   FlagRecord atr ( idx->User() );
#ifdef HAVE_BOTNET_H
   time_t tt;
#endif
    
   if (par[0]) 
     idx->Send( G->Lang(0x034, ""));
   else {
      int chan = (dcc->Channel() < 0) ? 0 : dcc->Channel();
      G->LOG_CMDS.Put(G->Lang(0x035, "s", idx->Nick()));
      if (!chan)
	idx->Send( G->Lang(0x010, "cccc", '*', '+', '%', '@'));
      else {
	 void * args[3] = { &chan, s, &k };
	 BindReturn ret = G->bt_name->Check("", 0, args);
	 
	 idx->Send( G->Lang((ret == EXEC_LOG) ? 0x012 : 0x011,
			       "ccccsds",  '*', '+', '%', '@', 
			       (chan < GLOBAL_CHANS) ? "" : "*", 
			       chan % GLOBAL_CHANS, s));
      }
      context;
      while ((ptr = ptr->Next()))
	if ((ptr->TypeFlags() & DCT_CHAT) && 
	    ((dcc = (DccChat *)ptr)->Channel() == chan)) {
	   if (G->now- dcc->Timeval() > 300) {
	      int days, hrs, mins;
	      days = (G->now- dcc->Timeval()) / 86400;
	      hrs = ((G->now- dcc->Timeval()) - (days * 86400)) / 3600;
	      mins = ((G->now- dcc->Timeval()) - (hrs * 3600)) / 60;
	      t = G->Lang((days > 0) ? 0x013 : ((hrs > 0) ? 0x014 : 0x015),
		       "ddd", days, hrs, mins);
	   } else 
	     t = 0;
	   C = dcc->GetIcon();
	   idx->Send( G->Lang(atr.GlobOwner() ? 0x016 : 0x017,
				 "cs sss2", C == '-' ? ' ' : C, dcc->Nick(),
				 HANDLEN - strlen(dcc->Nick()), dcc->Host(), t,
				 (atr.GlobMaster() && 
				  dcc->ConsoleFlags().Value())
				 ? G->Lang(0x018, "s",
					dcc->ConsoleFlags().Mask()) : "",
				 dcc->FD()));
	   if (dcc->Away())
	     idx->Send( G->Lang(0x019, "s", dcc->Away()));
	}
      context;
#ifdef HAVE_BOTNET_H
      ptr = G->socklist;
      while ((ptr = ptr->Next())) 
	if (ptr->TypeFlags() & DCT_REALBOT) {
	   if (!ok) {
	      ok = 1;
	      idx->Send( G->Lang(0x01A, ""));
	   }
	   tt = ptr->Timeval();
	   strcpy(s, ctime(&tt));
	   strcpy(s, &s[1]);
	   s[9] = 0;
	   strcpy(s, &s[7]);
	   s[2] = ' ';
	   strcpy(&s[7], &s[10]);
	   s[12] = 0;
	   idx->Send( G->Lang(atr.GlobOwner() ? 0x01B : 0x01C,
			   "scs s2", 
			   ptr->HasStatus(Bot::CALLED) ?  "<-" : "->", 
			   strcmp(ptr->Type(), "SHAREBOT") ? ' ' : '+', 
			   ptr->Nick(), HANDLEN - strlen(ptr->Nick()), s,
			   ptr->FD()));
	}
#endif
      ok = 0;
      context;
      ptr = G->socklist;
      while ((ptr = ptr->Next())) {
	 if ((ptr->TypeFlags() & DCT_CHAT) && 
	     (((dcc = (DccChat *)ptr)->Channel() != chan))) {
	    if (!ok) {
	       ok = 1;
	       idx->Send( G->Lang(0x01D, ""));
	    }
	    if (G->now- dcc->Timeval() > 300) {
	       int days, hrs, mins;
	       days = (G->now- dcc->Timeval()) / 86400;
	       hrs = ((G->now- dcc->Timeval()) - (days * 86400)) / 3600;
	       mins = ((G->now- dcc->Timeval()) - (hrs * 3600)) / 60;
	       t = G->Lang((days > 0) ? 0x013 : ((hrs > 0) ? 0x014 : 0x015),
			"ddd", days, hrs, mins);
	    } else 
	      t = 0;
	    C = dcc->GetIcon();
	    u = atr.GlobMaster() ? G->Lang(
					(dcc->Channel() < 0) ? 0x01E : 
					(!dcc->Channel() ? 0x01F : 0x020), "5",
					dcc->Channel()) : 0;
	    idx->Send( G->Lang(atr.GlobOwner() ? 0x021 : 0x022,
				  "cs ssss2", C == '-' ? ' ' : C, dcc->Nick(),
				  HANDLEN - strlen(dcc->Nick()), u, dcc->Host(), t,
				  (atr.GlobMaster() &&
				   dcc->ConsoleFlags().Value())
				  ? G->Lang(0x018, "s",
					 dcc->ConsoleFlags().Mask()) : "",
				  dcc->FD()));
	    if (dcc->Away())
	      idx->Send( G->Lang(0x019, "s", dcc->Away()));
	 } else if (atr.GlobMaster() && (ptr->TypeFlags() & DCT_SHOWWHO)
		    && (ptr->TypeFlags() & DCT_CHAT)) {
	    if (!ok) {
	       ok = 1;
	       idx->Send( G->Lang(0x01D, ""));
	    }
	    idx->Send( G->Lang(atr.GlobOwner() ? 0x023 : 0x024,
				  "cs s2", ptr->HasStatus(STAT_CHAT) ?  '+' : ' ',
				  ptr->Nick(), HANDLEN-strlen(ptr->Nick()),
				  ptr->Host(), ptr->FD()));
	 }
      }
   }
   context;
   return EXECUTED;
}

static BindReturn cmd_whois (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   UserRecord * u;
   
   if (!par[0]) 
     idx->Send( G->Lang(0x036, ""));
   else {
      G->LOG_CMDS.Put(G->Lang(0x037, "ss", idx->Nick(), par)); 
      if (!(u = G->userlist->Find(par))
	  && (!(u = G->userlist->FindHost(par)) || (u == G->userlist)))
	idx->Send(G->Lang(0x282, ""));
      else {
	 idx->Send(G->Lang(0x280, " ", HANDLEN - 6));
	 u->Display(idx);
      }     
   }
   return EXECUTED;
}
      
/* DCC CHAT COMMANDS */
#define DCC_COMMANDS 55
static command_binds C_dcc[DCC_COMMANDS]={
   { "+host",     "t|m",  cmd_pls_host,       0 },
   { "+ignore",   "m",    cmd_pls_ignore,     0 },
   { "+user",     "m",    cmd_pls_user,       0 },
   { "-host",     "",     cmd_mns_host,       0 },
   { "-ignore",   "m",    cmd_mns_ignore,     0 },
   { "-user",     "t",    cmd_mns_user,       0 },
   { "addlog",    "to|o", cmd_addlog,         0 },
   { "away",      "",     cmd_away,           0 },
   { "back",      "",     cmd_away,           0 },
   { "banner",    "t",    cmd_banner,         0 },
   { "binds",     "m",    cmd_binds,          0 },
   { "boot",      "t",    cmd_boot,           0 },
   { "chat",      "",     cmd_chat,           0 },
   { "chattr",    "m|m",  cmd_chattr,         0 },
   { "chnick",    "t",    cmd_chnick,         0 },
   { "chpass",    "t",    cmd_chpass,         0 },
   { "comment",   "m",    cmd_comment,        0 },
   { "console",   "to|o", cmd_console,        0 },
   { "dccstat",   "t",    cmd_dccstat,        0 },
   { "debug",     "m",    cmd_debug,          0 },
   { "die",       "n",    cmd_die,            0 },
   { "echo",      "",     cmd_echo,           0 },
   { "fixcodes",  "",     cmd_fixcodes,       0 },
   { "help",      "",     cmd_help,           0 },
   { "helpsearch","",     cmd_help,           0 },
   { "hooks",     "n",    cmd_hooks,          0 },
   { "ignores",   "m",    cmd_ignores,        0 },
   { "language",  "n",    cmd_loadlanguage,   0 },
   { "ldump",     "n",    cmd_languagedump,   0 },
   { "lstat",     "n",    cmd_languagestatus, 0 },
   { "loadconfig","n",    cmd_loadconfig,     0 },
   { "loadmod",   "n",    cmd_loadmod,        0 },
   { "match",     "to|o", cmd_match,          0 },
   { "me",        "",     cmd_me,             0 },
   { "module",    "m",    cmd_module,         0 },
   { "motd",      "",     cmd_motd,           0 },
   { "newpass",   "",     cmd_newpass,        0 },
   { "nick",      "",     cmd_chnick,         0 },
   { "note",      "",     cmd_note,           0 },
   { "page",      "",     cmd_page,           0 },
   { "quit",      "",     cmd_quit,           0 },
   { "rehash",    "m",    cmd_restart,        0 },
   { "rehelp",    "n",    cmd_rehelp,         0 },
   { "reload",    "m|m",  cmd_reload,         0 },
   { "restart",   "n",    cmd_restart,        0 },
   { "save",      "m|m",  cmd_save,           0 },
   { "saveconfig","n",    cmd_loadconfig,     0 },
   { "set",       "m|n",  cmd_set,            0 },
   { "simul",     "n",    cmd_simul,          0 },
   { "status",    "m|m",  cmd_status,         0 },
   { "strip",     "",     cmd_strip,          0 },
   { "su",        "",     cmd_su,             0 },
   { "unloadmod", "n",    cmd_loadmod,        0 },
   { "who",       "",     cmd_who,            0 },
   { "whois",     "to|o", cmd_whois,          0 },
};

static int script_wildmatch (ScriptContext * sc, int argc, 
			     ScriptArgument ** argv) {
   FUNCTION;
   BADARGS (3, 3, G->Lang(0x29C, ""));
   sc->SetInteger ( G->WildMatch(argv[1]->AsString(), argv[2]->AsString()));
   return SCRIPT_OK;
}

static int script_maskhost (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char newp[121];
   
   BADARGS(2, 2, G->Lang( 0x22E, ""));
   G->MaskHost(argv[1]->AsString(), newp);
   sc->SetString(newp);
   return SCRIPT_OK;
} 

static int script_putdcc (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * sock;
   char buf[8192];
   
   context;
   BADARGS(3, 3, G->Lang( 0x22F, ""));
   strncpyz(buf, argv[2]->AsString(), 8191);
   if ((sock = G->FindValidIdx(argv[1]->AsInteger()))) {
      sock->DumpLots("", buf);
      return SCRIPT_OK;
   }
   sc->SetString(G->Lang( 0x105, ""));
   return SCRIPT_ERROR;
}

static int script_dccsimul (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * idx;
   char buf[513];
   
   context;
   BADARGS(3, 3, G->Lang( 0x230, ""));
   if (enable_simul) {
      idx = G->FindValidIdx(argv[1]->AsInteger());
      if (idx && (idx->TypeFlags() & DCT_SIMUL)) {
	 strncpyz(buf, argv[2]->AsString(), 510);
	 idx->Activity(buf,strlen(buf));
	 return SCRIPT_OK;
      }
   }
   sc->SetString(G->Lang( 0x105, ""));
   return SCRIPT_ERROR;
}

static int script_dccbroadcast (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   void * args[3];
   int chan = -1, l;
   char * o, msg[401];
   DccChat * ptr = (DccChat *)G->socklist;

   context;
   BADARGS(2, 2, G->Lang( 0x231, ""));
   strncpy(msg, argv[1]->AsString(), 400);
   msg[400] = 0;
   l = strlen(o = G->Lang( 0x10C, "s", msg));
   o[l++] = '\n';
   o[l] = 0;
   while ((ptr = ptr->Next()))
     if (ptr->Channel() >= 0)
       ptr->Dump(o, l);
   args[0] = G->botnetnick;
   args[1] = &chan;
   args[2] = msg;
   G->bt_bcst->Check(msg, 0, args);   
   return SCRIPT_OK;
} 

static int script_hand2idx (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * ptr = G->socklist;
   
   context;
   BADARGS(2, 2, G->Lang( 0x232, ""));
   while ((ptr = ptr->Next()))
     if (!strcasecmp(argv[1]->AsString(), ptr->Nick()) &&
	 (ptr->TypeFlags() & DCT_VALIDIDX)) {
	sc->SetInteger(ptr->FD());
	return SCRIPT_OK;
     }
   sc->SetString("-1");
   return SCRIPT_OK;
}

static int script_getchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccChat * idx;

   context;
   BADARGS(2, 2, " idx");
   idx = (DccChat *)G->FindValidIdx(argv[1]->AsInteger());
   if (!idx || (idx->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   sc->SetInteger(idx->Channel());
   return SCRIPT_OK;
}

static int script_setchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   int chan = -1;
   DccChat * idx;
   void * args[3] = { &chan, 0, 0 };
   const char * set = argv[2]->AsString();
   
   context;
   BADARGS(3, 4, G->Lang( 0x0233, ""));
   if (!(idx = (DccChat *)G->FindValidIdx(argv[1]->AsInteger())) 
       || (idx->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   if ((set[0] < '0') || (set[0] > '9')) {
      if (strcmp(set, "-1") && strcasecmp(set, "off")
	  && (G->bt_name->Check("", 0, args) != EXEC_LOG)) {
	 sc->SetString(G->Lang( 0x234, ""));
	 return SCRIPT_ERROR;
      }
   } else
      chan = argv[2]->AsInteger();
   if ((chan < -1) || (chan > 199999)) {
      sc->SetString(G->Lang( 0x234, ""));
      return SCRIPT_ERROR;
   }
   args[1] = (void *)set;
   if ((idx->Channel() >= 0) && (chan < 0))
     idx->CheckLChpt(argc == 4 ? argv[3]->AsString() : G->Lang( 0x236, ""));
   idx->SetChannel(chan);
   idx->CheckLChjn();
   return SCRIPT_OK;
}

static int script_dccputchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   int chan;
   char msg[401];
   void * args[3];
   
   context;
   BADARGS(3, 3, G->Lang( 0x237, ""));
   chan = argv[1]->AsInteger();
   if ((chan < 0) || (chan > 199999)) {
      sc->SetString(G->Lang( 0x238, ""));
      return SCRIPT_ERROR;
   }
   strncpy(msg, argv[2]->AsString(), 400);
   msg[400] = 0;
   DccChat::Chanout(chan, G->Lang( 0x10C, "s", msg));
   args[0] = G->botnetnick;
   args[1] = &chan;
   args[2] = msg;
   G->bt_bcst->Check(argv[2]->AsString(), 0, args);
   return SCRIPT_OK;
}

static int script_console (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   int pls, arg,j;
   DccChat * ptr;
   const char * list;
   
   context;
   BADARGS(2, 4, G->Lang( 0x239, ""));
   if (!(ptr = (DccChat *)G->FindValidIdx(argv[1]->AsInteger())) 
       || (ptr->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   pls = 1;
   for (arg = 2; arg < argc; arg++) {
      const char * which = argv[arg]->AsString();
      if ((which[0] == '#') || (which[0] == '&') || (which[0] == '*')) {
	 if ((which[0] != '*') && !G->chanset->Find(which)) {
	    sc->SetString(G->Lang( 0x23A, ""));
	    return SCRIPT_ERROR;
	 }
	 ptr->SetConsoleChan(which);
      } else {
	 if ((which[0] != '+') && (which[0] != '-'))
	   ptr->ConsoleFlags().Clear();
	 for (j = 0; j < (int)strlen(which); j++) {
	    if (which[j] == '+')
	       pls = 1;
	    else if (which[j] == '-')
	       pls = -1;
	    else if (pls == 1)
	      ptr->ConsoleFlags().Set(which[j]);
	    else
	      ptr->ConsoleFlags().UnSet(which[j]);
	 }
      }
   }
   sc->SetArray(1, &(list = ptr->ConsoleChan()));
   sc->AppendArray(1, &(list = ptr->ConsoleFlags().Mask()));
   return SCRIPT_OK;
}

static int script_strip (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   int pls, arg, j;
   DccChat * ptr;
   
   context;
   BADARGS(2, 4, G->Lang( 0x23B, ""));
   if (!(ptr = (DccChat *)G->FindValidIdx(argv[1]->AsInteger())) 
       || (ptr->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   pls = 1;
   for (arg = 2; arg < argc; arg++) {
      const char * which = argv[arg]->AsString();
      
      if ((which[0] != '+') && (which[0] != '-'))
	 ptr->SetStripFlags( 0 );
      for (j = 0; j < (int)strlen(which); j++) {
	 if (which[j] == '+')
	    pls = 1;
	 else if (which[j] == '-')
	    pls = (-1);
	 else {
	    char s[2];
	    
	    s[0] = which[j];
	    s[1] = 0;
	    if (pls == 1)
	       ptr->AddStripFlags(G->StripModes(s));
	    else
	       ptr->DelStripFlags(G->StripModes(s));
	 }
      }
   }
   sc->SetString(G->StripMaskType(ptr->StripFlags()));
   return SCRIPT_OK;
}

static int script_echo (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket *ptr;
   
   context;
   BADARGS(2, 3, G->Lang( 0x23C, ""));
   if (!(ptr = G->FindValidIdx(argv[1]->AsInteger())) 
       || (ptr->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   if (argc == 3) {
      if (argv[2]->AsInteger())
	ptr->SetStatus(STAT_ECHO);
      else
	ptr->ClearStatus(STAT_ECHO);
   }
   sc->SetString(ptr->HasStatus(STAT_ECHO) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_page (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccChat * ptr;
   
   context;
   BADARGS(2, 3, G->Lang( 0x23C, ""));
   if (!(ptr = (DccChat *)G->FindValidIdx(argv[1]->AsInteger())) 
       || (ptr->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   if (argc == 3) {
      int l = argv[2]->AsInteger();
      if (l == 0)
	ptr->ClearStatus(STAT_PAGE);
      else {
	 ptr->SetStatus(STAT_PAGE);
	 ptr->SetMaxLines(l);
      }
   }
   sc->SetInteger(ptr->HasStatus(STAT_PAGE) ? ptr->MaxLines() : 0);
   return SCRIPT_OK;
}

static int script_valididx (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, G->Lang( 0x23D, ""));
   sc->SetString(G->FindValidIdx(argv[1]->AsInteger())  ? "1" : "0");
   return SCRIPT_OK;
}

static int script_killdcc (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * ptr;

   context;
   BADARGS(2, 3, G->Lang( 0x23E, ""));
   if (!(ptr = G->FindValidIdx(argv[1]->AsInteger()))) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   /* don't kill terminal socket */
   if ((ptr->FD() == STDIN) && !G->backgrd)
     return SCRIPT_OK;
   /* make sure 'whom' info is updated for other bots */
   if (ptr->TypeFlags() & DCT_CHAT) 
     ((DccChat *)ptr)->CheckChof(argc == 3 ? argv[2]->AsString() : "");
   ptr->Close();
   return SCRIPT_OK;
}

static int script_idx2hand (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * idx;

   context;
   BADARGS(2, 2, G->Lang( 0x23D, ""));
   if (!(idx = G->FindValidIdx(argv[1]->AsInteger()))) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   sc->SetString(idx->Nick());
   return SCRIPT_OK;
}

/* list of { idx nick host type } */
/* type: "chat", "files", "script" */
static int script_dcclist (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * ptr = G->socklist;
   const char *list[4];

   context;
   BADARGS(1, 2, G->Lang( 0x23F, ""));
   while ((ptr = ptr->Next())) {
      if (((argc == 1) || !strcasecmp(ptr->Type(),argv[1]->AsString()))
	  && !(ptr->Flags() & (Socket::NONSOCK|Socket::LOST))) {
	 list[0] = G->ToBase10(ptr->FD());
	 list[1] = ptr->Nick();
	 list[2] = ptr->Host();
	 list[3] = ptr->Type();
	 sc->AppendArray(4, list);
      }
   }
   return SCRIPT_OK;
}

static int script_dccused (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   sc->SetInteger(G->dcc_total);
   return SCRIPT_OK;
}

static int script_getdccidle (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * idx;

   context;
   BADARGS(2, 2, G->Lang( 0x23D, ""));
   if (!(idx = G->FindValidIdx(argv[1]->AsInteger()))) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   sc->SetInteger(G->now - idx->Timeval());
   return SCRIPT_OK;
}

static int script_getdccaway (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccChat * idx;

   context;
   BADARGS(2, 2, G->Lang( 0x23D, ""));
   if (!(idx = (DccChat *)G->FindValidIdx(argv[1]->AsInteger())) 
       || (idx->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   sc->SetString(idx->Away());
   return SCRIPT_OK;
}

static int script_setdccaway (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccChat * idx;

   context;
   BADARGS(3, 3, G->Lang( 0x240, ""));
   if (!(idx = (DccChat *)G->FindValidIdx(argv[1]->AsInteger())) 
       || (idx->TypeFlags() & DCT_CHAT)) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   if (!argv[2]->AsString()[0]) {
      /* un-away */
      if (idx->Away())
	idx->SetAway(0);
   } else
     /* away */
     idx->SetAway(argv[2]->AsString());
   return SCRIPT_OK;
}

static int script_listen (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   TelnetPort * tp = (TelnetPort *)G->socklist;
   context;
   BADARGS(3, 4, G->Lang( 0x241, ""));
   if (!G->do_restart) {
      if ((tp = tp->Find(argv[2]->AsInteger()))) {
	 sc->SetString(G->Lang( 0x1A9, "ss", argv[1]->AsString(), 
			       argv[2]->AsString()));
	 return SCRIPT_ERROR;
      }
      NEW ( TelnetPort ( sc, argv[1]->AsString(), argv[2]->AsInteger(),
			(argc == 3) ? "" : argv[3]->AsString(), "(users)"));
   }
   return SCRIPT_OK;
}

static int script_killlisten (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   TelnetPort * tp = (TelnetPort *)G->socklist;
   context;
   BADARGS(2, 2, G->Lang( 0x242, ""));
   if (!(tp = tp->Find(argv[2]->AsInteger()))) {
      sc->SetString(G->Lang( 0x1A8, "ss", argv[1]->AsString(),
			    argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   DELETE ( tp );
   return SCRIPT_OK;
}

static int script_rehash (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, " ");
   if (G->make_userfile) {
      G->LOG_MISC.Put(G->Lang( 0x06A, ""));
      G->make_userfile = 0;
   }
   G->userlist->WriteUserfile(NULL);
   G->LOG_MISC.Put(G->Lang( 0x068, ""));
   G->do_restart = -2;
   return SCRIPT_OK;
}

static int script_restart (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, " ");
   if (G->make_userfile) {
      G->LOG_MISC.Put(G->Lang( 0x06A, ""));
      G->make_userfile = 0;
   }
   G->userlist->WriteUserfile(NULL);
   G->LOG_MISC.Put(G->Lang( 0x2A3, ""));
   G->do_restart = -1;
   return SCRIPT_OK;
}

static int script_putloglev (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   LogModes lev;
   char logtext[501];
   Channel * chan = 0;
   
   context;
   BADARGS(4, 4, G->Lang( 0x243, ""));
   lev.Set(argv[1]->AsString());
   if (lev.Value() == 0) 
     sc->SetString(G->Lang( 0x244, ""));
   else if ((argv[2]->AsString()[0] != '*') && 
	    !(chan = G->chanset->Find(argv[2]->AsString())))
     sc->SetString(G->Lang( 0x234, ""));
   else {
      strncpy(logtext, argv[3]->AsString(), 500);
      logtext[500] = 0;
      lev.Put(logtext, chan);
      return SCRIPT_OK;
   } 
   return SCRIPT_ERROR;
}

static int script_unixtime (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   sc->SetInteger(G->now);
   return SCRIPT_OK;
}

static int script_time (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   char s[81];

   context;
   BADARGS(1, 1, "");
   strcpy(s, ctime(&G->now));
   s[16] = 0;
   sc->SetString(s + 11);
   return SCRIPT_OK;
}

static int script_date (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   char s[81], * t;
   
   context;
   BADARGS(1, 1, "");
   t = ctime(&G->now);
   s[0] = t[8];
   s[1] = t[9];
   s[2] = t[3];
   s[3] = t[4];
   s[4] = t[5];
   s[5] = t[6];
   s[6] = t[7];
   s[7] = t[20];
   s[8] = t[21];
   s[9] = t[22];
   s[10] = t[23];
   s[11] = 0;
   sc->SetString(s);
   return SCRIPT_OK;
}

static int script_ctime (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   time_t tt;
   char s[81];

   context;
   BADARGS(2, 2, G->Lang( 0x245, ""));
   tt = (time_t) argv[1]->AsInteger();
   strcpy(s, ctime(&tt));
   s[strlen(s) - 1] = 0;
   sc->SetString(s);
   return SCRIPT_OK;
}

static int script_myip (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   sc->SetInteger(iptolong(G->GetMyIp()));
   return SCRIPT_OK;
}

static int script_rand (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   unsigned long x;

   context;
   BADARGS(2, 2, G->Lang( 0x246, ""));
   if ((x = argv[1]->AsInteger()) <= 0) {
      sc->SetString(G->Lang( 0x247, ""));
      return SCRIPT_ERROR;
   }
   sc->SetInteger(random() % x);
   return SCRIPT_OK;
}

static int script_sendnote (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char from[NOTENAMELEN+1], to[NOTENAMELEN+1], msg[451];
   int v;
     
   context;
   BADARGS(4, 4, G->Lang( 0x248, ""));
   strncpy(from, argv[1]->AsString(), NOTENAMELEN);
   from[NOTENAMELEN] = 0;
   strncpy(to, argv[2]->AsString(), NOTENAMELEN);
   to[NOTENAMELEN] = 0;
   strncpy(msg, argv[3]->AsString(), 450);
   msg[450] = 0;
   v = G->AddNote(to, from, msg, NULL, 0);
   sc->SetString(v < 0 ? "-" : "");
   sc->AppendString(G->ToBase10(v < 0 ? -v : v));
   return SCRIPT_OK;
}

static int script_dccdumpfile (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char fn[DIRLEN+1];
   DccChatPass * idx;
   FlagRecord fr ( FlagRecord::GLOBAL | FlagRecord::CHANNEL
                  | FlagRecord::ANYWHERE);
   
   context;
   BADARGS(3, 3, G->Lang( 0x249, ""));
   G->SPrintF(fn, DIRLEN, "text/%s", argv[2]->AsString());
   if (!(idx = (DccChatPass *)G->FindValidIdx(argv[1]->AsInteger()))) {
      sc->SetString(G->Lang( 0x105, ""));
      return SCRIPT_ERROR;
   }
   fr.Set(idx->User());
   G->help_list->Tell(idx, fn, fr, HelpReference::TEXT);
   return SCRIPT_OK;
}

static int script_suidx ( ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccChatPass * idx;
   UserRecord * u;
   
   BADARGS(3, 3, G->Lang( 0x2A0, ""));
   if (!(idx = (DccChatPass *)G->FindValidIdx(argv[1]->AsInteger()))) 
     sc->SetString(G->Lang(0x105, ""));
   else if (!(u = G->userlist->Find(argv[2]->AsString()))) 
     sc->SetString(G->Lang(0x100, ""));
   else {
      if (idx->TypeFlags() & DCT_CHAT) 
	((DccChat *)idx)->CheckLChpt ( G->Lang(0x0B1, ""));
      idx->SetNick(argv[2]->AsString());
      idx->SetUser(u);
      if (idx->TypeFlags() & DCT_CHAT) 
	((DccChat *)idx)->CheckLChjn ();
      return SCRIPT_OK;
   }
   return SCRIPT_ERROR;
}

static int script_backup (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   G->userlist->Backup();
   return SCRIPT_OK;
}

static int script_die (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(1, 2, G->Lang( 0x24A, ""));
   G->Fatal((argc == 2) ? argv[1]->AsString() : G->Lang( 0x24B, ""));
   /* should never return, but, to keep gcc happy: */
   return SCRIPT_OK;
}

static int script_strftime (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char buf[512];
   struct tm *tm1;
   time_t t;

   context;
   BADARGS(2, 3, G->Lang( 0x24C, ""));
   t = (argc == 3) ? argv[2]->AsInteger() : G->now;
   tm1 = localtime(&t);
   if (strftime(buf, sizeof(buf) - 1, argv[1]->AsString(), tm1)) {
      sc->SetString(buf);
      return SCRIPT_OK;
   }
   sc->SetString(G->Lang( 0x24D, ""));
   return SCRIPT_ERROR;
}

static int script_loadmodule (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   const char *p;
   
   context;
   BADARGS(2, 2, G->Lang( 0x24E, ""));
   p = G->module_list->Load(argv[1]->AsString());
   if (p && (strcmp(p, G->Lang( 0x0F5, "" ))) && (*argv[0]->AsString() == 'l') && !G->do_restart) {
      G->LOG_MISC.Put(G->Lang( 0x0FC, "ss", argv[1]->AsString(), p));
      return SCRIPT_ERROR;
   }
   sc->SetString(p);
   return SCRIPT_OK;
}

static int script_unloadmodule (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, G->Lang( 0x24E, ""));
   sc->SetString(G->module_list->Unload(argv[1]->AsString()));
   return SCRIPT_OK;
}

static int script_unames (ScriptContext * sc, int, ScriptArgument **) {
   FUNCTION;
   const char * unix_n, * vers_n;

#ifdef HAVE_UNAME
   struct utsname un;
   if (uname(&un) < 0) {
#endif
         unix_n = G->Lang( 0x24F, "");
         vers_n = "";   
#ifdef HAVE_UNAME
   } else {
         unix_n = un.sysname;
         vers_n = un.release;
   }
#endif
   sc->SetArray(1, &unix_n);
   sc->SetArray(1, &vers_n);
   return SCRIPT_OK;
} 

static int script_modules (ScriptContext * sc, int argc, ScriptArgument **)  {
   FUNCTION;
   ModuleEntry * current = G->module_list;
   Dependancy * dep;
   const char * list[100];
   char s[100], s2[100][100];
   int i;
   
   context;
   BADARGS(1, 1, "");

   while ((current = current->Next())) {
      list[0] = current->Key();
      G->SPrintF(s, 99, "%d.%d.%d.%d",current->Vers()->Major(),
		 current->Vers()->Minor(),current->Vers()->Fix(),
		 current->Vers()->Patch());
      list[1] = s;
      list[2] = safe_str(current->VersionStr());
      list[3] = safe_str(current->FullVersion());
      i = 4;
      dep = G->dep_list;
      while ((dep = dep->Next()) && (i < 100)) 
	if (dep->Needing() == current) {
	   G->SPrintF(s2[i],99,"%s %d.%d.%d",dep->Needed()->Key(),
		      dep->Major(),dep->Minor(),dep->Fix()); 
	   i++;
	}
      sc->AppendArray(i, list);
   }
   return SCRIPT_OK;
}

static int script_loadhelp (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, G->Lang( 0x250, ""));
   NEW ( HelpReference ( argv[1]->AsString() ) );
   return SCRIPT_OK;
}

static int script_unloadhelp (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   HelpReference * ptr;
   
   context;
   BADARGS(2, 2, G->Lang( 0x250, ""));
   if ((ptr = G->help_list->Find(argv[1]->AsString())))
     DELETE ( ptr );
   return SCRIPT_OK;
}

static int script_reloadhelp (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   G->help_list->Reload();
   return SCRIPT_OK;
}

static int script_countusers (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   sc->SetInteger(G->userlist->Count());
   return SCRIPT_OK;
}

static int script_validuser (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, G->Lang( 0x251, ""));
   sc->SetString(G->userlist->Find(argv[1]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_finduser (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord * u;
   
   context;
   BADARGS(2, 2, G->Lang( 0x252, ""));
   u = G->userlist->FindHost(argv[1]->AsString());
   sc->SetString(u->Handle());
   return SCRIPT_OK;
}

static int script_passwdOk (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord * u;
   
   context;
   BADARGS(3, 3, G->Lang( 0x253, ""));
   sc->SetString(((u = G->userlist->Find(argv[1]->AsString()))
			  && u->PassMatch(argv[2]->AsString())) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_chattr (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   const char * chan = 0, *chg;
   char work[100];
   FlagRecord pls,mns,user;
   UserRecord * u = 0;
   
   context;
   BADARGS(2, 4, G->Lang( 0x254, ""));
   if ((argv[1]->AsString()[0] == '*') 
       || !(u = G->userlist->Find(argv[1]->AsString()))) {
      sc->SetString("*");
      return SCRIPT_OK;
   }
   chg = argv[2]->AsString();
   if (argc == 4) {
      user.SetMatch ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
      chan = argv[3]->AsString();
   } else if ((argc == 3) && ((chg[0] == '&') || (chg[0] == '#'))) {
      user.SetMatch ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
      chan = chg;
      chg = 0;
   }
   if (chan && !G->chanset->Find(chan)) {
      sc->SetString(G->Lang( 0x255, ""));
      return SCRIPT_ERROR;
   }
   user.Set(u,chan);
   /* make changes */
   if (chg) {
      pls.SetMatch (user.Match());
      pls.BreakDown(chg,&mns);
      /* no-one can change these flags on-the-fly */
      pls.ClearGlobBot();
      mns.ClearGlobBot();
      if (chan) {
	 pls.ClearChanShare();
	 mns.ClearChanShare();
      }
      user += pls;
      user -= mns;
      u->Set(user,chan);
    }
   /* retrieve current flags and return them */
   user.Build(work);
   sc->SetString(work);
   return SCRIPT_OK;
}

static int script_matchattr (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord * u;
   FlagRecord plus,minus,user;
   int f;
   
   context;
   BADARGS(3, 4, G->Lang( 0x256, ""));
   if ((u = G->userlist->Find(argv[1]->AsString()))
       && ((argc == 3) || G->chanset->Find(argv[3]->AsString()))) {
      user.SetMatch ( FlagRecord::GLOBAL | FlagRecord::BOT
		     | (argc == 4 ? FlagRecord::CHANNEL : 0));
      user.Set(u,(argc == 4) ? argv[3]->AsString() : 0);
      plus.SetMatch ( user.Match());
      plus.BreakDown(argv[2]->AsString(),&minus);;
      f = minus.AnyFlags();
      minus ^= plus;
      if ((plus == user) && (!f || !(minus == user))) {
	 sc->SetString("1");
	 return SCRIPT_OK;
      }
   }
   sc->SetString("0");
   return SCRIPT_OK;
}

static int script_adduser (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char tbuf[HANDLEN + 1], ubuf[UHOSTLEN + 1];
   
   context;
   BADARGS(3, 3, G->Lang( 0x257, ""));
   strncpyz(tbuf, argv[1]->AsString(), HANDLEN);
   if ((tbuf[0] == '*') || G->userlist->Find(tbuf))
     sc->SetString("0");
   else {
      FlagRecord fr;
      
      fr.SetGlobal(G->default_flags);
      strncpyz(ubuf, argv[2]->AsString(), UHOSTLEN);
      NEW ( UserRecord (G->userlist, tbuf, ubuf, fr) );
      sc->SetString("1");
   }
   return SCRIPT_OK;
}

static int script_deluser (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord * u = 0;
   context;
   BADARGS(2, 2, G->Lang( 0x251, ""));
   if ((argv[1]->AsString()[0] != '*') && (u = G->userlist->Find(argv[1]->AsString())))
     DELETE (u);
   sc->SetString(u ? "1" : "0");
   return SCRIPT_OK;
}

static int script_delhost (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord * u;
   context;
   BADARGS(3, 3, G->Lang( 0x257, ""));
   if (!(u = G->userlist->Find(argv[1]->AsString())) 
       || (argv[1]->AsString()[0] == '*')) {
      sc->SetString(G->Lang( 0x258, ""));
      return SCRIPT_ERROR;
   } 
   sc->SetString(u->DelHost(argv[2]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_userlist (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord *u = G->userlist;
   FlagRecord user,plus,minus;
   const char * x;
   int f = 0;
   
   context;
   BADARGS(1, 3, G->Lang( 0x259, ""));
   if ((argc == 3) && !G->chanset->Find(argv[2]->AsString())) {
      sc->SetString(G->Lang( 0x25A, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   if (argc >= 2) {
      plus.SetMatch(FlagRecord::GLOBAL | FlagRecord::CHANNEL 
		    | FlagRecord::BOT);
      plus.BreakDown(argv[1]->AsString(),&minus);
      f = minus.AnyFlags();
      minus ^= plus;
      user.SetMatch (FlagRecord::GLOBAL | FlagRecord::CHANNEL |
		     FlagRecord::BOT | (argc == 3 ? 0 : FlagRecord::ANYWHERE));
   }
   while ((u = u->Next())) 
     if ((argc < 2) || ((plus == *user.Set(u,argv[2]->AsString())) &&
			!(f && (minus == user))))
       sc->AppendArray(1, &(x = u->Handle()));
   return SCRIPT_OK;
}

static int script_save (ScriptContext *, int, ScriptArgument **) {
   FUNCTION;
   context;
   G->userlist->WriteUserfile(NULL);
   return SCRIPT_OK;
}

static int script_reload (ScriptContext *, int, ScriptArgument **) {
   FUNCTION;
   context;
   G->Reload();
   return SCRIPT_OK;
}

static int script_chnick (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord * u;
   char hand[HANDLEN+1];
   int x = 1, i;
   
   context;
   BADARGS(3, 3, G->Lang( 0x25B, ""));
   if (!(u = G->userlist->Find(argv[1]->AsString())))
     x = 0;
   else {
      strncpy(hand, argv[2]->AsString(), HANDLEN);
      hand[HANDLEN] = 0;
      for (i = 0; i < (int)strlen(hand); i++)
	if ((hand[i] <= 32) || (hand[i] >= 127) || (hand[i] == '@'))
	  hand[i] = '?';
      if (strchr("-,+*=:!.@#;$", hand[0]) != NULL)
	x = 0;
      else if (strlen(hand) < 1)
	x = 0;
      else if (G->userlist->Find(hand))
	x = 0;
      else if (!strcasecmp(G->origbotname, hand) || !strcasecmp(G->botnetnick, hand))
	x = 0;
      else if (hand[0] == '*')
	x = 0;
   }
   if (x) 
     x = u->ChangeHandle(hand);
   sc->SetString(x ? "1" : "0");
   return SCRIPT_OK;
}

static int script_isignore (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, G->Lang( 0x252, ""));
   sc->SetString(G->ignores->Match(argv[1]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_newignore (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   time_t expire_time;
   char ign[UHOSTLEN], cmt[66], from[HANDLEN+1];
   
   context;
   BADARGS(4, 5, G->Lang( 0x25C, ""));
   strncpy(ign, argv[1]->AsString(), UHOSTLEN - 1);
   ign[UHOSTLEN - 1] = 0;
   strncpy(from, argv[2]->AsString(), HANDLEN);
   from[HANDLEN] = 0;
   strncpy(cmt, argv[3]->AsString(), 65);
   cmt[65] = 0;
   if (argc == 4)
       expire_time = G->now + (60 * G->ignore_time);
   else {
      if (argv[4]->AsInteger() == 0)
	expire_time = 0L;
      else
	 expire_time = G->now + (60 * argv[4]->AsInteger());
   } 
   NEW ( IgnoreRecord ( ign, from, cmt, expire_time ) );
   return SCRIPT_OK;
}

static int script_killignore (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   IgnoreRecord * rec;
   char buf[UHOSTLEN + 1];
   
   context;
   BADARGS(2, 2, G->Lang( 0x25D, ""));
   strncpyz(buf, argv[1]->AsString(), UHOSTLEN);
   if ((rec = G->ignores->Find(buf))) 
     DELETE ( rec );
   sc->SetString(rec ? "1" : "0");
   return SCRIPT_OK;
}

/* { hostmask note expire-time create-time creator } */
static int script_ignorelist (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   IgnoreRecord * ptr = G->ignores;
   char ts[21];
   const char * list[5];
   
   context;
   BADARGS(1, 1, "");
   while ((ptr = ptr->Next())) {
      list[0] = ptr->Mask();
      list[1] = ptr->Reason();
      strcpy(ts, G->ToBase10(ptr->Expires()));
      list[2] = ts;
      list[3] = G->ToBase10(ptr->Added());
      list[4] = ptr->Setter();
      sc->AppendArray(5, list);
   }
   return SCRIPT_OK;
}

static int script_getuser (ScriptContext * sc, int argc, ScriptArgument ** argv)  {
   FUNCTION;
   UserEntry * et, * e;
   UserRecord * u;
   
   context;
   BADARGS(3,999, G->Lang( 0x25E, ""));

   if (!(et = G->entrytype_list->Find(argv[2]->AsString()))) {
      sc->SetString(G->Lang( 0x25F, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   if (!(u = G->userlist->Find(argv[1]->AsString()))) {
      if (argv[1]->AsString()[0] != '*') {
	 sc->SetString(G->Lang( 0x102, "s", argv[1]->AsString()));
	 return SCRIPT_ERROR;
      } else
	return SCRIPT_OK; /* silently ignore user * */
   }
   if ((e = u->FindEntry(et)))
     return e->ScriptGet(sc, argc, argv);
   return SCRIPT_OK;
}


static int script_setuser (ScriptContext * sc,  int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserEntry * et, *e;
   UserRecord * u;
   int r = SCRIPT_ERROR;
   
   BADARGS(3,999, G->Lang( 0x25E, ""));

   if (!(et = G->entrytype_list->Find(argv[2]->AsString())))
     sc->SetString(G->Lang( 0x25F, "s", argv[2]->AsString()));
   else if (!(u = G->userlist->Find(argv[1]->AsString()))) {
      if (argv[1]->AsString()[0] != '*') 
	sc->SetString(G->Lang( 0x102, "s", argv[1]->AsString()));
      else
	return SCRIPT_OK; /* silently ignore user * */
   } else if ((e = u->GetEntry(et))) 
     r = e->ScriptSet(sc,argc,argv);
   return r;
}

static int script_language (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BADARGS(2, 2, " language");
   if (!G->LoadLanguage(argv[1]->AsString())) {
      sc->SetString("Load failed.");
      return SCRIPT_ERROR;
   }
   return SCRIPT_OK;
}

/* logfile [<modes> <channel> <filename>] */
static int script_logfile (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   LogFile * lf;
   char s[DIRLEN+1];
   const char * p = s;
   LogModes lmask;
   
   BADARGS(1, 4, G->Lang(0x127,""));
   if (argc == 1) {
      lf = G->log_list;
      while ((lf = lf->Next())) {
	 G->SPrintF(s, DIRLEN, "%s %s %s", lf->Mask().Mask(),
		    lf->Channel(), lf->FileName());
	 sc->AppendArray(1, &p);
      }
      return SCRIPT_OK;
   }
   BADARGS(4, 4, G->Lang(0x127,""));
   lmask.Set(argv[1]->AsString());
   if (lmask.Value())
     NEW ( LogFile ( argv[3]->AsString(), argv[2]->AsString(), lmask) );
   else if ((lf = (LogFile *)G->log_list->Find(argv[3]->AsString())))
     DELETE ( lf );
   sc->SetString(argv[3]->AsString());
   return SCRIPT_OK;
}

static int script_unbind (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BindTable * tp;
  
   BADARGS (5, 5, G->Lang( 0x141, ""));
   if (!(tp = G->bind_list->Find(argv[1]->AsString()))) {
      sc->SetString(G->Lang( 0x143, ""));
      G->bind_list->Dump(sc);
      return SCRIPT_ERROR;
   }
   if (!tp->RemBind(argv[3]->AsString(), argv[4]->AsString())) {
      /* don't error if trying to re-unbind a builtin */
      if (*argv[4]->AsString() != '*') {
	 sc->SetString(G->Lang( 0x144, ""));
	 return SCRIPT_ERROR;
      } else
	return SCRIPT_OK;
   }
   sc->SetString(argv[3]->AsString());
   return SCRIPT_OK;
}

static int script_bind (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BindTable * tp;
  
   BADARGS (4, 5, G->Lang( 0x142, ""));
   if (!(tp = G->bind_list->Find(argv[1]->AsString()))) {
      sc->SetString(G->Lang( 0x143, ""));
      G->bind_list->Dump(sc);
      return SCRIPT_ERROR;
   }
   if (argc == 4) {
      BindMask * bm = tp->First()->Find(argv[3]->AsString());
      if (bm)
	bm->First()->Dump(sc);
      return SCRIPT_OK;
   } else if (tp->Flags() & BindTable::SCRIPT_FUNC)
	tp->AddBind ( argv[2]->AsString(), argv[3]->AsString(),
		     argv[4]->AsString(), sc, 0, 0 );
   else
     G->LOG_MISC.Put(G->Lang(0x145, ""));
   sc->SetString(argv[3]->AsString());
   return SCRIPT_OK;
}


static void got_bus (int) {
   write_debug();
   G->Fatal(G->Lang( 0x1CB, ""));
#ifdef SA_RESETHAND 
   kill(getpid(),SIGBUS);
#else
   exit(1);
#endif
}

static void got_segv (int) {
   write_debug();
   G->Fatal(G->Lang( 0x1CC, ""));
#ifdef SA_RESETHAND
   kill(getpid(),SIGSEGV);
#else
   exit(1);
#endif
}

static void got_fpe (int) {
   write_debug();
   G->Fatal(G->Lang( 0x1CD, ""));
}

static void got_term (int) {
   G->userlist->WriteUserfile(NULL);
   if (die_on_sigterm) 
     G->Fatal(G->Lang( 0x1CE, ""));
   else 
     G->LOG_MISC.Put(G->Lang( 0x1CF, ""));
}

static void got_quit (int) {
   G->LOG_MISC.Put(G->Lang( 0x1D0, ""));
}

static void got_hup (int) {
   G->userlist->WriteUserfile(NULL);
   if (die_on_sighup) 
     G->Fatal(G->Lang( 0x1D1, ""));
   else 
     G->LOG_MISC.Put(G->Lang( 0x1D2, ""));
   G->do_restart = -2;
}

static void got_alarm (int) {
   longjmp(alarmret,1);
}

static void got_usr1 (int) {
   Socket * sock = G->socklist;
   
   G->LOG_MISC.Put(G->Lang( 0x1D3, ""));
   write_debug();
   while ((sock = sock->Next())) {
      if (!(sock->Flags() & (Socket::NONSOCK | Socket::LOST))
	  && (fcntl(sock->FD(), F_GETFD, 0) == -1) && (errno == EBADF)) {
	 G->LOG_MISC.Put(G->Lang( 0x1D4, "cdss", sock->FD() < 0 ? '-' : '+',
				 sock->FD(), sock->Type(), sock->Nick()));
	 sock->Close();
      }
   }
   nested_debug = 0;
   G->LOG_MISC.Put(G->Lang( 0x1D5, ""));
}

/* got USR2 signal -- crash */
static void got_usr2 (int) {
   G->LOG_MISC.Put(G->Lang( 0x1D6, "s", G->context_stack[G->context_ptr]));
   write_debug();
   G->LOG_MISC.Put(G->Lang( 0x1D7, ""));
}

/* got ILL signal -- log context and continue */
static void got_ill (int) {
   G->LOG_MISC.Put(G->Lang( 0x1D6, "s", G->context_stack[G->context_ptr]));
}

static void do_arg (char * s) {
   FUNCTION;
   int i;
   char x[256], *z = x;
   
   if (s[0] == '-')
      for (i = 1; s[i]; i++) {
	 switch (s[i]) {
	   case 'n':
	    G->backgrd = 0;
	    break;
	   case 'c':
	    G->con_chan = 1;
	    G->term_z = 0;
	    break;
	   case 't':
	    G->con_chan = 0;
	    G->term_z = 1;
	    break;
#ifdef EBUG_MEM
	   case 'T':
	    G->con_chan = 0;
	    G->term_z = 2;
	    break;
#endif
	   case 'm':
	    G->make_userfile = 1;
	    break;
	   case 'v':
	    strcpy(x, G->FullVersion());
	    G->Split(&z);
	    G->Split(&z);
	    puts(G->Lang( 0x1E5, ""));
	    if (z[0])
	      puts(G->Lang( 0x1E6, "s", z));
	    exit(0);
	   case 'h':
	    for (i = 0x1D8; i < 0x1E5; i++)
	      puts(G->Lang( i, ""));
	    exit(0);
	 }
      }
   else
      strcpy(G->configfile, s);
}

/* timer info: */
static int lastmin = 99;
static time_t then;
static struct tm *nowtm;

/* rally BB, this is not QUITE as bad as it seems <G> */
/* ONCE A SECOND */
static void core_secondly (void) {
   FUNCTION;
   static int cnt = 0;
   int miltime;
   Socket * ptr = G->socklist;
   
   cnt++;
   if (cnt >= 10) {		/* every 10 seconds */
      cnt = 0;
      if (G->con_chan && !G->backgrd) {
	 G->Stdout->Printf("\033[2J\033[1;1H");
	 G->Stdout->TellVerboseStatus(0);
	 G->module_list->Report(G->Stdout,0,"server");
	 G->module_list->Report(G->Stdout,0,"irc");
	 G->memtype_list->Status(G->Stdout);
      }
      while ((ptr = ptr->Next())) 
	ptr->CheckTimeout();
   }
   context;
   nowtm = localtime(&G->now);
   if (nowtm->tm_min != lastmin) {
      int i = 0;
      
      /* once a minute */
      lastmin = (lastmin + 1) % 60;
      G->hook_minutely->Call();
      G->ignores->Expire();
      /* in case for some reason more than 1 min has passed: */
      while (nowtm->tm_min != lastmin) {
	 /* timer drift, dammit */
	 G->LOG_MISC.Put(G->Lang( 0x28F, "dd",lastmin, nowtm->tm_min));
	 context;
	 i++;
	 lastmin = (lastmin + 1) % 60;
	 G->hook_minutely->Call();
      }
      if (i > 1)
	G->LOG_MISC.Put(G->Lang( 0x1E8, "d", i));
      miltime = (nowtm->tm_hour * 100) + (nowtm->tm_min);
      context;
      if (((int) (nowtm->tm_min / 5) * 5) == (nowtm->tm_min)) {	/* 5 min */
	 G->hook_5minutely->Call();
	 G->log_list->Flush();
      }
      context;
      if (nowtm->tm_min == notify_users_at)
	G->hook_hourly->Call();
      context; /* these no longer need checking since they are
		* all check vs minutely settings and we only
		* get this far on the minute */
      if (miltime == switch_logfiles_at) {
	 char s[128];
	 
	 G->hook_daily->Call();
	 s[G->StrCpy(s, ctime(&G->now)) - 1] = 0;
	 s[11] = 0;
	 G->LOG_MISC.Put(G->Lang( 0x1EB, "ss", s, s+20));
	 G->userlist->Backup();
	 G->log_list->Close();
	 if (!G->keep_all_logs) {
	    G->LOG_MISC.Put(G->Lang( 0x1EC, ""));
	    G->log_list->Backup();
	 }
      }
   }
}

static void core_minutely (void) {
   FUNCTION;
   char y[21];
   void * args = &nowtm;
   G->SPrintF(y, 20, "%2 %2 %2 %2 %4", nowtm->tm_min, nowtm->tm_hour,
	      nowtm->tm_mday, nowtm->tm_mon + 1, nowtm->tm_year + 1900);
   G->bt_time->Check(y, 0, &args);
}

static void core_hourly (void) {
   FUNCTION;
   context;
   G->userlist->WriteUserfile(NULL);
}

static void init_bind (void) {
   FUNCTION;
   G->bind_list = NEW ( BindTable () );
   G->bt_unld = NEW ( BindTableString ( "unld" ) );
   G->bt_time = NEW ( BindTableTime () );
   G->bt_note = NEW( BindTableNote ( "note" ) );
   G->bt_strnote = NEW( BindTableNote ( "strnote" ) );
   G->bt_lnkch = NEW( BindTableLNkch () );
   G->bt_name = NEW( BindTable("name", BindTable::STACKABLE | BindTable::WANTRET));
   G->bt_load = NEW ( BindTableString ( "load" ) );
   G->bt_listen = NEW ( BindTableListen ( ) );
   G->bt_filt = NEW ( BindTableFilt () );
   G->bt_die = NEW ( BindTableString ( "die" ) );
   G->bt_dcc = NEW ( BindTableDcc ( "dcc" ) );
   G->bt_chon = NEW ( BindTableChon () );
   G->bt_chof = NEW ( BindTableChof () );
   G->bt_lchpt = NEW ( BindTableLChpt () );
   G->bt_lchjn = NEW ( BindTableLChjn () );
   G->bt_chat = NEW ( BindTableCAB ( "chat" ) );
   G->bt_bcst = NEW ( BindTableCAB ( "bcst" ) );
   G->bt_away = NEW ( BindTableAway ( ) );
   G->bt_act = NEW ( BindTableCAB ( "act" ) );
   G->bt_dcc->AddBinds(C_dcc, DCC_COMMANDS, 0);
}
      
int main (int argc, char ** argv) {
   FUNCTION;
   int xx, i;
   char s[520];
   FILE *f;
   struct sigaction sv;
   Socket * ptr;
   DccChat * dcc;
   UserRecord * u;
   LogModes ALL (LogModes::ALL);
   Channel * chan;
   /* statics */
   ModuleEntry me;
   ModuleEntryEggdrop megg ( me );
   MemType memtype_list;
   
   G = &megg;
   megg.cmd_die = cmd_die;
   megg.cmd_note = cmd_note;
   megg.cmd_boot = cmd_boot;
   G->memtype_list = &memtype_list;
   G->memtype_modules = NEW ( MemType ( "modules", (List **)&G->module_list ) );

   megg.Patches(&egg_numver, G->full_version, G->version, main_patches);
   megg.ver = egg_numver;
   egg_numver.AsNumber(num_version, 9);
   G->SPrintF(megg.long_version, VERSION_LEN, 
	      "E%s copyright 1998 Darrin Smith", G->version + 1);
   context;
#ifdef STOP_UAC
   {
      int nvpair[2];
      nvpair[0] = SSIN_UACPROC;
      nvpair[1] = UAC_NOPRINT;
      setsysinfo(SSI_NVPAIRS, (char *) nvpair, 1, NULL, 0);
   }
#endif
   /* set up error traps: */
   sv.sa_handler = got_bus;
   sigemptyset(&sv.sa_mask);
#ifdef SA_RESETHAND
   sv.sa_flags = SA_RESETHAND;
#else
   sv.sa_flags = 0;
#endif
   sigaction(SIGBUS, &sv, NULL);
   sv.sa_handler = got_segv;
   sigaction(SIGSEGV, &sv, NULL);
#ifdef SA_RESETHAND
   sv.sa_flags = 0;
#endif
   sv.sa_handler = got_fpe;
   sigaction(SIGFPE, &sv, NULL);
   sv.sa_handler = got_term;
   sigaction(SIGTERM, &sv, NULL);
   sv.sa_handler = got_hup;
   sigaction(SIGHUP, &sv, NULL);
   sv.sa_handler = got_quit;
   sigaction(SIGQUIT, &sv, NULL);
   sv.sa_handler = SIG_IGN;
   sigaction(SIGPIPE, &sv, NULL);
   sv.sa_handler = got_usr1;
   sigaction(SIGUSR1, &sv, NULL);
   sv.sa_handler = got_usr2;
   sigaction(SIGUSR2, &sv, NULL);
   sv.sa_handler = got_ill;
   sigaction(SIGILL, &sv, NULL);
   sv.sa_handler = got_alarm;
   sigaction(SIGALRM, &sv, NULL);

   /* initialize variables and stuff */
   G->now = time(NULL);
   nowtm = localtime(&G->now);
   lastmin = nowtm->tm_min;
   srandom(G->now);
   
   G->chanset = 0;
   G->memtype_hostcache = NEW ( MemType ( "hostcache", (List **)&G->host_cache) );
   G->host_cache = NEW ( HostnameCache () );
   G->memtype_socket = NEW ( MemType ( "socket", (List **)&G->socklist ) );
   G->socklist = NEW ( Socket () );
   G->Stdout = NEW ( SpecialSocket( 0 ) );
   G->Stderr = NEW ( SpecialSocket( 1 ) );
   G->Log = NEW ( SpecialSocket( 2 ) );
   FD_ZERO(&G->fds_read);
   FD_ZERO(&G->fds_write);
   G->memtype_language = NEW ( MemType ( "language", (List **)&G->langtab_list ) );
   G->langtab_list = NEW ( ModuleLanguage () );
   NEW ( ModuleLanguage ( "eggdrop" ) );
   if (!G->LoadLanguage("core.english")) 
      G->LOG_MISC.Put("Can't load default language file!!");
   if (argc > 1)
      for (i = 1; i < argc; i++)
	 do_arg(argv[i]);
   printf("\n%s\n", G->long_version);
   G->memtype_ban = NEW ( MemType ( "ban", (List **)&G->global_bans) );
   G->global_bans = NEW ( BanRecord () );
   G->memtype_user = NEW ( MemType ( "user", (List **)&G->userlist ) );
   G->userlist = NEW ( UserRecord ( ) );
   G->memtype_chanrec = NEW ( MemType ( "chanrec", (List **)&G->null_chanrec ) );
   G->null_chanrec = NEW ( ChannelRecord ( ) );
   G->memtype_ignore = NEW ( MemType ( "ignore", (List **)&G->ignores) );
   G->ignores = NEW ( IgnoreRecord () );
   G->memtype_userent = NEW ( MemType ( "userent", (List **)&G->entrytype_list ) );
   G->entrytype_list = NEW ( UserEntry () );
   G->entrytype_hosts = NEW ( UserEntryHosts ( ) );
   G->entrytype_xtra = NEW ( UserEntryXtra ( ) );
   G->entrytype_botfl = NEW ( UserEntryBotfl ( ) );
   G->entrytype_laston = NEW ( UserEntryLaston () );
   G->entrytype_info = NEW ( UserEntryInfo ( ) );
   G->entrytype_comment = NEW ( UserEntryComment ( ) );
   G->entrytype_pass = NEW ( UserEntryPass ( ) );
   G->entrytype_unknown = NEW ( UserEntryUnknown ( 1 ) ); 
   G->memtype_help = NEW ( MemType ( "help", (List **)&G->help_list ) );
   G->help_list = NEW ( HelpReference () );
   G->memtype_hook = NEW ( MemType ( "hooks", (List **)&G->hook_list ) );
   G->hook_list= NEW ( HookList ( ) );
   G->hook_secondly= NEW ( HookList ( "secondly" ) );
   G->hook_minutely= NEW ( HookList ( "minutely" ) );
   G->hook_5minutely= NEW ( HookList ( "5minutely" ) );
   G->hook_hourly= NEW ( HookList ( "hourly" ) );
   G->hook_daily= NEW ( HookList ( "daily" ) );
   G->hook_idle= NEW ( HookList ( "idle" ) );
   G->hook_pre_rehash= NEW ( HookList ( "pre_rehash" ) );
   G->hook_rehash= NEW ( HookList ( "rehash" ) );
   G->hook_read_userfile= NEW ( HookList ( "read_userfile" ) );
   G->hook_write_userfile= NEW ( HookList ( "write_userfile" ) );
   G->memtype_dep = NEW ( MemType ( "dependancies", (List **)&G->dep_list) );
   G->dep_list = NEW ( Dependancy () );
   G->memtype_static = NEW ( MemType ( "statics", (List **)&G->statics) );
   G->statics = NEW ( StaticMods () );
   G->memtype_script = NEW ( MemType ( "scripts", (List **)&G->scripts ) );
   G->scripts = NEW ( ScriptContext ( ) );
   NEW ( StringVariable ( "userfile", G->userfile, 120, VAR_PROTECT, 0 ) );
   NEW ( StringVariable ( "motd", G->motdfile, 120, VAR_PROTECT, 0 ) );
   NEW ( StringVariable ( "admin", G->admin, 120, 0, 0 ) );
   NEW ( StringVariable ( "help_path", G->helpdir, 120, VAR_DIR|VAR_PROTECT, 0 ) );
   NEW ( StringVariable ( "temp_path", G->tempdir, 120, VAR_DIR|VAR_PROTECT, 0 ) );
#ifndef STATIC_ONLY
   NEW ( StringVariable ( "mod_path", G->moddir, 120, VAR_DIR|VAR_PROTECT, 0 ) );
#endif
   NEW ( StringVariable ( "notify_newusers", G->notify_new, 120, 0, 0 ) );
   NEW ( StringVariable ( "owner", G->owner, 120, VAR_PROTECT, 0 ) );
   NEW ( StringVariable ( "my_ip", G->myip, 120, 0, 0 ) );
   NEW ( StringVariable ( "whois_fields", G->whois_fields, 120, 0, 0 ) );
   NEW ( StringVariable ( "nat_ip", G->natip, 120, 0, 0 ) );
   NEW ( StringVariable ( "username", G->botuser, 10, 0, 0 ) );
   NEW ( StringVariable ( "version", G->version, 0, 0, 0 ) );
   NEW ( StringVariable ( "config", G->configfile, 0, 0, 0 ) );
   NEW ( StringVariable ( "saved_config", G->savedconfigfile, 120, VAR_PROTECT, 0 ) );
   NEW ( StringVariable ( "console_user", console_user, HANDLEN, VAR_PROTECT, 0 ) );
   NEW ( StringVariable ( "numversion", num_version, 0, 0, 0 ) );
   NEW ( IntegerVariable ( "ignore_time", &G->ignore_time, 0, 0 ) );
   NEW ( IntegerVariable ( "dcc_flood_thr", &G->dcc_flood_thr, 0, 0 ) );
   NEW ( IntegerVariable ( "hourly_updates", &notify_users_at, 0, 0 ) );
   NEW ( IntegerVariable ( "switch_logfiles_at", &switch_logfiles_at, 0, 0 ) );
   NEW ( IntegerVariable ( "connect_timeout", (int *)&G->connect_timeout, 0, 0 ) );
   NEW ( IntegerVariable ( "password_timeout", (int *)&G->password_timeout, 0, 0 ) );
   NEW ( IntegerVariable ( "ident_timeout", (int *)&G->ident_timeout, 0, 0 ) );
   NEW ( IntegerVariable ( "reserved_port", &G->reserved_port, 0, 0 ) );
   NEW ( IntegerVariable ( "require_p", &G->require_p, 0, 0 ) );
   NEW ( IntegerVariable ( "keep_all_logs", &G->keep_all_logs, 0, 0 ) );
   NEW ( IntegerVariable ( "open_telnets", &G->allow_new_telnets, 0, 0 ) );
   NEW ( IntegerVariable ( "uptime", (int *) &G->online_since, VAR_READONLY, 0 ) );
   NEW ( IntegerVariable ( "console", (int *)&G->conmask, 0, 0 ) );
   NEW ( IntegerVariable ( "default_flags", &G->default_flags, 0, 0 ) );
   NEW ( IntegerVariable ( "die_on_sighup", &die_on_sighup, VAR_PROTECT, 0 ) );
   NEW ( IntegerVariable ( "die_on_sigterm", &die_on_sigterm, VAR_PROTECT, 0 ) );
   NEW ( IntegerVariable ( "max_dcc", &G->max_dcc, 0, 0 ) );
   NEW ( IntegerVariable ( "debug_output", &G->debug_output, VAR_PROTECT, 0 ) );
   NEW ( IntegerVariable ( "protect_telnet", &G->protect_telnet, 0, 0 ) );
   NEW ( IntegerVariable ( "log_time", &G->shtime, 0, 0 ) );
   NEW ( IntegerVariable ( "allow_dk_cmds", &G->allow_dk_cmds, 0, 0 ) );
   NEW ( IntegerVariable ( "max_sendq", &G->max_sendq, 0, 0 ) );
   NEW ( IntegerVariable ( "port_range", &G->port_range, 0, 0 ) );
   NEW ( IntegerVariable ( "enable_simul", &enable_simul, VAR_PROTECT, 0 ) );
   NEW ( IntegerVariable ( "resolve_timeout", &G->resolve_timeout, 0, 0 ) );
   NEW ( CoupletVariable ( "telnet_flood", &G->flood_telnet_thr, &G->flood_telnet_time, 0 ) );
   NEW ( CoupletVariable ( "hostcache", &G->hostcache_count, &G->hostcache_age, 0 ) );
   NEW ( Firewall () );
   NEW ( Command ( "wildmatch", script_wildmatch, 0 ) );
   NEW ( Command ( "bind", script_bind, 0 ) );
   NEW ( Command ( "unbind", script_unbind, 0 ) );
   NEW ( Command ( "logfile", script_logfile, 0 ) );
   NEW ( Command ( "language", script_language, 0 ) );
   NEW ( Command ( "maskhost", script_maskhost, 0 ) );
   NEW ( Command ( "putdcc", script_putdcc, 0 ) );
   NEW ( Command ( "dccsimul", script_dccsimul, 0 ) );
   NEW ( Command ( "dccbroadcast", script_dccbroadcast, 0 ) );
   NEW ( Command ( "hand2idx", script_hand2idx, 0 ) );
   NEW ( Command ( "getchan", script_getchan, 0 ) );
   NEW ( Command ( "setchan", script_setchan, 0 ) );
   NEW ( Command ( "dccputchan", script_dccputchan, 0 ) );
   NEW ( Command ( "console", script_console, 0 ) );
   NEW ( Command ( "strip", script_strip, 0 ) );
   NEW ( Command ( "echo", script_echo, 0 ) );
   NEW ( Command ( "page", script_page, 0 ) );
   NEW ( Command ( "valididx", script_valididx, 0 ) );
   NEW ( Command ( "killdcc", script_killdcc, 0 ) );
   NEW ( Command ( "idx2hand", script_idx2hand, 0 ) );
   NEW ( Command ( "dcclist", script_dcclist, 0 ) );
   NEW ( Command ( "dccused", script_dccused, 0 ) );
   NEW ( Command ( "getdccidle", script_getdccidle, 0 ) );
   NEW ( Command ( "getdccaway", script_getdccaway, 0 ) );
   NEW ( Command ( "setdccaway", script_setdccaway, 0 ) );
   NEW ( Command ( "listen", script_listen, 0 ) );
   NEW ( Command ( "killlisten", script_killlisten, 0 ) );
   NEW ( Command ( "rehash", script_rehash, 0 ) );
   NEW ( Command ( "restart", script_restart, 0 ) );
   NEW ( Command ( "putloglev", script_putloglev, 0 ) );
   NEW ( Command ( "unixtime", script_unixtime, 0 ) );
   NEW ( Command ( "time", script_time, 0 ) );
   NEW ( Command ( "date", script_date, 0 ) );
   NEW ( Command ( "ctime", script_ctime, 0 ) );
   NEW ( Command ( "myip", script_myip, 0 ) );
   NEW ( Command ( "rand", script_rand, 0 ) );
   NEW ( Command ( "sendnote", script_sendnote, 0 ) );
   NEW ( Command ( "dccdumpfile", script_dccdumpfile, 0 ) );
   NEW ( Command ( "backup", script_backup, 0 ) );
   NEW ( Command ( "die", script_die, 0 ) );
   NEW ( Command ( "strftime", script_strftime, 0 ) );
   NEW ( Command ( "unames", script_unames, 0 ) );
   NEW ( Command ( "unloadmodule", script_unloadmodule, 0 ) );
   NEW ( Command ( "loadmodule", script_loadmodule, 0 ) );
   NEW ( Command ( "checkmodule", script_loadmodule, 0 ) );
   NEW ( Command ( "modules", script_modules, 0 ) );
   NEW ( Command ( "loadhelp", script_loadhelp, 0 ) );
   NEW ( Command ( "unloadhelp", script_unloadhelp, 0 ) );
   NEW ( Command ( "reloadhelp", script_reloadhelp, 0 ) );
   NEW ( Command ( "countusers", script_countusers, 0 ) );
   NEW ( Command ( "validuser", script_validuser, 0 ) );
   NEW ( Command ( "finduser", script_finduser, 0 ) );
   NEW ( Command ( "passwdok", script_passwdOk, 0 ) );
   NEW ( Command ( "chattr", script_chattr, 0 ) );
   NEW ( Command ( "matchattr", script_matchattr, 0 ) );
   NEW ( Command ( "adduser", script_adduser, 0 ) );
   NEW ( Command ( "deluser", script_deluser, 0 ) );
   NEW ( Command ( "delhost", script_delhost, 0 ) );
   NEW ( Command ( "userlist", script_userlist, 0 ) );
   NEW ( Command ( "save", script_save, 0 ) );
   NEW ( Command ( "reload", script_reload, 0 ) );
   NEW ( Command ( "chnick", script_chnick, 0 ) ); 
   NEW ( Command ( "isignore", script_isignore, 0 ) );
   NEW ( Command ( "newignore", script_newignore, 0 ) );
   NEW ( Command ( "killignore", script_killignore, 0 ) );
   NEW ( Command ( "ignorelist", script_ignorelist, 0 ) );
   NEW ( Command ( "getuser", script_getuser, 0 ) );
   NEW ( Command ( "setuser", script_setuser, 0 ) );
   NEW ( Command ( "suidx", script_suidx, 0 ) );
   G->memtype_bind = NEW ( MemType ( "binds", (List **)&G->bind_list ) );
   G->memtype_log = NEW ( MemType ( "log", (List **)&G->log_list ) );
   G->log_list = NEW ( LogFile () );
   context;
   init_bind();
   context;
   strcpy(s, ctime(&G->now));
   s[strlen(s) - 1] = 0;
   strcpy(&s[11], &s[20]);
   ALL.Put(G->Lang( 0x1ED, "ss", G->VersionStr(), s));
   G->Chanprog();
   G->LoadConfig(0);
   context;
   if (!G->module_encrypt) {
      printf(G->Lang( 0x0F4, ""));
      exit(1);
   }
   for (i = 0, chan = G->chanset; chan; chan = chan->Next(), i++);
   ALL.Put(G->Lang(0x1EE,"sdd",G->origbotname, (i-1), G->userlist->Count()));
   context;
   G->SPrintF(G->pid_file, 40, "pid.%s", G->botnetnick);
   context;
   /* check for pre-existing eggdrop! */
   if ((f = fopen(G->pid_file, "r"))) {
      fgets(s, 10, f);
      xx = atoi(s);
      kill(xx, SIGCHLD);  /* meaningless kill to determine if pid is used */
      if (errno != ESRCH) {
	 printf(G->Lang( 0x201, "s", G->origbotname));
	 printf(G->Lang( 0x202, "s", G->pid_file));
	 exit(1);
      }
   }
   context;
   /* move into background? */
   if (G->backgrd) {
      xx = fork();
      if (xx == -1)
	G->Fatal(G->Lang( 0x1EF, ""));
      if (xx != 0) {
	 FILE *fp;
	 /* need to attempt to write pid now, not later */
	 unlink(G->pid_file);
	 if ((fp = fopen(G->pid_file, "w"))) {
	    fprintf(fp, "%u\n", xx);
	    if (fflush(fp)) {
	       /* kill bot incase a botchk is run from crond */
	       printf(G->Lang( 0x1F1, "d", G->pid_file));
	       G->Fatal("  Try freeing some disk space\n");
	    }
	    fclose(fp);
          } else
	   printf(G->Lang( 0x1F1, "d", G->pid_file));
	 printf(G->Lang( 0x1F0, "d", xx));
#if HAVE_SETPGID
	 setpgid(xx, xx);
#endif
	 exit(0);
      }
   }
   context;
   G->use_stderr = 0;		/* stop writing to stderr now */
   if ((xx = getpid()) && !G->backgrd) {
      FILE *fp;
      /* write pid to file */
      unlink(G->pid_file);
      if ((fp = fopen(G->pid_file, "w"))) {
	 fprintf(fp, "%u\n", xx);
	 if (fflush(fp)) {
	    printf(G->Lang( 0x1F1, "s", G->pid_file));
	    fclose(fp);
	    unlink(G->pid_file);
	 } else
	   fclose(fp);
      } else
	printf(G->Lang( 0x1F1, "s", G->pid_file));
   }
   context;
   if (G->backgrd) {
#if HAVE_SETPGID
      setpgid(0, 0);
#endif
      freopen("/dev/null", "r", stdin);
      freopen("/dev/null", "w", stdout);
      freopen("/dev/null", "w", stderr);
   }
   context;
   /* terminal emulating dcc chat */
   if (!G->backgrd && G->term_z) {
      dcc = NEW ( DccChat ( STDIN ) );
      dcc->SetAddress(iptolong(G->GetMyIp()));
      if (!(u = G->userlist->Find(console_user))) {
	 FlagRecord fr;
	 fr.SetGlobParty();
	 u = NEW ( UserRecord ( G->userlist, console_user, "none", fr) );
      }
      dcc->SetUser(u);
      dcc->SetNick(console_user);
      dcc->Send(G->Lang( 0x1F3, ""));
      dcc->Chatter();
      dcc->SetStatus(STAT_CHAT);
      dcc->SetConsoleChan("*");
      dcc->ConsoleFlags(G->conmask);
      FD_SET(dcc->FD(),&G->fds_read);
   }
   then = G->now;
   G->online_since = G->now;
   NEW ( HelpReference ( "cmds1.help" ) );
   NEW ( HelpReference ( "cmds2.help" ) );
   G->hook_secondly->Add(core_secondly, G->module_eggdrop);
   G->hook_minutely->Add(core_minutely, G->module_eggdrop);
   G->hook_hourly->Add(core_hourly, G->module_eggdrop);
   while (1) {
      context;
      G->now = time(NULL);
      random();
      if (G->now != then) {
	 G->hook_secondly->Call();
	 then = G->now;
      }
      if (Socket::Handle()) {
	 context;
	 G->LOG_MISC.Put(G->Lang( 0x1F4, "d", errno));
	 ptr = G->socklist;
	 while ((ptr = ptr->Next())) {
	    if (!(ptr->Flags() & (Socket::NONSOCK | Socket::LOST))
		&& (fcntl(ptr->FD(), F_GETFD, 0) == -1) && (errno == EBADF)) {
	       G->LOG_MISC.Put(G->Lang( 0x1D4, "cdss", ptr->FD() < 0 ? '-' : '+',
				       ptr->FD(), ptr->Type(), ptr->Nick()));
	       ptr->Eof();
	    }
	 }
      } else 
	G->hook_idle->Call();
      if (G->do_restart) {
	 if (G->do_restart == -2)
	    G->Rehash();
	 else {
	    /* unload as many modules as possible */
	    int f = 1;
	    ModuleEntry * p;
	    char xx[256];

	    while (f) {
	       f = 0;
	       p = G->module_list;
	       while ((p = p->Next())) {
		  Dependancy * d = G->dep_list;

		  int ok = 1;
		  while (ok && (d = d->Next()))
		     if (d->Needed() == p)
			ok = 0;
		  if (ok) {
		     strncpy(xx, p->Key(), (sizeof(xx)-1));
		     if (strcasecmp(xx, "eggdrop")) {
			if (!G->module_list->Unload(xx)) {
			   f = 1;
			   break;
			}
		     }
		  }
	       }
	    }
	    p = G->module_list;
	    if (p && p->Next() && p->Next()->Next())
	       G->LOG_MISC.Put(G->Lang(0x2A4, ""));
	    context;
	    G->log_list->Flush();
	    context;
	    DELETE ( G->bind_list );
	    init_bind();
	    G->Rehash();
	    ptr = G->socklist;
	    while ((ptr = ptr->Next()))
	       if (ptr->Type() == G->DCC_CHAT) {
		  DccChat * dcc = (DccChat *)ptr;
		  dcc->CheckChon();
	       }
	 }
	 G->do_restart = 0;
      }
   }
}
