/*
 * Main.h - the *one* include file
 * 
 */

#ifndef _MAIN_
# define _MAIN_

# ifdef HAVE_CONFIG_H
#  include "../config.h"
# endif
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <stdarg.h>
# include <sys/types.h>
# include <netinet/in.h>
# include <sys/stat.h>

/***********************************************************************
 * numeric defines                                                     *
 ***********************************************************************/

/*
 * define the maximum length a handle on the bot can be.
 * (standard is 9 characters long)
 * (DO NOT MAKE THIS VALUE LESS THAN 9 UNLESS YOU WANT TROUBLE!)
 * (beware that using lengths over 9 chars is 'non-standard' and if you
 * wish to link to other bots, they _must_ both have the same maximum
 * handle length)
 */

# define HANDLEN		     9 /* valid values 9->NICKMAX */
# define NICKMAX             15 /* valid values HANDLEN->32 */
# define MAXLEN              32 /* absolute maximum hand/nick length */
# define MISCLEN             32 /* length of 'misc' strings */
# define CHANLEN             81
# define UHOSTLEN           161 /* reasonable, i think? */
# define DIRLEN             255 /* paranoia */
# define VERSION_LEN        127 /* bytes to allocate for version # storage */
# define PATCH_LEN         1023 /* bytes to allow for patch string */
# define HELP_BUF_LEN       256

# define NOTENAMELEN     ((HANDLEN * 2) + 1)

/*************************************************************************
 * DONT TOUCH ANYTHING BELOW THIS LINE (unless you know what your doing) *
 * this next bit checks various things                                   *
 *************************************************************************/

# if !HAVE_VSPRINTF
#  include "error_you_need_vsprintf_to_compile_eggdrop"
# endif

# if HAVE_UNISTD_H
#  include <unistd.h>
# endif

# ifndef STATIC_ONLY
#  if (!defined(MODULES_OK) || !defined(HAVE_DLOPEN)) && !defined(HPUX_HACKS)
#   include "you_can't_compile_with_module_support_on_this_system_try_make_static"
#  endif
# endif

# if !defined(STDC_HEADERS)
#  include "you_need_to_upgrade_your_compiler_to_a_standard_c_one_mate!"
# endif

# if (NICKMAX < 9) || (NICKMAX > MAXLEN)
#  include "invalid NICKMAX value"
# endif

# if (HANDLEN < 9) || (HANDLEN > MAXLEN)
#  include "invalid HANDLEN value"
# endif

# if HANDLEN > NICKMAX
#  include "HANDLEN MUST BE <= NICKMAX"
# endif

# ifndef __cplusplus
#  include "Eggdrop now REQUIRES C++ to compile"
# endif

/* almost every module needs some sort of time thingy, so... */
# if TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
# else
#  if HAVE_SYS_TIME_H
#   include <sys/time.h>
#  else
#   include <time.h>
#  endif
# endif

# if !HAVE_RENAME
#  define rename movefile
# endif

# if !HAVE_SRANDOM
#  define srandom(x) srand(x)
# endif

# if !HAVE_RANDOM
#  define random() (rand()/16)
# endif

# if !HAVE_SIGACTION     /* old "weird signals" */
#  define sigaction sigvec
#  ifndef sa_handler
#   define sa_handler sv_handler
#   define sa_mask sv_mask
#   define sa_flags sv_flags
#  endif
# endif

# if !HAVE_SIGEMPTYSET
/* and they probably won't have sigemptyset, dammit */
#  define sigemptyset(x) ((*(int *)(x))=0)
# endif

/* Stupid Borg Cube crap ;p */
# ifdef BORGCUBES

/* net.h needs this */
#  define O_NONBLOCK      00000004        /* POSIX non-blocking I/O       */

/* mod/filesys.mod/filedb.c needs this */
#  define _S_IFMT         0170000         /* type of file */
#  define   _S_IFDIR      0040000         /*   directory */
#  define S_ISDIR(m)      (((m)&(_S_IFMT)) == (_S_IFDIR))

# endif
# ifdef DEBUG_MEM_PARANOID
#  ifndef EBUG_MEM
#   define EBUG_MEM
#  endif
# endif

typedef int (*Function) ();
/***********************************************************************
 * STRUCTURES GALORE                                                   *
 ***********************************************************************/

class BindTable;
class BindEntry;
class BindMask;
class ModuleLanguage;
class Socket;
class MemType;
class MessageQueue;
class UserRecord;
class HelpReference;
template <int len> class KeyedList;
class IgnoreRecord;
class BanRecord;
class UserEntryType;
class UserEntry;
class HookList;
class LogFile;
class HostnameCache;
class ChannelRecord;
class ModuleEntry;
class ModuleEntryEggdrop;
class ModuleEntryEncryption;
class Version;
class Dependancy;
class StaticMods;
class FlagRecord;
class DccChat;
class DccChatPass;
class DccChatLogin;
class DccChatNewUser;
class IdentWait;
class ScriptContext;
class ScriptArgument;
class Channel;
class Command;
class BoundCommand;
class Variable;

enum BindReturn {
   NOMATCH, AMBIGUOUS, EXECUTED, EXEC_LOG, EXEC_BRK
};

struct command_binds {
   char *name;
   char *flags;
   BindReturn (*func) ( void **, const char *);
   char *funcname;
};

struct patches {
   char * name;
   int number;
};

# ifndef _LOGMODES_
#  include "LogModes.h"
# endif

# define MEMORY_TYPE(A) \
inline void * operator new (size_t s, int x) { return A->Malloc(s, x); } \
inline void operator delete (void * ptr) { A->Free(ptr); }
# define DEF_NEXT(A) \
inline A * Next ( void ) { return (A *)next; } 
# define DEF_NEXT_EXPMEM(A) \
inline A * Next ( void ) { return (A *)next; } \
inline int ExpMem ( void ) { A * ptr = this; int tot = sizeof(A); \
   while ((ptr = ptr->Next())) { tot += sizeof(A); } return tot; }

inline int is_file (char * s) {
   struct stat ss;
   return ((stat(s, &ss) >= 0) && 
	   ((ss.st_mode & S_IFREG) || (ss.st_mode & S_IFLNK)));
}

/********************************************************************
 * FLAGS / DEFINES / ETC                                            *
 ********************************************************************/

/* flags about dcc types */
# define DCT_CHAT      0x00000001 /* this dcc type receives botnet chatter */
# define DCT_MASTER    0x00000002 /* received master chatter */
# define DCT_SHOWWHO   0x00000004 /* show the user in .who */
# define DCT_REMOTEWHO 0x00000008 /* show in remote who */
# define DCT_VALIDIDX  0x00000010 /* valid idx for outputting to in script */
# define DCT_SIMUL     0x00000020 /* can be script_simul'd */
# define DCT_CANBOOT   0x00000040 /* can be booted */
# define DCT_GETNOTES  DCT_CHAT   /* can received notes */
# define DCT_FILES     0x00000080 /* gratuitous hack ;) */
# define DCT_BOT       0x00000100 /* a bot connection of some sort... */
# define DCT_REALBOT   0x00000200 /* a currently connected bot */
# define DCT_FILETRAN  0x00000400 /* a file transfer of some sort */
# define DCT_FILESEND  0x00000800 /* a sending file transfer, getting = !this */

/* for dcc chat & files: */
# define STAT_ECHO    1      /* echo commands back? */
# define STAT_DENY    2      /* bad username (ignore password & deny access) */
# define STAT_CHAT    4      /* in file-system but may return */
# define STAT_TELNET  8      /* connected via telnet */
# define STAT_PARTY   16     /* only on party line via 'p' flag */
# define STAT_BOTONLY 32     /* telnet on bots-only connect */
# define STAT_USRONLY 64     /* telnet on users-only connect */
# define STAT_PAGE    128    /* page output to the user */

/* for stripping out mIRC codes */
# define STRIP_COLOR  1      /* remove mIRC color codes */
# define STRIP_BOLD   2      /* remove bold codes */
# define STRIP_REV    4      /* remove reverse video codes */
# define STRIP_UNDER  8      /* remove underline codes */
# define STRIP_ANSI   16     /* remove ALL ansi codes */
# define STRIP_BELLS  32     /* remote ctrl-g's */
# define STRIP_HILITE 64     /* don't show hilighting */
# define STRIP_ALL    127    /* remove every damn thing! */

/* for local console: */
# define STDIN      0
# define STDOUT     1
# define STDERR     2

/* return codes for add_note */
# define NOTE_ERROR      0   /* error */
# define NOTE_OK         1   /* success */
# define NOTE_STORED     2   /* not online; stored */
# define NOTE_FULL       3   /* too many notes stored */
# define NOTE_SCRIPT     4   /* script binding caught it */
# define NOTE_AWAY       5   /* away; stored */
# define NOTE_FWD        6   /* away; forwarded */

#define VAR_READONLY 1
#define VAR_PROTECT  2
#define VAR_DIR      4
#define VAR_LIST     8

#define SCRIPT_OK    0
#define SCRIPT_ERROR 1

# define CHANLIST_HAND    0 /* check the handle vs handles */
# define CHANLIST_GETHOST 1 /* check user@host */
# define CHANLIST_SETHOST 2 /* set a found user@host */
# define CHANLIST_DELHOST 3 /* removed anyone except the given user matching */
# define CHANLIST_DELHAND 4 /* remove any of the user matching */
# define CHANLIST_CLEAR   5 /* clear all */

/* minimum version that uses tokens & base64 ints */
/* for channel msg's */
# define GLOBAL_CHANS 100000
# define MAX_CHANS    200000

/*********************************************************************
 * USEFULL SHORT-CUT DEFINES                                         *
 *********************************************************************/

# ifndef MAKING_MODS
#  define MODULE_NAME "eggdrop"
# endif

inline int iptolong (struct in_addr * x) { return htonl(x->s_addr); }
inline const char * safe_str ( const char * a) { return a ? a : ""; }
inline void strncpyz (char * x, const char * y, int l) {
   strncpy(x, y, l);
   x[l] = 0;
}
# ifndef MEM_C
#  undef malloc
#  undef free
#  undef realloc
#  define malloc use-my_malloc-not-malloc
#  define free use-my_free-not-free
#  define realloc real-men-dont-need-to-use-realloc
# endif

/* these NEED to be #defines */
# define context \
List::G->SPrintF(List::G->context_stack[List::G->context_ptr=((List::G->context_ptr)+1)&15],\
63, "%s/%s/%d", MODULE_NAME,__FILE__,__LINE__)
# define CONTEXT(X) { \
List::G->SPrintF(List::G->context_stack[List::G->context_ptr=((List::G->context_ptr)+1)&15],63, \
	      "%s/%s/%d(%s)", MODULE_NAME,__FILE__,__LINE__, X); \
}

# define my_malloc(a,b) b->Malloc(a,context)
# define my_free(a,b) {context; b->Free(a);}
# define NEW(a) new (context) a
# define DELETE(a) {context; delete a;}
# define list_insert(a,b) { b->next = *a; *a = b; }
# define BADARGS(nl,nh,example) \
if ((argc<(nl)) || (argc>(nh))) { \
sc->SetString(List::G->Lang(0x101, "ss", sc->Key(), example)); \
return SCRIPT_ERROR; \
}

#endif
