/*
 * BlowfishBox.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "blowfish.h"
#include "BlowfishBox.h"
#include "ModuleEntryBlowfish.h"
#define CLASS "BlowfishBox"

/* each box takes up 4k so be very careful here */
#define BOXES  3
#include "bf_tab.h"		/* P-box P-array, S-box  */

BlowfishBox::BlowfishBox (const char * n, ModuleEntry * m) 
: KeyedList<MISCLEN> ( *((ModuleEntryBlowfish *)m)->boxlist, n, 
		      KeyedList<MISCLEN>::NODUPS ) {
   FUNCTION;
   ModuleEntryBlowfish * meb = ((ModuleEntryBlowfish *)m);
   BlowfishBox * b = meb->boxlist, *i = 0;
   int t = 0, c = 0;

   keybytes = strlen(n);
   while ((b = b->Next())) {
      c++;
      if ((b->lastuse - G->now) > t) {
	 i = b;
	 t = b->lastuse - G->now;
      }
   }
   if (c > BOXES) 
     DELETE ( i );
}
     
BlowfishBox::~BlowfishBox ( void ) {
   FUNCTION;
}

int BlowfishBox::ExpMem ( void ) {
   FUNCTION;
   int tot = sizeof(BlowfishBox);
   BlowfishBox * b = this;
   
   context;
   while ((b = b->Next())) 
     tot += sizeof(BlowfishBox);
   return tot;

}

void BlowfishBox::Init (void) {
   FUNCTION;
   int i, j;
   UWORD_32bits data;
   UWORD_32bits datal;
   UWORD_32bits datar;
   union aword temp;

   lastuse = G->now;
   /* robey: reset blowfish boxes to initial state */
   /* (i guess normally it just keeps scrambling them, but here it's */
   /* important to get the same encrypted result each time) */
   for (i = 0; i < bf_N + 2; i++)
     P[i] = initbf_P[i];
   for (i = 0; i < 4; i++)
     for (j = 0; j < 256; j++)
       S[i][j] = initbf_S[i][j];
   j = 0;
   for (i = 0; i < bf_N + 2; ++i) {
      temp.word = 0;
      temp.w.byte0 = key[j];
      temp.w.byte1 = key[(j + 1) % keybytes];
      temp.w.byte2 = key[(j + 2) % keybytes];
      temp.w.byte3 = key[(j + 3) % keybytes];
      data = temp.word;
      P[i] = P[i] ^ data;
      j = (j + 4) % keybytes;
   }
   datal = 0x00000000;
   datar = 0x00000000;
   for (i = 0; i < bf_N + 2; i += 2) {
      Encipher(&datal, &datar);
      P[i] = datal;
      P[i + 1] = datar;
   }
   for (i = 0; i < 4; ++i) 
     for (j = 0; j < 256; j += 2) {
	Encipher(&datal, &datar);
	S[i][j] = datal;
	S[i][j + 1] = datar;
     }
}

#define F(x) (((S[0][x.w.byte0] + S[1][x.w.byte1]) ^ S[2][x.w.byte2]) + S[3][x.w.byte3])
#define ROUND(a,b,n) (a.word ^= F(b) ^ P[n])

void BlowfishBox::Encipher (UWORD_32bits * xl, UWORD_32bits * xr) {
   FUNCTION;
   union aword Xl;
   union aword Xr;

   Xl.word = *xl;
   Xr.word = *xr;

   Xl.word ^= P[0];
   ROUND(Xr, Xl, 1);
   ROUND(Xl, Xr, 2);
   ROUND(Xr, Xl, 3);
   ROUND(Xl, Xr, 4);
   ROUND(Xr, Xl, 5);
   ROUND(Xl, Xr, 6);
   ROUND(Xr, Xl, 7);
   ROUND(Xl, Xr, 8);
   ROUND(Xr, Xl, 9);
   ROUND(Xl, Xr, 10);
   ROUND(Xr, Xl, 11);
   ROUND(Xl, Xr, 12);
   ROUND(Xr, Xl, 13);
   ROUND(Xl, Xr, 14);
   ROUND(Xr, Xl, 15);
   ROUND(Xl, Xr, 16);
   Xr.word ^= P[17];

   *xr = Xl.word;
   *xl = Xr.word;
}

void BlowfishBox::Decipher (UWORD_32bits * xl, UWORD_32bits * xr) {
   union aword Xl;
   union aword Xr;

   Xl.word = *xl;
   Xr.word = *xr;

   Xl.word ^= P[17];
   ROUND(Xr, Xl, 16);
   ROUND(Xl, Xr, 15);
   ROUND(Xr, Xl, 14);
   ROUND(Xl, Xr, 13);
   ROUND(Xr, Xl, 12);
   ROUND(Xl, Xr, 11);
   ROUND(Xr, Xl, 10);
   ROUND(Xl, Xr, 9);
   ROUND(Xr, Xl, 8);
   ROUND(Xl, Xr, 7);
   ROUND(Xr, Xl, 6);
   ROUND(Xl, Xr, 5);
   ROUND(Xr, Xl, 4);
   ROUND(Xl, Xr, 3);
   ROUND(Xr, Xl, 2);
   ROUND(Xl, Xr, 1);
   Xr.word ^= P[0];

   *xl = Xr.word;
   *xr = Xl.word;
}

int BlowfishBox::Boxes ( void ) { 
   FUNCTION;
   return BOXES;
}

void * BlowfishBox::operator new ( size_t s, int t ) {
   FUNCTION;
   return ((ModuleEntryBlowfish *)List::G->module_encrypt)->memtype_blowfish
     ->Malloc(s,t);
}

void BlowfishBox::operator delete ( void * x ) {
   FUNCTION;
   ((ModuleEntryBlowfish *)List::G->module_encrypt)->memtype_blowfish->Free(x);
}
