/*
 * ModuleEntryBotservice.C
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryBotservice.h"
#include "ModuleEntryBotnet.h"
#include "Bot.h"
#include "JupeList.h"
#include "Reconnections.h"
#include "IntegerVariable.h"
#include "CoupletVariable.h"
#include "StringVariable.h"
#include "TandemBot.h"
#include "FlagRecord.h"
#include "UserRecord.h"
#define CLASS "ModuleEntryBotservice"

static Version botservice_version;
static struct patches botservice_patches [] = {
#include "patch.h"
};

ModuleEntryBotservice::ModuleEntryBotservice ( ModuleEntry * old, char ** r) 
: ModuleEntry ("botservice", botservice_version.Set(2,0), botservice_patches, old) {
   FUNCTION;
   Version required;
   
   BS = this;
   if (!(BOTNET = (ModuleEntryBotnet *)Depend("botnet", required.Set(2,0))))
     *r = "This module requires the botnet module";
   else {
      memtype_botservice = NEW ( MemType ("botservice", (List **)&jupe_list,this));
      jupe_list = NEW( JupeList () );
      recons = NEW ( Reconnections () );
      logger = LogModes::LEV1;
      autojupe_time = 24;
      max_depth = 100;
      warn_flud = 1;
      leaf_old = 0;
      recon_cnt = 5;
      recon_time = 600;
      NEW ( StringVariable ( "service_name", service_name, HANDLEN, 
			    VAR_PROTECT, this) );
      NEW ( IntegerVariable ("warn_flud", &warn_flud, 0, this) );
      NEW ( IntegerVariable ("leaf_old", &leaf_old, 0, this) );
      NEW ( IntegerVariable ("autojupe_time", &autojupe_time, 0, this) );
      NEW ( IntegerVariable ("max_depth", & max_depth, 0, this) );
      NEW ( CoupletVariable ( "bot_recons", &recon_cnt, &recon_time, this ) );
      context;
   }
}

ModuleEntryBotservice::~ModuleEntryBotservice ( void ) {
   FUNCTION;
   char buf[512];

   G->SPrintF(buf, 511, "BOTNET SERVICE MODULE %s unloaded.", service_name);
   BOTNET->ChatOut(buf);
   no_respawn = 1;
   if (service_socket) {
      service_socket->SendUnLinked(BOTNET->tandbot->Find(service_name),buf,0);
      service_socket->Close();
      DELETE ( service_socket );
   }
   G->LOG_MISC.Put("Unloading botservice..");
   BOTNET->RemMassbotFuncs(massbots);
   DELETE ( jupe_list );
   DELETE ( recons );
}

int ModuleEntryBotservice::ExpMem (void ) {
   FUNCTION;
   return sizeof(ModuleEntryBotservice) + ModuleEntry::ExpMem ();
}

void ModuleEntryBotservice::CheckJupes(void) {
   FUNCTION;
   TandemBot * o = BOTNET->tandbot, *t;
   context;
   while ((t = o->Next())) {
      CheckJupe(NullBot(), t->Name(), t->UplinkName(), 0, 0);
      if (o->Next() == t)
	o = t;
   }
}

void ModuleEntryBotservice::CheckJupe (Bot * sock, const char * bot,
				       const char * next,
				       char flag, int vernum) {
   FUNCTION;
   JupeList * jp = jupe_list->Match(bot);
   TandemBot * botentry;
   char buf[512], *reason;
   UserRecord * u;
   Bot * ptr;
   FlagRecord fr ( FlagRecord::BOT );
   int depth = 1;
   
   /* only check recons on the linking bot, not all those behind it */
   if (recon_cnt && (flag == '!')) {

      Reconnections * re = recons->Find(bot);
      
      if (re)
	re->Again();
      else
	NEW ( Reconnections ( bot ) );
   }
   if (!jp) {
      if ((botentry = BOTNET->tandbot->Find(next)))
	while ((botentry = botentry->Uplink()))
	  depth++;
      if (max_depth && (depth > max_depth)) 
	reason = "Maximum tree level exceeded";
      else if ((botentry = BOTNET->tandbot->Find(next)) &&
	  (botentry->Version() < 1020000) && leaf_old) {
	 bot = next;
	 reason = "1.1.5's may only be leaves";
      } else if ((botentry = BOTNET->tandbot->Find(bot)) && 
		 (botentry->Version() < 1010500) && (leaf_old > 1))
	   reason = "pre-1.1.5's do not make a nice botnet";
      else if (!sock && vernum) 
	reason = "fludnet bots are not welcome here";
      else
	return;
      jp = NEW(JupeList(bot, reason, service_name, autojupe_time * 60) );
   }
   botentry = BOTNET->tandbot->Find(bot);
   while (botentry && (botentry->Share() == '+')) 
     botentry = botentry->Uplink();
   if (!botentry) {
      if (next)
	BOTNET->LOG_BOTS.Put(0, "Can't jupe %s, it's sharing with me!", bot);
   } else if ((u = G->userlist->Find(botentry->Name())) 
	      && fr.Set(u)->BotHub()) {
      if (next)
	BOTNET->LOG_BOTS.Put(0, "Can't jupe %s, it's my hub!",botentry->Name());
   } else {
      if (strcasecmp(botentry->Name(),bot))
	G->SPrintF(buf, 511, "Rejected, sharing with jupedbot %s, (%s: %s)", 
		   jp->Key(), jp->Who(), jp->Reason());
      else
	G->SPrintF(buf, 511, "Rejected, requested by %s: %s", jp->Who(),
		   jp->Reason());
      if ((ptr = botentry->NextBot()) && strcasecmp(ptr->Nick(),botentry->Name()))
	ptr->SendUnLink(service_name, botentry, buf);
      else
	BOTNET->Unlink(0, 0, buf);
   }
}

