/*
 * Channel2.C - Channel info related to channels module
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "channels.h"
#include "Channel2.h"
#include "ScriptArgument.h"
#include "ScriptContext.h"
#include "ModuleEntryChannels.h"
#ifdef HAVE_SERVER_H
#include "server.h"
#include "ModuleEntryServer.h"
#endif
#include "BanRecord.h"
#define CLASS "Channel2"

Channel2::Channel2 ( const char * n ) : Channel ( n ) {
   FUNCTION;
   ChannelSettings * set = &((Channel2 *)G->chanset)->settings;

   bans = NEW ( BanRecord ( this ) );
   while ((set = set->Next()))
     set->Create(this);
}

Channel2::Channel2 ( Channel2 & o ) : Channel ( o ) {
   FUNCTION;
   ChannelSettings * cs = &o.settings;
   
   while ((cs = cs->Next()))
     cs->chan = this;
   settings.SetNext(o.settings.Next());
   o.settings.SetNext(0);
}

Channel2::Channel2 ( void ) : Channel () {
   FUNCTION;
   if (G->chanset) {
      Channel2 * oc = (Channel2 *)G->chanset;
      ChannelSettings * cs = &oc->settings;

      while ((cs = cs->Next()))
	cs->chan = this;
      settings.SetNext(oc->settings.Next());
      oc->settings.SetNext(0);
      next = G->chanset->Next();
      oc->next = 0;
      G->chanset = this;
      DELETE ( oc );
      CHNS->CHANSET_LIST = &settings;
   }
}

Channel2::~Channel2 ( void ) {
   FUNCTION;
   if ( bans ) {
      DELETE ( bans );
      bans = 0;
   }
   if (!key[0]) 
     while (next) {
	DELETE (next);
     }
}

int Channel2::ExpMem ( void ) {
   FUNCTION;
   int tot = 0;
   Channel *chan;

   context;
   for (chan = G->chanset; chan; chan = chan->Next())
      tot += sizeof(Channel2);
   return tot;
}

/* set sticky attribute for a ban */
int Channel2::SetStickyBan (char * uhost, int sticky) {
   FUNCTION;
   int j;
   BanRecord * u = (this == G->chanset) ? G->global_bans : bans;
   
   if (!(j = atoi(uhost)))
      j = -1;
   while ((u = u->Next())) {
      if (j >= 0)
	j--;
      if (!j || ((j < 0) && !strcasecmp(u->Key(),uhost))) {
	 if (sticky > 0)
	   u->AddFlag(BanRecord::STICKY);
	 else if (!sticky)
	   u->RemFlag(BanRecord::STICKY);
	 else /* we don't actually want to change, just skip over */
	   return 0;
	 if (!j)
	   strcpy(uhost,u->Key());
	 if (!G->noshare && G->shareout)
	   G->shareout(this,"s%s %d %s\n", uhost, sticky, key);
	 return 1;
      }
   }
   return (j >= 0) ? -j : 0;
}

/* returns 1 if temporary ban, 2 if permban, 0 if not a ban at all */
int Channel2::EqualsBan (const char * uhost) {
   FUNCTION;
   BanRecord * b;
   
   b = ((this == G->chanset) ? G->global_bans : bans)->Find((char *)uhost);
   return b ? (4 + b->Flags()) 
     : ((this == G->chanset) ? 0 : ((Channel2 *)G->chanset)->EqualsBan(uhost));
}

int Channel2::MatchBan (const char * uhost) {
   FUNCTION;
   BanRecord * b = ((this == G->chanset) ? G->global_bans : bans)->Match(uhost);
   
   return b ? 1 : ((this == G->chanset) ? 0 : 
		   ((Channel2 *)G->chanset)->MatchBan(uhost));
}

int Channel2::DelBan (char * who, int doit) {
   FUNCTION;
   int j = 0;
   BanRecord * u = (this == G->chanset) ? G->global_bans : bans;
   
   if (!strchr(who,'!') && (j = atoi(who))) {
      while ((u = u->Next()) && --j);
      if (u) 
	strcpy(who, u->Key());
      else
	return - j;
   } else 
     /* find matching host, if there is one */
     u = u->Find(who);
   if (u && doit) 
     DELETE ( u );
   return u != 0;
}

int Channel2::AddBan (const char * ban, const char * from, char * note,
		      time_t expire_time, int flags) {
   FUNCTION;
   char host[UHOSTLEN+1], s[UHOSTLEN*2+1];

   strncpy(host, ban, UHOSTLEN);
   /* choke check: fix broken bans (must have '!' and '@') */
   if (!strchr(ban, '!') && !strchr(ban, '@'))
     G->SPrintF(host, UHOSTLEN, "%s!*@*",ban);
   else if (!strchr(host, '@'))
     G->SPrintF(host, UHOSTLEN, "%s@*", ban);
   else if (!strchr(host, '!')) 
     G->SPrintF(host, UHOSTLEN, "*!%s", ban);
#ifdef HAVE_SERVER_H
   if (CHNS->SERVER)
     G->SPrintF(s, UHOSTLEN, "%s!%s", CHNS->SERVER->botname, 
		CHNS->SERVER->botuserhost);
   else
#endif
     gethostname(s + G->SPrintF(s, UHOSTLEN, "%s!%s@", G->origbotname,
			       G->botuser), UHOSTLEN);
   if (G->WildMatch(host, s)) {
      G->LOG_MISC.Put( CHNS->Lang(0x04A, ""));
      return 0;
   }
   /* it shouldn't expire and be sticky also */
   if (note[0] == '*') {
      flags |= BanRecord::STICKY;
      note++;
   }
   if (expire_time != 0L)
     flags &= ~BanRecord::STICKY;
   else
     flags |= BanRecord::PERM;
   /* new format: */
   NEW ( BanRecord( this, ban, from, note, expire_time, flags) );
   if (!G->noshare && G->shareout)
     G->shareout(this, "b%s %s %lu %s%s %s %s\n",
		 key[0] ? key : "*", host, expire_time - G->now,
		 (flags & BanRecord::STICKY) ? "s" : "",
		 (flags & BanRecord::PERM) ? "p": "-", from, note);
   return 1;
}

/* parse options for a channel */
int Channel2::Modify (ScriptContext * sc, int items, ScriptArgument ** item) {
   FUNCTION;
   int i, ret;
   ChannelSettings * chan_sets;
   
   for (i = 0; i < items; i++) {
      ret = 0;
      chan_sets = &settings;
      while ((chan_sets = chan_sets->Next()))
	ret = chan_sets->Set(item[i]->AsString(), item[i + 1] 
			     ? item[i + 1]->AsString() : "" , NULL);
      switch (ret) {
	case -1: /* invalid setting */
	 sc->SetString(CHNS->Lang(0x07D, "ss", item[i]->AsString(), key));
	 return SCRIPT_ERROR;
	case 0: /* unknown setting, silently ignore - it may be used later */
	case 1: /* single setting, move on */
	 break;
	case 2: /* both arguments used */
	 i++;
	 break;
      }
   }
   return SCRIPT_OK;
}

Channel2 * Channel2::NewChannel ( const char * x ) {
   FUNCTION;
   return NEW ( Channel2 ( x ) );
}

/* create new channel and parse commands */
int Channel2::Add (ScriptContext * sc, const char * newname, int items,
		   ScriptArgument ** itemlist) {
   FUNCTION;
   Channel2 *chan;
   
   if ((newname[0] != '#') && (newname[0] != '&') && (newname[0] != '+'))
      return SCRIPT_ERROR;
   if ((chan = FindChan(newname))) 
     /* already existing channel, maybe a reload of the channel file */
     chan->UnsetFlagged();
   else 
     chan = NewChannel(newname);
   if (CHNS->setstatic)
     chan->SetStatic();
   if (sc)
     return chan->Modify(sc, items, itemlist);
   return SCRIPT_OK;
}

int Channel2::Write ( FILE * f ) {
   FUNCTION;
   ChannelSettings * chan_sets = &settings;
   if (!Static() && (fprintf(f, "channel add %s\n", key) == EOF))
      return 1;
   while ((chan_sets = chan_sets->Next()))
     if (chan_sets->Write(f) == EOF)
       return 1;
   return 0;
}

ChannelSettings * Channel2::FindSettings (ChannelSettings * cst) {
   FUNCTION;
   ChannelSettings * chansets = &settings;
   
   while ((chansets = chansets->Next())) 
     if (chansets->type == cst)
       return chansets;
   return NULL;
}
