/*
 * ModuleEntryChannels.C 
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryChannels.h"
#include "IntegerVariable.h"
#include "StringVariable.h"
#include "Channel2.h"
#include "HelpReference.h"
#include "FlagRecord.h"
#include "Socket.h"
#include "UserRecord.h"
#include "ChannelRecord.h"
#include "UserEntryLaston.h"
#include "BanRecord.h"
#include "ScriptContext.h"
#define CLASS "ModuleEntryChannels"

static Version channels_ver;
static struct patches channels_patches[] = {
#include "patch.h"
};

ModuleEntryChannels::ModuleEntryChannels ( ModuleEntry * old, char ** r ) 
: ModuleEntry ( "channels", channels_ver.Set(2,0), channels_patches, old ) {
   FUNCTION;
   Version required;
   
   if (!Depend("eggdrop", required.Set(2,0)))
     *r = "Your eggdrop core version is incompatible";
   else {
      CHNS = this;
      strcpy(chanfile, "chanfile");
      ban_time = 60;
      use_info = 1;
      chan_hack = 0;
      SERVER = 0;
      IRC = 0;
      NEW ( IntegerVariable ( "share_greet", &G->share_greet, 0,this ) );
      NEW ( IntegerVariable ( "use_info", &use_info, 0,this ) );
      NEW ( IntegerVariable ( "ban_time", &ban_time, 0,this ) );
      NEW ( StringVariable ("chanfile", chanfile, 120, VAR_PROTECT, this) );
      G->memtype_chanset = NEW ( MemType ("chanset", (List **)&G->chanset, this) );
      G->chanset = NEW ( Channel2 () );
      CHNS->memtype_settings = NEW (MemType("chansettings", (List **)&CHNS->CHANSET_LIST,this));
      CHNS->CHANSET_LIST = &((Channel2 *)G->chanset)->settings;
      NEW ( HelpReference ( "channels.help", this ) );
      NEW ( HelpReference ( "chaninfo", this ) );
      setstatic = 1; 
   }
}

ModuleEntryChannels::~ModuleEntryChannels ( void ) {
   FUNCTION;
   DELETE ( G->chanset );
   G->chanset = 0;
}

int ModuleEntryChannels::WriteBans (FILE * f, Socket * idx) {
   FUNCTION;
   Channel *chan = G->chanset;
   BanRecord * b = G->global_bans;
   FlagRecord flags (FlagRecord::GLOBAL | FlagRecord::BOT | FlagRecord::CHANNEL);
   
   if (b->Next() && (fprintf(f, "*b\n")==EOF))
     return 0;
   while ((b = b->Next()))
     if (!b->Write(f))
       return 0;
   while ((chan = chan->Next())) {
      if (idx)
	flags.Set(idx->User(),chan->Key());
      if (!idx || (flags.GlobBot() &&
		   (flags.BotGlobal() || flags.ChanShare()))) {
	 if (fprintf(f, ":%s\n", chan->Key()) == EOF)
	   return 0;
	 b = chan->Bans();
	 while ((b = b->Next())) 
	   if (!b->Write(f))
	     return 0;
      }
   }
   return 1;
}

void ModuleEntryChannels::DelChanrec (UserRecord * u, const char * chname) {
   FUNCTION;
   ChannelRecord *ch, *lst = NULL;
   FlagRecord fr (u);
   
   for (ch = u->ChanRecs(); ch; lst = ch, ch = ch->Next())
     if (!strcasecmp(chname, ch->Key())) {
	u->ChanRecs()->Remove(ch);
	if (!G->noshare && G->shareout && !fr.GlobUnshared())
	  G->shareout(FindChan(chname),"L%s %s\n",u->Handle(),chname);   
	break;
     }
}

ChannelRecord * ModuleEntryChannels::AddChanrec (UserRecord * u, const char * chname) {
   FUNCTION;
   ChannelRecord *ch;
   FlagRecord fr (u);
   
   ch = NEW ( ChannelRecord ( u, chname ) );
   if (!G->noshare && G->shareout && !fr.GlobUnshared())
     G->shareout(FindChan(chname),"l%s %s\n",u->Handle(),chname);
   return ch;
}

void ModuleEntryChannels::GetHandleChaninfo (const char * handle,
					     const char * chname,
					     char * s) {
   FUNCTION;
   UserRecord *u;
   ChannelRecord *ch = 0;
   
   s[0] = 0;
   if ((u = G->userlist->Find(handle))
       && (ch = u->ChanRecs()->Find(chname)) && ch->Info())
     strcpy(s, ch->Info());
}

void ModuleEntryChannels::SetHandleChaninfo (UserRecord * bu, const char * handle,
					     const char * chname, const char * info) {
   FUNCTION;
   UserRecord *u;
   ChannelRecord *ch = 0;
   int i, j = 0;
   Channel *cst;
   char info2[81];
   
   if ((u = bu->Find(handle))) {
      FlagRecord fr (u);
      
      if (!(ch = u->ChanRecs()->Find(chname)))
	ch = AddChanrec(u, chname);
      if (info) {
	 for (i = 0; info[i] && (j < 80); i++)
	   if ((info[i] >= 32) && (info[i] != 127))
	     info2[j++] = info[i];
      }
      info2[j] = 0;
      ch->SetInfo(info2);
      cst = FindChan(chname);
      if (!G->noshare && (bu == G->userlist) && G->share_greet && G->shareout 
	  && !fr.GlobUnshared() && !fr.GlobBot())
	G->shareout(cst, "C%s %s %s\n", handle, chname, info2);
   }
}

void ModuleEntryChannels::SetHandleLaston (const char * chan, UserRecord * u, time_t n) {
   FUNCTION;
   ChannelRecord *ch;
   char buf[UHOSTLEN + 1];
   
   if (u) {
      strncpy(buf, chan, UHOSTLEN);
      buf[UHOSTLEN] = 0;
      TouchLaston(u, buf, n);
      if ((ch = u->ChanRecs()->Find(chan)))
	ch->SetLaston(n);
   }
}

int ModuleEntryChannels::ExpMem ( void ) {
   FUNCTION;
   return ModuleEntry::ExpMem() + sizeof(ModuleEntryChannels);
}

void ModuleEntryChannels::ReadChannels(void) {
   FUNCTION;
   Channel2 *chan = (Channel2 *)G->chanset, *chan2;
   CHNS->setstatic = 0;
   if (CHNS->chanfile[0]) {
      while ((chan = chan->Next()))
	if (!chan->Static())
	  chan->SetFlagged();
      CHNS->chan_hack = 1; 
      if (G->scripts->ParseFile(CHNS->chanfile) == SCRIPT_ERROR) {
	 FILE * f;
	 /* assume file isn't there & therefore make it */
	 G->LOG_MISC.Put(CHNS->Lang( 0x089, ""));
	 if (!(f = fopen(CHNS->chanfile,"w")))
	   G->LOG_MISC.Put(CHNS->Lang( 0x08A, "s", CHNS->chanfile));
	 else
	   fclose(f);
      }
      CHNS->chan_hack = 0;
      chan2 = (Channel2 *)G->chanset;
      while ((chan = chan2->Next())) {
	 if (chan->Flagged()) {
	    G->LOG_MISC.Put(CHNS->Lang( 0x08B, "s", chan->Key()));
	    DELETE ( chan );
	 } else
	   chan2 = chan;
      }
   }
}
