/*
 * channels.c - part of channels.mod
 * support for channels withing the bot 
 * commands from a user via dcc that cause channel interaction.
 * script commands for channels
 */
/*
 * This file is part of the eggdrop source code
 * copyright (c) 1997 Robey Pointer
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#define MAKING_CHANNELS
#include "DccChat.h"
#include "UserRecord.h"
#include "UserEntryLaston.h"
#include "UserEntryString.h"
#include "BindTable.h"
#include "BanRecord.h"
#include "HelpReference.h"
#include "HookList.h"
#include "Version.h"
#include "ModuleEntryChannels.h"
#include "channels.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "Channel2.h"
#include "ChannelRecord.h"
#include "Command.h"
#include "ChannelSettingsFlood.h"
#include <sys/stat.h>
#ifdef HAVE_IRC_H
#include "irc.h"
#include "ModuleEntryIrc.h"
#include "Channel3.h"
static Version required_irc ( 2, 0 );
#endif
#ifdef HAVE_SERVER_H
#include "server.h"
#include "ModuleEntryServer.h"
static Version required_server ( 2, 0 );
#endif

ModuleEntryChannels * CHNS;
static ModuleEntryEggdrop * G;

#ifdef HAVE_IRC_H
#define LOAD1 1
#else
#define LOAD1 0
#endif
#ifdef HAVE_SERVER_H
#define LOAD2 1
#else
#define LOAD2 0
#endif
#define CLASS ""

FlagRecord user ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
FlagRecord victim ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );

extern "C" {
   char * channels_start (ModuleEntryEggdrop *, ModuleEntry *);
};

/****************************************************************************
 * CHANNEL BANS                                                             *
 ****************************************************************************/

/* check for expired timed-bans */
static void check_expired_bans() {
   FUNCTION;
   BanRecord * u = G->global_bans, * v = u->Next();
   Channel *chan = G->chanset;

   while ((u = v)) {
      v = u->Next();
      if (!u->IsPerm() && (G->now >= u->Expires())) {
	 G->LOG_MISC.Put( CHNS->Lang(0x084, "s", u->Key()));
#ifdef HAVE_IRC_H	   
	 if (CHNS->IRC) {
	    while ((chan = chan->Next()))
	      AddMode(chan, '-', 'b', u->Key());
	    chan = G->chanset;
	 }
#endif
	 DELETE ( u );
      }
   }
   /* check for specific channel-domain bans expiring */
   while ((chan = chan->Next())) {
      u = chan->Bans();
      v = u->Next();
      while ((u = v)) {
	 v = u->Next();
	 if (!u->IsPerm() && (G->now >= u->Expires())) {
	    G->LOG_MISC.Put(CHNS->Lang(0x085,"ss",u->Key(),chan->Key()),chan);
#ifdef HAVE_IRC_H	   
	    if (CHNS->IRC)
	      AddMode(chan, '-', 'b', u->Key());
#endif
	    DELETE ( u );
	 }
      }
   }
}

/****************************************************************************
 * CHANNELS                                                                 *
 ****************************************************************************/

static void write_channels(void){
   FUNCTION;
   FILE *f;
   char s[DIRLEN+1];
   Channel2 *chan = (Channel2 *)G->chanset;
   
   context;
   if (CHNS->chanfile[0]) {
      G->SPrintF(s, DIRLEN, "%s~new", CHNS->chanfile);
      if (!(f = fopen(s, "w"))) 
	G->LOG_MISC.Put(CHNS->Lang( 0x086, ""));
      else {
	 chmod(s, 0600);
	 G->LOG_MISC.Put(CHNS->Lang( 0x087, ""));
	 fprintf(f, "# %s\n", CHNS->Lang( 0x088, "sss", G->origbotname, 
					  G->VersionStr(), ctime(&G->now)));
	 while ((chan = chan->Next())) 
	   if (chan->Write(f)) {
	      G->LOG_MISC.Put(CHNS->Lang( 0x086, ""));
	      fclose(f);
	      return;
	   }
	 if (fflush(f)) {
	    G->LOG_MISC.Put(CHNS->Lang( 0x086, ""));
	    fclose(f);
	 } else {
	    fclose(f);
	    unlink(CHNS->chanfile);
#ifdef RENAME
	    rename(s, CHNS->chanfile);
#else
	    G->MoveFile(s, CHNS->chanfile);
#endif
	 }
      }
   }
}

/****************************************************************************
 * DCC COMMANDS                                                             *
 ****************************************************************************/

static BindReturn cmd_pls_ban (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * chname, * who, s[UHOSTLEN*2+1], s1[UHOSTLEN+1], * p;
   char * par = (char *)args[1], comment[66], * p_expire;
   Channel2 *chan = 0;
   int langfact = 0;
   time_t expire_time = 0, expire_work;
   
   if (!par[0]) 
     idx->Send(CHNS->Lang(0x001, ""));
   else {
      who = G->Split(&par);
      for (p = who; *p; p++)
	if (((*p < 32) || (*p == 127)) &&
	    (*p != 2) && (*p != 22) && (*p != 31))
	  break;
      if (*p) { 
	 idx->Send(CHNS->Lang( 0x002, ""));
	 return EXECUTED;
      }
      chname = ((par[0] == '#') || (par[0] == '&') || (par[0] == '+')) ?
	G->Split(&par) : 0;
      user.Set(idx->User());
      if (chname || !user.GlobMaster()) {
	 if (!chname)
	   chname = idx->ConsoleChan();
	 user.Set(idx->User(),chname);
	 chan = FindChan(chname);
	 if (!chan || !user.AnyValidOp())
	   return NOMATCH;
      }
      if (*par == '%') {
	 p = G->Split(&par);
	 p_expire = p++;
	 while (*(++p)) {
	    switch (*p) {
	      case 'y':
	      case 'Y':
	       *p = 0;
	       if ((expire_work = atoi(p_expire)) > 100)
		 expire_work = 100;
	       expire_time += expire_work * 31536000;
	       break;
	      case 'w':
	      case 'W':
	       *p = 0;
	       if ((expire_work = atoi(p_expire)) > 52)
		 expire_work = 52;
	       expire_time += expire_work * 604800;
	       break;
	      case 'd':
	      case 'D':
	       *p = 0;
	       if ((expire_work = atoi(p_expire)) > 6)
		 expire_work = 6;
	       expire_time += expire_work * 86400;
	       break;
	      case 'h':
	      case 'H':
	       *p = 0;
	       if ((expire_work = atoi(p_expire)) > 23)
		 expire_work = 23;
	       expire_time += expire_time * 3600;
	       break;
	      case 'm':
	      case 'M':
	       *p = 0;
	       if ((expire_work = atoi(p_expire)) > 59)
		 expire_work = 59;
	       expire_time += expire_time * 60;
	       break;
	      default:
	       continue;
	    }
	    p_expire = p + 1;
	 }
      }
      strncpy(comment, par[0] ? par : CHNS->Lang( 0x003, ""), 65);
      comment[65] = 0;
      /* fix missing ! or @ BEFORE checking against myself */
      if (!strchr(who, '!')) {
	 if (!strchr(who, '@')) 
	   G->SPrintF(s, UHOSTLEN, "%s!*@*", who);	/* lame nick ban */
	 else
	   G->SPrintF(s, UHOSTLEN, "*!%s",who);
      } else if (!strchr(who, '@'))
	   G->SPrintF(s, UHOSTLEN, "%s@*", who);   /* brain-dead? */
      else
	strncpyz(s, who, UHOSTLEN);
#ifdef HAVE_SERVER_H
      if (CHNS->SERVER)
	G->SPrintF(s1, UHOSTLEN, "%s!%s", CHNS->SERVER->botname, 
		   CHNS->SERVER->botuserhost);
      else
#endif
	gethostname(s1 + G->SPrintF(s1, UHOSTLEN, "%s!%s@", G->origbotname,
				   G->botuser), UHOSTLEN);
      if (G->WildMatch(s, s1)) {
	 idx->Send(CHNS->Lang( 0x004, ""));
	 G->LOG_MISC.Put(CHNS->Lang( 0x005, "ss", idx->Nick(), s));
      } else { 
	 if (strlen(s) > 70) {
	    s[69] = '*';
	    s[70] = 0;
	 }
	 if (comment[0] == '*') {
	    p = comment + 1;
	    langfact = 0x0A0;
	 } else
	   p = comment;
	   
	 /* irc can't understand bans longer than that */
	 if (chan) {
	    chan->AddBan(s, idx->Nick(), comment, 
			 expire_time ? G->now + expire_time : 0, 0);
	    G->LOG_CMDS.Put(CHNS->Lang( 0x006 + langfact, "sssss", idx->Nick(),
				     idx->ConsoleChan(), s, chan->Key(), p),
			     chan);
	    idx->Send(CHNS->Lang( 0x007+langfact,"sss",chan->Key(),s,p));
#ifdef HAVE_IRC_H	   
	    if (CHNS->IRC)
	      AddMode(chan, '+', 'b', s);
#endif
	 } else {
	    ((Channel2 *)G->chanset)->AddBan(s, idx->Nick(), comment, expire_time ? 
					     G->now + expire_time : 0, 0);
	    G->LOG_CMDS.Put(CHNS->Lang( 0x008 + langfact, "sss",idx->Nick(),s,p));
	    idx->Send(CHNS->Lang( 0x009 + langfact, "ss", s, p));
#ifdef HAVE_IRC_H	   
	    if (CHNS->IRC) {
	       chan = (Channel2 *)G->chanset;
	       while ((chan = chan->Next()))
		 AddMode(chan, '+', 'b', s);
	    }
#endif
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_mns_ban (void ** args, const char *) {
   FUNCTION;
   int i = 0, j;
   DccChat * idx = (DccChat *)args[0];
   Channel2 *chan = 0;
   char s[UHOSTLEN + 1], *ban, *chname, *par = (char *)args[1];
   
   if (!par[0]) 
     idx->Send(CHNS->Lang( 0x00A, ""));
   else {
      ban = G->Split(&par);
      chname = ((par[0] == '#') || (par[0] == '&') || (par[0] == '+')) ?
	G->Split(&par) : idx->ConsoleChan();
      user.Set(idx->User());
      if (chname || !user.GlobMaster()) {
	 if (!chname)
	   chname = idx->ConsoleChan();
	 user.Set(idx->User(),chname);
	 if (!user.AnyValidOp())
	   return EXECUTED;
      }
      strncpyz(s,ban,UHOSTLEN);
      if ((i = ((Channel2 *)G->chanset)->DelBan(s,user.GlobMaster())) > 0) {
	 G->LOG_CMDS.Put(CHNS->Lang( 0x00B, "ss", idx->Nick(), s));
	 idx->Send(CHNS->Lang( 0x00C, "s", s));
#ifdef HAVE_IRC_H	   
	 if (CHNS->IRC) {
	    chan = (Channel2 *)G->chanset;
	    while ((chan = chan->Next()))
	      AddMode(chan, '-', 'b', s);
	 }
#endif
	 return EXECUTED;
      } 
      /* channel-specific ban? */
      if (chname)
	chan = FindChan(chname);
      if (chan) {
	 if (atoi(ban) > 0) {
	    sprintf(s, "%d", -i);
	    ban = s;
	 }
	 if ((j = chan->DelBan(ban, 1)) > 0) {
	    G->LOG_CMDS.Put(CHNS->Lang( 0x00D, "sss", idx->Nick(),
				     chan->Key(), ban),chan);
	    idx->Send(CHNS->Lang( 0x00E, "ss", chan->Key(), ban));
#ifdef HAVE_IRC_H	   
	    if (CHNS->IRC)
	      AddMode(chan, '-', 'b', ban);
#endif
	    return EXECUTED;
	 } 
#ifdef HAVE_IRC_H	 
	 strncpyz(s, ban, UHOSTLEN);
	 if (CHNS->IRC && NukeChanBan(chan, s, -j)) {
	    G->LOG_CMDS.Put(CHNS->Lang( 0x00F, "sss", idx->Nick(),chan->Key(),
				     ban), chan);
	    idx->Send(CHNS->Lang( 0x010, "ss", ban, chan->Key()));
	    return EXECUTED;
	 }
#endif
      }
   }
   idx->Send(CHNS->Lang(0x011,""));
   return EXECUTED;
}

static BindReturn cmd_bans (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int show_inact = 0, k = 1, l, which = 0;
   char * match = (char *)args[1], *chname, s[UHOSTLEN+1], fill[UHOSTLEN+1];
   Channel2 *chan = NULL;
   BanRecord * b;
   
   context;
   G->LOG_CMDS.Put(CHNS->Lang( 0x012, "ss", idx->Nick(), match));
   /* was channel given? */
   if (match[0]) {
      chname = G->Split(&match);
      if ((chname[0] == '#') || (chname[0] == '+') || (chname[0] == '&')) {
	 if (!(chan = FindChan(chname))) {
	    idx->Send(CHNS->Lang( 0x100, "s", chname));
	    return EXECUTED;
	 }
      } else
	match = chname;
   }
   if (chan || (chan = FindChan(idx->ConsoleChan()))) {
      if (!strcasecmp(match, "all")) {
	 show_inact = 1;
	 match = "";
      }
      for (which = 0; which < 2; which++) {
	 context;
	 idx->Send(CHNS->Lang( show_inact + (which ? 0x015 : 0x013),
			   "ccs", '!',  '*', chan->Key()));
	 for (b = which ? chan->Bans() : G->global_bans; (b = b->Next()); k++)
	   if (!match[0] || G->WildMatch(match, b->Key()) ||
	       G->WildMatch(match, b->Reason()) ||
	       G->WildMatch(match, b->Setter())) {
	      if (b->IsPerm())
		l = G->StrCpy(s, CHNS->Lang( 0x019, ""));
	      else if (b->Expires()) {
		 G->Days(b->Expires(), G->now, fill);
		 l = G->StrCpy(s, CHNS->Lang( 0x01A, "s", fill));
	      } else 
		l = 0;
	      if (b->IsSticky())
		l += G->SPrintF(s + l, UHOSTLEN - l, " %s", 
				CHNS->Lang( 0x01B, ""));
#ifdef HAVE_IRC_H	   
	      if (CHNS->IRC && IsBanned(chan, b->Key()))
		idx->Send(CHNS->Lang( !match[0] ? 0x01C : 0x01D, "3ss",
				  k, b->Key(), s));
	      else
#endif
		if (show_inact) 
		idx->Send(CHNS->Lang( !match[0] ?  0x01E : 0x01F, "3ssc",
				  k, b->Key(), s, '!'));
	      else
		continue;
	      idx->Send(CHNS->Lang( 0x020, "ss", b->Setter(), b->Reason()));
	      if (b->Added()) {
		 G->DaysAgo(G->now, b->Added(), fill);
		 l = G->StrCpy(s, CHNS->Lang( 0x017, "s", fill));
		 if (b->Added() < b->LastActive()) {
		    G->DaysAgo(G->now, b->LastActive(), fill);
		    G->StrCpy(s + l, CHNS->Lang( 0x018, "s", fill));
		 }
		 idx->Send(CHNS->Lang( 0x021, "s", s));
	      }
	   }
      }
      context;
#ifdef HAVE_IRC_H	   
      if (chan->Active() && CHNS->IRC)
	k = ShowChanBan(chan, match, k, idx);
#endif
      if (k == 1)
	idx->Send(CHNS->Lang( 0x026, ""));
      if (!show_inact && !match[0])
	idx->Send(CHNS->Lang( 0x027, ""));
   } else
     idx->Send(CHNS->Lang( 0x101, ""));
   return EXECUTED;
}

static BindReturn cmd_stick_yn (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int i, j, yn = (mask && (mask[0] == 's')) ? 1 : 0, langbase=yn?0x028:0x02B;
   Channel2 *chan;
   char s[UHOSTLEN+1], *par = (char *)args[1];
   FlagRecord flags (idx->User());
   
   strncpyz(s, par, UHOSTLEN);
   if (!par[0]) 
     idx->Send(CHNS->Lang( langbase, ""));
   else if ((i = ((Channel2 *)G->chanset)->SetStickyBan(s, flags.GlobMaster() ? yn : -1)) > 0) {
      G->LOG_CMDS.Put(CHNS->Lang( langbase + 1, "ss", idx->Nick(), s));
      idx->Send(CHNS->Lang( langbase + 2, "s", s));
   } else if (!(chan = FindChan(idx->ConsoleChan())))
	idx->Send(CHNS->Lang( 0x101, ""));
   else {
      if (i)
	sprintf(s, "%d", -i);
      if ((j = chan->SetStickyBan(s, yn)) > 0) {
	 G->LOG_CMDS.Put(CHNS->Lang( langbase + 1, "ss", idx->Nick(), s));
	 idx->Send(CHNS->Lang( langbase + 2, "s", s));
      } else
	idx->Send(CHNS->Lang( 0x011, ""));
   }
   return EXECUTED;
}

static BindReturn cmd_info (void ** args, const char *) {
   FUNCTION;
   char s[512], *chname, *par = (char *)args[1];
   const char * s1;
   int locked = 0, langbase = 0x2F;
   DccChat * idx = (DccChat *)args[0];
   UserEntryString * uei;
   FlagRecord flags(idx->User());
   
   if (!CHNS->use_info) 
     idx->Send(CHNS->Lang( 0x02E, ""));
   else {
      uei = (UserEntryString *)idx->User()->GetEntry(G->entrytype_info);
      s1 = uei->Value ();
      if (s1 && s1[0] == '@')
	locked = 1;
      if ((par[0] == '#') || (par[0] == '+') || (par[0] == '&')) {
	 chname = G->Split(&par);
	 if (!FindChan(chname)) {
	    idx->Send(CHNS->Lang( 0x100, ""));
	    return EXECUTED;
	 }
	 CHNS->GetHandleChaninfo(idx->Nick(), chname, s);
	 if (s[0] == '@')
	   locked = 1;
	 s1 = s;
	 langbase = 0x035;
      } else
	chname = "*";
      if (!par[0]) {
	 if (s1 && s1[0] == '@')
	   s1++;
	 if (s1 && s1[0]) {
	    idx->Send(CHNS->Lang( langbase, chname, s1));
	    idx->Send(CHNS->Lang( langbase + 1, chname));
	 } else 
	   idx->Send(CHNS->Lang( 0x03B, ""));
	 G->LOG_CMDS.Put(CHNS->Lang( 0x03C, "ss", idx->Nick(), chname),
			  FindChan(chname));
      } else if (locked && !flags.GlobMaster())
	   idx->Send(CHNS->Lang( 0x03D, ""));
      else if (!strcasecmp(par, "none")) {
	 if (chname[0] == '*')
	   uei->Set(0);
	 else 
	   CHNS->SetHandleChaninfo(G->userlist, idx->Nick(), chname, "");
	 idx->Send(CHNS->Lang( langbase + 2, "s", chname));
	 G->LOG_CMDS.Put(CHNS->Lang( langbase + 3, "ss", idx->Nick(), chname),
			  FindChan(chname));
      } else {
	 if (par[0] == '@')
	   par++;
	 if (chname[0] == '*') 
	   uei->Set(par);
	 else
	   CHNS->SetHandleChaninfo(G->userlist, idx->Nick(), chname, par);
	 idx->Send(CHNS->Lang( langbase + 4, "ss", chname,par));
	 G->LOG_CMDS.Put(CHNS->Lang( langbase + 5, "ss", idx->Nick(), chname),
			  FindChan(chname));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_chinfo (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int langbase = 0x046;
   char * handle, * chname, *par = (char *)args[1];
   UserRecord * u1;
     
   if (!CHNS->use_info) 
     idx->Send(CHNS->Lang( 0x02E, ""));
   else if (!(handle = G->Split(&par))[0])
     idx->Send(CHNS->Lang( 0x03E, ""));
   else if (!(u1 = G->userlist->Find(handle)))
     idx->Send(G->Lang(0x100, ""));
   else {
      chname = ((par[0] == '#') || (par[0] == '+') || (par[0] == '&')) ?
	G->Split(&par) : "*";
      user.Set(idx->User(),chname);
      victim.Set(u1,chname);
      if (victim.GlobBot() && !user.GlobOwner())
	idx->Send(CHNS->Lang( 0x03F, ""));
      else if (victim.GlobOwner() && !user.GlobOwner())
	idx->Send(CHNS->Lang( 0x040, ""));
      else {
	 if ((chname[0] != '*') && victim.AnyOwner() && !user.AnyOwner()) {
	    idx->Send(CHNS->Lang( 0x041, ""));
	    return EXECUTED;
	 }
	 if (!strcasecmp(par, "none"))
	   par[0] = 0;
	 if (chname[0] != '*') {
	    CHNS->SetHandleChaninfo(G->userlist, handle, chname, par);
	    langbase = 0x042;
	 } else 
	   ((UserEntryString *)u1->GetEntry(G->entrytype_info))->Set(par);
	 G->LOG_CMDS.Put(CHNS->Lang( langbase, "ssss", idx->Nick(),
				  handle, chname, par), FindChan(chname));
	 if (par[0] == '@') {
	    langbase++;
	    par++;
	 } else if (par[0])
	   langbase += 2;
	 else
	   langbase += 3;
	 idx->Send(CHNS->Lang( langbase, "sss", handle,chname, par));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_chrec (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char *nick, *par = (char *)args[1];
   Channel *chan;
   UserRecord *u1;
   
   context;
   if (!par[0]) 
     idx->Send(CHNS->Lang( 0x04B, "c", mask[0]));
   else if (!(u1 = G->userlist->Find(nick = G->Split(&par))))
     idx->Send(G->Lang( 0x100, ""));
   else if (!(chan = FindChan(par[0] ? G->Split(&par) : 
			      idx->ConsoleChan())))
     idx->Send(CHNS->Lang( 0x0100, ""));
   else {
      user.Set(idx->User(), chan->Key());
      if (!user.AnyMaster())
	idx->Send(CHNS->Lang( 0x102, "s", chan->Key()));
      else if ((mask[0] == '+') && u1->ChanRecs()->Find(chan->Key()))
	idx->Send(CHNS->Lang( 0x04C, "ss", nick, chan->Key()));
      else if ((mask[0] == '-') && !u1->ChanRecs()->Find(chan->Key()))
	idx->Send(CHNS->Lang( 0x04D, "ss", nick, chan->Key()));
      else {
	 G->LOG_CMDS.Put(CHNS->Lang( 0x04E, "scss", idx->Nick(),
				       mask[0], nick, chan->Key()));
	 if (mask[0] == '+') 
	   CHNS->AddChanrec(u1,chan->Key());
	 else
	   CHNS->DelChanrec(u1, chan->Key());
	 idx->Send(CHNS->Lang( (mask[0] == '+') ? 0x04F :
					      0x050, "ss", chan->Key(), nick));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_pls_chan (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char *chname, *par = (char *)args[1];
   
   if (!par[0])
     idx->Send(CHNS->Lang( 0x051, ""));
   else if (FindChan(chname = G->Split(&par)))
     idx->Send(CHNS->Lang( 0x052, ""));
   else if (((Channel2 *)G->chanset)->Add(0, chname, 0, 0) == SCRIPT_ERROR)
     idx->Send(CHNS->Lang( 0x053, ""));
   else
     G->LOG_CMDS.Put(CHNS->Lang( 0x054, "ss", idx->Nick(), chname));
   return EXECUTED;
}

static BindReturn cmd_mns_chan (void ** args, const char *) {
   FUNCTION;
   char *chname, *par = (char *)args[1];
   Channel2 *chan;
   DccChat * idx = (DccChat *)args[0];
   
   if (!par[0]) 
     idx->Send(CHNS->Lang( 0x055, ""));
   else if (!(chan = FindChan(chname = G->Split(&par))))
     idx->Send(CHNS->Lang( 0x056, ""));
   else if (chan->Static())
     idx->Send(CHNS->Lang( 0x057, "s", chname));
   else {
      DELETE ( chan );
      idx->Send(CHNS->Lang( 0x058, "s", chname));
      idx->Send(CHNS->Lang( 0x059, ""));
      G->LOG_CMDS.Put(CHNS->Lang( 0x05A, "ss", idx->Nick(), chname));
      for (idx = (DccChat *)G->socklist; (idx = idx->Next()); )
	if (!strcasecmp(idx->ConsoleChan(), chname)) {
	   idx->Send(CHNS->Lang( 0x05B, "ss", chname, "*"));
	   strcpy(idx->ConsoleChan(), "*");
	}
   }
   return EXECUTED;
}

static BindReturn cmd_chaninfo (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char *chname, *par = (char *)args[1];
   Channel2 *chan;
   
   if (!par[0] && ((chname = idx->ConsoleChan())[0] == '*')) 
     idx->Send(CHNS->Lang( 0x101, ""));
   else {
      if (par[0]) {
	 chname = G->Split(&par);
	 user.Set(idx->User(),chname);
	 if (!user.AnyMaster()) {
	    idx->Send(CHNS->Lang( 0x102, "s", chname));
	    return EXECUTED;
	 }
      }
      if (!(chan = FindChan(chname))) 
	idx->Send(CHNS->Lang( 0x100, ""));
      else {
	 ChannelSettings * chan_sets = chan->Settings ();
	 
	 G->LOG_CMDS.Put(CHNS->Lang( 0x05C, "ss",idx->Nick(), chname));
	 idx->Send(CHNS->Lang( chan->Static() ? 0x05D : 0x05E,"s", chname));
	 while ((chan_sets = chan_sets->Next()))
	   chan_sets->View(idx);
      }
   }
   return EXECUTED;
}

static BindReturn cmd_chanset (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int ret;
   char *chname = NULL, answers[512], *arg, * par = (char *)args[1];
   Channel2 *chan = NULL;
   ChannelSettings * chan_sets;
   
   if (!par[0]) 
     idx->Send(CHNS->Lang( 0x05F, ""));
   else {
      if ((par[0] == '&') || (par[0] == '#') || (par[0] == '+')) {
	 chname = G->Split(&par);
	 user.Set(idx->User(),chname);
	 if (!user.AnyOwner()) {
	    idx->Send(CHNS->Lang( 0x102, "s", chname));
	    return EXECUTED;
	 }  else if (!(chan = FindChan(chname)) && (chname[0] != '+')) {
	    idx->Send(CHNS->Lang( 0x100, ""));
	    return EXECUTED;
	 } 
	 if (!chan) {
	    if (par[0])
	      *--par = ' ';
	    par = chname;
	 }
      } 
      if (!chan && !(chan = FindChan(chname = idx->ConsoleChan())))
	idx->Send(CHNS->Lang( 0x101, ""));
      else {
	 arg = G->Split(&par);
	 answers[0] = 0;
	 while (arg[0]) {
	    ret = 0;
	    chan_sets = chan->Settings();
	    while ((chan_sets =chan_sets->Next()) && !ret) 
	      ret = chan_sets->Set(arg, par, idx);
	    switch (ret) {
	      case -1:
	      case 0: /* invalid setting */
	       idx->Send(CHNS->Lang( 0x060, "sss", arg, par,chname));
	       arg[0] = 0;
	       break;
	      case 1: /* single setting, move on */
	       strcat(answers, arg);
	       strcat(answers, " ");
	       arg = G->Split(&par);
	       break;
	      case 2: /* whole argument used */
	       strcat(answers, arg);
	       strcat(answers, " { ");
	       strcat(answers, par);
	       strcat(answers, " }");
	       arg[0] = 0;
	       break;
	    }
	 }
	 if (answers[0]) {
	    G->LOG_CMDS.Put(CHNS->Lang( 0x061, "sss", idx->Nick(), chname, answers));
	    idx->Send(CHNS->Lang( 0x062, "ss", answers, chname));
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_chansave (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   
   if (!CHNS->chanfile[0])
     idx->Send(CHNS->Lang( 0x063, ""));
   else {
      G->LOG_CMDS.Put(CHNS->Lang( 0x064, "s", idx->Nick()));
      idx->Send(CHNS->Lang( 0x065, ""));
      write_channels();
   }
   return EXECUTED;
}

static BindReturn cmd_chanload (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   
   if (!CHNS->chanfile[0])
     idx->Send(CHNS->Lang( 0x063, ""));
   else {
      G->LOG_CMDS.Put(CHNS->Lang( 0x066, "s", idx->Nick()));
      idx->Send(CHNS->Lang( 0x067, ""));
      CHNS->setstatic = 0;
      CHNS->ReadChannels();
   }
   return EXECUTED;
}

#define COMMANDS 15
static command_binds channels_commands[COMMANDS]={
  { "+ban", "o|o", cmd_pls_ban, NULL },
  { "+chan", "n", cmd_pls_chan, NULL },
  { "+chrec", "m|m", cmd_chrec, NULL },
  { "-ban", "o|o", cmd_mns_ban, NULL },
  { "-chan", "n", cmd_mns_chan, NULL },
  { "-chrec", "m|m", cmd_chrec, NULL },
  { "bans", "o|o", cmd_bans, NULL },
  { "chaninfo", "m|m", cmd_chaninfo, NULL },
  { "chanload", "n|n", cmd_chanload, NULL },
  { "chanset", "n|n", cmd_chanset, NULL },
  { "chansave", "n|n", cmd_chansave, NULL },
  { "chinfo", "m|m", cmd_chinfo, NULL },
  { "info", "", cmd_info, NULL },
  { "stick", "o|o", cmd_stick_yn, NULL },
  { "unstick", "o|o", cmd_stick_yn, NULL },
};

/****************************************************************************
 * SCRIPT COMMANDS                                                          *
 ****************************************************************************/

static int script_killban (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel2 *chan = (Channel2 *)G->chanset;
   char buf[UHOSTLEN + 1];
   
   BADARGS(2, 3, CHNS->Lang( 0x068, ""));
   if ((argc == 3) && !(chan = FindChan(argv[2]->AsString()))) {
      sc->SetString(CHNS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   strncpy(buf, argv[1]->AsString(),UHOSTLEN);
   buf[UHOSTLEN] = 0;
   if (chan->DelBan(buf, 1) > 0) {
#ifdef HAVE_IRC_H	   
      if (CHNS->IRC) {
	 if (chan == G->chanset)
	   while ((chan = chan->Next()))
	     AddMode(chan, '-', 'b', buf);
	 else
 	   AddMode(chan, '-', 'b', buf);
      }
#endif
      sc->SetString("1");
   } else
     sc->SetString("0");
   return SCRIPT_OK;
}

static int script_isban (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel2 *chan = (Channel2 *)G->chanset;
   char buf[UHOSTLEN + 1];
   
   BADARGS(2, 3, CHNS->Lang( 0x068, ""));
   if ((argc == 3) && !(chan = FindChan(argv[2]->AsString()))) {
      sc->SetString(CHNS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   strncpyz(buf, argv[1]->AsString(), UHOSTLEN);
   sc->SetString(G->ToBase10(chan->EqualsBan(buf)));
   return SCRIPT_OK;
}

static int script_matchban (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel2 *chan = (Channel2 *)G->chanset;
   
   BADARGS(2, 3, CHNS->Lang( 0x069, ""));
   if ((argc == 3) && !(chan = FindChan(argv[2]->AsString()))) {
      sc->SetString(CHNS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(chan->MatchBan(argv[1]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_newban (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   time_t expire_time;
   Channel2 *chan = (Channel2 *)G->chanset;
   char ban[161], cmt[66], from[HANDLEN+1];
   int sticky = 0, argbase = 1;
   const char * ch = argv[1]->AsString();
   
   BADARGS(4, 7, CHNS->Lang( 0x06A, ""));
   if ((ch[0] == '#') || (ch[0] == '+')) {
      BADARGS(5, 7, CHNS->Lang( 0x06A, ""));
      if (!(chan = FindChan(ch))) {
	 sc->SetString(CHNS->Lang( 0x103, "s", ch));
	 return SCRIPT_ERROR;
      } 
      argbase++;
   }
   if (argc == argbase + 5) {
      if (!strcasecmp(argv[argbase + 4]->AsString(), "none"));
      else if (!strcasecmp(argv[argbase + 4]->AsString(), "sticky"))
	sticky = 1;
      else {
	 sc->SetString( CHNS->Lang( 0x06B, "s", argv[argbase + 4]->AsString()));
	 return SCRIPT_ERROR;
      }
   }
   strncpy(ban, argv[argbase]->AsString(), 160);
   ban[160] = 0;
   strncpy(from, argv[argbase + 1]->AsString(), HANDLEN);
   from[HANDLEN] = 0;
   strncpy(cmt, argv[argbase + 2]->AsString(), 65);
   cmt[65] = 0;
   if (argc == argbase + 3)
     expire_time = G->now + (60 * CHNS->ban_time);
   else if (!(expire_time = argv[argbase + 3]->AsInteger()))
     expire_time = 0L;
   else
     expire_time = G->now + (expire_time * 60);
   if (chan->AddBan(ban, from, cmt, expire_time,sticky)) {
#ifdef HAVE_IRC_H	   
      if (CHNS->IRC) {
	 if (chan != G->chanset) 
	   AddMode(chan, '+', 'b', ban);
	 else
	   while ((chan = chan->Next()))
	     AddMode(chan, '+', 'b', ban);
      }
#endif 
  }
   return SCRIPT_OK;
}

static int script_channel (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel2 *chan = (Channel2 *)G->chanset;;
   const char * cmd, * ch;
   ChannelSettings * chansets;
   
   BADARGS(2, 999, CHNS->Lang( 0x077, ""));
   cmd = argv[1]->AsString();
   ch = argv[2]->AsString();
   if (!strcmp(cmd, "add")) {
      BADARGS(3, 999, CHNS->Lang( 0x078, ""));
      return chan->Add(sc, ch, argc - 3, argv + 3);
   } else if (!strcmp(cmd, "set")) {
      BADARGS(3, 999, CHNS->Lang( 0x079, ""));
      if (!(chan = FindChan(ch))) {
	 if (CHNS->chan_hack == 1)
	   return SCRIPT_OK; /* ignore channel settings for a static
			   * channel which has been removed from the config*/
	 sc->SetString(CHNS->Lang( 0x103, "s", ch));
	 return SCRIPT_ERROR;
      }
      return chan->Modify(sc, argc - 3, argv + 3);
   } else if (!strcmp(cmd, "info")) {
      BADARGS(3, 3, CHNS->Lang( 0x07A, ""));
      if (!(chan = FindChan(ch))) {
	 sc->SetString(CHNS->Lang( 0x103, "s", ch));
	 return SCRIPT_ERROR;
      }
      chansets = chan->Settings();
      while ((chansets = chansets->Next()))
	chansets->Get(sc);
      return SCRIPT_OK;
   } else if (!strcmp(cmd, "remove")) {
      BADARGS(3, 3, CHNS->Lang( 0x07B, ""));
      if (!(chan = FindChan(ch))) {
	 sc->SetString(CHNS->Lang( 0x103, "s", ch));
	 return SCRIPT_ERROR;
      }
      DELETE ( chan );
      return SCRIPT_OK;
   }
   BADARGS(3, 3, CHNS->Lang( 0x07C, ""));
   return SCRIPT_ERROR;
}

static int script_banlist (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel2 *chan;
   BanRecord * u;
   char ts[21], ts1[21];
   const char *list[6];

   BADARGS(1, 2, CHNS->Lang( 0x07E, ""));
   if (argc == 2) {
      if (!(chan = FindChan(argv[1]->AsString()))) {
	 sc->SetString(CHNS->Lang( 0x103, "s", argv[1]->AsString()));
	 return SCRIPT_ERROR;
      }
      u = chan->Bans();
   } else
      u = G->global_bans;
   while ((u = u->Next())) {
      list[0] = u->Key();
      list[1] = u->Reason();
      strcpy(ts,G->ToBase10(u->Expires()));
      list[2] = ts;
      strcpy(ts1,G->ToBase10(u->Added()));
      list[3] = ts1;
      list[4] = G->ToBase10(u->LastActive());
      list[5] = u->Setter();
      sc->AppendArray(6, list);
   }
   return SCRIPT_OK;
}

static int script_channels (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   Channel *chan = G->chanset;
   const char * p;
   
   BADARGS(1, 1, "");
   while ((chan = chan->Next()))
     sc->AppendArray(1,&(p = chan->Key()));
   return SCRIPT_OK;
}

static int script_savechannels (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   if (!CHNS->chanfile[0]) {
      sc->SetString(CHNS->Lang( 0x07F, ""));
      return SCRIPT_ERROR;
   }
   write_channels();
   return SCRIPT_OK;
}

static int script_loadchannels (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   context;
   BADARGS(1, 1, "");
   if (!CHNS->chanfile[0]) {
      sc->SetString(CHNS->Lang( 0x07F, ""));
      return SCRIPT_ERROR;
   }
   CHNS->setstatic = 0;
   CHNS->ReadChannels();
   return SCRIPT_OK;
}

static int script_validchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BADARGS(2, 2, CHNS->Lang( 0x080, ""));
   sc->SetString(FindChan(argv[1]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_isdynamic (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel2 *chan;
   
   BADARGS(2, 2, CHNS->Lang( 0x080, ""));
   sc->SetString(((chan = FindChan(argv[1]->AsString()))
		  && !chan->Static()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_getchaninfo (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char s[161];
   UserRecord * u;
   FlagRecord fr;
   
   BADARGS(3, 3, CHNS->Lang( 0x081, ""));
   if ((u = G->userlist->Find(argv[1]->AsString())) && !fr.Set(u)->GlobBot()) {
      CHNS->GetHandleChaninfo(argv[1]->AsString(), argv[2]->AsString(), s);
      sc->SetString(s);
   }
   return SCRIPT_OK;
}

static int script_setchaninfo (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BADARGS(4, 4, CHNS->Lang( 0x082, ""));
   CHNS->SetHandleChaninfo(G->userlist, argv[1]->AsString(), 
			   argv[2]->AsString(), argv[3]->AsString());
   return SCRIPT_OK;
}

static int script_setlaston (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   time_t t = G->now;
   UserRecord * u;
   
   BADARGS(2, 4, CHNS->Lang( 0x083, ""));
   if (!(u = G->userlist->Find(argv[1]->AsString()))) {
      sc->SetString(G->Lang( 0x102, "s" ,argv[1]->AsString()));
      return SCRIPT_ERROR;
   } 
   if (argc == 4)
       t = (time_t) argv[3]->AsInteger();
   else if ((argc == 3) && !strchr("#&+",argv[2]->AsString()[0]))
       t = (time_t) argv[2]->AsInteger();
   if (argc == 2 || ((argc == 3) && !strchr("#&+",argv[2]->AsString()[0])))
       CHNS->SetHandleLaston("*", u, t);
   else
       CHNS->SetHandleLaston(argv[2]->AsString(), u, t);
   return SCRIPT_OK;
}

static int script_addchanrec (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord *u;
   
   context;
   BADARGS(3, 3, CHNS->Lang( 0x081, ""));
   if (!(u = G->userlist->Find(argv[1]->AsString())) 
       || !FindChan(argv[2]->AsString()) || u->ChanRecs()->Find(argv[2]->AsString()))
     sc->SetString("0");
   else {
      CHNS->AddChanrec(u, argv[2]->AsString());
      sc->SetString("1");
   }
   return SCRIPT_OK;
}

static int script_delchanrec (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   UserRecord *u;
   
   context;
   BADARGS(3, 3, CHNS->Lang( 0x081, ""));
   if (!(u = G->userlist->Find(argv[1]->AsString())) 
       && !FindChan(argv[2]->AsString()) 
       && !u->ChanRecs()->Find(argv[2]->AsString())) 
     sc->SetString("0");
   else {
      CHNS->DelChanrec(u, argv[2]->AsString());
      sc->SetString("1");
   }
   return SCRIPT_OK;
}

/****************************************************************************
 * MISC MODULE FUNCTIONS                                                    *
 ****************************************************************************/

static void channels_writeuserfile (void) {
   FUNCTION;
   char s[DIRLEN+1];
   FILE * f;
   
   G->SPrintF(s, DIRLEN, "%s~new", G->userfile);
   f = fopen(s, "a");
   CHNS->WriteBans(f, NULL);
   fclose(f);
   write_channels();
}

/* bind this to chon and *if* the users console channel == ***
 * then set it to a specific Channel2 */
static BindReturn channels_chon (void ** args, const char *) {
   FUNCTION;
   FlagRecord flags (FlagRecord::CHANNEL | FlagRecord::GLOBAL |
		     FlagRecord::ANYWHERE ), simp ( FlagRecord::CHANNEL );
   DccChat * idx = (DccChat *)args[0];
   Channel * chan = 0;
   
   if (idx->TypeFlags() & DCT_CHAT) 
     if (!FindChan(idx->ConsoleChan())
	 && ((idx->ConsoleChan()[0] != '*')
	     || (idx->ConsoleChan()[1] != 0))) {
	flags.Set(idx->User());
	if (!flags.GlobOp()) {
	   for (chan = G->chanset->Next(); chan; chan = chan->Next()) {
	      simp.Set(idx->User(),chan->Key());
	      if (flags.ChanOwner()) {
		 if (simp.ChanOwner())
		   break;
	      } else if (flags.ChanMaster()) {
		 if (simp.ChanMaster())
		   break;
	      } else if (simp.ChanOp())
		   break;
	   }
	}
	if (!chan)
	  chan = G->chanset;
	strcpy(idx->ConsoleChan(), chan ? chan->Key() : "*");
     }
   return EXECUTED;
}

static void channels_prerehash () {
   FUNCTION;
   Channel2 *chan = (Channel2 *)G->chanset;

   while ((chan = chan->Next())) {
     chan->SetFlagged();
     if (chan->Static())
	chan->UnsetStatic();
   }
   CHNS->setstatic = 1;
}

static void channels_rehash () {
   FUNCTION;
   CHNS->setstatic = 0;
   CHNS->ReadChannels();
}

static command_binds my_chon[] = {
     { "*", "", channels_chon, "channels:chon" },
};

/* there are cross calls to these module, so we need to know if they
 * are loaded / unloaded */
#ifdef HAVE_SERVER_H
static BindReturn channels_load_server (void ** args, const char *) {
   FUNCTION;
   CHNS->SERVER = (ModuleEntryServer *)args[0];
   if (*CHNS->SERVER->Vers() <= required_server)
     CHNS->SERVER = NULL;
   return EXECUTED;
};


static BindReturn channels_unld_server (void **, const char *) {
   FUNCTION;
   CHNS->SERVER = NULL;
   return EXECUTED;
};
#endif

#ifdef HAVE_IRC_H
static BindReturn channels_load_irc (void ** args, const char *) {
   FUNCTION;
   CHNS->IRC = (ModuleEntryIrc *)args[0];
   if (!(required_irc <= *CHNS->IRC->Vers()))
     CHNS->IRC = NULL;
   return EXECUTED;
};

static BindReturn channels_unld_irc (void **, const char *) {
   FUNCTION;
   CHNS->IRC = NULL;
   return EXECUTED;
};
#endif

static BindReturn channels_unld_any (void ** args, const char *) {
   FUNCTION;
   ((Channel2 *)G->chanset)->Settings()->Cleanup((ModuleEntry *)args[0]);
   return EXECUTED;
};

#if defined(HAVE_IRC_H) || defined(HAVE_SERVER_H)
static command_binds my_load[] = {
#ifdef HAVE_SERVER_H
   { "server", "", channels_load_server, "channels:server" },     
#endif
#ifdef HAVE_IRC_H
   { "irc", "", channels_load_irc, "channels:irc" },
#endif
};
#endif

static command_binds my_unld[] = {
   { "*", "", channels_unld_any, "channels:any" },
#ifdef HAVE_SERVER_H
   { "server", "", channels_unld_server, "channels:server" },
#endif
#ifdef HAVE_IRC_H
   { "irc", "", channels_unld_irc, "channels:irc" },
#endif
};

char *channels_start (ModuleEntryEggdrop * F, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   G = F;
   context;
   CHNS = NEW ( ModuleEntryChannels ( me, &r ) );
   if (!r) {
#ifdef HAVE_SERVER_H
      CHNS->SERVER = (ModuleEntryServer *)G->module_list->Find("server", &required_server);
#endif
#ifdef HAVE_IRC_H
      CHNS->IRC = (ModuleEntryIrc *)G->module_list->Find("irc", &required_irc);
#endif
      G->hook_write_userfile->Add(channels_writeuserfile,CHNS);
      G->hook_rehash->Add(channels_rehash,CHNS);
      G->hook_pre_rehash->Add(channels_prerehash,CHNS);
      G->hook_minutely->Add(check_expired_bans,CHNS);
      G->bt_chon->AddBinds(my_chon,1,CHNS);
#if defined(HAVE_IRC_H) || defined(HAVE_SERVER_H)
      G->bt_load->AddBinds(my_load,LOAD1+LOAD2,CHNS);
#endif
      G->bt_unld->AddBinds(my_unld,1+LOAD1+LOAD2,CHNS);
      NEW ( Command ("killban", script_killban, CHNS ) );
      NEW ( Command ("isban", script_isban, CHNS ) );
      NEW ( Command ("matchban", script_matchban, CHNS ) );
      NEW ( Command ("newban", script_newban, CHNS ) );
      NEW ( Command ("channel", script_channel, CHNS ) );
      NEW ( Command ("channels", script_channels, CHNS ) );
      NEW ( Command ("banlist", script_banlist, CHNS ) );
      NEW ( Command ("savechannels", script_savechannels, CHNS ) );
      NEW ( Command ("loadchannels", script_loadchannels, CHNS ) );
      NEW ( Command ("validchan", script_validchan, CHNS ) );
      NEW ( Command ("isdynamic", script_isdynamic, CHNS ) );
      NEW ( Command ("getchaninfo", script_getchaninfo, CHNS ) );
      NEW ( Command ("setchaninfo", script_setchaninfo, CHNS ) );
      NEW ( Command ("setlaston", script_setlaston, CHNS ) );
      NEW ( Command ("addchanrec", script_addchanrec, CHNS ) );
      NEW ( Command ("delchanrec", script_delchanrec, CHNS ) );
      G->bt_dcc->AddBinds(channels_commands,COMMANDS,CHNS);
   }
   return r;
}
