/*
 * UserEntryConsole.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "UserEntryLaston.h"
#include "UserEntryUnknown.h"
#include "UserEntryString.h"
#include "UserEntryConsole.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "ModuleEntryConsole.h"
#include "Socket.h"
#include "UserRecord.h"
#define CLASS "UserEntryConsole"

UserEntryConsole::~UserEntryConsole ( void ) {
   FUNCTION;
}

void UserEntryConsole::Unpack (UserRecord * u, MessageQueue * q) {
   FUNCTION;
   UserEntryConsole * ue = (UserEntryConsole *)u->GetEntry(this);
   MessageQueue * x = q->Next();
   char * w, * p, buf[512];
   
   if (x) {
      strncpy(buf, x->Data(), 511);
      buf[511] = 0;
      p = buf;
      w = G->Split(&p);
      ue->conflags.Set(G->Split(&p));
      ue->stripflags = G->StripModes(G->Split(&p));
      ue->echoflags = (G->Split(&p)[0] == '1') ? 1 : 0;
      ue->Set(w,atoi(G->Split(&p)));
      ue->conchan = atoi(G->Split(&p));
   } 
}

void UserEntryConsole::Pack (UserEntry * e) {
   FUNCTION;
   UserEntryUnknown * ueu = 
     (UserEntryUnknown *)e->Owner()->GetEntry(G->entrytype_unknown);
   UserEntryConsole * uec = (UserEntryConsole *)e;
   char work[512];

   G->SPrintF(work, 511, "%s %s %s %d %d %d", uec->string,
	      uec->conflags.Mask(), G->StripMaskType(uec->stripflags),
	      uec->echoflags, uec->when, uec->conchan);
   ueu->NewLine((char *)type, work);
}

UserEntry * UserEntryConsole::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryConsole ( u, this ) );
}

int UserEntryConsole::ExpMem ( void ) {
   FUNCTION;
   return UserEntryLaston::ExpMem() + sizeof(UserEntryConsole)
     - sizeof(UserEntryLaston);
}

int UserEntryConsole::Write (FILE * f, Socket * idx) {
   FUNCTION;
   if (!idx) 
     return fprintf(f,"=CONSOLE %s %s %s %d %lu %d\n", string, 
		    conflags.Mask(), G->StripMaskType(stripflags),
		    echoflags, when, conchan) != EOF;
   return 1;
}

int UserEntryConsole::ScriptGet (ScriptContext * sc, int, ScriptArgument **) {
   FUNCTION;
   char work[512];
   
   G->SPrintF(work, 511, "%s %s %s %d %d %d", string, conflags.Mask(),
	      G->StripMaskType(stripflags), echoflags, when, 
	      conchan);
   sc->SetString(work);
   return SCRIPT_OK;
}

int UserEntryConsole::ScriptSet (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char buf[81];
   
   BADARGS(4,9, CONSOLE->Lang(0x001, ""));
   strncpy(buf, argv[3]->AsString(), 80);
   buf[80] = 0;
   UserEntryString::Set(buf);
   if (argc > 4) {
      conflags.Set(argv[4]->AsString());
      if (argc > 5) {
	 stripflags = G->StripModes(argv[5]->AsString());
	 if (argc > 6) {
	    echoflags = argv[6]->AsInteger();
	    if (argc > 7) {
	       when = argv[7]->AsInteger();
	       if (argc > 8)
		 conchan = argv[8]->AsInteger();
	    }
	 }
      }
   }	 
   return SCRIPT_OK;
}

void UserEntryConsole::Display ( Socket * idx ) {
   FUNCTION;
   context;
   FlagRecord fr (idx->User());
   
   if (fr.GlobMaster()) {
      idx->Send(CONSOLE->Lang(0x002, ""));
      idx->Send(CONSOLE->Lang(0x003, "s", string));
      idx->Send(CONSOLE->Lang(0x004, "sss", conflags.Mask(),
			G->StripMaskType(stripflags),
			echoflags ? "yes" : "no"));
      idx->Send(CONSOLE->Lang(0x005, "dd", when, conchan));
   }
}
