/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#define _BSD_SOURCE
#define _XOPEN_SOURCE
#include "ModuleEntryCrypt.h"
#include "Socket.h"
#define CLASS "ModuleEntryCrypt"

static Version crypt_ver;
static struct patches crypt_patches[] = {
#include "patch.h"
};

ModuleEntryCrypt::ModuleEntryCrypt ( ModuleEntry * o, char ** r )
: ModuleEntryEncryption ( "crypt", crypt_ver.Set(2,0), crypt_patches, o, r) {
   FUNCTION;
   Version required_eggdrop;

   if (!*r) {
      required_eggdrop.Set(2,0);
      if (!Depend("eggdrop", &required_eggdrop))
	*r = "Your version of eggdrop is incompatible.";
      else {
	 G->module_encrypt = this;
      }
   }
}
			   
ModuleEntryCrypt::~ModuleEntryCrypt ( void ) {
   FUNCTION;
   G->module_encrypt = 0;
}

int ModuleEntryCrypt::ExpMem ( void ) {
   FUNCTION;
   return ModuleEntry::ExpMem () + sizeof(ModuleEntryCrypt);
}

void ModuleEntryCrypt::EncryptPass (char * text, char * newp) {
   FUNCTION;
   char saltdata[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";
   char salt[3];
   int salt1, salt2;

   salt1 = text[0];
   salt2 = text[1];
   while (salt1 >= 64)
      salt1 = salt1 / 2;
   while (salt2 >= 64)
      salt2 = salt2 / 2;
   salt[0] = saltdata[salt1];
   salt[1] = saltdata[salt2];
   salt[2] = '\0';
   G->SPrintF(newp, 32, "+%s", crypt(text, salt));
}

char * ModuleEntryCrypt::Encrypt (const char *, const char * str) {
   FUNCTION;
   char *dest;

   dest = (char *) my_malloc((strlen(str) + 1), G->memtype_misc);
   G->StrCpy(dest, str);
   return dest;
}

char * ModuleEntryCrypt::Decrypt (const char *, const char * str) {
   FUNCTION;
   char *dest;

   dest = (char *) my_malloc((strlen(str) + 1), G->memtype_misc);
   G->StrCpy(dest, str);
   return dest;
}

void ModuleEntryCrypt::Report (Socket * sock, int) {
   FUNCTION;
   sock->Printf("    Crypt encryption module\n");
}
