/*
 * ctcp.c - all the ctcp handling (except DCC, it's special ;)
 * this is an eggdrop-specific ctcp module, for emulation of another client
 * get a client-specific module
 * 
 */

#define MAKING_CTCP
#include "Socket.h"
#include "UserRecord.h"
#include "BindTable.h"
#include "HelpReference.h"
#include "ModuleEntryCtcp.h"
#include "ModuleEntryEggdrop.h"
#include "ModuleEntryServer.h"
#include "ScriptContext.h"
#include "StringVariable.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#define CLASS ""

static ModuleEntryCtcp * CTCP;
static ModuleEntryEggdrop * G;
static char ctcp_version[256];
static char ctcp_finger[256];
static char ctcp_userinfo[256];

static BindReturn ctcp_FINGER (void **, const char *) {
   FUNCTION;
   context;
   if (ctcp_finger[0]) 
     CTCP->SERVER->ctcp_reply_len += G->SPrintF(CTCP->SERVER->ctcp_reply + 
					  CTCP->SERVER->ctcp_reply_len,
					  511 - CTCP->SERVER->ctcp_reply_len,
					  "\001FINGER %s\001", ctcp_finger);
   return NOMATCH;
}

static BindReturn ctcp_ECHOPINGERR (void ** args, const char *) {
   FUNCTION;
   context;
   if (strlen((char *)args[5]) <= 80)
     CTCP->SERVER->ctcp_reply_len += G->SPrintF(CTCP->SERVER->ctcp_reply +
					  CTCP->SERVER->ctcp_reply_len,
					  511 - CTCP->SERVER->ctcp_reply_len,
					  "\001%s %s\001",
					  args[4],args[5]);
   return NOMATCH;
}

static BindReturn ctcp_VERSION (void **, const char *) {
   FUNCTION;
   context;
   if (ctcp_version[0]) 
     CTCP->SERVER->ctcp_reply_len += G->SPrintF(CTCP->SERVER->ctcp_reply + 
					  CTCP->SERVER->ctcp_reply_len,
					  511 - CTCP->SERVER->ctcp_reply_len,
					  "\001VERSION %s\001", ctcp_version);
   return NOMATCH;
}

static BindReturn ctcp_USERINFO (void **, const char *) {
   FUNCTION;
   context;
   if (ctcp_userinfo[0]) 
     CTCP->SERVER->ctcp_reply_len += G->SPrintF(CTCP->SERVER->ctcp_reply + 
					  CTCP->SERVER->ctcp_reply_len,
					  511 - CTCP->SERVER->ctcp_reply_len,
					  "\001USERINFO %s\001", 
					  ctcp_userinfo);
   return NOMATCH;
}

static BindReturn ctcp_CLIENTINFO (void ** args, const char *) {
   FUNCTION;
   int p;
   char * par = (char *)args[5];
   
   context;
   if (!par[0])
     p = 0x001;
   else if (!strcasecmp(par, "version"))
     p = 0x002;
   else if (!strcasecmp(par, "clientinfo"))
     p = 0x003;
   else if (!strcasecmp(par, "userinfo"))
     p = 0x004;
   else if (!strcasecmp(par, "errmsg"))
     p = 0x005;
   else if (!strcasecmp(par, "finger"))
     p = 0x006;
   else if (!strcasecmp(par, "time"))
     p = 0x007;
   else if (!strcasecmp(par, "action"))
     p = 0x008;
   else if (!strcasecmp(par, "dcc"))
     p = 0x009;
   else if (!strcasecmp(par, "ping"))
     p = 0x00A;
   else if (!strcasecmp(par, "echo"))
     p = 0x00B;
   else if (!strcasecmp(par, "chat"))
     p = 0x00C;
   else
     p = 0x00D;
   CTCP->SERVER->ctcp_reply_len += G->SPrintF (CTCP->SERVER->ctcp_reply + 
					 CTCP->SERVER->ctcp_reply_len,
					 511 - CTCP->SERVER->ctcp_reply_len, 
					 "\001%s %s\001",
					 (p == 0x00D) ? "ERRMSG" :
					 "CLIENTINFO",
					 CTCP->Lang(p, "s", par));
   return NOMATCH;
}

static BindReturn ctcp_TIME (void **, const char *) {
   FUNCTION;
   char tms[81];
   
   context;
   strcpy(tms, ctime(&G->now));
   tms[strlen(tms) - 1] = 0;
   CTCP->SERVER->ctcp_reply_len += G->SPrintF(CTCP->SERVER->ctcp_reply + 
					CTCP->SERVER->ctcp_reply_len, 
					511 - CTCP->SERVER->ctcp_reply_len,
					"\001TIME %s\001", tms);
   return NOMATCH;
}

static BindReturn ctcp_CHAT (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   FlagRecord atr ( u );
   Socket * ptr = G->socklist;
   char buf[512];
   
   context;
   if (atr.GlobParty() || atr.GlobXfer() ||
       (atr.GlobOp() && !G->require_p)) {
      while ((ptr = ptr->Next()))
	if ((ptr->Flags() & Socket::LISTENING) &&
	    (!strcmp(ptr->Nick(), "(telnet)") || 
	     !strcmp(ptr->Nick(), "(users)"))) {
	   struct in_addr addr;
	   
	   addr.s_addr = G->natip[0] ? inet_addr(G->natip) 
	     : G->GetMyIp()->s_addr;
	   G->SPrintF(buf, 511, "\001DCC CHAT chat %lu %u\001\n",
		      iptolong(&addr), ptr->LocalPort());
	   CTCP->SERVER->SendMsg(Q_SERVER, (char *)args[0], buf);
	   break;
	}
      if (!ptr)
	CTCP->SERVER->ctcp_reply_len += G->SPrintF (CTCP->SERVER->ctcp_reply + 
					      CTCP->SERVER->ctcp_reply_len, 
					      511 - CTCP->SERVER->ctcp_reply_len,
					      "\001%s\001",
					      CTCP->Lang(0x00E, ""));
   }
   return NOMATCH;
}

#define COMMANDS 9
static command_binds myctcp[COMMANDS] = {
     { "FINGER", "", ctcp_FINGER, NULL },
     { "ECHO", "", ctcp_ECHOPINGERR, NULL },
     { "PING", "", ctcp_ECHOPINGERR, NULL },
     { "ERRMSG", "", ctcp_ECHOPINGERR, NULL },
     { "VERSION", "", ctcp_VERSION, NULL },
     { "USERINFO", "", ctcp_USERINFO, NULL },
     { "CLIENTINFO", "", ctcp_CLIENTINFO, NULL },
     { "TIME", "", ctcp_TIME, NULL },
     { "CHAT", "", ctcp_CHAT, NULL },
};

extern "C" {
   char *ctcp_start (ModuleEntryEggdrop *, ModuleEntry *);
};

char *ctcp_start (ModuleEntryEggdrop * GG, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   G = GG;
   context;
   CTCP = NEW ( ModuleEntryCtcp ( me, &r ) );
   if (!r) {
      CTCP->SERVER->bt_ctcp->AddBinds(myctcp, COMMANDS, CTCP);
      strncpyz(ctcp_version, G->VersionStr(), 160);
      strncpyz(ctcp_finger, G->VersionStr(), 160);
      strncpyz(ctcp_userinfo, G->VersionStr(), 160);
      NEW ( StringVariable ("ctcp_version", ctcp_version, 120, 0, CTCP) );
      NEW ( StringVariable ("ctcp_finger", ctcp_finger, 120, 0, CTCP) );
      NEW ( StringVariable ("ctcp_userinfo", ctcp_userinfo, 120, 0, CTCP) );
   }
   return r;
}
