/*
 * DccFiles.C - a user in the filesystem
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "filesys.h"
#include "FileDB.h"
#include "ModuleEntryFilesys.h"
#include "UserRecord.h"
#include "DccFiles.h"
#include "DccFilesPass.h"
#include "UserEntryDccDir.h"
#include "UserEntryLaston.h"
#include "BindTable.h"
#define CLASS "DccFiles"

DccFiles::DccFiles ( DccFilesPass & old )
: DccChat ( &old ) {
   FUNCTION;
   if (CommonInit ( )) {
      G->LOG_MISC.Put(FILESYS->Lang(0x097, ""));
      Close();
   }
}

DccFiles::DccFiles ( DccChat & old )
: DccChat ( old ) {
   FUNCTION;
   status |= STAT_CHAT;
   if (CommonInit ( )) {
      DccChat * newone = NEW ( DccChat ( * this ) );
      newone->CheckChon();
   }
}

int DccFiles::CommonInit ( void ) {
   FUNCTION;
   FILE *f;
   UserEntryDccDir * ued = GetDccDir(user);
   int lang;
   
   type = FILESYS->DCC_FILES;
   type_flags = DCT_MASTER|DCT_VALIDIDX|DCT_SHOWWHO|DCT_SIMUL|DCT_CANBOOT|DCT_FILES;
   Printf("\n");
   if (!HasStatus(STAT_CHAT))
      ShowMotd();
   for (lang = 0x00F; lang < 0x010; lang++)
     Send(FILESYS->Lang(lang, ""));
   strncpy(dir, safe_str(ued->Value()), DIRLEN);
   /* does this dir even exist any more? */
   if (!(f = FILEDB->Open(dir,0))) {
      dir[0] = 0;
      if (!(f = FILEDB->Open(dir,0))) {
	 for (lang = 0x011; lang < 0x015; lang++)
	   Send(FILESYS->Lang(lang, ""));
	 FILESYS->dccdir[0] = 0;
	 return 1;
      }
   }
   TouchLaston(user,"filearea",G->now);
   FILEDB->Close(f);
   Send(FILESYS->Lang(0x015, "s", dir));
   return 0;
}

DccFiles::~DccFiles ( void ) {
   FUNCTION;
}

int DccFiles::ExpMem ( void ) {
   FUNCTION;
   return sizeof(DccFiles);
}

void DccFiles::Activity (char * buf, int) {
   FUNCTION;
   char * code, tbuf[512], * nbuf = tbuf;
   FlagRecord fr ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   void * args[2] = { buf, this };
   Socket * ptr = G->socklist;
   BindReturn x;
   
   context;
   if (buf[0] && DetectDccFlood())
     return;
   TimeNow();
   G->bt_filt->Check(buf, 0, args);
   strncpyz(nbuf, (char *)args[0], 511);
   if (nbuf[0]) {
      if (nbuf[0] == ',') {
	 code = G->Lang(0x106, "ss", nick, nbuf + 1);
	 while ((ptr = ptr->Next()))
	   if ((ptr->TypeFlags() & DCT_MASTER) && 
	       fr.Set(ptr->User())->GlobMaster() &&
	       ((ptr->Type() != G->DCC_CHAT) ||
		(((DccChat *)ptr)->Channel() >= 0)) &&
	       ((ptr != this) || HasStatus(STAT_ECHO)))
	     ptr->Send(code);
      } else {
	 context;
	 if (nbuf[0] == '.')
	   nbuf ++;
	 if (nbuf[0]) {
	    code = G->Split(&nbuf);
	    fr.Set(user,con_chan);
	    args[0] = this;
	    args[1] = nbuf;
	    if ((x = FILESYS->bt_fil->Check(code, &fr, args)) == EXEC_BRK) {
	       UserEntryDccDir * ued = GetDccDir(user);
	       
	       Send(FILESYS->Lang(0x09A, ""));
	       FlushLines();
	       FILESYS->TRANSFER->LOG_FILES.Put(FILESYS->Lang(0x09B, "ss",nick, host));
	       ued->Set(dir);
	       if (HasStatus(STAT_CHAT)) {
		  DccChat * newone;
		  
		  Send(FILESYS->Lang(0x09C, ""));
		  newone = NEW ( DccChat ( *this ) );
		  newone->CheckChon();
	       } else {
		  Send(FILESYS->Lang(0x09D, ""));
		  FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x09E, "ssd", nick,
								  host, remoteport));
		  if ((key != STDIN) || G->backgrd) 
		    Close();
		  else {
		     Send(G->Lang( 0x157, ""));
		     Chatter();
		  }
	       }
	       return;
	    } else if (x == AMBIGUOUS)
		 Send(G->Lang(0x107, ""));
	    else if (x == NOMATCH)
	      Send(G->Lang(0x108, ""));
	    else if (x == EXEC_LOG)
	      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x09F, "sss", nick,code, nbuf));
	 }
      }
   }
   if (HasStatus(STAT_PAGE))
     FlushLines();
}

void DccFiles::Display2 (Socket * where) {
   FUNCTION;
   where->Send(FILESYS->Lang(0x0A7, "ccccc",
		       HasStatus(STAT_CHAT) ? 'C' : 'c',
		       HasStatus(STAT_PARTY) ? 'P' : 'p',
		       HasStatus(STAT_TELNET) ? 'T' : 't',
		       HasStatus(STAT_ECHO) ? 'E' : 'e',
		       HasStatus(STAT_PAGE) ? 'P' : 'p'));
}

void DccFiles::Eof (void) {
   FUNCTION;
   con_flags = 0;
   G->LOG_MISC.Put(FILESYS->Lang(0x0A9, "ssd", nick, host, remoteport));
   Close();
}
