/*
 * UserEntryFstat.C - code for handling Filestats entries
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "filesys.h"
#include "ModuleEntryFilesys.h"
#include "UserEntryFstat.h"
#include "UserEntryUnknown.h"
#include "UserRecord.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#define CLASS "UserEntryFstat"

UserEntryFstat::~UserEntryFstat ( void ) {
   FUNCTION;
}

void UserEntryFstat::Unpack ( UserRecord * u, MessageQueue * x ) {
   FUNCTION;
   UserEntryFstat * uef = GetFstat(u);
   char buf[512], *q, *r;
   
   if ((x = x->Next())) {
      strncpy(buf, x->Data(), 511);
      buf[511] = 0;
      if ((q = strchr(buf,' '))) {
	 *q++ = 0;
	 uef->uploads = atoi(buf);
	 if ((r = strchr(q,' '))) {
	    *r++ = 0;
	    uef ->upload_ks = atoi(q);
	    if ((q = strchr(r,' '))) {
	       *q++ = 0;
	       uef->dnloads = atoi(r);
	       if ((r = strchr(q,' '))) {
		  *r++ = 0;
		  uef ->dnload_ks = atoi(q);
	       } 
	    }
	 } 
      }
   }
}

void UserEntryFstat::Pack ( UserEntry * u ) {
   FUNCTION;
   UserEntryFstat * uef = (UserEntryFstat *)u;
   UserEntryUnknown * uen = GetUnkown(u->Owner());
   char buf[40];
   
   context;
   G->SPrintF(buf, 39, "%9 %9 %9 %9", uef->uploads, uef->upload_ks,
	      uef->dnloads, uef->dnload_ks);
   uen->NewLine((char *)type, buf);
}

UserEntry * UserEntryFstat::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryFstat ( u, this ) );
}

int UserEntryFstat::ExpMem ( void ) {
   FUNCTION;
   return sizeof(UserEntryFstat);
}

int UserEntryFstat::Write ( FILE * f, Socket * ) {
   FUNCTION;
   context;
   return fprintf(f,"=FSTAT %09u %09u %09u %09u\n", uploads, upload_ks,
		  dnloads, dnload_ks) != EOF;
}

void UserEntryFstat::PruneUser ( Socket * ) {
   FUNCTION;
   DELETE ( this );
}

int UserEntryFstat::ScriptSet (ScriptContext * sc, int argc , ScriptArgument ** argv) {
   FUNCTION;
   int f = 0, k = 0;
   FlagRecord flags ( owner );
   
   BADARGS(4,6,FILESYS->Lang(0x08F, ""));
   if (argc > 4) 
     f = argv[4]->AsInteger();
   if (argc > 5)
     k = argv[5]->AsInteger();
   if (argv[3]->AsString()[0] == 'u') {
      uploads = f;
      upload_ks = k;
   } else if (argv[3]->AsString()[0] == 'd') {
      dnloads = f;
      dnload_ks = k;
   }
   if (G->shareout && !G->noshare && !flags.GlobBot() && !flags.GlobUnshared())
     G->shareout(NULL,"cfstat %c %09u %09u\n", argv[3]->AsString()[0], f, k);
   return SCRIPT_OK;
}

int UserEntryFstat::ScriptGet (ScriptContext * sc, int argc , ScriptArgument ** argv) {
   FUNCTION;
   char d[50];
   
   BADARGS(3,4,FILESYS->Lang(0x090, ""));
   if (argc == 3) 
     G->SPrintF(d, 49, "%u %u %u %u", uploads, upload_ks, dnloads, dnload_ks);
   else if (argv[3]->AsString()[0] == 'u') 
     G->SPrintF(d, 49, "%u %u", uploads, upload_ks);
   else if (argv[3]->AsString()[0] == 'd') 
     G->SPrintF(d, 49, "%u %u", dnloads, dnload_ks);
   sc->SetString(d);
   return SCRIPT_OK;
}

void UserEntryFstat::Display ( Socket * idx ) {
   FUNCTION;
   if (uploads || dnloads)
     idx->Send(FILESYS->Lang(0x091,"dddd",dnloads,dnload_ks,uploads,upload_ks));
}     

void UserEntryFstat::GotShare (char * par, Socket * ) {
   FUNCTION;
   G->noshare = 1;
   if (par[0] == '-') 
     Clear ();
   else if (par[0] == '+') {
      long x = (atoi(++par + 1) + 512) / 1024;

      if (par[0] == 'u') {
	 uploads++;
	 upload_ks += x;
      } else {
	 dnloads++;
	 dnload_ks += x;
      }
   } else {
      char c = *par++;
      int a = atoi(G->Split(&par));
      int b = atoi(par);
      
      if (c == 'u') {
	 uploads = a;
	 upload_ks = b;
      } else {
	 dnloads = a;
	 dnload_ks = b;
      }
   }
}

void UserEntryFstat::AddFileUp ( int val ) {
   FUNCTION;
   FlagRecord flags ( owner );
   
   uploads++;
   upload_ks += (val + 512) / 1024;
   if (G->shareout && !flags.GlobBot() && !flags.GlobUnshared())
     G->shareout(NULL,"cfstat %s +u%lu\n", owner->Handle(), val);
}

void UserEntryFstat::AddFileDown ( int val ) {
   FUNCTION;
   FlagRecord flags ( owner );
   
   dnloads++;
   dnload_ks += (val + 512) / 1024;
   if (G->shareout && !flags.GlobBot() && !flags.GlobUnshared())
     G->shareout(NULL,"cfstat %s +d%lu\n", owner->Handle(), val);
}

void UserEntryFstat::Clear ( void ) {
   FUNCTION;
   FlagRecord flags ( owner );
   
   dnloads = uploads = dnload_ks = upload_ks = 0;
   if (G->shareout && !flags.GlobBot() && !flags.GlobUnshared())
     G->shareout(NULL,"cfstat %s -\n", owner->Handle());
}
