/*
 * filesys.c - the file system module
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include "filesys.h"
#include "FileDB.h"
#include "Socket.h"
#include "UserRecord.h"
#include "UserEntryLaston.h"
#include "BindTable.h"
#include "MessageQueue.h"
#include "HelpReference.h"
#include "HookList.h"
#include "BindTableFiles.h"
#include "ModuleEntry.h"
#include "ModuleEntryBotnet.h"
#include "ModuleEntryServer.h"
#include "ModuleEntryFilesys.h"
#include "Bot.h"
#include "UserEntryFstat.h"
#include "UserEntryDccDir.h"
#include "FileGet.h"
#include "DccFiles.h"
#include "DccFilesPass.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "Command.h"
#include "Channel.h"
#include "FileQueue.h"
#include <sys/file.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#ifndef S_IFREG
#define S_IFREG 0100000
#endif
#ifndef S_IFLNK
#define S_IFLNK 0120000
#endif

#ifdef HAVE_SERVER_H
#include "server.h"
static Version required_server ( 2, 0 );
#define COUNT1 1
#else
#define COUNT1 0
#endif
#ifdef HAVE_BOTNET_H
#include "TandemBot.h"
static Version required_botnet ( 2, 0 );
#define COUNT2 1
#else
#define COUNT2 0
#endif
#define CLASS ""

ModuleEntryFilesys * FILESYS;
static ModuleEntryEggdrop * G;

static void filedb_add (FILE * f, const char * filename, const char * nick) {
   FUNCTION;
   long where = 0;
   filedb fdb;
   /* when the filedb was opened, a record was already created */
   if (!FILESYS->FindMatch(f, filename, &where, &fdb))
     return;
   strncpy(fdb.uploader, nick, HANDLEN);
   fdb.uploader[HANDLEN] = 0;
   fdb.uploaded = G->now;
   fseek(f, where, SEEK_SET);
   fwrite(&fdb, sizeof(filedb), 1, f);
}

#ifdef HAVE_BOTNET_H
static BindReturn filesys_zapf (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], buf[512], * param = buf, *whofrom;
   char dir[DIRLEN + 1], s[DIRLEN + 1], s1[DIRLEN + 1], *p, *path;
   int lang = 0x004, i = -1;
   FILE * f;
   filedb fdb;
   long where = 0;
   Socket * ptr = 0;
   TandemBot * tb;
   struct stat ss;
   Bot * bot = 0;
   
   context;
   if (!FILESYS->BOTNET)
     return EXEC_LOG;
   strncpy(param, (char *)args[1], 511);
   param[511] = 0;
   switch (*param++) {
     case 'r':
      /* filesys r<sock:nick> <file> */
      whofrom = G->Split(&param);
      if (!(p = strrchr(param, '/')))
	dir[0] = 0;
      else {
	 *p = 0;
	 strcpy(dir, param);
	 param = p + 1;
      }
      if (!(f = FILEDB->Open(dir, 0))) 
	 lang = 0x005;
      else if (!FILESYS->FindMatch(f, param, &where, &fdb)) {
	 lang = 0x006;
	 FILEDB->Close(f);
      } else if ((!(fdb.stat & FILE_SHARE)) || 
		 (fdb.stat & (FILE_HIDDEN | FILE_DIR))) {
	 lang = 0x007;
	 FILEDB->Close(f);
      } else {
	 FILEDB->Close(f);
	 /* copy to /tmp if needed */
	 G->SPrintF(s1, DIRLEN, "%s%s%s%s", FILESYS->dccdir, dir,
		    dir[0] ? "/" : "", param);
	 if (FILESYS->TRANSFER->copy_to_tmp) {
	    G->SPrintF(s, DIRLEN, "%s%s", G->tempdir, param);
	    G->CopyFile(s1, s);
	 } else
	   strcpy(s, s1);
	 if ((i = FILESYS->TRANSFER->RawDccSend(s, "*remote", "*remote")) < 0)
	   FILESYS->TRANSFER->WipeTmpFilename(s, NULL);
      }
      if (i < 0) 
	G->SPrintF(s1, DIRLEN, "!%s %s/%s %s", whofrom, dir, param,
		   FILESYS->Lang(lang, ""));
      else {
	 /* grab it from dcc struct and bounce real request across net */
	 stat ( s, &ss );
	 ptr = G->socklist->Find(i, FILESYS->TRANSFER->FILE_LISTEN);
	 G->SPrintF(s, DIRLEN, "s%s %s/%s %d %d %d", whofrom, dir, param,
		    iptolong(G->GetMyIp()), ptr->LocalPort(), ss.st_size);
	 FILESYS->TRANSFER->LOG_FILES.Put(FILESYS->Lang(0x008, "sss", dir, dir[0] ? "/" : "", param));
      }
      if ((tb = FILESYS->BOTNET->tandbot->Find(from)) && (bot = tb->NextBot()))
	bot->SendZapf(G->botnetnick, "filesys", s1);
      break;
     case '!':
      /* filesys !<sock:nick> <file> <reason> */
      whofrom = G->Split(&param);
      path = G->Split(&param);
      if ((p = strchr(whofrom, ':'))) {
	 *p = 0;
	 lang = atoi(whofrom);
	 for (ptr = G->socklist; (ptr = ptr->Next()); ) 
	   if (ptr->FD() == lang)
	     ptr->Send(FILESYS->Lang(0x009, "ss", path, param));
	 *p = ':';
      }
      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x00A, "ss", path, param));
      break;
     case 's':
      /* filesys !<sock:nick> <file> <IP#> <port> <size> */
      whofrom = G->Split(&param);
      path = G->Split(&param);
      if ((p = strrchr(path, '/')))
	path = p + 1;
      if ((p = strchr(whofrom, ':')))
	whofrom = p + 1;
      /* send it to 'nick' as if it's from me */
#ifdef HAVE_SERVER_H
      if (FILESYS->SERVER) {
	 G->SPrintF(s, DIRLEN, "\001DCC SEND %s %s\001", path, param);
	 FILESYS->SERVER->SendMsg(Q_SERVER, whofrom, s);
      }
#endif
      break;
     default:
      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x00B, "s", param));
   }
   return EXECUTED;
}

static command_binds myzapf [] = {
   { "filesys", "", filesys_zapf, "filesys:bot" }
};

static BindReturn load_botnet(void ** args, const char *) {
   FUNCTION;
   FILESYS->BOTNET = (ModuleEntryBotnet *)(args ? args[0]
	: G->module_list->Find("botnet", &required_botnet));
   if (!FILESYS->BOTNET || !(*FILESYS->BOTNET->Vers() <= required_botnet))
	FILESYS->BOTNET = NULL;
   else
	FILESYS->BOTNET->bt_bot->AddBinds(myzapf,1,FILESYS);
   return EXECUTED;
}

static BindReturn unload_botnet (void **, const char *) {
   FILESYS->BOTNET = NULL;
   return EXECUTED;
}
#endif

/* someone uploaded a file -- add it */
static BindReturn filesys_rcvd ( void ** args, const char *) {
   FUNCTION;
   FileSend * idx = (FileSend *)args[0];
   int l = strlen(FILESYS->dccdir);
   FILE * f;
   UserEntryFstat * fs;
   
   context;
   if ((idx->From()[0] != '*') && !strncmp(FILESYS->dccdir, idx->Dir(), l)) {
      if ((f = FILEDB->Open(idx->Dir() + l, 2))) {
	 filedb_add(f, idx->Filename(), idx->Nick());
	 FILEDB->Close(f);
      }
      if (idx->User()) {
	 fs = GetFstat ( idx->User() );
	 fs->AddFileUp(idx->Length());
      }
   }
   return EXECUTED;
}

static BindReturn filesys_sent ( void ** args, const char *) {
   FUNCTION;
   FileGet * idx = (FileGet *)args[0];
   int l = strlen(FILESYS->dccdir);
   filedb fdb;
   FILE *f;
   long where = 0;

   /* absolute dir?  probably a script sending it, and it might not */
   /* be in the file system at all, so just leave it alone */
   if ((idx->From()[0] != '*') && !strncmp(FILESYS->dccdir, idx->Dir(), l)) {
      if ((f = FILEDB->Open(idx->Dir() + l, 0))) {
	 if (FILESYS->FindMatch(f, idx->Filename(), &where, &fdb)) {
	    fdb.gots++;
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb, sizeof(filedb), 1, f);
	 }
	 FILEDB->Close(f);
      }
      if (idx->User()) {
	 UserEntryFstat * fs = GetFstat ( idx->User() );
	 fs->AddFileDown(idx->Length());
      }
   }
   return EXECUTED;
}

static command_binds myrcvd [] = {
   { "*", "", filesys_rcvd, "filesys:rcvd" }
};

static command_binds mysent [] = {
   { "*", "", filesys_sent, "filesys:sent" }
};

/* given current directory, and the desired changes, fill 'real' with */
/* the new current directory.  check directory parmissions along the */
/* way.  return 1 if the change can happen, 0 if not. */
static int resolve_dir (const char * current, const char * change, char * real,
			Socket * idx) {
   FUNCTION;
   char s[DIRLEN + 1], new2[DIRLEN + 1], *p, *q, *new1 = new2;
   FILE *f;
   filedb fdb;
   FlagRecord user ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   FlagRecord req ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   int ret;
   long i = 0;
   
   context;
   if (change[0]) {
      if (change[0] == '/') {
	 change ++;
	 real[0] = 0;
      } else 
	strcpy(real, current);
      G->SPrintF(new1, DIRLEN, "%s/", change);
      for (p = strchr(new1, '/'); p; new1 = p + 1, p = strchr(new1, '/')) {
	 *p++ = 0;
	 if (!new1[0] || (!new1[1] && (new1[0] == '.')))
	   continue;
	 else if (!strcmp(new1, "..")) {	/* go back */
	    if (!(q = strrchr(real, '/'))) {
	       /* can't go back from here? */
	       if (!real[0]) {
		  strcpy(real, current);
		  return 0;
	       }
	       real[0] = 0;
	    } else
	      *q = 0;
	 } else {
	    /* allowed access here? */
	    if ((f = FILEDB->Open(real, 0))) {
	       ret = FILESYS->FindMatch(f, new1, &i, &fdb);
	       FILEDB->Close(f);
	       if (ret && (fdb.stat & FILE_DIR) && !fdb.sharelink[0]) {
		  if (idx)
		    user.Set(idx->User(),fdb.chname);
		  else {
		     user.SetGlobOwner();
		     user.SetGlobBot();
		  }
		  req.BreakDown(fdb.flags_req);
		  if (req.Ok(user)) {
		     if ((i = strlen(real)) && (real[i - 1] != '/')) {
			real[i++] = '/';
			real[i] = 0;
		     }
		     G->SPrintF(real + i, DIRLEN - i, "%s", new1);
		     continue;
		  }
	       }
	    }
	    strcpy(real, current);
	    return 0;
	 }
      }
      /* sanity check: does this dir exist? */
      G->SPrintF(s, DIRLEN, "%s%s", FILESYS->dccdir, real);
      if (!(f = fopen(s, "r"))) 
	return 0;
      fclose(f);
   }
   return 1;
}

static void do_dcc_send (Socket * idx, char * dir, char * fn,
			 const char * tnick) {
   FUNCTION;
   char s[DIRLEN + 1], s1[DIRLEN + 1];
   FILE *f;
   int lang;
   
   context;
   if (!FILESYS->dccdir[0]) 
     lang = 0x016;
   else if (strchr(fn, '/'))
     lang = 0x018;
   else {
      if (dir[0])
	G->SPrintF(s, DIRLEN, "%s%s/%s", FILESYS->dccdir, dir, fn);
      else
	G->SPrintF(s, DIRLEN, "%s%s", FILESYS->dccdir, fn);
      if (!(f = fopen(s, "r"))) 
	lang = 0x01A;
      else {
	 fclose(f);
	 /* already have too many transfers active for this user?  queue it */
	 if (FILESYS->TRANSFER->AtLimit(tnick)) {
	    G->SPrintF(s, DIRLEN, "%d*%s%s", strlen(FILESYS->dccdir), FILESYS->dccdir, dir);
	    NEW ( FileQueue (s, fn, idx->Nick(), tnick) );
	    idx->Send(FILESYS->Lang(0x026, "ss", fn, tnick));
	    return;
	 } else {
	    if (FILESYS->TRANSFER->copy_to_tmp) {
	       /* copy this file to /tmp */
	       G->SPrintF(s, DIRLEN, "%s%s%s%s", FILESYS->dccdir, dir, 
			  dir[0] ? "/" : "", fn);
	       G->SPrintF(s1, DIRLEN, "%s%s", G->tempdir, fn);
	       if (G->CopyFile(s, s1)) {
		  idx->Send(FILESYS->Lang(0x024, ""));
		  FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x025, "ss", fn, G->tempdir));
		  return;
	       }
	    } else
	      G->SPrintF(s1, DIRLEN, "%s%s%s%s", FILESYS->dccdir, dir, 
			 dir[0] ? "/" : "", fn);
	    G->SPrintF(s, DIRLEN, "%s%s%s", dir, dir[0] ? "/" : "", fn);
	    switch (FILESYS->TRANSFER->RawDccSend(s1, tnick, idx->Nick())) {
	      case DCCSEND_FULL:
	       lang = 0x01C;
	       break;
	      case DCCSEND_NOSOCK:
	       lang = G-> reserved_port ? 0x01E : 0x020;
	       break;
	      case DCCSEND_BADFN:
	       lang = 0x022;
	       break;
	      default:
	       fn = (fn = strrchr(s1, '/')) ? fn + 1 : s1;
#ifdef HAVE_SERVER_H
	       if (FILESYS->SERVER && !strcasecmp(tnick, idx->Nick()))
		 FILESYS->SERVER->SendNotice(Q_HELP, tnick,
					FILESYS->Lang(0x027, "s", idx->Nick()));
	       idx->Send(FILESYS->Lang(0x028, "ss", FILESYS->SERVER->botname, fn));
#else
	       idx->Send(FILESYS->Lang(0x028, "ss", G->origbotname, fn));
#endif
	       idx->Send(FILESYS->Lang(0x029, "ss", fn, tnick));
	       return;
	    }
	    FILESYS->TRANSFER->WipeTmpFilename(s1, NULL);
	 }
      }
   }
   idx->Send(FILESYS->Lang(lang, ""));
   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(lang + 1, "ss", fn, idx->Nick()));
}

/*** COMMANDS ***/
static BindReturn cmd_sort (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   FILE *f;
   UserEntryDccDir * ued = GetDccDir ( idx->User() );
   
   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x02A, "s", idx->Nick()));
   /* does this dir even exist any more? */
   if (!(f = FILEDB->Open(ued->Value(), 1))) {
      ued->Set("");
      f = FILEDB->Open("", 1);
   }
   FILEDB->Close(f);   
   idx->Send(FILESYS->Lang(0x02B, ""));
   return EXECUTED;
}

static BindReturn cmd_pwd (void ** args, const char *) {
   FUNCTION;
   DccFiles * idx = (DccFiles *)args[0];
   
   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x02C, "s", idx->Nick()));
   idx->Send(FILESYS->Lang(0x015, "s", idx->Dir()));
   return EXECUTED;
}

static BindReturn cmd_pending (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x02D, "s", idx->Nick()));
   FILESYS->TRANSFER->ShowQueuedFiles(idx);
   return EXECUTED;
}

static BindReturn cmd_cancel (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1];
   
   if (!par[0]) 
     idx->Send(FILESYS->Lang(0x02E, ""));
   else {
      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x02F, "ss", idx->Nick(), par));
      FILESYS->TRANSFER->FileQCancel(idx, par);
   }
   return EXECUTED;
}

static BindReturn cmd_chdir  (void ** args, const char *) {
   FUNCTION;
   DccFiles * idx = (DccFiles *)args[0];
   char s[DIRLEN + 1], *par = (char *)args[1];
   
   if (!par) 
     idx->Send(FILESYS->Lang(0x030, ""));
   else if (!resolve_dir(idx->Dir(), par, s, idx))
     idx->Send(FILESYS->Lang(0x031, ""));
   else {
      UserEntryDccDir * ued = GetDccDir(idx->User());

      idx->SetDir(s);
      ued->Set(s);
      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x032, "ss", idx->Nick(), idx->Dir()));
      idx->Send(FILESYS->Lang(0x033, "s", idx->Dir()));
   }
   return EXECUTED;
}

static BindReturn cmd_ls (void ** args, const char * mask2) {
   FUNCTION;
   DccFiles * idx = (DccFiles *)args[0];
   int showall = !strcmp(mask2,"lsa"), cnt = 0, is = 0, l;
   char *mask = "*", s[DIRLEN + 1], destdir[DIRLEN + 1];
   char s2[81], *x, *p, s1[81], *par = (char *)args[1];
   FlagRecord user (FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   FlagRecord req (FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   FILE *f;
   filedb fdb;
   
   context;
   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x034, "sss", idx->Nick(),
				  showall ? "a" : "", par));
   if (par[0]) {
      if ((mask = strrchr(par, '/'))) {
	 *mask = 0;
	 strncpy(s, par, DIRLEN);
	 s[DIRLEN - 1] = 0;
	 mask++;
	 if (!resolve_dir(idx->Dir(), s, destdir, idx)) {
	    idx->Send(FILESYS->Lang(0x035, ""));
	    return EXECUTED;
	 }
      } else {
	 mask = "";
	 strcpy(destdir, idx->Dir());
      }
      /* might be 'ls dir'? */
      if (resolve_dir(destdir, mask, s, idx)) {
	 /* aha! it was! */
	 strcpy(destdir, s);
	 mask = "*";
      }
      f = FILEDB->Open(destdir, 0);
   } else 
     f = FILEDB->Open(idx->Dir(),0);
   rewind(f);
   for (fread(&fdb, sizeof(filedb), 1, f); !feof(f);
	fread(&fdb, sizeof(filedb), 1, f)) {
      if (fdb.stat & FILE_UNUSED) 
	continue;
      if (fdb.stat & FILE_DIR) {
	 /* check permissions */
	 req.BreakDown(fdb.flags_req);
	 user.Set(idx->User(),idx->ConsoleChan());
	 if (!req.Ok(user)) 
	   continue;
      }
      is = 1;
      if (!FILESYS->TRANSFER->WildMatchFile((unsigned char *)mask, 
				     (unsigned char *)fdb.filename) ||
	  ((fdb.stat & FILE_HIDDEN) && !(showall)))
	continue;
      /* display it! */
      if (!cnt) {
	 idx->Send(FILESYS->Lang(0x036, ""));
	 idx->Send(FILESYS->Lang(0x037, ""));
      }
      s[0] = 0;
      l = 0;
      if (fdb.stat & FILE_DIR) {
	 /* too long? */
	 if (strlen(fdb.filename) > 30) 
	   idx->Send(FILESYS->Lang(0x038, "s", fdb.filename));
	    /* causes filename to be displayed on its own line */
	 else
	   l = G->SPrintF(s, DIRLEN, "%s/", fdb.filename);
	 if (fdb.flags_req[0] && user.GlobJanitor()) {
	    strncpy(s1, FILESYS->Lang(0x039, "sss", fdb.flags_req,
					fdb.chname[0] ? " " : "", fdb.chname), 40);
	    s1[40] = 0;
	 } else
	   s1[0] = 0;
	 idx->Send(FILESYS->Lang(0x03A, "s s", s, 30 - l, s1));
      } else {
	 l = 0;
	 if (showall) {
	    if (fdb.stat & FILE_SHARE)
	      l = G->SPrintF(s2, 81, "%s", FILESYS->Lang(0x03B, ""));
	    if (fdb.stat & FILE_HIDDEN)
	      l += G->SPrintF(s2 + l, 81 - l, "%s", FILESYS->Lang(0x03C, ""));
	 }
	 s[0] = (p = ctime(&fdb.uploaded))[8];
	 s[1] = p[9];
	 s[2] = p[4];
	 s[3] = p[5];
	 s[4] = p[6];
	 s[5] = p[22];
	 s[6] = p[23];
	 s[7] = 0;
	 if (fdb.sharelink[0])
	   strcpy(s1, "     ");
	 else if (fdb.size < 1024)
	   G->SPrintF(s1, 80, "%5", (int)fdb.size);
	 else
	   G->SPrintF(s1, 80, "%4k", (int) (fdb.size / 1024));
	 /* too long? */
	 if ((l = strlen(fdb.filename)) > 30) {
	    idx->Send(FILESYS->Lang(0x03D, "s", fdb.filename));
	    fdb.filename[0] = 0;
	    l = 0;
	    /* causes filename to be displayed on its own line */
	 } 
	 idx->Send(FILESYS->Lang(0x3E, "s ss s6s", fdb.filename, 30 - l, s1,
			   fdb.uploader, HANDLEN - strlen(fdb.uploader),
			   s, fdb.gots, s2));
	 if (fdb.sharelink[0])
	   idx->Send(FILESYS->Lang(0x03F, "s", fdb.sharelink));
      }
      if ((x = fdb.desc)[0]) {
	 for (p = strchr(x, '\n'); p; p = strchr(x, '\n')) {
	    *p = 0;
	    if (x[0])
	      idx->Send(FILESYS->Lang(0x040, "s", x));
	    x = p + 1;
	 }
	 if (x[0])
	   idx->Send(FILESYS->Lang(0x040, "s", x));
      }
      cnt++;
   }
   idx->Send(FILESYS->Lang(!is ? 0x041 : (!cnt ? 0x042 : 
					(cnt > 2 ? 0x043 : 0x044)), "d", cnt));
   FILEDB->Close(f);
   context;
   return EXECUTED;
}

static BindReturn cmd_get (void ** args, const char *) {
   FUNCTION;
   DccFiles * idx = (DccFiles *)args[0];
   Bot * ptr = 0;
   int ok = 0;
   char *p, *what, destdir[DIRLEN + 1], s[DIRLEN + 1], *par = (char *)args[1];
   char whoto[NICKMAX + 1];
   filedb fdb;
   FILE *f;
   long where = 0;
   TandemBot * tb;
   
   if (!par[0]) 
     idx->Send(FILESYS->Lang(0x045, ""));
   else {
      what = G->Split(&par);
      if (strlen(par) > NICKMAX)
	idx->Send(FILESYS->Lang(0x046, ""));
      else {
	 strcpy(whoto, par[0] ? par : idx->Nick());
	 if ((p = strrchr(what, '/'))) {
	    *p = 0;
	    strncpy(s, what, 120);
	    s[120] = 0;
	    strcpy(what, p + 1);
	    if (!resolve_dir(idx->Dir(), s, destdir, idx)) {
	       idx->Send(FILESYS->Lang(0x035, ""));
	       return EXECUTED;
	    }
	 } else {
	    strncpy(destdir, idx->Dir(), DIRLEN);
	    destdir[DIRLEN] = 0;
	 }
	 for (f = FILEDB->Open(destdir,0); 
	      FILESYS->FindMatch(f, what, &where, &fdb);
	      where += sizeof(filedb)) {
	    if (!(fdb.stat & (FILE_HIDDEN | FILE_DIR))) {
	       ok = 1;
	       if (fdb.sharelink[0]) {
		  char bot[121];
		  /* this is a link to a file on another bot... */
		  G->SplitC(bot, fdb.sharelink, ':');
#ifdef HAVE_BOTNET_H
		  if (!strcasecmp(bot, G->botnetnick) ||
		      !FILESYS->BOTNET || !FILESYS->BOTNET->tandbot->Find(bot) )
#endif
		    idx->Send(FILESYS->Lang(0x047, "s",fdb.filename));
#ifdef HAVE_BOTNET_H
		  else {
		     G->SPrintF(s,255,"r%d:%s %s",idx->FD(),whoto,fdb.sharelink);
		     if ((tb = FILESYS->BOTNET->tandbot->Find(bot))
			 && (ptr = tb->NextBot()))
		       ptr->SendZapf(G->botnetnick, "filesys", s);
		     idx->Send(FILESYS->Lang(0x048, "ss", fdb.sharelink, bot));
		     /* increase got count now (or never) */
		     fdb.gots++;
		     G->SPrintF(s, DIRLEN, "%s:%s", bot, fdb.sharelink);
		     strcpy(fdb.sharelink, s);
		     fseek(f, where, SEEK_SET);
		     fwrite(&fdb, sizeof(filedb), 1, f);
		  }
#endif
	       } else 
		 do_dcc_send(idx, destdir, fdb.filename, whoto);
	       /* don't increase got count till later */
	    }
	 }
	 FILEDB->Close(f);
	 if (!ok)
	   idx->Send(FILESYS->Lang(0x042, ""));
	 else
	   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x049, "sss", idx->Nick(), what, whoto));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_file_help (void ** args, const char *) {
   FUNCTION;
   DccFiles * idx = (DccFiles *)args[0];
   char s[DIRLEN+1], *par = (char *)args[1];
   FlagRecord flags ( idx->User(), FlagRecord::GLOBAL | FlagRecord::CHANNEL,
		     idx->ConsoleChan());

   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x04A, "ss", idx->Nick(), par));
   if (par[0]) {
      G->SPrintF(s, DIRLEN, "filesys/%s", par);
      G->help_list->Tell(idx, s, flags,0);
   } else 
     G->help_list->Tell(idx, "filesys/help", flags,0);
   return EXECUTED;
}

static BindReturn cmd_hideshare (void ** args, const char * mask) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   int mode = (mask[0] == 'u' ? 1 : 0) + (mask[2] == 's' ? 2 : 0);
   int ok = 0, lang = 0x04B + (mode * 4);
   DccFiles * idx = (DccFiles *)args[0];
   char * par = (char *)args[1];
   
   if (!par[0]) 
     idx->Send(FILESYS->Lang(lang, ""));
   else {
      where = 0L;
      for (f = FILEDB->Open(idx->Dir(),0); 
	   FILESYS->FindMatch(f, par, &where, &fdb);
	   where += sizeof(filedb)) {
	 if (!mode && !(fdb.stat & FILE_HIDDEN)) 
	   fdb.stat |= FILE_HIDDEN;
	 else if ((mode == 1) && (fdb.stat & FILE_HIDDEN)) 
	   fdb.stat &= ~FILE_HIDDEN;
	 else if ((mode == 2) && !(fdb.stat & FILE_SHARE))
	   fdb.stat |= FILE_SHARE;
	 else if ((mode == 3) && (fdb.stat & FILE_SHARE))
	   fdb.stat &= ~FILE_SHARE;
	 else
	   continue;
	 ok++;
	 idx->Send(FILESYS->Lang(lang + 1, "s", fdb.filename));
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
      }
      FILEDB->Close(f);
      if (!ok)
	idx->Send(FILESYS->Lang(0x042, ""));
      else {
	 FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(lang + 2, "ss", idx->Nick(), par));
	 if (ok > 1)
	   idx->Send(FILESYS->Lang(lang + 3, "d", ok));
      }
   }
   return EXECUTED;
}

/* link a file from another bot */
static BindReturn cmd_ln (void ** args, const char *) {
   FUNCTION;
   char *share, newpath[121], newfn[81], *p, *par = (char *)args[1];
   FILE *f;
   filedb fdb;
   long where = 0;
   DccFiles * idx = (DccFiles *)args[0];
   
   share = G->Split(&par);
   if (strlen(share) > 60) 
     share[60] = 0;
   /* correct format? */
   if (!strchr(share, ':') || !par[0]) 
     idx->Send(FILESYS->Lang(0x05B, ""));
   else {
      if ((p = strrchr(par, '/'))) {
	 *p = 0;
	 strncpy(newfn, p + 1, 80);
	 newfn[80] = 0;
	 if (!resolve_dir(idx->Dir(), par, newpath, idx)) {
	    idx->Send(FILESYS->Lang(0x035, ""));
	    return EXECUTED;
	 }
      } else {
	 strncpy(newpath, idx->Dir(), 121);
	 newpath[121]=0;
	 strncpy(newfn, par, 80);
	 newfn[80] = 0;
      }
      f = FILEDB->Open(newpath, 0);
      if (FILESYS->FindMatch(f, newfn, &where, &fdb)) {
	 if (!fdb.sharelink[0]) 
	   idx->Send(FILESYS->Lang(0x05C, "s", newfn));
	 else {
	    strcpy(fdb.sharelink, share);
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb, sizeof(filedb), 1, f);
	    idx->Send(FILESYS->Lang(0x05D, "s", share));
	    FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x05E, "sss", idx->Nick(), par, share));
	 }
      } else {
	 /* new entry */
	 where = FILESYS->FindEmpty(f);
	 G->memset(&fdb, 0, sizeof(filedb));
	 fdb.version = FILEVERSION;
	 strncpy(fdb.filename, newfn, 30);
	 fdb.filename[30] = 0;
	 strncpy(fdb.uploader, idx->Nick(), HANDLEN);
	 fdb.uploader[HANDLEN] = 0;
	 fdb.uploaded = G->now;
	 strcpy(fdb.sharelink, share);
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
	 idx->Send(FILESYS->Lang(0x05F, "ss", fdb.filename, share));
	 FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x060, "sssss", idx->Nick(), newpath,
				newpath[0] ? "/" : "", newfn, share));
      }
      FILEDB->Close(f);
   }
   return EXECUTED;
}

static BindReturn cmd_desc (void ** args, const char *) {
   FUNCTION;
   char *fn, *par = (char *)args[1];
   int ok = 0, lin = 0, i, l = 0;
   DccFiles * idx = (DccFiles *)args[0];
   FILE *f;
   filedb fdb;
   long where = 0;
   FlagRecord flags (idx->User());
   
   if (!(fn = G->Split(&par))[0]) 
     idx->Send(FILESYS->Lang(0x061, ""));
   else {
      for (f = FILEDB->Open(idx->Dir(),0); 
	   FILESYS->FindMatch(f, fn, &where, &fdb);
	   where += sizeof(filedb)) 
	if (!(fdb.stat & FILE_HIDDEN)) {
	   ok = 1;
	   if (!flags.GlobJanitor() && !strcasecmp(fdb.uploader, idx->Nick()))
	     idx->Send(FILESYS->Lang(0x062, ""));
	   else {
	      for (i = 0; *par && i < 185; par++) {
		 if (*par == '|') {
		    if (++lin == 5)
		      break;
		    l = 0;
		    fdb.desc[i] = '\n';
		 } else if (i || (*par > 32)) {
		    fdb.desc[i] = *par;
		    if (++l == 60) {
		       for (l = 0; (l < 60) && (fdb.desc[i - l] != ' '); l--);
		       if (l == 60)
			 l = 0;
		       fdb.desc[i - l] = '\n';
		       if (++lin == 5)
			 break;
		    }
		 }
	      }
	      while (i && (fdb.desc[i - 1] == '\n'))
		i--;
	      fdb.desc[i] = 0;
	      fseek(f, where, SEEK_SET);
	      fwrite(&fdb, sizeof(filedb), 1, f);
	      idx->Send(FILESYS->Lang(par[0] ? 0x063 : 0x064,"s", fdb.filename));
	   }
	}
      FILEDB->Close(f);
      if (!ok)
	idx->Send(FILESYS->Lang(0x042, ""));
      else
	FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x065, "ss", idx->Nick(), fn));
   }
   return EXECUTED;
}

static BindReturn cmd_rm (void ** args, const char *) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   int ok = 0;
   DccFiles * idx = (DccFiles *)args[0];
   char s[DIRLEN+1], *par = (char *)args[1];
   
   if (!par[0]) 
     idx->Send(FILESYS->Lang(0x066, ""));
   else {
      for (f = FILEDB->Open(idx->Dir(), 0); 
	   FILESYS->FindMatch(f, par, &where, &fdb);
	   where += sizeof(filedb))	   
	if (!(fdb.stat & (FILE_HIDDEN | FILE_DIR))) {
	   G->SPrintF(s, DIRLEN, "%s%s/%s", FILESYS->dccdir, idx->Dir(),fdb.filename);
	   fdb.stat |= FILE_UNUSED;
	   ok++;
	   fseek(f, where, SEEK_SET);
	   fwrite(&fdb, sizeof(filedb), 1, f);
	   /* shared file links won't be able to be unlinked */
	   if (!(fdb.sharelink[0]))
	     unlink(s);
	   idx->Send(FILESYS->Lang(0x067, "s", fdb.filename));
	}
      FILEDB->Close(f);
      if (!ok)
	idx->Send(FILESYS->Lang(0x042, ""));
      else {
	 FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x068, "ss", idx->Nick(), par));
	 if (ok > 1)
	   idx->Send(FILESYS->Lang(0x069, "d", ok));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_mkdir (void ** args, const char *) {
   FUNCTION;
   char *name, *flags, *chan, s[DIRLEN+1], * par = (char *)args[1];
   FILE *f;
   filedb fdb;
   long where = 0;
   DccFiles * idx = (DccFiles *)args[0];
   FlagRecord fr (FlagRecord::CHANNEL | FlagRecord::CHANNEL );
   
   if (!par[0]) 
     idx->Send(FILESYS->Lang(0x06A, ""));
   else {
      name = G->Split(&par);
      if  (name[strlen(name) - 1] == '/')
	name[strlen(name) - 1] = 0;
      flags = G->Split(&par);
      chan = G->Split(&par);
      if (!chan[0] && ((flags[0] == '#') || (flags[0] == '&'))) {
	 chan = flags;
	 flags = par;
      }
      if (strchr(name, '/')) 
	idx->Send(FILESYS->Lang(0x0D3, ""));
      else if (chan[0] && !(G->chanset && G->chanset->Find(chan)))
	idx->Send(FILESYS->Lang(0x06B, ""));
      else {
	 f = FILEDB->Open(idx->Dir(), 0);
	 if (!FILESYS->FindMatch(f, name, &where, &fdb)) {
	    G->SPrintF(s, DIRLEN, "%s%s/%s", FILESYS->dccdir, idx->Dir(), name);
	    if (mkdir(s, 0755)) {
	       idx->Send(FILESYS->Lang(0x104, ""));
	       FILEDB->Close(f);
	       return EXECUTED;
	    } 
	    G->memset(&fdb, 0, sizeof(filedb));
	    fdb.version = FILEVERSION;
	    fdb.stat = FILE_DIR;
	    strcpy(fdb.filename, name);
	    fdb.uploaded = G->now;
	    idx->Send(FILESYS->Lang(0x06C, "sss", idx->Dir(),
			      idx->Dir()[0] ? "/" : "", name));
	    where = FILESYS->FindEmpty(f);
	 } else if (!(fdb.stat & FILE_DIR)) {
	    idx->Send(FILESYS->Lang(0x031, ""));
	    FILEDB->Close(f);
	    return EXECUTED;
	 }
	 if (flags[0]) {
	    fr.BreakDown(flags);
	    fr.Build(s);
	    strncpy(fdb.flags_req, s, 21);
	    fdb.flags_req[21] = 0;
	    idx->Send(FILESYS->Lang(0x06D, "ss", name,s));
	 } else if (!chan[0]) {
	    fdb.flags_req[0] = 0;
	    idx->Send(FILESYS->Lang(0x06E, "s", name));
	 }
	 if (chan[0]) {
	    strncpy(fdb.chname,chan,80);
	    fdb.chname[80] = 0;
	    idx->Send(FILESYS->Lang(0x06F, "s", chan));
	 } else if (!flags[0]) {
	    fdb.chname[0] = 0;
	    idx->Send(FILESYS->Lang(0x070, ""));
	 }
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
	 FILEDB->Close(f);
	 FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x071, "ssss", idx->Nick(), name,
				flags, chan));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_rmdir (void ** args, const char *) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   char s[DIRLEN+1], name[80], * par = (char *)args[1];
   DccFiles * idx = (DccFiles *)args[0];
   int len = strlen(name);
   
   if (len > 80)
     len = 80;
   while (name[len - 1] == '/')
     len--;
   strncpy(name, par, len);
   name[len] = 0;
   f = FILEDB->Open(idx->Dir(),0);
   if (strchr(name, '/')) 
     idx->Send(FILESYS->Lang(0x0D4, ""));
   else if (!FILESYS->FindMatch(f, name, &where, &fdb) || !(fdb.stat & FILE_DIR)) 
     idx->Send(FILESYS->Lang(0x031, ""));
   else {
      G->SPrintF(s, DIRLEN, "%s%s/%s/.filedb", FILESYS->dccdir, idx->Dir(), name);
      unlink(s);
      G->SPrintF(s, DIRLEN, "%s%s/%s", FILESYS->dccdir, idx->Dir(), name);
      if (!rmdir(s)) {
	 idx->Send(FILESYS->Lang(0x072, "sss", idx->Dir(),
					idx->Dir()[0] ? "/" : "", name));
	 fdb.stat |= FILE_UNUSED;
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
	 FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x073, "sss", idx->Nick(), name));
      } else 
	idx->Send(FILESYS->Lang(0x104, ""));
   }
   FILEDB->Close(f);
   return EXECUTED;
}

static BindReturn cmd_mv_cp (void ** args, const char * mask) {
   FUNCTION;
   char *p, *fn, oldpath[DIRLEN+1], s[DIRLEN+1], s1[DIRLEN+1];
   char newfn[DIRLEN+1], newpath[DIRLEN+1];
   int ok, only_first = 0, skip_this, ret, copy = mask[0] == 'c';
   FILE *f, *g;
   filedb fdb, z;
   long where, gwhere, wherez;
   char * par = (char *)args[1];
   DccFiles * idx = (DccFiles *)args[0];
   int lang = copy ? 0x078 : 0x07E;
   
   fn = G->Split(&par);
   if (!par[0]) 
     idx->Send(FILESYS->Lang(lang, ""));
   else {
      if ((p = strrchr(fn, '/'))) {
	 *p = 0;
	 strncpy(s, fn, DIRLEN);
	 s[DIRLEN] = 0;
	 fn = p+1;
	 if (!resolve_dir(idx->Dir(), s, oldpath, idx)) {
	    idx->Send(FILESYS->Lang(0x074, ""));
	    return EXECUTED;
	 }
      } else
	strcpy(oldpath, idx->Dir());
      strncpy(s, par, DIRLEN);
      s[DIRLEN] = 0;
      if (!resolve_dir(idx->Dir(), s, newpath, idx)) {
	 /* destination is not just a directory */
	 if (!(p = strrchr(s, '/'))) {
	    strcpy(newfn, s);
	    s[0] = 0;
	 } else {
	    *p = 0;
	    strcpy(newfn, p + 1);
	 }
	 if (!resolve_dir(idx->Dir(), s, newpath, idx)) {
	    idx->Send(FILESYS->Lang(0x075, ""));
	    return EXECUTED;
	 }
      } else
	newfn[0] = 0;
      /* stupidness checks */
      if (!strcmp(oldpath, newpath) && (!newfn[0] || !strcmp(newfn, fn))) {
	 idx->Send(FILESYS->Lang(lang + 1, ""));
	 return EXECUTED;
      }
      /* be aware of 'cp * this.file' possibility: ONLY COPY FIRST ONE */
      if ((strchr(fn, '?') || strchr(fn, '*')) && newfn[0])
	only_first = 1;
      f = FILEDB->Open(oldpath,0);
      g = strcmp(oldpath, newpath) ? FILEDB->Open(newpath,0) : f;
      where = 0L;
      ok = 0;
      for (ret = FILESYS->FindMatch(f, fn, &where, &fdb); ret; 
	   ret = (ok && only_first) ? 0 : FILESYS->FindMatch(f, fn, &where, &fdb)) {
	 skip_this = 0;
	 if (!(fdb.stat & (FILE_HIDDEN | FILE_DIR))) {
	    G->SPrintF(s, DIRLEN, "%s%s%s%s", FILESYS->dccdir, oldpath,
		       oldpath[0] ? "/" : "", fdb.filename);
	    G->SPrintF(s1, DIRLEN, "%s%s%s%s", FILESYS->dccdir, newpath, 
		    newpath[0] ? "/" : "", newfn[0] ? newfn : fdb.filename);
	    if (!strcmp(s, s1)) {
	       idx->Send(FILESYS->Lang(lang + 2,"sss", 
				 newpath,newpath[0] ? "/" : "",
				 newfn[0] ? newfn : fdb.filename));
	       skip_this = 1;
	    }
	    /* check for existence of file with same name in new dir */
	    wherez = 0;
	    if (FILESYS->FindMatch(g, newfn[0] ? newfn : fdb.filename, &wherez, &z)) {
	       /* it's ok if the entry in the new dir is a normal file (we'll */
	       /* just scrap the old entry and overwrite the file) -- but if */
	       /* it's a directory, this file has to be skipped */
	       if (z.stat & FILE_DIR) {
		  /* skip */
		  skip_this = 1;
		  idx->Send(FILESYS->Lang(0x076, "sss", newpath, 
				    newpath[0] ? "/" : "", 
				    newfn[0] ? newfn : fdb.filename));
	       } else {
		  z.stat |= FILE_UNUSED;
		  fseek(g, wherez, SEEK_SET);
		  fwrite(&z, sizeof(filedb), 1, g);
	       }
	    }
	    if (!skip_this) {
	       if ((fdb.sharelink[0]) || (G->CopyFile(s, s1) == 0)) {
		  /* raw file moved okay: create new entry for it */
		  ok++;
		  gwhere = FILESYS->FindEmpty(g);
		  z.version = FILEVERSION;
		  z.stat = fdb.stat;
		  strcpy(z.flags_req, fdb.flags_req);
		  strcpy(z.chname, fdb.chname);
		  strcpy(z.filename, fdb.filename);
		  strcpy(z.desc, fdb.desc);
		  if (newfn[0])
		    strcpy(z.filename, newfn);
		  strcpy(z.uploader, fdb.uploader);
		  z.uploaded = fdb.uploaded;
		  z.size = fdb.size;
		  z.gots = fdb.gots;
		  strcpy(z.sharelink, fdb.sharelink);
		  fseek(g, gwhere, SEEK_SET);
		  fwrite(&z, sizeof(filedb), 1, g);
		  if (!copy) {
		     unlink(s);
		     fdb.stat |= FILE_UNUSED;
		     fseek(f, where, SEEK_SET);
		     fwrite(&fdb, sizeof(filedb), 1, f);
		  }
		  idx->Send(FILESYS->Lang(lang + 3,"ssssss", oldpath, 
				    oldpath[0] ? "/" : "", fdb.filename,
				    newpath, newpath[0] ? "/" : "",
				    newfn[0] ? newfn : fdb.filename));
	       } else
		 idx->Send(FILESYS->Lang(0x077, "sss", newpath,
				   newpath[0] ? "/" : "", 
				   newfn[0] ? newfn : fdb.filename));
	    }
	 }
	 where += sizeof(filedb);
      }
      if (!ok)
	idx->Send(FILESYS->Lang(0x042, ""));
      else {
	 FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(lang + 4, "ssssss", idx->Nick(),
				copy ? "cp" : "mv", oldpath, 
				oldpath[0] ? "/" : "", fn, par));
	 if (ok > 1)
	   idx->Send(FILESYS->Lang(lang + 5, "d", ok));
      }
      FILEDB->Close(f);
      if (g != f)
	FILEDB->Close(g);
   }
   return EXECUTED;
}

static void tell_file_stats (Socket * idx, const char * hand) {
   FUNCTION;
   UserRecord *u;
   UserEntryFstat * fs = 0;
   
   float fr = (-1.0), kr = (-1.0);
   if (!(u = G->userlist->Find(hand)) 
       || !(fs = (UserEntryFstat *)u->GetEntry(FILESYS->entrytype_fstat))) 
     idx->Send(FILESYS->Lang(0x084, "s", hand));
   else {
      idx->Send(FILESYS->Lang(0x085, "46", fs->Uploads(), fs->UploadKs()));
      idx->Send(FILESYS->Lang(0x086, "46", fs->Dnloads(), fs->DnloadKs()));
      if (fs->Uploads())
	fr = ((float) fs->Dnloads() / (float) fs->Uploads());
      if (fs->UploadKs())
	kr = ((float) fs->DnloadKs() / (float) fs->UploadKs());
      idx->Send(FILESYS->Lang((fr < 0.0) ? 0x087 : 0x088, "f", fr));
      idx->Send(FILESYS->Lang((kr < 0.0) ? 0x089 : 0x08A, "f", kr));
   }
}

static BindReturn cmd_stats (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x08B, "s", idx->Nick()));
   tell_file_stats(idx, idx->Nick());
   return EXECUTED;
}

static BindReturn cmd_filestats (void ** args, const char *) {
   FUNCTION;
   char *nick, * par = (char *)args[1];
   Socket * idx = (Socket *)args[0];
   UserRecord * u;
   FlagRecord flags (idx->User());
   
   context;
   if (!par[0]) 
     idx->Send(FILESYS->Lang(0x08C, ""));
   else {
      nick = G->Split(&par);
      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(0x08D, "ss", idx->Nick(), nick));
      if (!nick[0])
	tell_file_stats(idx, idx->Nick());
      else if (!(u = G->userlist->Find(nick)))
	idx->Send(G->Lang(0x100, ""));
      else if (!strcmp(par, "clear") && flags.GlobJanitor()) {
	 UserEntryFstat * fs = GetFstat ( idx->User() );
	 fs->Clear();
      } else
	tell_file_stats(idx, nick);
   }
   return EXECUTED;
}

static BindReturn cmd_quit ( void ** , const char * ) {
   FUNCTION;
   return EXEC_BRK;
}

#define COMMANDS 25
static command_binds myfiles[COMMANDS] = {
   {"note", "", 0, 0 },
   {"cancel", "", cmd_cancel, 0 },
   {"cd", "", cmd_chdir, 0 },
   {"chdir", "", cmd_chdir, 0 },
   {"cp", "j", cmd_mv_cp, 0 },
   {"desc", "", cmd_desc, 0 },
   {"filestats", "j", cmd_filestats, 0 },
   {"get", "", cmd_get, 0 },
   {"help", "", cmd_file_help, 0 },
   {"hide", "j", cmd_hideshare, 0 },
   {"ln", "j", cmd_ln, 0 },
   {"ls", "", cmd_ls, 0 },
   {"lsa", "j", cmd_ls, 0 },
   {"mkdir", "j", cmd_mkdir, 0 },
   {"mv", "j", cmd_mv_cp, 0 },
   {"pending", "", cmd_pending, 0 },
   {"pwd", "", cmd_pwd, 0 },
   {"quit", "", cmd_quit, 0 },
   {"rm", "j", cmd_rm, 0 },
   {"rmdir", "j", cmd_rmdir, 0 },
   {"share", "j", cmd_hideshare, 0 },
   {"sort", "j", cmd_sort, 0 },
   {"stats", "", cmd_stats, 0 },
   {"unhide", "j", cmd_hideshare, 0 },
   {"unshare", "j", cmd_hideshare, 0 },
};

static BindReturn cmd_files (void ** args, const char *) {   FUNCTION;

   DccChat * idx = (DccChat *)args[0];
   FlagRecord atr (idx->User());
   
   context;
   if (!FILESYS->dccdir[0])
     idx->Send(FILESYS->Lang(0x0A0, ""));
   else if (FILESYS->TooManyFilers()) {
      idx->Send(FILESYS->Lang(0x0A1, "d", FILESYS->dcc_users));
      idx->Send(FILESYS->Lang(0x0A2, ""));
   } else {
      if (!atr.GlobXfer())
	idx->Send(FILESYS->Lang(0x0A3, ""));
      else {
	 G->LOG_CMDS.Put(FILESYS->Lang(0x0A4, "s", idx->Nick()));
	 idx->Send(FILESYS->Lang(0x0A5, ""));
	 idx->CheckChof(FILESYS->Lang(0x0D1, ""));
	 NEW ( DccFiles ( * idx ) );
      }
   }
   return EXECUTED;
}

static command_binds mydcc[] = {
   {"files", "-", cmd_files, 0 },
};

#ifdef HAVE_SERVER_H
/* received a ctcp-dcc */
static void filesys_dcc_send (char * nick, char * from, UserRecord * u,
			      char * text) {
   FUNCTION;
   char *param, *ip, *prt, buf[512], s1[DIRLEN + 1], xferdir[DIRLEN + 1];
   char s2[DIRLEN + 1], *msg = buf;
   FILE *f;
   FlagRecord atr (u);
   int lang = 0;
   FileSend * ptr = (FileSend *)G->socklist;
   struct in_addr in;
   
   context;
   strcpy(buf,text);
   param = G->Split(&msg);
   if (!atr.GlobXfer())
     lang = 0x0AA;
   else if (!FILESYS->dccin[0] && !FILESYS->upload_to_cd) 
     lang = 0x0AC;
   else if (strchr(param, '/')) 
     lang = 0x0AE;
   else {
      ip = G->Split(&msg);
      prt = G->Split(&msg);
      if (!atoi(msg)) 
	lang = 0x0B0;
      else if (atoi(msg) > (FILESYS->dcc_maxsize * 1024))
	lang = 0x0B2;
      else if (G->max_dcc == G->dcc_total) 
	lang = 0x0B4;
      else {
	 UserEntryDccDir * ued = GetDccDir(u);
	 
	 if (param[0] == '.')
	   param[0] = '_';
	 if (strlen(param) > DIRLEN)
	   param[DIRLEN] = 0;
	 if (FILESYS->upload_to_cd && ued->Value()) 
	   G->SPrintF(xferdir, DIRLEN, "%s%s/", FILESYS->dccdir, ued->Value());
	 else
	   G->SPrintF(xferdir, DIRLEN, "%s", FILESYS->upload_to_cd ? FILESYS->dccdir : FILESYS->dccin);
	 G->SPrintF(s1, DIRLEN, "%s%s", xferdir, param);
	 if ((f = fopen(s1, "r"))) {
	    fclose(f);
	    lang = 0x0B6;
	 } else {	 
	    /* check for dcc-sends in process with the same filename */
	    while ((ptr = ptr->Next()))
	      if (!strcmp(param, ptr->Filename())) {
		 lang = 0x0B8;
		 break;
	      }
	    if (!lang) {
	       /* put uploads in /tmp first */
	       G->SPrintF(s2, DIRLEN, "%s%s", G->tempdir, param);	 
	       if (!(f = fopen(s2, "w")))
		 lang = 0x0BA;
	       else {
		  in.s_addr = atoi(ip);
		  ptr = NEW ( FileSend( &in, atoi(prt), nick, u, atoi(msg),
				       from, s1, f));
		  if (ptr->Flags() & Socket::LOST)
		    lang = 0x0D3;
	       }
	    }
	 }
      }
   }
   if (lang) {
      FILESYS->TRANSFER->LOG_FILES.Put( FILESYS->Lang(lang, "sss", param, nick, from));
      if (!FILESYS->SERVER->quiet_reject)
	FILESYS->SERVER->SendNotice(Q_HELP, nick,  FILESYS->Lang(lang + 1, ""));
   }
}

/* this only handles CHAT requests, otherwise it's handled in 
 * filesys */
static BindReturn  filesys_DCC_CHAT( void ** args, const char *) {
   FUNCTION;
   char * param, * ip, * prt, buf[512], *msg = buf;
   char * nick = (char *)args[0], *from = (char *)args[1];
   char * text = (char *)args[5];
   int lang = 0;
   UserRecord * u = (UserRecord *)args[2];
   FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL | FlagRecord::ANYWHERE );
   struct in_addr in;
   DccFilesPass * newsock;
   
   context;
   if (!FILESYS->SERVER)
     return NOMATCH;
   if (!strncasecmp(text, "SEND ", 5)) {
      filesys_dcc_send(nick, from, u, text + 5);
      return EXEC_LOG;
   }
   if (strncasecmp(text, "CHAT ", 5) || !u)
     return NOMATCH;
   strcpy(buf, text + 5);
   fr.Set(u);
   param = G->Split(&msg);
   text = "FILE";
   if (G->dcc_total == G->max_dcc)
     lang = 0x071;
   else if (fr.GlobParty() || (!G->require_p && fr.ChanOp()))
     return NOMATCH; /* allow ctcp.so to pick up the chat */
   else if (!fr.GlobXfer() || !FILESYS->dccdir[0])
     lang = 0x073;
   else if (u->PassMatch("-")) 
     lang = 0x075;
   else {
      ip = G->Split(&msg);
      prt = G->Split(&msg);
      in.s_addr = atoi(ip);
      newsock = NEW ( DccFilesPass ( in, atoi(prt)));
      if ((newsock->Flags() & Socket::LOST)) {
	 param = strerror(errno);
	 lang = 0x077;
      } else {
	 newsock->SetNick(u->Handle());
	 newsock->SetUser(u);
	 newsock->SetHost(from);
	 newsock->SetStatus(STAT_ECHO);
	 newsock->TimeNow();
	 lang = 0x079;
	 newsock->Send(FILESYS->Lang(0x10A, ""));
      }
   }
   if (lang) {
      FILESYS->TRANSFER->LOG_FILES.Put(FILESYS->SERVER->Lang(lang, "ssss", text, param, nick, from));
      if (!FILESYS->SERVER->quiet_reject && (lang != 0x079))
	FILESYS->SERVER->SendNotice(Q_HELP, nick, FILESYS->SERVER->Lang(lang + 1, "s", 
						     param));
   }
   return EXECUTED;
}

static command_binds myctcp[] = {
     { "DCC", "", filesys_DCC_CHAT, "files:DCC" },
};

static BindReturn init_server_ctcps(void ** args, const char *) {
   FUNCTION;
   FILESYS->SERVER = (ModuleEntryServer *)(args ? args[0]
	: G->module_list->Find("server", &required_server));
   if (!FILESYS->SERVER || !(*FILESYS->SERVER->Vers() <= required_server))
	FILESYS->SERVER = NULL;
   else
	FILESYS->SERVER->bt_ctcp->AddBinds(myctcp,1,FILESYS);
   return EXECUTED;
}

static BindReturn unload_server (void **, const char *) {
   FUNCTION;
   FILESYS->SERVER = NULL;
   return EXECUTED;
}
#endif
   
#if defined(HAVE_SERVER_H) || defined(HAVE_BOTNET_H)
static command_binds myload [] = { 
#ifdef HAVE_SERVER_H
   { "server", "", init_server_ctcps, "filesys:server" },
#endif
#ifdef HAVE_BOTNET_H
   { "botnet", "", load_botnet, "filesys:botnet" },
#endif   
};

static command_binds myunld [] = {
#ifdef HAVE_SERVER_H
   { "server", "", unload_server, "filesys:server" },
#endif
#ifdef HAVE_BOTNET_H
   { "botnet", "", unload_botnet, "filesys:botnet" },
#endif   
};
#endif

static int script_getdesc (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long i = 0;

   context;
   BADARGS(3, 3, FILESYS->Lang(0x0BC, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(),0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &i, &fdb))
	sc->SetString(fdb.desc);
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_setdesc (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   
   context;
   BADARGS(4, 4, FILESYS->Lang(0x0BD, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(),0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &where, &fdb)) {
	 strncpy(fdb.desc, argv[3]->AsString(), 185);
	 fdb.desc[185] = 0;
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
      }
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_getowner (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long i = 0;

   context;
   BADARGS(3, 3, FILESYS->Lang(0x0BC, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(),0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &i, &fdb))
	sc->SetString(fdb.uploader);
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_setowner (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   
   context;
   BADARGS(4, 4, FILESYS->Lang(0x0BE, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(),0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &where, &fdb)) {
	 strncpy(fdb.uploader, argv[3]->AsString(), HANDLEN);
	 fdb.uploader[HANDLEN] = 0;
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
      }
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_getgots (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long i = 0;
   int gots = 0;
   
   context;   
   BADARGS(3, 3, FILESYS->Lang(0x0BC, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(), 0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &i, &fdb))
	gots = fdb.gots;
      FILEDB->Close(f);
   }
   sc->SetInteger(gots);
   return SCRIPT_OK;
}

static int script_setlink (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;

   context;   
   BADARGS(4, 4, FILESYS->Lang(0x0BF, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(), 0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &where, &fdb)) {
	 /* change existing one? */
	 if (!(fdb.stat & FILE_DIR) && fdb.sharelink[0]) {
	    if (!argv[3]->AsString()[0]) 
	       /* erasing file */
	      fdb.stat |= FILE_UNUSED;
	    else {
	       strncpy(fdb.sharelink, argv[3]->AsString(), 60);
	       fdb.sharelink[60] = 0;
	    }
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb, sizeof(filedb), 1, f);
	 }
      } else {
	 G->memset(&fdb, 0, sizeof(filedb));
	 fdb.version = FILEVERSION;
	 strcpy(fdb.uploader, G->botnetnick);
	 strncpy(fdb.filename, argv[2]->AsString(), 30);
	 fdb.uploaded = G->now;
	 strncpy(fdb.sharelink, argv[3]->AsString(), 60);
	 where = FILESYS->FindEmpty(f);
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
      }
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_getlink (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long i = 0;
   
   context;   
   BADARGS(3, 3, FILESYS->Lang(0x0BC, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(), 0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &i, &fdb) && !(fdb.stat & FILE_DIR))
	sc->SetString(fdb.sharelink);
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_setpwd (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccFiles * idx;
   char s[DIRLEN + 1];   
   
   context;   
   BADARGS(3, 3, FILESYS->Lang(0x0C0, ""));
   if (!(idx = (DccFiles *)G->socklist->Find(argv[1]->AsInteger(), FILESYS->DCC_FILES))) {
      sc->SetString(FILESYS->Lang(0x105, ""));
      return SCRIPT_ERROR;
   }
   if (resolve_dir(idx->Dir(), argv[2]->AsString(), s, idx)) {
      UserEntryDccDir * ued = GetDccDir(idx->User());
      
      idx->SetDir(s);
      ued->Set(s);
   }
   return SCRIPT_OK;
}

static int script_getpwd (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccFiles * idx;

   BADARGS(2, 2, FILESYS->Lang(0x0C1, ""));
     if (!(idx = (DccFiles *)G->socklist->Find(argv[1]->AsInteger(), FILESYS->DCC_FILES))) {
      sc->SetString(FILESYS->Lang(0x105, ""));
      return SCRIPT_ERROR;
   }
   sc->SetString(idx->Dir());
   return SCRIPT_OK;
}

static int script_getfiles (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   const char * p;
   
   BADARGS(2, 2, FILESYS->Lang(0x0C2, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(), 0))) {
      rewind(f);
      for (fread(&fdb, sizeof(filedb), 1, f); !feof(f);
	   fread(&fdb, sizeof(filedb), 1, f)) 
	if (!(fdb.stat & (FILE_DIR | FILE_UNUSED))) {
	   p = fdb.filename;
	   sc->AppendArray(1, &p);
	}
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_getdirs (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   const char * p;
   
   BADARGS(2, 2, FILESYS->Lang(0x0C2, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(), 0))) {
      rewind(f);
      for (fread(&fdb, sizeof(filedb), 1, f); !feof(f);      
	   fread(&fdb, sizeof(filedb), 1, f)) 
	if ((!(fdb.stat & FILE_UNUSED)) && (fdb.stat & FILE_DIR)) {
	   p = fdb.filename;
	   sc->AppendArray(1, &p);
	}
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_change (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   const char * cmd = argv[0]->AsString();
   
   BADARGS(3, 3, FILESYS->Lang(0x0BC, ""));
   if ((f = FILEDB->Open(argv[1]->AsString(),0))) {
      if (FILESYS->FindMatch(f, argv[2]->AsString(), &where, &fdb) && !(fdb.stat & FILE_DIR)) {
	 if (cmd[0] == 's') /* share */
	   fdb.stat |= FILE_SHARE;
	 else if (cmd[0] == 'h') /* hide */
	   fdb.stat |= FILE_HIDDEN;
	 else if (cmd[2] == 's') /* unshare */
	   fdb.stat &= ~FILE_SHARE;
	 else if (cmd[2] == 'h') /* unhide */
	   fdb.stat &= ~FILE_HIDDEN;
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
      }
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_setflags (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   char s[DIRLEN+1], *p, *d, *r = "-1";
   
   BADARGS(3, 4, FILESYS->Lang(0x0C3, ""));
   strncpy(s, argv[1]->AsString(), DIRLEN);
   s[DIRLEN] = 0;
   if (s[strlen(s) - 1] == '/')
       s[strlen(s) - 1] = 0;
   if (!(p = strrchr(s, '/'))) {
      p = s;
      d = "";
   } else {
      *p++ = 0;
      d = s;
   }
   if ((f = FILEDB->Open(d,0))) {
      if (FILESYS->FindMatch(f, p, &where, &fdb) && (fdb.stat & FILE_DIR)) {
	 if (argc >= 3) {
	    FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL );
	    
	    fr.BreakDown(argv[2]->AsString());
	    fr.Build(s);
	    strncpy(fdb.flags_req, s, 21);
	    fdb.flags_req[21] = 0;
	 } else
	   fdb.flags_req[0] = 0;
	 if (argc == 4) {
	    strncpy(fdb.chname,argv[3]->AsString(),80);
	    fdb.chname[80] = 0;
	 } 
	 r = "0";
      }
      fseek(f, where, SEEK_SET);
      fwrite(&fdb, sizeof(filedb), 1, f);
      FILEDB->Close(f);
   }
   sc->SetString(r);
   return SCRIPT_OK;
}

static int script_getflags (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   char s[DIRLEN+1], *p, *d;
   const char * list [2];
   
   BADARGS(2, 2, FILESYS->Lang(0x0C2, ""));
   strncpy(s, argv[1]->AsString(), DIRLEN);
   s[DIRLEN] = 0;
   if (s[strlen(s) - 1] == '/')
     s[strlen(s) - 1] = 0;
   if (!(p = strrchr(s, '/'))) {
      p = s;
      d = "";
   } else {
      *p++ = 0;
      d = s;
   }
   if ((f = FILEDB->Open(d, 0))) {
      if (FILESYS->FindMatch(f, p, &where, &fdb) && (fdb.stat & FILE_DIR)) {
	 strcpy(s, fdb.flags_req);
	 if (s[0] == '-')
	   s[0] = 0;
	 list[0] = s;
	 list[1] = fdb.chname;
	 sc->AppendArray(2, list);
      }
      FILEDB->Close(f);
   }
   return SCRIPT_OK;
}

static int script_mkdir (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char s[DIRLEN+1], t[DIRLEN+1], *d, *p, *r = NULL;
   FILE *f;
   filedb fdb;
   long where = 0;
   FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   
   BADARGS(3, 4, FILESYS->Lang(0x0C3, ""));
   strncpy(s, argv[1]->AsString(), DIRLEN);
   s[DIRLEN] = 0;
   if (s[strlen(s) - 1] == '/')
       s[strlen(s) - 1] = 0;
   if (!(p = strrchr(s, '/'))) {
      p = s;
      d = "";
   } else {
      *p = 0;
      p++;
      d = s;
   }
   if ((f = FILEDB->Open(d, 0))) {
      if (!FILESYS->FindMatch(f, p, &where, &fdb)) {
	 G->SPrintF(t, DIRLEN, "%s%s/%s", FILESYS->dccdir, d, p);
	 if (mkdir(t, 0755) != 0) 
	   r = "1";
	 else {
	    G->memset(&fdb, 0, sizeof(filedb));
	    fdb.version = FILEVERSION;
	    fdb.stat = FILE_DIR;
	    strcpy(fdb.filename, argv[1]->AsString());
	    fdb.uploaded = G->now;
	    where = FILESYS->FindEmpty(f);
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb, sizeof(filedb), 1, f);
	 }
      } else if (!(fdb.stat & FILE_DIR))
	   r = "2";
      if (!r) {
	 if (argc >= 3) {
	    fr.BreakDown(argv[2]->AsString());
	    fr.Build(s);
	    strncpy(fdb.flags_req, s, 21);
	    fdb.flags_req[21] = 0;
	 } else 
	   fdb.flags_req[0] = 0;
	 if (argc == 4) {
	    strncpy(fdb.chname,argv[3]->AsString(),80);
	    fdb.chname[80] = 0;
	 } else
	   fdb.chname[0] = 0;
	 r = "0";
	 fseek(f, where, SEEK_SET);
	 fwrite(&fdb, sizeof(filedb), 1, f);
      }
      FILEDB->Close(f);
   } else
     r = "1";
   sc->SetString(r);
   return SCRIPT_OK;
}

static int script_rmdir (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   filedb fdb;
   long where = 0;
   char s[DIRLEN+1], t[DIRLEN+1], *d, *p, *r = "1";
   
   BADARGS(2, 2, FILESYS->Lang(0x0C2, ""));
   strncpy(s, argv[1]->AsString(),DIRLEN);
   s[DIRLEN] = 0;
   if (s[strlen(s) - 1] == '/')
       s[strlen(s) - 1] = 0;
   if (!(p = strrchr(s, '/'))) {
      p = s;
      d = "";
   } else {
      *p++ = 0;
      d = s;
   }
   if ((f = FILEDB->Open(d,0))) {
      if  (FILESYS->FindMatch(f, p, &where, &fdb) && (fdb.stat & FILE_DIR)) {
	 G->SPrintF(t, DIRLEN, "%s%s/%s/.filedb", FILESYS->dccdir, d, p);
	 unlink(t);
	 G->SPrintF(t, DIRLEN, "%s%s/%s", FILESYS->dccdir, d, p);
	 if (!rmdir(t)) {
	    fdb.stat |= FILE_UNUSED;
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb, sizeof(filedb), 1, f);
	    r = "0";
	 }
      }
      FILEDB->Close(f);
   }
   sc->SetString(r);
   return SCRIPT_OK;
}

static int script_mv_cp (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char *p, fn[DIRLEN+1], oldpath[DIRLEN+1], s[DIRLEN+1], s1[DIRLEN+1],
     newfn[DIRLEN+1], newpath[DIRLEN+1];
   int ok, only_first, skip_this, ret, copy = (argv[0]->AsString()[0] == 'c');
   FILE *f, *g;
   filedb fdb, x;
   long where, gwhere, wherez;

   BADARGS(3, 3, FILESYS->Lang(0x0C4, ""));
   context;
   strncpy(fn, argv[1]->AsString(),DIRLEN);
   fn[DIRLEN] = 0;
   if ((p = strrchr(fn, '/'))) {
      *p = 0;
      strcpy(s, fn);
      strcpy(fn, p + 1);
      if (!resolve_dir("/", s, oldpath, NULL)) {
	 /* scripts can do * anything */
	 sc->SetString("-1");
	 return SCRIPT_OK;
      }
   } else
      strcpy(oldpath, "/");
   strncpy(s, argv[2]->AsString(), DIRLEN);
   s[DIRLEN] = 0;
   if (!resolve_dir("/", s, newpath, NULL)) {
      /* destination is not just a directory */
      if (!(p = strrchr(s, '/'))) {
	 strcpy(newfn, s);
	 s[0] = 0;
      } else {
	 *p = 0;
	 strcpy(newfn, p + 1);
      }
      if (!resolve_dir("/", s, newpath, NULL)) {
	 sc->SetString("-2");
	 return SCRIPT_OK;
      }
   } else
      newfn[0] = 0;
   /* stupidness checks */
   if (!strcmp(oldpath, newpath) && (!newfn[0] || !strcmp(newfn, fn))) {
      sc->SetString("-3");
      return SCRIPT_OK;
   }
   /* be aware of 'cp * this.file' possibility: ONLY COPY FIRST ONE */
   only_first =  ((strchr(fn, '?') || strchr(fn, '*')) && newfn[0]);
   f = FILEDB->Open(oldpath, 0);
   g = !strcmp(oldpath, newpath) ? f : FILEDB->Open(newpath, 0);
   where = 0L;
   ok = 0;
   for (ret = FILESYS->FindMatch(f, fn, &where, &fdb); ret; 
	ret = (ok && only_first) ? 0 : FILESYS->FindMatch(f, fn, &where, &fdb)) {
      skip_this = 0;
      if (!(fdb.stat & (FILE_HIDDEN | FILE_DIR))) {
	 G->SPrintF(s, DIRLEN, "%s%s%s%s", FILESYS->dccdir, oldpath,
		    oldpath[0] ? "/" : "", fdb.filename);
	 G->SPrintF(s1, DIRLEN, "%s%s%s%s", FILESYS->dccdir, newpath,
		    newpath[0] ? "/" : "", newfn[0] ? newfn : fdb.filename);
	 if (!strcmp(s, s1)) {
	    sc->SetString("-3");
	    skip_this = 1;
	 }
	 /* check for existence of file with same name in new dir */
	 wherez = 0;
	 if (FILESYS->FindMatch(g, newfn[0] ? newfn : fdb.filename, &wherez, &x)) {
	    /* it's ok if the entry in the new dir is a normal file (we'll */
	    /* just scrap the old entry and overwrite the file) -- but if */
	    /* it's a directory, this file has to be skipped */
	    if (x.stat & FILE_DIR) 
	      /* skip */
	      skip_this = 1;
	    else {
	       x.stat |= FILE_UNUSED;
	       fseek(g, wherez, SEEK_SET);
	       fwrite(&x, sizeof(filedb), 1, g);
	    }
	 }
	 if (!skip_this && (fdb.sharelink[0] || !G->CopyFile(s, s1))) {
	    /* raw file moved okay: create new entry for it */
	    ok++;
	    gwhere = FILESYS->FindEmpty(g);
	    x.version = FILEVERSION;
	    x.stat = fdb.stat;
	    strcpy(x.flags_req, fdb.flags_req);
	    strcpy(x.chname, fdb.chname);
	    strcpy(x.filename, fdb.filename);
	    strcpy(x.desc, fdb.desc);
	    if (newfn[0])
	      strcpy(x.filename, newfn);
	    strcpy(x.uploader, fdb.uploader);
	    x.uploaded = fdb.uploaded;
	    x.size = fdb.size;
	    x.gots = fdb.gots;
	    strcpy(x.sharelink, fdb.sharelink);
	    fseek(g, gwhere, SEEK_SET);
	    fwrite(&x, sizeof(filedb), 1, g);
	    if (!copy) {
	       unlink(s);     
	       fdb.stat |= FILE_UNUSED;
	       fseek(f, where, SEEK_SET);
	       fwrite(&fdb, sizeof(filedb), 1, f);
	    }
	 }
      }
      where += sizeof(filedb);
   }
   sc->SetString(!ok ? "-4" : G->ToBase10(ok));
   FILEDB->Close(f);
   if (g != f)
      FILEDB->Close(g);
   return SCRIPT_OK;
}

static int script_filesend (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccFiles * idx;
   int ret = 0;
   char *p, what[DIRLEN + 1], destdir[DIRLEN + 1], s[DIRLEN + 1];
   filedb fdb;
   FILE *f;
   long where = 0;
#ifdef HAVE_BOTNET_H
   Bot * ptr = 0;
   TandemBot * tb;
#endif
   
   BADARGS(3, 4, FILESYS->Lang(0x0C5, ""));
   if (!(idx = (DccFiles *)G->socklist->Find(argv[1]->AsInteger(),FILESYS->DCC_FILES))) {
      sc->SetString(FILESYS->Lang(0x105, ""));
      return SCRIPT_ERROR;
   }   
   if ((p = strrchr(argv[2]->AsString(), '/'))) {
      *p = 0;
      strncpy(s, argv[2]->AsString(), DIRLEN);
      s[DIRLEN] = 0;
      strncpy(what, p + 1, DIRLEN);
      what[DIRLEN] = 0;
      if (!resolve_dir(idx->Dir(), s, destdir, idx)) {
	 sc->SetString("0");
	 return SCRIPT_OK;
      }
   } else {
      strncpy(destdir, idx->Dir(), DIRLEN);
      destdir[DIRLEN] = 0;
      strncpy(what, argv[2]->AsString(), DIRLEN);
      what[DIRLEN] = 0;
   }
   f = FILEDB->Open(destdir,0);
   if (FILESYS->FindMatch(f, what, &where, &fdb) && 
       !(fdb.stat & (FILE_HIDDEN | FILE_DIR))) {
      if (fdb.sharelink[0]) {
	 char bot[121], whoto[NICKMAX + 1];
	 /* this is a link to a file on another bot... */
	 G->SplitC(bot, fdb.sharelink, ':');
#ifdef HAVE_BOTNET_H
	 if (strcasecmp(bot, G->botnetnick) &&
	     FILESYS->BOTNET && FILESYS->BOTNET->tandbot->Find(bot)) {
	    strncpy(whoto, (argc == 3) ? idx->Nick() : argv[3]->AsString(), NICKMAX);
	    whoto[NICKMAX] = 0;
	    G->SPrintF(s, 255, "r%d:%s %s", idx->FD(), whoto, fdb.sharelink);
	    if ((tb = FILESYS->BOTNET->tandbot->Find(bot)) && (ptr = tb->NextBot()))
	      ptr->SendZapf (G->botnetnick, "filesys", s);
	    idx->Send(FILESYS->Lang(0x01D, "ss", fdb.sharelink, bot));
	    /* increase got count now (or never) */
	    fdb.gots++;
	    G->SPrintF(s, DIRLEN, "%s:%s", bot, fdb.sharelink);
	    strcpy(fdb.sharelink, s);
	    fseek(f, where, SEEK_SET);
	    fwrite(&fdb, sizeof(filedb), 1, f);
	    FILEDB->Close(f);
	    ret = 1;
	 }
#endif
      } else {
	 do_dcc_send(idx, destdir, fdb.filename, (argc == 3) ? idx->Nick() :
		     argv[3]->AsString());
	 ret = 1;
      }
   }
   FILEDB->Close(f);
   sc->SetInteger(ret);
   return SCRIPT_OK;
}

extern "C" {
   char *filesys_start (ModuleEntryEggdrop *, ModuleEntry *);
};

char *filesys_start (ModuleEntryEggdrop * F, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;

   G = F;
   context;
   FILESYS = NEW( ModuleEntryFilesys ( me, &r ) );
   if (!r) {
      myfiles[0].func = G->cmd_note;
      NEW ( Command ( "getdesc", script_getdesc, FILESYS ) );
      NEW ( Command ( "getowner", script_getowner, FILESYS ) );
      NEW ( Command ( "setdesc", script_setdesc, FILESYS ) );
      NEW ( Command ( "setowner", script_setowner, FILESYS ) );
      NEW ( Command ( "getgots", script_getgots, FILESYS ) );
      NEW ( Command ( "getpwd", script_getpwd, FILESYS ) );
      NEW ( Command ( "setpwd", script_setpwd, FILESYS ) );
      NEW ( Command ( "getlink", script_getlink, FILESYS ) );
      NEW ( Command ( "setlink", script_setlink, FILESYS ) );
      NEW ( Command ( "getfiles", script_getfiles, FILESYS ) );
      NEW ( Command ( "getdirs", script_getdirs, FILESYS ) );
      NEW ( Command ( "hide", script_change, FILESYS ) );
      NEW ( Command ( "unhide", script_change, FILESYS ) );
      NEW ( Command ( "share", script_change, FILESYS ) );
      NEW ( Command ( "unshare", script_change, FILESYS ) );
      NEW ( Command ( "filesend", script_filesend, FILESYS ) );
      NEW ( Command ( "mkdir", script_mkdir, FILESYS ) );
      NEW ( Command ( "rmdir", script_rmdir, FILESYS ) );
      NEW ( Command ( "cp", script_mv_cp, FILESYS ) );
      NEW ( Command ( "mv", script_mv_cp, FILESYS ) );
      NEW ( Command ( "getflags", script_getflags, FILESYS ) );
      NEW ( Command ( "setflags", script_setflags, FILESYS ) );
      FILESYS->bt_fil = NEW ( BindTableFiles ( FILESYS ) );
      G->bt_dcc->AddBinds(mydcc,1,FILESYS);
      FILESYS->bt_fil->AddBinds(myfiles,COMMANDS,FILESYS);
#if defined(HAVE_SERVER_H) || defined(HAVE_BOTNET_H)
      G->bt_load->AddBinds(myload,COUNT1+COUNT2,FILESYS);
      G->bt_unld->AddBinds(myunld,COUNT1+COUNT2,FILESYS);
# ifdef HAVE_SERVER_H
      init_server_ctcps(0,0);
# endif
# ifdef HAVE_BOTNET_H
      load_botnet(0,0);
# endif
#endif
      FILESYS->TRANSFER->bt_rcvd->AddBinds(myrcvd, 1, FILESYS);
      FILESYS->TRANSFER->bt_sent->AddBinds(mysent, 1, FILESYS);
   }
   return r;
}

