/*
 * Channel3.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "irc.h"
#include "Socket.h"
#include "Channel3.h"
#include "server.h"
#include "UserRecord.h"
#include "BindTable.h"
#include "UserEntryComment.h"
#include "ModuleEntryIrc.h"
#include "ModuleEntryServer.h"
#include "ChannelSettingsFlood.h"
#include "ChannelSettingsChanmode.h"
#include "ChannelBan.h"
#include "ChannelMember.h"
#include "BanRecord.h"
#define CLASS "Channel3"

Channel3::~Channel3 ( void ) {
   FUNCTION;
   Clear ( 0 );
}

/* initialize out the channel record */
void Channel3::Init (void) {
   FUNCTION;
   UnsetActive();
   UnsetPending();
   mode = 0;
   chankey = 0;
   chanbans = NEW ( ChannelBan () );
   exceptions = NEW ( ChannelBan () );
   member = NEW ( ChannelMember () );
   topic = NULL;
   maxmembers = -1;
   cmode = (struct cmode_st *)
     my_malloc(sizeof(struct cmode_st)*I->modesperline,I->memtype_chaninfo);
}

Channel2 * Channel3::NewChannel ( const char * x) {
   FUNCTION;
   return NEW ( Channel3 ( x ) );
}

/* clear out channel data from memory */
void Channel3::Clear (int reset) {
   FUNCTION;
   /* channel key may be needed for rejoin */
   if (topic) 
     my_free(topic,I->memtype_chaninfo);
   topic = 0;
   DELETE ( member );
   member = 0;
   DELETE ( chanbans );
   chanbans = 0;
   DELETE ( exceptions );
   exceptions = 0;
   if (cmode) {
      while (lastmode) {
	 lastmode--;
	 if (cmode[lastmode].op)
	   my_free(cmode[lastmode].op,I->memtype_chaninfo);
      }
      my_free(cmode,I->memtype_chaninfo);
   }
   cmode = 0;
   if (reset)
     Init();
   else {
      I->SERVER->SendPart(Q_SERVER, key, I->Lang( 0x0BF, ""));
      if (chankey) 
	my_free(chankey,I->memtype_chaninfo);
      chankey = 0;
   }
}

void Channel3::RedoCmode ( void ) {
   FUNCTION;
   my_free(cmode,I->memtype_chaninfo);
   cmode = (struct cmode_st *)
     my_malloc(sizeof(struct cmode_st) * I->modesperline,I->memtype_chaninfo);
}

void Channel3::FlushMode ( int pri ) {
   FUNCTION;
   char *p, out[512], post[512];
   int i, posti, ok = 0;

   p = out;
   posti = 0;
   if (pls[0]) 
     for (*p++ = '+', i = 0; pls[i]; *p++ = pls[i++]);
   if (mns[0])  
     for (*p++ = '-', i = 0; mns[i]; *p++ = mns[i++]);
   pls[0] = 0;
   mns[0] = 0;
   bytes = 0;
   /* +k or +l ? */
   if (nkey[0]) {
      *p++ = '+';
      ok = 1;
      *p++ = 'k';
      posti += G->SPrintF(post + posti, 511 - posti, "%s ", nkey);
      nkey[0] = 0;
   }
   if (limit > 0) {
      if (!ok) {
	 *p++ = '+';
	 ok = 1;
      }
      *p++ = 'l';
      posti += G->SPrintF(post + posti, 511 - posti, "%d ", limit);
      limit = -1;
   }
   for (i = 0; i < lastmode; i++)
     if ((cmode[i].plus == '-') && ((cmode[i].type == 'b')
				    || (cmode[i].type == 'e'))) {
	if (ok < 2) {
	   *p++ = '-';
	   ok = 2;
	}
	*p++ = cmode[i].type;
	posti += G->SPrintF(post + posti, 511 - posti, "%s ", cmode[i].op);
	my_free(cmode[i].op,I->memtype_chaninfo);
	cmode[i].op = NULL;
     }
   ok &= 1;
   for (i = 0; i < lastmode; i++)
     if (cmode[i].plus == '+') {
	if (!ok) {
	   *p++ = '+';
	   ok = 1;
	}
	*p++ = cmode[i].type;
	posti += G->SPrintF(post + posti, 511 - posti, "%s ", cmode[i].op);
	my_free(cmode[i].op,I->memtype_chaninfo);
	cmode[i].op = NULL;
     }
   ok = 0;
   /* -k ? */
   if (rmkey[0]) {
      *p++ = '-';
      ok = 1;
      *p++ = 'k';
      posti += G->SPrintF(post + posti, 511 - posti, "%s ", rmkey);
      rmkey[0] = 0;
   }
   for (i = 0; i < lastmode; i++)
     if ((cmode[i].plus == '-') && (cmode[i].type != 'b') 
	 && (cmode[i].type != 'e')) {
	if (!ok) {
	   *p++ = '-';
	   ok = 1;
	}
	*p++ = cmode[i].type;
	posti += G->SPrintF(post + posti, 511 - posti, "%s ", cmode[i].op);
	my_free(cmode[i].op,I->memtype_chaninfo);
	cmode[i].op = NULL;
     }
   *p = 0;
   while (lastmode) {
      lastmode --;
      if (cmode[lastmode].op) {
	 my_free(cmode[lastmode].op,I->memtype_chaninfo);
	 cmode[lastmode].op = 0;
      }
   }
   if (posti) {
      post[--posti] = 0;
      G->SPrintF(p, 511 - (p - out), " %s", post);
   }
   if (out[0])
     I->SERVER->SendMode(pri, key, out);
}

void Channel3::DoBan (const char * ban ) {
   FUNCTION;
   ChannelBan *b;
   
   for (b = chanbans; (b = b->Next());)
     if (G->WildMatch(ban, b->Key()) && strcasecmp(ban, b->Key()))
       AddMode ('-', 'b', b->Key());
   AddMode('+', 'b', ban);
   FlushMode(QUICK);
}

void Channel3::SetMode ( char * msg ) {
   FUNCTION;
   char * md, *p;
   int i = 1;
   
   mode = 0;
   for (md = G->Split(&msg); md[i]; i++)
     switch (md[i]) {
       case 'i':
	mode |= CHANINV;
	break;
       case 'p':
	mode |= CHANPRIV;
	break;
       case 's':
	mode |= CHANSEC;
	break;
       case 'm':
	mode |= CHANMODER;
	break;
       case 't':
	mode |= CHANTOPIC;
	break;
       case 'n':
	mode |= CHANNOMSG;
	break;
       case 'a':
	mode |= CHANANON;
	break;
       case 'q':
	mode |= CHANQUIET;
	break;
       case 'z':
	mode |= CHANZOMBIE;
	break;
       case 'k':
	if ((p = G->Split(&msg))[0])
	  SetKey(p);
	break;
       case 'l':
	if ((p = G->Split(&msg))[0])
	  maxmembers = atoi(p);
     }
}

/* queue a channel mode change */
void Channel3::AddMode (char plus, char mode, const char * op2) {
   FUNCTION;
   int i, l;
   char * t, tbuf[40]; /* greater than 26 or NICKMAX + 2*/
   const char * op = op2;
   
   if (Meop()) {
      if (mode == 'c') { /* send c's alone */
	 /* lets not even bother sending on < 48hrs */
	 if (timestamp && ((G->now - timestamp) < 172800)) {
	    G->SPrintF(tbuf, 26, "%cc %s", plus, op2);
	    FlushMode(Q_SERVER);
	    I->SERVER->SendMode(Q_SERVER, key, tbuf);
	 }
      } else { /* break e's seperately */
	 if (lastmode && (((mode == 'e') && (cmode[lastmode - 1].type != 'e'))
			  || ((mode != 'e') && (cmode[lastmode - 1].type == 'e')))) 
	   FlushMode(Q_SERVER);
	 if ((mode == 'o') || (mode == 'b') || (mode == 'v') || (mode == 'e')) {
	    /* if -b'n a non-existant ban...nuke it */
	    if ((plus == '-') && (mode == 'b') && !chanbans->Find(op))
	      return;
	    /* likewise for exceptions */
	    if ((plus == '-') && (mode == 'e') && !exceptions->Find(op))
	      return;
	    /* make alowances for broken server ban handling */
	    if ((mode == 'b') && (op[0] == ':')) {
	       tbuf[0] = (plus == '+') ? '?' : ':';
	       strncpyz(tbuf + 1, op + ((plus == '+') ? 1 : 0), NICKMAX);
	       op = tbuf;
	    }
	    /* is it needed for exceptions too ? */
	    /* op-type mode change */
	    for (i = 0; i < lastmode; i++)
	      if ((cmode[i].type == mode) && (cmode[i].plus == plus)
		  && cmode[i].op && !strcasecmp(cmode[i].op, op))
		return;		/* already in there :- duplicate */
	    l = strlen(op) + 1;
	    if ((bytes + l) > I->mode_buf_len)
	      FlushMode(Q_SERVER);
	    cmode[lastmode].plus = plus;
	    cmode[lastmode].type = mode;
	    cmode[lastmode].op = (char *)my_malloc(l,I->memtype_chaninfo);
	    bytes += l; /* add 1 for safety */
	    strcpy(cmode[lastmode].op, op);
	    lastmode++;
	    if (lastmode == I->modesperline)
	      FlushMode(Q_SERVER);  /* full buffer!  flush modes */
	    /* if it's a +/- 'b' starting with ':' also flush */
	    else if ((mode == 'b') && (op2[0] == ':') && (plus == '-'))
	      FlushMode(Q_SERVER); 
	 } else if ((plus == '+') && (mode == 'k')) 
	      strncpyz(nkey, op, 80);
	 else if ((plus == '-') && (mode == 'k')) 
	   strncpyz(rmkey, op, 80);
	 else if ((plus == '+') && (mode == 'l')) 
	   limit = atoi(op);
	 else if (!(strchr(t = (plus == '+') ? pls : mns, mode))) {
	    while (*t)
	      t++;
	    *t++ = mode;
	    *t = 0;
	 }
      }
   }
}

int Channel3::NukeChanBan (char * ban, int j) {
   FUNCTION;
   ChannelBan * b;
   int i = 0;
   
   for (b = chanbans; (b = b->Next());) {
      if ((j && !EqualsBan(b->Key()) && (++i == j))
	  || (!j && !strcasecmp(b->Key(), ban))) {
	 if (j)
	   strncpyz(ban, b->Key(), UHOSTLEN);
	 AddMode('-', 'b', b->Key());
	 return 1;
      }
   }
   return 0;
}

int Channel3::ShowChanBan (char * match, int k, Socket * idx) {
   FUNCTION;
   ChannelBan * bl;
   char fill[512];
   
   context;
   for (bl = chanbans; (bl = bl->Next());)
     if (!EqualsBan(bl->Key())) {
	if (!match[0] || G->WildMatch(match, bl->Key())) {
	   int which = 0x025;
	   
	   if (bl->Timer() != 0) 
	     strcpy(fill, I->CHANNELS->Lang(0x022, "222",
					    (int)bl->Age()/3600,
					    (int)(bl->Age()/60)%60,
					    (int)bl->Age()%60));
	   else 
	     fill[0] = 0;
	   if (strchr(bl->Setter(),'!'))
	     which = 0x023;
	   else if (!strcasecmp(bl->Setter(), "existant"))
	     which = 0x024;
	   idx->Send(I->CHANNELS->Lang(which, "cdsss", '*', k, bl->Key(),
				       bl->Setter(), fill));
	   k++;
	}
     }
   if (I->SERVER->use_ts4) {
      int t = 0;
      for (bl = exceptions; (bl = bl->Next());)
	if (!match[0] || G->WildMatch(match, bl->Key())) {
	   int which = 0x025;
	   
	   if (!t) {
	      t++;
	      idx->Send(I->Lang(0x13A, ""));
	   }
	   if (bl->Timer() != 0) 
	     strcpy(fill, I->CHANNELS->Lang(0x022, "222",
					    (int)bl->Age()/3600,
					    (int)(bl->Age()/60)%60,
					    (int)bl->Age()%60));
	   else 
	     fill[0] = 0;
	   if (strchr(bl->Setter(),'!'))
	     which = 0x023;
	   else if (!strcasecmp(bl->Setter(), "existant"))
	     which = 0x024;
	   idx->Send(I->CHANNELS->Lang(which, "c sss", '*', 2, bl->Key(),
				       bl->Setter(), fill));
	}
   }
   return k;
}

/* set the key */
void Channel3::SetKey (const char * k) {
   FUNCTION;
   if (chankey)
     my_free(chankey, I->memtype_chaninfo);
   if (!k || !k[0])
     chankey = 0;
   else
     strcpy(chankey = (char *)my_malloc(strlen(k) + 1, I->memtype_chaninfo),k);
}

/* set the topic */
void Channel3::SetTopic (const char * k) {
   FUNCTION;
   if (topic)
     my_free(topic,I->memtype_chaninfo);
   if (k && k[0]) 
     strcpy(topic = (char *)my_malloc(strlen(k) + 1, I->memtype_chaninfo),k);
   else
     topic = 0;
}

/* are there any ops on the channel? */
int Channel3::AnyOps (void) {
   FUNCTION;
   ChannelMember * x = member;
   
   while ((x = x->Next()) && !x->HasOp());
   return x != NULL;
}

/* what the channel's mode CURRENTLY is */
const char *Channel3::GetMode (void) {
   FUNCTION;
   static char s[121];
   int i;
   
   s[0] = '+';
   i = 1;
   if (mode & CHANINV)
     s[i++] = 'i';
   if (mode & CHANPRIV)
     s[i++] = 'p';
   if (mode & CHANSEC)
     s[i++] = 's';
   if (mode & CHANMODER)
     s[i++] = 'm';
   if (mode & CHANTOPIC)
     s[i++] = 't';
   if (mode & CHANNOMSG)
     s[i++] = 'n';
   if (mode & CHANANON)
     s[i++] = 'a';
   if (mode & CHANZOMBIE)
     s[i++] = 'z';
   if (nkey[0])
     s[i++] = 'k';
   if (maxmembers > -1)
     s[i++] = 'l';
   s[i] = 0;
   if (nkey[0])
     i+= G->SPrintF(s+i, 120-i, " %s", nkey);
   if (maxmembers > -1)
     G->SPrintF(s+i, 120-i, " %d", maxmembers);
   return s;
}

/* this is a clone of detect_flood, but works for channel specificity now */
/* and handles kick & deop as well */

/* even I can't handle too many cascades */
int Channel3::HandleFlood (int which,char * floodnick, char *p, char * from) {
   FUNCTION;
   char h[UHOSTLEN + 1];
   ChannelMember * m;
   int lang = 0;
   
   switch (which) {
     case FLOOD_PRIVMSG:
     case FLOOD_NOTICE:
     case FLOOD_CTCP:
      if (Meop()) 
	lang = 0x0C4;
      break;
     case FLOOD_JOIN:
     case FLOOD_NICK:
      G->SPrintF(h, UHOSTLEN, "*!*@%s", p);
      if (!chanbans->Find(h) && Meop()) {
	 AddMode('-', 'o', floodnick);
	 if (I->SERVER->use_ts4)
	   AddMode('-', 'h', floodnick);
	 DoBan(h);
      }
      if (!MatchBan(from)) {
	 lang = (which == FLOOD_NICK) ? 0x0C8 : 0x0C6;
	 AddBan(h, G->origbotname, I->Lang( lang+1,""),
		G->now + (60 * I->CHANNELS->ban_time),0);
	 if (!EnforceBans() && Meop())
	   KickAll(h, I->Lang(0x0CA,""));
      }
     case FLOOD_KICK:
      if (Meop())
	 lang = 0xCB;
      break;
     case FLOOD_DEOP:
      if (Meop())
	lang = 0x0CD;
      break;
   }
   if (lang && Meop()) {
      I->LOG_MODES.Put(I->Lang( lang, "ss", floodnick, p, key), this);
      if ((m = member->Find(floodnick))) 
	m->Kick(I->Lang( lang+1,""));
   }
   return 1;
}

int Channel3::DetectFlood (char * floodnick, char * floodhost, char * from,
			   int which, char * victim) {
   FUNCTION;
   char * ftype = NULL, *p;
   UserRecord * u;
   int thr = 0, lapse = 0;
   ChannelSettingsFlood * flood;
   FlagRecord user (FlagRecord::GLOBAL | FlagRecord::CHANNEL);

   if ((which >= 0) && (which < FLOOD_CHAN_MAX)) {
      if (!(u = G->userlist->FindHost(from)))
	u = G->userlist;
      user.Set(u, key);
      context;
      if (!user.ChanBot() 
	  && !((which == FLOOD_DEOP) && user.AnyMaster())
	  && !((which != FLOOD_DEOP) && user.AnyFriend())
	  && !(DontKickOps() && user.AnyValidOp())) {
	 flood = (ChannelSettingsFlood *)FindSettings(I->CHANSET_FLOOD);
	 /* determine how many are necessary to make a flood */
	 switch (which) {
	   case FLOOD_PRIVMSG:
	   case FLOOD_NOTICE:
	    if (!(thr = flood->pub_thr) && !(lapse = flood->pub_time)) {
	       thr = I->SERVER->flood_thr;
	       lapse = I->SERVER->flood_time;
	    }
	    ftype = "pub";
	    break;
	   case FLOOD_CTCP:
	    if (!(thr = flood->ctcp_thr) && !(lapse = flood->ctcp_time)) {
	       thr = I->SERVER->flood_ctcp_thr;
	       lapse = I->SERVER->flood_ctcp_time;
	    }
	    ftype = "ctcp";
	    break;
	   case FLOOD_JOIN:
	   case FLOOD_NICK:
	    thr = flood->join_thr;
	    lapse = flood->join_time;
	    ftype = (which == FLOOD_JOIN) ? "join" : "nick";
	    break;
	   case FLOOD_DEOP:
	    thr = flood->deop_thr;
	    lapse = flood->deop_time;
	    ftype = "deop";
	    break;
	   case FLOOD_KICK:
	    thr = flood->kick_thr;
	    lapse = flood->kick_time;
	    ftype = "kick";
	    break;
	 }
	 if (thr && !I->SERVER->MatchMyNick(floodnick) &&
	     strcasecmp(floodhost, I->SERVER->botuserhost)) {
	    if ((p = ((which == FLOOD_KICK) || (which == FLOOD_DEOP)) 
		 ? floodnick : strchr(floodhost, '@'))) {
	       if (*p == '@') 
		 p++;
	       if (strcasecmp(floodwho[which], p)) {	/* new */
		  strncpy(floodwho[which], p,81);
		  floodwho[which][81] = 0;
		  floodtime[which] = G->now;
		  floodnum[which] = 1;
	       } else if (floodtime[which] < G->now - lapse) {
		  /* flood timer expired, reset it */
		  floodtime[which] = G->now;
		  floodnum[which] = 1;
	       } else if ((which != FLOOD_DEOP) || strcasecmp(deopd,victim)) {
		  if (which == FLOOD_DEOP) 
		    strcpy(deopd,victim);
		  floodnum[which]++;
		  context;
		  if (floodnum[which] >= thr) {	/* FLOOD */
		     void * args[5] = {  floodnick, from, u, ftype, this };
		     /* reset counters */
		     floodnum[which] = 0;
		     floodtime[which] = 0;
		     floodwho[which][0] = 0;
		     if (which == FLOOD_DEOP)
		       deopd[0] = 0;
		     if (I->SERVER->bt_flud->Check(ftype, 0, args) != EXEC_LOG) 
		       return HandleFlood(which,floodnick,p,from);
		  }
	       }
	    }
	 }
      }
   }
   context;
   return 0;
}

/* given a [nick!]user@host, place a quick ban on them on a chan */
const char * Channel3::QuickBan (const char * uhost) {
   FUNCTION;
   static char s1[UHOSTLEN+1];
   const char *pp;

   pp = (pp = strchr(uhost, '@')) ? pp + 1 : uhost;
   G->MaskHost(pp, s1);
   s1[2] = '*';
   DoBan(s1);
   return s1;
}

/* recheck_bans makes sure that all who are 'banned' on the userlist are
   actually in fact banned on the channel */
void Channel3::RecheckBans (void) {
   FUNCTION;
   BanRecord * b;
   int i;
   
   for (i = 0; i < 2; i++) {
      b = i ? bans : G->global_bans;
      while ((b = b->Next()))
	if (!chanbans->Find(b->Key()) && (!DynamicBans() || b->IsSticky()))
	  DoBan(b->Key());
   }
}

/* resets the bans on the channel */
void Channel3::ResetBans (void) {
   FUNCTION;
   ChannelBan *b;
   
   if (Meop()) 
     for (b = chanbans; (b = b->Next());)
       if (!EqualsBan(b->Key()))
	 AddMode('-', 'b', b->Key());
   RecheckBans();
}

/* resets the bans on the channel */
void Channel3::RemoveMatchingExceptions (const char * mask) {
   FUNCTION;
   ChannelBan *b;
   
   if (Meop()) 
     for (b = exceptions; (b = b->Next());)
       if (G->WildMatch(b->Key(),mask)) 
	 AddMode('-', 'e', b->Key());
}

/* reset the channel information */
void Channel3::ResetInfo (void) {
   FUNCTION;
   context;
   if (!Pending()) {
      ChannelSettingsChanmode * chanmode = 
	(ChannelSettingsChanmode *)FindSettings(I->CHANSET_CHANMODE);
      Clear(1);
      SetPending();
      UnsetActive();
      if (!AskedBans()) {
	 SetAskedBans();
	 if (I->SERVER->use_ts4)
	   I->SERVER->SendMode(Q_MODE, key, "+e");
	 I->SERVER->SendMode(Q_MODE, key, "+b");
      }
      I->SERVER->SendWho(Q_MODE, key);
      I->SERVER->SendMode(Q_MODE, key, "");
      if (I->SERVER->use_ts4 && chanmode->Pass() && Meop()) {
	 char buf[100];
	 
	 G->SPrintF(buf, 100, "=c %s", chanmode->Pass());
	 I->SERVER->SendMode(Q_MODE, key, buf);
      }
      I->SERVER->SendTopic(Q_SERVER, key, NULL);
   }
}

/* things to do when i just became a chanop: */
void Channel3::Recheck (int dobans) {
   FUNCTION;
   ChannelMember * m;
   int pls, mns, use_ts4 = I->SERVER->use_ts4;
   ChannelSettingsChanmode * chanmode;
   UserRecord * u;
   FlagRecord user ( FlagRecord::GLOBAL |FlagRecord::CHANNEL );
   
   if (!Meop() && !MeHelper())
     return;
   /* okay, sort through who needs to be deopped. */
   context;
   for (m = member; (m = m->Next());) {
      if (!(u = m->MemberUser()))
	u = G->userlist;
      user.Set(u, key);
      if (!I->SERVER->MatchMyNick(m->Key())) {
	 /* has ops, is +d or chan is +bitch & user not +o */
	 if (m->HasOp() && !m->SentDeop() && Meop() &&
	     (user.AnyValidDeop() || ((Bitch() || use_ts4)
				      && !user.AnyValidOp()))) 
	   m->TakeOp();
	 else if (use_ts4 && m->HasHelper() && !m->SentDeHelper()
		  && (Meop() || (MeHelper() && !NoHelper()))
		  && (user.AnyValidDeop() || (Bitch() && !user.AnyValidHelper())))
	   m->TakeHelper();
	 /* no ops, is +o & chan is +autop or user is + a */
	 else if (!m->HasOp() && !m->SentOp() && user.AnyValidOp() && Meop()
		  && (AutoOp() || user.AnyAutoop())) 
	   m->GiveOp();
	 else if (use_ts4 && !m->HasHelper() && !m->HasOp() && 
		  !m->SentHelper() && !m->SentOp() && user.AnyValidHelper()
		  && (Meop() || (MeHelper() && !NoHelper()))
		  && (AutoOp() || user.AnyAutoop()))
	   m->GiveHelper();
	 /* now lets check 'em vs bans */
	 /* if we're enforcing bans & they match a ban */
	 if (EnforceBans())
	   m->RefreshBanKick();
	 /* otherwise...are they +k ? */
	 else if (user.AnyKick()) {
	    UserEntryComment * uec = FindComment ( u );
	    QuickBan(m->Host());
	    m->Kick((uec && uec->Value() && uec->Value()[0]) ? 
		    uec->Value() : I->Lang( 0x0D1, ""));
	 } else if (AutoVoice()) {
	    /* -v on chan, +v on userlist */
	    if (!m->HasVoice() && !m->HasOp() && user.AnyValidVoice())
	      m->GiveVoice();
	    /* +v on chan, +q on userlist */
	    else if (m->HasVoice() && user.AnyValidQuiet())
	      m->TakeVoice();
	 }
      }
   }
   if (dobans)
     RecheckBans();
   chanmode = (ChannelSettingsChanmode *)
     FindSettings(I->CHANSET_CHANMODE);
   pls = chanmode->Plus();
   mns = chanmode->Minus();
   if ((pls & CHANINV) && !(mode & CHANINV))
     AddMode( '+', 'i', "");
   else if ((mns & CHANINV) && (mode & CHANINV))
     AddMode( '-', 'i', "");
   if (!use_ts4 || Meop()) {
      if ((pls & CHANPRIV) && !(mode & CHANPRIV))
	AddMode( '+', 'p', "");
      else if ((mns & CHANPRIV) && (mode & CHANPRIV))
	AddMode( '-', 'p', "");
   }
   if ((pls & CHANSEC) && !(mode & CHANSEC))
     AddMode( '+', 's', "");
   else if ((mns & CHANSEC) && (mode & CHANSEC))
     AddMode( '-', 's', "");
   if ((pls & CHANMODER) && !(mode & CHANMODER))
     AddMode( '+', 'm', "");
   else if ((mns & CHANMODER) && (mode & CHANMODER))
     AddMode( '-', 'm', "");
   if ((pls & CHANTOPIC) && !(mode & CHANTOPIC))
     AddMode( '+', 't', "");
   else if ((mns & CHANTOPIC) && (mode & CHANTOPIC))
     AddMode( '-', 't', "");
   if ((pls & CHANNOMSG) && !(mode & CHANNOMSG))
     AddMode( '+', 'n', "");
   else if ((mns & CHANNOMSG) && (mode & CHANNOMSG))
     AddMode( '-', 'n', "");
   if ((pls & CHANANON) && !(mode & CHANANON))
     AddMode( '+', 'a', "");
   else if ((mns & CHANANON) && (mode & CHANANON))
     AddMode( '-', 'a', "");
   if ((pls & CHANQUIET) && !(mode & CHANQUIET))
     AddMode( '+', 'q', "");
   else if ((mns & CHANQUIET) && (mode & CHANQUIET))
     AddMode( '-', 'q', "");
   if ((chanmode->Limit() > 0) &&
       (maxmembers != chanmode->Limit()))
     AddMode( '+', 'l', G->ToBase10(chanmode->Limit()));
   else if ((mns & CHANLIMIT) && (maxmembers > 0))
     AddMode( '-', 'l', "");
   if (chanmode->Key()[0]) {
      if (!chankey || strcasecmp(chankey, chanmode->Key())) {
	 if (chankey) 
	   AddMode( '-', 'k', chankey);
	 AddMode( '+', 'k', chanmode->Key());
      }
   } else if ((mns & CHANKEY) && chankey)
	AddMode( '-', 'k', chankey);
   if (use_ts4) {
      if ((mns & CHANPASS) && chanpass[0]) 
	AddMode( '-', 'c', "");
      else if (chanmode->Pass() && (!chanpass[0] || 
				    strcmp(chanpass, chanmode->Pass())))
	AddMode( '+', 'c', chanmode->Pass());
   }
}

/*  if i'm the only person on the channel, and i'm not op'd,   *
 *  might as well leave and rejoin. If i'm NOT the only person *
 *  on the channel, but i'm still not op'd, demand ops         */
void Channel3::CheckLonely (void) {
   FUNCTION;
   ChannelMember * m = member;
   int i = 0;
   static int whined = 0;
   UserRecord * u;
   void * args;
   FlagRecord fr (FlagRecord::GLOBAL);
   
   context;
   if (!Pending() && Active() && !Meop()) {
      /* count non-split channel members */
      context;
      while ((m = m->Next()))
	if (!m->IsSplit())
	  i++;
      context;
      if ((i == 1) && Cycle()) {
	 context;
	 G->LOG_MISC.Put(I->Lang( 0x0D2, "s", key));
	 I->SERVER->SendPart(Q_MODE, key, I->Lang( 0x0D3, ""));
	 I->SERVER->SendJoin(Q_MODE, key, safe_str(chankey));
	 whined = 0;
	 context;
      } else if (AnyOps()) {
	 context;
	 whined = 0;
	 args = this;
	 I->bt_need->Check("op", 0, &args);
	 context;
      } else {
	 context;
	 /* other people here, but none are ops */
	 /* are there other bots?  make them LEAVE. */
	 if (!whined) {
	    G->LOG_MISC.Put(I->Lang( 0x0D4, "s", key));
	    whined = 1;
	 }
	 context;
	 for (m = member; (m = m->Next());)
	   if (!I->SERVER->MatchMyNick(m->Key()) && 
	       (!(u = m->MemberUser()) || !fr.Set(u)->ChanBot()))
	     break;
	 context;
	 if (!m) {
	    char s[100], t[512];
	    int c, i;
	    
	    context;
	    G->SPrintF(s, 99, "go %s", key);
	    /* ALL bots!  make them LEAVE!!! */
	    for (c = 0, i = 0, m = member; (m = m->Next()); c++) {
	       if (c == 10) {
		  t[--i] = 0;
		  I->SERVER->SendMsg(Q_SERVER, t, s);
		  i = 0;
		  c = 0;
	       }
	       i += G->SPrintF(t + i, 511 - i, "%s,", m->Key());
	    }
	    context;
	    if (i) {
	       t[--i] = 0;
	       I->SERVER->SendMsg(Q_SERVER, t, s);
	    }
	    context;
	 } else {
	    context;
	    args = this;
	    /* some humans on channel, but still op-less */
	    I->bt_need->Check("op", 0, &args);
	    context;
	 }
	 context;
      }
      context;
   }
   context;
}

void Channel3::KickAll ( char * mask, char * reason ) {
   FUNCTION;
   ChannelMember * m = member;
   char s[UHOSTLEN+1];
   UserRecord * u;
   FlagRecord user ( FlagRecord::CHANNEL | FlagRecord::GLOBAL );
   
   while ((m = m->Next())) {
      if (!(u = m->MemberUser()))
	u = G->userlist;
      user.Set(u,key);
      G->SPrintF(s, UHOSTLEN, "%s!%s", m->Key(), m->Host());
      if (!m->SentKick() && !user.AnyFriend() 
	  && !(DontKickOps() && user.AnyValidOp())
	  && !I->SERVER->MatchMyNick(m->Key()) && G->WildMatch(mask, s)) {
	 m->SetSentKick();
	 I->SERVER->SendKick(Q_MODE,key,m->Key(),reason);
      }
   }
}

BindReturn Channel3::CheckTN ( const char * nick, char * uhost,
			       UserRecord * u, char * reason,
			       BindTable * table ) {
   FUNCTION;
   void * args[5] = { (void *)nick, uhost, u, this, reason };
   char buf[512];
   FlagRecord fr ( FlagRecord::CHANNEL | FlagRecord::GLOBAL);
   
   G->SPrintF(buf, 511, "%s %s", key, reason);
   fr.Set(u,key);
   return table->Check(buf, &fr, args);
}

BindReturn Channel3::CheckMode ( const char * nick, char * uhost,
				UserRecord * u, char * mode, char * victim ) {
   FUNCTION;
   void * args[6] = { (void *)nick, uhost, u, this, mode, victim };
   char buf[512];
   FlagRecord fr ( FlagRecord::CHANNEL | FlagRecord::GLOBAL);
   
   G->SPrintF(buf, 511, "%s %s %s", key, mode, victim);
   fr.Set(u,key);
   return I->bt_mode->Check(buf, &fr, args);
}


BindReturn Channel3::CheckJPSR (char * nick, char * uhost, UserRecord * u,
				BindTable * table) {
   FUNCTION;
   void * args[4] = { nick, uhost, u, this };
   char buf[512];
   FlagRecord fr ( FlagRecord::CHANNEL | FlagRecord::GLOBAL);
   
   G->SPrintF(buf, 511, "%s %s!%s", key, nick, uhost);
   fr.Set(u,key);
   return table->Check(buf, &fr, args);
}

void Channel3::GotOp (char * nick, char * who, FlagRecord * opper) {
   FUNCTION;
   ChannelMember * m;
   int check_chan = 0;
   FlagRecord victim (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
   
   if (!(m = member->Find(who))) {
      G->LOG_MISC.Put(I->Lang(0x0E2,"ss",key,who), this);
      I->SERVER->SendWho(Q_MODE, who);
   } else {
      victim.Set(m->MemberUser(), key);
      /* Did *I* just get opped? */
      if (!Meop() && I->SERVER->MatchMyNick(who))
	check_chan = 1;
      /* opped by not me or server & no deop sent */
      else if (Meop() && !I->SERVER->MatchMyNick(who) &&
	       !m->SentDeop() && nick[0] && (I->reversing ||
	       /* Channel is +bitch, and the opper !+m/+b and oped is !+o */
	       (((I->SERVER->use_ts4 || Bitch())
		 && !opper->AnyMaster() && !opper->ChanBot()
		 && !victim.AnyValidOp() && !victim.ChanBot())
		/* or +d user & opper not !+m */
		|| (victim.AnyValidDeop() && !opper->AnyMaster())))) 
	m->TakeOp();
      else if (!nick[0] && !m->WasOp() && StopNetHack() && !m->SentDeop()
	       && Meop()) {
	 m->TakeOp();
	 m->SetFakeOp();
      } else
	m->ClearFakeOp();
      m->SetOp();
      m->ClearSentOp();
      m->ClearWasOp();
      if (check_chan) {
	 SetMeop();
	 Recheck(1);
      }
   }
}

/* revenge tactic: person did something bad, if they are oplisted,
   remove them from the op list otherwise, deop them */
void Channel3::TakeRevenge (char * who, char * reason) {
   FUNCTION;
   char * nick, s[UHOSTLEN+1], s1[UHOSTLEN+1], ct[81];
   time_t tm;
   UserRecord * u;
   UserEntryComment * uec;
   FlagRecord fr (FlagRecord::CHANNEL | FlagRecord::GLOBAL);
   ChannelMember * m;
   
   if ((u = G->userlist->FindHost(who)) == G->userlist)
     u = NULL;
   fr.Set(u,key);
   fr.SetMatch (FlagRecord::CHANNEL);
   if (fr.AnyFriend())
     G->LOG_MISC.Put(I->Lang( 0x0E3, "ss", who, reason));
   else if (u && fr.AnyValidOp()) {
      if (fr.ChanOp())
	fr.ClearChanOp();
      else 
	fr.SetChanDeop();
      u->Set(fr,key);
      G->LOG_MISC.Put(I->Lang( 0x0E4, "sss", u->Handle(), who, reason));
      Recheck(0);
   } else if (!MatchBan(who)) {
      /* get current time into a string */
      tm = G->now;
      strcpy(ct, ctime(&tm));
      ct[10] = 0;
      ct[16] = 0;
      strcpy(ct, &ct[8]);
      strcpy(&ct[2], &ct[3]);
      strcpy(&ct[7], &ct[10]);
      if (fr.AnyValidDeop()) {
	 G->LOG_MISC.Put(I->Lang( 0x0E5, "ss", who, reason));
	 nick = G->SplitNick(&who);
	 G->MaskHost(who, s1);
	 G->SPrintF(s, UHOSTLEN, "(%s) %s", ct, reason);
	 AddBan(s1, G->origbotname, s, G->now + (60 * I->CHANNELS->ban_time),0);
	 if (Meop())
	   DoBan(s1);
      } else if (u) {
	 fr.SetChanDeop();
	 u->Set(fr,key);
	 G->SPrintF(s, UHOSTLEN, "(%s) %s", ct, reason);
	 G->LOG_MISC.Put(I->Lang( 0x0E6, "sss", u->Handle(), who, reason));
	 Recheck(0);
      } else {
	 nick = G->SplitNick(&who);
	 strcpy(s1, who);
	 G->MaskHost(s1, s);
	 strcpy(s1, nick);
	 while (G->userlist->Find(s1)) {
	    if (!strncmp(s1, "bad", 3)) 
	      strcpy(s1 + 3, G->ToBase10(atoi(s1+3)));
	    else
	      strcpy(s1, "bad1");
	 }
	 fr.SetGlobal(0);
	 u = NEW ( UserRecord (G->userlist, s1, s, fr) );
	 if ((m = M()->Find(who)))
	   m->SetUser(u);
	 fr.SetChanDeop();
	 u->Set(fr, key);
	 G->SPrintF(s, UHOSTLEN, "(%s) %s (%s)", ct, reason, who);
	 uec = GetComment(u);
	 uec->Set(s);
	 G->LOG_MISC.Put(I->Lang( 0x0E7, "ss", who, reason));
	 Recheck(0);
	 RecheckBans();
      }
   }
}

void Channel3::GotDeop (char * nick, char * from, char * who, FlagRecord * user) {
   FUNCTION;
   ChannelMember * m;
   char s1[UHOSTLEN+1];
   void * args;
   FlagRecord victim ( FlagRecord::CHANNEL | FlagRecord::GLOBAL );
   
   if (!(m = member->Find(who))) {
      G->LOG_MISC.Put(I->Lang( 0x0E2, "ss", key, who), this);
      I->SERVER->SendWho(Q_MODE, who);
   } else {
      G->SPrintF(s1, UHOSTLEN, "%s!%s", nick, from);
      victim.Set(m->MemberUser(),key);
      /* deop'd someone on my oplist? */
      if (Meop()) {
	 /* victim is a +o/+f */
	 if ((victim.AnyFriend() || victim.AnyValidOp())
	     /* who's not me, not themselves, and actually HAS ops ? */
	     && !I->SERVER->MatchMyNick(nick) &&
	     strcasecmp(who, nick) && m->HasOp()) {
	    /* reop - if the deopper is NOT a master or bot? and protectop */
	    if (I->reversing || ((!user->AnyMaster() && !user->ChanBot()
			       && !m->SentDeop() && ProtectOps())
			      /* provided it's not +bitch ... 
			       or the user is a valid op */
			      && (!(I->SERVER->use_ts4 || Bitch())
				  || victim.AnyValidOp())
			      /* and provied the users not a de-op*/
			      && !victim.AnyValidDeop())) 
	      m->GiveOp();
	    /* if the perpetrator is not a friend or bot, then do the
	     * vengeful thing */
	    if (!user->AnyFriend() && Revenge() && nick[0])
	      TakeRevenge(s1,I->Lang(0x0E8,"ss",m->Key(),m->Host()));
	 }
      }
      if (!nick[0])
	I->LOG_MODES.Put(I->Lang(0x0E9,"sss",key,who,from),this);
      else
	DetectFlood(nick, from, s1, FLOOD_DEOP, who);
      /* having op hides your +v status -- so now that someone's lost ops,
       check to see if they have +v */
      if (!(m->HasVoice() || m->HasHelper()) && !m->SentWho()) {
	 I->SERVER->SendWho(Q_MODE,m->Key());
	 m->SetSentWho();
      }
      if (I->SERVER->MatchMyNick(who)) {
	 if (Revenge()) {
	    /* deopped ME!  take revenge */
	    if (!user->AnyFriend() && nick[0] && !I->SERVER->MatchMyNick(nick)) 
	      TakeRevenge(s1, I->Lang( 0x0EA, "s", I->SERVER->botname));
	    if (!nick[0])
	      I->LOG_MODES.Put(I->Lang(0x0EB,"s",key), this);
	 }
	 args = this;
	 I->bt_need->Check("op", 0, &args);
	 UnsetMeop();
      }
      m->ClearFakeOp();
      m->ClearOp();
      m->ClearSentDeOp();
      m->ClearWasOp();
   }
}

void Channel3::GotHelper (char * nick, char *, char * who, FlagRecord * opper) {
   FUNCTION;
   ChannelMember * m;
   int check_chan = 0;
   FlagRecord victim (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
   
   if (!(m = member->Find(who))) {
      G->LOG_MISC.Put(I->Lang(0x0E2,"ss",key,who), this);
      I->SERVER->SendWho(Q_MODE, who);
   } else {
      victim.Set(m->MemberUser(), key);
      if (!MeHelper() && I->SERVER->MatchMyNick(who))
	check_chan = 1;
      else if ((Meop() || (MeHelper() && !NoHelper()))
	       && !I->SERVER->MatchMyNick(who) && !NoHelper() &&
	       !m->SentDeHelper() && nick[0] && (I->reversing ||
	       ((Bitch() && !opper->AnyMaster() && !opper->ChanBot()
		 && !victim.AnyValidHelper() && !victim.ChanBot())
		|| (victim.AnyValidDeop() && !opper->AnyMaster())))) 
	m->TakeHelper();
      m->ClearHelper();
      m->ClearSentDeHelper();
      m->SetHelper();
      m->ClearSentHelper();
      if (check_chan) {
	 SetMeHelper();
	 Recheck(1);
      }
   }
}

void Channel3::GotDeHelper (char * nick, char * from, char * who, FlagRecord * user) {
   FUNCTION;
   ChannelMember * m;
   char s1[UHOSTLEN+1];
   void * args;
   FlagRecord victim ( FlagRecord::CHANNEL | FlagRecord::GLOBAL );
   
   if (!(m = member->Find(who))) {
      G->LOG_MISC.Put(I->Lang( 0x0E2, "ss", key, who), this);
      I->SERVER->SendWho(Q_MODE, who);
   } else {
      G->SPrintF(s1, UHOSTLEN, "%s!%s", nick, from);
      victim.Set(m->MemberUser(),key);
      if (Meop() || (MeHelper() && !NoHelper())) {
	 if ((victim.AnyFriend() || victim.AnyValidHelper())
	     && !I->SERVER->MatchMyNick(nick) &&
	     strcasecmp(who, nick) && m->HasHelper()) {
	    if (I->reversing || ((!user->AnyMaster() && !user->ChanBot()
				  && !m->SentDeHelper() && ProtectOps())
				 && (!Bitch() || victim.AnyValidHelper())
				 && !victim.AnyValidDeop())) 
	      m->GiveHelper();
	    if (!user->AnyFriend() && Revenge() && nick[0])
	      TakeRevenge(s1,I->Lang(0x0E8,"ss",m->Key(),m->Host()));
	 }
      }
      if (!nick[0])
	I->LOG_MODES.Put(I->Lang(0x0E9,"sss",key,who,from),this);
      else
	DetectFlood(nick, from, s1, FLOOD_DEOP, who);
      if (!m->HasVoice() && !m->SentWho()) {
	 I->SERVER->SendWho(Q_MODE,m->Key());
	 m->SetSentWho();
      }
      if (I->SERVER->MatchMyNick(who)) {
	 if (Revenge()) {
	    if (!user->AnyFriend() && nick[0] && !I->SERVER->MatchMyNick(nick)) 
	      TakeRevenge(s1, I->Lang( 0x0EA, "s", I->SERVER->botname));
	    if (!nick[0])
	      I->LOG_MODES.Put(I->Lang(0x0EB,"s",key), this);
	    args = this;
	    I->bt_need->Check("op", 0, &args);
	 }
	 UnsetMeHelper();
      }
   }
}

void Channel3::GotBan (char * nick, char * from, char * who, FlagRecord * user) {
   FUNCTION;
   char me[UHOSTLEN+1], s[UHOSTLEN+1], s1[UHOSTLEN+1];
   int check, i;
   ChannelMember * m = 0;
   UserRecord * u;
   FlagRecord victim ( FlagRecord::CHANNEL | FlagRecord::GLOBAL );
   
   G->SPrintF(me, UHOSTLEN, "%s!%s", I->SERVER->botname, I->SERVER->botuserhost);
   G->SPrintF(s, UHOSTLEN, "%s!%s", nick, from);
   NEW ( ChannelBan ( this, who, s) );
   check = 1;
   if (!I->SERVER->MatchMyNick(nick)) {	/* it's not my ban */
      if (!UserBans() && !user->ChanBot() && !user->AnyMaster()) {
	 /* no bans made by users */
	 AddMode('-', 'b', who);
      } else {
	 for (i = 0; who[i]; i++) {
	    if ((who[i] < 32) || (who[i] == 127) || (!i && (who[i] == ':')))
	      break;
	    else
	      s1[i] = who[i];
	 }
	 if (who[i]) {
	    if (user->AnyFriend() || (DontKickOps() && user->AnyValidOp())) {
	       s1[i] = '?';
	       while (who[++i]) {
		  if ((who[i] < 32) || (who[i] == 127))
		    s1[i] = '?';
		  else 
		    s1[i] = who[i];
		  if ((s1[i] != '?') && (s1[i] != '*') && (s1[i] != '!') &&
		      (s1[i] != '@'))
		    check = 0;
	       }
	       s1[i] = 0;
	       AddMode('-', 'b', who);
	       FlushMode(Q_SERVER);
	       if (!check) /* only re-add it if something besides wildcards */
		 DoBan(s1);
	    } else {
	       AddMode('-', 'b', who);
	       if (KickBogus() && (m = member->Find(nick))) {
		  QuickBan(from);
		  m->Kick(I->Lang(0x0EC,""));
	       }
	    } 
	    return;
	 } 
	 /* don't enforce a server ban right away -- give channel users a
	  * chance to remove it, in case it's fake */
	 if (!nick[0])
	   check = 0;
	 /* does this remotely match against any of our hostmasks? */
	 /* just an off-chance... */
	 else if ((u = G->userlist->FindHost(who))->Handle()[0] != '*') {
	    victim.Set(u, key);
	    if (victim.AnyFriend() || victim.AnyValidOp()) {
	       if (!user->AnyMaster() && !user->ChanBot()) {
		  I->reversing = 1;
		  check = 0;
	       } else if (victim.AnyMaster())
		    check = 0;
	    } else if (G->WildMatch(who, me) && Meop()) {
		  /* ^ don't really feel like being banned today, thank you! */
	       I->reversing = 1;
	       check = 0;
	    }
	    /* banning an oplisted person who's on the channel? */
	 } else 
	   for (m = member; (m = m->Next());) {
	      G->SPrintF(s1, UHOSTLEN, "%s!%s", m->Key(), m->Host());
	      if (G->WildMatch(who, s1) && 
		  ((u = G->userlist->FindHost(s1)) != G->userlist)) {
		 victim.Set(u, key);
		 if (victim.AnyFriend() || victim.AnyValidOp()) {
		       /* remove ban on +o/f/m user, unless placed by
			another +m/b */
		    if (!user->AnyMaster() && !user->ChanBot()) {
		       AddMode('-', 'b', who);
		       check = 0;
		    } else if (victim.AnyMaster())
			 check = 0;
		 }
	      }
	   }
	 if (check && EnforceBans() && Meop())
	   KickAll(who, I->Lang(0x0D7,""));
      }
   }
   /* is it a server ban from nowhere? */
   if (I->reversing || (I->bounce_bans && !nick[0] && check && !EqualsBan(who)))
     AddMode('-', 'b', who);
}

void Channel3::GotUnban (char * nick, char * who, UserRecord * u, FlagRecord * user) {
   FUNCTION;
   int i;
   ChannelBan * b = chanbans->Find(who);
   ChannelMember * m = 0;
   
   DELETE ( b );
   for (i = 0; who[i]; i++)
     if ((who[i] < 32) || (who[i] == 127))
       break;
   /* it's bogus, not by me, and in fact didn't exist anyway.. */
   if (who[i] && !I->SERVER->MatchMyNick(nick) && !b &&
       /* not by valid +f/+b/+o */
       !user->AnyFriend() && !user->AnyValidOp()) {
      /* then lets kick the weenie */
      if (KickBogus() && (m = member->Find(nick))) 
        m->Kick(I->Lang( 0x0EC, ""));
   } else {
      FlagRecord bfl (FlagRecord::BOT);
      
      if ((i = EqualsBan(who)) == 7)
	/* that's a sticky ban! No point in being */
	/* sticky unless we enforce it!! */
	DoBan(who);
      else if (i && Meop() && !DynamicBans() &&
	       !user->ChanBot() && bfl.Set(u)->BotShare()) {
	 if (user->AnyValidOp())
	   I->SERVER->SendNotice(Q_SERVER, nick, I->Lang( 0x0ED, "s", who));
	 else 
	   DoBan(who);
      } else if (I->SERVER->use_ts4) 
	   RemoveMatchingExceptions ( who );
   }
}

int Channel3::ExpMem ( void ) {
   FUNCTION;
   int tot = 0;
   Channel *chan;

   context;
   for (chan = G->chanset; chan; chan = chan->Next())
      tot += sizeof(Channel3);
   return tot;
}

void Channel3::GotException (char * nick, char * from, char * who, FlagRecord * user) {
   FUNCTION;
   char me[UHOSTLEN+1], s[UHOSTLEN+1], s1[UHOSTLEN+1];
   int check, i;
   ChannelMember * m = 0;
   
   G->SPrintF(me, UHOSTLEN, "%s!%s", I->SERVER->botname, I->SERVER->botuserhost);
   G->SPrintF(s, UHOSTLEN, "%s!%s", nick, from);
   NEW ( ChannelBan ( this, who, s, 1) );
   check = 1;
   if (!I->SERVER->MatchMyNick(nick)) {	/* it's not my ban */
      if (!UserBans() && !user->ChanBot() && !user->AnyMaster()) {
	 /* no bans made by users */
	 AddMode('-', 'e', who);
      } else {
	 for (i = 0; who[i]; i++) {
	    if ((who[i] < 32) || (who[i] == 127) || (!i && (who[i] == ':')))
	      break;
	    else
	      s1[i] = who[i];
	 }
	 if (who[i]) {
	    if (user->AnyFriend()) {
	       s1[i] = '?';
	       while (who[++i]) {
		  if ((who[i] < 32) || (who[i] == 127))
		    s1[i] = '?';
		  else 
		    s1[i] = who[i];
		  if ((s1[i] != '?') && (s1[i] != '*') && (s1[i] != '!') &&
		      (s1[i] != '@'))
		    check = 0;
	       }
	       s1[i] = 0;
	       AddMode('-', 'e', who);
	       FlushMode(Q_SERVER);
	    } else {
	       AddMode('-', 'e', who);
	       if (KickBogus() && (m = member->Find(nick))) {
		  QuickBan(from);
		  m->Kick(I->Lang(0x127,""));
	       }
	    } 
	    return;
	 } 
      }
   }
   /* is it a server ban from nowhere? */
   if (I->reversing || (I->bounce_bans && !nick[0] && check))
     AddMode('-', 'e', who);
}

void Channel3::GotUnexception (char * nick, char * who, UserRecord *, FlagRecord * user) {
   FUNCTION;
   int i;
   ChannelBan * b = exceptions->Find(who);
   ChannelMember * m = 0;
   
   DELETE ( b );
   for (i = 0; who[i]; i++)
     if ((who[i] < 32) || (who[i] == 127))
       break;
   /* it's bogus, not by me, and in fact didn't exist anyway.. */
   if (who[i] && !I->SERVER->MatchMyNick(nick) && !b &&
       /* not by valid +f/+b/+o */
       !user->AnyFriend() && !user->AnyValidOp()) {
      /* then lets kick the weenie */
      if (KickBogus() && (m = member->Find(nick))) 
        m->Kick(I->Lang( 0x127, ""));
   }
}
