/*
 * Channel3.h - contains info relevant to irc module
 */

#ifndef _CHANNEL2_
#include "Channel2.h"
#endif

#ifndef _CHANNELSMODULE_
#include "channels.h"
#endif

#ifndef _IRCMODULE_
#include "irc.h"
#endif

#ifndef _CHANNEL3_
#define _CHANNEL3_

struct cmode_st {
   char *op;
   char type;
   char plus;
};

class Channel3 : public Channel2 {
  protected:
   ChannelMember * member;
   struct cmode_st * cmode;
   char * chankey;
   char chanpass[24];
   char * topic;
   ChannelBan * chanbans;
   ChannelBan * exceptions;
   unsigned short int mode;
   int maxmembers;
   char pls[21];             /* positive mode changes */
   char mns[21];             /* negative mode changes */
   char nkey[81];             /* new key to set */
   char rmkey[81];           /* old key to remove */
   int limit;                /* new limit to set */
   int bytes;                /* total bytes so far */
   int lastmode;
   char floodwho[FLOOD_CHAN_MAX][81];
   time_t floodtime[FLOOD_CHAN_MAX];
   int floodnum[FLOOD_CHAN_MAX];
   char deopd[NICKMAX + 1];      /* last person deop'd (must change) */
   time_t timestamp;
   
  public:
   enum { PLUS = 1, OP = 2, BAN = 4, EXC = 8, HELPR = 16 };
   enum { NORMAL = 0, QUICK = 1 };
   enum {
      CLEARBANS = 0x40, ENFORCEBANS = 0x80,
	DYNAMICBANS = 0x100, USERBANS = 0x200, OPONJOIN = 0x400, BITCH = 0x800,
	GREET = 0x1000, PROTECTOPS = 0x2000, LOGSTATUS = 0x4000, 
	STOPNETHACK = 0x8000, REVENGE = 0x10000, AUTOVOICE = 0x20000,
	CYCLE = 0x40000, BOGUS = 0x80000, DONTKICKOPS = 0x100000,
	NOCHANRECINFO = 0x200000, ASKEDBANS = 0x400000, MEHELPER = 0x800000,
	KICKFAKEMODES = 0x1000000,
   };
   
   void Init ( void );
   inline Channel3 ( void ) : Channel2 () { }
   inline Channel3 ( const char * x) : Channel2 ( x ) {
      Init();
   }
   inline Channel3 ( Channel2 & o) : Channel2 ( o ) {
      Init();
   }
   ~Channel3 ( void );
   
   int ExpMem ( void );
   void Clear ( int );
   void SetMode ( char * );
   int AnyOps ( void );
   const char * QuickBan ( const char * );
   void ResetBans ( void );
   void RecheckBans ( void );
   void SetTopic ( const char * );
   BindReturn CheckTN ( const char *, char *, UserRecord *, char *, BindTable * );
   BindReturn CheckMode ( const char *, char *, UserRecord *, char *, char *);
   BindReturn CheckJPSR ( char *, char *, UserRecord *, BindTable * );
   void GotOp ( char *, char *, FlagRecord * );
   void GotDeop ( char *, char *, char *, FlagRecord * );
   void GotHelper ( char *, char *, char *, FlagRecord * );
   void GotDeHelper ( char *, char *, char *, FlagRecord * );
   void GotBan ( char *, char *, char *, FlagRecord * );
   void GotUnban ( char *, char *, UserRecord *, FlagRecord * );
   void GotException ( char *, char *, char *, FlagRecord * );
   void GotUnexception ( char *, char *, UserRecord *, FlagRecord * );
   int DetectFlood (char *, char *, char *, int, char *);
   void ResetInfo ( void );
   void CheckLonely ( void );
   void TakeRevenge ( char *, char * );
   const char * GetMode ( void );
   void FlushMode ( int );
   void RedoCmode ( void );
   void DoBan ( const char * );
   void SetKey ( const char * );
   int HandleFlood (int, char *, char *, char * );
   void KickAll ( char *, char * );
   void RemoveMatchingExceptions ( const char * );
   
   virtual Channel2 * NewChannel ( const char * );
   virtual void AddMode ( char, char, const char * );
   virtual int NukeChanBan ( char *, int );
   virtual int ShowChanBan ( char * , int, Socket * );
   virtual void Recheck ( int );

   inline Channel3 * Next ( void ) { return (Channel3 *)next; }
   inline ChannelMember * M ( void ) { return member; }
   inline struct cmode_st * Cmode ( int i ) { return cmode + i; }
   inline const char * ChanKey ( void ) { return chankey; }
   inline const char * Topic ( void ) { return topic; }
   inline ChannelBan * ChanBans ( void ) { return chanbans; }
   inline ChannelBan * Exceptions ( void ) { return exceptions; }
   inline int MaxMembers ( void ) { return maxmembers; }
   inline void ClearMaxMembers ( void ) { maxmembers = -1; }
   inline void SetMaxMembers ( int x ) { maxmembers = x; }
   inline void AddMode ( int x ) { mode |= x; }
   inline void RemMode ( int x ) { mode &= ~x; }
   inline int Mode ( void ) { return mode; }
   inline int Hidden ( void ) { return mode & (CHANPRIV | CHANSEC); }
   inline int NoHelper ( void ) { return mode & CHANPRIV; }
   inline int OpTopic ( void ) { return mode & CHANTOPIC; }
   inline const char * Nkey ( void ) { return nkey; }
   inline void SetPass ( const char * p ) { strncpyz(chanpass, p, 23); }
   inline const char * Pass ( void ) { return chanpass; }
   inline void SetTs ( time_t val ) { timestamp = val; }
   inline int AskedBans ( void ) { return status & ASKEDBANS; }
   inline void SetAskedBans ( void ) { status |= ASKEDBANS; }
   inline void UnsetAskedBans ( void ) { status &= ~ASKEDBANS; }
   inline int MeHelper ( void ) { return status & MEHELPER; }
   inline void SetMeHelper ( void ) { status |= MEHELPER; }
   inline void UnsetMeHelper ( void ) { status &= ~MEHELPER; }
   
   inline int ClearBans ( void ) { return status & CLEARBANS; }
   inline int EnforceBans ( void ) { return status & ENFORCEBANS; }
   inline int DynamicBans ( void ) { return status & DYNAMICBANS; }
   inline int UserBans ( void ) { return status & USERBANS; }
   inline int AutoOp ( void ) { return status & OPONJOIN; }
   inline int Bitch ( void ) { return status & BITCH; }
   inline int Greet ( void ) { return status & GREET; }
   inline int ProtectOps ( void ) { return status & PROTECTOPS; }
   inline int LogStatus ( void ) { return status & LOGSTATUS; }
   inline int StopNetHack ( void ) { return status & STOPNETHACK; }
   inline int Revenge ( void ) { return status & REVENGE; }
   inline int AutoVoice ( void ) { return status & AUTOVOICE; }
   inline int Cycle ( void ) { return status & CYCLE; }
   inline int KickBogus ( void ) { return status & BOGUS; }
   inline int DontKickOps ( void ) { return status & DONTKICKOPS; }
   inline int NoChanRecInfo ( void ) { return status & NOCHANRECINFO; }
   inline int KickFakeModes ( void ) { return status & KICKFAKEMODES; }
};

inline void AddMode ( Channel * a, char b, char c, const char * d) {
   ((Channel3 *)a)->AddMode(b,c,d);
}

inline int NukeChanBan ( Channel * a, char * b, int c) {
   return ((Channel3 *)a)->NukeChanBan(b,c);
}

#ifndef _CHANNELBAN_
#include "ChannelBan.h"
#endif

inline int IsBanned ( Channel * a, const char * b) {
   return ((Channel3 *)a)->ChanBans()->Find(b) != 0;
}

inline int ShowChanBan ( Channel * a, char * b, int c, Socket * d) {
   return ((Channel3 *)a)->ShowChanBan(b,c,d);
}

inline Channel3 * FindChan3 ( const char * n ) {
   return (Channel3 *)List::G->chanset->Find(n);
}
#endif
