/*
 * ChannelMember.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ChannelMember.h"
#include "Channel3.h"
#include "UserRecord.h"
#include "ModuleEntryServer.h"
#include "BanRecord.h"
#include "Socket.h"
#define CLASS "ChannelMember"

ChannelMember::ChannelMember ( Channel3 * c, const char * n, const char * h,
			      UserRecord * u) 
: KeyedList<NICKMAX> ( *c->M(), n, APPEND | NODUPS ) {
   FUNCTION;
   strncpyz(userhost, h, UHOSTLEN);
   joined = 0;
   split = 0;
   last = G->now;
   flags = 0;
   user = u;
   chan = c;
}

ChannelMember::~ChannelMember ( void ) {
   FUNCTION;
   if (key[0])
     chan->M()->Remove(this);
   else 
     while (next) {
	DELETE ( next );
     }
}

ChannelMember * ChannelMember::FindUser ( const char * x) {
   FUNCTION;
   ChannelMember * ptr = this;
   
   while ((ptr = ptr->Next()) && (!ptr->user || strcasecmp(ptr->User()->Key(), x)));
   return ptr;
}

ChannelMember * ChannelMember::FindUser ( UserRecord * u) {
   FUNCTION;
   ChannelMember * ptr = this;
   
   while ((ptr = ptr->Next()) && (ptr->user != u));
   return ptr;
}

void ChannelMember::Kick ( const char * reason ) {
   FUNCTION;
   if (!SentKick()) {
      flags |= SENTKICK;
      I->SERVER->SendKick(Q_MODE,chan->Key(),key,reason);
   }
}
   
int ChannelMember::ExpMem ( void ) {
   FUNCTION;
   Channel3 * chan = (Channel3 *)List::G->chanset;
   ChannelBan * b;
   ChannelMember * m;
   int tot = sizeof(ChannelMember), i;
   
   while ((chan = chan->Next())) {
      tot += (sizeof(struct cmode_st) * I->modesperline);
      for (i = 0; i < I->modesperline; i++)
	if (chan->Cmode(i)->op)
	  tot += strlen(chan->Cmode(i)->op) + 1;
      if (chan->ChanKey())
	tot += strlen(chan->ChanKey()) + 1;
      if (chan->Topic())
	tot += strlen(chan->Topic()) + 1;
      for (m = chan->M(); m; m = m->Next())
	tot += sizeof(ChannelMember);
      for (b = chan->ChanBans(); b; b = b->Next())
	tot += sizeof(ChannelBan);
      for (b = chan->Exceptions(); b; b = b->Next())
	tot += sizeof(ChannelBan);
   }
   return tot;
}

/* if any bans match this wildcard expression, refresh them on the channel */
void ChannelMember::RefreshBanKick ( void ) {
   FUNCTION;
   BanRecord * b;
   int cycle;
   char s[UHOSTLEN+1];
   
   G->SPrintF(s, UHOSTLEN, "%s!%s", key, userhost);
   for (cycle = 0; cycle < 2; cycle++) {
      b = cycle ? chan->Bans() : G->global_bans;
      while ((b = b->Next())) 
	if (G->WildMatch(b->Key(),s) && (b->LastActive() < G->now - 60)) {
	   if (HasOp() && !SentDeop()) {
	      flags |= SENTDEOP;
	      chan->AddMode('-', 'o', key);  /* guess it can't hurt */
	   }
	   chan->DoBan(b->Key());
	   b->Activate();
	   Kick (I->Lang( (b->Reason() && (b->Reason()[0] != '@')) ?
			 0x0CF : 0x0D0, "s", b->Reason()));
	}
   }
}

void ChannelMember::DoEmbeddedMode (const char * mode) {
   FUNCTION;
   FlagRecord fr;
   Socket * servidx = G->socklist->Find(I->SERVER->serv);
   
   I->reversing = 0;
   while (*mode) {
      switch (*mode) {
      case 'o':
	 chan->CheckMode(servidx->Host(), "", NULL, "+o", key);
	 chan->GotOp("", key, &fr);
	 break;
      case 'v':
	 chan->CheckMode(servidx->Host(), "", NULL, "+v", key);
	 flags |= CHANVOICE;
	 break;
      }
      mode++;
   }
}

UserRecord * ChannelMember::MemberUser (void) {
   FUNCTION;
   char s[UHOSTLEN+1];
   
   if (this && !user) {
      G->SPrintF(s, UHOSTLEN, "%s!%s", key, userhost);
      user = G->userlist->FindHost(s);
   }
   return (!this || !user || user->Handle()[0] == '*') ? 0 : user;
}
