/*
 * ChannelMember.h
 */

#ifndef _KEYEDLIST_
#include "KeyedList.h"
#endif

#ifndef _MODULEENTRYIRC_
#include "ModuleEntryIrc.h"
#endif

#ifndef _CHANNEL3_
#include "Channel3.h"
#endif

#ifndef _CHANNELMEMBER_
#define _CHANNELMEMBER_

class ChannelMember : public KeyedList<NICKMAX> {
  protected:
   char userhost[UHOSTLEN + 1];
   time_t joined;
   time_t split;         /* in case they were just netsplit */
   time_t last;          /* for measuring idle time */
   int flags;
   UserRecord *user;
   Channel3 * chan;

   enum { CHANOP = 1, CHANVOICE = 2, FAKEOP = 4, SENTOP = 8,
	SENTDEOP = 0x10, SENTKICK = 0x20, SENTUWHO = 0x40, SENTDEVOICE = 0x80,
	CHANHELPER = 0x100, SENTHELPER = 0x200, SENTDEHELPER = 0x400,
	WASOP = 0x800
   };
   
  public:
   inline ChannelMember ( void ) : KeyedList<NICKMAX> () {
   }
   ChannelMember ( Channel3 *, const char *, const char *, UserRecord *);
   ~ChannelMember ( void );

   ChannelMember * FindUser ( const char * );
   ChannelMember * FindUser ( UserRecord * );
   void Kick ( const char * );
   UserRecord * MemberUser ( void );
   void DoEmbeddedMode ( const char * );
   void RefreshBanKick ( void );
   
   DEF_NEXT(ChannelMember);
   MEMORY_TYPE(I->memtype_chaninfo);
   int ExpMem ( void );
   inline ChannelMember * Find ( const char * x ) {
      return (ChannelMember *)KeyedList<NICKMAX>::Find(x);
   }
   inline UserRecord * User ( void ) { return user; }
   inline void SetUser ( UserRecord * x ) { user =  x; }
   inline const char * Host ( void ) { return userhost; }
   inline void SetNickHost ( const char * n, const char * h ) {
      strncpyz(key, n, NICKMAX);
      strncpyz(userhost, h, UHOSTLEN);
   }
   inline void Touch ( void ) { last = G->now; }
   inline void Join ( void ) { joined = G->now; }
   inline void Split ( void ) { split = G->now; }
   inline time_t SplitAge ( void ) { return G->now - split; }
   inline time_t Age ( void ) { return G->now - last; }
   inline time_t Joined ( void ) { return joined; }
   
   inline void ClearFlags ( void ) { flags = 0; }   
   inline void ClearVoice ( void ) { flags &= ~CHANVOICE; }
   inline void ClearSentWho ( void ) { flags &= ~SENTUWHO; }
   inline void ClearSentDevoice ( void ) { flags &= ~SENTDEVOICE; }
   inline void ClearSplit ( void ) { split = 0; }
   inline void ClearFakeOp ( void ) { flags &= ~FAKEOP; }
   inline void ClearOp ( void ) { flags &= ~CHANOP; }
   inline void ClearHelper ( void ) { flags &= ~CHANHELPER; }
   inline void ClearSentOp ( void ) { flags &= ~SENTOP; }
   inline void ClearSentDeOp ( void ) { flags &= ~SENTDEOP; }
   inline void ClearSentDeHelper ( void ) { flags &= ~SENTDEHELPER; }
   inline void ClearSentHelper ( void ) { flags &= ~SENTHELPER; }
   inline void ClearWasOp ( void ) { flags &= ~WASOP; }
   
   inline void SetVoice ( void ) { flags |= CHANVOICE; }
   inline void SetOp ( void ) { flags |= CHANOP; }
   inline void SetHelper ( void ) { flags |= CHANHELPER; }
   inline void SetSentOp ( void ) { flags |= SENTOP; }
   inline void SetSentKick ( void ) { flags |= SENTKICK; }
   inline void SetSentDeop ( void ) { flags |= SENTDEOP; }
   inline void SetSentWho ( void ) { flags |= SENTUWHO; }
   inline void SetFakeOp ( void ) { flags |= FAKEOP; }
   inline void SetSentDeHelper ( void ) { flags |= SENTDEHELPER; }
   inline void SetSentHelper ( void ) { flags |= SENTHELPER; }
   inline void SetWasOp ( void ) { flags |= WASOP; }
   
   inline int HasVoice ( void ) { return flags & CHANVOICE; }
   inline int HasOp ( void ) { return flags & CHANOP; }
   inline int HasHelper ( void ) { return flags & CHANHELPER; }
   inline int FakeOp ( void ) { return flags & FAKEOP; }
   inline int SentOp ( void ) { return flags & SENTOP; }
   inline int SentDeop ( void ) { return flags & SENTDEOP; }
   inline int SentKick ( void ) { return flags & SENTKICK; }
   inline int SentWho ( void ) { return flags & SENTUWHO; }
   inline int SentDevoice ( void ) { return flags & SENTDEVOICE; }
   inline int SentHelper ( void ) { return flags & SENTHELPER; }
   inline int SentDeHelper ( void ) { return flags & SENTDEHELPER; }
   inline int IsSplit ( void ) { return split > 0; }
   inline int WasOp ( void ) { return flags & WASOP; }
   
   inline void GiveOp ( void ) {
      chan->AddMode('+', 'o', key);
      SetSentOp();
      SetWasOp();
   }
   inline void TakeOp ( void ) {
      chan->AddMode('-', 'o', key);
      SetSentDeop();
   }
   inline void GiveHelper ( void ) {
      chan->AddMode('+', 'h', key);
      SetSentHelper();
   }
   inline void TakeHelper ( void ) {
      chan->AddMode('-', 'h', key);
      SetSentDeHelper();
   }
   inline void GiveVoice ( void ) {
      chan->AddMode('+', 'v', key);
   }
   inline void TakeVoice ( void ) {
      chan->AddMode('-', 'v', key);
   }
};
#endif
