/*
 * ChannelSettingsChanmode.C
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ChannelSettingsChanmode.h"
#include "ModuleEntryChannels.h"
#include "Channel2.h"
#include "Socket.h"
#include "ScriptContext.h"
#include "FlagRecord.h"
#include "ModuleEntryServer.h"
#define CLASS "ChannelSettingsChanmode"

ChannelSettingsChanmode::ChannelSettingsChanmode ( ModuleEntry * m ) 
: ChannelSettings ( m ) {
   FUNCTION;
   Channel2 * ptr = chan;
   
   pls_prot = mns_prot = limit_prot = 0;
   key_prot[0] = 0;
   while ((ptr = ptr->Next()))
     Create ( ptr );
}

ChannelSettingsChanmode::~ChannelSettingsChanmode ( void ) {
   FUNCTION;
}

int ChannelSettingsChanmode::Set (const char * arg, const char * arg2in, Socket *) {
   FUNCTION;
   int i, pos = 1;
   char *s, *s1, buf[121], * arg2 = buf;
   
   strncpyz(arg2, arg2in, 120);
   if (!strcmp(arg, "chanmode")) {
      if (!arg2)
	return -1;
      /* clear old modes */
      mns_prot = pls_prot = 0;
      limit_prot = -1;
      key_prot[0] = 0;
      for (s = G->Split(&arg2); *s; s++) {
	 i = 0;
	 switch(*s) {
	   case '+':
	    pos = 1;
	    break;
	   case '-':
	    pos = 0;
	    break;
	   case 'i':
	    i = CHANINV;
	    break;
	   case 'p':
	    i = CHANPRIV;
	    break;
	   case 's':
	    i = CHANSEC;
	    break;
	   case 'm':
	    i = CHANMODER;
	    break;
	   case 't':
	    i = CHANTOPIC;
	    break;
	   case 'n':
	    i = CHANNOMSG;
	    break;
	   case 'a':
	    i = CHANANON;
	    break;
	   case 'q':
	    i = CHANQUIET;
	    break;
	   case 'l':
	    i = CHANLIMIT;
	    limit_prot = -1;
	    if (pos && (s1 = G->Split(&arg2))[0])
	      limit_prot = atoi(s1);
	    break;
	   case 'k':
	    i = CHANKEY;
	    key_prot[0] = 0;
	    if (pos && (s1 = G->Split(&arg2))[0]) 
	      strncpyz(key_prot, s1, 120);
	    break;
	   case 'c':
	    i = CHANPASS;
	    pass_prot[0] = 0;
	    if (pos && (s1 = G->Split(&arg2))[0]) 
	      strncpyz(pass_prot, s1, 120);
	    break;
	 }
	 if (i) {
	    if (pos) {
	       pls_prot |= i;
	       mns_prot &= ~i;
	    } else {
	       pls_prot &= ~i;
	       mns_prot |= i;
	    }
	 }
      }
      if (CHNS->SERVER)
	CHNS->SERVER->SendMode(Q_MODE, chan->Key(), 0);
      return 2;
   }
   return 0;
}

void ChannelSettingsChanmode::View (Socket * idx) {
   FUNCTION;
   char work[121];
   FlagRecord user (idx->User(), FlagRecord::GLOBAL | FlagRecord::CHANNEL, 
		    chan->Key());
   
   GetModeProtect(work, user.AnyOwner());
   idx->Send(CHNS->Lang( 0x06F, "s", work));
}

int ChannelSettingsChanmode::Get ( ScriptContext * sc ) {
   FUNCTION;
   char s[121];
   
   GetModeProtect(s);
   sc->SetString(s);
   return SCRIPT_OK;
}

int ChannelSettingsChanmode::Write ( FILE * f ) {
   FUNCTION;
   char s[121];
   
   GetModeProtect(s);
   return s[0] ? fprintf(f, "channel set %s chanmode \"%s\"\n", 
			 chan->Key(), s) : 0;
}

int ChannelSettingsChanmode::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ChannelSettingsChanmode);
}

void ChannelSettingsChanmode::GetModeProtect (char * s, int owner) {
   FUNCTION;
   char *p = s, s1[121];
   int ok = 0, i, tst, l = 0;
   
   s1[0] = 0;
   for (i = 0; i < 2; i++) {
      ok = 0;
      if (i == 0) {
	 tst = pls_prot;
	 if (tst || (limit_prot != -1) || key_prot[0])
	   *p++ = '+';
	 if (limit_prot != (-1)) {
	    *p++ = 'l';
	    l += G->SPrintF(s1 + l, 120 - l, "%d ", limit_prot);
	 }
	 if (key_prot[0]) {
	    *p++ = 'k';
	    l += G->SPrintF(s1 + l, 120 - l, "%s ", key_prot);
	 }
      	 if (pass_prot[0]) {
	    *p++ = 'c';
	    l += G->SPrintF(s1 + l, 120 - l, "%s ", owner ? key_prot : "*HIDDEN*");
	 }
      } else {
	 if ((tst = mns_prot))
	   *p++ = '-';
	 if (tst & CHANKEY)
	   *p++ = 'k';
	 if (tst & CHANLIMIT)
	   *p++ = 'l';
	 if (tst & CHANPASS)
	   *p++ = 'c';
      }
      if (tst & CHANINV)
	 *p++ = 'i';
      if (tst & CHANPRIV)
	 *p++ = 'p';
      if (tst & CHANSEC)
	 *p++ = 's';
      if (tst & CHANMODER)
	 *p++ = 'm';
      if (tst & CHANTOPIC)
	 *p++ = 't';
      if (tst & CHANNOMSG)
	 *p++ = 'n';
      if (tst & CHANANON)
	 *p++ = 'a';
      if (tst & CHANQUIET)
	 *p++ = 'q';
   }
   *p = 0;
   if (s1[0]) {
      s1[l - 1] = 0;
      G->SPrintF(p, 120 - (p - s), " %s", s1);
   }
}

void ChannelSettingsChanmode::Create ( Channel2 * ch ) {
   FUNCTION;
   NEW ( ChannelSettingsChanmode ( ch, this ) );
}
