/*
 * ChannelSettingsFlags.C
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryIrc.h"
#include "ModuleEntryServer.h"
#include "ChannelSettingsFlags.h"
#include "ModuleEntryChannels.h"
#include "ScriptContext.h"
#include "Socket.h"
#include "ChannelSettingsChanmode.h"
#include "ChannelSettingsIdle.h"
#include "FlagRecord.h"
#define CLASS "ChannelSettingsChanmode"

FLAGTABLE flag_table [] = { 
   { "clearbans", Channel3::CLEARBANS, 0x13C },
   { "enforcebans", Channel3::ENFORCEBANS, 0x13D },
   { "dynamicbans", Channel3::DYNAMICBANS, 0x13E },
   { "userbans", Channel3::USERBANS, 0x13F },
   { "autoop", Channel3::OPONJOIN, 0x140 },
   { "bitch", Channel3::BITCH, 0x141 },
   { "greet", Channel3::GREET, 0x142 },
   { "protectops", Channel3::PROTECTOPS, 0x143 },
   { "statuslog", Channel3::LOGSTATUS, 0x144 },
   { "stopnethack", Channel3::STOPNETHACK, 0x145 },
   { "revenge", Channel3::REVENGE, 0x146 },
   { "secret", Channel3::SECRET, 0x147 },
   { "autovoice", Channel3::AUTOVOICE, 0x148 },
   { "cycle", Channel3::CYCLE, 0x149 },
   { "kickbogus", Channel3::BOGUS, 0x14A },
   { "dontkickops", Channel3::DONTKICKOPS, 0x14B },
   { "nochanrecinfo", Channel3::NOCHANRECINFO, 0x14C },
   { "kickfakemodes", Channel3::KICKFAKEMODES, 0x14D },
   { 0, 0 }
};

ChannelSettingsFlags::ChannelSettingsFlags ( ModuleEntry * m ) 
: ChannelSettings ( m ) {
   FUNCTION;
   Channel2 * ptr = chan;
   while ((ptr = ptr->Next()))
     Create ( ptr );
}

ChannelSettingsFlags::~ChannelSettingsFlags ( void ) {
   FUNCTION;
}

int ChannelSettingsFlags::Set (const char * arg, const char *, Socket *) {
   FUNCTION;
   int i;
   
   if ((arg[0] == '+') || (arg[0] == '-')) {
      for (i = 0; flag_table[i].name; i++) 
	if (!strcmp(flag_table[i].name,arg+1)) {
	   if ((I->SERVER->use_ts4) && 
	       (flag_table[i].value == Channel3::STOPNETHACK))
	     G->LOG_MISC.Put(I->Lang(0x13B, ""));
	   if (arg[0] == '+')
	     chan->AddStatus(flag_table[i].value);
	   else
	     chan->RemStatus(flag_table[i].value);
	   return 1;
	}
   }
   return 0;
}
	 
void ChannelSettingsFlags::View (Socket * idx) {
   FUNCTION;
   int status = chan->Status();
   idx->Send(I->Lang( 0x14E, ""));
   idx->Send(I->Lang( 0x14F, "cccc",
		     (status & Channel3::CLEARBANS) ? '+' : '-',
		     (status & Channel3::ENFORCEBANS) ? '+' : '-',
		     (status & Channel3::DYNAMICBANS) ? '+' : '-',
		     (status & Channel3::USERBANS) ? '+' : '-'));
   idx->Send(I->Lang( 0x150, "cccc", 
		     (status & Channel3::OPONJOIN) ? '+' : '-',
		     (status & Channel3::BITCH) ? '+' : '-',
		     (status & Channel3::GREET) ? '+' : '-',
		     (status & Channel3::PROTECTOPS) ? '+' : '-'));
   idx->Send(I->Lang( 0x151, "cccc", 
		     (status & Channel3::LOGSTATUS) ? '+' : '-',
		     (status & Channel3::STOPNETHACK) ? '+' : '-',
		     (status & Channel3::REVENGE) ? '+' : '-',
		     (status & Channel3::SECRET) ? '+' : '-'));
   idx->Send(I->Lang( 0x152, "cccc",
		     (status & Channel3::AUTOVOICE) ? '+' : '-',
		     (status & Channel3::CYCLE) ? '+' : '-',
		     (status & Channel3::BOGUS) ? '+' : '-',
		     (status & Channel3::DONTKICKOPS) ? '+' : '-'));
   idx->Send(I->Lang(0x153,"cc",
		     (status & Channel3::NOCHANRECINFO) ? '+' : '-',
		     (status & Channel3::KICKFAKEMODES) ? '+' : '-'));
}

int ChannelSettingsFlags::Get ( ScriptContext * sc) {
   FUNCTION;
   int i;
   char s[100];
   const char *t = s;
   
   for (i = 0; flag_table[i].name; i++) {
      G->SPrintF(s, 99, "%c%s",  
		 (chan->Status() & flag_table[i].value) ? '+' : '-',
		 flag_table[i].name);
      sc->AppendArray(1, &t);
   }
   return SCRIPT_OK;
}

int ChannelSettingsFlags::Write ( FILE * f) {
   FUNCTION;
   int i;
   
   fprintf(f, "channel set %s ", chan->Key());
   for (i = 0; flag_table[i].name; i++) 
     if (fprintf(f, "%c%s ",  (chan->Status() & flag_table[i].value) 
		 ? '+' : '-', flag_table[i].name) == EOF)
       return EOF;
   return fprintf(f, "\n");
}

void ChannelSettingsFlags::Create ( Channel2 * ch ) {
   FUNCTION;
   NEW ( ChannelSettingsFlags ( ch, this ) );
}

int ChannelSettingsFlags::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ChannelSettingsFlags);
}
