/*
 * ModuleEntryIrc.C 
*/
/*
 * This file is part of the eggdrop source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
*/

#include "ModuleEntryIrc.h"
#include "ModuleEntryServer.h"
#include "ModuleEntryChannels.h"
#include "Channel3.h"
#include "IntegerVariable.h"
#include "HelpReference.h"
#include "BindTableSTNM.h"
#include "BindTableJPSR.h"
#include "BindTableKick.h"
#include "BindTableNeed.h"
#include "FlagRecord.h"
#include "Socket.h"
#include "UserRecord.h"
#include "ChannelSettingsChanmode.h"
#include "ChannelSettingsFlood.h"
#include "ChannelSettingsFlags.h"
#include "ChannelSettingsIdle.h"
#include "ChannelMember.h"
#define CLASS "ModuleEntryIrc"

/* version numbers and all that */
static Version irc_ver;
static struct patches irc_patches[] = {
#include "patch.h"
};

ModuleEntryIrc::ModuleEntryIrc ( ModuleEntry * old, char ** r) 
: ModuleEntry ( "irc", irc_ver.Set(2,0), irc_patches, old ) {
   FUNCTION;
   Version required_channels, required_server;

   I = this;
   if (!(SERVER = (ModuleEntryServer *) 
	 Depend("server", required_server.Set(2,0))))
     *r = "You need the server module to use the irc module.";
   else if (!(CHANNELS = (ModuleEntryChannels *) 
	      Depend("channels", required_channels.Set(2,0))))
     *r = "You need the channels module to use the irc module.";
   else {
      Channel2 * chan, *cn = (Channel2 *)G->chanset->Next();

      LOG_JOIN = LogModes::JOIN;
      LOG_MODES = LogModes::MODES;
      LOG_PUBLIC = LogModes::PUBLIC;
      modesperline = 3;
      mode_buf_len = 200;
      invite_key = 0;
      reversing = 0;
      bounce_bans = 0;
      wait_split = 300;
      learn_users = 0;
      wait_info = 15;
      memtype_chaninfo = NEW ( MemType ("chaninfo", (List **)&null_member, this) );
      null_member = NEW ( ChannelMember () );
      NEW ( IntegerVariable ("learn_users", &learn_users, 0, this) );
      NEW ( IntegerVariable ("wait_split", &wait_split, 0, this) );
      NEW ( IntegerVariable ("wait_info", &wait_info, 0, this) );
      NEW ( IntegerVariable ("bounce_bans", &bounce_bans, 0, this) );
      NEW ( IntegerVariable ("modes_per_line", &modesperline, VAR_PROTECT, this) );
      NEW ( IntegerVariable ("mode_buf_length" , &mode_buf_len, 0, this) );
      NEW ( IntegerVariable ("invite_key", &invite_key, 0, this) );
      NEW ( HelpReference("irc.help",this) );
      context;
      CHANSET_FLOOD = NEW ( ChannelSettingsFlood ( this ) );
      CHANSET_CHANMODE = NEW ( ChannelSettingsChanmode ( this ) );
      CHANSET_IDLE = NEW ( ChannelSettingsIdle ( this ) );
      CHANSET_FLAGS = NEW ( ChannelSettingsFlags ( this ) );
      context;
      NEW ( Channel3 () );
      while ((chan = cn)) {
	 cn = chan->Next();
	 NEW ( Channel3 ( *chan ) );
	 DELETE ( chan );
      }
      context;
      bt_topc = NEW ( BindTableSTNM ( "topc", this ) );
      bt_splt = NEW ( BindTableJPSR ( "splt", this ) );
      bt_sign = NEW ( BindTableSTNM ( "sign", this ) );
      bt_rejn = NEW ( BindTableJPSR ( "rejn", this ) );
      bt_part = NEW ( BindTableSTNM ( "part", this ) );
      bt_nick = NEW ( BindTableSTNM ( "nick", this ) );
      bt_mode = NEW ( BindTableKick ( "mode", this ) );
      bt_kick = NEW ( BindTableKick ( "kick", this ) );
      bt_join = NEW ( BindTableJPSR ( "join", this ) );
      bt_pubm = NEW ( BindTableSTNM ( "pubm", this ) );
      bt_pub = NEW ( BindTableSTNM ( "pub", this, 1 ) );
      bt_need = NEW ( BindTableNeed ( this ) );
      context;
   }
}

ModuleEntryIrc::~ModuleEntryIrc ( void ) {
   FUNCTION;
   Channel3 * cn = (Channel3 *)G->chanset->Next(), * chan;
   
   context;
   while ((chan = cn)) {
      cn = chan->Next();
      SERVER->SendPart(Q_MODE, chan->Key(), Lang( 0x0D3, ""));
      NEW ( Channel2 ( *chan ) );
      DELETE ( chan );
   }
   NEW ( Channel2 ( ) );
   context;
   G->check_chanlist = 0;
   context;
   DELETE ( null_member );
}

void ModuleEntryIrc::Report (Socket * idx, int details) {
   FUNCTION;
   char s[512], s2[121];
   int i, j;
   Channel3 * chan = (Channel3 *)G->chanset;
   FlagRecord user ( FlagRecord::GLOBAL | FlagRecord::CHANNEL );
   
   while ((chan = chan->Next())) {
      if (idx != G->Stdout)
	user.Set(idx->User(),chan->Key());
      if ((idx == G->Stdout) || user.AnyMaster()) {
	 if (!chan->Active()) 
	   strcpy(s, Lang( 0x11C, ""));
	 else if (chan->Pending())
	   strcpy(s, Lang( 0x11D, ""));
	 else if (!chan->Meop())
	   strcpy(s, Lang( 0x11E, ""));
	 else 
	   s[0] = 0;
	 idx->Send(Lang( 0x11F, "sds", chan->Key(),chan->M()->Count(), s));
      }
      if ((idx == G->Stdout) || user.AnyMaster()) {
	 ChannelSettingsChanmode * chansets = 
	   (ChannelSettingsChanmode *)chan->FindSettings(CHANSET_CHANMODE);
	 chansets->GetModeProtect(s2);
	 if (details) {
	    i = 0;
	    for (j = 0; flag_table[j].name; j++)
	      if (chan->Status() & flag_table[j].value)
		i += G->SPrintF(s + i, 511 - i, "%s ",
				Lang( flag_table[j].lang, ""));
	    if (!chan->Static())
	      i += G->SPrintF(s + i, 511 - i, "%s ",
			      Lang( 0x8F, ""));
	 } else 
	   s[0] = 0;
	 idx->Send(CHANNELS->Lang( details ? 0x8D : 0x08E, "sss", 
			   chan->Key(), s, s2));
	 if (details) {
	    ChannelSettingsIdle * chansets =
	      (ChannelSettingsIdle *)chan->FindSettings(CHANSET_IDLE);
	    chansets->View(idx);
	 }
      }
   }
   if (details)
     idx->Send(CHANNELS->Lang( 0x08C, "d", CHANNELS->ban_time));
}

int ModuleEntryIrc::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ModuleEntryIrc) + ModuleEntry::ExpMem ();
}
