/*
 * irc.c - part of irc.mod
 * support for channel intereaction on irc
 */

#define MAKING_IRC
#include "DccChat.h"
#include "UserEntryComment.h"
#include "UserEntryXtra.h"
#include "UserEntryInfo.h"
#include "UserEntryLaston.h"
#include "BindTable.h"
#include "BanRecord.h"
#include "HelpReference.h"
#include "HookList.h"
#include "irc.h"
#include "server.h"
#include "channels.h"
#include "BindTableNeed.h"
#include "BindTableSTNM.h"
#include "BindTableJPSR.h"
#include "BindTableKick.h"
#include "ModuleEntryIrc.h"
#include "ModuleEntryServer.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "Channel3.h"
#include "ChannelRecord.h"
#include "ChannelSettingsChanmode.h"
#include "ChannelSettingsIdle.h"
#include "Command.h"
#include "ChannelBan.h"
#include "ChannelMember.h"
#define CLASS ""

static ModuleEntryEggdrop * G;
ModuleEntryIrc * I;

extern "C" {
   char * irc_start (ModuleEntryEggdrop *, ModuleEntry *);
};

static FlagRecord user (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
static FlagRecord victim (FlagRecord::GLOBAL | FlagRecord::CHANNEL);

/***************************************************************************
 * RUNTIME CHANNEL INFO                                                    *
 ***************************************************************************/

/* chanlist checking multi-hook */
static UserRecord * irc_chanlist (char * nick, UserRecord * rec, int mode) {
   FUNCTION;
   Channel3 *chan = (Channel3 *)G->chanset;
   ChannelMember * m = 0;
   char * uhost, s[UHOSTLEN + 1];
   
   context;
   switch (mode) {   
     case CHANLIST_HAND:
      while ((chan = chan->Next()))
	if ((m = chan->M()->FindUser(nick)))
	  return m->User();
      break;
     case CHANLIST_GETHOST:
      if ((uhost = strchr(nick, '!'))) {
	 *uhost ++ = 0;
	 while ((chan = chan->Next()))
	   if ((m = chan->M()->Find(nick)) && !strcasecmp(uhost, m->Host()))
	     break;
	 *--uhost = '!';
	 if (m)
	   return m->User();
      }
      break;
     case CHANLIST_SETHOST:
      if ((uhost = strchr(nick, '!'))) {
	 *uhost ++ = 0;
	 while ((chan = chan->Next()))
	   if ((m = chan->M()->Find(nick)) && !strcasecmp(uhost, m->Host()))
	     m->SetUser(rec);
	 *--uhost = '!';
      }
      break;
     case CHANLIST_DELHOST:
      while ((chan = chan->Next()))
	for (m = chan->M(); (m = m->Next());)
	  if (m->User() != rec) {
	     G->SPrintF(s, UHOSTLEN, "%s!%s", m->Key(), m->Host());
	     if (G->WildMatch(nick, s))
	       m->SetUser(0);
	  }
      break;
     case CHANLIST_DELHAND:
      while ((chan = chan->Next()))
	for (m = chan->M(); (m = m->Next());)
	  if (m->User() == rec) {
	     if (!nick)
	       m->SetUser(0);
	     else {
		G->SPrintF(s, UHOSTLEN, "%s!%s", m->Key(), m->Host());
		if (G->WildMatch(nick, s))
		  m->SetUser(0);
	     }
	  }
      break;
     case CHANLIST_CLEAR:
      while ((chan = chan->Next()))
	for (m = chan->M(); (m = m->Next());)
	  m->SetUser(0);
      break;
   }
   return 0;
}

/**************************************************************************
 * SERVER MESSAGES                                                        *
 **************************************************************************/

static BindReturn got001 (void **, const char *) {
   FUNCTION;
   Channel3 * chan = (Channel3 *)G->chanset;
   
   while ((chan = chan->Next())) {
      chan->Clear(1); /* make sure it's all cleaned up */
      chan->UnsetActive();
      chan->UnsetPending();
      I->SERVER->SendJoin(Q_SERVER,chan->Key(),chan->ChanKey());
   }
   return EXECUTED;
}

/* <server> 324 <to> <channel> <mode> */
static BindReturn got324 (void ** args, const char *) {
   FUNCTION;
   char *chname, *msg = (char *)args[2];
   Channel3 *chan;
   
   G->Split(&msg);
   if (!(chan = FindChan3(chname = G->Split(&msg)))) {
      G->LOG_MISC.Put(I->Lang( 0x0D5, ""));
      I->SERVER->SendPart(Q_SERVER, chname, I->Lang( 0x0D6, ""));
   } else 
     chan->SetMode(msg);
   chan->Recheck(0);
   return EXECUTED;
}

/* <server> 325 <user> <channel> <chanpass> */
/* <server> 326 <user> <channel> :No password set */
/* <server> 326 <user> <channel> :Invalid password */
static BindReturn got325or6 ( void ** args, const char * mask ) {
   FUNCTION;
   char * chname, * msg = (char *)args[2], *pass;
   Channel3 * chan;
   ChannelSettingsChanmode * chanmode;
   
   context;
   G->Split(&msg);
   if ((chan = FindChan3(chname = G->Split(&msg)))) {
      pass = (mask[2] == '5') ? G->Split(&msg) : "";
      chan->SetPass(pass);
      chanmode = (ChannelSettingsChanmode *)
	chan->FindSettings(I->CHANSET_CHANMODE);
      if (chanmode->Pass()[0] && strcmp(chanmode->Pass(), pass))
	chan->AddMode('+', 'c', chanmode->Pass());
      else if ((mask[2] == '5') && (chanmode->Minus() & CHANPASS))
	chan->AddMode('-', 'c', "");
   }
   return EXECUTED;
}

/* <server> 329 <channel> <channelts> [passts] [oplessts] */
static BindReturn got329 ( void ** args, const char * ) {
   FUNCTION;
   char * chname, *msg = (char *)args[2];
   Channel3 * chan;
   
   G->Split(&msg);
   if ((chan = FindChan3(chname = G->Split(&msg))))
     chan->SetTs(atoi(G->Split(&msg)));
   return EXECUTED;
}
   
/* 352/354 who info */
static BindReturn got352or4 (void ** args, const char * mask) {
   FUNCTION;
   char *nick, *user, *host, *chname, *flags, *msg = (char *)args[2];
   char userhost[UHOSTLEN+1], userhost2[UHOSTLEN+1];
   Channel3 *chan;
   ChannelMember *m;
   int waschanop, isme = 0, washelper, use_ts4 = I->SERVER->use_ts4;
   void * argv;
   
   if ((mask[2] == '2') || I->SERVER->use_354) {
      G->Split(&msg);
      if (((msg[0] == '#') || (msg[0] == '&'))
	  && (chan = FindChan3(chname = G->Split(&msg)))) {
	 user = G->Split(&msg);
	 host = G->Split(&msg);
	 if (mask[2] == '2')
	   G->Split(&msg);	
	 nick = G->Split(&msg);	
	 flags = G->Split(&msg);
	 G->SPrintF(userhost2, UHOSTLEN, "%s@%s", user, host);
	 I->SERVER->FixFrom(userhost2);
	 if (!(m = chan->M()->Find(nick))) 	/* In my channel list copy? */
	   m = NEW ( ChannelMember (chan, nick, userhost2, 0 ) );
	 else 
	   m->SetNickHost(nick, userhost2);
	 G->SPrintF(userhost, UHOSTLEN, "%s!%s", nick, userhost2);
	 if ((isme = I->SERVER->MatchMyNick(nick)))
	   strcpy(I->SERVER->botuserhost, userhost2);
	 waschanop = chan->Meop();	/* Am I opped here? */
	 washelper = chan->MeHelper();
	 m->ClearFlags();
	 if (isme) {
	    chan->UnsetMeop();
	    chan->UnsetMeHelper();
	 }
	 while (*flags) {
	    switch (*flags) {
	      case '@':
	       m->SetOp();
	       if (isme)
		 chan->SetMeop();
	       break;
	      case '+':
	       m->SetVoice();
	       break;
	      case '%':
	       if (use_ts4) {
		  m->SetHelper();
		  if (isme)
		    chan->SetMeHelper();
	       }
	       break;
	    }
	    flags++;
	 }
	 if (isme && ((!waschanop && chan->Meop()) ||
		      (use_ts4 && !washelper && chan->MeHelper())))
	   chan->Recheck(1);
	 if (isme && !chan->Meop() && chan->AnyOps()) {
	    argv = chan;
	    I->bt_need->Check("op", 0, &argv);
	 }
	 victim.Set(m->MemberUser(),chan->Key());
	 /* are they a chanop, and me too & they're +d or non+o on bitch*/
	 if (m->HasOp() && chan->Meop() && !isme && 
	     (victim.AnyValidDeop() || ((chan->Bitch() || use_ts4)
					&& !victim.AnyValidOp())))
	   m->TakeOp();
	 else if (use_ts4 && m->HasHelper() && !isme
		  && (chan->Meop() || (!chan->NoHelper() && chan->MeHelper()))
		  && (victim.AnyValidDeop() || (chan->Bitch() 
						&& !victim.AnyValidHelper())))
	   m->TakeHelper();
	 /* if channel is enforce bans & user matches a ban */
	 if (chan->EnforceBans() && !isme && 
	     (chan->Meop() || (use_ts4 && chan->MeHelper() && !m->HasOp())
	      && !victim.AnyFriend() && !(chan->DontKickOps() 
					  && victim.AnyValidOp()))
	     && IsBanned(chan,userhost))
	   m->Kick(I->Lang( 0x0D7,""));
	 else if (victim.AnyKick() && !isme && 
		  (chan->Meop() || (use_ts4 && chan->MeHelper() && !m->HasOp()))) {
	    UserEntryComment * uec = FindComment(m->User());
	    
	    chan->QuickBan(userhost);
	    m->Kick((uec && uec->Value() && uec->Value()[0]) ? 
		    uec->Value() : I->Lang( 0x0D1, ""));
	 }
      }
   }
   return EXECUTED;
}

/* got 315: end of who */
/* <server> 315 <to> <chan> :End of /who */
static BindReturn got315 (void ** args, const char *) {
   FUNCTION;
   char * msg = (char *)args[2];
   Channel3 *chan;
   
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg))) && chan->Pending()) {
      chan->SetActive();
      chan->UnsetPending();
      if (!chan->Meop()) {
	 void * argv = chan;
	 
	 I->bt_need->Check("op", 0, &argv);
      }
      if (!chan->M()->Find(I->SERVER->botname)) {
	 LogModes x ( LogModes::MISC | LogModes::JOIN );
	 
	 x.Put(I->Lang(0x0D8,"s",chan->Key()), chan);
	 chan->Clear(1);
	 chan->UnsetActive();
	 I->SERVER->SendJoin(Q_MODE, chan->Key(), safe_str(chan->ChanKey()));
      }
   }
   return EXECUTED;
}

/* got 348: exception info */
/* <server> 348 <to> <chan> <ban> [placed-by] [timestamp] */
static BindReturn got348 (void ** args, const char *) {
   FUNCTION;
   char * exc, * who, * msg = (char *)args[2];
   Channel3 *chan;
   
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg))) && 
       (chan->Pending() || chan->Active())) {
      exc = G->Split(&msg);
      who = G->Split(&msg);
      /* extended timestamp format? */
      NEW ( ChannelBan ( chan, exc, who[0] ? who : "existent", 1) );
   }
   return EXECUTED;
}

/* got 367: ban info */
/* <server> 367 <to> <chan> <ban> [placed-by] [timestamp] */
static BindReturn got367 (void ** args, const char *) {
   FUNCTION;
   char s[UHOSTLEN+1], * ban, * who, * msg = (char *)args[2];
   Channel3 *chan;
   UserRecord * u;
   
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg))) && 
       (chan->Pending() || chan->Active())) {
      ban = G->Split(&msg);
      who = G->Split(&msg);
      /* extended timestamp format? */
      NEW ( ChannelBan ( chan,ban, who[0] ? who : "existent") );
      G->SPrintF(s, UHOSTLEN, "%s!%s", I->SERVER->botname, I->SERVER->botuserhost);
      if (G->WildMatch(ban, s))
	chan->AddMode('-', 'b', ban);
      if (((u = G->userlist->FindHost(ban))->Handle()[0] != '*') 
	  && user.Set(u,chan->Key())->AnyValidOp())
	chan->AddMode('-', 'b', ban);
   }
   return EXECUTED;
}

/* got 368: end of ban list */
/* <server> 368 <to> <chan> :etc */
static BindReturn got368 (void ** args, const char *) {
   FUNCTION;
   Channel3 *chan;
   char * msg = (char *)args[2];
   const char * bc;
   ChannelBan *b;
   
   /* ok, now add bans that i want, which aren't set yet */
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg)))) {
      chan->UnsetAskedBans();
      if (chan->ClearBans())
	chan->ResetBans();
      else if (chan->Meop() || (I->SERVER->use_ts4 && chan->MeHelper())) {
	 for (b = chan->ChanBans(); b; b = b->Next()) {
	    if (*(b->Key()) == ':') 
	      chan->AddMode('-', 'b', b->Key());
	    else
	      for (bc = b->Key(); *bc; bc++)
		if ((*bc < 32) || (*bc > 126)) {
		   chan->AddMode('-', 'b', b->Key());
		   break;
		}
	 }
	 chan->RecheckBans();
      }
   }
   return EXECUTED;
}

/* too many channels */
static BindReturn got405 (void **, const char *) {
   FUNCTION;
   G->LOG_MISC.Put(I->Lang( 0x0D9, ""));
   return EXECUTED;
}

/* got 471: can't join channel, full */
/* got 473: can't join channel, invite only */
/* got 474: can't join channel, banned */
/* got 475: can't goin channel, bad key */
static BindReturn got47x (void ** args, const char * mask) {
   FUNCTION;
   char * msg = (char *)args[2], *which = 0;
   Channel3 *chan;
   void * argv;
   
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg)))) {
      argv = chan;
      switch (mask[3]) {
	case '1':
	 which = "limit";
	 break;
	case '3':
	 which = "invite";
	 break;
	case '4':
	 which = "unban";
	 break;
	case '5':
	 which = "key";
	 break;
      }
      if (which)
	I->bt_need->Check(which, 0, &argv);
   }
   I->LOG_JOIN.Put(I->Lang( 0x0DA + (mask[3] - '1'), "s", chan->Key()), chan);
   return EXECUTED;
}

/* got 442: not on channel */
static BindReturn got442 (void ** args, const char *) {
   FUNCTION;
   char *msg = (char *)args[2];
   Channel3 *chan;
   
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg)))) {
      G->LOG_MISC.Put(I->Lang( 0x0DB, "s", chan->Key()), chan);
      chan->Clear(1);
      chan->UnsetActive();
      I->SERVER->SendJoin(Q_MODE, chan->Key(), safe_str(chan->ChanKey()));
   }
   return EXECUTED;
}

/* got invitation */
static BindReturn gotinvite (void ** args, const char *) {
   FUNCTION;
   char * nick, *msg = (char *)args[2], *from = (char *)args[0];
   Channel3 *chan;
   static char lastnick[NICKMAX + 1] = "*";
   static time_t lastinv = (time_t) 0L;
   
   G->Split(&msg);
   msg = FixColon(msg);
   nick = G->SplitNick(&from);
   /* come on.  more than one invite from someone within 30 seconds? */
   /* just ignore them -- don't even log it. */
   if (strcasecmp(lastnick, nick) || (G->now - lastinv >= 30)) {
      G->LOG_MISC.Put(I->Lang( 0x0DF, "sss", nick, from, msg));
      strncpy(lastnick, nick, NICKMAX);
      lastnick[NICKMAX] = 0;
      lastinv = G->now;
      if ((chan = FindChan3(msg)) && (chan->Pending() || chan->Active()))
	I->SERVER->SendNotice(Q_SERVER, nick, I->Lang( 0x0E0, ""));
      else if (chan) 
	I->SERVER->SendJoin(Q_MODE,chan->Key(),safe_str(chan->ChanKey()));
   }
   return EXECUTED;
}

/* topic change */
static BindReturn gottopic (void ** args, const char *) {
   FUNCTION;
   char * nick, *from = (char *)args[0], *msg = (char *)args[2];
   ChannelMember *m;
   Channel3 *chan;
   UserRecord * u;
   
   if ((chan = FindChan3(G->Split(&msg)))) {
      msg = FixColon(msg);
      nick = G->SplitNick(&from);
      I->LOG_JOIN.Put(I->Lang(0x0E1,"ssss",chan->Key(),nick,from,msg), chan);
      if ((m = chan->M()->Find(nick)))
	m->Touch();
      if ((u = G->userlist->FindHost(from))->Handle()[0] == '*')
	u = 0;
      chan->SetTopic(msg);
      chan->CheckTN(nick,from,u,msg,I->bt_topc);
   }
   return EXECUTED;
}

/* 331: no current topic for this channel */
/* <server> 331 <to> <chname> :etc */
/* 332: topic on a channel i've just joined */
/* <server> 332 <to> <chname> :topic goes here */
static BindReturn got33x (void ** args, const char *) {
   FUNCTION;
   Channel3 *chan;
   char * msg = (char *)args[2];
   
   G->Split(&msg);
   if ((chan = FindChan3(G->Split(&msg)))) {
      msg = FixColon(msg);
      chan->SetTopic(msg);
      chan->CheckTN("*", "*", 0, msg, I->bt_topc);
   }
   return EXECUTED;
}

static BindReturn gotmode (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], buf[512], * msg = buf;
   char * nick, *ch, *op, *chg, ms2[3];
   UserRecord * u;
   ChannelMember * m;
   Channel3 *chan;
   ChannelSettingsChanmode * chanmode;
   int i = 0, use_ts4 = I->SERVER->use_ts4;
   
   strcpy(msg, (char *)args[2]);
   if ((msg[0] == '#') || (msg[0] == '&')) {
      if (!(chan = FindChan3(ch = G->Split(&msg)))) {
	 G->LOG_MISC.Put(I->Lang( 0x0EE, "s", ch));
	 I->SERVER->SendPart(Q_SERVER, ch, I->Lang( 0x0EF, ""));
      } else if (!chan->Pending()) {
	 i = strlen(msg);
	 while (msg[--i] == ' ')
	   msg[i] = 0;
	 chg = G->Split(&msg);
	 I->reversing = 0;
	 I->LOG_MODES.Put(I->Lang( 0x0F0, "ssss",ch,chg,msg,from),chan);
	 if ((u = G->userlist->FindHost(from)) == G->userlist)
	   u = 0;
	 user.Set(u,ch);
	 nick = G->SplitNick(&from);
	 m = chan->M()->Find(nick);
	 if (m) {
	    i = 0;
	    if (use_ts4) {
	       if (chan->Meop() || (chan->MeHelper() && !m->HasOp()))
		 i = m->FakeOp() ? 0x0F1 : (m->HasHelper() || m->HasOp()) 
		   ? 0 : 0x0F3;
	    }  else if (chan->Meop())
		 i = m->FakeOp() ? 0x0F1 : (m->HasOp() ? 0 : 0x0F3);
	    if (i) {
	       I->LOG_MODES.Put(I->Lang( i, "s", ch), chan);
	       if (chan->KickFakeModes()) {
		  chan->QuickBan(from);
		  m->Kick(I->Lang(i + 1, ""));
	       }
	       I->reversing = 1;
	    }
	 }
	 chanmode = (ChannelSettingsChanmode *)
	   chan->FindSettings(I->CHANSET_CHANMODE);
	 ms2[0] = '+';
	 ms2[2] = 0;
	 while ((ms2[1] = *chg)) {
	    int todo =0;
	    
	    switch (*chg) {
	     case '+':
	       ms2[0] = '+';
	       break;
	     case '-':
	       ms2[0] = '-';
	       break;
	     case 'i':
	       todo = CHANINV;
	       break;
	     case 'p':
	       todo = CHANPRIV;
	       break;
	     case 's':
	       todo = CHANSEC;
	       break;
	     case 'm':
	       todo = CHANMODER;
	       break;
	     case 't':
	       todo = CHANTOPIC;
	       break;
	     case 'n':
	       todo = CHANNOMSG;
	       break;
	     case 'a':
	       todo = CHANANON;
	       break;
	     case 'q':
	       todo = CHANQUIET;
	       break;
	      case 'z':
	       todo = CHANZOMBIE;
	       break;
	     case 'l':
	       if (ms2[0] == '-') {
		  chan->CheckMode(nick, from, u, ms2, "");
		  if (I->reversing && (chan->MaxMembers() != -1))
		    chan->AddMode('+', 'l', 
				   G->ToBase10(chan->MaxMembers()));
		  else if ((chanmode->Limit() != -1) && !user.AnyMaster())
		    chan->AddMode('+', 'l',  G->ToBase10(chanmode->Limit()));
		  chan->ClearMaxMembers();
	       } else {
		  op = G->Split(&msg);
		  chan->SetMaxMembers(atoi(op));
		  chan->CheckMode(nick, from, u, ms2, 
				  G->ToBase10(chan->MaxMembers()));
		  if (I->reversing ||
		      ((chanmode->Minus() & CHANLIMIT) && !user.AnyMaster())) {
		     if (chan->MaxMembers() <= 0)
		       chan->AddMode('+', 'l', "23"); /* wtf? 23 ??? */
		     chan->AddMode('-', 'l', "");
		  }
		  if ((chanmode->Limit() != chan->MaxMembers()) &&
		      (chanmode->Plus() & CHANLIMIT) && !user.AnyMaster())
		    chan->AddMode('+', 'l',  G->ToBase10(chanmode->Limit()));
	       }
	       break;
	      case 'k':
	       op = G->Split(&msg);
	       chan->CheckMode(nick, from, u, ms2, op);
	       if (ms2[0] == '+') {
		  chan->SetKey(op);
		  for (i = 0; op[i]; i++)
		    if ((op[i] < 32) || (op[i] == 127))
		      break;
		  if (op[i] && !I->SERVER->MatchMyNick(nick)) {
		     I->LOG_MODES.Put(I->Lang( 0x0F5, "s",chan->Key()), chan);
		     if (chan->KickBogus() && (m = chan->M()->Find(nick))
			 && !user.AnyFriend()) {
			chan->QuickBan(from);
			m->Kick(I->Lang( 0x0F6, ""));
		     }
		  }
		  if (I->reversing || op[i] || 
		      ((chanmode->Minus() & CHANKEY) && !user.AnyMaster()
		       && !user.ChanBot()))
		    chan->AddMode('-', 'k', op);
	       } else {
		  if (I->reversing && chan->Key())
		    chan->AddMode('+', 'k', chan->Key());
		  else if (chanmode->Key() && !user.AnyMaster())
		    chan->AddMode('+', 'k', chanmode->Key());
		  chan->SetKey(0);
	       }
	       break;
	      case 'c':
	       if (use_ts4 && chan->Meop()) {
		  chan->CheckMode(nick, from, u, ms2, "");
		  if (ms2[0] == '+') {
		     if (I->reversing || 
			 ((chanmode->Minus() & CHANPASS) && !user.AnyMaster()
			  && !user.ChanBot()))
		       chan->AddMode('-', 'c', "");
		     else
		       chan->AddMode('+', 'c', "");
		  } else {
		     if (I->reversing && chan->Pass())
		       chan->AddMode('+', 'c', chan->Pass());
		     else if (chanmode->Pass()[0] && !user.AnyMaster())
		       chan->AddMode('+', 'c', chanmode->Pass());
		     chan->SetPass("");
		  }
	       }
	       break;
	      case 'o':
	       op = G->Split(&msg);
	       chan->CheckMode(nick, from, u, ms2, op);
	       if (ms2[0] == '+')
		 chan->GotOp(nick, op, &user);
	       else
		 chan->GotDeop(nick, from, op, &user);
	       break;
	      case 'h':
	       op = G->Split(&msg);
	       if (use_ts4) {
		  chan->CheckMode(nick, from, u, ms2, op);
		  if (ms2[0] == '+')
		    chan->GotHelper(nick, from, op, &user);
		  else
		    chan->GotDeHelper(nick, from, op, &user);
	       }
	       break;
	      case 'v':
	       op = G->Split(&msg);
	       if (!(m = chan->M()->Find(op))) {
		  G->LOG_MISC.Put(I->Lang(0x0E2,"ss", chan->Key(),op),chan);
		  I->SERVER->SendWho(Q_MODE, op);
	       } else {
		  chan->CheckMode(nick, from, u, ms2, op);
		  victim.Set(m->User(), chan->Key());
		  if (ms2[0] == '+') {
		     m->SetVoice();
		     if (user.AnyMaster() && (I->reversing ||
					      (chan->AutoVoice() 
					       && victim.AnyValidQuiet())))
		       m->TakeVoice();
		  } else {
		     m->ClearVoice();
		     if (!user.AnyMaster() && (I->reversing || 
					       (chan->AutoVoice() 
						&& victim.AnyValidVoice())))
		       m->GiveVoice();
		  }
	       }
	       break;
	      case 'b':
	       op = G->Split(&msg);
	       chan->CheckMode(nick, from, u, ms2, op);
	       if (ms2[0] == '+')
		 chan->GotBan(nick, from, op, &user);
	       else
		 chan->GotUnban(nick, op, u, &user);
	       break;
	      case 'e':
	       if (I->SERVER->use_ts4) {
		  op = G->Split(&msg);
		  chan->CheckMode(nick, from, u, ms2, op);
		  if (ms2[0] == '+')
		    chan->GotException(nick, from, op, &user);
		  else 
		    chan->GotUnexception (nick, op, u, &user);
	       }
	       break;
	    }
	    if (todo) {
	       ms2[1] = *chg; 
	       chan->CheckMode(nick, from, u, ms2, "");
	       if (ms2[0] == '+')
		 chan->AddMode(todo);
	       else
		 chan->RemMode(todo);
	       if ((((ms2[0] == '+') && (chanmode->Minus() & todo)) ||
		    ((ms2[0] == '-') && (chanmode->Plus() & todo)))
		   && !user.AnyMaster())
		 chan->AddMode(ms2[0] == '+' ? '-' : '+', *chg, "");
	       else if (I->reversing && 
			((ms2[0] == '+') || (chanmode->Plus() & todo))
			&& ((ms2[0] == '-') || (chanmode->Minus() & todo)))
		 chan->AddMode(ms2[0] == '+' ? '-' : '+', *chg, "");
	    }
	    chg++;
	 }
      }  
   }
   return EXECUTED;
}

/* join */
static BindReturn gotjoin (void ** args, const char *) {
   FUNCTION;
   char * nick, *p, *newmode = 0, buf[UHOSTLEN], *uhost = buf;
   char * from = (char *)args[0], * chname = (char *)args[2];
   Channel3 *chan;
   ChannelMember * m = 0;
   UserRecord * u;
   int lang = 0, use_ts4 = I->SERVER->use_ts4;
   
   context;
   chname = FixColon(chname);
   /* ircd 2.9 sometimes does '#chname^Gmodes' when returning from splits */
   if ((p = strchr(chname, 7))) {
      *p = 0;
      newmode = (++p);
   }
   if (!(chan = FindChan3(chname))) {
      G->LOG_MISC.Put(I->Lang( 0x0F7, "s", chname));
      I->SERVER->SendPart(Q_MODE, chname, I->Lang( 0x0F8, ""));
   } else if (!chan->Pending()) {
      strcpy(uhost,from);
      nick = G->SplitNick(&uhost);
      chan->DetectFlood(nick, uhost, from, FLOOD_JOIN, 0);
      /* grab last time joined before we update it */
      if ((u = G->userlist->FindHost(from)) == G->userlist)
	u = 0;
      user.Set(u,chname);
      if (!chan->Active() && !I->SERVER->MatchMyNick(nick)) {
	 /* uh, what?!  i'm on the channel?! */
	 G->LOG_MISC.Put(I->Lang( 0x0F9, "s", chname), chan);
	 chan->SetActive();
	 chan->UnsetPending();
	 chan->UnsetMeop();
	 chan->UnsetMeHelper();
	 chan->ResetInfo();
      } else {
	 I->CHANNELS->SetHandleLaston(chname, u, G->now);
	 if ((m = chan->M()->Find(nick)) && m->IsSplit() &&
	     !strcasecmp(m->Host(),uhost)) {
	    int wasop = m->HasOp(), washelper = m->HasHelper();
	    
	    chan->CheckJPSR(nick, uhost, u, I->bt_rejn);
	    m->ClearSplit();
	    m->Touch();
	    m->ClearFlags();
	    if (wasop)
	      m->SetWasOp();
	    m->SetUser(u);
	    /* had ops before split, Im an op */
	    if (m->WasOp() && chan->Meop() && !m->SentOp()
		&& (chan->AutoOp() || user.AnyAutoop())) 
	      m->GiveOp();
	    else if (use_ts4 && washelper && 
		     (chan->Meop() || (!chan->NoHelper() && chan->MeHelper()))
		     && !m->SentHelper() && !m->SentOp() &&
		     (chan->AutoOp() || user.AnyAutoop()))
	       m->GiveHelper();
	    lang = 0x0FA;
	    if (newmode) 
	       m->DoEmbeddedMode(newmode);
	 } else {
	    if (I->SERVER->MatchMyNick(nick))
	      chan->ResetInfo();
	    if (m) {
	      DELETE ( m );
	    }
	    m = NEW ( ChannelMember ( chan, nick, uhost, u ) );
	    m->Join();
	    chan->CheckJPSR (nick, uhost, u, I->bt_join);
	    if (newmode)
	      m->DoEmbeddedMode(newmode);
	    if (I->SERVER->MatchMyNick(nick)) 
	       lang = 0x0FC;
	    else {
	       ChannelRecord * cr;
	       const char *p;
	       
	       lang = 0x0FE;
	       if (chan->Meop() || (use_ts4 && chan->MeHelper()
				    && !m->HasOp())) {
		  for (p = m->Host(); *p; p++)
		    if (((unsigned char )*p) < 32) {
		       if (chan->KickBogus()) {
			  chan->QuickBan(from);
			  m->Kick(I->Lang(0x100, ""));
		       }
		       return EXECUTED;
		    }
		  if (user.AnyValidOp() && !m->SentOp() && chan->Meop()
		      && (chan->AutoOp() || user.AnyAutoop())) {
		     m->GiveOp();
		     m->SetWasOp();
 		  } else if (use_ts4 && user.AnyValidHelper() && !m->SentHelper()
			     && (chan->AutoOp() || user.AnyAutoop())) 
		       m->GiveHelper();
		  else if (chan->MatchBan(from))
		    m->RefreshBanKick();
		  else if (user.AnyKick()) {
		     UserEntryComment * uec = FindComment(m->User());
		     chan->QuickBan(from);
		     m->Kick((uec && uec->Value() && 
			      (uec->Value()[0] != '@') ? uec->Value() :
			      I->Lang( 0x101, "")));
		  } else if (chan->AutoVoice() && user.AnyValidVoice())
		       m->GiveVoice();
		  
	       }
	       /* don't re-display greeting if they've been on the channel recently */
	       if (u && !user.GlobBot() && chan->Greet() 
		   && I->CHANNELS->use_info) {
		  UserEntryInfo * uei = FindInfo(u);
		  time_t laston = 0;
		  UserEntryLaston * uel = 0;
		  
		  cr = u->ChanRecs()->Find(chan->Key());
		  if (cr)
		    laston = cr->Laston();
		  else if (!cr && chan->NoChanRecInfo() && (uel = FindLaston(u))) 
		    laston = uel->When();
		  
		  if (laston && (G->now - laston > I->wait_info)) {
		     const char * s = uei ? uei->Value() : 0;
		     char s1[512], s2[512];
		     
		     I->CHANNELS->GetHandleChaninfo(u->Handle(), chan->Key(), s1);
		     /* locked info line overides non-locked channel specific info line */
		     if (!s || (s1[0] && (s[0] != '@' || s1[0] == '@')))
		       s = s1;
		     if (s[0] == '@')
		       s++;
		     if (s && s[0]) {
			G->SPrintF(s2, 511, "[%s] %s", nick, s);
			I->SERVER->SendMsg(Q_HELP, chan->Key(), s2);
		     }
		  }
	       }
	    }
	 }
	 I->LOG_JOIN.Put(I->Lang( newmode ? lang : lang + 1, "ssss", nick, 
			      uhost, chname, newmode), chan);
      }
   }
   return EXECUTED;
}

/* part */
static BindReturn gotpart (void ** args, const char *) {
   FUNCTION;
   char * nick, *from = (char *)args[0], *msg = (char *)args[2], *chname;
   Channel3 *chan;
   UserRecord * u;
   ChannelMember * m;
     
   chname = G->Split(&msg);
   chname = FixColon(chname);
   msg = FixColon(msg);
   if ((chan = FindChan3(chname)) && !chan->Pending()) {
      if ((u = G->userlist->FindHost(from)) == G->userlist)
	u = 0;
      nick = G->SplitNick(&from);
      if (!chan->Active()) {
	 G->LOG_MISC.Put(I->Lang( 0x0F9, "s", chname),chan);
	 chan->SetActive();
	 chan->UnsetPending();
	 chan->ResetInfo();
      }
      chan->CheckTN(nick, from, u, msg, I->bt_part);
      I->CHANNELS->SetHandleLaston(chname,u,G->now);
      if ((m = chan->M()->Find(nick))) {
	 DELETE (m);
      }
      I->LOG_JOIN.Put(I->Lang( (msg && msg[0]) ? 0x102 : 0x103, "ssss", nick,
			   from, chname, msg), chan);
      if (I->SERVER->MatchMyNick(nick)) {
	 chan->Clear(1);
	 chan->UnsetActive();
	 chan->UnsetPending();
	 I->SERVER->SendJoin(Q_MODE, chan->Key(), safe_str(chan->ChanKey()));
      } else
	chan->CheckLonely();
   }
   return EXECUTED;
}

/* kick */
static BindReturn gotkick (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], * msg = (char *)args[2], buf[512], buf2[512];
   char * nick, * whodid, * chname, *uhost = buf;
   ChannelMember * m, * m2;
   Channel3 *chan;
   UserRecord * u;
   void * argv[6];
   
   chname = G->Split(&msg);
   if ((chan = FindChan3(chname)) && !chan->Pending()) {
      nick = G->Split(&msg);
      msg = FixColon(msg);
      if ((u = G->userlist->FindHost(from)) == G->userlist)
	u = 0;
      strcpy(uhost,from);
      whodid = G->SplitNick(&uhost);
      chan->DetectFlood(whodid, uhost, from, FLOOD_KICK, nick);
      if ((m = chan->M()->Find(whodid)))
	m->Touch();
      G->SPrintF(buf2, 511, "%s %s", chname, nick);
      user.Set(u,chname);
      argv[0] = whodid;
      argv[1] = uhost;
      argv[2] = u;
      argv[3] = chan;
      argv[4] = nick;
      argv[5] = msg;
      I->bt_kick->Check(buf2, &user, argv);
      /* kicking an oplisted person?  KICK THEM. */
      if ((m2 = chan->M()->Find(nick))) {
	 victim.Set(m->MemberUser(),chname);
	 I->CHANNELS->SetHandleLaston(chname,m2->MemberUser(),G->now);
	 /* kicked has +o ? & kicker doesnt */
	 if (m && victim.AnyValidOp() && !user.AnyValidOp()
	     /* channel is revenge ? & wasnt me doing it */
	     && chan->Revenge() && !I->SERVER->MatchMyNick(whodid)
	     /* not kicking themselves?  & im opped ?*/
	     && strcasecmp(whodid,nick) && chan->Meop())
	   m->Kick(I->Lang( 0x104, ""));
	 I->LOG_MODES.Put(I->Lang( 0x105, "sssss", m2->Key(), m2->Host(),
			       chname, from, msg), chan);
      }
      /* kicked ME?!? the sods! */
      if (I->SERVER->MatchMyNick(nick)) {
	 chan->UnsetActive();
	 chan->UnsetPending();
	 I->SERVER->SendJoin(Q_MODE, chan->Key(), safe_str(chan->ChanKey()));
	 chan->Clear(1);
	 if (chan->Revenge() && !user.AnyFriend()) 
	   chan->TakeRevenge(from, I->Lang( 0x106, "s", I->SERVER->botname));
      } else {
	 if (m2)
	   DELETE (m2);
	 chan->CheckLonely();
      }
   }
   return EXECUTED;
}

/* nick change */
static BindReturn gotnick (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], * msg = (char *)args[2];
   char * nick, s1[UHOSTLEN], buf[UHOSTLEN], *uhost = buf;
   ChannelMember * m, * mm = 0;
   Channel3 *chan = (Channel3 *)G->chanset;
   UserRecord * u;
   
   if ((u = G->userlist->FindHost(from)) == G->userlist)
     u = 0;
   strcpy(uhost,from);
   nick = G->SplitNick(&uhost);
   msg = FixColon(msg);
   while ((chan = chan->Next())) 
     if ((m = chan->M()->Find(nick))) {
	I->LOG_JOIN.Put(I->Lang( 0x107, "ss", nick, msg), chan);
	m->Touch();
	if (strcasecmp(nick, msg) && (mm = chan->M()->Find(msg))) {
	   (mm->IsSplit() ? I->LOG_JOIN : G->LOG_MISC).Put
	     (I->Lang( mm->IsSplit() ? 0x108 : 0x109, "s", msg), chan);
	   DELETE ( mm );
	}
	chan->DetectFlood(nick, uhost, from, FLOOD_NICK, 0);
	G->SPrintF(s1, UHOSTLEN, "%s!%s", msg, uhost);
	if (chan->EnforceBans() && !m->SentKick())
	  m->RefreshBanKick();
	m->SetKey(msg);
	m->SetUser(0);
	chan->CheckTN(nick, uhost, u, msg, I->bt_nick);
     }
   return EXECUTED;
}

/* signoff, similar to part */
static BindReturn gotquit (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], * msg = (char *)args[2], * nick, buf[512];
   char *p, *z1, *z2;
   int split = 0;
   ChannelMember * m;
   Channel3 *chan = (Channel3 *)G->chanset;
   UserRecord * u;
   
   if ((u = G->userlist->FindHost(from)) == G->userlist)
     u = 0;
   nick = G->SplitNick(&from);
   msg = FixColon(msg);
   /* Fred1: instead of expensive G->WildMatch on signoff, quicker method */
   /* determine if signoff string matches "%.% %.%", and only one space */
   if ((p = strchr(msg, ' ')) && (p == strrchr(msg,' '))) {
      *p = 0;
      z1 = strchr(p + 1, '.');
      z2 = strchr(msg, '.');
      if (z1 && z2 && (*(z1 + 1) != 0) && (z1 - 1 != p) &&
	  (z2 + 1 != p) && (z2 != msg)) {
	 /* server split, or else it looked like it anyway */
	 /* (no harm in assuming) */
	 split = 1;
      } else
	*p = ' ';
   }
   while ((chan = chan->Next()))
     if ((m = chan->M()->Find(nick))) {
	I->CHANNELS->SetHandleLaston(chan->Key(),u,G->now);
	if (split) {
	   m->Split();
	   chan->CheckJPSR(nick, from, u, I->bt_splt);
	   I->LOG_JOIN.Put(I->Lang( 0x10A, "sss", nick,from), chan);
	} else {
	   void * args[5] = { nick, from, u, chan, msg };
	   
	   G->SPrintF(buf, 511, "%s %s!%s", chan->Key(), nick, from);
	   user.Set(u,chan->Key());
	   I->bt_sign->Check(buf, &user, args);
	   I->LOG_JOIN.Put(I->Lang(0x10B,"sss",nick,from,msg),chan);
	   DELETE ( m );
	   chan->CheckLonely();
	}
     }
   return EXECUTED;
}

/* public message */
static BindReturn gotmsg (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], * msg = (char *)args[2];
   char *to, *realto, buf[UHOSTLEN], *nick, buf2[512], *uhost = buf;
   char *p, *p1, *code, *ctcp;
   int ctcp_count = 0;
   Channel3 *chan;
   int ignoring;
   UserRecord * u;
   ChannelMember * m;
   
   if (!strchr("&#@+",msg[0])) 
     return NOMATCH;
   to = G->Split(&msg);
   if (to[0] == '@') {
      if ((to[1] == '+') || (to[1] == '%'))
	realto = to + 2;
      else 
	realto = to + 1;
   } else
     realto = to;
   if (!(chan = FindChan3(realto))) 
     return NOMATCH; /* privmsg to an unknown channel ?? */
   ignoring = G->ignores->Match(from) != 0;
   msg = FixColon(msg);
   strcpy(uhost, from);
   nick = G->SplitNick(&uhost);
   if ((m = chan->M()->Find(nick)))
     m->Touch();
   context;
   if (I->SERVER->flood_ctcp_thr && I->SERVER->DetectAvalanche(msg)) {
      context;
      if (m) {
	 user.Set(m->MemberUser(), realto);
	 if (chan->Meop() && !user.AnyFriend()
	     && !(chan->DontKickOps() && user.AnyValidOp()))
	   m->Kick(I->Lang( 0x10C, ""));
      }
      if (!ignoring) {
	 I->LOG_MODES.Put(I->Lang( 0x10D, "ss", nick, uhost));
	 G->SPrintF(buf2, UHOSTLEN, "*!*@%s", 
		    (p = strchr(uhost, '@')) ? p + 1 : uhost);
	 NEW ( IgnoreRecord ( buf2, G->origbotname, I->Lang( 0x10E, ""), 
			     G->now + (60 * G->ignore_time)) );
      }
      context;
   } else {
      context;
   /* check for CTCP: */
      I->SERVER->ctcp_reply[0] = 0;
      I->SERVER->ctcp_reply_len = 0;
      if ((u = G->userlist->FindHost(from)) == G->userlist)
	u = 0;
      p = strchr(msg,1);
      while (p && *p) {
	 p++;
	 p1 = p;
	 while ((*p != 1) && *p)
	   p++;
	 if (*p == 1) {
	    *p = 0;
	    ctcp = buf2;
	    strcpy(ctcp, p1);
	    strcpy(p1 - 1, p + 1);
	    p = strchr(p1 - 1, 1);
	    chan->DetectFlood(nick, uhost, from, strncmp(ctcp, "ACTION ", 7) ? 
			      FLOOD_CTCP : FLOOD_PRIVMSG, 0);
	    if ((ctcp_count++ < I->SERVER->answer_ctcp) && (ctcp[0] != ' ') &&
		(!ignoring || I->SERVER->trigger_on_ignore)) {
	       code = G->Split(&ctcp);
	       if ((I->SERVER->CheckCtcpr(nick, uhost, u, to, code, ctcp,
					  I->SERVER->bt_ctcp) == NOMATCH)
		   && !ignoring) 
		 I->LOG_PUBLIC.Put(I->Lang(strcmp(code, "ACTION")?0x110:0x10F,
					"sssss",nick, ctcp, code, from, to),
				chan);
	    }
	 }
      }
      /* send out possible ctcp responses */
      if (I->SERVER->ctcp_reply_len)
	I->SERVER->SendNotice(Q_SERVER, nick, I->SERVER->ctcp_reply);
      if (msg[0]) {
	 if (!ignoring)
	   chan->DetectFlood(nick, uhost, from, FLOOD_PRIVMSG, 0);
	 if (!ignoring || I->SERVER->trigger_on_ignore) {
	    void * args[5] = { nick, uhost, u, chan, 0 };
	    BindReturn x;
	    char work[512], *rest = work, *cmd;
	    
	    user.Set(u,chan->Key());
	    strcpy(rest, msg);
	    cmd = G->Split(&rest);
	    args[4] = rest;
	    if ((x = I->bt_pub->Check(cmd, &user, args)) == NOMATCH) {
	       G->SPrintF(work, 511, "%s %s", chan->Key(), msg);
	       args[4] = msg;
	       I->bt_pubm->Check(work, &user, args);
	    } else {
	       if (x == EXEC_LOG)
		 G->LOG_CMDS.Put(I->Lang(0x111,"ssss",nick,u?u->Handle():"*",
					 cmd,rest),chan);
	       return EXECUTED;
	    }
	 }
	 if (!ignoring) 
	   I->LOG_PUBLIC.Put(I->Lang((to[0]=='@')?0x0112:0x113,"ss", nick,msg),
			  chan);  
      }
   }
   return EXECUTED;
}

/* public notice */
static BindReturn gotnotice (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], buf3[512], * msg = (char *)buf3, *ctcp, *code;
   char *to, *realto, *nick, buf2[512], *p, *p1, buf[512], *uhost = buf;
   UserRecord * u;
   Channel3 * chan;
   int ignoring;
   ChannelMember * m;
   
   strcpy(msg,(char *)args[2]);
   if (!strchr("#&+@",*msg))
     return NOMATCH;
   to = G->Split(&msg);
   if (to[0] == '@') {
      if ((to[1] == '+') || (to[1] == '%'))
	realto = to + 2;
      else 
	realto = to + 1;
   } else
     realto = to;
   if (!(chan = FindChan3(realto)))
     return NOMATCH;
   ignoring = G->ignores->Match(from) != 0;
   msg = FixColon(msg);
   strcpy(uhost,from);
   nick = G->SplitNick(&uhost);
   if ((m = chan->M()->Find(nick)))
     m->Touch();
   if (I->SERVER->flood_ctcp_thr && I->SERVER->DetectAvalanche(msg)) {
      if (m) {
	 user.Set(m->MemberUser(), realto);
	 /* discard -- kick user if it was to the channel */
	 if (chan->Meop() && !user.AnyFriend()
	     && !(chan->DontKickOps() && user.AnyValidOp()))
	   m->Kick(I->Lang( 0x10C, ""));
      }
      if (!ignoring)
	I->LOG_MODES.Put(I->Lang( 0x10D, "ss", nick, uhost));
      return EXECUTED;
   }
   if ((u = G->userlist->FindHost(from)) == G->userlist)
     u = 0;
   /* check for CTCP: */
   p = strchr(msg,1);
   while (p && *p) {
      p++;
      p1 = p;
      while ((*p != 1) && *p)
	 p++;
      if (*p == 1) {
	 *p = 0;
	 ctcp = buf2;
	 strcpy(ctcp, p1);
	 strcpy(p1 - 1, p + 1);
	 p = strchr(p1 - 1, 1);
	 if (!ignoring)
	   chan->DetectFlood(nick, uhost, from, FLOOD_CTCP, 0);
	 if (ctcp[0] != ' ') {
	    code = G->Split(&ctcp);
	    if (!ignoring || I->SERVER->trigger_on_ignore) {
	       I->SERVER->CheckCtcpr(nick,uhost,u,to,code,msg,I->SERVER->bt_ctcp);
	       if (!ignoring)
		 I->LOG_PUBLIC.Put(I->Lang( 0x114, "sssss", code,
					msg, nick, from, to), chan);
	    }
	 }
      }
   }
   if (msg[0] && (!ignoring || I->SERVER->trigger_on_ignore)) {
      void * args[4] = { nick, uhost, u, msg };
      
      chan->DetectFlood(nick,uhost,from, FLOOD_NOTICE, 0);
      user.Set(u);
      I->SERVER->bt_notc->Check(msg, &user, args);
      I->LOG_PUBLIC.Put(I->Lang( 0x115, "sss", nick, to, msg), chan);
   }
   return EXECUTED;
}

#define RAW_COMMANDS 30
static command_binds irc_raw [RAW_COMMANDS] = {
   { "001", "", got001, "irc:001" },
   { "324", "", got324, "irc:324" },
   { "325", "", got325or6, "irc:325or6" },
   { "326", "", got325or6, "irc:325or6" },
   { "329", "", got329, "irc:329" },
   { "339", "", got325or6, "irc:325or6" },
   { "352", "", got352or4, "irc:352or4" },
   { "354", "", got352or4, "irc:352or4" },
   { "315", "", got315, "irc:315" },
   { "348", "", got348, "irc:348" },
   { "367", "", got367, "irc:367" },
   { "368", "", got368, "irc:368" },
   { "405", "", got405, "irc:405" },
   { "471", "", got47x, "irc:47x" },
   { "473", "", got47x, "irc:47x" },
   { "474", "", got47x, "irc:47x" },
   { "475", "", got47x, "irc:47x" },
   { "442", "", got442, "irc:442" },
   { "INVITE", "", gotinvite, "irc:invite" },
   { "TOPIC", "", gottopic, "irc:topic" },
   { "331", "", got33x, "irc:33x" },
   { "332", "", got33x, "irc:33x" },
   { "JOIN", "", gotjoin, "irc:join" },
   { "PART", "", gotpart, "irc:part" },
   { "KICK", "", gotkick, "irc:kick" },
   { "NICK", "", gotnick, "irc:nick" },
   { "QUIT", "", gotquit, "irc:quit" },
   { "PRIVMSG", "", gotmsg, "irc:msg" },
   { "NOTICE", "", gotnotice, "irc:notice" },
   { "MODE", "", gotmode, "irc:mode" },
};

/*************************************************************************
 * DCC COMMANDS                                                          *
 *************************************************************************/


/* Do we have any flags that will allow us ops on a channel? */
static Channel3 * is_ok (DccChat * idx, char ** buf, int op) {
   FUNCTION;
   Channel3 *chan;
   char * chname = (buf && *buf && ((**buf == '#') || (**buf == '&'))) ?
     G->Split(buf) : 0;
   
   if (chname && !(chan = FindChan3(chname))) 
     idx->Send(I->CHANNELS->Lang( 0x100, ""));
   else if (!(chan = FindChan3(chname = idx->ConsoleChan())))
     idx->Send(I->CHANNELS->Lang( 0x101, ""));
   else {
      user.Set(idx->User(),chname);
      if ((op && user.AnyValidOp()) || (!op && user.AnyValidHelper()))
	return chan;
      idx->Send(I->CHANNELS->Lang( 0x102, "s", chan->Key()));
   }
   return 0;
}

static BindReturn cmd_channel (void ** args, const char *) {
   FUNCTION;
   int i, len, len2;
   DccChat * idx = (DccChat *)args[0];
   char s[121], s1[UHOSTLEN+1], atrflag, chanflag, *par = (char *)args[1], *t;
   const char * hand;
   Channel3 *chan;
   ChannelMember * m;
   UserRecord * u;
   time_t tt;
   
   G->LOG_CMDS.Put(I->Lang(0x05C,"sss",idx->Nick(), idx->ConsoleChan(), par));
   if ((chan = is_ok(idx, &par, 0))) {
      int mem = chan->M()->Count();
      strcpy(s, chan->GetMode());
      idx->Send(I->Lang( chan->Pending() ? 0x05D :
			(chan->Active() ? 0x05E : 0x05F), "sdss",
			chan->Key(), mem, (mem == 1) ? "" : "s", s));
      if (chan->Topic())
	idx->Send(I->Lang( 0x060, "s", chan->Topic()));
      i = 0;
      if (chan->Active()) {
	 idx->Send(I->Lang( 0x061, ""));
	 idx->Send(I->Lang( 0x062, ""));
	 for (m = chan->M(); (m = m->Next()); ) {
	    if ((tt = m->Joined())) {
	       t = ctime(&tt);
	       if ((G->now - tt) > 86400) {
		  s[0] = t[8];
		  s[1] = t[9];
		  s[2] = t[4];
		  s[3] = t[5];
		  s[4] = t[6];
		  s[5] = 0;
	       } else 
		 strncpyz(s, t + 11, 5);
	    } else
	      strcpy(s, " --- ");
	    u = m->MemberUser();
	    user.Set(u,chan->Key());
	    /* determine status char to use */
	    if (user.ChanBot())
	      atrflag = 'B';
	    else if (user.GlobBot())
	      atrflag = 'b';
	    else if (user.GlobOwner())
	      atrflag = 'N';
	    else if (user.ChanOwner())
	      atrflag = 'n';
	    else if (user.GlobMaster())
	      atrflag = 'M';
	    else if (user.ChanMaster())
	      atrflag = 'm';
	    else if (user.GlobOp() && !user.ChanDeop())
	      atrflag = 'O';
	    else if (user.ChanOp())
	      atrflag = 'o';
	    else if (user.GlobHelper() && !user.ChanDeop())
	      atrflag = 'H';
	    else if (user.ChanHelper())
	      atrflag = 'h';
	    else if (user.GlobDeop())
	      atrflag = 'D';
	    else if (user.ChanDeop())
	      atrflag = 'd';
	    else if (user.GlobVoice() && !user.ChanQuiet())
	      atrflag = 'V';
	    else if (user.ChanVoice())
	      atrflag = 'v';
	    else if (user.GlobQuiet())
	      atrflag = 'Q';
	    else if (user.ChanQuiet())
	      atrflag = 'q';
	    else
	      atrflag = ' ';
	    if (m->HasOp())
	      chanflag = '@';	
	    else if (m->HasHelper())
	      chanflag = '%';
	    else if (m->HasVoice())
	      chanflag = '+';
	    else
	      chanflag = ' ';
	    hand = u ? u->Handle() : "*";
	    len = NICKMAX - strlen(m->Key());
	    len2 = HANDLEN - strlen(hand);
	    if (m->IsSplit())
	      idx->Send(I->Lang( 0x063, "cs s scl",chanflag,m->Key(),len,
				   hand, len2, s, atrflag, m->SplitAge()));
	    else if (I->SERVER->MatchMyNick(m->Key()))
	      idx->Send(I->Lang( 0x064, "cs s sc", chanflag, m->Key(),
				   len, hand, len2, s, atrflag));
	    else {
	       /* determine idle time */
	       if (m->Age() > 86400)
		 G->SPrintF(s1, UHOSTLEN, "%2d",
			    (int)(m->Age() / 86400));
	       else if (m->Age() > 3600)
		 G->SPrintF(s1, UHOSTLEN, "%2h", 
			    (int)(m->Age() / 3600));
	       else if (m->Age() > 180)
		 G->SPrintF(s1, UHOSTLEN, "%2m", 
			    (int)(m->Age() / 60));
	       else
		 strcpy(s1, "   ");
	       idx->Send(I->Lang( 0x065, "cs s scss", chanflag, m->Key(),
				 len, hand, len2, s, atrflag, s1,m->Host()));
	    }
	    if (m->FakeOp())
	      idx->Send(I->Lang( 0x066, ""));
	    if (m->SentOp())
	      idx->Send(I->Lang( 0x067, ""));
	    if (m->SentDeop())
	      idx->Send(I->Lang( 0x068, ""));
	    if (m->SentHelper())
	      idx->Send(I->Lang( 0x128, ""));
	    if (m->SentDeHelper())
	      idx->Send(I->Lang( 0x129, ""));
	    if (m->SentWho())
	      idx->Send(I->Lang( 0x12A, ""));
	    if (m->SentKick())
	      idx->Send(I->Lang( 0x069, ""));
	 }
      }
      idx->Send(I->Lang( 0x06A, ""));
   }
   return EXECUTED;
}

static BindReturn cmd_actsay (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int lang = mask[0] == 'a' ? 0x06B : 0x06E;
   char * par = (char *)args[1];
   Channel3 * chan;
   
   if (!par[0]) 
     idx->Send(I->Lang( lang, ""));
   else if ((chan = is_ok(idx, &par,1))) {
      G->LOG_CMDS.Put(I->Lang(lang+1,"sss",idx->Nick(),chan->Key(),par));
      if (mask[0] == 'a') 
	I->SERVER->SendAct(Q_HELP, chan->Key(), par);
      else
	I->SERVER->SendMsg(Q_HELP, chan->Key(), par);
      idx->Send(I->Lang(lang+2, "ss", chan->Key(), par));
   }
   return EXECUTED;
}

static BindReturn cmd_msg (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * nick, *par = (char *)args[1];
   
   if (!par[0]) 
     idx->Send(I->Lang( 0x071, ""));
   else {
      nick = G->Split(&par);
      G->LOG_CMDS.Put(I->Lang( 0x072, "sss", idx->Nick(), nick, par));
      I->SERVER->SendMsg(Q_HELP, nick, par);
      idx->Send(I->Lang(0x073, "ss", nick, par));
   }
   return EXECUTED;
}

static BindReturn cmd_kickban ( void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int ban = !strncmp(mask, "kickb", 5), lang = ban ? 0x074 : 0x08D;
   Channel3 *chan;
   char *nick, *par = (char *)args[1];
   const char * s1;
   ChannelMember * m;
   
   if (!par[0]) 
     idx->Send(I->Lang( lang, ""));
   else if ((chan = is_ok(idx, &par,0))) {
      if (!chan->Meop() || (I->SERVER->use_ts4 && !chan->MeHelper())) 
	idx->Send(I->Lang( 0x075, "s", chan->Key()));
      else {
	 G->LOG_CMDS.Put(I->Lang(lang+1,"sss",idx->Nick(),chan->Key(),par));
	 nick = G->Split(&par);
	 if (I->SERVER->MatchMyNick(nick)) 
	   idx->Send(I->Lang(lang+2, ""));
	 else if (!(m = chan->M()->Find(nick))) 
	   idx->Send(I->Lang(0x078,"ss",nick, chan->Key()));
	 else {
	    victim.Set(m->MemberUser(), chan->Key());
	    if (victim.AnyValidHelper() && !user.AnyValidOp())
	      idx->Send(I->Lang(0x12B, "s", nick));
	    if (victim.AnyValidOp() && !user.AnyMaster())
	      idx->Send(I->Lang(0x079,"s",nick));
	    else if (victim.AnyMaster() && !user.AnyOwner())
	      idx->Send(I->Lang(0x07A,"ss",nick,chan->Key()));
	    else if (victim.ChanBot() && !user.AnyOwner())
	      idx->Send(I->Lang(0x07B,"s",nick));
	    else {
	       if (!par[0]) 
		 par = I->Lang( 0x090, "");
	       if (ban) {
		  if (m->HasOp() && chan->Meop())
		    chan->AddMode('-', 'o', m->Key());
		  else if (I->SERVER->use_ts4 && m->HasHelper()
			   && (chan->Meop() && (!chan->NoHelper() && chan->MeHelper())))
		    chan->AddMode('-', 'h', m->Key());
		  s1 = chan->QuickBan(m->Host());
		  chan->AddBan(s1, idx->Nick(), par, G->now + 
			       (60 * I->CHANNELS->ban_time),0);
	       }
	       m->Kick(par);
	       idx->Send(I->Lang( 0x07C, ""));
	    }
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_voice (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int de = mask[0] == 'd', lang = de ? 0x080 : 0x07D;
   Channel3 *chan;
   char * nick, *par = (char *)args[1];
   ChannelMember * m;

   if (!(nick = G->Split(&par))[0]) 
     idx->Send(I->Lang( lang, ""));
   else if ((chan = is_ok(idx,&par, 0))) {
      if (!chan->Meop() || (I->SERVER->use_ts4 && !chan->MeHelper())) 
	idx->Send(I->Lang( 0x075, "s", chan->Key()));
      else {
	 G->LOG_CMDS.Put(I->Lang( lang+1, "ssss", idx->Nick(),
			       idx->ConsoleChan(),nick,chan->Key()));
	 if (!(m = chan->M()->Find(nick)))
	   idx->Send(I->Lang( 0x078, "ss", nick, chan->Key()));
	 else {
	    chan->AddMode(de ? '-' : '+', 'v', nick);
	    idx->Send(I->Lang( lang + 2, "ss", nick, chan->Key()));
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_op (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int de = mask[0] == 'd', lang = de ? 0x88 : 0x083;
   Channel3 *chan;
   char * nick, *par = (char *)args[1];
   ChannelMember * m;
   
   if (!*(nick = G->Split(&par)))
     idx->Send(I->Lang( lang, ""));
   else if ((chan = is_ok(idx,&par,1))) {
      if (!(chan->Meop() || (I->SERVER->use_ts4 && chan->MeHelper())))
	idx->Send(I->Lang( 0x075, "s", chan->Key()));
      else {
	 G->LOG_CMDS.Put(I->Lang( lang+1, "ssss", idx->Nick(),
				  idx->ConsoleChan(),nick,chan->Key()));
	 if (!(m = chan->M()->Find(nick))) 
	   idx->Send(I->Lang( 0x078, "ss", nick, chan->Key()));
	 else {
	    victim.Set(m->MemberUser(),chan->Key());
	    if (!de) {
	       if (victim.AnyValidDeop())
		 lang = 0x085;
	       else if ((chan->Bitch() || I->SERVER->use_ts4)
			&& !victim.AnyValidOp()) 
		 lang = 0x086;
	       else {
		  chan->AddMode('+', 'o', nick);
		  lang = 0x087;
	       }
	    } else if (victim.AnyMaster() && !user.AnyOwner())
		 lang = 0x08A;
	    else if (victim.AnyValidOp() && !user.AnyMaster())
	      lang = 0x08B;
	    else {
	       chan->AddMode('-', 'o', nick);
	       lang = 0x08C;
	    }
	    idx->Send(I->Lang( lang, "ss", nick, chan->Key()));
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_helper (void ** args, const char * mask) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int de = mask[0] == 'd', lang = de ? 0x131 : 0x12C;
   Channel3 *chan;
   char * nick, *par = (char *)args[1];
   ChannelMember * m;
   
   if (!(nick = G->Split(&par))[0]) 
     idx->Send(I->Lang( lang, ""));
   else if ((chan = is_ok(idx,&par,1))) {
      if (!I->SERVER->use_ts4)
	idx->Send(I->Lang( 0x136, ""));
      else if (!chan->MeHelper() && !chan->Meop())
	idx->Send(I->Lang( 0x075, ""));
      else {
	 G->LOG_CMDS.Put(I->Lang( lang + 1, "ssss", idx->Nick(),
				 idx->ConsoleChan(),nick,chan->Key()));
	 if (!(m = chan->M()->Find(nick))) 
	   idx->Send(I->Lang( 0x078, "ss", nick, chan->Key()));
	 else {
	    victim.Set(m->MemberUser(),chan->Key());
	    if (!de) {
	       if (victim.AnyValidDeop())
		 lang = 0x085;
	       else if (chan->Bitch() && !victim.AnyValidHelper()) 
		 lang = 0x086;
	       else {
		  chan->AddMode('+', 'h', nick);
		  lang = 0x087;
	       }
	    } else if (victim.AnyMaster() && !user.AnyOwner())
		 lang = 0x08A;
	    else if (victim.AnyValidHelper() && !user.AnyMaster())
	      lang = 0x08B;
	    else {
	       chan->AddMode('-', 'h', nick);
	       lang = 0x08C;
	    }
	    idx->Send(I->Lang( lang, "ss", nick, chan->Key()));
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_invite (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int lang = 0x093;
   const char * nick;
   char * par = (char *)args[1];
   Channel3 *chan;
   ChannelMember * m;
   
   nick = G->Split(&par);
   if ((chan = is_ok(idx, &par,0))) {
      if (!nick[0]) 
	nick = idx->Nick();
      G->LOG_CMDS.Put(I->Lang(0x091,"sss",idx->Nick(),chan->Key(),nick));
      if (!chan->Active() || (!chan->Meop() && (chan->Mode() & CHANINV)))
	lang = 0x075;
      else if ((m = chan->M()->Find(nick)) && !m->IsSplit())
	lang = 0x092;
      else 
	I->SERVER->SendInvite(Q_SERVER, nick, chan->Key());
      idx->Send(I->Lang( lang, "ss", chan->Key(), nick));
   }
   return EXECUTED;
}

static BindReturn cmd_topic (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int lang = 0x097;
   char * par = (char *)args[1];
   Channel3 *chan;

   if ((chan = is_ok(idx,&par,0))) {
      if (!par[0]) 
	lang = chan->Topic() ? 0x094 : 0x095;
      else if (chan->OpTopic() && !chan->Meop() && 
	       (!I->SERVER->use_ts4 || !chan->MeHelper()))
	lang = 0x096;
      else {
	 I->SERVER->SendTopic(Q_SERVER, chan->Key(), par);
	 G->LOG_CMDS.Put(I->Lang(0x098,"sss",idx->Nick(),chan->Key(),par));
      }
      idx->Send(I->Lang( lang, "ss", chan->Key(), chan->Topic()));
   }
   return EXECUTED;
}

static BindReturn cmd_resetbans (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   char * par = (char *)args[1];
   Channel3 *chan = is_ok(idx, &par,1);
   
   if (chan) {
      G->LOG_CMDS.Put(I->Lang(0x09A,"ss",idx->Nick(),chan->Key()));
      idx->Send(I->Lang(0x09B,"s",chan->Key()));
      chan->ResetBans();
   }
   return EXECUTED;
}

static BindReturn cmd_adduser (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int ok = 1;
   char *nick, *hand, * par = (char *)args[1], * p, s[UHOSTLEN+1];
   Channel3 *chan;
   ChannelMember * m;
   UserRecord * u;
   FlagRecord fr (idx->User());
   
   context;
   if (!par[0])
     idx->Send(I->Lang( 0x09C, ""));
   else if (!(chan = FindChan3(idx->ConsoleChan())))
     idx->Send(I->CHANNELS->Lang( 0x101, ""));
   else if (!chan->Active()) 
     idx->Send(I->Lang( 0x09D, "s", chan->Key()));
   else {
      hand = nick = G->Split(&par);
      if (par[0]) 
	hand = par;
      for (p = hand; *p && ((p - hand) < HANDLEN); p++)
	if ((*p <= 32) || (*p >= 127)) {
	   ok = 0;
	   break;
	}
      *p = 0;
      if (!ok) 
	idx->Send(I->Lang( 0x09E, ""));
      else if (strchr("-,+*=:!.@#;$", hand[0]))
	idx->Send(I->Lang( 0x09F, "c", hand[0]));
      else if (!(m = chan->M()->Find(nick)))
	idx->Send(I->Lang( 0x078, "ss", nick, chan->Key()));
      else if ((u = m->MemberUser())) 
	idx->Send(I->Lang( 0x0A0, "ss", nick, u->Handle()));
      else if ((u = G->userlist->Find(hand)) && !fr.GlobOwner()
	       && fr.Set(u)->GlobOwner() && !strcasecmp(idx->Nick(),hand))
	idx->Send(I->Lang( 0x0A1, ""));
      else {
	 G->MaskHost(m->Host(), s);
	 if (!u) {
	    idx->Send(I->Lang( 0x0A2, "ss", hand, s));
	    fr.Clear();
	    fr.SetGlobal(G->default_flags);
	    NEW ( UserRecord ( G->userlist, hand, s, fr) );
	 } else {
	    idx->Send(I->Lang( 0x0A3, "ss", s, u->Handle()));
	    u->AddHost(s);
	    user.Set(u,chan->Key());
	    if (!m->HasOp() && !m->SentOp() && user.AnyValidOp()
		&& (chan->AutoOp() || user.AnyAutoop()) && chan->Meop()) {
	       chan->AddMode('+', 'o', nick);
	       m->SetSentOp();
	    }
	 }
	 G->LOG_CMDS.Put(I->Lang( 0x0A4, "sss",idx->Nick(),nick,hand));
      }
   }
   return EXECUTED;
}

static BindReturn cmd_deluser (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int lang = 0x0AD;
   char *nick, *par = (char *)args[1];
   Channel3 *chan;
   ChannelMember * m;
   UserRecord * u;
   FlagRecord victim (FlagRecord::GLOBAL | FlagRecord::CHANNEL 
		      | FlagRecord::ANYWHERE);
   
   if (!par[0]) 
     idx->Send(I->Lang( 0x0A5, ""));
   else if (!(chan = FindChan3(idx->ConsoleChan())))
     idx->Send(I->CHANNELS->Lang( 0x101, ""));
   else if (!chan->Active())
     idx->Send(I->Lang( 0x09D, "s", chan->Key()));
   else if (!(m = chan->M()->Find(nick = G->Split(&par))))
     idx->Send(I->Lang( 0x078, "ss", nick, chan->Key()));
   else {
      user.Set(idx->User(),chan->Key());
      if (!(u = m->MemberUser()))
	lang = 0x0A6;
      else {
	 victim.Set(u);
	 if (victim.GlobOwner() && !strcasecmp(idx->Nick(), nick))
	   lang = 0x0A7;
	 else if (victim.GlobBotmast() && !user.GlobOwner())
	   lang = 0x0A8;
	 else if (victim.ChanOwner() && !user.GlobOwner())
	   lang = 0x0A9;
	 else if (victim.ChanMaster() && !user.AnyOwner())
	   lang = 0x0AA;
	 else if (victim.GlobBot() && !user.GlobOwner())
	   lang = 0x0AB;
	 else {
	    DELETE ( u );
	    G->LOG_CMDS.Put(I->Lang( 0x0AE, "ss", idx->Nick(), nick));
	 }
      }
      idx->Send(I->Lang( lang, "s", nick));
   }
   return EXECUTED;
}

static BindReturn cmd_reset (void ** args, const char *) {
   FUNCTION;
   DccChat * idx = (DccChat *)args[0];
   int lang = 0xB2;
   char * par = (char *)args[1];
   Channel3 *chan = (Channel3 *)G->chanset;

   if (par[0]) {
      if (!(chan = FindChan3(par)))
	lang = 0x0AF;
      else {
	 user.Set(idx->User(),par);
	 if (!user.AnyMaster())
	   lang = 0x0B0;
	 else if (!chan->Active())
	   lang = 0x0B1;
	 else {
	    G->LOG_CMDS.Put(I->Lang( 0x0B3, "ss", idx->Nick(), par));
	    chan->ResetInfo();
	 }
      }
      idx->Send(I->Lang( lang, "s", par));
   } else if (!user.Set(idx->User())->GlobMaster())
	idx->Send(I->Lang( 0x0B4, ""));
   else {
      G->LOG_CMDS.Put(I->Lang( 0x0B5, "s", idx->Nick()));
      idx->Send(I->Lang( 0x0B6, ""));
      while ((chan = chan->Next()))
	if (chan->Active())
	  chan->ResetInfo();
   }
   return EXECUTED;
}

#define DCC_COMMANDS 18
static command_binds irc_dcc[DCC_COMMANDS]={
   { "act", "o|o", cmd_actsay, 0 },
   { "adduser", "m|m", cmd_adduser, 0 },
   { "channel", "h|h", cmd_channel, 0 },
   { "deluser", "m|m", cmd_deluser, 0 },
   { "dehelper", "h|h", cmd_helper, 0 },
   { "deop", "o|o", cmd_op, 0 },
   { "devoice", "o", cmd_voice, 0 },
   { "helper", "h|h", cmd_helper, 0 },
   { "invite", "h|h", cmd_invite, 0 },
   { "kick", "h|h", cmd_kickban, 0 },
   { "kickban", "h|h", cmd_kickban, 0 },
   { "msg", "o", cmd_msg, 0 },
   { "op", "o|o", cmd_op, 0 },
   { "reset", "m|m", cmd_reset, 0 },
   { "resetbans", "o|o", cmd_resetbans, 0 },
   { "say", "o|o", cmd_actsay, 0 },
   { "topic", "h|h", cmd_topic, 0 },
   { "voice", "o", cmd_voice, 0 },
};

/*************************************************************************
 * MSG COMMANDS                                                          *
 *************************************************************************/

#define NOTICE(a,b) I->SERVER->SendNotice(Q_SERVER, a, b)

static BindReturn msg_hello (void ** args, const char *) {
   FUNCTION;
   char host[UHOSTLEN+1], s[UHOSTLEN+1], s1[161];
   char *p1, *nick = (char *)args[0], *h = (char *)args[1];
   UserRecord * u = (UserRecord *)args[2];
   int common = 0;
   Channel3 * chan = (Channel3 *)G->chanset;
   FlagRecord fr;
   
   if (!I->learn_users && !G->make_userfile)
      return EXECUTED;
   else if (!I->SERVER->MatchMyNick(nick)) {
      fr.Set(u);
      if (u && !fr.GlobCommon())
	NOTICE(nick,I->Lang( 0x001, "s", u->Handle()));
      else if (G->userlist->Find(nick)) {
	 NOTICE(nick,I->Lang(0x002,""));
	 NOTICE(nick,I->Lang(0x003,"s",I->SERVER->botname));
      } else {
	 G->SPrintF (s, UHOSTLEN, "%s!%s", nick, h);
	 if (chan->MatchBan(s)) 
	   NOTICE(nick, I->Lang(0x004,""));
	 else if (strlen(nick) > HANDLEN) 
	   NOTICE(nick, I->Lang(0x005,""));
	 else {
	    if (fr.GlobCommon()) {
	       G->MaskHost(s, host);
	       strcpy(s, host);
	       G->SPrintF(host, UHOSTLEN, "%s!%s", nick, s + 2);
	       G->LOG_MISC.Put(I->Lang( 0x006, "ss", nick, host));
	       common = 1;
	    } else {
	       G->MaskHost(s, host);
	       G->LOG_MISC.Put(I->Lang( 0x007, "ss", nick, host));
	    }
	    fr.SetGlobal(G->default_flags);
	    if (G->make_userfile) 
	      fr.SetGlobOwner();
	    fr.Sanity();
	    u = NEW ( UserRecord ( G->userlist, nick, host, fr));
	    while ((chan = chan->Next()))
	      if (chan->M()->Find(nick))
		I->CHANNELS->AddChanrec(u, chan->Key());
	    NOTICE(nick, I->Lang(0x008,"ss", nick, I->SERVER->botname));
	    NOTICE(nick, I->Lang(0x009,"s", host));
	    if (common) {
	       NOTICE(nick, I->Lang(0x00A,""));
	       NOTICE(nick, I->Lang(0x00B,""));
	    }
	    if (G->make_userfile) {
	       NOTICE(nick, I->Lang( 0x00C,""));
	       NOTICE(nick,I->Lang(0x00D,"s",I->SERVER->botname));
	       NOTICE(nick,I->Lang(0x00E,""));
	       G->LOG_MISC.Put(I->Lang( 0x00F, "s", nick));
	       G->make_userfile = 0;
	       G->userlist->WriteUserfile(0);
	       G->AddNote(nick, G->origbotname, I->Lang( 0x010, ""), 0, 0);
	    } else 
	      NOTICE(nick, I->Lang( 0x011, "s", I->SERVER->botname));
	    if (G->notify_new[0]) {
	       strncpy(s, I->Lang( 0x012, "ss", nick, host), UHOSTLEN);
	       s[UHOSTLEN] = 0;
	       strcpy(s1, G->notify_new);
	       while (s1[0]) {
		  if ((p1 = strchr(s1, ','))) {
		     *p1 = 0;
		     p1++;
		     G->RMSpace(p1);
		  }
		  G->RMSpace(s1);
		  G->AddNote(s1, G->origbotname, s, 0, 0);
		  if (!p1)
		    s1[0] = 0;
		  else
		    strcpy(s1, p1);
	       }
	    }
	 }
      }
   }
   return EXEC_LOG;
}

static BindReturn msg_ident (void ** args, const char *) {
   FUNCTION;
   char s[UHOSTLEN+1], s1[UHOSTLEN+1], * pass, who[NICKMAX+1];
   char *nick = (char *)args[0], buf[512], *par = buf;
   UserRecord * u2, *u = (UserRecord *)args[2];
   Channel3 *chan = (Channel3 *)G->chanset;
   FlagRecord fr (u);
   
   if (!I->SERVER->MatchMyNick(nick) && u && !fr.GlobBot()) {
      if (u && fr.GlobCommon()) {
	 if (!I->SERVER->quiet_reject)
	   NOTICE(nick,I->Lang(0x01C,""));
      } else {
	 strcpy(par,(char *)args[3]);
	 pass = G->Split(&par);
	 if (!par[0])
	   strcpy(who, nick);
	 else {
	    strncpy(who, par, NICKMAX);
	    who[NICKMAX] = 0;
	 }
	 u2 = G->userlist->Find(who);
	 fr.Set(u2);
	 if (u && (u == u2)) {
	    if (!I->SERVER->quiet_reject)
	      NOTICE(nick,I->Lang(0x01D,""));
	 } else if (u) {
	    if (!I->SERVER->quiet_reject)
	      NOTICE(nick,I->Lang(0x01E,"ss",nick,u->Handle(),nick));
	 } else if (u2 && strcasecmp(who, G->origbotname) && !fr.GlobBot()) {
	    if (u2->PassMatch("-")) {
	       if (!I->SERVER->quiet_reject)
		 NOTICE(nick,I->Lang(0x014,""));
	    } else if (!u2->PassMatch(pass)) {
	       if (!I->SERVER->quiet_reject)
		 NOTICE(nick,I->Lang(0x01F,""));
	    } else {
	       G->LOG_CMDS.Put(I->Lang(0x020, "sss",nick,(char *)args[1],who));
	       G->SPrintF(s, UHOSTLEN, "%s!%s", nick, (char *)args[1]);
	       G->MaskHost(s, s1);
	       NOTICE(nick,I->Lang(0x021,"s",s1));
	       u->AddHost(s1);
	       while ((chan = chan->Next())) {
		  user.Set(u,chan->Key());
		  if (chan->AutoOp() || user.AnyAutoop()) {
		     if (user.AnyValidOp() && chan->Meop())
		       chan->AddMode('+', 'o', nick);
		     else if (I->SERVER->use_ts4 && user.AnyValidHelper()
			      && (chan->Meop() || (!chan->NoHelper() && 
						   chan->MeHelper())))
		       chan->AddMode('+', 'h', nick);
		     else if (user.AnyValidVoice() && chan->Meop())
		       chan->AddMode('+', 'v', nick);
		  }
	       }
	       return EXECUTED;
	    }
	 }
	 G->LOG_CMDS.Put(I->Lang(0x022,"sss",nick,(char *)args[1],who));
      }
   }
   return EXECUTED;
}

static BindReturn msg_op (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   Channel3 *chan = (Channel3 *)G->chanset;
   ChannelMember * mx = 0;
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0])) {
      char buf[512], * par = buf, * pass;
      
      strcpy(par,(char *)args[3]);
      pass = G->Split(&par);
      if (u->PassMatch(pass) && !u->PassMatch("-")) {
	 if (par[0]) {
	    if ((chan = FindChan3(par)) && chan->Active() && chan->Meop()) {
	       user.Set(u,par);
	       if (chan->M()->FindUser(u) &&
		   (mx = chan->M()->Find((char *)args[0])) &&
		   !mx->HasOp() && user.AnyValidOp()) {
		  chan->AddMode('+', 'o', (char *)args[0]);
		  G->LOG_CMDS.Put(I->Lang(0x23,"ssss", (char *)args[0],
					(char *)args[1], u->Handle(), par));
	       }
	       return EXECUTED;
	    }
	 } else {
	    while ((chan = chan->Next())) {
	       user.Set(u,chan->Key());
	       if (chan->M()->FindUser(u) && (mx = chan->M()->Find((char *)args[0]))
		   && !mx->HasOp() && chan->Meop() && user.AnyValidOp())
		 chan->AddMode('+', 'o', (char *)args[0]);
	    }
	    G->LOG_CMDS.Put(I->Lang(0x24,"sss",args[0],args[1], u->Handle()));
	    return EXECUTED;
	 }
      }
      G->LOG_CMDS.Put(I->Lang(0x25,"sss",args[0],args[1],u->Handle()));
   }
   return EXECUTED;
}

static BindReturn msg_helper (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   Channel3 *chan = (Channel3 *)G->chanset;
   ChannelMember * mx = 0;
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0]) 
       && I->SERVER->use_ts4) {
      char buf[512], * par = buf, * pass;
      
      strcpy(par,(char *)args[3]);
      pass = G->Split(&par);
      if (u->PassMatch(pass) && !u->PassMatch("-")) {
	 if (par[0]) {
	    if ((chan = FindChan3(par)) && chan->Active() && 
		(chan->Meop() || (!chan->NoHelper() && chan->MeHelper()))) {
	       user.Set(u,par);
	       if (chan->M()->FindUser(u) &&
		   (mx = chan->M()->Find((char *)args[0])) &&
		   !mx->HasHelper() && user.AnyValidHelper()) {
		  chan->AddMode('+', 'h', (char *)args[0]);
		  G->LOG_CMDS.Put(I->Lang(0x137,"ssss", (char *)args[0],
					  (char *)args[1], u->Handle(), par));
	       }
	       return EXECUTED;
	    }
	 } else {
	    while ((chan = chan->Next())) {
	       user.Set(u,chan->Key());
	       if (chan->M()->FindUser(u) && (mx = chan->M()->Find((char *)args[0]))
		   && !mx->HasHelper() && user.AnyValidHelper() &&
		   (chan->Meop() || (!chan->NoHelper() && chan->MeHelper())))
		 chan->AddMode('+', 'h', (char *)args[0]);
	    }
	    G->LOG_CMDS.Put(I->Lang(0x138,"sss",args[0],args[1], u->Handle()));
	    return EXECUTED;
	 }
      }
      G->LOG_CMDS.Put(I->Lang(0x139,"sss",args[0],args[1],u->Handle()));
   }
   return EXECUTED;
}

static BindReturn msg_voice (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   Channel3 *chan = (Channel3 *)G->chanset;
   ChannelMember * mx = 0;
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0])) {
      char buf[512], * par = buf, * pass;
      
      strcpy(par, (char *)args[3]);
      pass = G->Split(&par);
      if (u->PassMatch(pass) && !u->PassMatch("-")) {
	 if (par[0] && (chan = FindChan3(par)) && chan->Active() && chan->Meop()) {
	    user.Set(u,par);
	    if (chan->M()->FindUser(u) && (mx = chan->M()->Find((char *)args[0]))
		&& !mx->HasVoice() && user.AnyValidVoice()) {
	       chan->AddMode('+', 'v', (char *)args[0]);
	       G->LOG_CMDS.Put(I->Lang(0x026,"ssss",args[0],args[1],
					u->Handle(),par));
	    }
	    return EXECUTED;
	 } else if (!par[0]) {
	    while ((chan = chan->Next())) {
	       user.Set(u,par);
	       if (chan->M()->FindUser(u) && 
		   (mx = chan->M()->Find((char *)args[0]))
		   && !mx->HasVoice() && user.AnyValidVoice()) {
		  chan->AddMode('+', 'v', (char *)args[0]);
	       }
	    }
	    G->LOG_CMDS.Put(I->Lang(0x026,"ssss",args[0],args[1],
				    u->Handle(),par));
	    return EXECUTED;
	 }
      }
      G->LOG_CMDS.Put(I->Lang(0x027,"sss",args[0],args[1],u->Handle()));
   }
   return EXECUTED;
}

static BindReturn msg_invite (void ** args, const char *) {
   FUNCTION;
   char * pass, buf[512], *par = buf;
   Channel3 * chan = (Channel3 *)G->chanset;
   UserRecord * u = (UserRecord *)args[2];
   int lang = 0x02B;
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0])) {
      strcpy(par,(char *)args[3]);
      pass = G->Split(&par);
      if (u->PassMatch(pass)) {
	 if (par[0] == '*') {
	    FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
	    
	    while ((chan = chan->Next())) {
	       if (chan->Mode() & CHANINV) {
		  fr.Set(u, chan->Key());
		  if (fr.AnyValidHelper()) {
		     I->SERVER->SendInvite(Q_SERVER,(char *)args[0],par);
		     lang = 0x0126;
		  }
	       }
	    }
	 } else if (!(chan = FindChan3(par))) 
	   NOTICE((char *)args[0],I->Lang(0x028,"s",I->SERVER->botname));
	 else if (!chan->Active()) 
	   NOTICE((char *)args[0],I->Lang(0x029,"s",par));
	 else {
	    I->SERVER->SendInvite(Q_SERVER,(char *)args[0],par);
	    lang = 0x02A;
	 }
      } 
      G->LOG_CMDS.Put(I->Lang(lang,"ssss",args[0],args[1],u->Handle(),par));
   }
   return EXECUTED;
}

static BindReturn msg_key (void ** args, const char *) {
   FUNCTION;
   Channel3 *chan = 0;
   char *pass, buf[512], *par = buf;
   FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL);
   UserRecord * u = (UserRecord *)args[2];
   int lang = 0x122;
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0])) {
      strcpy(par,(char *)args[3]);
      pass = G->Split(&par);
      if (!u->PassMatch(pass) || !(chan = FindChan3(par)))
	NOTICE((char *)args[0],I->Lang(0x120,"s",I->SERVER->botname));
      else if (!chan->Active()) 
	NOTICE((char *)args[0],I->Lang(0x121,"s",par));
      else if (fr.Set(u,par)->AnyValidHelper()) {
	 lang = 0x123;
	 if (chan->ChanKey()[0]) {
	    NOTICE((char *)args[0],I->Lang( 0x124, "sss", par, chan->ChanKey()));
	    if (I->invite_key && (chan->Mode() & CHANINV)) 
	      I->SERVER->SendInvite(Q_SERVER,(char *)args[0], par);
	 } else
	   NOTICE((char *)args[0],I->Lang( 0x125, "s"));
      }
      G->LOG_CMDS.Put(I->Lang(lang,"ssss",args[0],args[1],u->Handle(),par));
   }
   return EXECUTED;
}

static BindReturn msg_info (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   char s[121], *chname = 0, *nick = (char *)args[0], buf[512];
   char * par = buf;
   const char * p;
   int locked = 0, lang = 0;
   FlagRecord fr (u);
   
   if (!I->SERVER->MatchMyNick(nick) && I->CHANNELS->use_info && u
       && !fr.GlobCommon() && !fr.GlobBot()) {
      strcpy(par, (char *)args[3]);
      if (!u->PassMatch("-") && !u->PassMatch(G->Split(&par))) 
	lang = 0x02C;
      else {
	 if ((par[0] == '#') || (par[0] == '+') || (par[0] == '&'))
	   chname = G->Split(&par);
	 else
	   chname = 0;
	 if (par[0]) {
	    UserEntryInfo * uei = GetInfo(u);
	    if ((p = uei->Value()) && (p[0] == '@'))
	      locked = 1;
	    if (chname) {
	       I->CHANNELS->GetHandleChaninfo(u->Handle(), chname, s);
	       p = s;
	       if (s[0] == '@')
		 locked = 1;
	    }
	    if (locked) 
	      NOTICE(nick,I->Lang(0x033,""));
	    else if (!strcasecmp(par, "none")) {
	       par[0] = 0;
	       if (chname) 
		 I->CHANNELS->SetHandleChaninfo(G->userlist, u->Handle(), chname, par);
	       else
		 uei->Set(0);
	       NOTICE(nick,I->Lang(chname?0x034:0x035,"s",chname));
	       lang = chname ? 0x031 : 0x032;
	    } else if (par[0]) {
	       if (par[0] == '@')
		 par++;
	       NOTICE(nick,I->Lang(0x038,"s",par));
	       if (chname) {
		  I->CHANNELS->SetHandleChaninfo(G->userlist, u->Handle(), chname, par);
		  lang = 0x02D;
	       } else {
		  uei->Set(par);
		  lang = 0x02E;
	       }
	    } else {
	       lang = chname ? 0x02F : 0x030;
	       if (p && p[0]) {
		  NOTICE(nick,I->Lang(0x039,"s",par));
		  NOTICE(nick,I->Lang(chname?0x03A:0x03B,"ss",
					       I->SERVER->botname,chname));
	       } else
		 NOTICE(nick,I->Lang(chname?0x036:0x037,"s",chname));
	    }
	 }
      }
      if (lang)
	G->LOG_CMDS.Put(I->Lang(lang,"ssss",nick,args[1],u->Handle(),chname));
   }
   return EXECUTED;
}

#define HNOTICE(a,b) I->SERVER->SendNotice(Q_HELP, a, b)
static BindReturn msg_who (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   Channel3 *chan;
   ChannelMember * m;
   char s[UHOSTLEN+1], also[512], * par = (char *)args[3];
   const char * info;
   int i, lang;
   
   if (!I->SERVER->MatchMyNick((char *)args[0]) && I->CHANNELS->use_info && u)  {
      if (!par[0]) 
	HNOTICE((char *)args[0], I->Lang(0x03C, "s",I->SERVER->botname));
      else if (!(chan = FindChan3(par))) 
	HNOTICE((char *)args[0], I->Lang(0x03D, "s",par));
      else {
	 user.Set(u,par);
	 if (chan->Hidden() && !chan->M()->FindUser(u) &&
	     !user.AnyValidOp() && !user.AnyFriend()) 
	   HNOTICE((char *)args[0], I->Lang(0x03E, ""));
	 else {
	    context;
	    G->LOG_CMDS.Put(I->Lang(0x03F,"sss",args[0],args[1],u->Handle()));
	    also[0] = 0;
	    i = 0;
	    for (m = chan->M(); (m = m->Next());) {
	       user.Set((u = m->MemberUser()));
	       info = (!u || user.Set(u)->GlobBot()) ? 0 : GetInfo(u)->Value();
	       if (info && (info[0] == '@'))
		 info++;
	       else if (u) {
		  I->CHANNELS->GetHandleChaninfo(u->Handle(), chan->Key(), s);
		  if (s[0]) {
		     info = s;
		     if (info[0] == '@')
		       info++;
		  }
	       }
	       lang = 0;
	       if (info && info[0])
		 lang = 0x040;
	       else if (I->SERVER->MatchMyNick(m->Key()))
		 lang = 0x041;
	       else if (u && user.GlobBot()) {
		  FlagRecord bfl (u, FlagRecord::BOT);
		  lang = bfl.BotShare() ? 0x042 : 0x043;
	       }
	       if (lang) 
		 HNOTICE((char *)args[0],I->Lang(lang,"s s",m->Key(),
						 NICKMAX - strlen(m->Key()),
						 info));
	       else {
		  i += G->SPrintF(also + i, 450 - i, "%s, ", m->Key());
		  if (i > 400) {
		     also[i - 2] = 0;
		     HNOTICE((char *)args[0],I->Lang(0x044,"s",also));
		     i = 0;
		     also[0] = 0;
		  }
	       }
	    }
	    if (i) {
	       also[i - 2] = 0;
	       HNOTICE((char *)args[0],I->Lang(0x044,"s",also));
	    }
	    context;
	 }
      }
   }
   return EXECUTED;
}

static BindReturn msg_whois (void ** args, const char *) {
   FUNCTION;
   char s[UHOSTLEN+1], par[NICKMAX+1];
   const char * s2;
   Channel3 *chan = (Channel3 *)G->chanset;
   ChannelMember * m;
   ChannelRecord * cr;
   UserRecord * u2, *u = (UserRecord *)args[2];
   UserEntryInfo * uei;
   UserEntryXtra * uex;
   XtraKeys * xk = 0;
   int i;
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0])) {
      strncpy(par,(char *)args[3], NICKMAX);
      par[NICKMAX] = 0;
      G->LOG_CMDS.Put(I->Lang(0x045,"ssss",args[0],args[1],u->Handle(),par));
      if (!(u2 = G->userlist->Find((char *)args[2]))) {
	 /* no such handle -- maybe it's a nickname of someone on a chan? */
	 while ((chan = chan->Next()))
	   if ((m = chan->M()->Find(par)) && (u2 = m->MemberUser())) {
	      HNOTICE((char *)args[0],I->Lang(0x046,"ss",par,u2->Handle()));
	      break;
	   }
	 if (!chan) {
	    HNOTICE((char *)args[0],I->Lang(0x047,"s",args[3]));
	    return EXECUTED;
	 }
      }
      uei = FindInfo(u2); 
      victim.Set(u2);
      if ((s2 = uei ? uei->Value() : 0) && (s2[0] == '@'))
	s2++;
      if (s2 && s2[0] && !victim.GlobBot())
	HNOTICE((char *)args[0],I->Lang(0x048,"ss", u2->Handle(), s2));
      if ((uex = FindXtra(u2)) && (xk = uex->FindKey("EMAIL")))
	HNOTICE((char *)args[0],I->Lang(0x049,"ss", u2->Handle(), s2));
      chan = (Channel3 *)G->chanset;
      while ((chan = chan->Next())) {
	 user.Set(u,chan->Key());
	 if (chan->M()->FindUser(u2) && (!chan->Hidden() || chan->M()->FindUser(u) ||
				      user.AnyValidOp() || user.AnyFriend())) {
	    HNOTICE((char *)args[0],I->Lang(0x04A,"ss", u2->Handle(),
					    chan->Key()));
	    s[0] = 1;
	 } else if ((cr = u2->ChanRecs()->Find(chan->Key())) && cr->Laston()
		&& (!chan->Secret() || chan->M()->FindUser(u) ||
		    user.AnyValidOp() || user.AnyFriend())) {
	    time_t x = cr->Laston();
	    
	    strcpy(s, ctime(&x));
	    s[16] = 0;
	    HNOTICE((char *)args[0],I->Lang(0x04B,"sss", u2->Handle(), s,
					    chan->Key()));
	 }
      }
      if (!s[0])
	HNOTICE((char *)args[0],I->Lang(0x04C,"s", u2->Handle()));
      i = 0;
      if (victim.GlobMaster())
	i += G->SPrintF(s + i,UHOSTLEN-i,"%s, ",I->Lang(0x04F,""));
      else if (victim.GlobOp())
	i = G->SPrintF(s, UHOSTLEN, "%s, ", I->Lang(0x04D,""));
      if (victim.GlobBot())
	i += G->SPrintF(s + i,UHOSTLEN-i,"%s, ",I->Lang(0x04E,""));
      if (i)
	HNOTICE((char *)args[0],I->Lang(0x050,"ss",u2->Handle(),s));
   }
   return EXECUTED;
}

static BindReturn msg_reset (void ** args, const char *) {
   FUNCTION;
   Channel3 *chan = (Channel3 *)G->chanset;
   int lang = 0;
   
   if (!I->SERVER->MatchMyNick((char *)args[0])) {
      if (((char *)args[3])[0]) {
	 if (!(chan = FindChan3((char *)args[3]))) {
	    HNOTICE((char *)args[0], I->Lang(0x03D, "s",args[3]));
	 } else {
	    lang = 0x051;
	    chan->ResetInfo();
	 }
      } else {
	 lang = 0x053;
	 while ((chan = chan->Next()))
	   chan->ResetInfo();
      }
      if (lang) {
	 G->LOG_CMDS.Put(I->Lang(lang,"ssss",args[0],args[1],
				  ((UserRecord *)args[2])->Handle(), args[3]));
	 HNOTICE((char *)args[0], I->Lang(lang+1, "s", args[3]));
      }
   }
   return EXECUTED;
}

static BindReturn msg_go (void ** args, const char *) {
   FUNCTION;
   Channel3 *chan = (Channel3 *)G->chanset;
   int ok = 0, ok2 = 0, lang = 0;
   UserRecord * u = (UserRecord *)args[2];
   
   if (u && !I->SERVER->MatchMyNick((char *)args[0])) {
      if (((char *)args[3])[0]) {
	 if ((chan = FindChan3((char *)args[3]))) {
	    user.Set(u,(char*) args[3]);
	    if (user.AnyValidOp())
	      lang = 0x055;
	    else if (!chan->Meop()) {
	       I->SERVER->SendPart(Q_SERVER, chan->Key(), I->Lang(0x05B,""));
	       lang = 0x056;
	    } else 
	      lang = 0x057;
	 }
      } else {
	 while ((chan = chan->Next())) 
	   if (chan->M()->Find((char *)args[0])) {
	      user.Set(u,chan->Key());
	      if (user.AnyValidOp()) {
		 ok2 = 1;
		 if (!chan->Meop()) {
		    I->SERVER->SendPart(Q_SERVER, chan->Key(), I->Lang(0x05B,""));
		    ok = 1;
		 }
	      }
	   }
	 lang = ok ? 0x058 : ok2 ? 0x059 : 0x05A;
      }
      if (lang) 
	G->LOG_CMDS.Put(I->Lang(lang, "ssss", args[0], args[1],
				 u->Handle(), args[3]));
   }
   return EXECUTED;
}

BindReturn msg_addhost (void ** args, const char * ) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   FlagRecord fr;
   char * pass, buf [512], * par = buf, * nick = (char *)args[0];
   int lang = 0;

   buf[0] = 0;
   if (u && !fr.Set(u)->GlobBot() && !I->SERVER->MatchMyNick(nick)) {
      if (fr.GlobCommon()) 
	lang = 0x10C;
      else {
	 strncpyz(buf, (char *)args[3], 511);
	 pass = G->Split(&par);
	 if (!par[0]) 
	   lang = 0x013;
	 else if (strcasecmp(u->Key(), G->origbotname)) {
	    if (u->PassMatch("-")) 
	      lang = 0x014;
	    else if (!u->PassMatch(pass)) 
	      lang = 0x01F;
	    else if (G->userlist->FindHost(par)) 
	      lang = 0x015;
	    else {
	       G->LOG_CMDS.Put(G->Lang(0x016, "ssss", nick, args[1],
				       u->Key(), par));
	       lang = 0x017;
	       u->AddHost(par);
	    }
	 }
      }
      if (lang && !I->SERVER->quiet_reject)
	NOTICE(nick,I->Lang(lang,"s", par));
   }
   return EXECUTED;
}

/* MSG COMMANDS */
/*  function call should be:
 *  int msg_cmd("handle","nick","user@host","params");
 *  function is responsible for any logging
 *  (return 1 if successful, 0 if not) */
#define MSG_COMMANDS 13
static command_binds C_msg[MSG_COMMANDS]={
   { "addhost", "", msg_addhost, 0},
   { "go", "", msg_go, 0 },
   { "hello", "", msg_hello, 0 },
   { "helper", "h|h", msg_helper, 0 },
   { "ident", "", msg_ident, 0 },
   { "info", "", msg_info, 0 },
   { "invite", "h|h", msg_invite, 0 },
   { "key", "h|h", msg_key, 0 },
   { "op", "", msg_op, 0 },
   { "reset", "m", msg_reset, 0 },
   { "voice", "", msg_voice, 0 },
   { "who", "", msg_who, 0 },
   { "whois", "", msg_whois, 0 },
};

/*************************************************************************
 * SCRIPT COMMANDS                                                       *
 *************************************************************************/

/* streamlined by answer */
static int script_chanlist (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   int f;
   ChannelMember * m;
   Channel3 *chan;
   FlagRecord plus (FlagRecord::CHANNEL | FlagRecord::GLOBAL 
		    | FlagRecord::BOT);
   FlagRecord minus (FlagRecord::CHANNEL | FlagRecord::GLOBAL 
		    | FlagRecord::BOT);
   FlagRecord user (FlagRecord::CHANNEL | FlagRecord::GLOBAL 
		    | FlagRecord::BOT);
   const char * p;

   BADARGS(2, 3, I->Lang( 0x0B7, ""));
   context;
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   context;
   if (argc == 2) {
      for (m = chan->M(); (m = m->Next());) {
	 p = m->Key();
	 sc->AppendArray(1, &p);
      }
      return SCRIPT_OK;
   }
   plus.BreakDown(argv[2]->AsString(),&minus);
   f = minus.AnyFlags();
   if (!plus.AnyFlags() && !f)
      return SCRIPT_OK;
   minus ^= plus;
   user.SetMatch(plus.Match());
   for (m = chan->M(); (m = m->Next());) {
      user.Set(m->MemberUser(),argv[1]->AsString());
      if ((plus == user) && (!f || (minus == user))) {
	 p = m->Key();
	 sc->AppendArray(1, &p);
      }
   }
   return SCRIPT_OK;
}

static int script_botisop (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(chan->Meop() ? "1" : "0");
   return SCRIPT_OK;
}

static int script_isop (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * mx;

   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(((mx = chan->M()->Find(argv[1]->AsString())) && mx->HasOp())
		    ? "1" : "0");
   return SCRIPT_OK;
}

static int script_botishelper (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString((I->SERVER->use_ts4 && chan->MeHelper()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_ishelper (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * mx;

   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(((mx = chan->M()->Find(argv[1]->AsString())) && 
		  I->SERVER->use_ts4 && mx->HasHelper()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_isvoice (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * mx;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(((mx = chan->M()->Find(argv[1]->AsString()))
		  && mx->HasVoice()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_onchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(chan->M()->Find(argv[1]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_handonchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   UserRecord * u;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(((u = G->userlist->Find(argv[1]->AsString())) &&
			  chan->M()->FindUser(u)) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_ischanban (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(3, 3, I->Lang( 0x0BA, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(chan->ChanBans()->Find(argv[1]->AsString()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_ischanexception (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(3, 3, I->Lang( 0x0BA, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString((I->SERVER->use_ts4 && chan->Exceptions()->Find(argv[1]->AsString()))
		 ? "1" : "0");
   return SCRIPT_OK;
}

static int script_getchanhost (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * m;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   if ((m = chan->M()->Find(argv[1]->AsString())))
     sc->SetString(m->Host());
   return SCRIPT_OK;
}

static int script_onchansplit (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * m;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(((m = chan->M()->Find(argv[1]->AsString()))
		  && m->IsSplit()) ? "1" : "0");
   return SCRIPT_OK;
}

static int script_getchanidle (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * m;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString((m = chan->M()->Find(argv[1]->AsString())) ? 
		 G->ToBase10(m->Age() / 60) : 0);
   return SCRIPT_OK;
}

static int script_chanbans (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   ChannelBan *b;
   Channel3 *chan;
   const char * list[3];

   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   for (b = chan->ChanBans(); b; b = b->Next()) {
      list[0] = b->Key();
      list[1] = b->Setter();
      list[2] = G->ToBase10(b->Age());
      sc->AppendArray(3, list);
   }
   return SCRIPT_OK;
}

static int script_chanexceptions (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   ChannelBan *b;
   Channel3 *chan;
   const char * list[3];

   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   if (I->SERVER->use_ts4) {
      for (b = chan->Exceptions(); b; b = b->Next()) {
	 list[0] = b->Key();
	 list[1] = b->Setter();
	 list[2] = G->ToBase10(b->Age());
	 sc->AppendArray(3, list);
      }
   }
   return SCRIPT_OK;
}

static int script_getchanmode (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(chan->GetMode());
   return SCRIPT_OK;
}

static int script_getchanjoin (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   ChannelMember * m;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString())))
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
   else if (!(m = chan->M()->Find(argv[1]->AsString())))
     sc->SetString(I->Lang( 0x078, "ss", argv[1]->AsString(), 
			   argv[2]->AsString()));
   else {
      sc->SetInteger(m->Joined());
      return SCRIPT_OK;
   }
   return SCRIPT_ERROR;
}

/* flushmode <chan> */
static int script_flushmode (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   chan->FlushMode(Q_SERVER);
   return SCRIPT_OK;
}

static int script_pushmode (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   char plus, mode;
   
   BADARGS(3, 4, I->Lang( 0x0BB, ""));
   if (!(chan = FindChan3(argv[1]->AsString())))
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
   else {
      plus = argv[2]->AsString()[0];
      mode = argv[2]->AsString()[1];
      if ((plus != '+') && (plus != '-')) {
	 mode = plus;
	 plus = '+';
      }
      if ((mode < 'a') || (mode > 'z')) 
	sc->SetString(I->Lang( 0x0BC, "s", argv[2]->AsString()));
      else if ((argc < 4) && strchr("bvo", mode)) 
	sc->SetString(I->Lang( 0x0BD, "c", mode));
      else {
	 chan->AddMode(plus, mode, (argc == 4) ? argv[3]->AsString() : "");
	 return SCRIPT_OK;
      }
   }
   return SCRIPT_ERROR;
}

static int script_resetbans (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   chan->ResetBans();
   return SCRIPT_OK;
}

static int script_resetchan (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   context;
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   chan->ResetInfo();
   return SCRIPT_OK;
}

static int script_topic (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   
   context;
   BADARGS(2, 2, I->Lang( 0x0B8, ""));
   if (!(chan = FindChan3(argv[1]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   sc->SetString(chan->Topic());
   return SCRIPT_OK;
}

static int script_hand2nick (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   ChannelMember *m;
   Channel3 *chan;
   UserRecord * u;
   
   BADARGS(3, 3, I->Lang( 0x0BE, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[1]->AsString()));
      return SCRIPT_ERROR;
   }
   for (m = chan->M(); (m = m->Next());)
     if ((u = m->MemberUser()) && !strcasecmp(u->Handle(), argv[1]->AsString())) {
	sc->SetString(m->Key());
	break;
     }
   return SCRIPT_OK;		/* blank */
}

static int script_nick2hand (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Channel3 *chan;
   UserRecord * u;
   ChannelMember *m;
   
   BADARGS(3, 3, I->Lang( 0x0B9, ""));
   if (!(chan = FindChan3(argv[2]->AsString()))) {
      sc->SetString(I->CHANNELS->Lang( 0x103, "s", argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   if ((m = chan->M()->Find(argv[1]->AsString()))) {
      u = m->MemberUser();
      sc->SetString(u ? u->Handle() : "*");
   }
   return SCRIPT_OK;
}

/**************************************************************************
 * MISC IRC THINGS                                                        *
 **************************************************************************/

/* log the channel members */
static void log_chans() {
   FUNCTION;
   ChannelMember * m;
   Channel3 *chan = (Channel3 *)G->chanset;
   int chops = 0, ok = 0;
   
   while ((chan = chan->Next())) 
     if (chan->Active() && chan->LogStatus()) {
	int mem = chan->M()->Count();
	for (m = chan->M(); (m = m->Next());)
	  if (m->HasOp())
	    chops++;
	G->LOG_MISC.Put(I->Lang( chan->Meop() ? 0x116 : 0x117, "sddd",
				chan->Key(), mem, chops,
				chan->Bans()->Count()), chan);
	if (chan->Active() && (mem == 1))
	  ok = 1;
     }
   if (ok && I->SERVER) 
     I->SERVER->SendLusers();
}

static void check_expired_chanstuff() {
   FUNCTION;
   ChannelBan * b;
   ChannelMember * n, *m;
   char s[UHOSTLEN+1];
   Channel3 *chan = (Channel3 *)G->chanset;
   ChannelSettingsIdle * idle_kick;
   
   if (!I->SERVER->server_online)
     return;
   while ((chan = chan->Next())) {
      if (!chan->Meop()) {
	 void * argv = chan;
	 
	 I->bt_need->Check("op", 0, &argv);
      }
      if (I->SERVER->server_online && !(chan->Active() | chan->Pending()))
	I->SERVER->SendJoin(Q_MODE, chan->Key(),safe_str(chan->Nkey()));
      if (chan->DynamicBans() && chan->Meop())
	for (b = chan->ChanBans(); (b = b->Next()); )
	  if ((b->Age() > (60 * I->CHANNELS->ban_time)) &&
	      !(chan->MatchBan(b->Key()) == 2)) {
	     I->LOG_MODES.Put(I->Lang(0x118,"ss",chan->Key(), b->Key()),chan);
	     chan->AddMode('-', 'b', b->Key());
	     b->Touch();
	  }
      for (m = chan->M()->Next(); m; m = n) {
	 n = m->Next();
	 if (m->IsSplit()) {
	    if (!chan->Active()) {
	       DELETE ( m );
	    } else if (m->SplitAge() > I->wait_split) {
	       void * args[5];
	       
	       args[0] = (void *)m->Key();
	       args[1] = (void *)m->Host();
	       args[2] = m->MemberUser();
	       args[3] = chan;
	       args[4] = I->Lang( 0x119, "");
	       G->SPrintF(s, 511, "%s %s", chan->Key(), args[4]);
	       user.Set(m->User(),chan->Key());
	       I->bt_sign->Check(s, &user, args);
	       I->LOG_JOIN.Put(I->Lang( 0x11A, "ss", m->Key(), m->Host()),chan);
	       DELETE ( m );
	    }
	 }
      }
      idle_kick = (ChannelSettingsIdle *)
	chan->FindSettings(I->CHANSET_IDLE);
      if (chan->Active() && idle_kick->Time() && chan->Meop())
	for (m = chan->M(); (m = m->Next());)
	  if (m->Age() >= (idle_kick->Time() * 60) &&
	      !I->SERVER->MatchMyNick(m->Key())) {
	     user.Set(m->MemberUser(), chan->Key());
	     if (!user.AnyValidOp() && !user.AnyFriend())
	       m->Kick(I->Lang( 0x11B, "d", idle_kick->Time()));
	  }
      chan->CheckLonely();
   }
}

static void irc_postrehash (void) {
   FUNCTION;
   Channel3 * chan = (Channel3 *)G->chanset;
   
   while ((chan = chan->Next())) {
      chan->FlushMode(Q_SERVER);
      chan->RedoCmode();
   }
   I->CHANNELS->ReadChannels();
}

/* for EVERY channel */
static void flush_modes() {
   FUNCTION;
   Channel3 *chan = (Channel3 *)G->chanset;
   
   while ((chan = chan->Next()))
     chan->FlushMode(Q_SERVER);
}

char * irc_start (ModuleEntryEggdrop * F, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   G = F;
   I = NEW ( ModuleEntryIrc ( me, &r ) );
   if (!r) {
      NEW ( Command ("botisop", script_botisop, I) );
      NEW ( Command ("botishelper", script_botishelper, I) );
      NEW ( Command ("isop", script_isop, I) );
      NEW ( Command ("isvoice", script_isvoice, I) );
      NEW ( Command ("ishelper", script_ishelper, I) );
      NEW ( Command ("onchan", script_onchan, I) );
      NEW ( Command ("nick2hand", script_nick2hand, I) );
      NEW ( Command ("handonchan", script_handonchan, I) );
      NEW ( Command ("hand2nick", script_hand2nick, I) );
      NEW ( Command ("chanbans", script_chanbans, I) );
      NEW ( Command ("chanexceptions", script_chanexceptions, I) );
      NEW ( Command ("ischanban", script_ischanban, I) );
      NEW ( Command ("ischanexception", script_ischanexception, I) );
      NEW ( Command ("resetbans", script_resetbans, I) );
      NEW ( Command ("resetchan", script_resetchan, I) );
      NEW ( Command ("getchanhost", script_getchanhost, I) );
      NEW ( Command ("getchanjoin", script_getchanjoin, I) );
      NEW ( Command ("onchansplit", script_onchansplit, I) );
      NEW ( Command ("chanlist", script_chanlist, I) );
      NEW ( Command ("getchanidle", script_getchanidle, I) );
      NEW ( Command ("getchanmode", script_getchanmode, I) );
      NEW ( Command ("flushmode", script_flushmode, I) );
      NEW ( Command ("pushmode", script_pushmode, I) );
      NEW ( Command ("topic", script_topic, I) );
      context;
      G->hook_minutely->Add(check_expired_chanstuff,I);
      G->hook_5minutely->Add(log_chans,I);
      G->hook_idle->Add(flush_modes,I);
      G->hook_rehash->Add(irc_postrehash,I);
      G->check_chanlist = irc_chanlist;
      context;
      G->bt_dcc->AddBinds(irc_dcc,DCC_COMMANDS,I);
      I->SERVER->bt_msg->AddBinds(C_msg,MSG_COMMANDS,I);
      I->SERVER->bt_raw->AddBinds(irc_raw,RAW_COMMANDS,I);
      context;
   }
   return r;
}
