/*
 * DccLedit.C - a user in line edit mode
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryLedit.h"
#include "DccLedit.h"
#include "UserEntryLaston.h"
#include "LeditList.h"
#define CLASS "DccLedit"

DccLedit::DccLedit ( DccChat & old, const char * fn)
: DccChat ( old ) {
   FUNCTION;
   strncpy(fname, safe_str(fn), DIRLEN);
   type = LE->DCC_LEDIT;
   type_flags = DCT_MASTER;
   Printf("\n");
   context;
   insert = 1;
   if (!ReadFile()) {
      int lang;
      
      for (lang = 0x002; lang < 0x005; lang++)
	Send(LE->Lang(lang, "ss", fname));
      TouchLaston(user,"line editing",G->now);
      file->Display(this, 0, 10, 0);
   } else {
      DccChat * newone;
      
      Send(LE->Lang(0x008, "s", fname));
      newone = NEW (DccChat (* this ) );
      newone->CheckChon();
   }
}

DccLedit::~DccLedit ( void ) {
   FUNCTION;
   if (file)
     DELETE ( file );
}

int DccLedit::ExpMem ( void ) {
   FUNCTION;
   return sizeof(DccLedit);
}

void DccLedit::Activity (char * buf, int) {
   FUNCTION;
   DccChat * newone;
   int i, start = 0, count = 10;
   char * x;
   
   if (*buf == '.') {
      char * cmd = G->Split(&buf), *arg = G->Split(&buf);
      
      switch (cmd[1]) {
	case 'q':
	 Send(LE->Lang(0x00A, "" ));
	 newone = NEW ( DccChat (* this ) );
	 newone->CheckChon();
	 break;
	case 'h':
	 for (i = 0x00B; i < 0x010; i++)
	   Send(LE->Lang(i, ""));
	 break;
	case 'l':
	 if (*arg) 
	   NewFile ( arg );
	 Send(LE->Lang( ReadFile (  ) ? 0x010 : 0x011, "s", fname));
	 break;
	case 's':
	 if (*arg) 
	   NewFile ( arg );
	 Send(LE->Lang( WriteFile (  ) ? 0x012 : 0x013, "s", fname));
	 break;
	case 'k': 
	 count = 1;
	case 'd': 
	 if ((x = strchr(arg, ','))) {
	    *x = 0;
	    count = atoi(x + 1);
	 }
	 if (*arg)
	   start = atoi(arg);
	 current_line = start;
	 if (cmd[1] == 'd') 
	   file->Display(this, start, count, start);
	 else 
	   while ( count-- ) 
	     file->RemoveData( start );
	 break;
	case 'e':
	case 'i':
	 if (*arg)
	   current_line = atoi(arg);
	 insert = (cmd[1] == 'i');
	 break;
	case 'x':
	 file->Debug(this);
	 break;
	default:
	 Send(G->Lang(0x108, ""));
      }
   } else if (insert) 
	file->AddData(current_line++, buf);
   else 
     file->ModifyData(current_line++, buf);
   Send(LE->Lang(insert ? 0x014 : 0x015, ""));
   file->Display(this, current_line, 1, current_line);
}

int DccLedit::ReadFile ( void ) {
   FUNCTION;
   FILE * f = fopen(fname, "r");
   char buf[512]; /* if the lines are > 512 then uses should sodoff */
   int line = 0, l;
   
   context;
   if (file)
     DELETE ( file );
   file = NEW ( LeditList ( 0 ) );
   if (!f) 
     return (fopen(fname, "w") ? 0 : 1);
   while (fgets(buf, 511, f) && !feof(f)) {
      for (l = strlen(buf); buf[--l] == '\n'; buf[l] = 0);
      file->AddData(line ++, buf);
   }
   fclose(f);
   return 0;
}

int DccLedit::WriteFile ( void ) {
   FUNCTION;
   FILE * f = fopen(fname, "w");
   int ret = 1;
   
   context;
   if (*fname && f) {
      ret = file->Dump(f);
      fclose(f);
   }
   return ret;
}

void DccLedit::Display2 (Socket * where) {
   FUNCTION;
   where->Send(LE->Lang(0x005, "s", fname));
}

void DccLedit::Eof (void) {
   FUNCTION;
   con_flags = 0;
   G->LOG_MISC.Put(LE->Lang(0x006, "ssd", nick, host, remoteport));
   Close();
}
