/*
 * LeditList.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryLedit.h"
#include "LeditList.h"
#include "LeditData.h"
#include "Socket.h"
#define CLASS "LeditData"

LeditData::~LeditData ( void ) {
   FUNCTION;
   if (key)
     my_free(key, G->memtype_socket);
}

int LeditData::ExpMem ( void ) {
   FUNCTION;
   return 0;
}

LeditList::LeditList ( LeditList & s, int base, int hold )
: IntList (s, base, INSERT) {
   FUNCTION;
   held = NEW ( LeditList ( hold ) );
}

LeditList::~LeditList ( void ) {
   FUNCTION;
   if (held)
     DELETE ( held );
}
   
int LeditList::ExpMem ( void ) {
   FUNCTION;
   int tot = sizeof(LeditList);
   LeditData * pt;
   LeditList * le = held;
   
   if (held->Key()) {
      tot += sizeof(LeditList);
      while ((le = le->Next()))
	tot += le->ExpMem();
   } else 
     for (pt = (LeditData *)held; pt; pt = pt->Next())
       tot += sizeof(LeditData) + strlen(pt->Key()) + 1;
   return tot;
}
   
LeditData * LeditList::AddData ( int index, const char * text) {
   FUNCTION;
   LeditData * p = 0;
   if (!held) {
      if (!next)
	NEW ( LeditList ( *this, 0, 0 ) );
      p = Next()->AddData (index, text);
   } else if ((index < key) || (!next && (index == key))) {
      if (held->Key()) {
	 LeditList * pt = held;
	 
	 while ((pt = pt->Next())) {
	    if ((index < pt->Key()) || (!pt->Next() && (index == pt->Key()))) {
	       p = pt->AddData(index, text);
	       break;
	    } else 
	      index -= pt->Key();
	 }
	 if (p && pt) {
	    int limit = 16;
	    
	    key ++;
	    if (pt->held->Key()) 
	      limit = 256;
	    if (pt->Key() > (limit * 2)) {
	       LeditList * newone =
		 NEW ( LeditList ( *pt, pt->Key() - limit, pt->held->key ) );
	       
	       pt->key = limit;
	       pt = pt->held;
	       while (limit-- && (pt = pt->Next()));
	       newone->held->next = pt->next;
	       pt->next = 0;
	    }
	 }
      } else {
	 LeditData * pt = (LeditData *)held;
	 
	 key++;
	 while (index-- && pt->Next())
	   pt = pt->Next();
	 p = NEW ( LeditData ( *pt, text ) );
      }
   } else if (next)
	p = Next()->AddData( index - key, text );
   return p;
}

LeditData * LeditList::ModifyData ( int index, const char * text) {
   FUNCTION;
   LeditData * p = FindData ( index );

   if (p)
     p->SetData ( text );
   return p;
}

LeditData * LeditList::FindData ( int index ) {
   FUNCTION;
   LeditData * p = 0;

   if (!held) {
      if (!next)
	NEW ( LeditList ( *this, 0, 0 ) );
      p = Next()->FindData( index );
   } else if (index < key) {
      if (held->Key()) {
	 LeditList * pt = held;

	 while ((pt = pt->Next())) {
	    if (index < pt->Key()) {
	       p = pt->FindData(index);
	       break;
	    } else
	      index -= pt->Key();
	 }
      } else {
	 p = (LeditData *)held;
	 while ((p = p->Next()) && index--);
      }
   } else if (next) 
	p = Next()->FindData( index - key );
   return p;
}

int LeditList::RemoveData ( int index ) {
   FUNCTION;
   int p = 0;
   
   if (!held) {
      if (!next)
	NEW ( LeditList ( *this, 0, 0 ) );
      p = Next()->RemoveData( index );
   } else if (index < key) {
      if (held->Key()) {
	 LeditList * o = held, *pt;
	 
	 while ((pt = o->Next())) {
	    if (index < pt->Key()) {
	       p = pt->RemoveData(index);
	       break;
	    } else {
	       o = pt;
	       index -= pt->Key();
	    }
	 }
	 if (p && pt) {
	    key --;
	    if (pt->Key() < 9) {
	       LeditList * q;
	       
	       if (!pt->Key()) {
		  o->next = pt->next;
		  pt->next = 0;
		  DELETE ( pt );
		  return p;
	       } else if (pt->Next() && ((pt->Key() + pt->Next()->Key()) < 17)) {
		  o = pt;
		  pt = pt->Next();
	       } else if ((o->Key() + pt->Key()) > 16)
		    return p;
	       q = o->held;
	       while (q && q->next)
		 q = q->Next();
	       q->next = pt->held->next;
	       pt->held->next = 0;
	       o->next = pt->next;
	       pt->next = 0;
	       o->key += pt->key;
	       DELETE ( pt );
	    }
	 }
      } else {
	 LeditData * o = (LeditData *)held, *pt;
	 
	 while ((pt = o->Next()) && index--)
	   o = pt;
	 if (pt) {
	    o->SetNext(pt->Next());
	    pt->SetNext(0);
	    DELETE ( pt );
	    p = 1;
	 }
      }
   } else if (next)
	p = Next()->RemoveData( index - key );
   return p;
}

int LeditList::Dump ( FILE * f ) {
   FUNCTION;
   int fail = 0;
   
   if (held && held->Key()) {
      LeditList * p = held;
      
      while (!fail && (p = p->Next()))
	fail = p->Dump(f);
   } else if (held) {
      LeditData * p = (LeditData *)held;
      
      while (!fail && (p = p->Next())) {
	 if (fprintf(f, "%s\n", p->Key()) == EOF)
	   fail = 1;
      }
   } else if (next)
	fail = Next()->Dump(f);
   else 
     fail = 1;
   return fail;
}

void LeditList::Display ( Socket * idx, int real, int count, int st) {
   FUNCTION;
   if (held && (st < key)) {
      int howmuch = count;
      if ((st + count) > key) {
	 howmuch = key - st;
	 count -= howmuch;
      }
      if (held->Key()) {
	 LeditList * pt = held;
	 
	 while ((pt = pt->Next()) && (st > pt->Key())) 
	   st -= pt->Key();
	 if (pt) {
	    pt->Display(idx, real, howmuch, st);
	    real += howmuch;
	 }
      } else {
	 LeditData * p = (LeditData *)held;

	 while ((p = p->Next()) && st)
	   st--;
	 while (p && howmuch) {
	    idx->Send(LE->Lang(0x007, "5s", real++, p->Key()));
	    p = p->Next();
	    howmuch--;
	 }
      }
   } else
	st -= key;
   if (next && count)
     Next()->Display(idx, real, count, st);
}

void LeditList::Debug ( Socket * idx ) {
   FUNCTION;
   idx->Printf("LeditList: base %d\n", key);
   if (held) {
      idx->Printf(" HeldFactor: %d\n", held->Key());
      if (held->Key()) 
	held->Next()->Debug(idx);
      else {
	 LeditData * ptr = (LeditData *)held;
	 int id = key;
	 
	 while ((ptr = ptr->Next())) 
	   idx->Printf("Data item #%d: %s\n", id++, ptr->Key());
      }
   }
   idx->Printf("Next (%d)\n", key);
   if (next)
     Next()->Debug(idx);
   idx->Printf("End (%d)\n", key);
}
