/*
 * mkdep.c - automatically create dependancies for modules
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "mkdep.h"

static struct dependtree * root = NULL;

int main (int argc, char **argv) {
   int depnumber;
   char *p, modname[63], modentry[128], depfile[128];

   if (argv[1]) {
      strncpy(modname, argv[1], 63);
      p = argv[1];
      p++;
      sprintf(depfile, "./%s.so.dep", argv[1]);
      sprintf(modentry, "./ModuleEntry%c%s.C", toupper(modname[0]), p);
      depnumber = scan(modentry, depfile);
      if (depnumber)
	 save(depfile, 0);
      else
	 save(depfile, 1);
   } else {
      printf("ERROR: No module name specified!\n");
      exit(0);
   }
}

int scan(char *modentry, char *depfile) {
   FILE *f;
   char workbuf[1024], depend[64];
   char *p;
   int count, depnumber = 0;

   f = fopen(modentry, "r");
   while (!feof(f)) {
      if (feof(f))
	 break;
      fgets(workbuf, sizeof(workbuf) - 1, f);
      p = strstr(workbuf, "Depend(\"");
      if (p) {
	 p += 8;
	 strncpy(workbuf, p, sizeof(workbuf) - 1);
	 for (count = 0; count<strlen(workbuf); count++)
	    if (workbuf[count] == '"') {
	       workbuf[count] = 0;
	       strncpy(depend, workbuf, 63);
	       if (strcmp(depend, "eggdrop")) {
		  depnumber++;
		  add(depend);
	       }
	    }
      }
   }
   fclose(f);
   return depnumber;
}

void add(char *name) {
   struct dependtree * depend = NULL, * dependold = NULL;

   for (depend = root; depend; dependold = depend, depend = depend->next)
      if (!strcmp(root->name, name))
	 break;

   if (!depend) {
      depend = (struct dependtree *)malloc(sizeof(struct dependtree));
      depend->next = root;
      root = depend;
   }
   strncpy(root->name, name, 63);
}

void save(char *depfile, int writenull) {
   FILE *f;
   char filedepend[strlen(depfile)+2];
   struct dependtree * depend = NULL, * dependold = NULL;

   f = fopen(depfile, "w");
   if (!writenull)
      printf("Dependancies created: ");
   for (depend = root; depend; dependold = depend, depend = depend->next) {
      sprintf(filedepend, "%s\n", depend->name);
      fputs(filedepend, f);
      if (!writenull)
	 printf("%s ", depend->name);
   }
   if (!writenull)
      printf("\n");
   fclose(f);
}
