/* 
 * ModuleEntryNotes.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryNotes.h"
#include "IntegerVariable.h"
#include "StringVariable.h"
#include "Socket.h"
#include "HelpReference.h"
#include "UserEntryFwd.h"
#include "UserEntryNotes.h"
#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
#endif
#define CLASS "ModuleEntryNotes"

static Version notes_version, required_eggdrop;
static struct patches notes_patches[] = { 
#include "patch.h"
};

ModuleEntryNotes::ModuleEntryNotes( ModuleEntry * old, char ** r) 
: ModuleEntry ( "notes", notes_version.Set(2,0), notes_patches, old ) {
   FUNCTION;
   NOTES = this;
   if (!Depend("eggdrop", required_eggdrop.Set(2,0)))
     *r = "Your eggdrop core is incompatible.";
   else {
      maxnotes = 50;
      note_life = 60;
      allow_fwd = 0;
      notify_users = 0;
      notefile[0] = 0;
      NEW ( IntegerVariable ( "note_life", &note_life, 0, this ) );
      NEW ( IntegerVariable ( "max_notes", &maxnotes, 0, this ) );
      NEW ( IntegerVariable ( "allow_fwd", &allow_fwd, 0, this ) );
      NEW ( IntegerVariable ( "notify_users", &notify_users, 0, this ) );
      NEW ( StringVariable ( "notefile", notefile, 120, 0 , this ) );
      NEW ( HelpReference("notes.help",this) );
      entrytype_fwd = NEW (UserEntryFwd (this) );
      entrytype_notes = NEW (UserEntryNotes (this) );
   }
}

ModuleEntryNotes::~ModuleEntryNotes ( void ) {
   FUNCTION;
}

int ModuleEntryNotes::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ModuleEntryNotes) + ModuleEntry::ExpMem ();
}

void ModuleEntryNotes::Report (Socket * idx, int details) {
   FUNCTION;
   if (details) 
     idx->Send(Lang(notefile[0] ? 0x039 : 0x03A, "s", notefile));
}

int ModuleEntryNotes::CountNotes (const char * user) {
   FUNCTION;
   int tot = 0, l;
   FILE *f;
   char s[513], *s1;
   
   if (notefile[0] && (f = fopen(notefile, "r"))) {
      for (fgets(s, 512, f); !feof(f); fgets(s, 512, f)) {
	 if (s[(l = strlen(s) - 1)] == '\n')
	   s[l] = 0;
	 G->RMSpace(s);
	 if (s[0] && (s[0] != '#') && (s[0] != ';')) {
	    s1 = s;
	    if (!strcasecmp(G->Split(&s1), user))
	      tot++;
	 }
      }
      fclose(f);
   }
   return tot;
}

/* convert a string like "2-4;8;16-" in an array {2, 4, 8, 8, 16, maxnotes, -1} */
void ModuleEntryNotes::Parse (int * dl, const char * s) {
   FUNCTION;
   int i=0;
   int idl=0;
   
   while ( s[i] && (idl < 124) ) {
      while (s[i] == ';')
	i++;
      if (s[i]) {
	 dl[idl] = (s[i] == '-') ? 1 : atoi(s + i);
	 idl++;
	 while ( s[i] && (s[i] != '-') && (s[i] != ';') )
	   i++;
	 if (s[i] == '-') {
	    dl[idl] = atoi(s + i + 1); /* will be 0 if not a number */
	    if (!dl[idl])
	      dl[idl] = maxnotes;
	 } else
	   dl[idl] = dl[idl - 1];
	 idl++;
	 while ( s[i] && (s[i] != ';') )
	   i++;
      }
   }
   dl[idl] = -1;
}

/* return true if 'in' is in intervals of 'dl' */
int ModuleEntryNotes::Contains (int * dl, int in) {
   FUNCTION;
   int i;
   
   for (i = 0; dl[i] != -1; i+= 2) 
     if ( (dl[i] <= in) && (in <= dl[i+1]) )
       return 1;
   return 0;
}

void ModuleEntryNotes::Read (const char * hand, char * nick, char * srd,
			     Socket * idx) {
   FUNCTION;
   FILE *f;
   char s[601], *to, *dt, *from, *s1, wt[100], *temp;
   time_t tt;
   int ix = 1, ir = 0, lang = 0, rd[128], l;

   if (!notefile[0] || !(f = fopen(notefile, "r"))) 
     lang = 0x00C;
   else {
      if (!srd[0])
	srd = "-";
      Parse(rd, srd);
      for (fgets(s, 600, f); !feof(f); fgets(s, 600, f)) {
	 if (s[l = strlen(s) - 1] == '\n')
	   s[l] = 0;
	 G->RMSpace(s);
	 if (s[0] && (s[0] != '#') & (s[0] != ';')) { /* not comment */
	    s1 = s;
	    to = G->Split(&s1);
	    if (!strcasecmp(to, hand)) {
	       int lapse;
	       from = G->Split(&s1);
	       dt= G->Split(&s1);
	       tt = atoi(dt);
	       strcpy(wt, ctime(&tt));
	       wt[16] = 0;
	       dt = wt+4;
	       lapse = ((G->now - tt) / 86400);
	       lang = (lapse >= note_life) ? 0x00D : ((lapse > note_life - 7)
						      ? 0x00E : 0x00F);
	       if (srd[0] == '+') {
		  if (idx && (ix == 1))
		    idx->Send(Lang(0x013, ""));
		  temp = Lang(lang+3, "2ssd", ix, from, dt, note_life - lapse);
	       } else if (Contains(rd, ix)) {
		  temp = Lang(lang, "2sssd", ix, from,dt,s1,note_life - lapse);
		  ir++;
	       } else
		 temp = NULL;
	       if (temp && idx) 
		 idx->Send(temp);
#ifdef HAVE_SERVER_H
	       else if (temp && SERVER)
		 SERVER->SendNotice(Q_HELP,nick, temp);
#endif
	       ix++;
	       lang = 0;
	    }
	 }
      }
      fclose(f);
      if (srd[0] == '+') {
	 if (ix == 1) 
	   lang = 0x00C;
	 else if (idx)
	   idx->Send(Lang(0x015, ""));
#ifdef HAVE_SERVER_H
	 else if (SERVER)
	   SERVER->SendNotice(Q_HELP, nick, Lang(0x016, "d", ix - 1));
#endif
      } else if (!ir && (ix > 1))
	   lang = 0x014;
      else if (!ir && (ix == 1)) 
	lang = 0x00C;
   }
   if (lang) {
      if (idx) {
	 idx->Send(Lang(lang, ""));
      }
   }
#ifdef HAVE_SERVER_H
   else if (SERVER)
     SERVER->SendNotice(Q_HELP, nick, Lang(lang, ""));
#else
   else
     nick = nick; /* shutup compilet */
#endif
}

void ModuleEntryNotes::Delete (const char * hand, char * nick, char * sdl,
			       Socket * idx) {
   FUNCTION;
   FILE *f, *g;
   char s[513], s2[DIRLEN+1], *to, *s1;
   int in = 1, lang = 0, er = 0, dl[128], l;

   if (!notefile[0] || !(f = fopen(notefile, "r"))) 
     lang = 0x00C;
   else {
      G->SPrintF(s2, 511, "%s~new", notefile);
      if (!(g = fopen(s2, "w"))) {
	 lang = 0x017;
	 fclose(f);
      } else {
	 if (sdl[0]==0)
	   sdl = "-";
	 Parse(dl, sdl);
	 while (fgets(s, 512, f) && !feof(f)) {
	    if (s[l = strlen(s) - 1] == '\n')
	      s[l] = 0;
	    G->RMSpace(s);
	    if (s[0] && (s[0] != '#') && (s[0] != ';')) {
	       /* not comment */
	       s1 = s;
	       to = G->Split(&s1);
	       if (!strcasecmp(to, hand)) {
		  if (!Contains(dl, in))
		    fprintf(g, "%s %s\n", to, s1);
		  else 
		    er++;
		  in++;
	       } else 
		 fprintf(g, "%s %s\n", to, s1);
	    } else
	      fprintf(g, "%s\n", s);
	 }
	 fclose(f);
	 fclose(g);
	 unlink(notefile);
#ifdef RENAME
	 rename(s2, notefile);
#else
	 G->MoveFile(s2, notefile);
#endif
	 if ((er == 0) && (in > 1))
	   lang = 0x014;
	 else if (in == 1) 
	   lang = 0x00C;
	 else
	   lang = (er == in - 1) ? 0x018 : 0x019;
      }
   }
   if (lang && idx) 
     idx->Send(Lang(lang, "dd", er, in - 1 - er));
#ifdef HAVE_SERVER_H
   else if (lang && SERVER)
     SERVER->SendNotice(Q_HELP, nick, Lang(lang, "dd", dl, in - 2));
#else
   else
     nick = nick; /* shut that !@&*$(@! compiler up */
#endif
}
