/* 
 * UserEntryNotes.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "UserEntryNotes.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "ModuleEntryNotes.h"
#include "Socket.h"
#include "UserRecord.h"
#define CLASS "UserEntryNotes"

UserEntryNotes::~UserEntryNotes ( void ) {
   FUNCTION;
}

UserEntry * UserEntryNotes::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW (UserEntryNotes (u, this ) );
}
   
int UserEntryNotes::ScriptGet (ScriptContext * sc, int argc,
			     ScriptArgument ** argv) {
   FUNCTION;
   FILE *f;
   char s[601], *to, *from, *dt, *s1;
   int count, read, nl[128], l;
   const char *list[3];
   
   context;
   BADARGS(3, 4, NOTES->Lang(0x02B, ""));
   if (argc == 3)
     sc->SetInteger(NOTES->CountNotes(owner->Key()));
   else if (!NOTES->notefile[0] || !(f = fopen(NOTES->notefile, "r")))
     sc->SetString("-2");
   else {
      count = 0;
      read = 0;
      if (argc == 4)
	NOTES->Parse(nl, (argv[3]->AsString()[0] == 0) ? "-" : 
		     argv[3]->AsString());
      while (fgets(s, 600, f) && !feof(f)) {
	 if (s[l = strlen(s) - 1] == '\n')
	   s[l] = 0;
	 G->RMSpace(s);
	 if (s[0] && (s[0] != '#') & (s[0] != ';')) {		/* not comment */
	    s1 = s;
	    to = G->Split(&s1);
	    if (!strcasecmp(to, owner->Handle())) {
	       read++;
	       if ((argc == 3) || NOTES->Contains(nl,read)) {
		  count++;
		  from = G->Split(&s1);
		  dt = G->Split(&s1);
		  list[0] = from;
		  list[1] = dt;
		  list[2] = s1;
		  sc->AppendArray(3, list);
	       }
	    }
	 }
      }
      if (count == 0)
	sc->SetString("0");
      fclose(f);
   }
   return SCRIPT_OK;
}

int UserEntryNotes::ScriptSet ( ScriptContext * sc, int argc, 
			     ScriptArgument ** argv) {
   FUNCTION;
   FILE *f, *g;
   char s[DIRLEN + 1], *to, *s1, * ret, inbuf[512];
   int read, erased, l, nl[128]; /* is it enough ? */
   
   context;
   BADARGS(4, 4, NOTES->Lang(0x041, ""));
   if (!NOTES->notefile[0]) 
     ret = "-2";
   else if (!(f = fopen(NOTES->notefile, "r"))) 
     ret = "-2";
   else {
      G->SPrintF(s, DIRLEN, "%s~new", NOTES->notefile);
      if (!(g = fopen(s, "w")))
	ret = "-2";
      else {
	 read = 0;
	 erased = 0;
	 NOTES->Parse(nl, (argv[3]->AsString()[0] == 0) ? "-" : argv[3]->AsString());
	 while (!feof(f)) {
	    fgets(s, 512, f);
	    if (s[l = strlen(s) - 1] == '\n')
	      s[l] = 0;
	    if (!feof(f)) {
	       G->RMSpace(s);
	       if (s[0] && (s[0] != '#') & (s[0] != ';')) { /* not comment */
		  s1 = s;
		  to = G->Split(&s1);
		  if (!strcasecmp(to, owner->Key())) {
		     read++;
		     if (!NOTES->Contains(nl, read)) 
		       fprintf(g, "%s %s\n", to, s1);
		     else 
		       erased++;
		  } else 
		    fprintf(g, "%s %s\n", to, s1);
	       }
	    }
	 }
	 G->SPrintF(inbuf, 511, "%d", erased);
	 ret = inbuf;
	 fclose(f);
	 fclose(g);
	 unlink(NOTES->notefile);
#ifdef RENAME
	 rename(s, NOTES->notefile);
#else
	 G->MoveFile(s, NOTES->notefile);
#endif
      }
   }
   sc->SetString(ret);
   return SCRIPT_OK;
}

void UserEntryNotes::Display (Socket * idx) {
   FUNCTION;
   int i = NOTES->CountNotes(owner->Key());
   
   if (i)
     idx->Send(NOTES->Lang(0x040, "d", i));
}

int UserEntryNotes::ExpMem ( void ) {
   FUNCTION;
   return sizeof(UserEntryNotes);
}
