/* 
 * Notes.mod -- module handles: (the old notes.c + extras)
 * reading and sending notes
 * killing old notes and changing the destinations
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include <fcntl.h>
#include "DccChat.h"
#include "BindTable.h"
#include "UserRecord.h"
#include "UserEntryComment.h"
#include "HookList.h"
#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
#endif
#ifdef HAVE_IRC_H
#include "irc.h"
#include "Channel3.h"
#include "ChannelMember.h"
#endif
#ifdef HAVE_BOTNET_H
#include "ModuleEntryBotnet.h"
#include "TandemBot.h"
#endif
#include "ModuleEntryNotes.h"
#include "UserEntryFwd.h"
#define CLASS ""

ModuleEntryNotes * NOTES;
static ModuleEntryEggdrop * G;

/* change someone's handle */
static BindReturn notes_change (void ** args, const char * oldnick) {
   FUNCTION;
   FILE *f, *g;
   char s[513], s2[DIRLEN+1], *to, *s1, * newnick = (char *)args[1];
   int tot = 0, l;
   
   if (strcasecmp(oldnick, newnick)) {
      if (NOTES->notefile[0] && (f = fopen(NOTES->notefile, "r"))) {
	 G->SPrintF(s2, DIRLEN, "%s~new", NOTES->notefile);
	 if ((g = fopen(s2, "w"))) {
	    for (fgets(s, 512, f); !feof(f); fgets(s, 512, f)) {
	       if (s[l = strlen(s) - 1] == '\n')
		 s[l] = 0;
	       G->RMSpace(s);
	       if (s[0] && (s[0] != '#') && (s[0] != ';')) {	/* not comment */
		  s1 = s;
		  to = G->Split(&s1);
		  if (!strcasecmp(to, oldnick)) {
		     tot++;
		     fprintf(g, "%s %s\n", newnick, s1);
		  } else
		    fprintf(g, "%s %s\n", to, s1);
	       } else
		 fprintf(g, "%s\n", s);
	    }
	    fclose(g);
	    fclose(f);
	    unlink(NOTES->notefile);
#ifdef RENAME
	    rename(s2, NOTES->notefile);
#else
	    G->MoveFile(s2, NOTES->notefile);
#endif
	    if (tot)
	      G->LOG_MISC.Put(NOTES->Lang((tot == 1) ? 0x001 : 0x002, "dss",
					  tot, oldnick, newnick));
	 } else
	   fclose(f);
      }
   }
   return EXECUTED;
}

/* add note to notefile */
static BindReturn storenote (void ** argv, const char *) {
   FUNCTION;
   FILE * f;
   Socket * idx = (Socket *)argv[4];
   char u[HANDLEN+2], work[512], work2[512], * msg = (char *)argv[2];
   char * toreal = (char *)argv[1], * fromreal = (char *)argv[0];
   const char * to = 0;
   UserRecord * ur;
   UserEntryFwd * f1 = 0;
   int lang;
   BindReturn ret = EXECUTED;
   
   ur = G->userlist->Find(toreal);
   if (ur && NOTES->allow_fwd && (f1 = GetFwd(ur)) && f1->Value()) {
      char fwd[161], *q, *r;
      const char * p;
      int ok = 1;
#ifdef HAVE_BOTNET_H
      Version required (2, 0);
      ModuleEntryBotnet * me = 
	(ModuleEntryBotnet *)G->module_list->Find("botnet", &required);
#endif
      ur->GetEntry(NOTES->entrytype_notes); /* make sure a notes entry exists */
      /* user is valid & has a valid forwarding address */
      strcpy(fwd,f1->Value()); /* only 40 bytes are stored in the userfile */
      q = strchr(fwd, '@');
      if (q && !strcasecmp(q + 1,G->botnetnick)) {
	 *q = 0;
	 if (!strcasecmp(fwd,toreal)) /* they're forward to themselves on
				       * the same bot, llama's */
	   ok = 0;
	 p = 0;
      } else 
	p = q;
      if ((fromreal[0] != '@') && ((msg[0] == '<') || (msg[0] == '>')))
	ok = 0; /* probablly fake pre 1.3 hax0r */
      if (ok && (!p
#ifdef HAVE_BOTNET_H
		 || (me && !me->tandbot->Find(p+1))
#endif
		 )) {
	 if (p)
	   p++;
	 strncpy(work2, msg, 511);
	 work2[511] = 0;
	 q = work2;
	 while (ok && q && (q = strchr(q,'<'))) {
	    q++;
	    if ((r = strchr(q,' '))) {
	       *r = 0;
	       if (!strcasecmp(fwd,q))
		 ok = 0;
	       *r = ' ';
	    }
	 }
	 if (ok) {
	    if (p && strchr(fromreal,'@')) {
	       G->SPrintF(work, 511, "<%s@%s >%s %s", toreal,
			  G->botnetnick, fromreal, msg);
	       G->SPrintF(u, HANDLEN+1, "@%s",G->botnetnick);
	       p = u;
	    } else {
	       G->SPrintF(work, 511, "<%s@%s %s", toreal, G->botnetnick, msg);
	       p = toreal;
	    }
	 }
      } else 
	   ok = 0;
      if (ok) {
	 if ((G->AddNote(fwd,p,work,idx,0) == NOTE_OK) && idx)
	   idx->Send(NOTES->Lang(0x006, "s", f1->Value()));
	 to = 0;
      } else {
	 strcpy(work,msg);
	 strncpyz(msg, fwd, NOTENAMELEN);
	 to = toreal;
	 ret = EXEC_LOG;
      } 
   } else
     to = toreal;
   if (to) {
      if (!NOTES->notefile[0]) 
	lang = 0x007;
      else if (NOTES->CountNotes(to) >= NOTES->maxnotes) 
	lang = 0x008;
      else { /* time to unpack it meaningfully */
	 if (!(f = fopen(NOTES->notefile, "a")) && 
	     !(f = fopen(NOTES->notefile, "w"))) {
	    G->LOG_MISC.Put(NOTES->Lang(0x009, ""));
	    lang = 0x00A;
	 } else {
	    char * p, *q;
	    const char *from = fromreal;
	    int l = 0;
	    	
	    strncpyz(work2, msg, 511);
	    q = work2;
	    while ((*q == '<') || (*q == '>')) {
	       p = G->Split(&q);
	       if (*p == '<')
		 l += G->SPrintF(work + l, 511 - l, "via %s, ", p + 1);
	       else if (toreal[0] == '@')
		 from = p + 1;
	    }  
	    fprintf (f, "%s %s %lu %s%s\n", to, from, G->now,
		     l ? work : "", msg);
	    fclose(f);
	    lang = 0x00B;
	 }
      }
      if (idx && lang)
	idx->Send(NOTES->Lang(lang, ""));
   }
   return ret;
}

static BindReturn cmd_notes (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char *fcn, *par = (char *)args[1];
   int i;
   
   if (!par[0]) 
     for (i = 0x02C; i < 0x031; i++) 
	idx->Send(NOTES->Lang(i, ""));
   else {
      fcn = G->Split(&par);
      if (!strcasecmp(fcn, "index"))
	NOTES->Read(idx->Nick(), "", "+", idx);
      else if (!strcasecmp(fcn, "read")) 
	NOTES->Read(idx->Nick(), "", strcasecmp(par, "all") ? par : "-", idx);
      else if (!strcasecmp(fcn, "erase")) 
	NOTES->Delete(idx->Nick(), "", strcasecmp(par, "all") ? par : "-", idx);
      else {
	 idx->Send(NOTES->Lang(0x03E, ""));
	 return EXECUTED;
      }
      G->LOG_CMDS.Put(NOTES->Lang(0x03F, "sss", idx->Nick(), fcn, par));
   }
   return EXECUTED;
}

static BindReturn cmd_fwd (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * handle = 0, buf[128], *par = buf;
   UserRecord * u1 = 0;
   int lang = 0;
   FlagRecord fr;
   
   strncpy(par, (char *)args[1],127);
   par[127] = 0;
   if (!par[0]) 
     lang = 0x031;
   else if (!(u1 = G->userlist->Find(handle = G->Split(&par))))
     lang = 0x032;
   else if (fr.Set(u1)->GlobOwner() && strcasecmp(handle, idx->Nick()))
     lang = 0x033;
   else {
      lang = par[0] ? 0x036 : 0x034;
      G->LOG_CMDS.Put(NOTES->Lang(lang + 1, "sss", idx->Nick(), handle, par));
      GetFwd(u1)->Set(par[0] ? par : 0);
   }
   idx->Send(NOTES->Lang(lang, "ss", handle, par));
   return EXECUTED;
}

#ifdef HAVE_SERVER_H
/* notes <pass> <func> */
static BindReturn msg_notes (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   char *fcn, buf[512], *par = buf, *nick = (char *)args[0];
   char *host = (char *)args[1], *to;
   int lang = 0;  
   FILE *f;
   UserRecord * u2;
   Socket * ptr = G->socklist;
   FlagRecord fr;
     
   if (!NOTES->SERVER) /* somethings very very wrong */
     return NOMATCH;
   if (u && !fr.Set(u)->GlobBot() && !fr.GlobCommon()) {
      strcpy(par,(char *)args[3]);
      if (!par[0]) {
	 for (lang = 0x01A; lang < 0x01D; lang++) 
	   NOTES->SERVER->SendNotice(Q_HELP, (char *)args[0], NOTES->Lang(lang, ""));
      } else if (u->PassMatch("-") || u->PassMatch(G->Split(&par))) {
	 fcn = G->Split(&par);
	 if (!strcasecmp(fcn, "INDEX"))
	   NOTES->Read(u->Handle(), nick, "+", 0);
	 else if (!strcasecmp(fcn, "READ")) 
	   NOTES->Read(u->Handle(), nick, strcasecmp(par, "ALL") ? par : "-",0);
	 else if (!strcasecmp(fcn, "ERASE")) 
	   NOTES->Delete(u->Handle(), nick, strcasecmp(par, "ALL") ? par : "-",0);
	 else if (!strcasecmp(fcn, "TO")) {
	    to = G->Split(&par);
	    if (!par[0]) 
	      lang = 0x01E;
	    else if (!(u2 = G->userlist->Find(to))) 
	      lang = 0x01F;
	    else if (fr.Set(u2)->GlobBot())
	      lang = 0x020;
	    else {
	       while ((ptr = ptr->Next()))
		 if (!strcasecmp(ptr->Nick(), to) &&
		     (ptr->TypeFlags() & DCT_GETNOTES) &&
		     (ptr->TypeFlags() & DCT_CHAT) && 
		     !((DccChat *)ptr)->Away()) {
		    ptr->Send(NOTES->Lang(0x021, "css", 7, u->Handle(), par));
		    lang = 0x022;
		 }
	       if (!lang) {
		  if (!NOTES->notefile[0]) 
		    lang = 0x007;
		  else if (!(f = fopen(NOTES->notefile, "a")) 
			   && !(f = fopen(NOTES->notefile, "w"))) {
		     lang = 0x00A;
		     G->LOG_MISC.Put(NOTES->Lang(0x009, ""));
		  } else {
		     fprintf(f, "%s %s %lu %s\n", to, u->Handle(), G->now, par);
		     fclose(f);
		     lang = 0x022;
		  }
	       }
	    }
	 } else
	   lang = 0x023;
	 G->LOG_CMDS.Put(NOTES->Lang(0x023, "sssss", nick, host, u->Handle(), 
				     fcn, par[0] ? "..." : ""));
      }
   }
   if (lang)
     NOTES->SERVER->SendNotice(Q_HELP, (char *)args[0], NOTES->Lang(lang, ""));
   return EXECUTED;
}
#endif

static void notes_hourly() {
   FUNCTION;
   int k;
#if defined(HAVE_IRC_H) && defined(HAVE_SERVER_H)
   Channel3 * chan = (Channel3 *)G->chanset;
   ChannelMember * m;
   UserRecord * u;
#endif
   DccChat * ptr;
   FILE *f, *g;
   char s[513], s2[DIRLEN+1], *to, *from, *ts, *s1;
#ifdef HAVE_BOTNET_H
   char out[512], out1[HANDLEN+1], *p;
#endif
   int tot = 0, fc = 0, lapse, l;
   UserRecord * ur;
   UserEntryFwd * fwd = 0;
#ifdef HAVE_BOTNET_H
   Version required (2,0);
   ModuleEntryBotnet * me = (ModuleEntryBotnet *)
     G->module_list->Find ("botnet", &required);
#endif
   
   if (NOTES->notefile[0] && (f = fopen(NOTES->notefile, "r"))) {
      G->SPrintF(s2, DIRLEN, "%s~new", NOTES->notefile);
      if ((g = fopen(s2, "w"))) {
	 for (fgets(s, 512, f); !feof(f); fgets(s, 512, f)) {
	    if (s[l = strlen(s) - 1] == '\n')
	      s[l] = 0;
	    G->RMSpace(s);
	    if (s[0] && (s[0] != '#') && (s[0] != ';')) {	/* not comment */
	       s1 = s;
	       to = G->Split(&s1);
	       from = G->Split(&s1);
	       ts = G->Split(&s1);
	       lapse = (G->now - (time_t) atoi(ts)) / 86400;
	       if ((lapse > NOTES->note_life) || !(ur = G->userlist->Find(to))) {
		  tot ++;
		  continue;
	       } else if (NOTES->allow_fwd && (fwd = FindFwd(ur))) {
#ifdef HAVE_BOTNET_H
		  if ((p = strchr(fwd->Value(), '@')) && me
		      && strcasecmp(p+1,G->botnetnick) && !me->tandbot->Find(p+1)) {
		     if (strchr(from, '@'))
		       G->SPrintF(out, 511, "<%s@%s >%s %s", to, G->botnetnick,
				  from, s1);
		     else
		       G->SPrintF(out, 511, "<%s@%s >%s@%s %s", to, G->botnetnick,
				  from, G->botnetnick, s1);
		     G->SPrintF(out1, HANDLEN+1, "@%s", G->botnetnick);
		     if (G->AddNote(fwd->Value(), out1, out, 0, 0) == NOTE_OK) {
			fc ++;
			continue;
		     }
		  }
#endif
	       }
	       fprintf(g, "%s %s %s %s\n", to, from, ts, s1);
	    } else
	      fprintf(g, "%s\n", s);
	 }
	 fclose(f);
	 fclose(g);
	 unlink(NOTES->notefile);
#ifdef RENAME
	 rename(s2, NOTES->notefile);
#else
	 G->MoveFile(s2, NOTES->notefile);
#endif
	 if (tot > 0)
	   G->LOG_MISC.Put( NOTES->Lang((tot == 1) ? 0x003 : 0x004, "d", tot));
	 if (fc > 0)
	   G->LOG_MISC.Put( NOTES->Lang((fc == 1) ? 0x029 : 0x02A, "d", fc));
      } else
	fclose(f);
   }
   if (NOTES->notify_users) {
#if defined(HAVE_IRC_H) && defined(HAVE_SERVER_H)
      if (NOTES->SERVER)
	while ((chan = chan->Next())) 
	  for (m = chan->M(); (m = m->Next()); ) 
	    if ((u = m->User()) && (u->Handle()[0] != '*')) {
	       k = NOTES->CountNotes(u->Handle());
	       for (ptr = (DccChat *)G->socklist; (ptr = ptr->Next()); )
		 if (ptr->User() == u)
		   break;
	       if (k && !ptr) {
		  NOTES->SERVER->SendNotice(Q_HELP, m->Key(), 
				     NOTES->Lang(0x025,"ds",k,NOTES->SERVER->botname));
		  NOTES->SERVER->SendNotice(Q_HELP, m->Key(),
				     NOTES->Lang(0x026,"s",NOTES->SERVER->botname));
	       }
	    }
#endif
      for (ptr = (DccChat *)G->socklist; (ptr = ptr->Next()); )
	if ((k = NOTES->CountNotes(ptr->Nick()))) {
	   ptr->Send(NOTES->Lang(0x027, "d", k));
	   ptr->Send(NOTES->Lang(0x015, ""));
	}
   }
}

static BindReturn away_notes ( void ** args, const char *) {
   FUNCTION;
   Socket * idx = 0;
   
   context;
   if (!strcasecmp((char *)args[0],G->botnetnick)
       && (idx = G->socklist->Find(atoi((char *)args[1])))) {
      if (args[3] && ((char *)args[3])[0])
	idx->Send(NOTES->Lang(0x038, ""));
      else
	NOTES->Read(idx->Nick(), 0, "+", idx);
   }
   return NOMATCH; /* allow futher processing */
}

static BindReturn chon_notes ( void ** args, const char *) {
   FUNCTION;
   Socket * sock = (Socket *)args[0];
   if (sock->TypeFlags() & DCT_CHAT)
     NOTES->Read(sock->Nick(), 0, "+", sock);
   return NOMATCH;
}

#if defined(HAVE_IRC_H) && defined(HAVE_SERVER_H)
static BindReturn join_notes (void ** args, const char *) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   DccChat * ptr = (DccChat *)G->socklist;
   int i;
	
   if (u && NOTES->SERVER) {
      i = NOTES->CountNotes(u->Handle());
      while ((ptr = ptr->Next()))
	if (ptr->User() == u)
	  break;
      if (i && !ptr) {
	 NOTES->SERVER->SendNotice(Q_HELP, (char *)args[0], 
			    NOTES->Lang(0x025, "ds", i, NOTES->SERVER->botname));
	 NOTES->SERVER->SendNotice(Q_HELP, (char *)args[0], 
			    NOTES->Lang(0x026, "s", NOTES->SERVER->botname));
      }
   }
   return NOMATCH;
}

static command_binds notes_join [] = {
     { "*", "", join_notes, "notes" },
};
#endif

static command_binds notes_lnkch [] = {
     { "*", "", notes_change, "notes" },
};

static command_binds notes_away [] = {
     { "*", "", away_notes, "notes" },
};

static command_binds notes_chon [] = {
     { "*", "", chon_notes, "notes" },
};

#ifdef HAVE_SERVER_H
static command_binds notes_msgs [] = {
     { "notes", "", msg_notes, 0 },
};
#endif

static command_binds notes_cmds [] = {
     { "fwd", "m", cmd_fwd, 0 },
     { "notes", "", cmd_notes, 0 },
};

#ifdef HAVE_SERVER_H
#ifdef HAVE_IRC_H
static BindReturn notes_irc_setup (void **, const char *) {
   FUNCTION;
   BindTable * bt_temp;
   
   if ((bt_temp = G->bind_list->Find("join")))
     bt_temp->AddBinds(notes_join,1,NOTES);
   return NOMATCH;
}
#endif

static BindReturn notes_server_setup (void ** args, const char *) {
   FUNCTION;
   Version required(2,0);
   NOTES->SERVER = (ModuleEntryServer *)
     (args ? args[0] : G->module_list->Find("server", &required));
   
   if (NOTES->SERVER) {
      if (*NOTES->SERVER->Vers() <= required)
	NOTES->SERVER = 0;
      else 
	NOTES->SERVER->bt_msg->AddBinds(notes_msgs,1,NOTES);
   }
   return EXECUTED;
};

static BindReturn notes_server_unload (void **, const char *) {
   FUNCTION;
   NOTES->SERVER = 0;
   return EXECUTED;
}
#endif

#if defined(HAVE_IRC_H) || defined(HAVE_SERVER_H)
static command_binds notes_load [] = {
#ifdef HAVE_SERVER_H
     { "server", "", notes_server_setup, "notes:server" },
#ifdef HAVE_IRC_H
     { "irc", "", notes_irc_setup, "notes:irc" }
#define THING1 2
#else
#define THING1 1 
#endif
#else    
#define THING1 0
#endif
};
#endif

#ifdef HAVE_SERVER_H
static command_binds notes_unload[] = {
   { "server", "", notes_server_unload, "notes:server" },
};
#endif

static command_binds notes_strnote[] = {
   { "*", "", storenote , "notes:strnote" },
};

extern "C" {
   char * notes_start (ModuleEntryEggdrop *, ModuleEntry *);
};

char * notes_start (ModuleEntryEggdrop * global_funcs, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   G = global_funcs;
   context;
   NEW ( ModuleEntryNotes ( me, &r ) );
   if (!r) {
      G->hook_hourly->Add(notes_hourly,NOTES);
      G->bt_dcc->AddBinds(notes_cmds,2,NOTES);
      G->bt_chon->AddBinds(notes_chon,1,NOTES);
      G->bt_away->AddBinds(notes_away,1,NOTES);
      G->bt_lnkch->AddBinds(notes_lnkch,1,NOTES);
      G->bt_strnote->AddBinds(notes_strnote,1,NOTES);
#if defined(HAVE_SERVER_H) || defined(HAVE_IRC_H)
      G->bt_load->AddBinds(notes_load,THING1,NOTES);
#endif
#ifdef HAVE_SERVER_H
#ifdef HAVE_IRC_H
      notes_irc_setup(0,0);
#endif
      notes_server_setup(0,0);
      G->bt_unld->AddBinds(notes_unload,1,NOTES);
#endif
   }
   context;
   return r;
}
