/*
 * Relay.C - outgoing connection for a relay
 */
/*
 * This file is part of the eggdrop source code
 * copyright (c) 1997 Robey Pointer
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Relaying.h"
#include "Relay.h"
#include "DccChat.h"
#define CLASS "Relay"

Relay::Relay ( DccChat * old, const char * host, int port )
: Socket ( CONNECTING | STRONG, RLY->RELAY, 0, &G->connect_timeout, 0,
	  host, port) {
   FUNCTION;
   Relaying * rly;
   
   old->CheckChof(RLY->Lang(0x01B, "s", host));
   rly = NEW ( Relaying ( old , this ) );
   relay = rly->FD();
   baseport = port;
   SetUser(rly->User());
   SetNick(rly->Nick());
}

Relay::~Relay ( void ) {
   FUNCTION;
}

int Relay::ExpMem ( void ) {
   FUNCTION;
   return sizeof(Relay);
}

void Relay::Display2 (Socket * where) {
   FUNCTION;
   where->Send (RLY->Lang(timeout_val ? 0x003 : 0x001, "d", relay));
}

void Relay::Eof ( void ) {
   FUNCTION;
   Relaying * uidx = (Relaying *)G->socklist->Find(relay, RLY->RELAYING);
   DccChat * dc;
   
   context;
   if (!uidx) {
      G->LOG_MISC.Put(RLY->Lang(0x005, "dd", key, relay));
      Close();
   } else if (timeout_val) {
      remoteport++;
      if (remoteport >= (baseport + G->port_range)) { 
	 uidx->Send(RLY->Lang(0x006, "s", host));
	 dc = NEW ( DccChat ( uidx ) );
	 dc->SetStatus(uidx->oldstatus);
	 dc->CheckChon();
	 Close();	 
      } else {
	 TimeNow();
	 if (Scan(baseport + G->port_range))
	   Eof();
	 else 
	   uidx->Redo(this);
      }
   } else {
      /* in case echo was off, turn it back on: */
      if (uidx->HasStatus(STAT_TELNET))
	uidx->Send("\377\374\001\r\n");
      G->LOG_MISC.Put(RLY->Lang(0x01E, "ss", uidx->Nick(), nick));
      uidx->Send("\n\n");
      uidx->Send(RLY->Lang(0x01F, ""));
      uidx->Send(RLY->Lang(0x01D, "s", G->botnetnick));
      dc = NEW ( DccChat ( uidx ) );
      uidx->SetValues(dc);
      dc->CheckChon();
      Close();	 
   } 
} 

void Relay::Activity ( char * buf, int i) {
   FUNCTION;
   Relaying * tidx = (Relaying *)G->socklist->Find(relay, RLY->RELAYING);
   unsigned char * p = (unsigned char *)buf, *q = p;
   
   context;
   if (!tidx) {
      G->LOG_MISC.Put(RLY->Lang(0x005, "dd", key, relay));
      Close();
   } else {
      if (timeout_val) {
	 timeout_val = 0;
	 tidx->status = 1;
	 for (i = 0x016; i < 0x01A; i++)
	   tidx->Send(RLY->Lang(i, "sd", host, baseport));
	 G->LOG_MISC.Put(RLY->Lang(0x01A, "ss", tidx->Nick(), nick));
      }
      if (!tidx->HasStatus(STAT_TELNET)) {
	 while (*p) {
	    if (*p == 255) {
	       if (!*(p + 1))
		 p += 1;
	       else if (((*(p + 1) == 251) || (*(p + 1) == 252))
			&& (*(p + 2) == 1)) 
		 /* 252/1 & 253/1 = echo on/off, dont strip */
		 *q++ = *p++;
	       else if ((*(p + 1) >= 251) || (*(p + 1) <= 254)) {
		  if (!*(p + 2))
		    p += 2;	/* bogus */
		  else
		    p += 3;
	       } else
		 p += 2;
	    } else
	      *q++ = *p++;
	 }
	 *q = 0;
      }
      tidx->Printf("%s\n", *buf ? buf : " ");
   }
}
