/*
 * Relaying.C - a dcc chatter relaying to a remote host/port
 */
/*
 * This file is part of the eggdrop source code
 * copyright (c) 1997 Robey Pointer
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Relaying.h"
#include "Relay.h"
#include "DccChat.h"
#define CLASS "Relaying"

Relaying::Relaying ( DccChat * old, Relay * conn ) 
: DccChatPass ( *old ) {
   FUNCTION;
   type = RLY->RELAYING;
   oldstatus = status;
   status = 0;
   relay = conn->FD();
   timeout_val = 0;
   oldchan = old->Channel();
   strncpyz(old_conchan, old->ConsoleChan(),80);
   old_modes = old->ConsoleFlags();
}

Relaying::~Relaying ( void ) {
   FUNCTION;
}

int Relaying::ExpMem ( void ) {
   FUNCTION;
   return sizeof(Relaying) + (away ? strlen(away) + 1 : 0);
}

void Relaying::Display2 ( Socket * where ) {
   FUNCTION;
   where->Send(RLY->Lang(status ? 0x004 : 0x002 , "d", relay));
}

void Relaying::Eof (void) {
   FUNCTION;
   Relay * tidx = (Relay *)G->socklist->Find(relay, RLY->RELAY);

   context;
   if (!tidx) 
     G->LOG_MISC.Put(RLY->Lang(0x005, "dd", key, relay));
   else {
      G->LOG_MISC.Put(RLY->Lang(0x013, "ssd", nick, host, remoteport));
      G->LOG_MISC.Put(RLY->Lang(0x014, "s", tidx->Nick()));
      if (tidx->FD() || G->backgrd) 
	tidx->Close();
      else 
	G->Fatal(RLY->Lang(0x015, ""));
   }
   Close();
}

void Relaying::Activity (char * buf, int) {
   FUNCTION;
   Relay * tidx =(Relay *) G->socklist->Find(relay, RLY->RELAY);
   DccChat * dc;
   
   context;
   if (!tidx) 
     G->LOG_MISC.Put(RLY->Lang(0x005, "dd", key, relay));
   else if (!strcasecmp(buf, "*bye*")) {
      int lang = status ? 0x01C : 0x010;
      
      if (HasStatus(STAT_TELNET))
	Send("\377\374\001\r\n");
      Send(RLY->Lang(lang, "s", tidx->Nick()));
      Send(RLY->Lang(lang + 1, "s", G->botnetnick));
      G->LOG_MISC.Put(RLY->Lang(0x012, "ss", nick, tidx->Nick()));
      tidx->Close();
      dc = NEW ( DccChat ( this ) );
      SetValues(dc);
      dc->CheckChon ();
   } else if (status)  
	tidx->Printf("%s\n", buf);
}

void Relaying::SetValues ( DccChat * newone ) {
   FUNCTION;
   newone->SetChannel(oldchan);
   newone->SetConsoleChan(old_conchan);
   newone->ConsoleFlags(old_modes);
}
