/*
 * ScriptSocket.C
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ScriptSocket.h"
#include "ModuleEntryScriptsock.h"
#include "DccChat.h"
#include "IdentWait.h"
#include "ScriptContext.h"
#include "FlagRecord.h"
#define CLASS "ScriptSocket"

ScriptSocket::ScriptSocket ( struct in_addr ad, int port, const char * c,
			    ScriptContext * s)
: DccChatPass (ad, port) {
   FUNCTION;
   sc = s;
   type = SS->SCRIPT_SOCKET;
   SetNick("*");
   strncpy(command, c, 120);
   command[120]  = 0;
   backtodcc = 0;
}

ScriptSocket::ScriptSocket ( DccChat & dc, const char * c, ScriptContext * s )
: DccChatPass ( dc ) {
   FUNCTION;
   sc = s;
   type = SS->SCRIPT_SOCKET;
   strncpy(command, c, 120);
   command[120]  = 0;
   backtodcc = 1;
}

ScriptSocket::ScriptSocket (IdentWait & o, const char * c, ScriptContext * s )
: DccChatPass ( o, SS->SCRIPT_SOCKET ) {
   FUNCTION;
   sc = s;
   strncpy(command, c, 120);
   command[120]  = 0;
   backtodcc = 0;
}

ScriptSocket::~ScriptSocket ( void ) {
   FUNCTION;
}

int ScriptSocket::CallFunc (char * mode, const char * data) {
   FUNCTION;
   if (sc->Evaluate(command, 3, G->ToBase10(key), mode, data) == SCRIPT_ERROR) {
      G->LOG_MISC.Put(SS->Lang(0x001, "sss", sc->Key(), command, sc->AsString()));
      return -1;
   }
   return sc->AsInteger();
}

void ScriptSocket::DropBack ( void ) {
   FUNCTION;
   if (backtodcc) {
      DccChat * newone = NEW ( DccChat ( this ) );
      FlagRecord atr ( user );
      
      newone->ConsoleFlags(atr.GlobMaster() ? G->conmask.Value() : 0);
      newone->CheckChon();
   } else
     Close();
}

void ScriptSocket::Activity (char * buf,int len) {
   FUNCTION;
   StripTelnet(buf, &len);
   TimeNow();
   if (CallFunc("D", buf)) 
     DropBack();
}

void ScriptSocket::Eof (void) {
   FUNCTION;
   /* tell the script they're gone: */
   CallFunc("E", "");
   DropBack();
}

void ScriptSocket::Display2 (Socket * where) {
   FUNCTION;
   where->Send(SS->Lang(0x003, "ss", command, sc->Key()));
}

int ScriptSocket::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ScriptSocket);
}
