/*
 * Scriptsock.c 
 * 
 * the script socket module, allows script users to use sockets
 */
/*
 * extracted from the eggdrop core (C) 1997 Robey Pointer
 * by beldin (and enhanced a bit too)
 */

#include "Socket.h"
#include "BindTable.h"
#include "DccChat.h"
#include "IdentWait.h"
#include "ModuleEntryScriptsock.h"
#include "HostnameCache.h"
#include "TelnetPort.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "ScriptSocket.h"
#include "Command.h"
#include <errno.h>
#define CLASS ""

ModuleEntryScriptsock * SS;
static ModuleEntryEggdrop * G;

static int script_connect (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Socket * sock;
   int ret;
   struct in_addr * in;
   
   context;
   BADARGS(4, 4, SS->Lang(0x004, ""));
   if (!(in = G->host_cache->Find(argv[1]->AsString())))
     sc->SetString(SS->Lang( 0x005, ""));
   else if (!(sock = NEW (ScriptSocket( *in, argv[2]->AsInteger(),
				       argv[3]->AsString(), sc))))
     sc->SetString(SS->Lang( 0x006, ""));
   else if ((ret = sock->Scan(sock->RemotePort() + 1)))
     sc->SetString(strerror(errno));
   else {
      sc->SetInteger(sock->FD());
      return SCRIPT_OK;
   }
   return SCRIPT_ERROR;
}

static int script_control (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   DccChat * idx;

   context;
   BADARGS(3, 3, SS->Lang( 0x007, ""));
   if ((idx = (DccChat *)G->socklist->Find(argv[1]->AsInteger(), G->DCC_CHAT))) {
      idx->CheckChof(SS->Lang( 0x008, ""));
      NEW ( ScriptSocket ( *idx, argv[2]->AsString(), sc));
      return SCRIPT_OK;
   }
   sc->SetString(SS->Lang( 0x009, ""));
   return SCRIPT_ERROR;
}

static BindReturn scriptsock_listen (void ** args, const char * mask) {
   FUNCTION;
   IdentWait * idx = (IdentWait *)args[0];
   TelnetPort * telnet = (TelnetPort *)args[1];
   ScriptSocket * newone;
   
   if (strcmp(mask,"(script)"))
     G->LOG_MISC.Put(SS->Lang( 0x00A, "dds", idx->FD(), telnet->FD(), mask));
   else {
      newone = NEW ( ScriptSocket ( *idx, telnet->Host(), telnet->Script()));
      if (newone->CallFunc("C", newone->Host()))
	newone->Close();
   }
   return EXECUTED;
}

static int script_scriptport (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   TelnetPort * tp = (TelnetPort *)G->socklist;
   context;
   BADARGS(4, 4, SS->Lang(0x00B, ""));
   if ((tp = tp->Find(argv[2]->AsInteger()))) {
      sc->SetString(G->Lang(0x1A9, "ss", argv[1]->AsString(),argv[2]->AsString()));
      return SCRIPT_ERROR;
   }
   NEW ( TelnetPort ( sc, argv[1]->AsString(), argv[2]->AsInteger(), 
		     argv[3]->AsString(), "(script)"));
   return SCRIPT_OK;
}

static command_binds my_list[1] =  {
   { "(script)", "", scriptsock_listen, NULL }
};

extern "C" {
   char * scriptsock_start ( ModuleEntryEggdrop *, ModuleEntry * );
};

char * scriptsock_start ( ModuleEntryEggdrop * globals, ModuleEntry * me ) {
   FUNCTION;
   char * r = 0;
   G = globals;
   context;
   NEW ( ModuleEntryScriptsock ( me, &r ) );
   if (!r) {
      NEW ( Command ("control", script_control, SS ) );
      NEW ( Command ("connect", script_connect, SS ) );
      NEW ( Command ("scriptport", script_scriptport, SS ) );
      G->bt_listen->AddBinds(my_list , 1, SS );
   }
   return r;
}
