/*
 * seen.c   - Implement the seen.tcl script functionality via module.
 *
 *            by ButchBub - Scott G. Taylor (staylor@mrynet.com) 
 *
 *      REQUIRED: Eggdrop Module version 1.5.0
 *
 *      0.1     1997-07-29      Initial. [BB]
 *      1.0     1997-07-31      Release. [BB]
 *      1.1     1997-08-05      Add nick->handle lookup for NICK's. [BB]
 *      1.2     1997-08-20      Minor fixes. [BB]
 *      1.2a    1997-08-24	Minor fixes. [BB]
 *      1.5     1998-01-12      Bringing it in line with 1.5 [beldin]
 */

/*
 *  Currently, PUB, DCC and MSG commands are supported.  No party-line
 *      filtering is performed.
 *
 *  For boyfriend/girlfriend support, this module relies on the XTRA
 *      fields in the userfile to use BF and GF, respectively, for
 *      these fields.  
 *
 *  userinfo1.0.tcl nicely compliments this script by providing 
 *      the necessary commands to facilitate modification of these
 *      fields via DCC and IRC MSG commands.
 *
 *  A basic definition of the parsing syntax follows:
 *
 *      trigger :: seen [ <key> [ [ and | or ] <key> [...]]]
 *
 *        <key> :: <keyword> [ <keyarg> ]
 *
 *    <keyword> :: god | jesus | shit | me | yourself | my | <nick>'s |
 *                 your
 *       <nick> :: (any current on-channel IRC nick or userlist nick or handle)
 *
 *     <keyarg> :: (see below)
 *
 *              KEYWORD KEYARG
 *
 *              my      boyfriend
 *                      bf
 *                      girlfriend
 *                      gf
 *              your    owner
 *                      admin
 *                      (other)
 *              NICK's  boyfriend   
 *                      bf
 *                      girlfriend
 *                      gf
 *
 */

#include "DccChat.h"
#include "UserRecord.h"
#include "UserEntryXtra.h"
#include "UserEntryLaston.h"
#include "BindTable.h"
#include "HelpReference.h"
#include "ModuleEntrySeen.h"
#include "ModuleEntry.h"
#include "ScriptContext.h"
#include "ChannelMember.h"
#include "ChannelRecord.h"
#include "ChannelSettingsSeen.h"
#ifdef HAVE_CHANNELS_H
#include "channels.h"
static Version required_channels ( 2, 0 );
#endif
#ifdef HAVE_SERVER_H
#include "server.h"
#include "ModuleEntryServer.h"
static Version required_server ( 2, 0 );
#endif
#ifdef HAVE_IRC_H
#include "irc.h"
#include "ModuleEntryIrc.h"
#include "Channel3.h"
static Version required_irc ( 2, 0 );
#else
class Channel3;
#endif
#define CLASS ""

ModuleEntrySeen * SEEN;
static ModuleEntryEggdrop * G;

static int match_trigger( char *word) {
   FUNCTION;
   int lang = 0x0FE;
   char * data = G->botnetnick;
   
   while (strncasecmp(data, "MSG:", 4) && strcasecmp(word, data)) {
      lang += 2;
      data = SEEN->Lang(lang, "");
   }
   return strncasecmp(data, "MSG:", 4) ? lang + 1 : 0;
}

static void wordshift (char * first, char *rest) {
   FUNCTION;
   char * p, * q = rest;
   
   do {
      p = G->Split(&q);
      strcpy(first,p);
      strcpy(rest,q);
   } while (!strcasecmp(first, "and") || !strcasecmp(first, "or"));
}

static char fixit[512];
static char *fixnick (char *nick) {
   FUNCTION;
   int l = G->StrCpy(fixit, nick);
   fixit[l++] = '\'';
   switch (nick[l - 1]) {
     case 's':
     case 'S':
     case 'x':
     case 'X':
     case 'z':
     case 'Z':
      break;
     default:
      fixit[l++] = 's';
      break;
   }
   fixit[l] = 0;
   return fixit;
}

static void output (Socket * idx, const char * desto, char * data) {
   FUNCTION;
   if (idx)
     idx->Send(data);
#ifdef HAVE_SERVER_H
   else if (SEEN->SERVER) 
     SEEN->SERVER->SendMsg(Q_HELP, desto, data);
#endif
}

static void do_seen(Socket * idx, const char *desto, const char *nick,
		    UserRecord * u, Channel3 * channel,
		    char * text) { 
   FUNCTION;
   char word1[512], word2[512], whotarget[512], object[512];
   char * oix, whoredirect[512];
   const char * lastonplace = 0;
   UserEntryLaston * li;
   ChannelRecord * cr = NULL;
#ifdef HAVE_IRC_H
   Channel3 * chan;
   ChannelMember *m;
#endif
   UserRecord * urec;
   int i, onchan = 0;
   time_t laston = 0, work;
   DccChat * ptr = (DccChat *)G->socklist;
   
   whotarget[0] = 0;
   whoredirect[0] = 0;
   object[0] = 0;

/* Was ANYONE specified */

   if (!text[0]) {
      output(idx, desto, SEEN->Lang(0x001, "s", nick));
      return;
   }
   
   wordshift(word1, text);
   oix = strchr(word1, '\'');
   /* Have we got a NICK's target? */
   if (oix != word1) {       /* Skip anything starting with ' */
      context;
      if( oix && *oix &&
	 ((oix[1] && ( oix[1] == 's' || oix[1]  == 'S') && !oix[2]) ||
	  (!oix[1] && (oix[-1] == 's' || oix[-1] == 'z' || oix[-1] == 'x'
		       || oix[-1] == 'S' || oix[-1] == 'Z'  || oix[-1] == 'X')))) {
	 strncpy(object, word1, oix - word1);
	 object[oix - word1] = 0;
	       context;
	 wordshift(word1, text);
	 if(!word1[0]) {
	    output(idx, desto, SEEN->Lang(0x002, "ss", object, nick));
	    return;
	 }
	 onchan = 0;
	 urec = G->userlist->Find(object);
#ifdef HAVE_IRC_H
	 if (SEEN->IRC && !urec) {
	    for (chan = (Channel3 *)G->chanset; (chan = chan->Next());)
	      if ((m = chan->M()->Find(object))) {
		 onchan = 1;
		 urec = m->MemberUser();
		 if (!urec || !strcasecmp(object, urec->Handle())) 
		   break;
		 strcpy(whoredirect, SEEN->Lang(0x003, "ss", object, 
					     urec->Handle()));
		 strcpy(whotarget, urec->Handle());
		 break;
	      }
	    if (!onchan) {
	       output(idx, desto, SEEN->Lang(0x004, "ss", object, nick));
	       return;
	    }
	 }
#endif
	 context;
	 if(!strcasecmp(word1, "bf") || !strcasecmp(word1, "boyfriend")) {
	    strcpy(whotarget, safe_str(GetXtra(urec, "BF")));
	    if (whotarget[0]) {
	       strcpy(whoredirect, SEEN->Lang(0x005, "ss", fixnick(object),
					   whotarget));
	       goto TARGETCONT;
	    }
	    output(idx, desto, SEEN->Lang(0x006, "ss", fixnick(object), nick));
	 } else if(!strcasecmp(word1, "gf") || !strcasecmp(word1, "girlfriend")) {
	    strcpy(whotarget, safe_str(GetXtra(urec, "GF")));
	    if (whotarget[0]) {
	       strcpy(whoredirect, SEEN->Lang(0x007, "ss", fixnick(object),
					   whotarget));
	       goto TARGETCONT;
	    }
	    output(idx, desto, SEEN->Lang(0x008, "ss", fixnick(object), nick));
	 } else
	   output(idx, desto, SEEN->Lang(0x009, "sss", fixnick(object), word1,
				      nick));
	 context;
	 return;
      }
      
      /* Keyword "my" */
      if (!strcasecmp(word1, "my")) {
	 
	 wordshift(word1, text);
	 if (!word1[0]) {
	    output(idx, desto, SEEN->Lang(0x00A, "s", nick));
	    return;
	 } 
	 
	 /* Do I even KNOW the requestor? */
	 
	 if (!u) {
	    output(idx, desto, SEEN->Lang(0x00B, "ss", nick, word1));
	    return;
	 } 
	 
	 /* "my boyfriend" */
	 
	 if (!strcasecmp(word1, "boyfriend") || !strcasecmp(word1, "bf")) {
	    strcpy(whotarget, GetXtra(u, "BF"));
	    if (whotarget[0]) 
	      strcpy(whoredirect, SEEN->Lang(0x00C, "ss", nick, whotarget));
	    else {
	       output(idx, desto, SEEN->Lang(0x00D, "s", nick));
	       return;
	    }
	 } 

	 /* "my girlfriend" */
	 else if (!strcasecmp(word1, "girlfriend") || !strcasecmp(word1, "gf")) {
	    strcpy(whotarget, GetXtra(u, "GF"));
	    if (whotarget[0]) 
	      strcpy(whoredirect, SEEN->Lang(0x00E, "ss", nick, whotarget));
	    else {
	       output(idx, desto, SEEN->Lang(0x00F, "s", nick));
	       return;
	    }
	 } else {
	    output(idx, desto, SEEN->Lang(0x010, "ss", word1, nick));
	    return;
	 }
      } 
      
      /* "your" keyword */
      
      else if (!strcasecmp(word1, "your")) {
	 wordshift(word1, text);
	 
	 /* "your admin" */
	 
	 if (!strcasecmp(word1, "owner") || !strcasecmp(word1, "admin")
	     || !strcasecmp(word1, "llama")) {
	    if(G->admin[0]) {
	       strcpy(word2, G->admin);
	       wordshift(whotarget, word2);
	       if (!strcasecmp(whotarget, u->Handle())) {
		  output(idx, desto, SEEN->Lang(strcasecmp(u->Handle(), nick)
					     ? 0x011 : 0x012, "ss", whotarget,
					     nick));
		  return;
	       }
	       strcpy(whoredirect, SEEN->Lang(0x013, "s", whotarget));
	    } else {          /* owner variable munged or not set */
	       output(idx, desto, SEEN->Lang(0x014, ""));
	       return;
	    }
	 } else {             /* no "your" target specified */
	    output(idx, desto, SEEN->Lang(0x015, "s", nick));
	    return;
	 }
      } 
      
      /* Check for keyword match in the internal table */

      else if ((i = match_trigger(word1))) {
	 output(idx, desto, SEEN->Lang(i, "s", nick));
	 return;
      } 
      
      /* Otherwise, make the target to the first word and continue */
      
      else 
	strcpy(whotarget, word1);
      
      TARGETCONT:
      
      context;
      /* Looking for ones own nick? */
      
      if(!strcasecmp(nick, whotarget)) {
	 output(idx, desto, SEEN->Lang(0x016, "ss", whoredirect, nick));
	 return;
      }
      
      /* Check if nick is on a channel */
#ifdef HAVE_IRC_H     
      if (SEEN->IRC) {
	 for (chan = (Channel3 *)G->chanset; (chan = chan->Next()); ) 
	   if ((m = chan->M()->Find(whotarget))) {
	      onchan = 1;
	      urec = m->MemberUser();
	      if (!urec || !strcasecmp(whotarget, urec->Handle()))
		break;
	      strcpy(whoredirect, SEEN->Lang(0x003, "ss", 
					  whotarget, urec->Handle()));
	      strcpy(whotarget, urec->Handle());
	      break;
	   }
      /* Check if nick is on a channel by xref'ing to handle */
      
	 if (!onchan) 
	   for (chan = (Channel3 *)G->chanset; (chan = chan->Next());) 
	     for (m = chan->M(); (m = m->Next());)
	       if ((urec = m->MemberUser())
		   && !strcasecmp(urec->Handle(), whotarget)) {
		  onchan = 1;
		  strcpy(whoredirect, SEEN->Lang(0x003, "ss", whotarget,
					      m->Key()));
		  strcpy(whotarget, m->Key());
		  break;
	       }
      
	 /* Check if the target was(&split) or is on the channel */
	 context;
	 
	 if (channel && (m = channel->M()->Find(whotarget))) {
	    output(idx, desto, SEEN->Lang(m->IsSplit() ? 0x017 : 0x018,
				       "ss", whoredirect, whotarget));
	    return;
	 } 
	 
	 /* Target not on this channel.   Check other channels */
	 
	 for (chan = (Channel3 *)G->chanset; (chan = chan->Next());)
	   if ((m = chan->M()->Find(whotarget))) {
	      output(idx, desto, SEEN->Lang(m->IsSplit() ? 0x019 : 0x01A,
					 "ss", whoredirect, whotarget,
					 chan->Key()));
	      return;
	   }
      }
#endif      
      
      /*
       * Target isn't on any of my channels.  See if target matches a handle
       * in my userlist 
       */
      
      if (!(urec = G->userlist->Find(whotarget))) {
	 
	 /* No match, then bail out */
	 output(idx, desto, SEEN->Lang(0x01B, "ss", whoredirect, whotarget));
	 return;
      }
      
      /* We had a userlist match to a handle */
      
      /* Is the target currently DCC CHAT to me on the botnet? */
      
      while ((ptr = ptr->Next())) {
	 if (!strcasecmp(whotarget, ptr->Nick())) {
	    if (channel && !strcasecmp(channel->Key(), ptr->ConsoleChan()) &&
		(ptr->ConsoleFlags().Value() & LogModes::PUBLIC))
	      output(idx, desto, SEEN->Lang(0x01C, "ss", whoredirect,
					 whotarget));
	    else
	      output(idx, desto, SEEN->Lang(0x01D, "ss", whoredirect,
					 whotarget));
	    return;
	 }
      }
      
      /* Target known, but nowhere to be seen.   Give last IRC and botnet time */
      
      wordshift(word1, text);
      if (!strcasecmp(word1,"anywhere"))
	cr = NULL;
      else if (channel) {
	for (cr = urec->ChanRecs(); cr; cr = cr->Next())
	  if (!strcasecmp(cr->Key(),channel->Key()) && cr->Laston()) {
	     laston = cr->Laston();
	     lastonplace = channel->Key();
	     break;
	  }
      } else 
	cr = NULL;
      if (!cr) {
	 li = FindLaston(urec);
	 if (!li || !li->Where() || !li->Where()[0]) {
	    output(idx, desto, SEEN->Lang(0x01E, "ss", whoredirect, whotarget));
	    return;
	 }	
	 lastonplace = li->Where();
	 laston = li->When();
      }
      word1[0] = 0;
      word2[0] = 0;
      work = G->now - laston;
      i = 0;
      if (work >= 86400) {
	 i += G->StrCpy(word2, SEEN->Lang(((work/86400) == 1) ? 0x01F : 0x020,
				       "d", work / 86400));
	 work = work % 86400;
      }
      if (work >= 3600) {
	 i += G->StrCpy(word2 + i, SEEN->Lang(((work/3600) == 1) ? 0x021 : 0x022,
					   "d", work /3600));
	 work = work % 3600;
      }
      if (work >= 60) 
	i += G->StrCpy(word2 + i, SEEN->Lang(((work/60) == 1) ? 0x023 : 0x024,
					  "d", work /60));
      if (!i && (work < 60))
	strcpy(word2, SEEN->Lang(0x026, ""));
      else 
        G->StrCpy(word2 + i, SEEN->Lang(0x025, ""));
      if ((lastonplace[0] == '#') || (lastonplace[0] == '&'))
	strcpy(word1, SEEN->Lang(0x027, "s", lastonplace));
      else if ((lastonplace[0] == '@') && SEEN->display_bots)
	strcpy(word1, SEEN->Lang(0x028, "s", lastonplace + 1));
      else if ((lastonplace[0] != 0) && (lastonplace[0] != '@'))
	strcpy(word1, SEEN->Lang(0x029, "s", lastonplace));
      else 
	strcpy(word1, SEEN->Lang(0x02A, ""));
      output(idx, desto, SEEN->Lang(0x02B, "ssss", whoredirect, whotarget,
				 word1, word2));
   }
}

/*
 * PUB `seen' trigger.
 */

#ifdef HAVE_IRC_H
static BindReturn pub_seen ( void ** args, const char *) {
   FUNCTION;
   ChannelSettingsSeen * chansets;
   
   if (SEEN->CHANNELS && (chansets = (ChannelSettingsSeen *)
			  ((Channel3 *)args[3])->FindSettings(SEEN->CHANSET_SEEN))
       && !chansets->Flag())
     return NOMATCH;
   do_seen(NULL, ((Channel3 *)args[3])->Key(), (char *)args[0], 
	   (UserRecord *)args[2], (Channel3 *)args[3],
	   (char *)args[4]);
   return EXECUTED;
}
#endif

#ifdef HAVE_SERVER_H
static BindReturn msg_seen ( void ** args, const char *) {
   FUNCTION;
   if (!args[2]) 
     G->LOG_MISC.Put(SEEN->Lang(0x02D, "sss", args[0], args[1], args[3]));
   else {
      G->LOG_MISC.Put(SEEN->Lang(0x02E, "ssss", args[0], args[1], 
			       ((UserRecord *)args[2])->Handle(), args[3]));
      do_seen(NULL, (char *)args[0], (char *)args[0], (UserRecord *)args[2],
	      NULL, (char *)args[3]);
   }
   return EXECUTED;
}
#endif

static BindReturn dcc_seen ( void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   context;
   G->LOG_CMDS.Put(SEEN->Lang(0x02F, "ss", idx->Nick(), args[1]));
   do_seen(idx, "", idx->Nick(), idx->User(), NULL, (char *)args[1]);
   return EXECUTED;
}

/*
 * PUB channel builtin commands.
 */

#ifdef HAVE_IRC_H
static command_binds seen_pub[] = {
   {"seen",   "", pub_seen, 0},
};

static BindReturn irc_seen_setup(void ** args, const char *) {
   FUNCTION;
   SEEN->IRC = (ModuleEntryIrc *)(args ? args[0]
     : G->module_list->Find("irc", &required_irc));
   if (!SEEN->IRC || !(*SEEN->IRC->Vers() <= required_irc) ||
       !(SEEN->CHANNELS = (ModuleEntryChannels *)
	 SEEN->Depend("channels", &required_channels)))
     SEEN->IRC = NULL;
   else {
      SEEN->IRC->bt_pub->AddBinds(seen_pub, 1,SEEN);
      SEEN->CHANSET_SEEN = NEW ( ChannelSettingsSeen ( SEEN ) );
   }
   return EXECUTED;
}

static BindReturn irc_unload (void **, const char *) {
   FUNCTION;
   SEEN->IRC = 0;
   if (SEEN->CHANSET_SEEN) {
      DELETE (SEEN->CHANSET_SEEN);
      SEEN->CHANSET_SEEN = 0;
   }
   SEEN->CHANNELS = 0;
   return EXECUTED;
}
#define PART1 1
#else
#define PART1 0
#endif

static command_binds seen_dcc[] = {
   {"seen",   "", dcc_seen, 0},
};

#ifdef HAVE_SERVER_H
static command_binds seen_msg[] = {
   {"seen",  "", msg_seen, 0},
};

static BindReturn server_seen_setup(void ** args, const char *) {
   FUNCTION;
   SEEN->SERVER = (ModuleEntryServer *)(args ? args[0]
     : G->module_list->Find("server", &required_server));
   if (!SEEN->SERVER || !(*SEEN->SERVER->Vers() <= required_server))
     SEEN->SERVER = NULL;
   else 
     SEEN->SERVER->bt_msg->AddBinds(seen_msg,1,SEEN);
   return EXECUTED;
}

static BindReturn server_unload (void **, const char *) {
   FUNCTION;
   SEEN->SERVER = NULL;
   return EXECUTED;
}
#define PART2 1
#else
#define PART2 0
#endif

#if defined(HAVE_SERVER_H) || defined(HAVE_IRC_H)
static command_binds seen_load[] = {
#ifdef HAVE_SERVER_H
   {"server", "", server_seen_setup, "seen:server"},
#endif
#ifdef HAVE_IRC_H
   {"irc", "", irc_seen_setup, "seen:irc"},
#endif
};

static command_binds seen_unload[] = {
#ifdef HAVE_SERVER_H
   {"server", "", server_unload, "seen:server"},
#endif
#ifdef HAVE_IRC_H
   {"irc", "", irc_unload, "seen:irc"},
#endif
};
#endif

extern "C" {
   char * seen_start (ModuleEntryEggdrop *, ModuleEntry *);
};

char *seen_start (ModuleEntryEggdrop * F, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;

   G = F;
   context;
   SEEN = NEW( ModuleEntrySeen ( me, &r ) );
   if (!r) {
#if defined(HAVE_SERVER_H) || defined(HAVE_IRC_H)
      G->bt_load->AddBinds(seen_load,PART1+PART2,SEEN);
      G->bt_unld->AddBinds(seen_unload,PART1+PART2,SEEN);
#endif
      G->bt_dcc->AddBinds(seen_dcc,1,SEEN);
#ifdef HAVE_SERVER_H
      server_seen_setup(0,0);
#endif
#ifdef HAVE_IRC_H
      irc_seen_setup(0,0);
#endif
   }
   return r;
}
