/*
 * ModuleEntryEggdrop.C
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "server.h"
#include "ModuleEntryServer.h"
#include "IrcNick.h"
#include "BotName.h"
#include "Servers.h"
#include "ServerVar.h"
#include "BindTableRaw.h"
#include "BindTableWall.h"
#include "BindTableMsg.h"
#include "BindTableCtcpr.h"
#include "BindTableFlud.h"
#include "ServerHeads.h"
#include "Server.h"
#include "StringVariable.h"
#include "IntegerVariable.h"
#include "CoupletVariable.h"
#include "ServerList.h"
#include "HelpReference.h"
#include <netdb.h>
#include <ctype.h>
#define CLASS "ModuleEntryServer"

/* version info */
static Version server_ver, required_eggdrop;
static struct patches server_patches[] = {
#include "patch.h"
};

ModuleEntryServer::ModuleEntryServer ( ModuleEntry * o, ModuleEntryEggdrop *,
				      char ** r )
: ModuleEntry ( "server", server_ver.Set(2,0), server_patches, o ) {
   FUNCTION;
   int i;
   
   if (!Depend("eggdrop", required_eggdrop.Set(2,0))) 
     *r = "You eggdrop version is incompatible";
   else {
      SVR = this;
      trying_server = 0;
      botname[0] = altnick[0] = newbotname[0] = botuserhost[0] =0;
      strict_host = keepnick = check_stoned = serverror_quit = 1;
      check_mode_r = quiet_reject = answer_ctcp = use_ison = 1;
      server_lag = waiting_for_awake = never_give_up = 0;
      newserverport = strict_servernames = lowercase_ctcp = 0;
      trigger_on_ignore = use_silence = burst = min_servs = 0;
      last_time = use_ts4 = 0;
      server_cycle_wait = server_timeout = 60;
      oldnick[0] = bothost[0] = newserver[0] = 0;
      maxqmsg = 300;
      flood_thr = 5;
      flood_ctcp_thr = 6;
      cookie_expire = 15;
      flood_time = flood_ctcp_time = 60;
      serv = -1;
      context;
      strncpyz(botrealname, Lang(0x07F, ""), 60);
      context;
      bt_wall = NEW ( BindTableWall ( this ) );
      bt_raw = NEW ( BindTableRaw ( "raw", this ) );
      bt_notc = NEW ( BindTableMsg ("notc", this ) );
      bt_msgm = NEW ( BindTableMsg ("msgm", this ) );
      bt_msg = NEW ( BindTableMsg( this ) );
      bt_flud = NEW ( BindTableFlud ( this ) );
      bt_ctcr = NEW ( BindTableCtcpr ( "ctcr", this ) );
      bt_ctcp = NEW ( BindTableCtcpr ( "ctcp", this ) );
      bt_initsvr = NEW ( BindTableRaw ( "initsvr", this ) );
      context;
      memtype_server = NEW ( MemType ("server", (List **)&serverlist, this) );
      for (i = 0; i < 3; i++)
	queues[i] = NEW ( ServerHeads ( i ) );
      serverlist = NEW ( ServerList () );
      context;
      GetMyHostname(bothost);
      G->SPrintF(botuserhost, UHOSTLEN, "%s@%s", G->botuser, bothost);
      curserv = 0;
      context;
      LOG_MSGS = LogModes::MSGS;
      LOG_SERV = LogModes::SERV;
      default_port = 6667;
      SERVER = "SERVER";
      serv_sock = NEW ( Server () );
      NEW ( StringVariable ( "altnick", altnick, NICKMAX, 0 , this ) );
      NEW ( StringVariable ( "realname", botrealname, 80, 0 , this ) );
      NEW ( StringVariable ( "botnick", botname, 0, VAR_PROTECT , this ) );      
      NEW ( IntegerVariable ( "use_console_r", &G->use_console_r, VAR_PROTECT, this ) );
      NEW ( IntegerVariable ( "use_354", &use_354, 0, this ) );
      NEW ( IntegerVariable ( "servlimit", &min_servs, 0, this ) );
      NEW ( IntegerVariable ( "server_timeout", (int *)&server_timeout, 0, this ) );
      NEW ( IntegerVariable ( "server_online", (int *)&server_online, VAR_READONLY, this ) );
      NEW ( IntegerVariable ( "never_give_up", &never_give_up, 0, this ) );
      NEW ( IntegerVariable ( "keep_nick", &keepnick, 0, this ) );
      NEW ( IntegerVariable ( "check_stoned", &check_stoned, 0, this ) );
      NEW ( IntegerVariable ( "quiet_reject", &quiet_reject, 0, this ) );
      NEW ( IntegerVariable ( "max_queue_msg", &maxqmsg, 0, this ) );
      NEW ( IntegerVariable ( "trigger_on_ignore", &trigger_on_ignore, 0, this ) );
      NEW ( IntegerVariable ( "answer_ctcp", &answer_ctcp, 0, this ) );
      NEW ( IntegerVariable ( "server_cycle_wait", (int *)&server_cycle_wait, 0, this ) );
      NEW ( IntegerVariable ( "default_port", &default_port, 0, this ) );
      NEW ( IntegerVariable ( "check_mode_r", &check_mode_r, 0, this ) );
      NEW ( IntegerVariable ( "strict_host", &strict_host, 0, this ) );
      NEW ( IntegerVariable ( "serverror_quit", &serverror_quit, 0, this ) );
      NEW ( IntegerVariable ( "use_silence", &use_silence, 0, this ) );
      NEW ( CoupletVariable ( "flood_msg", &flood_thr, &flood_time, this ) );
      NEW ( CoupletVariable ( "flood_ctcp", &flood_ctcp_thr, &flood_ctcp_time, this ) );
      NEW ( IntegerVariable ( "lowercase_ctcp", &lowercase_ctcp, 0, this ) );
      NEW ( IntegerVariable ( "strict_servernames", &strict_servernames, 0, this ) );
      NEW ( IntegerVariable ( "use_ts4", &use_ts4, VAR_PROTECT, this ) );
      NEW ( IntegerVariable ( "cookie_expire", &cookie_expire, 0, this ) );
      NEW ( IntegerVariable ( "use_ison", &use_ison, 0, this) );
      NEW ( IrcNick ( this ) );
      NEW ( BotName ( this ) );
      NEW ( ServerVar ( this ) );
      NEW ( Servers ( this ) );
      NEW ( HelpReference ( "server.help", this ) );
   }
}

ModuleEntryServer::~ModuleEntryServer ( void ) {
   FUNCTION;
   int i ;
   
   context;
   cycle_time = 100;
   NukeServer(Lang(0x08D, ""));
   context;
   for (i = 0; i < 3; i++)
     DELETE (queues[i]);
   context;
   DELETE ( serv_sock );
   DELETE ( serverlist );
}

int ModuleEntryServer::ExpMem ( void ) {
   FUNCTION;
   return ModuleEntry::ExpMem () + sizeof(ModuleEntryServer);
}

/* puts full hostname in s */
void ModuleEntryServer::GetMyHostname (char * s) {
   FUNCTION;
   struct hostent *hp;
   char *p;
   int l = strlen(G->myip);
   
   if (l && ((G->myip[l - 1] < '0') || (G->myip[l - 1] > '9')))
     strncpyz(s, G->myip, UHOSTLEN);
   else if ((p = getenv("HOSTNAME")) && strchr(p,'.')) 
     strncpyz(s, p, UHOSTLEN);
   else {
      gethostname(s, UHOSTLEN);
      if (!strchr(s, '.')) {
	 if (!(hp = gethostbyname(s))) 
	   G->Fatal(Lang(0x07D, ""));
	 strncpyz(s, hp->h_name, UHOSTLEN);
	 if (!strchr(s, '.')) {
	    if (!hp->h_aliases[0])
	      G->Fatal(Lang(0x07E, ""));
	    strncpyz(s, hp->h_aliases[0], UHOSTLEN);
	    if (!strchr(s, '.'))
	      G->Fatal(Lang(0x07E, ""));
	 }
      }
   }
}

/* a report on the module status */
void ModuleEntryServer::Report (Socket * idx, int details) {
   FUNCTION;
   char s1[128], s[512];
   Socket * servidx;
   int i;
   
   idx->Send(Lang(0x081, "sss", botname, botuserhost, botrealname));
   if (server_online) {
      if (!trying_server) {
	 G->DaysDur(G->now, server_online, s1);
	 strcpy(s,Lang((server_lag == -1) ? 0x083 : 0x082,"sd",s1,server_lag));
      } else
	strcpy(s, Lang(0x084, ""));
      if ((servidx = G->socklist->Find(serv)))
	idx->Send(Lang(0x085, "sds", servidx->Host(), servidx->RemotePort(), s));
   } else
     idx->Send(Lang(0x086, ""));
   for (i = 0; i < 3; i++) 
     idx->Send(Lang(0x087 + i, "d", (queues[i]->Tot() * 100) / maxqmsg));
   if (details) {
      if (min_servs)
	idx->Send(Lang(0x08A, "d", min_servs));
      idx->Send(Lang(0x08C, "dddd", flood_thr, flood_time,
		     flood_ctcp_thr, flood_ctcp_time));
   }
}

void ModuleEntryServer::PostRehash ( void ) {
   FUNCTION;
   void * argv[3];
   FILE * f;
   
   strncpyz(botname, G->origbotname, NICKMAX);
   if (!botname[0])
     G->Fatal(Lang(0x07B, ""));
   if (!serverlist)
      G->Fatal(Lang(0x07C, ""));
   if (!strcasecmp(oldnick, botname) && !strcasecmp(oldnick, altnick) 
       && oldnick[0]) {
      /* change botname back, don't be premature */
      strcpy(botname, oldnick);
      SendNick(G->origbotname);
   } else if (oldnick[0])
     strcpy(botname, oldnick);
   argv[0] = "rehash";
   argv[1] = "";
   argv[2] = "server-rehashing";
   bt_initsvr->Check("", 0, argv);
   if ((f = fopen(".lastcookie","r"))) {
      char buf[201], * server, * cookie;
      int i;
      ServerList * svr;
      
      fgets(buf, 200, f);
      if (buf[i = strlen(buf) - 1] == '\n')
	buf[i] = 0;
      cookie = buf;
      server = G->Split(&cookie);
      printf("Found lastcookie file : %s %s\n", server, cookie);
      if ((svr = (ServerList *)SVR->serverlist->Find(server))) {
	 strncpyz(SVR->newserver, server, 120);
	 SVR->newserverport = svr->Port();
	 strncpyz(SVR->newserverpass, svr->Pass(), 120);
	 svr->SetCookie(cookie);
	 printf("restorting old connection to %s:%d (cookie: %s)\n",
		server, svr->Port(), cookie);
	 SVR->cycle_time = 0;
      }
      unlink(".lastcookie");
   } else {
      if (!server_online) {
	 SVR->cycle_time = 0;
	 SVR->NukeServer(SVR->Lang(0x100, ""));
      }
   }
}

void ModuleEntryServer::FiveMins ( void ) {
   FUNCTION;
   char workbuf[512];
   
   if (check_stoned) {
      if (waiting_for_awake) 
	NukeServer(Lang(0x07A, ""));
      else if (!trying_server) {
	 QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511, "PING :%l\n",
						  (unsigned long) G->now));
	 waiting_for_awake = 1;
      }
   }
}

void ModuleEntryServer::DeqMsg(void) {
   FUNCTION;
   int ok = 0;
   Socket * servidx;
   
   if (G->now != last_time) {
      last_time = G->now;
      if (burst > 0)
	burst--;
      ok = 1;
   }
   if ((serv >= 0) && (servidx = G->socklist->Find(serv))) {
      if (burst < 8) 
	burst += (2 * queues[Q_MODE]->Flush(servidx, 4 - (burst / 2)));
      if (burst < 4)
	burst += (2 * queues[Q_SERVER]->Flush(servidx, 2 - (burst / 2)));
      if (!burst && ok)
	burst += (2 * queues[Q_HELP]->Flush(servidx, 1));
   }
}

/* use when sending msgs... will spread them out so there's no flooding */
void ModuleEntryServer::QueueServer (int which, char * buf, int len) {
   FUNCTION;
   if (serv >= 0) {
      if ((which < 0) || (which > 2)) 
	 G->LOG_MISC.Put( SVR->Lang(0x06A, "d", which));
      else 
	queues[which]->Queue(buf, len);
   }
}

int ModuleEntryServer::MatchMyNick(const char * nick) {
   FUNCTION;
   return ((newbotname[0] && !strcasecmp(nick,newbotname)) ||
	   !strcasecmp(nick,botname));
}

/* make nick!~user@host into nick!user@host if necessary */
/* also the new form: nick!+user@host or nick!-user@host */
void ModuleEntryServer::FixFrom (char * s) {
   FUNCTION;
   char * p;
   if (!strict_host && s) {
      p = (p = strchr(s, '!')) ? p + 1 : s;
      if (strchr("~+-^=", *p))
	strcpy(p, p + 1);
   }
}

/* check for more than 8 control characters in a line */
/* this could indicate: beep flood or CTCP avalanche */
int ModuleEntryServer::DetectAvalanche (char * msg) {
   FUNCTION;
   int count = 0;
   unsigned char *p;
   for (p = (unsigned char *) msg; *p && (count < 8); p++)
      if ((*p == 7) || (*p == 1))
	 count++;
   return (count >= 8);
}

BindReturn ModuleEntryServer::CheckCtcpr (char * nick, char * uhost,
					  UserRecord * u, char * dest,
					  char * keyword, char * args,
					  BindTable * table)  {
   FUNCTION;
   FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL | FlagRecord::ANYWHERE);
   char buf[512], *key = keyword;
   void * argv[6] = { nick, uhost, u, dest, keyword, args };
   
   context;
   fr.Set(u);
   if (lowercase_ctcp) {
      int i = 0;
      
      while (*key && (i < 511)) 
	buf[i++] = toupper(*key++);
      buf[i] = *key;
      key = buf;
   }
   return table->Check(key, &fr, argv);
}

void ModuleEntryServer::SendMode (int which_q, const char * where, const char * what) {
   FUNCTION;
   char workbuf[512];
   
   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "MODE %s%s%s\n",
					    where, (what && what[0]) ? 
					    
					    " " : "", what));
}

void ModuleEntryServer::SendJoin (int which_q, const char * chan, const char * key) {
   FUNCTION;
   char workbuf[512];

   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "JOIN %s %s\n",
					     chan, safe_str(key)));
}

void ModuleEntryServer::SendPart (int which_q, const char * chan, char * reason) {
   FUNCTION;
   char workbuf[512];

   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "PART %s%s%s\n",
					     chan, (reason && reason[0]) ?
					     " :" : "", reason));
}

void ModuleEntryServer::SendKick (int which_q, const char * chan, const char * who, const char * why) {
   FUNCTION;
   char workbuf[512];
   
   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "KICK %s %s :%s\n",
					     chan, who, why));
}

void ModuleEntryServer::SendNotice (int which_q, const char * where, char * what) {
   FUNCTION;
   char workbuf[512];
   
   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "NOTICE %s :%s\n",
					     where, what));
}
   
void ModuleEntryServer::SendTopic (int which_q, const char * chan, char * topic) {
   FUNCTION;
   char workbuf[512];

   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "TOPIC %s%s%s\n",
					     chan, topic ? " :" : "", topic));
}

void ModuleEntryServer::SendMsg (int which_q, const char * where, char * what) {
   FUNCTION;
   char workbuf[512];

   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "PRIVMSG %s :%s\n",
					     where, what));
};

void ModuleEntryServer::SendAct (int which_q, const char * where, char * what) {
   FUNCTION;
   char workbuf[512];

   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, 
					     "PRIVMSG %s :\001ACTION %s\001\n",
					     where, what));
}

void ModuleEntryServer::SendInvite (int which_q, const char * who, const char * where) {
   FUNCTION;
   char workbuf[512];
   
   QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "INVITE %s %s\n",
					    who, where));
}

void ModuleEntryServer::SendSilence (int which_q, char * what) {
   FUNCTION;
   char * p, workbuf[512];
      
   if (use_silence) {
      p = (p = strchr(what,'!')) ? p + 1 : what;
      QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "SILENCE *@%s\n", 
						p));
   }
}

void ModuleEntryServer::SendUnSilence (int which_q, char * what) {
   FUNCTION;
   char * p, workbuf[512];

   if (use_silence) {
      p = (p = strchr(what,'!')) ? p + 1 : what;
      QueueServer(which_q, workbuf, G->SPrintF(workbuf, 511, "SILENCE -%s\n", 
						p));
   }
}

void ModuleEntryServer::SendLusers (void) {
   FUNCTION;
   char buf[8];
   
   strcpy(buf, "LUSERS\n"); /* it gets overwritten */
   QueueServer(Q_SERVER, buf, 7);
}

void ModuleEntryServer::SendNick (const char * nick) {
   FUNCTION;
   char workbuf[512];

   QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511, "NICK %s\n", nick));
}

void ModuleEntryServer::SendWho (int which_q, const char * who) {
   FUNCTION;
   char workbuf[512];
   int l;
   
   if (use_354) 
     l = G->SPrintF(workbuf, 511, "WHO %s %%c%%h%%n%%u%%f\n", who);
   else
     l = G->SPrintF(workbuf, 511, "WHO %s\n", who);
   QueueServer(which_q, workbuf, l);
}

/* close the current server connection */
void ModuleEntryServer::NukeServer (char * reason) {
   FUNCTION;
   if (serv >= 0) {
      Server * servidx = (Server *)G->socklist->Find(serv);
      server_online = 0;
      if (servidx) {
	 if (reason && strncmp(reason, "!!!", 3)) 
	   servidx->Printf("QUIT :%s\n",reason);
	 servidx->Reset();
      }
   }
}

void ModuleEntryServer::ShowHelp (const char * who, char * file, FlagRecord * flags, int fl) {
   FUNCTION;
   int lines = 0, l;
   char s[HELP_BUF_LEN+1];
   FILE * f = G->help_list->Resolve(fl,file);

   if (f) {
      G->HelpSubst(NULL, NULL, 0, 0, NULL);	/* clear flags */
      while (!feof(f)) {
	 fgets(s, HELP_BUF_LEN, f);
	 if (!feof(f)) {
	    if (s[l = (strlen(s) - 1)] == '\n')
	      s[l] = 0;
	    if (!s[0])
	      strcpy(s, " ");
	    G->HelpSubst(s, who, flags, 0, file);
	    if (s[0]) {
	       SendNotice(Q_HELP,who, s);
	       lines++;
	    }
	 }
      }
      fclose(f);
   }
   if (!lines && !(fl & HelpReference::TEXT))
     SendNotice(Q_HELP,who, G->Lang(0x10B, ""));
}
