/*
 * Server.C
*/
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
*/

#include "Server.h"
#include "ServerHeads.h"
#include "IgnoreRecord.h"
#include "BindTable.h"
#include "ServerList.h"
#include <errno.h>
#define CLASS "Server"

Server::~Server ( void ) {
   FUNCTION;
}

void Server::Eof ( void ) {
   FUNCTION;
   SVR->LOG_SERV.Put(SVR->Lang( 0x054, "s", host));
   Reset();
}

void Server::Display2 (Socket * where) {
   FUNCTION;
   where->Send(SVR->Lang( SVR->trying_server ? 0x055 : 0x056, "d", 
		       SVR->server_lag));
}

void Server::Reset ( void ) {
   FUNCTION;
   SVR->server_online = 0;
   SVR->serv = -1;
   if (key >= 0) {
      close(key);
      FD_CLR(key,&G->fds_read);
      FD_CLR(key,&G->fds_write);
      key = -1;
   }
   flags = NONSOCK;
   Connect();
}

void Server::Timeout( void ) {
   FUNCTION;
   SVR->LOG_SERV.Put(SVR->Lang( 0x057, "s", host ));
   Reset();
}

void Server::Connect(void ) {
   FUNCTION;
   char workbuf[512];
   static ServerList * oldserv = 0;
   int r;

   SVR->waiting_for_awake = 0;
   SVR->trying_server = G->now;
   for (r = 0; r < 3; r++)
     SVR->queues[r]->Empty();
   /* start up the counter (always reset it if "never-give-up" is on) */
   if (!oldserv || SVR->never_give_up)
      oldserv = SVR->curserv;
   if (!SVR->cycle_time) { 
      if (SVR->newserverport) {
	 SVR->serverlist->NextServer(SVR->newserver, SVR->newserverport, 
				     SVR->newserverpass);
	 SVR->newserver[0] = 0;
	 SVR->newserverport = 0;
	 SVR->newserverpass[0] = 0;
      } else
	SVR->serverlist->NextServer("", 0, "");
      SVR->LOG_SERV.Put(SVR->Lang( 0x05B, "sd", SVR->curserv->Key(),
				  SVR->curserv->Port()));
      if (!Init (CONNECTING, 0)) 
	G->Fatal(G->Lang(0x001, ""));
      timeout_val = &SVR->server_timeout;
      strncpyz(host, SVR->curserv->Key(), UHOSTLEN);
      remoteport = SVR->curserv->Port();
      if ((r = Scan(remoteport + 1))) {
	 SVR->LOG_SERV.Put(SVR->Lang( 0x05D, "ss", host, (r == 2) ?
				    SVR->Lang( 0x05C, "") : strerror(errno)));
	 if ((oldserv == SVR->curserv) && !SVR->never_give_up)
	   G->Fatal(SVR->Lang( 0x05E, ""));
	 flags = NONSOCK;
      } else {
	 SVR->serv = key;
	 SetNick(SVR->Lang( 0x058, ""));
	 TimeNow();
	 strcpy(SVR->botname,G->origbotname);
	 SVR->SendNick(SVR->botname);
	 if (SVR->curserv->Pass()[0])
	   SVR->QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511, "PASS %s\n",
							SVR->curserv->Pass()));
	 SVR->QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511, 
						       "USER %s %s %s :%s\n",
						       G->botuser, SVR->bothost,
						       host, SVR->botrealname));
	 if (SVR->use_ts4 && SVR->curserv->Cookie() &&
	     (SVR->curserv->CookieAge() < (SVR->cookie_expire * 60))) 
	   SVR->QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511,
							"COOKIE %s\n",
							SVR->curserv->Cookie()));
	 SVR->cycle_time = 0;
      }
      if (SVR->server_cycle_wait) 
	SVR->cycle_time = SVR->server_cycle_wait;
   } 
}

void Server::Activity (char * msg, int) {
   FUNCTION;
   char * from, * code;
   void * args[3];
   
   if (SVR->trying_server) {
      SVR->LOG_SERV.Put(SVR->Lang( 0x059, "s", host));
      SVR->trying_server = 0;
      SVR->waiting_for_awake = 0;
      timeout_val = 0;
   }
   from = "";
   if (msg[0] == ':') {
      msg++;
      from = G->Split(&msg);
   }
   code = G->Split(&msg);
   if (G->use_console_r && ((strcmp(code, "PRIVMSG") && strcmp(code, "NOTICE"))
			 || !G->ignores->Match(from))) {
      LogModes LOG_RAW ( "r" );
      
      LOG_RAW.Put(SVR->Lang( 0x05A, "sss", from, code, msg));
   }
   if (from[0])
      SVR->FixFrom(from);
   CONTEXT(code);
   args[0] = from;
   args[1] = code;
   args[2] = msg;
   SVR->bt_raw->Check(code, 0, args);
}

int Server::ExpMem ( void ) {
   FUNCTION;
   return sizeof(Server);
}
