/*
 * ServerList.C
 */
/*
 * This file is part of the eggdrop 2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "server.h"
#include "ModuleEntryServer.h"
#include "ServerList.h"
#include "Socket.h"
#include "ServerHeads.h"
#include "ScriptContext.h"
#define CLASS "ServerList"

ServerList::ServerList (const char * ss) : KeyedList<UHOSTLEN> () {
   FUNCTION;
   char * p, *q, buf[512];
   
   context;
   strncpyz(buf, ss, 511);
   realname = 0;
   cookie = 0;
   cookie_time = 0;
   if (!(q = strchr(buf,':'))) {
      port = SVR->default_port;
      pass = 0;
   } else {
      *q = 0;
      SetKey(buf);
      p = ++q;
      if (!(q = strchr(p,':'))) 
	pass = 0;
      else {
	 *q++ = 0;
	 strcpy(pass = (char *)my_malloc(strlen(q)+1,SVR->memtype_server), q);
      }
      port = atoi(p);
   }
   KeyedList_Cons( *SVR->serverlist, buf, KeyedList<UHOSTLEN>::NODUPS |
		  KeyedList<UHOSTLEN>::APPEND );
}

ServerList::~ServerList ( void ) {
   FUNCTION;
   if (pass) 
     my_free(pass,SVR->memtype_server);
   if (realname)
     my_free(realname,SVR->memtype_server);
   if (cookie)
     my_free(cookie,SVR->memtype_server);
   SVR->curserv = 0;
}

void ServerList::NextServer (const char * serv, int port, char * pass) {
   FUNCTION;
   ServerList * x = this;
   
   if (SVR->curserv)
     SVR->curserv->cookie_time = G->now;
   if (serv[0]) 
     while ((x = x->Next()))
       if ((x->port == port) && (!strcasecmp(x->key, serv)
				 || (x->realname &&
				     !strcasecmp(x->realname,serv)))) {
	  SVR->curserv = x;
	  return;
       }
   if (serv[0]) {
      x = NEW ( ServerList ( serv ) );
      x->port = port;
      if (pass && pass[0])
	strcpy(x->pass = (char *)my_malloc(strlen(pass)+1,SVR->memtype_server),pass);
      SVR->curserv = x;
   } else {
      if (SVR->curserv && SVR->curserv->Next())
	SVR->curserv = SVR->curserv->Next();
      else
	SVR->curserv = this->Next();
   }
}

void ServerList::SetRealName ( const char * from ) {
   FUNCTION;
   if (realname)
     my_free(realname,SVR->memtype_server);
   if (SVR->strict_servernames) {
      realname = 0;
      SetKey(from);
   } else
     strcpy(realname = (char *)my_malloc(strlen(from)+1,SVR->memtype_server),from);
}

void ServerList::Dump ( Socket * idx ) {
   FUNCTION;
   ServerList * ptr = this;
   char * s;
   int l;
   
   if (!next) 
     idx->Send(SVR->Lang(0x061, ""));
   else {
      idx->Send(SVR->Lang(0x062, ""));
      while ((ptr = ptr->Next())) {
	 s = SVR->Lang((ptr == SVR->curserv) ? 0x063 : 0x064, "s 4", ptr->key,
		       20 - strlen(ptr->key), ptr->port ? ptr->port :
		       SVR->default_port);
	 if (realname) {
	    l = strlen(s);
	    l += G->SPrintF(s + l, 22, " (%s", ptr->realname);
	    s[l++] = ')';
	    s[l] = 0;
	 }
	 idx->Send(s);
      }
   }
}
   
int ServerList::ExpMem (void) {
   FUNCTION;
   int tot = sizeof(ServerList), i;
   ServerList *s = this;
   
   context;
   while ((s = s->Next())) {
      tot += sizeof(ServerList);
      if (s->pass)
	tot += strlen(s->pass) + 1;
      if (s->realname)
	tot += strlen(s->realname) + 1;
      if (s->cookie)
	tot += strlen(s->cookie) + 1;
   }
   for (i = 0; i < 3; i++) 
     tot += SVR->queues[i]->ExpMem();
   return tot;
}

void ServerList::Dump (ScriptContext * sc) {
   FUNCTION;
   ServerList * s = this;
   char buf[512];
   
   sc->Clear();
   while ((s = s->Next())) {
      G->SPrintF(buf, 511, "%s:%d%s%s %s", s->key, s->port ? s->port 
		 : SVR->default_port, s->pass ? ":" : "", safe_str(s->pass),
		 safe_str(s->realname));
      sc->AppendString(buf);
   }
}

void ServerList::SetCookie ( const char * c ) {
   FUNCTION;
   cookie_time = G->now;
   if (cookie)
     my_free(cookie, SVR->memtype_server);
   cookie = (char *)my_malloc(strlen(c) + 1, SVR->memtype_server);
   strcpy(cookie, c);
}
