/*
 * server.c - part of server.mod
 * basic irc server support
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#define MAKING_SERVER
#include "MessageQueue.h"
#include "DccChatPass.h"
#include "UserRecord.h"
#include "UserEntryPass.h"
#include "BindTable.h"
#include "IgnoreRecord.h"
#include "HookList.h"
#include "server.h"
#include <errno.h>
#ifdef HAVE_UNAME
#include <sys/utsname.h>
#endif
#include "ModuleEntryServer.h"
#include "Server.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "Channel.h"
#include "Command.h"
#include "BindTableRaw.h"
#include "BindTableWall.h"
#include "BindTableMsg.h"
#include "BindTableCtcpr.h"
#include "BindTableFlud.h"
#include "HelpReference.h"
#include "ServerHeads.h"
#include "ServerList.h"
#define CLASS ""

static ModuleEntryEggdrop * G;
ModuleEntryServer * SVR;

/* flood tables */
static int lastmsgs[FLOOD_GLOBAL_MAX];
static char lastmsghost[FLOOD_GLOBAL_MAX][81];
static time_t lastmsgtime[FLOOD_GLOBAL_MAX];

extern "C" {
   char *server_start (ModuleEntryEggdrop *,ModuleEntry *);
};

/**************************************************************************
 * MSG COMMANDS                                                           *
 **************************************************************************/

static BindReturn msg_pass (void ** args, const char * ) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   char * nick = (char *)args[0], buf[512], *par = buf, *old, *newp;
   FlagRecord fr (u);
   
   if (!SVR->MatchMyNick(nick) && u && !fr.GlobBot() && !fr.GlobCommon()) {
      strcpy(par, (char *)args[3]);
      if (!par[0]) {
	 SVR->SendNotice(Q_SERVER,nick,SVR->Lang(u->PassMatch("-")?0x002:0x001,""));
	 G->LOG_CMDS.Put(SVR->Lang(0x003,"sss",nick,args[1],u->Handle()));
      } else {
	 old = G->Split(&par);
	 if (!u->PassMatch("-") && !par[0]) 
	   SVR->SendNotice(Q_SERVER,nick,SVR->Lang(0x004,""));
	 else {
	    if (par[0]) {
	       if (!u->PassMatch(old)) {
		  SVR->SendNotice(Q_SERVER,nick,SVR->Lang(0x005,""));
		  return EXECUTED;
	       }
	       newp = G->Split(&par);
	    } else
	      newp = old;
	    G->LOG_CMDS.Put(SVR->Lang(0x006,"sss",nick,args[1],u->Handle()));
	    if (strlen(newp) > 15)
	      newp[15] = 0;
	    if (strlen(newp) < 6) {
	       SVR->SendNotice(Q_SERVER,nick,SVR->Lang(0x007,""));
	       return EXECUTED;
	    }
	    SetPassword(u, newp);
	    SVR->SendNotice(Q_SERVER,nick,SVR->Lang(newp==old?0x008:0x009,"s",newp));
	 }
      }
   }
   return EXECUTED;
}

static BindReturn msg_help (void ** args, const char * nick) {
   FUNCTION;
   char *p, *q, buf[512];
   UserRecord * u = (UserRecord *)args[2];
   
   nick = (char *)args[0];
   if (!SVR->MatchMyNick(nick)) {
      if (!u) {
	 if (!SVR->quiet_reject) {
	    SVR->SendNotice(Q_HELP,nick, SVR->Lang( 0x00A, ""));
	    SVR->SendNotice(Q_HELP,nick, SVR->Lang( 0x00B, "s", SVR->botname));
	 }
      } else if (G->helpdir[0]) {
	 FlagRecord fr (u,FlagRecord::ANYWHERE | FlagRecord::GLOBAL |
			FlagRecord::CHANNEL);
	 if (!((char *)args[3])[0])
	   SVR->ShowHelp(nick, "help", &fr,0);
	 else {
	    for (p = (char *)args[3], q = buf; *p; p++, q++) {
	       if ((*p >= 'A') && (*p <= 'Z')) 
		 *q = *p + ('a' - 'A');
	       else
		 *q = *p;
	    }
	    *q = *p;
	    SVR->ShowHelp(nick, buf, &fr,0);
	 }
      } else
	SVR->SendNotice(Q_HELP,nick, G->Lang(0x10B, ""));
   }
   return EXECUTED;
}

static BindReturn msg_status (void ** args, const char * nick) {
   FUNCTION;
   char s[256], p[160];
   char * ve_t, * un_t;
   int i, k, l, m;
   Channel *chan;
#ifdef HAVE_UNAME
   struct utsname un;
   if (!uname(&un) < 0) {
#endif
      ve_t = " ";
      un_t = "*unkown*";
#ifdef HAVE_UNAME
   } else {
      ve_t = un.release;
      un_t = un.sysname;
   }
#endif
   nick = (char *)args[0];
   if (!SVR->MatchMyNick(nick)) {
      G->LOG_CMDS.Put(SVR->Lang(0x00C, "sss", nick, args[1],
			   ((UserRecord *)args[2])->Handle()));
      SVR->SendNotice(Q_HELP,nick, SVR->Lang( 0x00D, "ss", SVR->botname, 
					     G->VersionStr()));
      SVR->SendNotice(Q_HELP,nick, SVR->Lang( 0x00E, "ss", un_t, ve_t));
      if (G->admin[0])
	SVR->SendNotice(Q_HELP,nick, SVR->Lang( 0x00F, "s", G->admin));
      strcpy(p, SVR->Lang( 0x010, ""));
      k = m = strlen(p);
      for (chan = G->chanset; chan; chan = chan->Next()) {
	 l = G->StrCpy(s, chan->Key());
	 if (!chan->Active())
	   l += G->StrCpy(s + l, SVR->Lang( 0x011, ""));
	 else if (chan->Pending())
	   l += G->StrCpy(s + l, SVR->Lang( 0x012, ""));
	 else if (!chan->Meop())
	   l += G->StrCpy(s + l, SVR->Lang( 0x013, ""));
	 s[l++] = ',';
	 s[l++] = ' ';
	 if ((k + l) > 70) {
	    SVR->SendNotice(Q_HELP, nick, p);
	    for (k = 0; k < m; k++)
	      p[k] = ' ';
	    k = 10;
	 }
	 strcpy(p + k, s);
	 k += l;
      }
      if (k > 10) {
	 p[k - 2] = 0;
	 SVR->SendNotice(Q_HELP, nick, p);
      }
      i = G->userlist->Count();
      SVR->SendNotice(Q_HELP, nick, SVR->Lang( i == 1 ? 0x014 : 0x015, "dd", i,
					   (int) (G->memtype_list->ExpectedMemory()
						  / 1024)));
      G->Days(G->now, SVR->server_online, p);
      SVR->SendNotice(Q_HELP, nick, SVR->Lang( 0x016, "s", p));
      SVR->SendNotice(Q_HELP, nick, SVR->Lang( 0x017, "ss", SVR->botname,
					      SVR->botuserhost));
   }
   return EXECUTED;
}
   
static BindReturn msg_memory (void ** args, const char * nick) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   
   nick = (char *)args[0];
   if (!SVR->MatchMyNick(nick)) {
      G->LOG_CMDS.Put( SVR->Lang( 0x018, "sss", nick, args[1], u->Handle()));
      SVR->SendNotice(Q_HELP, nick, SVR->Lang( 0x019, "d", 
					G->memtype_list->ExpectedMemory()));
   }
   return EXECUTED;
}

static BindReturn msg_die (void ** args, const char * nick) {
   FUNCTION;
   char *pass, *msg = (char *)args[3];
   UserRecord * u = (UserRecord *)args[2];
   
   nick = (char *)args[0];
   if (!SVR->MatchMyNick(nick)) {
      pass = G->Split(&msg);
      if (((UserRecord *)args[2])->PassMatch(pass)) {
	 Socket * servidx = G->socklist->Find(SVR->serv);
	 G->LOG_CMDS.Put( SVR->Lang( 0x01A, "sss", nick, args[1], u->Handle()));
	 servidx->Printf("NOTICE %s :%s\n", nick, SVR->Lang( 0x01B, ""));
	 G->userlist->WriteUserfile(NULL);
	 sleep(3);		 /* give the server time to understand */
	 G->LOG_MISC.Put(SVR->Lang( 0x01D, "ss", nick, args[1]));
	 G->Fatal(msg[0] ? msg : safe_str(nick));
      }
      G->LOG_CMDS.Put( SVR->Lang( 0x01E, "sss", nick, args[1], u->Handle()));
   }
   return EXECUTED;
}

static BindReturn msg_rehash (void ** args, const char * nick) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   
   nick = (char *)args[0];
   if (!SVR->MatchMyNick(nick)) {
      if (u->PassMatch((char *)args[3])) {
	 G->LOG_CMDS.Put( SVR->Lang( 0x01F, "sss", nick, args[1], u->Handle()));
	 SVR->SendNotice(Q_SERVER, nick, G->Lang(0x10E, ""));
	 G->do_restart = -2;
      } else
	G->LOG_CMDS.Put( SVR->Lang( 0x020, "sss", nick, args[1], u->Handle()));
   }
   return EXECUTED;
}

static BindReturn msg_jump (void ** args, const char * nick) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   char * s, buf[512], *par = buf;
   int port;
   
   nick = (char *)args[0];
   if (!SVR->MatchMyNick(nick)) {
      strcpy(par, (char *)args[3]);
      if (u->PassMatch(G->Split(&par))) {
	 if (par[0]) {
	    s = G->Split(&par);
	    port = atoi(G->Split(&par));
	    if (!port)
	      port = SVR->default_port;
	    G->LOG_CMDS.Put( SVR->Lang( 0x021, "ssssds", nick, args[1],
				       u->Handle(), s, port, par));
	    strcpy(SVR->newserver, s);
	    SVR->newserverport = port;
	    strcpy(SVR->newserverpass, par);
	 } else
	   G->LOG_CMDS.Put( SVR->Lang( 0x022, "sss", nick, args[1], u->Handle()));
	 G->socklist->Find(SVR->serv)->Printf("NOTICE %s :%s\n", nick,
						SVR->Lang( 0x023, ""));
	 SVR->cycle_time = 0;
	 SVR->NukeServer(SVR->Lang( 0x100, ""));
      } else
	G->LOG_CMDS.Put( SVR->Lang( 0x024, "sss", nick, args[1], u->Handle()));
   }
   return EXECUTED;
}

#define MSG_COMMANDS 7
static command_binds my_msg[MSG_COMMANDS]={ 
   { "die", "n", msg_die, NULL },
   { "help", "", msg_help, NULL },
   { "jump", "m", msg_jump, NULL },
   { "memory", "m", msg_memory, NULL },
   { "pass", "", msg_pass, NULL },
   { "rehash", "m", msg_rehash, NULL },
   { "status", "m|m", msg_status, NULL }
};

/***************************************************************************
 * SCRIPT COMMANDS                                                         *
 ***************************************************************************/

static int script_jump (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   BADARGS(1, 4, SVR->Lang(0x02B, ""));
   if (argc >= 2) {
      strncpyz(SVR->newserver, argv[1]->AsString(),120);
      SVR->newserverport = (argc >= 3) ? argv[2]->AsInteger() : SVR->default_port;
      if (argc == 4)
	strncpyz(SVR->newserverpass, argv[3]->AsString(), 120);
   }
   SVR->cycle_time = 0;
   SVR->NukeServer(SVR->Lang(0x100, ""));
   return SCRIPT_OK;
}

static int script_putserv (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char s[511], *p;
   int l;
   
   context;
   BADARGS(2, 2, SVR->Lang(0x02C, ""));
   strncpy(s, argv[1]->AsString(), 510);
   s[510] = 0;
   if ((p = strchr(s, '\n')))
     *p = 0;
   if ((p = strchr(s, '\r')))
     *p = 0;
   s[l = strlen(s)] = '\n';
   s[++l] = 0;
   SVR->QueueServer(argv[0]->AsString()[3] == 's' ? Q_SERVER : Q_HELP, s, l);
   return SCRIPT_OK;
}

static int script_dumpfile (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char nick[NICKMAX + 1], fn[DIRLEN+1];
   
   context;
   BADARGS(3, 3, SVR->Lang(0x02D, ""));
   strncpy(nick, argv[1]->AsString(), NICKMAX);
   nick[NICKMAX] = 0;
   G->SPrintF(fn, DIRLEN, "text/%s", argv[2]->AsString());
   SVR->ShowHelp(argv[1]->AsString(), fn, 0, HelpReference::TEXT);
   return SCRIPT_OK;
}

/************************************************************************
 * RAW SERVER HANDLING                                                  *
 ************************************************************************/

/* 001: welcome to IRC (use it to fix the server name) */
static BindReturn got001 (void ** args, const char *) {
   FUNCTION;
   int i;
   Socket * servidx = G->socklist->Find(SVR->serv);
   char * p  = (char *)args[2], *from = (char *)args[0];
   void * argv[3];
   /* ok...param #1 of 001 = what server thinks my nick is */
   
   SVR->server_online = G->now;
   if (*p == ':') 
     p++;
   for (i = 0; *p && (*p != ' ') && i < NICKMAX; i++, p++)
     SVR->botname[i] = *p;
   SVR->botname[i] = 0;
   /* init-server */
   argv[0] = from;
   argv[1] = "";
   argv[2] = p;
   SVR->bt_initsvr->Check("", 0, argv);
   if (strcasecmp(from, SVR->curserv->Key())) {
      G->LOG_MISC.Put( SVR->Lang( 0x02E, "ss", servidx->Host(), from));
      SVR->curserv->SetRealName ( from );
   }
   return EXECUTED;
}

/* 014: <cookie> :Is you connection cookie */
static BindReturn got014 (void ** args, const char *) {
   FUNCTION;
   char * p  = (char *)args[2];
   
   context;
   G->Split(&p);
   if (SVR->use_ts4) {
      FILE * f = fopen(".lastcookie", "w");
      
      p = FixColon(p);
      if (f) {
	 fprintf(f, "%s %s\n", SVR->curserv->Key(), p);
	 fclose(f);
      }
      SVR->curserv->SetCookie(FixColon(p));
   }
   return EXECUTED;
}

/* do on NICK, PRIVMSG, and NOTICE -- and JOIN */
static int detect_flood (char * floodnick, char * floodhost, UserRecord * u,
			 char * from, int which) {
   FUNCTION;
   char *p;
   char ftype[10], h[UHOSTLEN+1];
   int thr = 0, lapse = 0;
   FlagRecord atr(u);
   
   context;
   if (!atr.GlobBot() && !atr.GlobFriend()) {
      /* determine how many are necessary to make a flood */
      switch (which) {
	case FLOOD_PRIVMSG:
	case FLOOD_NOTICE:
	 thr = SVR->flood_thr;
	 lapse = SVR->flood_time;
	 strcpy(ftype, "msg");
	 break;
	case FLOOD_CTCP:
	 lapse = SVR->flood_ctcp_time;
	 thr = SVR->flood_ctcp_thr;
	 strcpy(ftype, "ctcp");
	 break;
      }
      if (thr && !SVR->MatchMyNick(floodnick) &&
	  strcasecmp(floodhost, SVR->botuserhost)) {
	 if ((p = strchr(floodhost, '@'))) {
	    p++;
	    if (!strcasecmp(lastmsghost[which], p)) {	/* new */
	       strcpy(lastmsghost[which], p);
	       lastmsgtime[which] = G->now;
	       lastmsgs[which] = 0;
	    } else if (lastmsgtime[which] < G->now - lapse) {
	       /* flood timer expired, reset it */
	       lastmsgtime[which] = G->now;
	       lastmsgs[which] = 0;
	    } else {
	       lastmsgs[which]++;
	       if (lastmsgs[which] >= thr) {	/* FLOOD */
		  void * args[5] = { floodnick, from, u, ftype, 0 };
		  /* reset counters */
		  lastmsgs[which] = 0;
		  lastmsgtime[which] = 0;
		  lastmsghost[which][0] = 0;
		  if (SVR->bt_flud->Check(ftype, 0, args)
		      != EXEC_LOG) {
		     /* private msg */
		     G->SPrintF(h, UHOSTLEN, "*!*@%s", p);
		     G->LOG_MISC.Put( SVR->Lang( 0x02F, "s", p));
		     NEW ( IgnoreRecord (h,G->origbotname,
					 SVR->Lang(0x030+which,""),
					 G->now + (60 * G->ignore_time)) );
		     SVR->SendSilence(Q_MODE, p);
		     return 1;
		  }
	       }
	    }
	 }
      }
   }
   return 0;
}

/* private message */
static BindReturn gotmsg (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], buf2[512], *msg = buf2;
   char *to, buf[UHOSTLEN], *nick, ctcpbuf[512], *uhost = buf, *ctcp;
   char *p, *p1, *code;
   UserRecord * u;
   int ctcp_count = 0;
   int ignoring;
   
   if (strchr("#&+@",*((char *)args[2])))
     return NOMATCH;
   strncpy(msg, (char *)args[2], 511);
   msg[511] = 0;
   ignoring = G->ignores->Match(from) != 0;
   to = G->Split(&msg);
   msg = FixColon(msg);
   /* only check if flood-ctcp is active */
   strcpy(uhost,from);
   nick = G->SplitNick(&uhost);
   if (SVR->flood_ctcp_thr && SVR->DetectAvalanche(msg)) {
      if (!ignoring) {
	 LogModes LOG_MODES ( LogModes::MODES );
	 
	 LOG_MODES.Put( 0, "Avalanche from %s - ignoring", from);
	 p = (p = strchr(uhost, '@')) ? p + 1 : uhost;
	 G->SPrintF(ctcpbuf, 511, "*!*@%s", p);
	 NEW ( IgnoreRecord (ctcpbuf, G->origbotname, "ctcp avalanche", 
			     G->now + (60 * G->ignore_time)) );
      }
      return EXECUTED;
   } 
   /* check for CTCP: */
   SVR->ctcp_reply[0] = 0;
   SVR->ctcp_reply_len = 0;
   p = strchr(msg, 1);
   if ((u = G->userlist->FindHost(from)) == G->userlist)
     u = NULL;
   while (p && *p) {
      p++;
      p1 = p;
      while ((*p != 1) && (*p != 0))
	 p++;
      if (*p == 1) {
	 *p = 0;
	 ctcp = strcpy(ctcpbuf, p1);
	 strcpy(p1 - 1, p + 1);
	 if (!ignoring)
	   detect_flood(nick, uhost, u, from, FLOOD_CTCP);
	 /* respond to the first answer_ctcp */
	 p = strchr(msg, 1);
	 if (ctcp_count < SVR->answer_ctcp) {
	    ctcp_count++;
	    if (ctcp[0] != ' ') {
	       code = G->Split(&ctcp);
	       if ((to[0] == '$') || strchr(to, '.')) {
		  LogModes LOG_PUBLIC ( LogModes::PUBLIC );
		  
		  if (!ignoring)
		    LOG_PUBLIC.Put(SVR->Lang( 0x033, "sssss", code, ctcp,
					   nick, uhost, to));
	       } else {
		  if (!ignoring || SVR->trigger_on_ignore) {
		     if ((SVR->CheckCtcpr(nick, uhost, u, to, code, ctcp,
				      SVR->bt_ctcp) == NOMATCH)
			 && !ignoring && 
			 ((SVR->lowercase_ctcp && !strcasecmp(code, "DCC"))
			    || (!SVR->lowercase_ctcp && !strcmp(code, "DCC")))) {
			/* if it gets this far unhandled, it means that 
			 * the user is totally unknown */
			code = G->Split(&ctcp);
			if (!strcmp(code,"CHAT")) {
			   if (!SVR->quiet_reject) 
			     SVR->SendNotice(Q_HELP, nick, SVR->Lang( u ?
							       0x034:0x035, ""));
			   G->LOG_MISC.Put( SVR->Lang( 0x072, "s", from));
			} else 
			  G->LOG_MISC.Put( SVR->Lang( 0x037, "ss", code, from));
		     }
		     SVR->LOG_MSGS.Put(SVR->Lang( strcmp(code, "ACTION") ?
						 0x038 : 0x039, "sssss", to, nick,
						 ctcp, code, uhost));
		  }
	       }
	    }
	 }
      }
   }
   /* send out possible ctcp responses */
   if (SVR->ctcp_reply_len)
     SVR->SendNotice(Q_SERVER, nick, SVR->ctcp_reply);
   if (msg[0]) {
      if ((to[0] == '$') || strchr(to, '.')) {	
	 /* msg from oper */
	 if (!ignoring) {
	    LogModes x ( LogModes::MSGS | LogModes::SERV );
	    detect_flood(nick, uhost, u, from, FLOOD_PRIVMSG);
	    x.Put(SVR->Lang( 0x03A, "ssss", nick, uhost, to, msg));
	 }
      } else if (!ignoring) {
	 FlagRecord fr (u, FlagRecord::GLOBAL | FlagRecord::CHANNEL
			| FlagRecord::ANYWHERE);
	 void * argv[4] = { nick, uhost, u, msg };
	 BindReturn x;
	 
	 detect_flood(nick, uhost, u, from, FLOOD_PRIVMSG);
	 if (SVR->trigger_on_ignore) 
	   SVR->bt_msgm->Check(msg, &fr, argv);
	 code = G->Split(&msg);
	 argv[3] = msg;
	 if ((x = SVR->bt_msg->Check(code, &fr, argv))
	     == EXEC_LOG)
	   G->LOG_CMDS.Put( SVR->Lang( 0x03B, "sssss", nick, uhost, 
		     u ? u->Handle() : "*", code, msg));
	 else if ((x == NOMATCH) || (x == EXEC_BRK))
	   SVR->LOG_MSGS.Put(SVR->Lang( 0x03C, "sss", from, code, msg));
      }
   }
   return EXECUTED;
}

/* private notice */
static BindReturn gotnotice (void ** args, const char *) {
   FUNCTION;
   char * from = (char *)args[0], buf2[512], *msg = buf2;
   char *to, *nick, ctcpbuf[512], *p, *p1, buf[512], *uhost = buf, *ctcp;
   UserRecord * u;
   int ignoring;
   
   if (strchr("#&+@",*(char *)args[2]))
     return NOMATCH;
   strncpy(msg, (char *)args[2], 511);
   msg[511] = 0;
   ignoring = G->ignores->Match(from) != 0;
   to = G->Split(&msg);
   msg = FixColon(msg);
   strcpy(uhost,from);
   nick = G->SplitNick(&uhost);
   if (SVR->flood_ctcp_thr && SVR->DetectAvalanche(msg)) {
      LogModes LOG_MODES ( LogModes::MODES );
      /* discard -- kick user if it was to the channel */
      if (!ignoring)
	LOG_MODES.Put(SVR->Lang( 0x03D, "s", from));
      return EXECUTED;
   }
   if ((u = G->userlist->FindHost(from)) == G->userlist)
     u = NULL;
   /* check for CTCP: */
   p = strchr(msg, 1);
   while (p && *p) {
      p++;
      p1 = p;
      while ((*p != 1) && (*p != 0))
	p++;
      if (*p == 1) {
	 *p = 0;
	 ctcp = strcpy(ctcpbuf, p1);
	 strcpy(p1 - 1, p + 1);
	 if (!ignoring)
	   detect_flood(nick,uhost,u,from, FLOOD_CTCP);
	 p = strchr(msg, 1);
	 if (ctcp[0] != ' ') {
	    char * code = G->Split(&ctcp);
	    if ((to[0] == '$') || strchr(to, '.')) {
	       LogModes LOG_PUBLIC ( LogModes::PUBLIC );
	       if (!ignoring)
		 LOG_PUBLIC.Put(SVR->Lang( 0x03E, "sssss", code, ctcp, nick,
					from, to));
	    } else if (!ignoring || SVR->trigger_on_ignore) {
	       SVR->CheckCtcpr(nick, from, u, to, code, ctcp, SVR->bt_ctcr);
	       if (!ignoring)
		 SVR->LOG_MSGS.Put(SVR->Lang( 0x03F, "sssss", code, ctcp, nick,
				      from, to));
	    }
	 }
      }
   }
   if (msg[0]) {
      if (((to[0] == '$') || strchr(to, '.')) && !ignoring) {
	 LogModes x ( "ms" );
	 
	 detect_flood(nick,uhost,u,from, FLOOD_NOTICE);
	 x.Put(SVR->Lang( 0x040, "ssss", nick, uhost, to, msg));
      } else {
	 detect_flood(nick, uhost, u,from, FLOOD_NOTICE);
	 /* server notice? */
	 if (!from[0] || !nick[0]) {
	    /* hidden `250' connection count message from server */
	    if (strncmp(msg, "Highest connection count:", 25))
	      SVR->LOG_SERV.Put(SVR->Lang( 0x041, "s", msg));
	 } else if (!ignoring || SVR->trigger_on_ignore) {
	    FlagRecord fr (u,FlagRecord::GLOBAL | FlagRecord::CHANNEL |
			   FlagRecord::ANYWHERE);
	    void * args[4] = { nick, uhost, u, msg };
	    
	    SVR->bt_notc->Check(msg, &fr, args);
	    SVR->LOG_MSGS.Put(SVR->Lang( 0x042, "sss", nick, from, msg));
	 }
      }
   }
   return EXECUTED;
}

/* <server> 251 <to> :there are 2258 users and 805 invisible on 127 servers */
static BindReturn got251 (void ** args, const char *) {
   FUNCTION;
   int i;
   char * servs, buf[512], *msg = buf;
   
   if (SVR->min_servs) {
      strncpy(msg, (char *)args[2], 511);
      msg[511] = 0;
      G->Split(&msg);
      msg = FixColon(msg);
      for (i = 0; i < 8; i++)
	G->Split(&msg);
      servs = G->Split(&msg);
      if (!strncmp(msg, "servers", 7)) {
	 while (*servs && (*servs < 32)) 
	   servs++; /* I've seen some lame nets put bolds & stuff in here :/ */
	 if ((i = atoi(servs)) < SVR->min_servs) {
	    SVR->LOG_SERV.Put(SVR->Lang( 0x043, "dd", SVR->min_servs, i));
	    SVR->NukeServer(SVR->Lang( 0x100, ""));
	 }
      }
   }
   return EXECUTED;
}

/* WALLOPS: oper's nuisance */
static BindReturn gotwall (void ** args, const char *) {
   FUNCTION;
   char *nick, *p, buf[512], *from = buf, *msg = (char *)args[2];
   void * argv[2];
   
   strncpy(from, (char *)args[0], 511);
   from[511] = 0;
   if (*msg == ':')
     msg++;
   if ((p = strchr(from,'!')) && (p == strrchr(from,'!'))) 
     nick = G->SplitNick(&from);
   else 
     nick = from;
   argv[0] = nick;
   argv[1] = msg;
   if (SVR->bt_wall->Check(msg, 0, argv) == EXEC_LOG) {
      LogModes LOG_WALL ( LogModes::WALL );
      
      LOG_WALL.Put(SVR->Lang( (nick == from) ? 0x044 : 0x045, "sss", 
			   nick, from, msg));
   }
   return EXECUTED;
}

/* ping from server */
static BindReturn gotpong (void ** args, const char * ) {
   FUNCTION;
   char buf[512], *msg = buf;
   
   strcpy(msg, (char *)args[2]);
   G->Split(&msg);
   msg = FixColon(msg);		/* scrap server name */
   SVR->waiting_for_awake = 0;
   SVR->server_lag = G->now - atol(msg);
   if (SVR->server_lag > 99999)
     SVR->server_lag = -1;
   return EXECUTED;
}

/* ison failed! meaning my nick is not in use! */
static BindReturn got303 (void ** args, const char *) {
   FUNCTION;
   char buf[512], *msg = buf;
   
   if (SVR->use_ison) {
      strcpy(msg, (char *)args[2]);
      G->Split(&msg);
      msg = FixColon(msg);		/* scrap server name */
      if (!msg[0] && strcasecmp(SVR->botname, G->origbotname)) {
	 G->LOG_MISC.Put( SVR->Lang( 0x046, "s", G->origbotname));
	 strcpy(SVR->newbotname, SVR->botname);
	 strcpy(SVR->botname, G->origbotname);
	 SVR->SendNick(SVR->botname);
      }
   }
   return EXECUTED;
}

/* trace failed! meaning my nick is not in use! 206/401/402 */
static BindReturn trace_fail (void **, const char *) {
   FUNCTION;
   if (!SVR->use_ison) {
      if (strcasecmp(SVR->botname, G->origbotname)) {
	 G->LOG_MISC.Put( SVR->Lang( 0x046, "s", G->origbotname));
	 strcpy(SVR->newbotname, SVR->botname);
	 strcpy(SVR->botname, G->origbotname);
	 SVR->SendNick(SVR->botname);
      }
   }
   return EXECUTED;
}

/* 432 : bad nickname */
static BindReturn got432 (void **, const char *) {
   FUNCTION;
   G->LOG_MISC.Put( SVR->Lang( 0x047, ""));
   if (!SVR->newbotname[0]) {	
      strcpy(SVR->newbotname, SVR->botname);	
      G->MakePass(SVR->botname);
      SVR->botname[NICKMAX] = 0;
      SVR->SendNick(SVR->botname);
   } else {
      strcpy(SVR->botname, SVR->newbotname);
      SVR->newbotname[0] = 0;
   }
   return EXECUTED;
}

/* 433 : nickname in use */
/* change nicks till we're acceptable or we give up */
static BindReturn got433 (void **, const char *) {
   FUNCTION;
   char c, *oknicks = "^-_\\[]`", *p;
   if (SVR->newbotname[0]) {
      strcpy(SVR->botname, SVR->newbotname);
      SVR->newbotname[0] = 0;
   }  else { /* alternate nickname defined? */
      if (SVR->altnick[0] && strcasecmp(SVR->altnick, SVR->botname)) 
	strcpy(SVR->botname, SVR->altnick);
      /* if alt nickname failed, drop thru to here */
      else {
	 int l = strlen(SVR->botname) - 1;
	 c = SVR->botname[l];
	 p = strchr(oknicks, c);
	 if (((c >= '0') && (c <= '9')) || (p != NULL)) {
	    if (p == NULL) {
	       if (c == '9')
		 SVR->botname[l] = oknicks[0];
	       else
		 SVR->botname[l] = c + 1;
	    } else {
	       p++;
	       if (!*p)
		 SVR->botname[l] = 'a' + random() % 26;
	       else
		 SVR->botname[l] = (*p);
	    }
	 } else {
	    if (l+1 == NICKMAX + 1)
	      SVR->botname[l] = '0';
	    else {
	       SVR->botname[++l] = '0';
	       SVR->botname[++l] = 0;
	    }
	 }
      }
      G->LOG_MISC.Put( SVR->Lang( 0x048, "s", SVR->botname));
   }
   SVR->SendNick(SVR->botname);
   return EXECUTED;
}

/* 437 : nickname juped (IRCnet) */
static BindReturn got437 (void ** args, const char *) {
   FUNCTION;
   char *s, buf[512], *msg = buf;
   Channel *chan;
   int lang = 0;
   
   strcpy(msg,(char *)args[2]);
   G->Split(&msg);
   s = G->Split(&msg);
   if (strchr("#&+", s[0])) {
      if ((chan = G->chanset->Find(s))) {
	if (chan->Active()) {
	   lang = 0x049;
	   got433(NULL, NULL);
	} else
	   lang = 0x04A;
      }
   } else {
      lang = 0x04B;
      got433(NULL, NULL);
   }
   if (lang)
     G->LOG_MISC.Put( SVR->Lang( lang, "s", s));
   return EXECUTED;
}

/* 451 : Not registered */
static BindReturn got451 (void ** args, const char *) {
   FUNCTION;
   G->LOG_MISC.Put( SVR->Lang( 0x04C, "s", args[0]));
   SVR->NukeServer(SVR->Lang( 0x04D, ""));
   return EXECUTED;
}

/* error notice */
static BindReturn goterror (void ** args, const char *) {
   FUNCTION;
   char buf[512], *msg = buf;
   LogModes x ( "ms" );
   
   strcpy(msg, (char *)args[2]);
   msg = FixColon(msg);
   x.Put(SVR->Lang( 0x04E, "s", msg));
   if (SVR->serverror_quit)
     SVR->NukeServer(SVR->Lang( 0x04F, ""));
   return EXECUTED;
}

/* nick change */
static BindReturn gotnick (void ** args, const char *) {
   FUNCTION;
   char * nick, *msg = (char *)args[2], buf[512], *from = buf;
   
   strcpy(from, (char *)args[0]);
   nick = G->SplitNick(&from);
   if (*msg == ':') 
     msg++;
   if (SVR->MatchMyNick(nick)) {
      LogModes x ( "os" ); 
      /* regained nick! */
      strncpy(SVR->botname, msg, NICKMAX);
      SVR->botname[NICKMAX] = 0;
      SVR->waiting_for_awake = 0;
      x.Put(SVR->Lang( SVR->newbotname[0] ? 0x050 : 0x051, "s", msg));
      SVR->newbotname[0] = 0;
   }
   return EXECUTED;
}

static BindReturn gotmode (void ** args, const char *) {
   FUNCTION;
   char *ch, buf[512], *msg = buf;
   
   strcpy(msg, (char *)args[2]);
   ch = G->Split(&msg);
   /* usermode changes? */
   if ((ch[0] != '#') && (ch[0] != '&') && SVR->MatchMyNick(ch) && SVR->check_mode_r) {
      msg = FixColon(msg);
      if ((msg[0] == '+') && strchr(msg, 'r')) {
	 Socket * servidx = G->socklist->Find(SVR->serv);
	 LogModes x ( "jo" );
	 
	 x.Put(SVR->Lang( 0x052, "s", servidx->Host()));
	 SVR->NukeServer(SVR->Lang( 0x053, ""));
      }
   }
   return EXECUTED;
}

static BindReturn gotping (void ** args, const char *) {
   FUNCTION;
   char buf[512], *msg = buf, workbuf[512];
   
   strcpy(msg, (char *)args[2]);
   msg = FixColon(msg);
   SVR->QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511, "PONG :%s\n", msg));
   return EXECUTED;
}

/***************************************************************************
 * SERVER CONNECTION HANDLING                                              *
 ***************************************************************************/

#define RAW_COMMANDS 20
static command_binds my_raw_binds [RAW_COMMANDS] = {
   {"PRIVMSG", "", gotmsg, NULL} ,
   {"NOTICE", "", gotnotice, NULL}, 
   {"MODE", "", gotmode, NULL},
   {"PING", "", gotping, NULL},
   {"PONG", "", gotpong, NULL},
   {"WALLOPS", "", gotwall, NULL},
   {"001", "", got001, NULL},
   {"014", "", got014, NULL},
   {"206", "", trace_fail, NULL},
   {"251", "", got251, NULL},
   {"303", "", got303, NULL},
   {"401", "", trace_fail, NULL},
   {"402", "", trace_fail, NULL},
   {"432", "", got432, NULL},
   {"433", "", got433, NULL},
   {"437", "", got437, NULL},
   {"451", "", got451, NULL},
   {"453", "", got433, NULL},
   {"NICK", "", gotnick, NULL},
   {"ERROR", "", goterror, NULL},
};

/*************************************************************************
 * DCC COMMANDS                                                          *
 *************************************************************************/

static BindReturn cmd_servers (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   
   G->LOG_CMDS.Put( SVR->Lang(0x060, "s", idx->Nick()));
   SVR->serverlist->Dump(idx);
   return EXECUTED;
}

static BindReturn cmd_dump (void ** args, const char *) {
   FUNCTION;
   char buf[512];
   int l;
   Socket * idx = (Socket *)args[0];
   
   if (!((char *)args[1])[0]) 
     idx->Send(SVR->Lang(0x065, ""));
   else {
      G->LOG_CMDS.Put( SVR->Lang(0x066, "ss", idx->Nick(), args[1]));
      l = G->SPrintF(buf, 511, "%s\n", args[1]);
      SVR->QueueServer(Q_SERVER, buf, l);
   }
   return EXECUTED;
}

static BindReturn cmd_jump (void ** args, const char *) {
   FUNCTION;
   int port;
   Socket * idx = (Socket *)args[0];
   char * other, * par = (char *)args[1];

   if (par[0]) {
      other = G->Split(&par);
      if (!(port = atoi(G->Split(&par))))
	port = SVR->default_port;
      G->LOG_CMDS.Put(SVR->Lang(0x067,"ssds",idx->Nick(),other,port,par));
      strncpyz(SVR->newserver, other,120);
      SVR->newserverport = port;
      strncpyz(SVR->newserverpass, par,120);
   } else
     G->LOG_CMDS.Put(SVR->Lang(0x068,"s",idx->Nick()));
   idx->Send(SVR->Lang(0x069, ""));
   SVR->cycle_time = 0;
   SVR->NukeServer(SVR->Lang(0x100, ""));
   return EXECUTED;
}

#define DCC_COMMANDS 3
static command_binds my_dcc_cmds[DCC_COMMANDS]={
  { "dump", "m", cmd_dump, NULL },
  { "jump", "m", cmd_jump, NULL },
  { "servers", "o", cmd_servers, NULL },
};

/* this overrides the default die, handling extra server stuff */
static BindReturn server_die (void ** args, const char *) {
   FUNCTION;
   char * message = (char *)args[0];
   
   SVR->cycle_time = 100;
   SVR->NukeServer(message);
   return EXECUTED;
}

static command_binds my_die[]={
   { "*", "", server_die, "server:die"},
};

/************************************************************************
 * MISC MODULE STUFF                                                    *
 ************************************************************************/

/* this only handles CHAT requests, otherwise it's handled in 
 * filesys */
static BindReturn ctcp_DCC_CHAT(void ** args, const char *) {
   FUNCTION;
   char * nick = (char *)args[0], *from = (char *)args[1];
   char * param, * ip, * prt,buf[512],*msg = buf;
   int lang = 0;
   UserRecord * u = (UserRecord *)args[2];
   FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL
		  | FlagRecord::ANYWHERE);
   struct in_addr in;
   DccChatPass * newsock;
   
   context;
   if (strncasecmp((char *)args[5],"CHAT ",5) || !u)
     return NOMATCH;
   strcpy(msg,((char *)args[5])+5);
   fr.Set(u);
   param = G->Split(&msg);
   if (G->dcc_total == G->max_dcc) 
     lang = 0x071;
   else if (!fr.GlobParty() && !(!G->require_p && fr.ChanOp())) {
      if (fr.GlobXfer())
	return NOMATCH; /* allow filesys to pick up the chat */
      lang = 0x073;
   } else if (u->PassMatch("-"))
	lang = 0x075;
   else {
      ip = G->Split(&msg);
      prt = G->Split(&msg);
      in.s_addr = strtoul(ip, NULL, 10);
      if (!(newsock = NEW ( DccChatPass (in, atoi(prt)) ))) {
	 param = strerror(errno);
	 lang = 0x077;
      } else {
	 newsock->SetNick(u->Handle());
	 newsock->SetHost(from);
	 newsock->SetStatus(STAT_ECHO | (fr.GlobParty() ? STAT_PARTY : 0));
	 newsock->TimeNow();
	 newsock->SetUser(u);
	 lang = 0x079;
	 newsock->Send(G->Lang(0x10A, ""));
      }
   }
   if (lang) {
      G->LOG_MISC.Put( SVR->Lang(lang, "ssss", "CHAT", param, nick, from));
      if (!SVR->quiet_reject && (lang != 0x079))
	SVR->SendNotice(Q_HELP, nick, SVR->Lang(lang+1, "s", param));
   }
   return EXECUTED;
}
 
static void server_secondly() {
   FUNCTION;
   ServerList * x = SVR->serverlist;

   if (!x->Next())
      return;
   if (SVR->cycle_time)
      SVR->cycle_time--;
   SVR->DeqMsg();
   if (SVR->serv < 0)
      SVR->serv_sock->Connect();
}

static void minutely_checks() {
   FUNCTION;
   char workbuf[512];
   if (SVR->keepnick && strncmp(SVR->botname, G->origbotname, strlen(SVR->botname)))
     SVR->QueueServer(Q_MODE, workbuf, G->SPrintF(workbuf, 511,
						  SVR->use_ison ? "ISON :%s\n"
						  : "TRACE %s\n", G->origbotname));
}

static void server_5minutely() {
   FUNCTION;
   SVR->FiveMins();
}

static void server_prerehash () {
   FUNCTION;
   strcpy(SVR->oldnick, SVR->botname);
}

static void server_postrehash () {
   FUNCTION;
   SVR->PostRehash();
}

#define CTCP_COMMANDS 1
static command_binds my_ctcps[CTCP_COMMANDS] = {
   { "DCC", "", ctcp_DCC_CHAT, "server:DCC" }
};

char *server_start (ModuleEntryEggdrop * G, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   ::G = G;
   NEW ( ModuleEntryServer ( me, G, &r ) );
   if (!r) {
      context;
      SVR->bt_raw->AddBinds(my_raw_binds,RAW_COMMANDS,SVR);
      G->bt_dcc->AddBinds(my_dcc_cmds,DCC_COMMANDS,SVR);
      G->bt_die->AddBinds(my_die,1,SVR);
      SVR->bt_ctcp->AddBinds(my_ctcps,CTCP_COMMANDS,SVR);
      SVR->bt_msg->AddBinds(my_msg,MSG_COMMANDS,SVR);
      context;
      NEW ( Command ( "jump" , script_jump , SVR ) );
      NEW ( Command ( "putserv", script_putserv , SVR ) );
      NEW ( Command ( "puthelp", script_putserv , SVR ) );
      NEW ( Command ( "dumpfile", script_dumpfile , SVR ) );
      context;
      G->hook_secondly->Add(server_secondly,SVR);
      G->hook_minutely->Add(minutely_checks,SVR);
      G->hook_5minutely->Add(server_5minutely,SVR);
      G->hook_pre_rehash->Add(server_prerehash,SVR);
      G->hook_rehash->Add(server_postrehash,SVR);
      context;
   }
   return r;
}
   
