/*
 * ModuleEntryShare.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryShare.h"
#include "IntegerVariable.h"
#include "ShareQueue.h"
#include "HelpReference.h"
#include "ShareBot.h"
#include "UserRecord.h"
#include "ChannelRecord.h"
#include "Channel2.h"
#include "FileSend.h"
#include "BanRecord.h"
#include "ChannelRecord.h"
#include "UserRecord.h"
#define CLASS "ModuleEntryShare"

/* minimum version I will share with */
static Version share_ver, required_transfer, required_botnet, required_channels;
static struct patches share_patches [] = {
#include "patch.h"
};

ModuleEntryShare::ModuleEntryShare ( ModuleEntry * old, char ** r ) 
: ModuleEntry("share", share_ver.Set(2,0), share_patches, old) {
   FUNCTION;
   SH = this;
   if (!(TRANSFER = (class ModuleEntryTransfer *)
	 Depend("transfer", required_transfer.Set(2,0))))
     *r = "You need the transfer module to use userfile sharing.";
   else if (!(CHANNELS = (class ModuleEntryChannels *)
	      Depend("channels", required_channels.Set(2,0))))
     *r = "You need the channels module it use userfile sharing";
   else if (!(BOTNET = (class ModuleEntryBotnet *)
	      Depend("botnet", required_botnet.Set(2,0))))
     *r = "You need the botnet module it use userfile sharing";
   else {
      memtype_resync = NEW ( MemType ("resyncbuffers", (List **)&null_entry, this) );
      null_entry = NEW ( ShareQueue() );
      NEW ( HelpReference("share.help", this) );
      SHAREBOT = "SHAREBOT";
      allow_resync = 0;
      resync_time = 900;
      NEW ( IntegerVariable ("allow_resync", &allow_resync, 0, this ) );
      NEW ( IntegerVariable ("resync_time", &resync_time, 0, this ) );
   }
}

ModuleEntryShare::~ModuleEntryShare ( void ) {
   FUNCTION;
   ShareBot * ptr;
   LogModes x ( LogModes::MISC | LogModes::BOTS );
   
   context;
   x.Put(Lang( 0x052, ""));
   x.Put(Lang( 0x053, ""));
   while ((ptr = ((ShareBot *)G->socklist)->Next())) {
      ptr->Printf("se%s\n", Lang( 0x054, ""));
      ptr->CancelUserXfer();
      NEW ( Bot ( BOTNET->BOT, *ptr ) );
      delete ptr;
   }
   G->shareout = 0;
   DELETE ( null_entry );
}

int ModuleEntryShare::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ModuleEntryShare) + ModuleEntry::ExpMem ( );
}

void ModuleEntryShare::QueueResync (char * s, Channel2 * chan) {
   FUNCTION;
   FlagRecord fr (FlagRecord::BOT | FlagRecord::CHANNEL);
   ShareBot * ptr = (ShareBot *)G->socklist;
   
   while ((ptr = ptr->Next())) 
     if (ptr->queue) {
	if (chan) 
	  fr.Set(ptr->User(), chan->Key());
	if (!chan || fr.ChanShare() || fr.BotGlobal()) 
	  NEW ( ShareQueue ( *ptr->queue, chan, s) );
     }
}

void ModuleEntryShare::Report (Socket * idx, int details) {
   FUNCTION;
   ShareBot * ptr = (ShareBot *)G->socklist;
   FileSend * j;
   int off = 0;
   char s[512];
   
   if (details) {
      idx->Send(Lang( 0x048, "d", null_entry->ExpMem()));
      idx->Send(Lang( 0x049, "s", Lang( allow_resync ? 0x04A : 0x04B, "")));
   }
   while ((ptr = ptr->Next())) {
      if (details) {
	 if (ptr->HasShareStatus(ShareBot::GETTING)) {
	    if ((j = (FileSend *)G->socklist->Find(idx->Nick(),DCT_FILETRAN))) 
	      idx->Send(Lang(0x04C, "sF", ptr->Nick(),
			     (int) (100.0 * ((float) j->Status()) /
				    ((float) j->Length()))));
	    else 
	      idx->Send(Lang(0x04D, "s", ptr->Nick()));
	 } else if (ptr->HasShareStatus(ShareBot::SENDING)) {
	    if ((j = (FileSend *)G->socklist->Find(idx->Nick(),DCT_FILETRAN))) {
	       if (j->Flags() & Socket::CONNECTING) 
		 idx->Send(Lang( 0x04F, "s", ptr->Nick()));
	       else 
		 idx->Send(Lang( 0x04E, "sF", ptr->Nick(),
				(int) (100.0 * ((float) j->Status()) /
				       ((float) j->Length()))));
	    }
	 } else if (ptr->HasShareStatus(ShareBot::AGGRESSIVE))
	      idx->Send(Lang( 0x050, "s", ptr->Nick()));
	 else 
	   idx->Send(Lang( 0x051, "s", ptr->Nick()));
      }
      if (ptr->queue && (off < (110 - HANDLEN))) {
	 off += G->StrCpy(s + off, ptr->Nick());
	 off += G->SPrintF(s + off, 511 - off, " (%d), ", ptr->queue->Count());
      }      
   }
   if (off) {
      s[off - 2] = 0;
      idx->Send(Lang(0x038, "s", s));
   }
}
