/*
 * ShareBot.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ShareBot.h"
#include "BotLink.h"
#include "ShareQueue.h"
#include "Channel2.h"
#include "UserRecord.h"
#include "FileGetListen.h"
#include "TandemBot.h"
#include "UserEntryBotaddr.h"
#include "UserEntryHosts.h"
#include "ChannelRecord.h"
#include "FileSend.h"
#include "IgnoreRecord.h"
#define CLASS "ShareBot"

ShareBot::ShareBot ( Bot * old) : Bot ( SH->SHAREBOT, *old ) {
   FUNCTION;
   Printf("su\n");
   SetShareStatus(ShareBot::OFFERED);
}

ShareBot::ShareBot ( BotLink & old, char * x ) : Bot ( old, x ) {
   FUNCTION;
   SetType(SH->SHAREBOT);
   delete & old;
}

void ShareBot::Display2 (Socket * where) {
   FUNCTION;
   char buf[20];
   int i = 0;

   buf[i++] = HasStatus(PINGED) ? 'P' : 'p';
   buf[i++] = HasStatus(CALLED) ? 'C' : 'c';
   buf[i++] = HasStatus(WARNED) ? 'W' : 'w';
   buf[i++] = HasStatus(LEAF) ? 'L' : 'l';
   buf[i++] = HasStatus(LINKING) ? 'I' : 'i';
   buf[i++] = HasShareStatus(OFFERED) ? 'O' : 'o';
   buf[i++] = HasShareStatus(SENDING) ? 'S' : 's';
   buf[i++] = HasShareStatus(GETTING) ? 'G' : 'g';
   buf[i++] = HasShareStatus(AGGRESSIVE) ? 'a' : 'A';
   buf[i++] = 0;
   where->Send(SH->Lang(0x02D,"sss",nick,version,buf));
}
   
ShareBot::~ShareBot ( void ) {
   FUNCTION;
   CancelUserXfer();
   if (queue) {
      DELETE ( queue );
      queue = 0;
   }
}

int ShareBot::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ShareBot);
}

void ShareBot::ShareOutBut(Channel2 * chan, char * format, ...) {
   FUNCTION;
   int l;
   va_list va;
   char s[601];
   ShareBot * ptr = (ShareBot *)G->socklist;
   FlagRecord fr (FlagRecord::CHANNEL | FlagRecord::BOT);
   
   va_start(va, format);
   s[0] = 's';
#ifdef HAVE_VSNPRINTF
   if ((l = vsnprintf(s + 1, 510, format, va))<0)
     s[1 + (l = 510)] = 0;   
#else
   l = vsprintf(s+2, format, va);
#endif
   while ((ptr = ptr->Next()))
     if ((ptr != this) && !ptr->HasShareStatus(ShareBot::GETTING
					      | ShareBot::SENDING)) {
	if (chan) 
	   fr.Set(ptr->User(),chan->Key());
	if (!chan || fr.ChanShare() || fr.BotGlobal())
	  ptr->Dump(s, l + 1);
     }
   SH->QueueResync(s,chan);
   va_end(va);
}

void ShareBot::NewResync ( void ) {
   FUNCTION;
   if (queue)
     DELETE ( queue );
   queue = NEW ( ShareQueue );
   queue_age = G->now;
   G->LOG_MISC.Put( SH->Lang( 0x036, "s", nick));
}

int ShareBot::FlushResync (void) {
   FUNCTION;
   if (queue) {
      DELETE ( queue );
      queue = 0;
      return 1;
   }
   return 0;
}

void ShareBot::QueueResync ( Channel2 * chan, char * s ) {
   FUNCTION;
   FlagRecord fr (FlagRecord::BOT | FlagRecord::CHANNEL);

   if (chan) 
     fr.Set(user, chan->Key());
   if (!chan || fr.ChanShare() || fr.BotGlobal()) 
     NEW ( ShareQueue ( *queue, chan, s ) );
}

void ShareBot::DumpResync (void) {
   FUNCTION;
   if (queue) {
      queue->Flush(-1, this);
      DELETE ( queue );
      queue = 0;
   }
}

void ShareBot::StartSendingUsers (void) {
   FUNCTION;
   char s[DIRLEN+1], s1[64], *s2;
   int i = 1;
   ChannelRecord *ch;
   Channel *cst;
   FileGetListen * ptr;
   TandemBot * tb;
   struct in_addr in;
   FlagRecord botfr (FlagRecord::GLOBAL | FlagRecord::BOT);
   FlagRecord bfl (FlagRecord::BOT);
   FlagRecord fr (FlagRecord::CHANNEL);
   FlagRecord fl;
   struct stat ss;
   UserRecord * u;
   
   context;
   G->SPrintF(s, DIRLEN, ".share.%s.%lu", nick, G->now);
   WriteTmpUserfile(s, bfl.Set(user)->BotShareflags());
   if ((i = SH->TRANSFER->RawDccSend(s, nick, "*users")) < 0) {
      unlink(s);
      s2 = SH->Lang( (i == -1) ? 0x041 : (i == -2) ? 0x042 : 0x043, "");
      Printf("se%s\n",s2);
      G->LOG_MISC.Put(s2);
      NEW ( Bot ( SH->BOTNET->BOT, *this ) );
   } else {
      if ((tb = SH->BOTNET->tandbot->Find(nick)))
	tb->Update(0,'+',0);
      if (!(ptr = (FileGetListen *)G->socklist->Find(i,SH->TRANSFER->FILE_LISTEN)))
	G->Fatal("Lost Share Send file!!");
      SetShareStatus(ShareBot::SENDING);
      in.s_addr = ptr->Address().s_addr;
      stat (s, &ss);
      Printf("sS%lu %d %lu\n", iptolong(&in), ptr->LocalPort(), ss.st_size);
      /* start up a tbuf to queue outgoing changes for this bot until the */
      /* userlist is done transferring */
      NewResync();
      /* immediately, queue bot hostmasks & addresses (jump-start) */
      u = G->userlist;
      while ((u = u->Next()))
	if (botfr.Set(u)->GlobBot() && !botfr.GlobUnshared()) {
	   UserEntryBotaddr * bi = 
	     (UserEntryBotaddr *)u->FindEntry(SH->BOTNET->entrytype_botaddr);
	   UserEntryHosts * t = (UserEntryHosts *)u->FindEntry(G->entrytype_hosts);
	   KeyedList<UHOSTLEN> * l;
	   
	   /* send hostmasks */
	   if (t && (l = t->First()))
	     while ((l = l->Next())) {
		G->SPrintF(s, DIRLEN, "s#%s %s\n", u->Handle(), l->Key());
		QueueResync(0, s);
	     }
	   /* send address */
	   if (bi) {
	      G->SPrintF(s, DIRLEN, "scBOTADDR %s %s %d %d\n", u->Handle(), 
			 bi->Value(), bi->Telnet(), bi->Relay());
	      QueueResync(0, s);
	   }
	   /* send user-flags */
	   if (bfl.BotShareflags()) {
	      fl.Set(u);
	      fl.Build(s1);
	      G->SPrintF(s, DIRLEN, "sa%s %s\n", u->Handle(), s1);
	      QueueResync(0, s);
	   }
	   ch = u->ChanRecs();
	   while ((ch = ch->Next())) {
	      if ((cst = FindChan(ch->Key()))) {
		 if (!bfl.BotGlobal())
		   fr.Set(user,ch->Key());
		 if (bfl.BotGlobal() || fr.ChanShare()) {
		    fr.Set(u, ch->Key());
		    fr.ChanMaskShare();
		    fr.Build(s1);
		    G->SPrintF(s, DIRLEN, "sa%s %s %s\n", u->Handle(), s1,
			       ch->Key());
		    QueueResync(0, s);
		 }
	      }
	   }
	}
      QueueResync(0, "s!\n");
      /* wish could unlink the file here to avoid possibly leaving it lying */
      /* around, but that messes up NFS clients. */
      context;
   }
}

void ShareBot::CancelUserXfer (void) {
   FUNCTION;
   FileSend * ptr = (FileSend *)G->socklist;
   TandemBot * tb;
   
   context;
   if ((tb = SH->BOTNET->tandbot->Find(nick)))
     tb->Update(0,'-',0);
   FlushResync();
   if (HasShareStatus(ShareBot::GETTING | ShareBot::SENDING)) {
      while ((ptr = ptr->Next()))
	if (!strcasecmp(ptr->Host(), nick)) {
	   unlink(ptr->Filename());
	   ptr->Close();
	   break;
	}
      if (ptr)
	SH->BOTNET->LOG_BOTS.Put( SH->Lang( 0x044, "s", ptr->Nick()));
   }
   if (SH->allow_resync && !HasShareStatus(ShareBot::GETTING|ShareBot::SENDING))
     NewResync();
}

int ShareBot::WriteTmpUserfile (char * fn, int f1) {
   FUNCTION;
   FILE *f;
   UserRecord *u = G->userlist;
   int ok = 0;
   
   if ((f = fopen(fn, "w"))) {
      chmod(fn, 0600);		/* make it -rw------- */
      fprintf(f, "%s\n", SH->Lang( 0x039, "css", '5', G->VersionStr(),
				  G->origbotname));
      ok = 1;
      while ((u = u->Next()) && ok) 
	ok = u->Write(f, this, f1);
      if (ok)
	ok = G->ignores->Write(f); 
      if (ok)
	ok = SH->CHANNELS->WriteBans(f,this);
      fclose(f);
   }   
   if (!ok) 
     G->LOG_MISC.Put( SH->Lang( 0x03A, ""));
   return ok;
}	   
