/*
 * share.c - part of share module, provides userfile sharing between bots
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Socket.h"
#include "UserRecord.h"
#include "UserEntryPass.h"
#include "UserEntryHosts.h"
#include "BindTable.h"
#include "BanRecord.h"
#include "HelpReference.h"
#include "HookList.h"
#include "channels.h"
#include "botnet.h"
#include "TandemBot.h"
#include "UserEntryBotaddr.h"
#include "ModuleEntryShare.h"
#include "ModuleEntryTransfer.h"
#include "ModuleEntryChannels.h"
#include "BotLink.h"
#include "FileSend.h"
#include "FileGetListen.h"
#include "ScriptContext.h"
#include "Channel2.h"
#include "ChannelRecord.h"
#include "ShareBot.h"
#include "Command.h"
#ifdef HAVE_IRC_H
#include "irc.h"
#include "Channel3.h"
#endif
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#define CLASS ""

static const int min_share = 2000000;
#ifdef HAVE_IRC_H
static Version required_irc ( 2, 0 );
#endif

ModuleEntryShare * SH;
static ModuleEntryEggdrop * G;

/* botnet commands */

static void share_stick (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *host, *val;
   int yn, lang;

   host = G->Split(&par);
   val = G->Split(&par);
   lang = (yn = atoi(val)) ? 0x001 : 0x004;
   G->noshare = 1;
   if (!par[0]) {		/* global ban */
      if (((Channel2 *)G->chanset)->SetStickyBan(host, yn) > 0) {
	 G->LOG_CMDS.Put( SH->Lang( lang, "ss", idx->Nick(), host));
	 idx->ShareOutBut(0,"s%s %d\n", host, yn);
      }
   } else {
      Channel2 *chan = FindChan(par);
      FlagRecord flags (FlagRecord::CHANNEL | FlagRecord::GLOBAL);
      
      if (chan && (flags.Set(idx->User(),chan->Key())->BotGlobal()
		   || flags.ChanShare())
	  && (chan->SetStickyBan(host, yn) > 0)) {
	 G->LOG_CMDS.Put( SH->Lang( lang+1, "sss", idx->Nick(), host, par));
	 idx->ShareOutBut(chan,"s%s %d %s\n", host, yn, chan->Key());
      } else 
	G->LOG_CMDS.Put(SH->Lang( lang+2, "sss", idx->Nick(), host, par));
   }
   G->noshare = 0;
}

static void share_chhand (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *hand;
   UserRecord * u;
   FlagRecord fr;
   
   hand = G->Split(&par);
   u = G->userlist->Find(hand);
   if (u && !fr.Set(u)->GlobUnshared()) {
      idx->ShareOutBut(0, "h%s %s\n", hand, par);
      G->noshare = 1;
      if (u->ChangeHandle(par))
	G->LOG_CMDS.Put( SH->Lang( 0x007, "sss", idx->Nick(), hand, par));
      G->noshare = 0;
   }
}

static void share_chattr (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *hand, *atr, s[100];
   Channel2 *cst;
#ifdef HAVE_IRC_H
   Channel3 * c1 = (Channel3 *)G->chanset;
#endif
   UserRecord * u;
   FlagRecord fr2 (FlagRecord::GLOBAL | FlagRecord::CHANNEL | FlagRecord::BOT);
   FlagRecord fr (FlagRecord::CHANNEL);
   
   hand = G->Split(&par);
   u = G->userlist->Find(hand);
   fr2.Set(u);
   if (u && !fr2.GlobUnshared() && !(fr2.GlobBot() && fr2.BotShare())) {
      atr = G->Split(&par);
      G->noshare = 1;
      if (!par[0] && fr.BotShareflags()) {
	 fr.SetMatch(FlagRecord::GLOBAL);
	 fr.BreakDown(atr);
	 if (fr2.GlobBot())
	   fr.SetGlobBot();
	 else
	   fr.ClearGlobBot();
	 fr.Sanity();
	 u->Set(fr);
	 fr.Build(s);
	 fr.SetMatch(FlagRecord::CHANNEL);
	 if (!idx->HasShareStatus(ShareBot::GETTING)) {
	    G->LOG_CMDS.Put( SH->Lang( 0x00A, "sss", idx->Nick(), hand, s));
	    idx->ShareOutBut(0,"a%s %s\n", hand, s);
	 }
#ifdef HAVE_IRC_H
	 if (SH->IRC)
	   while ((c1 = c1->Next()))
	     c1->Recheck(0);
#endif
      } else if (!par[0]) 
	   G->LOG_CMDS.Put( SH->Lang( 0x055, "ss", idx->Nick(), hand));	    
      else {
	 fr2.SetMatch(FlagRecord::CHANNEL);
	 if ((cst = FindChan(par)) && !fr2.BotGlobal())
	   fr2.Set(idx->User(),par);
	 if (cst && (fr2.BotGlobal() || fr2.ChanShare())) {
	    fr2.Set(u,par);
	    fr.BreakDown(atr);
	    if (fr2.ChanShare())
	      fr.SetChanShare();
	    else
	      fr.ClearChanShare();
	    fr.Sanity();
	    u->Set(fr,par);
	    fr.Build(s);
	    if (!idx->HasShareStatus(ShareBot::GETTING)) {
	       G->LOG_CMDS.Put( SH->Lang( 0x008, "ssss", idx->Nick(), hand, 
					 s, par));
	       idx->ShareOutBut(cst,"a%s %s %s\n", hand, s, par);
	    }
#ifdef HAVE_IRC_H
	    if (SH->IRC) 
	      ((Channel3 *)cst)->Recheck(0);
#endif
	 } else
	   G->LOG_CMDS.Put( SH->Lang( 0x009, "sss", idx->Nick(), par, hand));
      }
      G->noshare = 0;
   }
}

static void share_chrec (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *user;
   Channel2 * chan;
   UserRecord * u;
   int pls = code == 'l' ? 1 : 0;
   FlagRecord fr (FlagRecord::CHANNEL), fl, bfl(FlagRecord::BOT);

   user = G->Split(&par);
   if ((u = G->userlist->Find(user)) && !fl.Set(u)->GlobUnshared()) {
      chan = FindChan(par);
      fr.Set(idx->User(),par);
      if (!chan || (!fr.ChanShare() && !bfl.Set(idx->User())->BotGlobal())) 
	G->LOG_CMDS.Put( SH->Lang( 0x00B + pls, "sss", idx->Nick(), par, user));
      else {
	 G->noshare = 1;
	 idx->ShareOutBut(chan,"%c%s %s\n", code, user, par);
	 if (pls) 
	   SH->CHANNELS->DelChanrec(u,par);
	 else if (!pls && !u->ChanRecs()->Find(par))
	   SH->CHANNELS->AddChanrec(u,par);
	 G->LOG_CMDS.Put( SH->Lang( 0x00D + pls, "sss", idx->Nick(), user, par));
	 G->noshare = 0;
      }
   }
}

static void share_newuser (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *etc, *etc2;
   UserRecord * u;
   FlagRecord fr ( idx->User(), FlagRecord::BOT ), fr2, fl;
   FlagRecord bfl (FlagRecord::BOT);

   etc = G->Split(&par);
   if (!(u = G->userlist->Find(etc)) || !fl.Set(u)->GlobUnshared()) {
      etc2 = G->Split(&par);
      idx->ShareOutBut(0,"n%s %s %s\n", etc, etc2, fr.BotShareflags() ? par
		   : (fr2.Set(u)->GlobBot() ? "b" : "-"));
      /* If user already exists, ignore command */
      if (!u) {
	 G->noshare = 1;
	 if (strlen(etc) > HANDLEN)
	   etc[HANDLEN] = 0;
	 fl.BreakDown(par);
	 if (!bfl.Set(idx->User())->BotShareflags()) {
	    if (fl.GlobBot()) {
	       fl.Clear();
	       fl.SetGlobBot();
	    } else 
	      fl.SetGlobal(G->default_flags);
	 }
	 NEW ( UserRecord ( G->userlist, etc, etc2, fl) );
	 G->noshare = 0;
	 G->LOG_CMDS.Put( SH->Lang( 0x00F, "ss", idx->Nick(), etc));
      }
   }
}

static void share_killuser (ShareBot * idx, char * par, char code) {
   FUNCTION;
   UserRecord * u = 0;
   FlagRecord fr ( FlagRecord::BOT | FlagRecord::GLOBAL );
   
   /* If user is a share bot, ignore command */
   if ((u = G->userlist->Find(par)) && !fr.Set(u)->GlobUnshared() 
       && !(fr.GlobBot() && fr.BotShare())) {
      G->noshare = 1;
      DELETE (u);
      idx->ShareOutBut(0,"k%s\n", par);
      G->LOG_CMDS.Put( SH->Lang( 0x010, "ss", idx->Nick(), par));
      G->noshare = 0;
   }
}

static void share_host (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *hand;
   UserRecord * u;
   int pls = (code == '+') ? 1 : 0;
   FlagRecord fr ( FlagRecord::BOT | FlagRecord::GLOBAL );
   
   hand = G->Split(&par);
   if ((u = G->userlist->Find(hand)) && !fr.Set(u)->GlobUnshared() &&
       !(fr.GlobBot() && fr.BotShare())) {
      idx->ShareOutBut(0,"%c%s %s\n", code, hand, par);
      G->noshare = 1;
      if (pls)
	u->AddHost(par);
      else
	u->DelHost(par);
      G->noshare = 0;
      G->LOG_CMDS.Put( SH->Lang( 0x011 + pls, "sss", idx->Nick(), hand, par));
   }
}

static void share_pls_bothost (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *hand, p[32];
   UserRecord * u;
   FlagRecord fl;
   
   context;
   hand = G->Split(&par);
   fl.Clear();
   if (!(u = G->userlist->Find(hand)) || !fl.Set(u)->GlobUnshared()) {
      if (!idx->HasShareStatus(ShareBot::GETTING))
	idx->ShareOutBut(0,"#%s %s\n", hand, par);
      G->noshare = 1;
      if (u) {
	 if (fl.GlobBot())
	   u->AddHost(par);
      } else {
	 fl.Clear();
	 fl.SetGlobBot();
	 u = NEW ( UserRecord ( G->userlist, hand, par, fl) );
	 G->MakePass(p);
	 SetPassword(u, p);
      }
      G->noshare = 0;
      if (!idx->HasShareStatus(ShareBot::GETTING) && fl.GlobBot())
	G->LOG_CMDS.Put( SH->Lang( 0x013, "sss", idx->Nick(), hand, par));
   }
}

static void share_change (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char * key, *hand;
   UserRecord * u;
   UserEntry * uet;
   FlagRecord fl;
   
   key = G->Split(&par);
   hand = G->Split(&par);
   if (!(u = G->userlist->Find(hand)) || !fl.Set(u)->GlobUnshared()) {
      if (!(uet = G->entrytype_list->Find(key))) 
	G->LOG_CMDS.Put( SH->Lang( 0x014, "sss", key, idx->Nick(), hand));
      else {
	 if (!idx->HasShareStatus(ShareBot::GETTING))
	   idx->ShareOutBut(0,"c%s %s %s\n",key, hand, par);
	 G->noshare = 1;
	 if (!u && (uet == SH->BOTNET->entrytype_botaddr)) {
	    char pass[30];
	    
	    fl.Clear();
	    fl.SetGlobBot();
	    u = NEW ( UserRecord ( G->userlist, hand, "none", fl) );
	    G->MakePass(pass);
	    SetPassword(u, pass);
	 } else if (!u)
	      return;
	 u->GetEntry(uet)->GotShare(par, idx);
	 G->noshare = 0;
      }
   }
}

static void share_chchinfo (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *hand, *chan;
   Channel2 *cst;
   UserRecord * u;
   FlagRecord fr (FlagRecord::BOT | FlagRecord::CHANNEL ), fl;
   
   hand = G->Split(&par);
   if ((u = G->userlist->Find(hand)) && !fl.Set(u)->GlobUnshared()
       && G->share_greet) {
      chan = G->Split(&par);
      cst = FindChan(chan);
      fr.Set(idx->User(),chan);
      if (!cst || (!fr.ChanShare() && fr.BotGlobal()))
	G->LOG_CMDS.Put( SH->Lang( 0x015, "sss", idx->Nick(), chan, hand));
      else {
	 idx->ShareOutBut(cst,"C%s %s %s\n", hand, chan, par);
	 G->noshare = 1;
	 SH->CHANNELS->SetHandleChaninfo(G->userlist, hand, chan, par);
	 G->noshare = 0;
	 G->LOG_CMDS.Put( SH->Lang( 0x016, "sss", idx->Nick(), chan, hand));
      }
   }
}

static void share_mns_ban (ShareBot * idx, char * par, char code) {
   FUNCTION;
   char *chname;
   Channel2 *chan = NULL;
   FlagRecord fr (FlagRecord::BOT | FlagRecord::CHANNEL );
   
   chname = G->Split(&par);
   if ((chname[0] != '*') && (chan = FindChan(chname))) 
     fr.Set(idx->User(),chname);
   if (!chan || fr.ChanShare() || fr.BotGlobal()) {
      idx->ShareOutBut(chan,"B%s %s\n", chname, par);
      G->LOG_CMDS.Put( SH->Lang( chan ? 0x017 : 0x018, "sss", idx->Nick(),
				par, chname));
      G->noshare = 1;
      if (!chan)
	chan = (Channel2 *)G->chanset;
      chan->DelBan(par, 1);
      G->noshare = 0;
   }
}

static void share_mns_ignore (ShareBot * idx, char * par, char code) {
   FUNCTION;
   IgnoreRecord * irec;
   
   idx->ShareOutBut(0,"I%s\n", par);
   G->LOG_CMDS.Put( SH->Lang( 0x019, "ss", idx->Nick(), par));
   G->noshare = 1;
   if ((irec = G->ignores->Find(par)))
     DELETE ( irec );
   G->noshare = 0;
}

static void share_pls_ban (ShareBot * idx, char * par, char code) {
   FUNCTION;
   time_t expire_time;
   char *ban, *tm, *from, *chname;
   int flags = 0;
   Channel2 *chan = NULL;
   FlagRecord fr (FlagRecord::BOT | FlagRecord::CHANNEL );
   
   chname = G->Split(&par);
   if ((chname[0] != '*') && (chan = FindChan(chname))) {
      fr.Set(idx->User(),chname);
      if (!fr.ChanShare() && !fr.BotGlobal()) {
	 ban = G->Split(&par);
	 G->LOG_CMDS.Put( SH->Lang( 0x01A, "sss", idx->Nick(), ban, chname));
	 return;
      }
   }
   idx->ShareOutBut(chan,"b%s %s\n", chname, par); 
   G->noshare = 1;
   ban = G->Split(&par);
   tm = G->Split(&par);
   from = G->Split(&par);
   if (strchr(from,'s'))
     flags |= BanRecord::STICKY;
   if (strchr(from,'p'))
     flags |= BanRecord::PERM;
   from = G->Split(&par);
   if ((expire_time = atoi(tm)))
     expire_time += G->now;
   G->LOG_CMDS.Put( SH->Lang( chan ? 0x01B : 0x01C, "sssss", idx->Nick(),
			     ban, from, par, chname));
   if (!chan)
     chan = (Channel2 *)G->chanset;
   chan->AddBan(ban, from, par, expire_time, flags);
   G->noshare = 0;
}
      
/* +ignore <host> +<seconds-left> <from> <note> */
static void share_pls_ignore (ShareBot * idx, char * par, char code) {
   FUNCTION;
   time_t expire_time;
   char *ign, *from, *ts;
   
   idx->ShareOutBut(0,"i%s\n", par);
   G->noshare = 1;
   ign = G->Split(&par);
   ts = G->Split(&par);
   if (!atoi(ts))
     expire_time = 0L;
   else
     expire_time = G->now + atoi(ts);
   from = G->Split(&par);
   if (strchr(from,'p'))
     expire_time = 0;
   from = G->Split(&par);
   if (strlen(from) > HANDLEN+1)
     from[HANDLEN+1] = 0;
   par[65] = 0;
   G->LOG_CMDS.Put( SH->Lang( 0x01D, "ssss", idx->Nick(), ign, from, par));
   NEW ( IgnoreRecord (ign, from, par, expire_time) );
   G->noshare = 0;
}
   
static void share_ufno (ShareBot * idx, char * par, char code) {
   FUNCTION;
   SH->BOTNET->LOG_BOTS.Put( SH->Lang( 0x01E, "ss", idx->Nick(), par));
   idx->Send("se");
   NEW ( Bot ( SH->BOTNET->BOT, *idx ) );
}

static void share_ufyes (ShareBot * idx, char *, char code) {
   FUNCTION;
   context;
   if (idx->HasShareStatus(ShareBot::OFFERED)) {
      idx->ClearShareStatus(ShareBot::OFFERED);
      idx->SetShareStatus(ShareBot::SENDING);
      idx->StartSendingUsers();
      SH->BOTNET->LOG_BOTS.Put(SH->Lang( 0x01F, "s", idx->Nick()));
   }
}

static void share_userfileq (ShareBot * idx, char *, char code) {
   FUNCTION;
   FlagRecord bfl (idx->User(),FlagRecord::BOT);
   int lang = 0;
   ShareBot * ptr = (ShareBot *)G->socklist;
   
   idx->FlushResync();
   if (bfl.BotAggressive())
     lang = 0x020;
   else if (!bfl.BotPassive())
     lang = 0x021;
   else if (min_share > idx->Numver())
     idx->Printf("sU%s\n", SH->Lang( 0x022, "ddd", (min_share / 1000000),
				   (min_share / 10000) % 100, 
				   (min_share / 100) % 100));
   else {
      while ((ptr = ptr->Next()))
	if (ptr->HasShareStatus(ShareBot::AGGRESSIVE) && (ptr != idx))
	  break;
      if (ptr)
	lang = 0x023;
      else {
	 idx->Printf("sY\n");
	 /* set stat-getting to astatic void race condition (robey 23jun96) */
	 idx->SetShareStatus(ShareBot::GETTING | ShareBot::AGGRESSIVE);
	 SH->BOTNET->LOG_BOTS.Put( SH->Lang( 0x024, "s", idx->Nick()));
      }
   }
   if (lang) {
      idx->Printf("sU%s\n", SH->Lang( lang, ""));
      idx->Send("se");
      NEW ( Bot ( SH->BOTNET->BOT, *idx ) );
   }
}

/* ufsend <ip> <port> <length> */
static void share_ufsend (ShareBot * idx, char * par, char code) {
   FUNCTION;
   FILE * f;
   FileSend * ptr;
   struct in_addr in;
   int lang = 0;
   char buf[DIRLEN+1], *pt;
   
   G->SPrintF(buf, DIRLEN, ".share.%s.users", G->botnetnick);
   if (!idx->HasShareStatus(ShareBot::GETTING | ShareBot::AGGRESSIVE)
       == (ShareBot::GETTING | ShareBot::AGGRESSIVE)
       || (ptr = (FileSend *)G->socklist->Find(idx->Nick(), DCT_FILETRAN))) 
     lang = 0x025;
   else if (G->dcc_total == G->max_dcc) 
     lang = 0x027;
   else if (!(f = fopen(buf, "w")))
     lang = 0x029;
   else {
      in.s_addr = strtoul(G->Split(&par),NULL,10);
      pt = G->Split(&par);
      ptr = NEW ( FileSend (&in,  atoi(pt), "*users*", 
			    idx->User(), atoi(par), 0, buf, f));
      if (ptr->Flags() & Socket::LOST)
	lang = 0x02B;
   }
   if (lang) {
      G->LOG_MISC.Put( SH->Lang( lang, "s", idx->Nick()));
      idx->Printf("se%s\n", SH->Lang( lang+1, ""));
      NEW ( Bot ( SH->BOTNET->BOT, *idx ) );
   }
}

static void share_resyncq (ShareBot * idx, char *, char code) {
   FUNCTION;
   int lang = 0;
   TandemBot * tb;
   
   if (!SH->allow_resync)
     lang = 0x02F;
   else {
      FlagRecord bfl (idx->User());
      
      if (!bfl.BotShare())
	lang = 0x030;
      else if (idx->CanResync()) {
	 idx->Printf("sy\n");
	 idx->DumpResync();
	 idx->ClearShareStatus(ShareBot::OFFERED);
	 SH->BOTNET->LOG_BOTS.Put( SH->Lang( 0x032, "s", idx->Nick()));
	 if ((tb = SH->BOTNET->tandbot->Find(idx->Nick())))
	   tb->Update(NULL,'+',0);
      } else 
	lang = 0x031;
   }
   if (lang) {
      idx->Printf("sR%s\n", SH->Lang( lang, ""));
      idx->Send("se");
      NEW ( Bot ( SH->BOTNET->BOT, *idx ) );
   }
}

static void share_resync (ShareBot * idx, char *, char code) {
   FUNCTION;
   TandemBot * tb;
   
   if (idx->HasStatus(ShareBot::OFFERED) && idx->CanResync()) {
      idx->DumpResync();
      idx->ClearStatus(ShareBot::OFFERED);
      if ((tb = SH->BOTNET->tandbot->Find(idx->Nick())))
	tb->Update(NULL,'+',0);
      SH->BOTNET->LOG_BOTS.Put( SH->Lang( 0x032, "s", idx->Nick()));
   }
}

static void share_resync_no (ShareBot * idx, char * par, char code) {
   FUNCTION;
   SH->BOTNET->LOG_BOTS.Put( SH->Lang( 0x033, "ss", idx->Nick(), par));
   idx->FlushResync();
   idx->Printf("su\n");
   idx->Send("se");
   NEW ( Bot ( SH->BOTNET->BOT, *idx ) );
}

static void hook_read_userfile () {
   FUNCTION;
   ShareBot * ptr = (ShareBot *)G->socklist;
   
   if (!G->noshare) 
     while ((ptr = ptr->Next()))
       if (!ptr->HasShareStatus(ShareBot::AGGRESSIVE)) {
	  /* cancel any existing transfers */
	  if (ptr->HasShareStatus(ShareBot::SENDING))
	    ptr->CancelUserXfer();
	  ptr->Printf("su\n");
	  ptr->SetShareStatus(ShareBot::OFFERED);
       }
}

static void share_endstartup (ShareBot * idx, char *, char code) {
   FUNCTION;
   idx->ClearShareStatus(ShareBot::GETTING);
   hook_read_userfile();
};

static void share_end (ShareBot * idx, char * par, char code) {
   FUNCTION;
   SH->BOTNET->LOG_BOTS.Put(SH->Lang( 0x034, "ss", idx->Nick(), par));
   NEW ( Bot ( SH->BOTNET->BOT, *idx ) );
}


/* these MUST be sorted */
static struct { 
   char name; 
   void (*func) (ShareBot *, char *, char); 
} sharein_table[] = {
  { '!', share_endstartup },
  { '#', share_pls_bothost },
  { '+', share_host },
  { '-', share_host },
  { 'B', share_mns_ban },
  { 'C', share_chchinfo },
  { 'I', share_mns_ignore },
  { 'L', share_chrec },
  { 'R', share_resync_no },
  { 'S', share_ufsend },
  { 'U', share_ufno },
  { 'Y', share_ufyes },
  { 'a', share_chattr },
  { 'b', share_pls_ban },
  { 'c', share_change },
  { 'e', share_end },
  { 'h', share_chhand },
  { 'i', share_pls_ignore },
  { 'k', share_killuser },
  { 'l', share_chrec },
  { 'n', share_newuser },
  { 'r', share_resyncq },
  { 's', share_stick },
  { 'u', share_userfileq },
  { 'y', share_resync },
  { 0, 0 }
};

/* this is a cop out - it should be in ShareBot.C */
void ShareBot::Activity (char * msg, int i) {
   FUNCTION;
   char code;

   if (msg[0] == 's') {
      context;
      if (G->debug_output) {
	 LogModes LOG_BOTSHARE ( LogModes::BOTSHARE );
	 LOG_BOTSHARE.Put(SH->Lang( 0x035, "ss", nick, ++msg));
      }
      code = *msg++;
      for (i = 0; sharein_table[i].name; i++) 
	if (code == sharein_table[i].name) {
	   sharein_table[i].func(this, msg, code);
	   break;
	} else if (code < sharein_table[i].name)
	     break;
   } else
     Bot::Activity ( msg, i );
}

static BindReturn share_version (void ** args, const char *) {
   FUNCTION;
   BotLink * idx = (BotLink *)args[0];
   ShareBot *newone;
   char buffer[512], *par = buffer;
   int l;
   FlagRecord bfl(FlagRecord::BOT);
   
   strncpy(buffer, (char *)args[1], 511);
   buffer[511] = 0;
   idx->SetNumver(atoi(G->Split(&par)));
   if ((idx->Numver() >= min_share) && (bfl.Set(idx->User())->BotShare())) {
      idx->Printf("B%s %s\n", G->botnetnick, 
		  BNET->BotnetFlags((int)(G->debug_output ? Bot::TRAFWATCH : 0)));
      l = atoi(G->Split(&par));
      if (l != HANDLEN) {
	 idx->Printf("error %s\n", SH->BOTNET->Lang(0x00A, "dd", HANDLEN, l));
	 idx->Close();
      } else {
	 newone = NEW ( ShareBot ( *idx, par ) );
	 if (bfl.Set(newone->User())->BotAggressive()) {
	    newone->Printf(newone->CanResync() ? "sr\n" : "su\n");
	    newone->SetShareStatus(ShareBot::OFFERED);
	 }
      }
      return EXEC_LOG;
   }
   return NOMATCH;
}

static command_binds my_version [] = {
   { "02*", "", share_version, "share" },
};

static void shareout_mod(Channel * chan, char * format, ...) {
   FUNCTION;
   int  l;
   va_list va;
   char s[601];
   ShareBot * ptr = (ShareBot *)G->socklist;
   FlagRecord fr (FlagRecord::CHANNEL | FlagRecord::BOT);
   
   va_start(va,format);
   s[0] = 's';
#ifdef HAVE_VSNPRINTF
   if ((l = vsnprintf(s+1, 510, format, va)) < 0) 
     s[1 + (l = 510)] = 0;
#else
   l = vsprintf(s+1, format, va);
#endif
   while ((ptr = ptr->Next()))
     if (!ptr->HasShareStatus(ShareBot::GETTING|ShareBot::SENDING)) {
	if (chan)
	  fr.Set(ptr->User(),chan->Key());
	if (!chan || fr.ChanShare() || fr.BotGlobal()) 
	  ptr->Dump(s, l + 1);
     }
   SH->QueueResync(s,(Channel2 *)chan);
   va_end(va);
}

/* flush all tbufs older than 15 minutes */
static void share_minutely ( void ) {
   FUNCTION;
   Bot * ptr = (Bot *)G->socklist;
   FlagRecord bfl(FlagRecord::BOT); 
  
   while ((ptr = ptr->Next())) {
      if (ptr->Type() == SH->SHAREBOT) {
	 ShareBot * sptr = (ShareBot *)ptr;

	 if (sptr->QueueAge() > SH->resync_time) {
	    sptr->FlushResync();
	    G->LOG_MISC.Put( SH->Lang( 0x037, "s", sptr->Nick()));
	 }
      }
      if (bfl.Set(ptr->User())->BotAggressive()) {
	 if ((ptr->Type() == SH->SHAREBOT) &&
	     ((ShareBot *)ptr)->HasShareStatus(ShareBot::OFFERED) && (ptr->Age() > 120))
	   ptr->Printf("su\n");
	 /* ^ send it again in case they missed it */
	 /* if it's a share bot that hasnt been sharing, ask again */
	 else if (ptr->Type() == SH->BOTNET->BOT) 
	   NEW ( ShareBot ( ptr ) );
      }
   }
}

static BindReturn share_eof (void ** args, const char * mask) {
   FUNCTION;
   FileSend * idx = (FileSend *)args[0];
   ShareBot * bidx = (ShareBot *)G->socklist->Find(mask,DCT_BOT);
   const char * filename;
   char buf[4] = "se";
   
   if (bidx && (bidx->Type() == SH->SHAREBOT) && (bidx->Numver() > min_share)) {
      if (idx->Flags() & Socket::CONNECTING) {
	 FileGetListen * idx2 = (FileGetListen *)args[0];
	 
	 filename = idx2->Filename(); 
      } else
	filename = idx->Filename();
      
      unlink(filename);
      G->LOG_MISC.Put( SH->Lang( bidx->HasShareStatus(ShareBot::GETTING)
				   ? 0x03B : 0x03C, "s", bidx->Nick()));
      bidx->ClearShareStatus(ShareBot::GETTING|ShareBot::SENDING);
      bidx->Send(buf);
      NEW ( Bot ( SH->BOTNET->BOT, *bidx ) );
   }
   return EXECUTED;
}

static command_binds my_eof [] = {
   { "*", "b", share_eof, "share" }
};

static BindReturn share_sent ( void ** args, const char * mask) {
   FUNCTION;
   FileSend * idx = (FileSend *)args[0];
   ShareBot * bidx = (ShareBot *)G->socklist->Find(mask, DCT_BOT);
   
   if (bidx && (bidx->Type() == SH->SHAREBOT) && (bidx->Numver() > min_share)) {
      unlink(idx->Filename());
      bidx->ClearShareStatus(ShareBot::SENDING);
      G->LOG_MISC.Put(SH->Lang(0x03D, "s", bidx->Nick()));
      bidx->DumpResync();
   }
   return EXECUTED;
}

static command_binds my_sent [] = {
   { "*", "b", share_sent, "share" }
};

/* erase old user list, switch to new one */
static BindReturn share_rcvd ( void ** args, const char * mask) {
   FUNCTION;
   FileSend * idx = (FileSend *)args[0];
   Socket * ptr = G->socklist;
   ShareBot * bidx = (ShareBot *)G->socklist->Find(mask,DCT_BOT);
   UserRecord *u, *ou, *u2, *u3, *bu;
   TandemBot * tb;
   ChannelRecord *ch, *oh;
   UserEntry * ue, *ue2;
   FlagRecord gbfr (FlagRecord::BOT | FlagRecord::GLOBAL);
   FlagRecord chfl (FlagRecord::CHANNEL), gfr, bfl(FlagRecord::BOT);
   int ushr;
   Channel2 * chan;
   
   if (bidx && (bidx->Type() == SH->SHAREBOT) && (bidx->Numver() > min_share)) {
      bidx->ClearShareStatus(ShareBot::GETTING);
      G->LOG_MISC.Put( SH->Lang( 0x03E, "s", bidx->Nick())); 
      bfl.Set(bu = bidx->User());
      ou = G->userlist;
      G->userlist = NULL; /* do this to prevent .user messups */
      u = NEW ( UserRecord () );
      G->noshare = 1;
      DELETE ( G->global_bans );
      NEW ( BanRecord () );
      DELETE ( G->ignores );
      G->ignores = NEW ( IgnoreRecord () ) ;
      G->noshare = 0;
      /* read the rest in */
      if (G->check_chanlist)
	G->check_chanlist(0, 0, CHANLIST_CLEAR);
      if (!u->ReadUserfile(idx->Filename(), bidx)) 
	G->LOG_MISC.Put( SH->Lang( 0x03F, ""));
      else {
	 context;
	 G->noshare = 1;
	 G->LOG_MISC.Put( SH->Lang( 0x040, ""));
	 G->lastuser = NULL;
	 while ((ptr = ptr->Next()))
	   ptr->SetUser(0);
	 u2 = ou;
	 while ((u2 = ou->Next())) {
	    u3 = u->Find(u2->Handle());
	    gbfr.Set(u2);
	    ou->Remove(u2);
	    ushr = gbfr.GlobUnshared() || gbfr.GlobBot();
	    if (!u3) {
	       u->Add(u2);
	       u3 = u2;
	       if (!ushr) {
		  char buf[100];
		  UserEntryHosts * ueh = 
		    (UserEntryHosts *)u2->FindEntry(G->entrytype_hosts);
		  KeyedList<UHOSTLEN> * l;
		  UserEntryPass * uep = 
		    (UserEntryPass *)u2->FindEntry(G->entrytype_pass);
		  /* very simple, send only handle, flags, hostmasks &
		   * channel flags, all userentries are ignored */
		  
		  gfr.Set(u2);
		  gfr.Build(buf);
		  bidx->Printf("sn%s none %s\n", u2->Handle(), buf);
		  if (uep && uep->Value()) 
		    bidx->Printf("scPASS %s %s\n", u2->Handle(), uep->Value());
		  if (ueh && (l = ueh->First()))
		    while ((l = l->Next())) 
		      bidx->Printf("s+%s %s\n", u2->Handle(), l->Key());
		  ch = u2->ChanRecs();
		  while ((ch = ch->Next()))
		    if ((chan = FindChan (ch->Key()))) {
		       if (!bfl.BotGlobal())
			 chfl.Set(bu,chan->Key());
		       if (bfl.BotGlobal() || chfl.ChanShare()) {
			  chfl.Set(u2, chan->Key());
			  chfl.Build(buf);
			  bidx->Printf("sa%s %s %s\n", u2->Handle(), buf, chan->Key());
		       }
		    }
	       }
	    } else if (u3) {
	       if (bu == u2)
		 bu = u3;
	       if (!bfl.BotShareflags()) {
		  gfr.Set(u2);
		  u3->Set(gfr);
	       }
	       while ((ch = u2->ChanRecs()->Next())) 
		 if ((chan = FindChan (ch->Key()))) {
		    if (!ushr && !bfl.BotGlobal())
		      chfl.Set(bu,chan->Key());
		    if (ushr || !(bfl.BotGlobal() && chfl.ChanShare())) {
		       u2->ChanRecs()->Remove(ch);
		       if ((oh = u3->ChanRecs()->Find(ch->Key()))) {
			  u3->ChanRecs()->Remove ( oh );
			  DELETE ( oh );
		       }
		       u3->ChanRecs()->Add(ch);
		    }
		 } 
	       for (ue = u2->Entries()->Next(); ue; ue = ue2) {
		  ue2 = ue->Next();
		  ue->PruneUser(bidx);
	       }
	       while ((ue = u2->Entries()->Next())) {
		  u2->Entries()->Remove(ue);
		  if ((ue2 = u3->Entries()->Find(ue->Type()))) {
		     u3->Entries()->Remove(ue2);
		     DELETE ( ue2 );
		  }
		  u3->Entries()->Add(ue);
		  ue->SetOwner(u3);
	       }
	       DELETE ( u2 );
	    } 
	 }
	 DELETE ( ou );
	 G->userlist = u;
	 G->noshare = 0;
	 FlagRecord::ReaffirmOwners(); /* make sure my owners are +n */
	 if ((tb = SH->BOTNET->tandbot->Find(bidx->Nick())))
	   tb->Update(NULL,'+',0);	
	 ptr = G->socklist;
	 while ((ptr = ptr->Next()))
	   if (!ptr->User())
	     ptr->SetUser(G->userlist->Find(ptr->Nick()));
      }
   }
   unlink(idx->Filename());
   return EXECUTED;
}

static command_binds my_rcvd [] = {
   { "*", "b", share_rcvd, "share" }
};

static BindReturn cmd_flush (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   ShareBot * bot;
   char * par = (char *)args[1];
   int lang = 0x047;
   
   if (!par[0]) 
     lang = 0x045;
   else if (!(bot = (ShareBot *)G->socklist->Find(par, DCT_BOT))
	    || (bot->Type() != SH->SHAREBOT) || !bot->FlushResync())
     lang = 0x046;
   idx->Send(SH->Lang( lang, "s", par));
   return EXECUTED;
}

static command_binds my_cmds[1] = {
  { "flush", "n", cmd_flush, NULL },
};

#ifdef HAVE_IRC_H
static BindReturn channels_load_irc (void ** args, const char *) {
   FUNCTION;
   SH->IRC = (ModuleEntryIrc *)args[0];
   if (*SH->IRC->Vers() <= required_irc)
     SH->IRC = 0;
   return EXECUTED;
};

static BindReturn channels_unld_irc (void **, const char *) {
   FUNCTION;
   SH->IRC = 0;
   return EXECUTED;
};

static command_binds my_load[] = {
   { "irc", "", channels_load_irc, "share:irc" },
};

static command_binds my_unld[] = {
   { "irc", "", channels_unld_irc, "share:irc" },
};
#endif

static int script_getting_users (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   ShareBot* ptr = (ShareBot *)G->socklist;

   context;
   BADARGS(1, 1, "");
   while ((ptr = ptr->Next()))
     if (ptr->HasShareStatus(ShareBot::GETTING)) 
       break;
   sc->SetString(ptr ? "1" : "0");
   return SCRIPT_OK;
}

extern "C" {
   char *share_start (ModuleEntryEggdrop *, ModuleEntry * me);
};

char *share_start (ModuleEntryEggdrop * global_funcs, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   G = global_funcs;
   context;
   SH = NEW (ModuleEntryShare ( me, &r ));
   if (!r) {
#ifdef HAVE_IRC_H
      SH->IRC = (ModuleEntryIrc *)G->module_list->Find("irc", &required_irc);
      G->bt_load->AddBinds(my_load,1,SH);
      G->bt_unld->AddBinds(my_unld,1,SH);
#endif
      G->shareout = shareout_mod;
      G->hook_minutely->Add(share_minutely,SH);
      G->hook_read_userfile->Add(hook_read_userfile,SH);
      context;
      NEW ( Command ( "getting_users", script_getting_users, SH ) );
      context;
      G->bt_dcc->AddBinds(my_cmds,1,SH);
      SH->TRANSFER->bt_rcvd->AddBinds(my_rcvd,1,SH);
      SH->TRANSFER->bt_sent->AddBinds(my_sent,1,SH);
      SH->TRANSFER->bt_eof->AddBinds(my_eof,1,SH);
      SH->BOTNET->bt_version->AddBinds(my_version, 1,SH);
      context;
   }
   return r;
}
