/* 
 * TclScript.h - Script Context for Tcl
 */

#include "tcl.h"

#ifndef _SCRIPTCONTEXT_
#include "ScriptContext.h"
#endif

#ifndef _TCLSCRIPT_
#define _TCLSCRIPT_

class TclScript : public ScriptContext {
  protected:
   Tcl_Interp * interp;
  public:
   inline TclScript ( ModuleEntry * mod ) : ScriptContext ( "tcl", mod ) {
      interp = Tcl_CreateInterp();
      Tcl_Init(interp);
      SetMeUp();
   }
   ~TclScript ( void );
   
   int ExpMem ( void );
   
   void SetString ( const char *);
   void SetInteger ( int );
   void SetArray ( int, const char ** );
   void AppendString ( const char * );
   void AppendInteger ( int );
   void AppendArray ( int, const char ** );

   void AddStringVar ( StringVariable * );
   void RemStringVar ( StringVariable * );
   void AddIntVar ( IntegerVariable * );
   void RemIntVar ( IntegerVariable * );
   void AddCoupletVar ( CoupletVariable * );
   void RemCoupletVar ( CoupletVariable * );
   void AddMiscVar ( MiscVariable * );
   void RemMiscVar ( MiscVariable * );
   
   void AddCommand ( Command * );
   void RemCommand ( Command * );
   
   BindReturn CallBind ( BindEntry *, const char *, int, va_list );
   int Evaluate ( const char *, int, ... );
   const char * AsString ( void );
   int AsInteger ( void );
   void Clear ( void );
      
   inline Tcl_Interp * Interp ( void ) { return interp; }
};

#endif
