/*
 * ModuleEntryTimers.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Timer.h"
#include "Socket.h"
#define CLASS "ModuleEntryTimers"

static Version timers_version, required_eggdrop;
static struct patches timers_patches [] = {
#include "patch.h"
};

ModuleEntryTimers :: ModuleEntryTimers ( ModuleEntry * old, char ** r )
: ModuleEntry("timers", timers_version.Set(2,0), timers_patches, old) {
   FUNCTION;
   TM = this;
   if (!Depend("eggdrop", required_eggdrop.Set(2,0)))
     *r = "Your eggdrop core version is incompatible";
   else {
      memtype_timer = NEW ( MemType ("timers", (List **)&timer, this) );
      utimer = NEW ( Timer () );
      timer = NEW ( Timer () );
   }
}

ModuleEntryTimers::~ModuleEntryTimers ( void ) {
   FUNCTION;
   DELETE ( utimer );
   DELETE ( timer );
}

int ModuleEntryTimers::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ModuleEntryTimers) + ModuleEntry::ExpMem ();
}

int ModuleEntryTimers::NewTimer (TimerStack ts, int t, ScriptContext * sc,
				 void (*f)(int, char *), const char * x,
				 ModuleEntry * m) {
   FUNCTION;
   int ret = -1;
   
   if ((ts == SECONDS) || (ts == MINUTES)) {
      Timer * p = NEW ( Timer ( ts, t, sc, f, x, m) );
      ret = p->Key();
   }
   return ret;
}
     
int ModuleEntryTimers::KillTimer (TimerStack ts, int t) {
   FUNCTION;
   Timer * stack = utimer;
   
   if (ts == MINUTES)
     stack = timer;
   if ((ts == MINUTES) || (ts == SECONDS)) {
      Timer * p = (Timer *)stack->Find(t);
      
      if (p) {
	 stack->Remove(p);
	 DELETE ( p );
	 return 1;
      }
   }
   return 0;
}

int ModuleEntryTimers::TimerRunning (TimerStack ts, int t) {
   FUNCTION;
   Timer * stack = utimer;
   
   if (ts == MINUTES)
     stack = timer;
   if ((ts == MINUTES) || (ts == SECONDS)) {
      Timer * p = (Timer *)stack->Find(t);
      if (p)
	return 1;
   }
   return 0;
}

void ModuleEntryTimers::Report (Socket * idx, int details) {
   FUNCTION;
   int timers = 0, ctimers = 0, stimers = 0;
   Timer * mark = utimer;
   
   if (details) {
      while ((mark = mark->Next())) {
	 timers++;
	 if (mark->HasFunc())
	   ctimers++;
	 else
	   stimers++;
      }
      idx->Printf("     Secondly Timers: %d  (%d script, %d C)\n", timers, 
		  stimers, ctimers);
      ctimers = stimers = timers = 0;
      mark = timer;
      while ((mark = mark->Next())) {
	 timers++;
	 if (mark->HasFunc())
	   ctimers++;
	 else
	   stimers++;
      }
      idx->Printf("     Minutely Timers: %d  (%d script, %d C)\n", timers, 
		  stimers, ctimers);
   }
}
