/*
 * Timer.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Timer.h"
#include "ScriptContext.h"
#define CLASS "Timer"

int Timer::next_timer = 0;

Timer::Timer ( TimerStack stack, int elapse, ScriptContext * s,
	      void (*f) (int, char *), const char * x, ModuleEntry * m)
: IntList (*((stack == SECONDS) ? TM->utimer : TM->timer), ++next_timer, 
	   NODUPS | APPEND, m ) {
   FUNCTION;
   length = elapse;
   func = f;
   sc = s; 
   extra = (char *) my_malloc(strlen(x) + 1,TM->memtype_timer);
   strcpy(extra, x);
}

Timer::~Timer ( void ) {
   FUNCTION;
   if (extra)
     my_free(extra, TM->memtype_timer);
}

int Timer::ExpMem ( void ) {
   FUNCTION;
   int tot = 0;
   Timer * t;
   
   context;
   for (t = TM->timer; t; t = (Timer *)t->next)
     tot += sizeof(Timer) + (extra ? strlen(t->extra) + 1 : 0);
   for (t = TM->utimer; t; t = t->Next()) 
     tot += sizeof(Timer) + (t->extra ? strlen(t->extra) + 1 : 0);
   return tot;
}

void Timer::Check ( void ) {
   FUNCTION;
   Timer *mark = Next(), *old = this;
   context;
   next = 0;
   
   while (mark) {
      context;
      if (mark->length > 0)
	mark->length--;      
      old = mark;
      mark = mark->Next();
      if (!old->length) {
	 context;
	 if (old->func)
	   old->func(old->key,old->extra);
	 else
	   old->sc->Evaluate(old->extra, 0);
	 old->next = 0;
	 DELETE (old);
      } else {
	 context;
	 old->next = next;
	 next = old;
      }
   }
}
