/*
 * timers.c - handles:
 * timers, utimers for C & scripts
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "BindTable.h"
#include "Socket.h"
#include "HookList.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "Timer.h"
#include "Command.h"
#include "BindTable.h"
#define CLASS ""

static ModuleEntryEggdrop * G;
ModuleEntryTimers * TM;

static void timer_secondly ( void ) {
   FUNCTION;
   TM->utimer->Check();
}

static void timer_minutely ( void ) {
   FUNCTION;
   TM->timer->Check();
}

static int script_addtimer (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   long x;
   char buf[128];
   int u = argv[0]->AsString()[0]== 'u';
   
   context;
   if (argc != 3) {
      G->SPrintF(buf, 128, "wrong # args: should be \"%s comand \"",
		 u ? " seconds" : " minutes");
      sc->SetString(buf);
   } else if ((x = argv[1]->AsInteger()) < 0)
	sc->SetString("time value must be positive");
   else if (argv[2]->AsString()[0] != '#') {
      x = TM->NewTimer(u ? SECONDS : MINUTES, x, sc, 0, argv[2]->AsString(),0);
      G->SPrintF(buf, 128, "timer%d", x);
      sc->SetString(buf);
      return SCRIPT_OK;
   } 
   return SCRIPT_ERROR;
}

static int script_killtimer (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, " timerID");
   if (strncmp(argv[1]->AsString(), "timer", 5))
     sc->SetString("argument is not a timerID");
   else if (TM->KillTimer(argv[0]->AsString()[0] == 'u' ? SECONDS : MINUTES,
			  atol(argv[1]->AsString()+5)) > 0)
     return SCRIPT_OK;
   else
     sc->SetString("invalid timerID");
   return SCRIPT_ERROR;
}

/* return list of timers */
static int script_timers (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   Timer * mark;
   char id[20];
   const char *list[4];
   
   context;
   BADARGS(1, 1, "");
   for (mark = (argv[0]->AsString()[0] == 'u' ? TM->utimer : TM->timer); mark;
	mark = mark->Next()) {
      G->SPrintF(id, 19, "timer%d", mark->Key());
      list[0] = G->ToBase10(mark->ToGo());
      if (!mark->HasFunc()) {
	 list[1] = mark->SC()->Key();
	 list[2] = mark->String();
      } else {
	 list[1] = "C";
	 list[2] = "*INTERNAL*";
      }
      list[3] = id;
      sc->AppendArray(4, list);
   }
   return SCRIPT_OK;
}

BindReturn timer_unload ( void ** args, const char * ) {
   FUNCTION;
   ModuleEntry * m = (ModuleEntry *)args[0];
   
   TM->timer->Cleanup(m);
   TM->utimer->Cleanup(m);
   return EXECUTED;
}

static command_binds time_unld [] = {
   { "*", "", timer_unload, "timer" }
};
	
extern "C" {
   char * timers_start (ModuleEntryEggdrop *, ModuleEntry *);
};
   
char *timers_start (ModuleEntryEggdrop * global_funcs, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   G = global_funcs;
   context;
   NEW ( ModuleEntryTimers ( me, &r ) );
   if (!r) {
      NEW ( Command ( "timer", script_addtimer, TM ) );
      NEW ( Command ( "utimer", script_addtimer, TM ) );
      NEW ( Command ( "killtimer", script_killtimer, TM ) );
      NEW ( Command ( "killutimer", script_killtimer, TM ) );
      NEW ( Command ( "timers", script_timers, TM ) );
      NEW ( Command ( "utimers", script_timers, TM ) );
      G->hook_secondly->Add( timer_secondly, TM );
      G->hook_minutely->Add( timer_minutely, TM );
      G->bt_unld->AddBinds( time_unld, 1, TM );
   }
   return r;
}
