/*
 * FileGet.C - code handling sending of a file to a user
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "FileGet.h"
#include "FileGetListen.h"
#include "FlagRecord.h"
#include "BindTable.h"
#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
#endif
#define CLASS "FileGet"

FileGet::FileGet ( FileGetListen * old, int fd)
: FileSend (FD_ | RAW, TR->FILE_GET, DCT_FILETRAN, &TR->transfer_timeout,
	    fd, old->Host(), 0) {
   FUNCTION;
   char * p;
   struct stat ss;

   SetNick(old->Nick());
   SetUser(old->User());
   strcpy(dir, old->Filename());
   if ((p = strrchr(dir, '/'))) {
      strcpy (filename, p + 1);
      *p = 0;
   } else {
      dir[0] = 0;
      strcpy(filename, old->Filename());
   }
   stat(old->Filename(), &ss);
   length = ss.st_size;
   strcpy(from, old->From());
   f = fopen(old->Filename(), "r");
   /* file was already opened */
   if ((TR->dcc_block == 0) || (length < TR->dcc_block))
     SetStatus2(length);
   else
     SetStatus2(TR->dcc_block);
   p = (char *) my_malloc(status + 1,G->memtype_misc);
   fread(p, status, 1, f);
   Dump(p, status);
   my_free(p,G->memtype_misc);
   TimeNow();
}

FileGet::~FileGet ( void ) {
   FUNCTION;
}

int FileGet::ExpMem ( void ) {
   FUNCTION;
   return sizeof(FileGet);
}

void FileGet::Display2 ( Socket * where) {
   FUNCTION;
   int lang, data;
   
   if ((long)status == length) {
      lang = 0x018;
      data = acked;
   } else {
      lang = 0x019;
      data = status;
   }
   where->Send(TR->Lang( lang, "dlss", data, length, from, filename));
}

void FileGet::Eof (void) {
   FUNCTION;
   char xnick[NICKMAX + 1];
   
   context;
   if (from[0] != '*') {
      TR->LOG_FILES.Put(TR->Lang(0x011,"sss", filename, from, host));
      TR->WipeTmpFilename(filename,this);
      strcpy(xnick, from);
   } else
     xnick[0] = 0;
   CheckEof();
   Close();
   /* send next queued file if there is one */
   if (!TR->AtLimit(xnick))
     TR->SendNextFile(xnick);
   context;
}

void FileGet::Timeout (void) {
   FUNCTION;
   char xx[NICKMAX + 1];

   context;
   if (from[0] == '*') {
      G->LOG_MISC.Put(TR->Lang( 0x014, "s", from));
      xx[0] = 0;
   } else {
      char *p;

      p = strrchr(filename, '/');
#ifdef HAVE_SERVER_H
      if (TR->SERVER) 
	 TR->SERVER->SendNotice(Q_HELP,from,TR->Lang(0x015,"s",p?p+1:xx));
#endif
      TR->LOG_FILES.Put(TR->Lang( 0x016, "ssll", p ? p+1 : xx,from,status,length));
      TR->WipeTmpFilename(filename,this);
      strcpy(xx, from);
   }
   CheckEof();
   Close();
   if (!TR->AtLimit(xx))
     TR->SendNextFile(xx);
}

void FileGet::Activity (char * buf, int len) {
   FUNCTION;
   char xnick[NICKMAX + 1];
   unsigned char *bf, bbuf[4];
   long cmp, l;
   int w = len + sofar, p = 0;

   context;
   if (w < 4) {
      G->MemCpy(buf + sofar,buf,len);
      sofar += len;
   } else {
      if (w == 4) {
	 G->MemCpy((char *)bbuf,buf,sofar);
	 G->MemCpy((char *)(bbuf + sofar),buf,len);
      } else {      
	 p = ((w-1) & ~3) - sofar;
	 w = w - ((w - 1) & ~3);
	 if (w < 4) {
	    G->MemCpy(buf,buf + p,w);
	    return;
	 }
	 G->MemCpy((char *)bbuf,buf + p,w);
      } /* go back and read it again, it *does* make sense ;) */
      sofar = 0;
      /* this is more compatable than ntohl for machines where an int */
      /* is more than 4 bytes: */
      cmp = ((unsigned int) bbuf[0] << 24) + ((unsigned int) bbuf[1] << 16) +
	((unsigned int) bbuf[2] << 8) + bbuf[3];
      acked = cmp;
      if ((cmp > (long)status) && (cmp <= length)) {
	 /* attempt to resume I guess */
	 if (from[0] == '*')
	   G->LOG_MISC.Put(TR->Lang( 0x00E, "s", from));
	 else {
	    fseek(f, cmp, SEEK_SET);
	    SetStatus2(cmp);
	    TR->LOG_FILES.Put(TR->Lang( 0x00F, "dss", (int) (cmp / 1024),
				      filename, from));
	 }
      }
      if (cmp == (long)status) {
	 if ((long)status == length) {
	    void * arg = this;
	    FlagRecord fr (FlagRecord::GLOBAL | FlagRecord::CHANNEL |
			   FlagRecord::ANYWHERE );
	    
	    context;
	    /* successful send, we're done */
	    if (from[0] != '*') {
	       TR->LOG_FILES.Put(TR->Lang( 0x010, "ss", filename, from));
	       TR->WipeTmpFilename(filename,this);
	       strcpy((char *) xnick, from);
	    } else
	      xnick[0] = 0;
	    fr.Set(user);
	    TR->bt_sent->Check(nick, &fr, &arg);
	    Close();
	    /* any to dequeue? */
	    if (!TR->AtLimit(xnick))
	      TR->SendNextFile(xnick);
	 } else {
	    context;
	    /* note: is this fseek necessary any more? */
	    /*    fseek(xfer(idx)->f,idx->status,0);   */
	    l = TR->dcc_block;
	    if ((l == 0) || (((long)status + l) > length))
	      l = length - status;
	    bf = (unsigned char *) my_malloc(l + 1,G->memtype_misc);
	    fread(bf, l, 1, f);
	    Dump((char *)bf, l);
	    my_free(bf,G->memtype_misc);
	    SetStatus2(status + l);
	    TimeNow();
	 }
      }
   }
}
