/*
 * FileGetListen.C - code handling sending of a file to the bot (listen port)
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "FileGetListen.h"
#include "FileGet.h"
#include "UserRecord.h"
#include "BindTable.h"
#include <errno.h>
#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
#endif
#define CLASS "FileGetListen"

FileGetListen::FileGetListen ( const char * file, const char * frm,
			      const char * hst, const char * nck)
: ListeningSocket ( TR->FILE_LISTEN, 0, G->reserved_port ) {
   FUNCTION;
   if (Enable()) {
      type_flags |= DCT_FILETRAN;
      timeout_val = &TR->transfer_timeout;
      SetNick(nck);
      SetUser(G->userlist->Find(nick));
      SetHost(hst);
      strncpy(from, frm, NICKMAX);
      from[NICKMAX] = 0;
      strncpy (filename, file, DIRLEN);
      filename[DIRLEN] = 0;
   } else 
     Close();
}

FileGetListen::~FileGetListen ( void ) {
   FUNCTION;
}

void FileGetListen::Eof (void) {
   FUNCTION;
   char xnick[NICKMAX + 1];
   context;

   if (from[0] != '*') {
      TR->LOG_FILES.Put(TR->Lang(0x03B,"sss", filename, from, host));
      TR->WipeTmpFilename(filename,this);
      strcpy(xnick, from);
   } else
     xnick[0] = 0;
   CheckEof();
   Close();
   /* send next queued file if there is one */
   if (!TR->AtLimit(xnick))
      TR->SendNextFile(xnick);
   context;
}

void FileGetListen::Timeout ( void ) {
   FUNCTION;
   char xx[NICKMAX + 1];

   context;
   if (from[0] == '*') {
      G->LOG_MISC.Put(TR->Lang(0x03C, "s", from));
      xx[0] = 0;
   } else {
      char *p = strrchr(filename, '/');
      
#ifdef HAVE_SERVER_H
      if (TR->SERVER) 
	TR->SERVER->SendNotice(Q_HELP,from,TR->Lang(0x03D,"s", p ? p + 1 : xx));
#endif
      TR->LOG_FILES.Put(TR->Lang(0x016, "ssll", p ? p + 1 : xx, from));
      TR->WipeTmpFilename(filename,this);
      strcpy(xx, from);
   }
   CheckEof();
   Close();
   if (!TR->AtLimit(xx))
     TR->SendNextFile(xx);
}

void FileGetListen::CheckEof (void) {
   FUNCTION;
   void * arg = this;
   FlagRecord fr ( FlagRecord::GLOBAL | FlagRecord::CHANNEL | FlagRecord::ANYWHERE );
   
   fr.Set(user);
   TR->bt_eof->Check(nick, &fr, &arg);
}

void FileGetListen::Activity ( char *, int ) {
   FUNCTION;
   FileGet * newsock;
   context;
   
   if (!(newsock = (FileGet *)Answer())) {
#ifdef HAVE_SERVER_H
      if (TR->SERVER) 
	TR->SERVER->SendNotice(Q_HELP, from, TR->Lang(0x01E, "s",
						      strerror(errno)));
#endif
      TR->LOG_FILES.Put(TR->Lang(0x01F, "sss", filename, from, host));
      CheckEof();
   }
   Close();
}

void FileGetListen::Display2 (Socket * where) {
   FUNCTION;
   where->Send(TR->Lang(0x01A, "lss", Age(), from, filename));
}


Socket * FileGetListen::Create ( int fd ) {
   FUNCTION;
   return NEW ( FileGet ( this, fd ) );
}
	       
int FileGetListen::ExpMem ( void ) {
   FUNCTION;
   return sizeof(FileGetListen);
}

