/*
 * FileSend.C - code handling sending of a file to the bot
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "FileSend.h"
#include "UserRecord.h"
#include "BindTable.h"
#include <errno.h>
#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
#endif
#define CLASS "FileSend"

FileSend::FileSend ( struct in_addr * in, int port, const char * frm,
		    UserRecord * u, long len, const char * hst,
		    char * fnm, FILE * f1)
: Socket ( TR->FILE_SEND, RAW | CONNECTING | STRONG, *in, 
	  port, DCT_FILETRAN | DCT_FILESEND, &TR->transfer_timeout ) {
   FUNCTION;
   char * p;

   user = u;
   SetNick( u ? u->Handle() : "*");
   length = len;
   strncpy(from, frm, NICKMAX);
   from[NICKMAX] = 0;
   if (hst)
     SetHost(hst);
   TimeNow();
   if ((p = strrchr(fnm, '/'))) {
      *p++ = 0;
      strncpy(dir, fnm, DIRLEN);
      dir[DIRLEN] = 0;
      strncpy (filename, p, DIRLEN);
      p--;
   } else {
      strncpy(filename, fnm, DIRLEN);
      dir[0] = 0;
   }
   filename[DIRLEN] = 0;
   f = f1;
}

FileSend::~FileSend ( void ) {
   FUNCTION;
   if (f)
     fclose (f);
}

int FileSend::ExpMem ( void ) {
   FUNCTION;
   return sizeof(FileSend);
}

void FileSend::Display2 ( Socket * where ) {
   FUNCTION;
   where->Send(TR->Lang( (status < 0) ? 0x01B : 0x01C,
		       from, status, length, filename ));
}

void FileSend::Eof ( void ) {
   FUNCTION;
   int ok;
   char ofn[DIRLEN+1], nfn[DIRLEN+1], *s1;
   Socket * ptr = G->socklist;
   
   fclose(f);
   f = 0;
   if (from[0] != '*')
     G->SPrintF(ofn, DIRLEN, "%s%s", G->tempdir, filename);
   else
     ofn[0] = 0;
   if (status < 0) {
      if (from[0] != '*') {
	 s1 = strerror(errno);
#ifdef HAVE_SERVER_H
	 if (TR->SERVER) 
	   TR->SERVER->SendNotice(Q_HELP, from, TR->Lang( 0x008, "s", s1));
#endif
	 G->LOG_MISC.Put(TR->Lang( 0x009, "ssss", filename, from, host, s1));
	 unlink(ofn);
      }
      CheckEof ();
   } else if (length == (long)status) {
      /* success */
      ok = 1;
      G->SPrintF(nfn, DIRLEN, "%s%s", dir, filename);
      if (from[0] != '*') {
	 TR->LOG_FILES.Put(TR->Lang(0x00A, "sss", filename, from, host));
	 /* move the file from /tmp */
	 if (G->MoveFile(ofn, nfn)) {
	    LogModes x ( LogModes::MISC | LogModes::FILES );
	    
	    x.Put(TR->Lang(0x00B, "ss", filename, G->tempdir));
	    ok = 0;
	 } else {	
	    while ((ptr = ptr->Next()))
	      if ((ptr->TypeFlags() & (DCT_GETNOTES | DCT_FILES)) &&
		  (ptr->User() == user)) {
		 ptr->Send(TR->Lang(0x00C, ""));
		 break;
	      }
#ifdef HAVE_SERVER_H
	    if (!ptr && TR->SERVER)
	      TR->SERVER->SendNotice(Q_HELP, from, TR->Lang(0x00C,""));
#endif
	    ok = 1;
	 }
      }
      if (ok) {
	 void * arg = this;
	 FlagRecord fr (user, FlagRecord::GLOBAL | FlagRecord::CHANNEL | 
			FlagRecord::ANYWHERE);

	 TR->bt_rcvd->Check(nick, &fr, &arg);
      }
   } else {
      /* failure :( */
      if (from[0] != '*') {
	 TR->LOG_FILES.Put(TR->Lang(0x00D, "sssll", filename, nick, host,
			       status, length));
	 unlink(ofn);
      }
      CheckEof ();
   }
   Close();
}

void FileSend::Timeout ( void ) {
   FUNCTION;
   char xx[DIRLEN+1];
   
   if (from[0] == '*')
     G->LOG_MISC.Put(TR->Lang( 0x014, "s", from));
   else {
#ifdef HAVE_SERVER_H
      if (TR->SERVER) 
	TR->SERVER->SendNotice(Q_HELP, from, TR->Lang( 0x015, "s", filename));
#endif
      TR->LOG_FILES.Put(TR->Lang( 0x017, "ssll", filename, from, status, length));
      G->SPrintF(xx, DIRLEN, "%s%s", G->tempdir, filename);
      unlink(xx);
   }
   CheckEof ( );
   Close();
}

void FileSend::CheckEof (void) {
   FUNCTION;
   void * arg = this;
   FlagRecord fr ( FlagRecord::GLOBAL | FlagRecord::CHANNEL | FlagRecord::ANYWHERE );
   
   fr.Set(user);
   TR->bt_eof->Check(nick, &fr, &arg);
}

void FileSend::Activity (char * buf, int len) {
   FUNCTION;
   unsigned char s[DIRLEN + 1];
   long sent;

   if (status < 0) {
      status = 0;
      if (from[0] != '*') 
	G->LOG_MISC.Put(TR->Lang( 0x01D, "sss", filename, from, host));
   } else {
      context;
      fwrite(buf, len, 1, f);
      sent = status + len;
      SetStatus2(sent);
      /* put in network byte order */
      s[0] = (sent / (1 << 24));
      s[1] = (sent % (1 << 24)) / (1 << 16);
      s[2] = (sent % (1 << 16)) / (1 << 8);
      s[3] = (sent % (1 << 8));
      Dump((char *)s, 4);
      TimeNow();
      if ((sent > length) && (length > 0)) {
#ifdef HAVE_SERVER_H
	 if (TR->SERVER)
	   TR->SERVER->SendNotice(Q_HELP, from, TR->Lang(0x012,""));
#endif
	 TR->LOG_FILES.Put(TR->Lang( 0x013, "sss", filename, nick, host));
	 G->SPrintF((char *)s, DIRLEN, "%s%s", G->tempdir, filename);
	 unlink((char *)s);
	 CheckEof();
	 Close();
      }
   }
}
