/*
 * ModuleEntryTransfer.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryTransfer.h"
#include "IntegerVariable.h"
#include "HelpReference.h"
#include "FileQueue.h"
#include "BindTableSentRcvd.h"
#include "Socket.h"
#include "FileGet.h"
#include "FileGetListen.h"
#include "FileSend.h"
#include "Socket.h"
#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
#endif
#define CLASS "ModuleEntryTransfer"

/* version stuff */
static Version transfer_version, required_eggdrop;
static struct patches transfer_patches[] = {
#include "patch.h"
};

ModuleEntryTransfer::ModuleEntryTransfer ( ModuleEntry * old, char ** r ) 
: ModuleEntry ("transfer", transfer_version.Set(2,0), transfer_patches,old) {
   FUNCTION;
   TR = this;
   if (!Depend("eggdrop", required_eggdrop.Set(2,0)))
     *r = "Your eggdrop core version is incompatible";
   else {
      context;
      memtype_fileq = NEW ( MemType ("fileq", (List **)&fileq, this) );
      fileq = NEW ( FileQueue () );
      context;
      copy_to_tmp = 1;
      dcc_block = 1024;
      transfer_timeout = 300;
      dcc_limit = 3;
      FILE_SEND = "FILE_SEND";
      FILE_GET = "FILE_GET";
      FILE_LISTEN = "FILE_LISTEN";
      LOG_FILES = LogModes::FILES;
      NEW ( HelpReference ( "transfer.help", this ) );
      context;
      bt_rcvd = NEW ( BindTableSentRcvd ("rcvd", this ) );
      bt_sent = NEW ( BindTableSentRcvd ("sent", this ) );
      bt_eof = NEW ( BindTableSentRcvd ("eof", this ) );   
      NEW ( IntegerVariable ( "max_dloads", &dcc_limit, 0, this ) );
      NEW ( IntegerVariable ( "dcc_block", &dcc_block, 0, this ) );
      NEW ( IntegerVariable ( "copy_to_tmp", &copy_to_tmp, 0, this ) );
      NEW ( IntegerVariable ( "xfer_timeout", (int *)&transfer_timeout, 0, this) );
   }
}

ModuleEntryTransfer::~ModuleEntryTransfer ( void ) {
   FUNCTION;
   Socket * ptr = G->socklist, * next = ptr->Next();
   
   context;
   G->LOG_MISC.Put(Lang(0x033, ""));
   DELETE ( fileq );
   while ((ptr = next)) {
      next = ptr->Next();
      if (ptr->TypeFlags() & DCT_FILETRAN) {
	 ptr->Eof();
	 delete ptr;
      }
   }
}

int ModuleEntryTransfer::ExpMem ( void ) {
   FUNCTION;
   return sizeof(ModuleEntryTransfer) + ModuleEntry::ExpMem ( );
}

void ModuleEntryTransfer::Report (Socket * idx, int details) {
   FUNCTION;
   if (details) {
      idx->Send(Lang(0x034, "dsd", dcc_block, !dcc_block ?
			Lang(0x035, "") : "", dcc_limit));
      idx->Send(Lang(0x036, "d", fileq->ExpMem()));
   }
}

/* originally from core match.c */
/*========================================================================*
 * Features:  Forward, case-sensitive, ?, *                               *
 * Best use:  File mask matching, as it is case-sensitive                 *
 *========================================================================*/
int ModuleEntryTransfer::WildMatchFile(register unsigned char * m, 
				       register unsigned char * n){
   FUNCTION;
   unsigned char *ma = m, *lsm = 0, *lsn = 0;
   int match = 1;
   register unsigned int sofar = 0;

   /* take care of null strings (should never match) */
   if (!m || !n || !*n)
      return 0;
   /* (!*m) test used to be here, too, but I got rid of it.  After all,
    * If (!*n) was false, there must be a character in the name (the
    * second string), so if the mask is empty it is a non-match.  Since
    * the algorithm handles this correctly without testing for it here
    * and this shouldn't be called with null masks anyway, it should be
    * a bit faster this way */

   while (*n) {
      /* Used to test for (!*m) here, but this scheme seems to work better */
      switch (*m) {
      case 0:
	 do
	    m--;		/* Search backwards      */
	 while ((m > ma) && (*m == '?'));	/* For first non-? char  */
	 if ((m > ma) ? ((*m == '*') && (m[-1] != '\\')) : (*m == '*'))
	    return (match +sofar);	/* nonquoted * = match   */
	 break;
      case '*':
	 do
	    m++;
	 while (*m == '*');	/* Zap redundant wilds   */
	 lsm = m;
	 lsn = n;		/* Save * fallback spot  */
	 match += sofar;
	 sofar = 0;
	 continue;		/* Save tally count      */
      case '?':
	 m++;
	 n++;
	 continue;		/* Match one char        */
      case '\\':
	 m++;			/* Handle quoting        */
      }
      if (*m == *n) {		/* If matching           */
	 m++;
	 n++;
	 sofar++;
	 continue;		/* Tally the match       */
      }
      if (lsm) {		/* Try to fallback on *  */
	 n = ++lsn;
	 m = lsm;		/* Restore position      */
	 /* Used to test for (!*n) here but it wasn't necessary so it's gone */
	 sofar = 0;
	 continue;		/* Next char, please     */
      }
      return 0;		/* No fallbacks=No match */
   }
   while (*m == '*')
      m++;			/* Zap leftover *s       */
   return (*m) ? 0 : (match + sofar);	/* End of both = match   */
}

void ModuleEntryTransfer::SendNextFile (const char * to) {
   FUNCTION;
   FileQueue * which = (FileQueue *)fileq->Find(to);
   char s1[DIRLEN + 1], s[DIRLEN + 1];

   if (which) {
      /* copy this file to /tmp */
      if (*which->Dir() == '*')	/* absolute path */
	G->SPrintF(s, DIRLEN, "%s/%s", which->Dir() + 1, which->File());
      else {
	 const char *p = strchr(which->Dir(), '*');
	 if (!p) {		/* if it's messed up */
	    DELETE ( which );
	    SendNextFile(to);
	    return;
	 }
	 p++;
	 G->SPrintF(s, DIRLEN, "%s%s%s", p, p[0] ? "/" : "", which->File());
	 strcpy(which->Dir(), &(p[atoi(which->Dir())]));
      }
      if (copy_to_tmp) {
	 G->SPrintF(s1, DIRLEN, "%s%s", G->tempdir, which->File());
	 if (G->CopyFile(s, s1) != 0) {
	    LogModes x ( LogModes::FILES | LogModes::MISC );
	    
	    x.Put(Lang(0x001,"ss",which->File(),G->tempdir));
#ifdef HAVE_SERVER_H
	    if (SERVER)
	      SERVER->SendNotice(Q_HELP, which->Key(),Lang(0x002,""));
#endif
	    fileq->Flush(to);
	    return;
	 }
      } else
	strcpy(s1, s);
      switch (RawDccSend(s1, which->Key(), which->From())) {
	case -1:
	 WipeTmpFilename(s1, NULL);
	 LOG_FILES.Put(Lang(0x003,"ss",s1,which->From()));
#ifdef HAVE_SERVER_H
	 if (SERVER)
	   SERVER->SendNotice(Q_HELP, which->Key(),Lang(0x004,""));
#endif
	 fileq->Flush(to);
	 break;
	case -2:
	 WipeTmpFilename(s1, NULL);
	 LOG_FILES.Put(Lang(0x005,"ss",s1,which->From()));
#ifdef HAVE_SERVER_H
	 if (SERVER)
	   SERVER->SendNotice(Q_HELP, which->Key(),Lang(0x006,""));
#endif
	 fileq->Flush(to);
	 break;
	default:
#ifdef HAVE_SERVER_H
	 if (!strcasecmp(which->Key(), which->From()) && SERVER)
	   SERVER->SendNotice(Q_HELP, which->Key(),Lang(0x007,"s",
							       which->Key()));
#endif
	 DELETE ( which );
      }
   }
}

void ModuleEntryTransfer::WipeTmpFilename (const char * fn, Socket * idx) {
   FUNCTION;
   FileGet * ptr = (FileGet *)G->socklist;
   FileGetListen * ptr2 = (FileGetListen *)G->socklist;
   
   if (copy_to_tmp) {
      while ((ptr = ptr->Next()))
	if ((ptr != idx) && !strcmp(ptr->Filename(), fn))
	  return;
      while ((ptr2 = ptr2->Next()))
	if ((ptr2 != idx) && !strcmp(ptr2->Filename(), fn))
	  return;
      unlink(fn);
   }
}

/* return true if this user has >= the maximum number of file xfers going */
int ModuleEntryTransfer::AtLimit (const char * nick) {
   FUNCTION;
   FileGet * ptr = (FileGet *)G->socklist;
   FileGetListen * ptr2 = (FileGetListen *)G->socklist;
   int x = 0;
   
   while ((ptr = ptr->Next()))
     if (!strcasecmp(ptr->From(), nick))
       x++;
   while ((ptr2 = ptr2->Next()))
     if (!strcasecmp(ptr2->From(), nick))
       x++;
   return (x >= dcc_limit);
}

void ModuleEntryTransfer::ShowQueuedFiles (Socket * idx) {
   FUNCTION;
   int cnt = 0;
   FileQueue * q = fileq;
   FileGet * ptr = (FileGet *)G->socklist;
   FileGetListen * ptr2 = (FileGetListen *)G->socklist;
   
   while ((q = q->Next()))
     if (!strcasecmp(q->From(), idx->Nick())) {
	if (!cnt) {
	   idx->Send(Lang( 0x023, " ", HANDLEN - 9));
	   idx->Send(Lang( 0x024, " ", HANDLEN - 9));
	}
	cnt++;
	idx->Send(Lang( (*q->Dir() == '*') ? 0x025 : 0x026,
		       "s sss", q->Key(), HANDLEN-strlen(q->Key()),
		       q->Dir() + ((*q->Dir() == '*') ? 1 : 0),
		       q->File(), (*q->Dir() == '*') ? "/" : ""));
     }
   while ((ptr = ptr->Next())) 
     if ((!strcasecmp(ptr->Nick(), idx->Nick()) ||
	  !strcasecmp(ptr->From(), idx->Nick()))) {
	const char *nfn;
	
	if (!cnt) {
	   idx->Send(Lang( 0x023, " ", HANDLEN - 9));
	   idx->Send(Lang( 0x024, " ", HANDLEN - 9));
	}
	nfn = (nfn = strrchr(ptr->Filename(), '/')) ? nfn + 1 
	  : ptr->Filename();
	cnt++;
	idx->Send(Lang( 0x028, "s sF", ptr->From(),
			  HANDLEN-strlen(ptr->From()),
			  nfn,(100.0 * ((float) ptr->Status() /
					(float) ptr->Length()))));
     }
   while ((ptr2 = ptr2->Next())) 
     if ((!strcasecmp(ptr2->Nick(), idx->Nick()) ||
	  !strcasecmp(ptr2->From(), idx->Nick()))) {
	const char *nfn;
	
	if (!cnt) {
	   idx->Send(Lang( 0x023, " ", HANDLEN - 9));
	   idx->Send(Lang( 0x024, " ", HANDLEN - 9));
	}
	nfn = (nfn = strrchr(ptr2->Filename(), '/')) ? nfn + 1 
	  : ptr2->Filename();
	cnt++;
	idx->Send(Lang( 0x027, "s s", ptr2->From(), 
			  HANDLEN - strlen(ptr2->From()),nfn));
     }
   idx->Send(Lang( cnt ? 0x02A : 0x029, "d", cnt));
}

void ModuleEntryTransfer::FileQCancel (Socket * idx, char * par) {
   FUNCTION;
   int matches = 0, atot = 0, i;
   Socket * ptr = G->socklist;
   FileQueue * q, *q2 = fileq->Next();
   char s[DIRLEN+1];
   const char *filename = 0, *nfn , *from;
   
   while ((q = q2)) {
      q2 = q->Next();
      if (!strcasecmp(idx->Nick(), q->From())) {
	 if (*q->Dir() == '*')
	   G->SPrintF(s, DIRLEN, "%s/%s", q->Dir() + 1, q->File());
	 else
	   G->SPrintF(s, DIRLEN, "/%s%s%s", q->Dir(), *q->Dir() ? "/" : "",
		      q->File());
	 if (WildMatchFile((unsigned char *)par, (unsigned char *)s) || 
	     WildMatchFile((unsigned char *)par, (unsigned char *)q->File())) {
	    idx->Send(Lang( 0x02B, "ss", s, q->Key()));
	    DELETE (q);
	    matches++;
	 }
      }
   }
   while ((ptr = ptr->Next())) {
      if ((ptr->Type() == FILE_GET) &&
	  (!strcasecmp(ptr->Nick(), idx->Nick()) ||
	   !strcasecmp(((FileGet *)ptr)->From(), idx->Nick()))) {
	 filename = ((FileGet *)ptr)->Filename();
	 from = ((FileGet *)ptr)->From();
      } else if ((ptr->Type() == FILE_LISTEN) && 
		 (!strcasecmp(ptr->Nick(), idx->Nick()) ||
		  !strcasecmp(((FileGetListen *)ptr)->From(), idx->Nick()))) {
	 filename = ((FileGetListen *)ptr)->Filename();
	 from = ((FileGetListen *)ptr)->From();
      } else
	continue;
      nfn = strrchr(filename, '/');
      nfn = nfn ? nfn + 1 : filename;
      if (WildMatchFile((unsigned char *)par, (unsigned char *)nfn)) {
	 idx->Send(Lang( 0x02C, "s", nfn));
#ifdef HAVE_SERVER_H
	 if (!strcasecmp(from, idx->Nick()) && SERVER)
	   SERVER->SendNotice(Q_HELP,from,Lang(0x02D,"ss",nfn,idx->Nick())); 
#endif
	 if (ptr->Type() == FILE_GET)
	   LOG_FILES.Put(Lang( 0x02E, "ssll", nfn, from,
				 ptr->Status(), ((FileGet *)ptr)->Length()));
	 WipeTmpFilename(filename,ptr);
	 atot++;
	 matches++;
	 ptr->Close();
      }
   }
   idx->Send(Lang( !matches ? 0x02F : (matches == 1) ? 0x030 : 0x031,
		     "d", matches));
   for (i = 0; i < atot; i++)
     if (!AtLimit(idx->Nick()))
       SendNextFile(idx->Nick());
}

int ModuleEntryTransfer::RawDccSend (char * filename, const char * nick,
				     const char * from) {
   FUNCTION;
   char *nfn;
   FILE * f;
   Socket * ptr;
   
   context;
   if (G->dcc_total == G->max_dcc)
     return DCCSEND_FULL;
   nfn = (nfn = strrchr(filename, '/')) ? nfn + 1 : filename;
   if (!(f = fopen(filename, "r")))
     return DCCSEND_BADFN;
   if (!(ptr = NEW ( FileGetListen(filename, from, "irc", nick)))) {
      fclose(f);
      return DCCSEND_NOSOCK;
   }
   if (from[0] != '*') {
#ifdef HAVE_SERVER_H
      if (SERVER) {
	 char x[512];
	 struct stat ss;
	 
	 ::stat(filename,&ss);
	 G->SPrintF(x, 511, "\001DCC SEND %s %lu %d %lu\001\n",
		    nfn, ptr->Address().s_addr, ptr->LocalPort(), ss.st_size);
	 SERVER->SendMsg(Q_HELP, from, x);
      }
#endif
      LOG_FILES.Put(Lang( 0x032, "ss", nfn, from));
   }
   return ptr->FD();
}
