/* 
 * transfer.c (tansfer module) 
 * handles: all low level file transfers (DCC GET/SEND)
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#define MAKING_TRANSFER

/* sigh sunos */
#include <sys/types.h>
#include <sys/stat.h>
#include "ListeningSocket.h"
#include "BindTable.h"
#include "UserRecord.h"
#include "BindTableSentRcvd.h"
#include "ModuleEntryTransfer.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "FileSend.h"
#include "FileGet.h"
#include "FileGetListen.h"
#include "HostnameCache.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "FileQueue.h"
#include "Command.h"
#define CLASS ""

#ifdef HAVE_SERVER_H
#include "ModuleEntryServer.h"
static Version required_server ( 2, 0 );
#endif
static ModuleEntryEggdrop * G;
ModuleEntryTransfer * TR;

/* a few forward references */
extern "C" {
   char * transfer_start (ModuleEntryEggdrop *, ModuleEntry *);
};

/***********************************************************************
 * SCRIPT COMMANDS                                                     *
 ***********************************************************************/

static int script_getfileq (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char s[DIRLEN+1];
   const char * l = s;
   FileQueue * q = TR->fileq;
   
   BADARGS(2, 2, TR->Lang(0x020, ""));
   while ((q = q->Next()))
     if (!strcasecmp(q->From(), argv[1]->AsString())) {
	if (*q->Dir() == '*')
	  G->SPrintF(s, DIRLEN, "%s %s/%s", q->Key(), q->Dir() + 1, q->File());
	else
	  G->SPrintF(s, DIRLEN, "%s /%s%s%s", q->Key(), q->Dir(),
		     *q->Dir() ? "/" : "", q->File());
	sc->AppendArray(1, &l);
     }
   return SCRIPT_OK;
}

static int script_dccsend (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   char sys[DIRLEN+1], *nfn, buf[DIRLEN + 1];
   int ret;
   FILE *f;
   
   BADARGS(3, 3, TR->Lang(0x021, ""));
   strncpyz(buf, argv[1]->AsString(), DIRLEN);
   if (!(f = fopen(buf, "r"))) 
     ret = 3;
   else {
      fclose(f);
      nfn = (nfn = strrchr(buf, '/')) ? nfn + 1 : buf;
      if (TR->AtLimit(argv[2]->AsString())) {
	 /* queue that mother */
	 if (nfn == buf)
	   strcpy(sys, "*");
	 else {
	    nfn--;
	    *nfn = 0;
	    nfn++;
	    G->SPrintF(sys, DIRLEN, "*%s", buf);
	 }
	 NEW ( FileQueue (sys, nfn, TR->Lang(0x022, ""),argv[2]->AsString()));
	 ret = 4;
      } else {
	 if (TR->copy_to_tmp) {  /* new filename, in /tmp */
	    G->SPrintF(sys, DIRLEN, "%s%s", G->tempdir, nfn); 
	    G->CopyFile(buf, sys);
	 } else 
	   strncpyz(sys, buf, DIRLEN);
	 if ((ret = TR->RawDccSend(sys, argv[2]->AsString(), "*")) < 0) {
	    TR->WipeTmpFilename(sys, NULL);
	    ret = -ret;
	 } else
	   ret = 0;
      }
   }
   sc->SetInteger(ret);
   return SCRIPT_OK;
}

/* dccget <host> <port> <from> <user> <size> <filename> */
static int script_dccget (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   struct in_addr * in;
   const char * realhost;
   FILE * f;
   UserRecord * u;
   FileSend * fs;
   
   BADARGS(6, 6, TR->Lang(0x037, ""));
   if ((realhost = strchr(argv[1]->AsString(), '@')))
     realhost++;
   else
     realhost = argv[1]->AsString();
   if (!(in = G->host_cache->Find(realhost)))
     sc->SetString(TR->Lang(0x038, ""));
   else if (!(u = G->userlist->Find(argv[4]->AsString())))
     sc->SetString(TR->Lang(0x039, ""));
   else if (!(f = fopen(argv[6]->AsString(),"w"))) 
     sc->SetString(TR->Lang(0x040, ""));
   else {
      char buf[DIRLEN + 1];
      
      strncpyz(buf, argv[6]->AsString(), DIRLEN);
      fs = NEW ( FileSend ( in, argv[2]->AsInteger(), argv[3]->AsString(), u,
			   argv[5]->AsInteger(), argv[1]->AsString(), buf, f));
      sc->SetInteger(fs->FD());
      return SCRIPT_OK;
   }
   return SCRIPT_ERROR;
}

#ifdef HAVE_SERVER_H
static BindReturn transfer_load_server (void ** args, const char *) {
   FUNCTION;
   TR->SERVER = (ModuleEntryServer *)args[0];
   if (*TR->SERVER->Vers() <= required_server)
     TR->SERVER = 0;
   return EXECUTED;
};

static BindReturn transfer_unld_server (void **, const char *) {
   FUNCTION;
   TR->SERVER = 0;
   return EXECUTED;
};

static command_binds my_load[] = {
   { "server", "", transfer_load_server, "transfer:server" },     
};

static command_binds my_unld[] = {
   { "server", "", transfer_unld_server, "transfer:server" },
};
#endif

char *transfer_start (ModuleEntryEggdrop * global_funcs, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;
   
   G = global_funcs;
   context;
   TR = NEW ( ModuleEntryTransfer ( me, &r ) );
   if (!r) {
#ifdef HAVE_SERVER_H
      TR->SERVER = (ModuleEntryServer *)
	G->module_list->Find("server", &required_server);
#endif
#ifdef HAVE_SERVER_H
      G->bt_load->AddBinds(my_load,1,TR);
      G->bt_unld->AddBinds(my_unld,1,TR);
#endif
      context;
      NEW ( Command ( "dccsend", script_dccsend, TR ) );
      NEW ( Command ( "dccget", script_dccget,TR ) );
      NEW ( Command ( "getfileq", script_getfileq,TR ) );
      context;
   }
   return r;
}
