/* 
 * userinfo module, based on userinfo.tcl 1.03 by butchbub/beldin/kirk
 *
 * userinfo.tcl v1.03 for Eggdrop 2.0.0 and higher 
 *			Scott G. Taylor -- ButchBub!staylor@mrynet.com
 * 
 * V1.00	     ButchBub	  14 July      1997  Original release.  Based on
 *                                             whois.tcl "URL" commands.
 * v1.01      Beldin       11 November  1997  1.3 only version
 * v1.02      Kirk         31 January   1998  1.5.3 'only' version
 * v1.03      Kirk         17 February  1998  2.0.0 'only' version
 * v2.00      Kirk/Beldin  29 August	1998  upgrade to new module code
 *
 * TO USE:  o	Set the desired userinfo field keywords to the 
 *		`userinfo-fields' line below where indicated.
 *          o    Load this script on a 2.0.0 or later Eggdrop bot.
 *	    o	Begin having users save the desired information.  If you
 *		chose to add the default "IRL" field, they just use
 *		the IRC command: /MSG <botnick> irl Joe Blow.
 *	    o	See the new information now appear with the whois command.
 *
 * This script enhances the `whois' output utilising the `whois-fields'
 * option of eggdrop 1.1-grant and later versions.  It adds the functionality
 * of whois.tcl used in pre-1.1-grant versions.
 *
 * The fields desired to be maintained in the userfile `xtra' information
 * should be put in `userinfo-fields'.  This is different than the Eggdrop
 * configuration variable `whois-fields' in that this script will add the
 * commands to change these fields.   It will also add these desired fields
 * to the `whois-fields' itself as long as they're not already defined there.
 * The fields added in `userinfo-fields' will be converted to upper case for
 * aesthetics in the `whois' command output.
 *
 * The commands that will be added to the running eggdrop are:
 *  (<info> will be the respective userfile field added in `userinfo-fields')
 *
 *	TYPE	COMMAND		USAGE
 *	======	==============	========================================
 *	msg	<info>		To change your <info> via /MSG.
 *	dcc	.<info>		To change your <info> via DCC.
 *	dcc	.ch<info>	To change someone else's <info> via DCC.
 *
 * Currently supported fields and commands:
 *
 *	FIELD	USAGE
 *	=====	=====================
 *	URL	WWW page URL
 *	IRL	In Real Life name
 *	BF	Boyfriend
 *	GF	Girlfriend
 *	DOB     Birthday (Date Of Birth)
 *       EMAIL   Email address
 * 
 */

#define MAKING_USERINFO
#include "Socket.h"
#include "BindTable.h"
#include "HelpReference.h"
#include "UserRecord.h"
#include "UserEntryXtra.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "Command.h"
#include "ModuleEntryServer.h"
#include "ModuleEntryUserinfo.h"
#include "InfoField.h"
#ifdef HAVE_SERVER_H
#include "server.h"
static Version required_server ( 2, 0 );
#endif
ModuleEntryUserinfo * UI;
static ModuleEntryEggdrop * G;
#define CLASS ""

/***************************************************************************
 * actualy interaction with user commands                                  *
 ***************************************************************************/

static void userinfo_norm (UserRecord * u, char * buf,
			   const char * mask, char * par) {
   FUNCTION;
   UserEntryXtra * uex = GetXtra(u);
   XtraKeys * xk = uex->FindKey(mask);

   if (!strcasecmp(par, "none")) {
      if (xk)
	xk->Set(NULL);
      G->SPrintF(buf, 511, UI->Lang(0x001, "s", mask));
   } else if (par[0]) {
      if (!xk)
	xk = uex->GetKey(mask);
      if (strlen(par) > 160)
	par[160] = 0;
      xk->Set(par);
      G->SPrintF(buf, 511, UI->Lang(0x002, "s", xk->Data()));
   } else if (xk)
	G->SPrintF(buf, 511, UI->Lang(0x003, "s", xk->Data()));
   else
     G->SPrintF(buf, 511, UI->Lang(0x004, "s", mask));
}


#ifdef HAVE_SERVER_H
static BindReturn userinfo_msg (void ** args, const char * mask) {
   FUNCTION;
   UserRecord * u = (UserRecord *)args[2];
   char buf[512], *par = (char *)args[3];
   
   if (u && UI->SERVER) {
      userinfo_norm (u, buf, mask, par);
      UI->SERVER->SendNotice(Q_HELP, (char *)args[0], buf);
      G->LOG_MISC.Put(0, "(%s!%s) !%s! %s %s", args[1], args[2], 
		       u->Handle(), mask, par[0] ? "..." : "");
      return EXECUTED;
   }
   return NOMATCH;
}
#endif

static BindReturn userinfo_dcc (void ** args, const char * mask) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char buf[512], * par = (char *)args[1];
   
   userinfo_norm (idx->User(), buf, mask, par);
   G->LOG_MISC.Put(0, "#%s# %s %s", idx->Nick(), mask, par[0] ? "..." : "");
   idx->Send(buf);
   return EXECUTED;
}

static BindReturn userinfo_dcc_ch (void ** args, const char * mask) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   char * par = (char *)args[1], buf[512], * who;
   UserRecord * u;
   UserEntryXtra * uex;
   XtraKeys * xk;
   
   buf[0] = 0;
   who = G->Split(&par);
   mask += 2;
   if (!who[0]) 
     G->SPrintF(buf, 511, UI->Lang(0x005, "ss", mask, mask));
   else if (!(u = G->userlist->Find(who)))
     G->SPrintF(buf, 511, UI->Lang(0x006, "s", who));
   else {
      uex = GetXtra(u);
      xk = uex->FindKey(mask + 2);
      if (!strcasecmp(par, "none")) {
	 G->LOG_MISC.Put(0, "#%s# ch%s %s none", idx->Nick(), mask, who);
	 if (xk)
	   xk->Set(NULL);
	 G->SPrintF(buf, 511, UI->Lang(0x007, "ss", who, mask));
      } else if (par[0]) {
	 if (!xk)
	   xk = uex->GetKey(mask);
	 if (strlen(par) > 160)
	   par[160] = 0;
	 xk->Set(par);
	 G->LOG_MISC.Put(0, "#%s# ch%s %s ...", idx->Nick(), mask, who);
	 G->SPrintF(buf, 511, UI->Lang(0x002, "s", xk->Data()));
      } else {
	 G->LOG_MISC.Put(0, "#%s# ch%s %s", idx->Nick(), mask, who);
	 if (xk)
	   G->SPrintF(buf, 511, UI->Lang(0x003, "s", xk->Data()));
	 else
	   G->SPrintF(buf, 511, UI->Lang(0x008, "s", who, mask));
      }
   }
   if (buf[0])
     idx->Send(buf);
   return EXECUTED;
}

/***************************************************************************
 * SCRIPT COMMANDS                                                         *
 ***************************************************************************/
 
static int script_addinfofield (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, "field-name");
   if (!UI->info_fields->Find(argv[1]->AsString())) {
      NEW( InfoField (argv[1]->AsString()) );
      return SCRIPT_OK;
   } else 
     sc->SetString(UI->Lang(0x009, ""));
   return SCRIPT_ERROR;
}

static int script_delinfofield (ScriptContext * sc, int argc, ScriptArgument ** argv) {
   FUNCTION;
   context;
   BADARGS(2, 2, "field-name");
   
   if (!UI->info_fields->Find(argv[1]->AsString())) {
      sc->SetString(UI->Lang(0x010, ""));
      return SCRIPT_ERROR;
   }
   UI->info_fields->Remove(argv[1]->AsString());
   return SCRIPT_OK;
}

static int script_infofields (ScriptContext * sc, int argc, ScriptArgument **) {
   FUNCTION;
   InfoField * ptr;
   const char * list;
   
   context;
   BADARGS(1, 1, "");
   for (ptr = UI->info_fields->Next(); ptr; ptr = ptr->Next()) 
     sc->AppendArray(1, &(list = ptr->Key()));
   return SCRIPT_OK;
}

/************************************************************************
 * MISC module stuff                                                    *
 ************************************************************************/

#ifdef HAVE_SERVER_H
static BindReturn userinfo_load_server (void ** args, const char *) {
   FUNCTION;
   UI->SERVER = (ModuleEntryServer *)(args ? args[0]
    : G->module_list->Find("server", &required_server));
   if (!UI->SERVER || !(*UI->SERVER->Vers() <= required_server))
     UI->SERVER = NULL;
   return EXECUTED;
};

static BindReturn userinfo_unld_server (void **, const char *) {
   FUNCTION;
   UI->SERVER = NULL;
   return EXECUTED;
};

static command_binds my_load[] = {
   { "server", "", userinfo_load_server, "userinfo:server" },     
};

static command_binds my_unld[] = {
   { "server", "", userinfo_unld_server, "userinfo:server" },
};
#endif

extern "C" {
   char * userinfo_start (ModuleEntryEggdrop *, ModuleEntry *);
};

char * userinfo_start (ModuleEntryEggdrop * F, ModuleEntry *me) {
   FUNCTION;
   char *r = 0;

   G = F;
   context;
   UI = NEW( ModuleEntryUserinfo ( me, &r ) );
   if (!r) {
      UI->userinfo_dcc = userinfo_dcc;
      UI->userinfo_msg = userinfo_msg;
      UI->userinfo_dcc_ch = userinfo_dcc_ch;
      NEW( InfoField("URL") );
      NEW( InfoField("IRL") );
      NEW( InfoField("BF") );
      NEW( InfoField("GF") );
      NEW( InfoField("DOB") );
      NEW( InfoField("EMAIL") );
#ifdef HAVE_SERVER_H
      UI->SERVER = (ModuleEntryServer *)
	G->module_list->Find("server", &required_server);
      G->bt_load->AddBinds(my_load,1, UI);
      G->bt_unld->AddBinds(my_unld,1, UI);
#endif
      NEW( Command ( "addinfofield", script_addinfofield, UI ) );
      NEW( Command ( "delinfofield", script_delinfofield, UI ) );
      NEW( Command ( "infofields", script_infofields, UI ) );
   }
   return r;
}
