/*
 * ModuleEntryWire.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "ModuleEntryWire.h"
#include "HelpReference.h"
#include "Socket.h"
#include "WireEntry.h"
#define CLASS "WireEntry"

/* version info */
static Version wire_ver, required_botnet, required_eggdrop, required_encryption;
static struct patches wire_patches[] = {
#include "patch.h"
};

ModuleEntryWire::ModuleEntryWire ( ModuleEntry * o, char ** r )
: ModuleEntry ( "wire", wire_ver.Set(2,0), wire_patches, o ) {
   FUNCTION;
   WIRE = this;
   if (!Depend("eggdrop", required_eggdrop.Set(2,0)))
     *r = "Your eggdrop core version is incompatible";
   else if (!(BOTNET = (ModuleEntryBotnet *) Depend("botnet", 
						    required_botnet.Set(2,0))))
     *r = "You need the botnet module to use the wire module.";
   else if (!(ENCRYPT = (ModuleEntryEncryption *) 
	      Depend("encryption", required_encryption.Set(2,0))))
     *r = "How on earth did you get the bot started without encryption!?";
   else {
      NEW ( HelpReference("wire.help", this) );
      memtype_wire = NEW ( MemType("wires", (List **)&wirelist, this) ); 
      wirelist = NEW ( WireEntry () );
   }
}

ModuleEntryWire::~ModuleEntryWire ( void ) {
   FUNCTION;
   WireEntry * w;

   context;
   if (wirelist) {
      while ((w = wirelist->Next())) {
	 Socket * idx = G->socklist->Find(w->Sock());
	 if (idx) {
	    idx->Send(WIRE->Lang(0x01B, ""));
	    idx->Send(WIRE->Lang(0x00F, ""));
	 }
	 w->Leave();
      }
      DELETE ( wirelist );
   }
}

int ModuleEntryWire::ExpMem ( void ) {
   FUNCTION;
   
   return ModuleEntry::ExpMem() + sizeof(ModuleEntryWire);
}

void ModuleEntryWire::Report ( Socket * sock, int details ) {
   FUNCTION;
   context;
   if (details)
     sock->Send(WIRE->Lang(0x01A, "dd", wirelist->Count(), wirelist->ExpMem()));
}
