/*
 * Wire.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "BindTable.h"
#include "Bot.h"
#include "IntList.h"
#include "WireEntry.h"
#include "ModuleEntryEncryption.h"
#include "ModuleEntryWire.h"
#include "Socket.h"
#define CLASS "WireEntry"

extern BindReturn wire_filter (void **, const char *);

WireEntry::WireEntry ( Socket * idx, char * nkey ) :
IntList ( *WIRE->wirelist, idx->FD(), IntList::NODUPS ) {
   FUNCTION;
   char *enctmp, wirecmd[512], wiretmp[512];
   Bot * ptr = NULL;
   WireEntry * w = WIRE->wirelist;
   Socket * idx2 = 0;
   
   context;
   ckey = (char *) my_malloc(strlen(nkey)+1,WIRE->memtype_wire);
   strcpy(ckey, nkey);

   context;
   enctmp = WIRE->ENCRYPT->Encrypt(ckey, "wire");
   crypt = (char *) my_malloc(strlen(enctmp)+1,WIRE->memtype_wire);
   strcpy(crypt, enctmp);
   G->SPrintF(wirecmd, 511, "!wire%s", enctmp);
   my_free(enctmp,G->memtype_misc);

   enctmp = WIRE->ENCRYPT->Encrypt(ckey, 
				    WIRE->Lang(0x016, "ss", idx->Nick(), ckey));
   G->SPrintF(wiretmp, 511, "%s %s", G->botnetnick, enctmp);
   ptr->SendZapfBroad(G->botnetnick, wirecmd, wiretmp);
   my_free(enctmp,G->memtype_misc);

   while ((w = w->Next())) 
     if (!strcmp(ckey, w->Key()) && 
	 (idx2 = G->socklist->Find(w->Sock())))
       idx2->Send(WIRE->Lang(0x017, "ss",idx->Nick(),ckey)); 
   if (w != this)
     enctmp = NULL;
   if (enctmp) { /* Someone else is NOT using this key, so we add a bind */
      wire_bot[0].name = wirecmd;
      wire_bot[0].flags = "";
      wire_bot[0].func = wire_filter;
      WIRE->BOTNET->bt_bot->AddBinds(wire_bot,1, WIRE);
   }
}

WireEntry::~WireEntry ( void ) {
   FUNCTION;
   if (crypt)
     my_free(crypt,WIRE->memtype_wire);
   if (ckey)
     my_free(ckey,WIRE->memtype_wire);
}

WireEntry * WireEntry::Find ( const char * cryptkey ) {
   FUNCTION;
   WireEntry * ptr = Next();
   while (ptr && strcmp(cryptkey, ptr->crypt))
     ptr = ptr->Next();
   return ptr;
}

int WireEntry::ExpMem ( void ) {
   FUNCTION;
   WireEntry * ptr = this;
   int tot = sizeof(WireEntry);
   
   while ((ptr = ptr->Next()))
     tot += sizeof(WireEntry) + strlen(crypt) + strlen(ckey) + 2;
   return tot;
}
   
void WireEntry::Leave ( void ) {
   FUNCTION;
   char *enctmp = "", wirecmd[512], wiretmp[512];
   Bot * ptr = NULL;
   WireEntry *w2;
   Socket * idx = G->socklist->Find(key);
   
   context;
   if (idx) {
      enctmp = WIRE->ENCRYPT->Encrypt(ckey, "wire");
      G->SPrintF(wirecmd, 511, "!wire%s", enctmp);
      my_free(enctmp,G->memtype_misc);
      
      enctmp = WIRE->ENCRYPT->Encrypt(ckey, WIRE->Lang(0x018, "s", idx->Nick()));
      G->SPrintF(wiretmp, 511, "%s %s", wirecmd, G->botnetnick);
      ptr->SendZapfBroad(G->botnetnick, wiretmp, enctmp);
      my_free(enctmp,G->memtype_misc);  
   }
   
   w2 = WIRE->wirelist;
   while ((w2 = w2->Next())) 
     if (!strcmp(w2->ckey, ckey)) {
	if (idx)
	  G->socklist->Find(w2->key)->Send(WIRE->Lang(0x019, "s", idx->Nick()));
	enctmp = NULL;
     }
   /* Check to see if someone else is using this wire key.   If so, */
   /* then don't remove the wire filter binding */
   if (!enctmp) { /* Someone else is NOT using this key */
      wire_bot[0].name = wirecmd;
      wire_bot[0].flags = "";
      wire_bot[0].func = wire_filter;
      WIRE->BOTNET->bt_bot->RemBinds(wire_bot,1);
   }
   WIRE->wirelist->Remove(this);
   DELETE( this );
}
