/*
 * wire.c   - An encrypted partyline communication.
 *            Compatible with wire.tcl.
 *
 *            by ButchBub - Scott G. Taylor (staylor@mrynet.com) 
 *
 *      REQUIRED: Eggdrop Modules (see dependancies below).  
 *		  Wire will only compile and work on the latest Required
 *		  Eggdrop version listed below.
 *
 *    Version   Date            Req'd Eggver	Notes			 Who
 *    .......   ..........      ............    ....................     ......
 *	1.0	1997-07-17	1.2.0		Initial.		 BB
 *	1.1	1997-07-28	1.2.0		Release version.	 BB
 *      1.2     1997-08-19      1.2.1		Update and bugfixes.	 BB
 *      1.3     1997-09-24      1.2.2.0		Reprogrammed for 1.2.2	 BB
 *      1.4     1997-11-25      1.2.2.0         Added language addition  Kirk
 *      1.1.0   1997-12-15      1.4.0.0         upgraded to 1.4          Beldin
 *
 *      2.0     1998-08-31      2.0.0           new eggdrop module code Beldin/
 *                                                                       Kirk
 *
 *   This module does not support wire usage in the files area.
 *
 */

#include "Bot.h"
#include "DccChat.h"
#include "BindTable.h"
#include "HelpReference.h"
#include "TandemBot.h"
#include "ModuleEntry.h"
#include "ModuleEntryEncryption.h"
#include "ModuleEntryBotnet.h"
#include "ModuleEntryWire.h"
#include "Socket.h"
#include "WireEntry.h"
#include "botnet.h"
#include <time.h>
#define CLASS ""

static ModuleEntryEggdrop * G;
ModuleEntryWire * WIRE;
command_binds wire_bot[] = {
   {0, 0, 0, 0}, /* Saves having to malloc :P */
};

static void wire_display (Socket * idx, const char *key, char *from, char *message);

BindReturn wire_filter (void ** args, const char * wirecrypt) {
   FUNCTION;
   char * from = (char *)args[0], buf[512], * param = buf, *wirereq, *wirewho;
   Bot * ptr = 0;
   DccChat * idx = 0;
   char workarea[512], *enctmp;
   int len;
   WireEntry * w;
   TandemBot * tb;
   
   context;
   wirecrypt += 5;
   strcpy(param, (char *)args[1]);
   wirereq = G->Split(&param);

/*
 * !wire<crypt"wire"> !wirereq <destbotsock> <crypt"destbotnick">
 * -----  wirecrypt    wirereq    wirewho         param
 */

   if (!strcmp(wirereq, "!wirereq")) {
      context;
      wirewho = G->Split(&param);
      len = (HANDLEN - strlen(G->botnetnick));
      if ((w = WIRE->wirelist->Find(wirecrypt))) {
	 if ((idx = (DccChat *)G->socklist->Find(w->Sock(),G->DCC_CHAT))) {
	    if (idx->Age() > 300) {
	       int Days, hrs, mins;
	       Days = idx->Age() / 86400;
	       hrs = (idx->Age() - (Days * 86400)) / 3600;
	       mins = (idx->Age() - (hrs * 3600)) / 60;
	       enctmp = WIRE->Lang((Days > 0) ? 0x001 : (hrs > 0) ? 0x002 :
				0x003, "ddd", Days, hrs, mins);
	    } else
	      enctmp = 0;
	    enctmp = WIRE->ENCRYPT->Encrypt(w->Key(), 
					     WIRE->Lang(0x004, "cs s ss", 
						     idx->GetIcon(),idx->Nick(), 
						     HANDLEN-strlen(idx->Nick()),
						     G->botnetnick, len, idx->Host(),
						     enctmp));
	    G->SPrintF(workarea, 511, "!wire%s !wireresp %s %s %s", wirecrypt,
		       wirewho, param, enctmp);
	    my_free(enctmp,G->memtype_misc);
	    if ((tb = WIRE->BOTNET->tandbot->Find(from)) && (ptr = tb->NextBot())) {
	       ptr->SendZapf(G->botnetnick, from, workarea);
	       if(idx->Away()) {
		  enctmp = WIRE->ENCRYPT->Encrypt(w->Key(), 
						   WIRE->Lang(0x005, "s",
							   idx->Away()));
		  G->SPrintF(workarea, 511, "!wire%s !wireresp %s %s %s",
			     wirecrypt, wirewho, param, enctmp);
		  my_free(enctmp,G->memtype_misc);
		  ptr->SendZapf(G->botnetnick, from, workarea);
	       }
	    }
	 }
      }
   } else if (!strcmp(wirereq, "!wireresp")) {
      context;
      len = atoi(G->Split(&param));
      wirewho = G->Split(&param);
      if ((w = WIRE->wirelist->Find(len))) {
	 enctmp = WIRE->ENCRYPT->Decrypt(w->Key(), wirewho);
	 idx = (DccChat *)G->socklist->Find(len, G->DCC_CHAT);
	 if (idx && !strcmp(idx->Nick(), enctmp)) {
	    my_free(enctmp,G->memtype_misc);
	    enctmp = WIRE->ENCRYPT->Decrypt(w->Key(), param);
	    idx->Printf("%s\n", enctmp);
	 }
	 my_free(enctmp,G->memtype_misc);
      }
   } else {
      context;
      if ((w = WIRE->wirelist->Find(wirecrypt)) && 
	  (idx = (DccChat *)G->socklist->Find(w->Sock(),G->DCC_CHAT)))
	wire_display(idx, w->Key(), wirereq, param);
   }
   return EXECUTED;
}

static void wire_display (Socket * idx, const char *key, char *from, char *message) {
   FUNCTION;
   char * enctmp;
   
   context;
   enctmp = WIRE->ENCRYPT->Decrypt(key, message);
   idx->Send(WIRE->Lang(from[0] == '!' ? 0x006 : 0x007, "ss",
		     from[0] == '!' ? from + 1 : from, enctmp));
   my_free(enctmp,G->memtype_misc);
}

static BindReturn cmd_wirelist (void ** args, const char *) {
   FUNCTION;
   WireEntry *w = WIRE->wirelist;
   int entry = 0;
   Socket * idx = (Socket *)args[0], *idx2;
   
   context;
   idx->Send(WIRE->Lang(0x008, "X", WIRE->wirelist));
   while ((w = w->Next()))
     if ((idx2 = G->socklist->Find(w->Sock())))
       idx->Send(WIRE->Lang(0x009, "dXXdX", ++entry, w, idx2, w->Sock(),
			 w->Next()));
   return EXECUTED;
}

static BindReturn cmd_onwire (void ** args, const char *) {
   FUNCTION;
   WireEntry *w, *w2;
   Bot * ptr = NULL;
   char tmp[512], tmp2[512], *enctmp;
   DccChat * idx = (DccChat *)args[0], * idx2;
   int len;
   
   context;
   if (!(w = WIRE->wirelist->Find (idx->FD()))) 
     idx->Send(WIRE->Lang(0x00A, ""));
   else {
      idx->Send(WIRE->Lang(0x00B, "s", w->Key()));
      idx->Send(WIRE->Lang(0x00C, " ", HANDLEN - 9)); 
      idx->Send(WIRE->Lang(0x00D, " ", HANDLEN - 9)); 
      
      enctmp = WIRE->ENCRYPT->Encrypt(w->Key(), idx->Nick());
      G->SPrintF(tmp2, 511, "!wire%s", w->Crypt());
      G->SPrintF(tmp, 511, "!wirereq %d %s", idx->FD(), enctmp);
      ptr->SendZapfBroad(G->botnetnick, tmp2, tmp);
      my_free(enctmp,G->memtype_misc);
      
      len = HANDLEN - strlen(G->botnetnick);
      w2 = WIRE->wirelist;
      while ((w2 = w2->Next())) {
	 if (!strcmp(w2->Key(), w->Key())) {
	    if ((idx2 = (DccChat *)G->socklist->Find(w2->Sock(), G->DCC_CHAT))) {
	       if (idx2->Age() > 300) {
		  int Days, hrs, mins;
		  Days = idx2->Age() / 86400;
		  hrs = (idx2->Age() - (Days * 86400)) / 3600;
		  mins = (idx2->Age() - (hrs * 3600)) / 60;
		  enctmp = WIRE->Lang((Days > 0) ? 0x001 : (hrs > 0) ? 0x002 :
				   0x003, "ddd", Days, hrs, mins);
	       } else
		 enctmp = 0;
	       idx->Send(WIRE->Lang(0x004, "cs s ss", idx2->GetIcon(),
				 idx2->Nick(), HANDLEN-strlen(idx2->Nick()),
				 G->botnetnick, len, idx2->Host(), enctmp));
	       if(idx2->Away()) 
		 idx->Send(WIRE->Lang(0x005, "s", idx2->Away()));
	    }
	 }
      }
   }
   return EXECUTED;
}

static BindReturn cmd_wire (void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[0];
   WireEntry *w;
   int lang = 0;
   
   context;
   if(!((char *)args[1])[0])
     lang = 0x00E;
   else {
      w = WIRE->wirelist->Find(idx->FD());
      if (!strcasecmp((char *)args[1], "OFF")) {
	 if (w)
	   w->Leave();
	 lang = w ? 0x00F : 0x010;
      } else if(!strcasecmp((char *)args[1], "info")) {
	 if (w) 
	   idx->Send(WIRE->Lang(0x015, "s", w->Key()));
	 else
	   lang = 0x010;
      } else {
	 if (w) {
	    lang = 0x011;
	    w->Leave();
	 } else 
	   for (lang = 0x012; lang < 0x014; lang++)
	     idx->Send(WIRE->Lang(lang, ""));
	 NEW ( WireEntry ( idx, (char *)args[1]) );
	 cmd_onwire(args,"");
      }
   }
   if (lang)
     idx->Send(WIRE->Lang(lang, "s", args[1]));
   return EXECUTED;
}

static BindReturn chof_wire (void ** args, const char *) {
   FUNCTION;
   WireEntry * w = WIRE->wirelist->Find(((Socket *)args[0])->FD());
   if (w)
     w->Leave();
   return EXECUTED;
}

static BindReturn cmd_putwire ( void ** args, const char *) {
   FUNCTION;
   Socket * idx = (Socket *)args[1];
   char * message = (char *)args[0], tmp[512], *enctmp, wirecmd[512];
   Bot * ptr = NULL;
   WireEntry * w = WIRE->wirelist->Find(idx->FD()), *w2 = WIRE->wirelist;

   context;
   if (w && message[1])  {
      enctmp = WIRE->ENCRYPT->Encrypt(w->Key(), "wire");
      G->SPrintF(wirecmd, 511, "!wire%s", enctmp);
      my_free(enctmp,G->memtype_misc);
      
      if ((strlen(message) > 3) && !strncmp(message + 1, "me ", 3)) {
	 enctmp = WIRE->ENCRYPT->Encrypt(w->Key(), message + 4);
	 G->SPrintF(tmp, 511, "!%s@%s %s", idx->Nick(),G->botnetnick,enctmp);
      } else {
	 enctmp = WIRE->ENCRYPT->Encrypt(w->Key(), message + 1);
	 G->SPrintF(tmp, 511, "%s@%s %s", idx->Nick(),G->botnetnick,enctmp);
      }
      ptr->SendZapfBroad(G->botnetnick, wirecmd, tmp);	  
      strcpy(tmp + ((tmp[0] == '!') ? 1 : 0), idx->Nick());
      while ((w2 = w2->Next())) 
	if (!strcmp(w2->Key(), w->Key())
	    && (idx = G->socklist->Find(w2->Sock())))
	  wire_display(idx, w2->Key(), tmp, enctmp);
      my_free(enctmp,G->memtype_misc);
   }
   return (BindReturn)(int)"";
}

static command_binds wire_dcc[] = {
   {"wire", "", cmd_wire, 0},
   {"onwire", "", cmd_onwire, 0},
   {"wirelist","n", cmd_wirelist, 0},
};

static command_binds wire_chof[] = {
   {"*", "", chof_wire, "wire:chof" },
};

static command_binds wire_filt[] = {
     {";*", "", cmd_putwire, "wire:filt"},
};

extern "C" {
   char *wire_start (ModuleEntryEggdrop *, ModuleEntry *);
};

char *wire_start (ModuleEntryEggdrop * F, ModuleEntry * me) {
   FUNCTION;
   char * r = 0;

   G = F;
   context;
   WIRE = NEW( ModuleEntryWire ( me, &r ) );
   if (!r) {
      G->bt_dcc->AddBinds(wire_dcc,3, WIRE);
      G->bt_filt->AddBinds(wire_filt,1, WIRE);
      G->bt_chof->AddBinds(wire_chof,1, WIRE);
   }
   return r;
}
