/*
 * Bot.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Bot.h"
#include "BotLink.h"
#include "TandemBot.h"
#include "BindTable.h"
#include "PartyLiner.h"
#include "DccChat.h"
#include "Channel.h"
#include "UserRecord.h"
#include "UserEntryXtra.h"
#include "UserEntryPass.h"
#include "UserEntryLaston.h"
#include "HelpReference.h"
#include "BotFork.h"
#define CLASS "Bot"

static char OBUF[512], TBUF[512];

Bot::Bot ( BotFork & other, const char * t ) 
: Socket ( other, t, DCT_BOT, &BNET->bot_timeout) {
   FUNCTION;
   if (other.crosslink) { 
      other.crosslink->crosslink = this;
      crosslink = other.crosslink;
   }     
}

Bot::Bot (const char * t, Bot & old  )
: Socket ( old, t, DCT_BOT, &BNET->bot_timeout) {
   FUNCTION;
   strcpy(version, old.version);
   port = old.port;
   numver = old.numver;
   strcpy(linker, old.linker);
   call_version = old.call_version;
   bot_table = old.bot_table;
   size = old.size;
   if (old.crosslink) { 
      old.crosslink->crosslink = this;
      crosslink = old.crosslink;
   }
}

Bot::Bot (BotLink & old, char * par, char * v, int b) 
: Socket ( old, BNET->BOT, DCT_BOT | DCT_REALBOT, 0) {
   FUNCTION;
   char * x =  BNET->Lang(0x001, "s", nick);
   TandemBot * tb = 0, * up = 0;
   
   context;
   call_version = v;
   bot_table = BNET->bot_table;
   size = BOT_TABLE_SIZE;
   strcpy(version, par);
   port = old.port;
   numver = old.numver;
   strcpy(linker, old.linker);
   SetStatus(LINKING);
   BNET->LOG_BOTS.Put(x);
   BNET->ChatOut(G->Lang(0x10C, "s", x));
   context;
   if (old.crosslink && (tb = BNET->tandbot->Find(nick))
       && (up = tb->Uplink())) {
      up->RemoveChild(tb);
      tb->peer = up->peer;
      up->peer = tb;
      tb->uplink = up->uplink;
      tb->crosslink = up;
      up->crosslink = tb;
      SendCrossLink(tb, BNET->tandbot);
      crosslink = old.crosslink;
      old.crosslink->crosslink = this;
   } else
     NEW ( TandemBot ( BNET->tandbot, nick, '-', numver, b) );
}

Bot::~Bot ( void ) {
   FUNCTION;
   if (crosslink && (crosslink->crosslink == this))
     crosslink->crosslink = 0;
}

int Bot::Dump ( char * buf, int len ) {
   FUNCTION;
   if (G->debug_output) {
      LogModes L ( LogModes::BOTNET_SEND );
   
      L.Put(BNET->Lang(0x0E1, "ss", nick, buf));
   }
   return Socket::Dump (buf, len);
}

void Bot::Eof ( void ) {
   FUNCTION;
   char * x;
   TandemBot * ptr = BNET->tandbot->Find(nick);
   
   x = BNET->Lang(0x00F, "s", nick);
   BNET->LOG_BOTS.Put(x);
   Close();
   if (ptr) 
     SendUnLinked(ptr, x, 0);
}

void Bot::Display2 (Socket * where) {
   FUNCTION;
   char buf[20];
   int i = 0;
   
   buf[i++] = HasStatus(PINGED) ? 'P' : 'p';
   buf[i++] = HasStatus(CALLED) ? 'C' : 'c';
   buf[i++] = HasStatus(WARNED) ? 'W' : 'w';
   buf[i++] = HasStatus(LEAF) ? 'L' : 'l';
   buf[i++] = HasStatus(LINKING) ? 'I' : 'i';
   buf[i++] = 0;
   where->Send(BNET->Lang(0x010,"ssss",nick,version,buf,call_version));
   if (crosslink) 
     where->Send(BNET->Lang(0x0D9, "s", crosslink->Nick()));
}

void Bot::Activity (char * msg, int i) {
   FUNCTION;
   int min = 0, max = size, y;
   LogModes LOG_BOTNET ( LogModes::BOTNET );
   
   context;
   StripTelnet(msg, &i);
   if (G->debug_output)
     LOG_BOTNET.Put(BNET->Lang(0x0C2, "ss", nick, msg));
   if (bot_table == BNET->bot_table) {
      char v = msg[0];
      
      msg++;
      while (max != min) {
	 i = (max + min) / 2;
	 if (v == bot_table[i].name[0]) {
	    (this->*(bot_table[i].func))(msg);
	    break;
	 } else if (v < bot_table[i].name[0])
	      max = i;
	 else
	   min = i+1;
      }
   } else {
      char * code = G->Split(&msg);
      
      while (max != min) {
	 i = (max + min) / 2;
	 if (!(y = strcasecmp(code, bot_table[i].name))) {
	    (this->*(bot_table[i].func))(msg);
	    break;
	 } else if (y < 0)
	      max = i;
	 else
	   min = i+1;
      }
   }
}

int Bot::ExpMem ( void ) {
   FUNCTION;
   return sizeof(Bot);
}

void Bot::SendInfoQ (char * par) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int i = G->SPrintF(OBUF, 511, "I%s\n",par);
   
   SendTandBut(OBUF,i);
   for (t = BNET->masscall_list; t; t = t->next)
     t->infoq(this, par);
}

void Bot::SendChan (PartyLiner * pl, char * data) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int i;
   
   if (BNET->tands && (pl->Chan() < GLOBAL_CHANS)) {
      i = G->SPrintF(OBUF, 511, "c%s %D %s\n",pl->BotName(),pl->Sock(),data);
      SendTandButIso(OBUF,i);
      for (t = BNET->masscall_list; t; t = t->next)
	t->chan(this, pl,data);
   }
}

void Bot::SendChatter (char * bt, int chan, char * data) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int i;
   
   if (BNET->tands && (chan < GLOBAL_CHANS)) {
      i = G->SPrintF(OBUF, 511, "e%s %D %s\n",bt,chan,data);
      SendTandButIso(OBUF,i);
      for (t = BNET->masscall_list; t; t = t->next)
	t->chatter(this, bt, chan, data);
   }
}

void Bot::SendChat (char * botnick, char * data) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int i;
   
   if (BNET->tands) {
      i = G->SPrintF(OBUF, 511, "C%s %s\n",botnick,data);
      SendTandButIso(OBUF,i);
      for (t = BNET->masscall_list; t; t = t->next)
	t->chat(this, botnick,data);
   }
}

void Bot::SendAct (PartyLiner * pl, char * data) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int i;
   
   if (BNET->tands && (pl->Chan() < GLOBAL_CHANS)) {
      i = G->SPrintF(OBUF, 511, "a%s %D %s\n",pl->BotName(),pl->Sock(),data);
      SendTandButIso(OBUF,i);
      for (t = BNET->masscall_list; t; t = t->next)
	t->act(this, pl, data);
   }
}

void Bot::SendUnLinked (TandemBot * bot, const char * args, int remote) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;

   context;
   if (bot) {
      if (args && args[0])
	BNET->ChatOut(G->Lang(remote ? 0x10D : 0x10C , "ss", 
			remote ? bot->UplinkName() : args, args));
      if (BNET->tands) {
	 if (!crosslink || !bot->crosslink || strcasecmp(nick, bot->Key())) {
	    l = G->SPrintF(OBUF, 511, "N%s %s\n", bot->Name(), args ? args:"");
	    SendTandBut(OBUF, l);
	    for (t = BNET->masscall_list; t; t = t->next)
	      t->unlinked(this, bot, args);
	    DELETE ( bot );
	 } else {
	    TandemBot * other = bot->crosslink;
	    
	    other->crosslink = 0;
	    bot->crosslink = 0;
	    BNET->tandbot->RemoveChild(bot);
	    bot->peer = other->child;
	    bot->uplink = other;
	    other->child = bot;
	    crosslink->crosslink = 0;
	    crosslink = 0;
	    SendUnCrossLink(bot, BNET->tandbot);
	 }
      }
   }
}   

void Bot::SendNLinked (const char * bot, const char * next, char flag, int vernum,
		       const char * bflgs) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;
   
   if (BNET->tands) {
      l = G->SPrintF(OBUF, 511, "n%s %s %c%D %s\n",bot,next,flag,vernum,bflgs);
      SendTandBut(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->nlinked(this, bot, next, flag, vernum, bflgs);
   }
}   

void Bot::SendZapfBroad (char * a, char * b, char * c) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;

   if (BNET->tands) {
      l = G->SPrintF(OBUF, 511, "Z%s %s%s%s\n",a,b?b:"",b?" ":"",c);
      SendTandBut(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->zapf_broad(this, a, b, c);
   }
}

void Bot::SendNkCh (PartyLiner * ptr, char * oldnick) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;
   
   if (BNET->tands > 0) {
      l = G->SPrintF(OBUF, 511, "k%s %D %s\n",ptr->BotName(),ptr->Sock(),
		     ptr->Nick());
      SendTandButIso(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->nkch(this, ptr, oldnick);
   }
}

void Bot::SendIdle (char * bot, int socki, int idle, char * away) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;
   
   context;
   if (BNET->tands > 0) {
      l = G->SPrintF(OBUF, 511, "i%s %D %D %s\n", bot, socki, idle, safe_str(away));
      SendTandButIso(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->idle(this, bot, socki, idle, away);
   }
}
   
void Bot::SendAway (PartyLiner * partyidx, int linking) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;

   if (BNET->tands) {
      l = G->SPrintF(OBUF, 511, "A%s%s %D %s\n", (this && linking) ? "!":"",
		     partyidx->BotName(), partyidx->Sock(), safe_str(partyidx->Away()));
      SendTandButIso(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->away(this, partyidx, linking);
   }
}

void Bot::SendJoin (int linking, PartyLiner * ptr, int oldchan) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;
   
   context;
   if (BNET->tands){
      l = G->SPrintF(OBUF, 511, "j%s%s %s %D %c%D %s\n", linking ? "!" : "",
		     ptr->BotName(), ptr->Nick(), ptr->Chan(), 
		     ptr->Flag(), ptr->Sock(), safe_str(ptr->From()));
      SendTandButIso(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->join(this, linking, ptr, oldchan);
   }
}

void Bot::SendPart(PartyLiner * ptr, char * reason, int silent) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;
   
   if (BNET->tands) {
      l = G->SPrintF(OBUF, 511, "J%s%s %s %D %s\n", silent ? "!" : "", 
		     ptr->BotName(), ptr->Nick(), ptr->Sock(), reason?reason:"");
      SendTandButIso(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->part(this, ptr, reason, silent);
   }  
}

void Bot::SendUpdate (TandemBot * ptr) {
   FUNCTION;
   struct massbot_call_funcs * t;
   int l;
   
   if (BNET->tands > 0) {
      l = G->SPrintF(OBUF, 511, "u%s %c%D %s\n", ptr->Name(), ptr->Share(),
		     ptr->Version(), BNET->BotnetFlags(ptr->BotnetFlags()));
      SendTandBut(OBUF,l);
      for (t = BNET->masscall_list; t; t = t->next)
	t->update(this, ptr);
   }
}

void Bot::SendBye (const char * reason) {
   FUNCTION;
   struct massbot_call_funcs * t;

   if (BNET->tands > 0) {
      SendTandBut(OBUF, G->SPrintF(OBUF, 511, "b%s\n", reason));
      for (t = BNET->masscall_list; t; t = t->next)
	t->bye(reason);
   }
}

void Bot::SendCrossLink (TandemBot * b1, TandemBot * b2) {
   FUNCTION;
   struct massbot_call_funcs * t;

   if (BNET->tands > 0) {
      SendTandBut(OBUF, G->SPrintF(OBUF, 511, "x%s %s\n", b1->Key(), 
				   b2->Key()[0] ? b2->Key() : G->botnetnick));
      for (t = BNET->masscall_list; t; t = t->next)
	if (t->crosslink)
	  t->crosslink(b1,b2);
   }
}

void Bot::SendUnCrossLink (TandemBot * b1, TandemBot * b2) {
   FUNCTION;
   struct massbot_call_funcs * t;
   
   if (BNET->tands > 0) {
      SendTandBut(OBUF, G->SPrintF(OBUF, 511, "X%s %s\n", b1->Key(), 
				   b2->Key()[0] ? b2->Key() : G->botnetnick));
      for (t = BNET->masscall_list; t; t = t->next)
	if (t->uncrosslink)
	  t->uncrosslink(b1,b2);
   }
}

void Bot::ProcessCrossLink ( char * n1, char * n2 ) {
   FUNCTION;
   TandemBot * b1 = BNET->tandbot->Find(n1);
   TandemBot * b2 = BNET->tandbot->Find(n2);
   
   if (!b1)
     SendFakeAlert("crosslink-source",n1);
   else if (!b2)
     SendFakeAlert("crosslink-desto", n2);
   else if (b1->CrossLink() && (b1->CrossLink() != b2))
     SendFakeAlert("crosslink-already-source",n1);
   else if (b2->CrossLink() && (b2->CrossLink() != b1))
     SendFakeAlert("crosslink-already-desto",n2);
   else if (!b2->CrossLink()) {
      if (!b1->Uplink() && !b2->Uplink()) {
	 Bot * bt1 = (Bot *)G->socklist->Find(n1, DCT_BOT);
	 Bot * bt2 = (Bot *)G->socklist->Find(n2, DCT_BOT);
	 
	 if (bt1 && bt2) {
	    bt1->crosslink = bt2;
	    bt2->crosslink = bt1;
	 }
	 BNET->LOG_BOTS.Put(BNET->Lang(0x0D1, "ss", n1, n2));
      } else
	BNET->LOG_BOTS.Put(BNET->Lang(0x0D2, "ss", n1, n2));
      b1->crosslink = b2;
      b2->crosslink = b1;
      if ((b1->peer != b2) && (b2->peer != b1)) {
	 TandemBot * b3 = b1->uplink ? b1->uplink->child : BNET->tandbot->child;
	 
	 if (b3 == b2) {
	    if (b1->uplink)
	      b1->uplink->child = b2->peer;
	    else
	      BNET->tandbot->child = b2->peer;
	 } else {
	    while (b3->peer != b2)
	      b3 = b3->peer;
	    b3->peer = b2->peer;
	 }
	 b2->peer = b1->peer;
	 b1->peer = b2;
      }
      SendCrossLink(b1, b2);
   }
}

void Bot::ProcessUnCrossLink (const char * n1, const char * n2 ) {
   FUNCTION;
   TandemBot * b1 = BNET->tandbot->Find(n1);
   TandemBot * b2 = BNET->tandbot->Find(n2);
   
   if (!b1)
     SendFakeAlert("uncrosslink-source",n1);
   else if (!b2)
     SendFakeAlert("uncrosslink-desto", n2);
   else if (b1->CrossLink() && (b1->CrossLink() != b2) && (b2->Uplink() != b1))
     SendFakeAlert("uncrosslink-not-source",n1);
   else if (b2->CrossLink() && (b2->CrossLink() != b1) && (b1->Uplink() != b2))
     SendFakeAlert("uncrosslink-not-desto",n2);
   else {
      if (b1->CrossLink())
	SendUnCrossLink(b1, b2);
      if (!b1->Uplink() || !b2->Uplink()) {
	 Bot * bt1 = (Bot *)G->socklist->Find(n1, DCT_BOT);
	 Bot * bt2 = (Bot *)G->socklist->Find(n2, DCT_BOT);
	 
	 if (bt1)
	   bt1->crosslink = 0;
	 if (bt2)
	   bt2->crosslink = 0;
	 BNET->LOG_BOTS.Put(BNET->Lang(0x0D3, "ss", n1, n2));
	 b1->crosslink = 0;
	 b2->crosslink = 0;
      } else {
	 BNET->LOG_BOTS.Put(BNET->Lang(0x0D4, "ss", n1, n2));
	 if (b1->Uplink() == b2->Uplink()) {
	    b1->crosslink = 0;
	    b2->crosslink = 0;
	 } else {
	    TandemBot * other;
	    
	    if (b1->Uplink() == b2) {
	       other = b2;	       
	       b2 = b1;
	       b1 = other;
	    }
	    other = b2->crosslink;
	    b1->RemoveChild(b2);
	    b2->uplink = other;
	    b2->crosslink = 0;
	    b2->peer = other->child;
	    other->child = b2;
	    other->crosslink = 0;
	 }
      }
   }
}

void Bot::ProcessEndLink (char * ) {
   FUNCTION;
   context;
   ClearStatus(LINKING);
}

void Bot::ProcessPing (char * ) {
   FUNCTION;
   context;
   SendPong();
}

void Bot::ProcessPong (char * ) {
   FUNCTION;
   context;
   ClearStatus(PINGED);
}

void Bot::ProcessChatter (char * bot, int chan, char * msg) {
   FUNCTION;
   void * args[3];
   FlagRecord bfl (FlagRecord::BOT);
   TandemBot * tb;
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (tb->NextBot() != this))
	SendFakeAlert("chatter-direction",bot);
      else if ((chan >= 0) && (chan <= GLOBAL_CHANS)) {
	 DccChat::Chanout(chan, G->Lang(0x10D, "ss", bot, msg));
	 SendChatter(bot, chan, msg);
	 args[0] = bot;
	 args[1] = &chan;
	 args[2] = msg;
	 G->bt_bcst->Check(msg, 0, args);
      }
   }
}

void Bot::ProcessChan (char * bot, int socki, char * msg) {
   FUNCTION;
   void * args[3];
   char tbuf[NOTENAMELEN];
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   PartyLiner * pl;
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (tb->NextBot() != this))
	SendFakeAlert("chan-direction",bot);
      else if (!(pl = tb->FindParty(socki))) 
	SendFakeAlert("chan-user",G->ToBase10(socki));
      else {
	 pl->TimeNow();
	 G->SPrintF(tbuf, NOTENAMELEN, "%s@%s", pl->Nick(), bot);
	 DccChat::Chanout(pl->Chan(), G->Lang(0x10F, "ss", tbuf, msg));
	 SendChan(pl, msg);
	 args[0] = tbuf;
	 args[1] = &(socki = pl->Chan());
	 args[2] = msg;
	 G->bt_chat->Check(msg, 0, args);
      }
   }
}
   
void Bot::ProcessChat (char * from, char * msg) {
   FUNCTION;
   TandemBot * tb;
   
   if (strchr(from, '@')) 
     SendFakeAlert("chat-bot",from);
   else if (!(tb = BNET->tandbot->Find(from)) || (tb->NextBot() != this))
     SendFakeAlert("chat-direction",from);
   else {
      BNET->ChatOut(G->Lang(0x10D, "ss", from, msg));
      SendChat(from,msg);
   } 
}

void Bot::ProcessActChan(char * bot, int socki, char * msg) {
   FUNCTION;
   char tbuf[NOTENAMELEN];
   void * args[3];
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   PartyLiner * pl;
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {   
      if (!(tb = BNET->tandbot->Find(bot)) && (tb->NextBot() != this)) 
	SendFakeAlert("actchan-direction",bot);
      else if (!(pl = tb->FindParty(socki))) 
	SendFakeAlert("actchan-user@bot",G->ToBase10(socki));
      else {
	 pl->TimeNow();
	 G->SPrintF(tbuf, NOTENAMELEN, "%s@%s", pl->Nick(), bot);
	 DccChat::Chanout(pl->Chan(), G->Lang(0x110, "ss", tbuf, msg));
	 SendAct(pl, msg);
	 args[0] = tbuf;
	 args[1] = &(socki = pl->Chan());
	 args[2] = msg;
	 G->bt_act->Check(msg, 0, args);
      }
   }
}

void Bot::ProcessPriv (char * from, char * frombot, char * to, char * tobot,
		       char * par) {
   FUNCTION;
   int i, lang = 0;
   Bot * next = 0;
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {      
      if ((to[0] == '!') || (to[0] == '#'))
	to++;
      if (!(tb = BNET->tandbot->Find(frombot)) && (tb->NextBot() != this))
	SendFakeAlert("priv-direction",frombot);
      else if (to[0]) {
	 if (!strcasecmp(tobot, G->botnetnick)) {	/* for me! */
	    if (frombot == from)
	      G->AddNote(to, from, par, NULL, -1);
	    else {
	       i = G->AddNote(to, from, par, NULL, 0);
	       if (from[0] != '@')
		 switch (i) {
		   case NOTE_ERROR:
		    lang = 0x089;
		    break;
		   case NOTE_STORED:
		    lang = 0x08A;
		    break;
		   case NOTE_FULL:
		    lang = 0x08B;
		    break;
		   case NOTE_AWAY:
		    lang = 0x08C;
		    break;
		   case NOTE_FWD:
		    lang = 0x08E;
		    break;
		   case NOTE_OK:
		    lang = 0x08D;
		    break;
		 }
	       if (lang)
		 SendPriv(G->botnetnick,from,BNET->Lang(lang,"s",to));
	    }
	 } else if ((tb = BNET->tandbot->Find(tobot)) 
		    && (next = tb->NextBot())) {
	    G->SPrintF(TBUF+128, 128, "%s@%s", to, tobot);
	    next->SendPriv(from,TBUF+128,par);
	 }
      }
   }
}

void Bot::ProcessBye (char * par) {
   FUNCTION;
   char * s = BNET->Lang(par[0] ? 0x08F : 0x090, "ss", nick, par);
   context;
   BNET->LOG_BOTS.Put(s);
   SendUnLinked (BNET->tandbot->Find(nick), s,0);
   Close();
}

void Bot::ProcessWho (char * from, char * to, int chan) {
   FUNCTION;
   char s[512], *t, C;
   int i, ok = 0;
   Channel *c;
   Socket * ptr = 0;
   Bot * bot = 0;
   TandemBot * tb;
   
   context;
   if (strcasecmp(to, G->botnetnick)) {
      if ((tb = BNET->tandbot->Find(to)) && (bot = tb->NextBot()))
	bot->SendWho(from, to, chan);
   } else {
      context;
      i = 0;
      for (c = G->chanset; c; c = c->Next())
	if (!c->Secret()) 
	  i += G->SPrintF(s + i, 511 - i, "%s, ", c->Key());
      if (i)
	s[i - 2] = 0;
      SendPriv( G->botnetnick, from, 
			  BNET->Lang(i ? 0x091 : 0x092, "ss", s, G->VersionStr()));
      if (G->admin[0])
	SendPriv( G->botnetnick, from, BNET->Lang(0x093, "s",
							       G->admin));
      if (!chan)
	SendPriv( G->botnetnick, from, 
			    BNET->Lang(0x094, "cccc", '*', '+', '%', '@'));
      else {
	 char buf[100];
	 int x = 100;
	 void * args[3] = { &chan, buf, &x };
	 BindReturn ret = G->bt_name->Check("", 0, args);
	 
	 SendPriv( G->botnetnick, from,
			     BNET->Lang((ret == EXEC_LOG) ? 0x096 : 0x095,
				     "ccccsds",  '*', '+', '%', '@',
				     (chan < GLOBAL_CHANS) ? "" : "*",
				     chan % GLOBAL_CHANS, buf));				     
      }
      for (ptr = G->socklist; (ptr = ptr->Next()); )
	if (ptr->Type() && (ptr->TypeFlags() & DCT_CHAT)
	    && (((DccChat *)ptr)->Channel() == chan)) {
	   if (ptr->Age() > 300) {
	      unsigned long days, hrs, mins;
	      days = ptr->Age() / 86400;
	      hrs = (ptr->Age() - (days * 86400)) / 3600;
	      mins = (ptr->Age() - (hrs * 3600)) / 60;
	      t = BNET->Lang((days > 0) ? 0x052 : ((hrs > 0) ? 0x053 : 0x054),
			  "ddd", days, hrs, mins);
	   } else 
	     t = 0;
	   C = ptr->GetIcon();
	   SendPriv( G->botnetnick, from,
			       BNET->Lang(0x097, "cs ss", C == '-' ? ' ' : C,
				       ptr->Nick(), 
				       HANDLEN - strlen(ptr->Nick()),
				       ptr->Host(), t));
	   if (((DccChat *)ptr)->Away())
	     SendPriv( G->botnetnick, from,
				 BNET->Lang(0x056, "s", 
					 ((DccChat *)ptr)->Away()));
	}
      for (ptr = G->socklist; (ptr = ptr->Next()); )
	if ((ptr->Type() == BNET->BOT) || 
	   !strcmp(ptr->Type(),"SHAREBOT")) {
	   if (!ok) {
	      ok = 1;
	      SendPriv(G->botnetnick,from,BNET->Lang(0x098,""));
	   }
	   SendPriv( G->botnetnick, from,
			       BNET->Lang(0x099, "scs s",	
				       ptr->HasStatus(Bot::CALLED) ? "<-" : "->",
				       ptr->Type() != BNET->BOT ? '+' : ' ',
				       ptr->Nick(), HANDLEN-strlen(ptr->Nick()),
				       ((Bot *)ptr)->Version()));
	}
      ok = 0;
      for (ptr = G->socklist; (ptr = ptr->Next()); )
	if ((ptr->TypeFlags() & DCT_CHAT) 
	    && (((DccChat *)ptr)->Channel() != chan)) {
	   if (!ok) {
	      ok = 1;
	      SendPriv(G->botnetnick,from,BNET->Lang(0x09A,""));
	   }
	   if (ptr->Age() > 300) {
	      unsigned long days, hrs, mins;
	      days = ptr->Age() / 86400;
	      hrs = (ptr->Age() - (days * 86400)) / 3600;
	      mins = (ptr->Age() - (hrs * 3600)) / 60;
	      t = BNET->Lang((days > 0) ? 0x052 : ((hrs > 0) ? 0x053 : 0x054),
			  "ddd", days, hrs, mins);
	   } else 
	     t = 0;
	   C = ptr->GetIcon();
	   SendPriv( G->botnetnick, from,
			       BNET->Lang(0x097, "cs ss", C == '-' ? ' ' : C,
				       ptr->Nick(), HANDLEN - strlen(ptr->Nick()),
				       ptr->Host(), t));
	   if (((DccChat *)ptr)->Away())
	     SendPriv( G->botnetnick, from,
				 BNET->Lang(0x056, "s", 
					 ((DccChat *)ptr)->Away()));
	}
      
   }
}

void Bot::ProcessInfoQ (char * par) {
   FUNCTION;
   char s[1024],s2[32];
   Channel *chan;
   int l = 0, day = 0, hr, min;
   time_t now;
   
   context;
   now = G->now - G->online_since;
   if (now > 86400) {
      day = (now / 86400);
      now = now % 86400;
   } 
   hr = now / 3600;
   min = (now % 3600) / 60;
   strncpyz(s2, BNET->Lang(day ? ((day == 1) ? 0x0DF : 0x0DE) : 0x0E0, "ddd",
			   hr, min, day), 31);
   s[0] = 0;
   if (G->module_list->Find("server", &BNET->required_server)) {
      s[0] = '*';
      for (chan = G->chanset; chan; chan = chan->Next())
	if (!chan->Secret()) 
	  l += G->SPrintF(s + l, 511 - l, "%s, ", chan->Key());
      if (l) 
	s[l - 2] = 0;
   }
   SendPriv( G->botnetnick, par,
		       BNET->Lang(s[0] ? (l ? 0x044 : 0x045) : 0x046,
				  "ssss", G->VersionStr(), BNET->network, s, s2));
   SendInfoQ(par);
}

void Bot::ProcessLink (char * from, char * who, char * towho) {
   FUNCTION;
   char * rfrom;
   TandemBot * tb;
   Bot * next = 0;
   
   context;
   if (!strcasecmp(who, G->botnetnick)) {
      rfrom = (rfrom = strchr(from,':')) ? rfrom + 1 : from;
      G->LOG_CMDS.Put(BNET->Lang(0x074, "ss", rfrom, towho));
      SendPriv( G->botnetnick, from, 
		  BNET->Lang(BNET->BotLink(from, NULL, towho) ? 0x09B : 0x09C,
			     "s", towho));
   } else if ((tb = BNET->tandbot->Find(who)) && (next = tb->NextBot()))
	next->SendLink(from,who,towho);
}
   
void Bot::ProcessUnLink (char * from, char * bot, TandemBot * undes, char * why) {
   FUNCTION;
   char *rfrom, *p;
   int lang = 0;
   TandemBot * tb;
   Bot * next = 0;
   
   context;
   if (!strcasecmp(bot, G->botnetnick)) {
      rfrom = (rfrom = strchr(from, ':')) ? rfrom + 1 : from;
      G->LOG_CMDS.Put(BNET->Lang(why[0] ? 0x09D : 0x076, "sss", rfrom, 
			       undes->Name(), why));
      switch (BNET->Unlink(undes->Name(),NULL, why[0] ? why : 0)) {
	case 1:
	 lang = 0x034;
	 break;
	case 0:
	 lang = 0x09F;
	 NullBot()->SendUnLinked(undes, "",0);
	 break;
	default:
	 lang = 0x0A0;
      }
      if (lang && (p = strchr(from, '@')) && 
	  (tb = BNET->tandbot->Find(p + 1)) && (next = tb->NextBot()))
	next->SendPriv(G->botnetnick, from, 
		       BNET->Lang(lang, "s", undes->Name()));
   } else if ((tb = BNET->tandbot->Find(bot)) && (next = tb->NextBot())) 
	next->SendUnLink(from, undes, why);
}

void Bot::ProcessNLinked (const char * newbot, char * next, char flag, 
			  int vernum, int bnfl) {
   FUNCTION;
   const char *p;
   char s[512];
   int bogus = 0, lang = 0;
   void * args[2];
   UserRecord * u;
   TandemBot * tb;
   FlagRecord bfl;
     
   s[0] = 0;
   if (!next[0]) 
     lang = 0x0A1;
   else if (BNET->tandbot->Find(newbot) || !strcasecmp(newbot, G->botnetnick)) 
     lang = 0x0A4;
   else {
      for (p = newbot; *p; p++)
	if ((*p < 32) || (*p == 127) || ((p - newbot) >= HANDLEN))
	    bogus = 1;
      if (!bogus && (!(tb = BNET->tandbot->Find(next))
		     || (this != tb->NextBot())))
	bogus = 1;
      if (bogus) 
	lang = 0x0A7;
      else if (user && bfl.Set(user)->BotLeaf())
	lang = 0x0AA;
      else if ((bnfl & Bot::TRAFWATCH) && BNET->dont_allow_watchers)
	lang = 0x0DA;
      else {
	 if (flag == '!') 
	   BNET->ChatOut(G->Lang(0x10D, "ss", next, 
			   BNET->Lang(0x001, "s", newbot)));
	 NEW ( TandemBot ( tb, newbot,  flag == '!' ? '-' : flag, vernum, bnfl) );
	 SendNLinked(newbot, next, flag, vernum, BNET->BotnetFlags(bnfl));
	 args[0] = (void *)newbot;
	 args[1] = next;
	 BNET->bt_link->Check(newbot, 0, args);
	 u = G->userlist->Find(newbot);
	 if (u && bfl.Set(u)->BotReject()) {
	    UserEntryXtra * uex =
	      (UserEntryXtra *)u->FindEntry(G->entrytype_xtra);
	    XtraKeys * x = uex ? uex->FindKey("reject") : 0;
	    SendReject(G->botnetnick, newbot, x ? x->Data() : "");
	    BNET->LOG_BOTS.Put(BNET->Lang(0x0AD, "ss", newbot, nick));
	 }
	 return;
      }
   }
   BNET->LOG_BOTS.Put(BNET->Lang(lang, "sss", nick, next, newbot));
   SendBye2(BNET->Lang(lang + 1, "ss", next, newbot));
   SendUnLinked (BNET->tandbot->Find(nick),
		 BNET->Lang(lang + 2, "sss", nick, next, newbot),0);
   Close();
}

void Bot::ProcessUnLinked (TandemBot * bot, char * why) {
   FUNCTION;
   Socket * next;
   
   context;
   if ((next = bot->NextBot()) == this) 
     SendUnLinked(bot,  why[0] ?
		  why : BNET->Lang(0x034, "s", bot->Name()), 1);
   else if (next) 
	SendFakeAlert("unlinked-direction",bot->Name());
}
   
void Bot::ProcessTrace (char * from, char * dest, char * chain) {
   FUNCTION;
   TandemBot * tb;
   Bot * next = 0;
   
   context;
   G->SPrintF(TBUF, 511, "%s:%s",chain,G->botnetnick);
   SendTraced(from,TBUF);
   if (strcasecmp(dest, G->botnetnick) &&
       (tb = BNET->tandbot->Find(dest)) && (next = tb->NextBot()))
     next->SendTrace(from, dest, chain);
}

void Bot::ProcessTraced (char * to, char * chain) {
   FUNCTION;
   char *p;
   int socki;
   TandemBot * tb;
   Socket * sptr;
   Bot * next;
   
   context;
   if (!(p = strchr(to, '@')))
     p = to;
   else 
     *p++ = 0;
   if (!strcasecmp(p, G->botnetnick)) {
      time_t t = 0;
      char * p = chain, *ss = TBUF;
      
      G->SplitC(ss, to, ':');
      socki =  ss[0] ? atoi(ss) : -1;
      if (chain[0] == ':') {
	 t = atoi(chain+1);
	 p = (p = strchr(chain+1,':')) ? p + 1 : chain + 1;
      }
      for (sptr = G->socklist; (sptr = sptr->Next()); )
	if ((sptr->TypeFlags() & DCT_CHAT) && !strcasecmp(sptr->Nick(), to) &&
	    ((socki == -1) || (socki == sptr->FD())))
	  sptr->Send(BNET->Lang(t ? 0x09E : 0x0AE, "sl", p, G->now - t));
   } else {
      tb = BNET->tandbot->Find(p);
      if (p != to) 
	*--p = '@';
      if ((next = tb->NextBot()))
	next->SendTraced(to,chain);
   }
}

void Bot::ProcessReject (char * from, char * par) {
   FUNCTION;
   char *who, *destbot, *frombot;
   Socket * next = 0;
   Bot * nptr = 0;
   TandemBot * tb;
   UserRecord * u;
   FlagRecord bfl (FlagRecord::BOT);
   Bot * nxt;
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      frombot = (frombot = strchr(from, '@')) ? frombot + 1 : from;
      if (!(tb = BNET->tandbot->Find(frombot)) && (this != tb->NextBot()))
	SendFakeAlert("reject-direction",frombot);
      else {
	 who = G->Split(&par);
	 if (!(destbot = strchr(who, '@'))) {
	    /* rejecting a bot */
	    if (!(tb = BNET->tandbot->Find(who)) && !(nptr = tb->NextBot()))
	      SendPriv( G->botnetnick, from,
				  BNET->Lang(0x0B0, "s", who));
	    else if (!strcasecmp(next->Nick(), who)) {
	       /* i'm the connection to the rejected bot */
	       BNET->LOG_BOTS.Put(BNET->Lang(0x0B1, "ss", from, next->Nick()));
	       BNET->Unlink(who, NULL, BNET->Lang(par[0] ? 0x0B2 : 0x0B3, "sss",
						  next->Nick(), from, par));
	    } else 
	      nptr->SendReject(from, who, par);
	 } else {			/* rejecting user */
	    *destbot++ = 0;
	    if (!strcasecmp(destbot, G->botnetnick)) {
	       /* kick someone here! */
	       int ok = 0;
	       switch (BNET->remote_boots) {
		 case 1:
		  frombot = (frombot = strchr(from, '@')) ? frombot + 1 : from;
		  if (!(u = G->userlist->Find(frombot))
		      || !bfl.Set(u)->BotShare()) {
		     SendPriv( G->botnetnick, from,
					 BNET->Lang(0x0B4, ""));
		     ok = 1;
		  }
		  break;
		 case 0:
		  SendPriv( G->botnetnick, from,
				      BNET->Lang(0x0B5, ""));
		  ok = 1;
		  break;
	       }
	       bfl.SetMatch(FlagRecord::GLOBAL);
	       for (next = G->socklist; (next = next->Next()); )
		 if (!strcasecmp(who, next->Nick()) && 
		     (next->TypeFlags() & DCT_CHAT)) {
		    if (next->User() && bfl.Set(next->User())->GlobOwner()) {
		       SendPriv( G->botnetnick, from,
					   BNET->Lang(0x0B6, ""));
		       return;
		    }
		    ((DccChat *)next)->DoBoot(from, par);
		    G->LOG_CMDS.Put(BNET->Lang(0x05B, "ssss", from, 
						next->Nick(), destbot, par));
		    break;
		 }
	    } else {
	       tb = BNET->tandbot->Find(destbot);
	       *--destbot = '@';
	       if (tb && (nxt = tb->NextBot()))
		 nxt->SendReject(from, who, par);
	    }
	 }
      }
   }
}

/* thisbot <botname> */
void Bot::ProcessThisBot (char * par) {
   FUNCTION;
   char * bot = G->Split(&par), * bnfl = G->Split(&par);
   context;
   if (strcasecmp(bot, nick)) {
      BNET->LOG_BOTS.Put(BNET->Lang(0x0B7, "ss", nick, bot));
      SendBye(BNET->Lang(0x0CE, ""));
      SendUnLinked(BNET->tandbot->Find(nick),
		   BNET->Lang(0x0B8, "s", nick),0);
      Close();
   } else {
      FlagRecord bfl (FlagRecord::BOT);
      TandemBot * tb = BNET->tandbot->Find(nick);
      
      bfl.Set(user);
      if (bfl.BotLeaf())
	SetStatus(Bot::LEAF);
      if (tb)
	tb->Update(this, 0, 0, BNET->BotnetFlags(bnfl));
      /* set capitalization the way they want it */
      G->noshare = 1;
      user->SetKey(bot);
      G->noshare = 0;
      SetNick(bot);
   }
}

/* handshake <newpass> */
void Bot::ProcessHandshake (char * par) {
   FUNCTION;
   /* only set a new password if no old one exists */
   context;
   if (user->PassMatch("-")) {
      G->noshare = 1; /* we *don't* want botnet passwords migrating */
      SetPassword(user,par);
      G->noshare = 0;
   }
}

void Bot::ProcessZapf (char * from, char * to, char * par) {
   FUNCTION;
   char * opcode;
   void * argv[2];
   TandemBot * tb = BNET->tandbot->Find(from);
   Bot * next = 0;
   
   context;
   if (this != tb->NextBot()) 
     SendFakeAlert( "zapf-direction",from);
   else if (!strcasecmp(to, G->botnetnick)) {
      /* for me! */
      opcode = G->Split(&par);
      argv[0] = from;
      argv[1] = par;
      BNET->bt_bot->Check(opcode, 0, argv);
   } else if ((tb = BNET->tandbot->Find(to)) && (next = tb->NextBot())) 
	next->SendZapf(from,to,par);
}

void Bot::ProcessZapfBroad (char * from, char * par) {
   FUNCTION;
   char *opcode;
   void * argv[2];
   TandemBot * tb = BNET->tandbot->Find(from);
   
   context;
   if (this != tb->NextBot())
      SendFakeAlert("zapf-broad-direction",from);
   else {
      opcode = G->Split(&par);
      argv[0] = from;
      argv[1] = par;
      BNET->bt_bot->Check(opcode, 0, argv);
      SendZapfBroad(from,opcode,par);
   }
}

void Bot::ProcessMotd (char * who, char * bot) {
   FUNCTION;
   FILE *vv;
   char *p;
   FlagRecord fr;
   TandemBot * tb;
   Bot * next = 0;
   
   context;
   fr.SetGlobBot();
   if (!bot[0] || !strcasecmp(bot, G->botnetnick)) {
      int fl = 0;
      p = (p = strchr(who, ':')) ? p + 1 : who;
      if (who[0] == '!') {
	 fl = HelpReference::IRC | HS_HILITE;
	 who++;
      } else if (who[0] == '#') {
	 fl = HS_HILITE;
	 who++;
      }
      G->LOG_CMDS.Put(BNET->Lang(0x057, "s", p));
      if ((vv = fopen(G->motdfile, "r"))) {
	 SendPriv( G->botnetnick, who, BNET->Lang(0x0B9, ""));
	 G->HelpSubst(NULL, NULL, 0, fl, NULL);
	 while (!feof(vv)) {
	    fgets(TBUF, 120, vv);
	    if (!feof(vv)) {
	       if (TBUF[strlen(TBUF) - 1] == '\n')
		  TBUF[strlen(TBUF) - 1] = 0;
	       if (!TBUF[0])
		  strcpy(TBUF, " ");
	       G->HelpSubst(TBUF, who, &fr, HelpReference::DCC, nick);
	       if (TBUF[0])
		 SendPriv( G->botnetnick, who,
				     BNET->Lang(0x0BA, "s",TBUF));
	    }
	 }
	 SendPriv( G->botnetnick, who,
		     BNET->Lang(0x0BB, ""));
	 fclose(vv);
      } else
	SendPriv( G->botnetnick, who,
			    BNET->Lang(0x0BC, ""));
   } else if ((tb = BNET->tandbot->Find(bot)) && (next = tb->NextBot()))
	next->SendMotd(who, bot);
}

/* nc <bot> <sock> <newnick> */
void Bot::ProcessNickChange (char * bot, int socki, char * newnick) {
   FUNCTION;
   char work[HANDLEN+1];
   PartyLiner * pptr;
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (this != tb->NextBot()))
	SendFakeAlert( "nickchange-direction", bot);
      else {
	 if (!(pptr = tb->FindParty(socki))) 
	   SendFakeAlert("nickchange-sock#",G->ToBase10(socki));
	 else {
	    strcpy(work,pptr->Nick());
	    pptr->SetKey(newnick);
	    DccChat::Chanout(pptr->Chan(), 
			     G->Lang(0x10D, "ss", bot,
				     G->Lang(0x11B, "ss", work, pptr->Nick())));
	    SendNkCh(pptr, work);
	 }
      }
   }   
}

/* join <bot> <nick> <chan> <flag><sock> <from> */
void Bot::ProcessJoin (char * bot, char * nick, int chan,
		  char flag, int socki, char * from, int linking) {
   FUNCTION;
   UserRecord * u;
   int i = 0;
   PartyLiner * ptr;
   void * args[1];
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (this != tb->NextBot()))
	SendFakeAlert( "join-direction", bot);
      else {
	 if ((u = G->userlist->Find(nick))) {
	    G->SPrintF(TBUF, 511, "@%s",bot);
	    TouchLaston(u,TBUF,G->now);
	 }
	 ptr = NEW ( PartyLiner ( tb, nick, chan, flag, socki, from, &i));
	 SendJoin(linking, ptr, i);
	 if (i != chan) {
	    args[0] = ptr;
	    if (i >= 0)  {
	       if (Numver() >= NEAT_BOTNET) 
		 DccChat::Chanout(i,
				G->Lang(0x10D, "ss", bot,
					G->Lang(i ? 0x117 : 0x116,"ss",
						ptr->Nick(),
						G->Lang(0x118,""))));
	       strcpy(TBUF, G->ToBase10(i));
	       BNET->bt_chpt->Check(TBUF, NULL, args); 
	    }
	    if ((Numver() >= NEAT_BOTNET) && !linking) 
	      DccChat::Chanout(chan,
			       G->Lang(0x10D, "ss", bot,
				       G->Lang(chan ? 0x113 : 0x112, "s",
					       ptr->Nick())));
	    strcpy(TBUF, G->ToBase10(chan));
	    BNET->bt_chjn->Check(TBUF, NULL, args); 
	 }
      }
   }
   context;
}

/* part <bot> <nick> <sock> [etc..] */
void Bot::ProcessPart (char * bot, char * nick, int socki, 
		       char * reason, int silent) {
   FUNCTION;
   UserRecord * u;
   PartyLiner * partyidx;
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (this != tb->NextBot())) {
	 SendFakeAlert( "party-direction", bot);
      } else {
	 if ((u = G->userlist->Find(nick))) {
	    G->SPrintF(TBUF, 511, "@%s",bot);
	    TouchLaston(u,bot,G->now); 
	 }
	 context;
	 if (!(partyidx = tb->FindParty(socki))) 
	    SendFakeAlert( "part-sock#",G->ToBase10(socki));
	 else {
	    context;
	    if ((Numver() >= NEAT_BOTNET) && !silent) 
	      DccChat::Chanout(partyidx->Chan(), 
			       G->Lang(0x10D, "ss", bot,
				       G->Lang( 0x114 + (partyidx->Chan() ? 0 : 1)
					       + ((reason && reason[0]) ? 2 : 0),
					       "ss", nick, reason)));
	    SendPart(partyidx, reason, silent);
	    DELETE ( partyidx );
	 }
      }
   }
}

/* away <bot> <sock> <message> */
void Bot::ProcessAway (char * bot, int socki, char * msg, int linking) {
   FUNCTION;
   void * args[3];
   PartyLiner * ptr;
   TandemBot * tb;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (this != tb->NextBot()))
	SendFakeAlert( "away-direction", bot);
      else {
	 args[0] = bot;
	 args[1] = &socki;
	 args[2] = msg;
	 G->bt_away->Check(bot, 0, args);
	 if (!(ptr = tb->FindParty(socki)))
	   SendFakeAlert( "away-sock#", G->ToBase10(socki));
	 else {
	    ptr->SetAway ( msg );
	    if ((Numver() >= NEAT_BOTNET) && !linking) 
	      DccChat::Chanout(ptr->Chan(),
			       G->Lang(0x10D, "ss", bot, 
				       G->Lang(msg[0] ? 0x119 : 0x11A,
					       "ss", ptr->Nick(), msg)));
	    SendAway(ptr, linking);
	 }
      }
   }
}

/* idle <bot> <sock> <#secs> [away msg]*/
void Bot::ProcessIdle (char * bot, int socki, int idle, char * msg) {
   FUNCTION;
   TandemBot * tb;
   PartyLiner * ptr;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      if (!(tb = BNET->tandbot->Find(bot)) || (this != tb->NextBot()))
	SendFakeAlert( "idle-direction", bot);
      else if (!(ptr = tb->FindParty(socki)))
	SendFakeAlert("idle-sock#", G->ToBase10(socki));
      else {
	 ptr->SetIdle ( idle );
	 ptr->SetAway ( msg );
	 SendIdle(bot, socki, idle, msg);
      }
   }
}
   
/* v <frombot> <tobot> <idx:nick> */
void Bot::ProcessVersions (char * frombot, char * tobot, char * from) {
   FUNCTION;
   ModuleEntry * me = G->module_list;
   TandemBot * tb = BNET->tandbot->Find(frombot);
   Bot * next = 0;
   
   if (!tb || (tb->NextBot() != this))
     SendFakeAlert( "versions-direction", frombot);
   else if (strcasecmp(tobot, G->botnetnick)) {
      if ((tb = BNET->tandbot->Find(tobot)) && (next = tb->NextBot()))
	next->Printf(OBUF, 511,"v%s %s %s\n", frombot, tobot, from);
   } else {
      G->SPrintF(TBUF, 511, "%s@%s", from, frombot);
      SendPriv( G->botnetnick, TBUF, BNET->Lang(0x0BD, ""));
      while ((me = me->Next())) {
	 char x[128];
	 me->Vers()->AsString(x,127);
	SendPriv( G->botnetnick, TBUF, 
			    BNET->Lang(0x0BE, "ss", me->Key(), x));
      }
      SendPriv( G->botnetnick, TBUF, BNET->Lang(0x0BF, ""));
   }
}

/* chan <from> <chan> <text> */
void Bot::BotChan (char * msg) {
   FUNCTION;
   char * from = G->Split(&msg);
   char * p = G->Split(&msg);
   ProcessChan (from, G->Base64To(p), msg);
}

/* chan <from> <chan> <text> */
void Bot::BotChatter (char * msg) {
   FUNCTION;
   char * from = G->Split(&msg);
   char * p = G->Split(&msg);
   ProcessChatter (from, G->Base64To(p), msg);
}

/* actchan <from> <chan> <text> */
void Bot::BotActChan (char * msg) {
   FUNCTION;
   char * from = G->Split(&msg);
   char * p = G->Split(&msg);
   ProcessActChan (from, G->Base64To(p), msg);
}

/* chat <from> <notice>  -- only from bots */
void Bot::BotChat (char * par) {
   FUNCTION;
   char *from = G->Split(&par);
   ProcessChat(from, par);
}

/* priv <from> <to> <message> */
void Bot::BotPriv (char * par) {
   FUNCTION;
   char *from, *p, *tobot, to[512];
   from = G->Split(&par);
   tobot = G->Split(&par);
   G->SplitC(to, tobot, '@');
   p = (p = strchr(from, '@')) ? p + 1 : from;
   ProcessPriv(from, p, to, tobot, par);
}

void Bot::BotWho (char * par) {
   FUNCTION;
   char * nick = G->Split(&par);
   char * to = G->Split(&par);
   char * num = G->Split(&par);
   ProcessWho(nick, to, G->Base64To(num));
}

/* link <from@bot> <who> <to-whom> */
void Bot::Bot_Link (char * par) {
   FUNCTION;
   char *from, *bot, *which;

   context;
   from = G->Split(&par);
   bot = G->Split(&par);
   which = G->Split(&par);
   ProcessLink(from, bot, which);
}

/* unlink <from@bot> <linking-bot> <undesired-bot> <reason> */
void Bot::BotUnLink (char * par) {
   FUNCTION;
   char *from, *bot;
   TandemBot * undes;
   
   context;
   from = G->Split(&par);
   bot = G->Split(&par);
   undes = BNET->tandbot->Find(G->Split(&par));
   ProcessUnLink(from, bot, undes, par);
}

/* bot next share? flags */
void Bot::BotUpdate (char * par)  {
   FUNCTION;
   char * bot = G->Split(&par), *vn = G->Split(&par), *fl = G->Split(&par), x;
   TandemBot * tb;
   int bfl = BNET->BotnetFlags(fl);
   
   context;
   if ((bfl & Bot::TRAFWATCH) && BNET->dont_allow_watchers) {
      BNET->LOG_BOTS.Put(BNET->Lang(0x0DA, "s s", nick, 0, bot));
      SendBye2(BNET->Lang(0x0DB, " s", 0, bot));
      SendUnLinked (BNET->tandbot->Find(nick),
		    BNET->Lang(0x0DC, "s 0s", nick, 0, bot),0);
      Close();
   } else {
	if ((x = par[0]))
	  par++;
	if ((tb = BNET->tandbot->Find(bot)))
	  tb->Update(this, x, G->Base64To(vn), bfl);
     }
}

/* nlinked <newbot> <next> <share?version>  <botnetflags>*/
void Bot::BotNLinked (char * par) {
   FUNCTION;
   char *newbot, *next, *ver, *flgs, x;
   
   context;
   newbot = G->Split(&par);
   next = G->Split(&par);
   if ((x = par[0])) 
     par++;
   else 
     x = '-';
   ver = G->Split(&par);
   flgs = G->Split(&par);
   ProcessNLinked(newbot, next, x, G->Base64To(ver), BNET->BotnetFlags(flgs));
}

/* unlinked <bot> <Reason> */
void Bot::BotUnLinked (char * par) {
   FUNCTION;
   char * bot;
   
   context;
   bot = G->Split(&par);
   ProcessUnLinked(BNET->tandbot->Find(bot), par);
}

/* traced <from@bot> <to> <chain:chain..> */
void Bot::BotTrace(char * par) {
   FUNCTION;
   char * from = G->Split(&par);
   char * dest = G->Split(&par);
   
   ProcessTrace(from, dest, par);
}

/* traced <to@bot> <chain:chain..> */
void Bot::BotTraced (char * par) {
   FUNCTION;
   char *to = G->Split(&par);
   
   ProcessTraced(to, par);
}

/* reject <from> <bot> */
void Bot::BotReject (char * par) {
   FUNCTION;
   char *from = G->Split(&par);
   
   ProcessReject(from, par);
}

/* zapf <frombot> <tobot> <code [param]>   */
void Bot::BotZapf (char * par) {
   FUNCTION;
   char * from = G->Split(&par);
   char * to = G->Split(&par);
   ProcessZapf(from, to, par);
}

/* zapf-broad <frombot> <code [param]> */
void Bot::BotZapfBroad (char * par) {
   FUNCTION;
   char * from = G->Split(&par);
   
   ProcessZapfBroad(from, par);
}

/* motd <from> <bot> */
void Bot::BotMotd (char * par) {
   FUNCTION;
   char * who = G->Split(&par);
   
   ProcessMotd(who, par);
}

/* nc <bot> <sock> <newnick> */
void Bot::BotNickChange (char * par ) {
   FUNCTION;
   char * bot = G->Split(&par);
   char *ssock = G->Split(&par);
   char *newnick = G->Split(&par);

   ProcessNickChange(bot, G->Base64To(ssock), newnick);
}

/* join <bot> <nick> <chan> <flag><sock> <from> */
void Bot::BotJoin (char * par) {
   FUNCTION;
   int linking = HasStatus(LINKING);
   char * bot = G->Split(&par);
   char * nick = G->Split(&par);
   char * x = G->Split(&par);
   char * y = G->Split(&par);

   if (bot[0] == '!') {
      linking = 1;
      bot++;
   }
   ProcessJoin(bot, nick, G->Base64To(x), y[0], G->Base64To(y+1), par, linking);
}

/* part <bot> <nick> <sock> [etc..] */
void Bot::BotPart (char * par) {
   FUNCTION;
   int silent = 0;
   char * bot = G->Split(&par);
   char * nick = G->Split(&par);
   char * etc = G->Split(&par);
   
   if (bot[0] == '!') {
      silent = 1;
      bot++;
   }
   ProcessPart(bot, nick, G->Base64To(etc), par, silent);
}

/* away <bot> <sock> <message> */
void Bot::BotAway (char * par) {
   FUNCTION;
   int linking = HasStatus(Bot::LINKING);
   char * bot = G->Split(&par);
   char * etc = G->Split(&par);

   if (bot[0] == '!') {
      linking = 1;
      bot++;
   }
   ProcessAway(bot, G->Base64To(etc), par, linking);
}

/* idle <bot> <sock> <#secs> [away msg]*/
void Bot::BotIdle (char * par) {
   FUNCTION;
   char * bot = G->Split(&par);
   char * ssock = G->Split(&par);
   char * idle = G->Split(&par);
   
   ProcessIdle(bot, G->Base64To(ssock), G->Base64To(idle), par);
}

/* v <frombot> <tobot> <idx:nick> */
void Bot::BotVersions (char * par) {
   FUNCTION;
   char * frombot = G->Split(&par);
   char * tobot = G->Split(&par);
   char * from = G->Split(&par);
   
   ProcessVersions(frombot, tobot, from);
}

void Bot::BotCrossLink ( char * par ) {
   FUNCTION;
   char * a = G->Split(&par);
   char * b = G->Split(&par);
   
   ProcessCrossLink ( a,b);
}

void Bot::BotUnCrossLink ( char * par ) {
   FUNCTION;
   char * a = G->Split(&par);
   char * b = G->Split(&par);
   
   ProcessUnCrossLink ( a,b);
}

void Bot::SendTandBut(char * buf, int len) {
   FUNCTION;
   Bot * ptr = (Bot *)G->socklist;
   
   while ((ptr = ptr->Next()))
     if ((ptr->TypeFlags() & DCT_BOT) && (ptr != this) && 
	 (ptr->Numver() > NEATER_BOTNET) && 
	 (!this || (ptr != crosslink)))
       ptr->Dump(buf, len);
}

void Bot::SendTandButIso(char * buf, int len) {
   FUNCTION;
   Bot * ptr = (Bot *)G->socklist;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   while ((ptr = ptr->Next()))
     if ((ptr->TypeFlags() & DCT_BOT) && (ptr != this)
	 && (ptr->Numver() > NEATER_BOTNET)
	 && (!this || (ptr != crosslink))
	 && !bfl.Set(ptr->User())->BotIsolate())
       ptr->Dump(buf, len);
}

void Bot::SendMotd (const char * from, const char * to) {
   FUNCTION;
   Printf("M%s %s\n", from, to);
}

void Bot::SendReject (const char * from, const char * to, const char * reason) {
   FUNCTION;
   Printf("r%s %s %s\n", from, to, reason);
}

void Bot::SendTrace (const char * to, const char * from, const char * buf) {
   FUNCTION;
   Printf("t%s %s %s:%s\n", to, from, buf, G->botnetnick);
}

void Bot::SendWho (const char * from, const char * to, int chan) {
   FUNCTION;
   Dump(OBUF, G->SPrintF(OBUF, 511,"w%s %s %D\n", from, to,chan));
}

void Bot::SendPing (void) {
   FUNCTION;
   Dump("p\n",2);
}

void Bot::SendPong (void) {
   FUNCTION;
   Dump("P\n",2);
}

void Bot::SendUnLink (const char * who, TandemBot * tb, const char * reason) {
   FUNCTION;
   Printf("L%s %s %s %s\n", who, tb->UplinkName(), tb->Name(), reason);
}   

void Bot::SendLink (const char * who, const char * via, const char * bot) {
   FUNCTION;
   Printf("l%s %s %s\n", who, via, bot);
}   

void Bot::SendTraced (const char * bot, const char * buf) {
   FUNCTION;
   Printf("T%s %s\n", bot, buf);
}
			 
void Bot::SendZapf (const char * a, const char * b, const char * c) {
   FUNCTION;
   Printf("z%s %s %s\n",a,b,c);
}

/* dump list of links to a new bot */
void Bot::SendLinkInfo (void) {
   FUNCTION;
   TandemBot * bot = BNET->tandbot;
   PartyLiner * ptr = BNET->party;
   FlagRecord bfl (FlagRecord::BOT);
   
   context;
   while ((bot = bot->Next()))
     if (strcasecmp(bot->Key(), nick))
       Dump(OBUF, G->SPrintF(OBUF, 511,"n%s %s %c%D\n", bot->Name(),
			     bot->UplinkName (), bot->Share(), bot->Version()));
   bfl.Set(user);
   if (!bfl.BotIsolate()) {
      while ((ptr = ptr->Next())) {
	 Dump(OBUF, G->SPrintF(OBUF, 511,"j%s %s %D %c%D %s\n", 
			       ptr->BotName(), ptr->Nick(), ptr->Chan(),
			       ptr->Flag(), ptr->Sock(), ptr->From()));
	 if (ptr->Away() || ptr->Timer()) 
	   Dump(OBUF, G->SPrintF(OBUF, 511,"i%s %D %D %s\n", ptr->BotName(),
				 ptr->Sock(), ptr->Age(), safe_str(ptr->Away())));
      }
   }
   Dump ( "E\n", 2);
   context;
}

void Bot::SendPriv (const char * from, const char * to, const char * msg) {
   FUNCTION;
   Printf("m%s %s %s\n",from,to, msg);
}
	      
void Bot::SendFakeAlert (const char * item, const char * extra) {
   FUNCTION;
   char * p = BNET->Lang(0x009, "sss", nick, item, extra);
   Printf("C%s %s\n", G->botnetnick, p);
   BNET->LOG_BOTS.Put(p);
}

void Bot::SendBye2 (const char * reason) {
   FUNCTION;
   Printf("b%s\n",reason);
}
