/*
 * ModuleEntryBotnet.C
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "botnet.h"
#include "ModuleEntryBotnet.h"
#include "TandemBot.h"
#include "PartyLiner.h"
#include "StringVariable.h"
#include "IntegerVariable.h"
#include "HelpReference.h"
#include "BotNetNick.h"
#include "UserEntryBotaddr.h"
#include "BindTableBot.h"
#include "BindTableChjn.h"
#include "BindTableChpt.h"
#include "BindTableLink.h"
#include "BindTableString.h"
#include "Bot.h"
#include "DccChat.h"
#include "UserEntryXtra.h"
#include "BotFork.h"
#define CLASS "ModuleEntryBotnet"

static Version botnet_version;
static struct patches botnet_patches [] = {
#include "patch.h"
};

ModuleEntryBotnet::ModuleEntryBotnet ( ModuleEntry * old, ModuleEntryEggdrop *,
				      char ** r)
: ModuleEntry ( MODULE_NAME, botnet_version.Set(2,0), botnet_patches, old ) {
   FUNCTION;
   Version required;
   
   BNET = this;
   if (!Depend("eggdrop", required.Set(2,0)))
     *r = "Your eggdrop core version is incompatible";
   LOG_BOTS = LogModes::BOTS;
   masscall_list = 0;
   tands = share_unlinks = remote_boots = dont_allow_watchers = 0;
   bot_timeout = 60;
   required_server.Set(2);
   BOT = "BOT";
   BOT_FORK = "BOT_FORK";
   BOT_LINK = "BOT_LINK";
   BOT_LOGIN = "BOT_LOGIN";
   BOT_PASS = "BOT_PASS";
   memtype_party = NEW ( MemType ("party", (List **)&party, this) );
   memtype_botnet = NEW ( MemType ("botnet", (List **)&tandbot, this) );
   tandbot = NEW ( TandemBot ( ) );
   party = NEW ( PartyLiner ( ) );
   entrytype_botaddr = NEW ( UserEntryBotaddr ( this ) );
   strcpy(network, "unknown-net");
   NEW ( StringVariable ( "network", network, 40, 0,this ) );
   NEW ( IntegerVariable ( "share_unlinks", &share_unlinks, 0, this ) );
   NEW ( IntegerVariable ( "bot_timeout", (int *)&bot_timeout, 0, this ) );
   NEW ( IntegerVariable ( "remote_boots", &remote_boots, VAR_PROTECT, this ) );
   NEW ( IntegerVariable ( "dont_allow_watchers", &dont_allow_watchers, 0, this ) );
   NEW ( BotNetNick ( this ) );
   NEW ( HelpReference ( "botnet.help", this ) );
   bt_link = NEW( BindTableLink ( this ) );
   bt_disc = NEW( BindTableString ("disc", this ) );
   bt_bot = NEW( BindTableBot ( this ) );
   bt_chpt = NEW( BindTableChpt ( this ) );
   bt_chjn = NEW( BindTableChjn ( this ) );
   bt_version = NEW( BindTable ("version", BindTable::MASK | BindTable::WANTRET
				| BindTable::STACKABLE, this));
}

ModuleEntryBotnet::~ModuleEntryBotnet ( void ) {
   FUNCTION;
   Bot * ptr;
   
   while ((ptr = ((Bot *)G->socklist)->Next()))
     DELETE ( ptr );
   DELETE ( tandbot );
   DELETE ( party );
}

int ModuleEntryBotnet::ExpMem ( void ) {
   FUNCTION;
   return ModuleEntry::ExpMem () + sizeof ( ModuleEntryBotnet );
}

void ModuleEntryBotnet::AddMassbotFuncs ( struct massbot_call_funcs * which ) {
   FUNCTION;
   struct massbot_call_funcs * ptr;
   
   for (ptr = masscall_list; ptr; ptr = ptr->next)
     if (ptr == which)
       return;
   which->next = masscall_list;
   masscall_list = which;
}

void ModuleEntryBotnet::RemMassbotFuncs ( struct massbot_call_funcs * which ) {
   FUNCTION;
   struct massbot_call_funcs * ptr, * last = NULL;
   
   for (ptr = masscall_list; ptr; last = ptr, ptr = ptr->next)
     if (ptr == which) {
	if (last)
	  last->next = last;
	else
	  masscall_list = last;
	break;
     }
}

void ModuleEntryBotnet::ChatOut ( char * what ) {
   FUNCTION;
   DccChat * ptr = (DccChat *)G->socklist;
   int l = strlen(what);
   
   what[l++] = '\n';
   what[l] = 0;
   while ((ptr = ptr->Next()))
     if (ptr->Channel() >= 0)
       ptr->Dump(what, l);
}

/* New methodology - cycle through list 3 times */
/* 1st time scan for +sh bots and link if none connected */
/* 2nd time scan for +h bots */
/* 3rd time scan for +a/+h bots */
void ModuleEntryBotnet::AutoLink (char * start) {
   FUNCTION;
   UserRecord *u = G->userlist, *autc = NULL;
   static int cycle = 0;
   int got_hub = 0, got_alt = 0, got_shared = 0, linked, ready = 0;
   FlagRecord bfl (FlagRecord::BOT | FlagRecord::GLOBAL);
   Bot * ptr = (Bot *)G->socklist;
   TandemBot * tb = 0;
   int x = 0;
   
   context;
   /* don't start a new cycle if some links are still pending */
   if (!start) 
     while ((ptr = ptr->Next()))
       if ((ptr->Type() == BNET->BOT_LINK) || (ptr->Type() == BNET->BOT_FORK))
	 return;
   if (!start) {
      ready = 1;
      cycle = 0;
   }				/* new run through the user list */
   context;
   while (u && !autc) {
      context;
      while (u && !autc) {
	 context;
	 bfl.Set(u);
	 if (bfl.GlobBot()) {
	    context;
	    if (bfl.BotHub() || bfl.BotAlt()) {
	       context;
	       linked = 0;
	       for (ptr = (Bot *)G->socklist; (ptr = ptr->Next()); )
		 if ((ptr->User() == u))
		   linked = 1;
	       if (bfl.BotHub() && bfl.BotShare()) {
		  if (linked)
		    got_shared = 1;
		  else if ((cycle == 0) && ready && !autc)
		    autc = u;
	       } else if (bfl.BotHub() && cycle > 0) {
		  if (linked)
		    got_hub = 1;
		  else if ((cycle == 1) && ready && !autc)
		    autc = u;
	       } else if (bfl.BotAlt() && (cycle == 2)) {
		  if (linked)
		    got_alt = 1;
		  else if (!BNET->tandbot->Find(u->Handle()) && ready && !autc)
		    autc = u;
	       }
	       /* did we make it where we're supposed to start?  yay! */
	       if (!ready && !strcasecmp(u->Handle(), start)) {
		  context;
		  ready = 1;
		  autc = NULL;
		  /* if starting point is a +h bot, must be in 2nd cycle */
		  if (bfl.BotHub() && !bfl.BotShare())
		    cycle = 1;
		  /* if starting point is a +a bot, must be in 3rd cycle */
		  if (bfl.BotAlt()) 
		    cycle = 2;
	       }
	    }
	    if (!cycle && bfl.BotReject() &&
		(tb = BNET->tandbot->Find(u->Handle()))) {
	       context;
	       if ((ptr = tb->NextBot())
		   && !strcasecmp(ptr->Nick(), u->Handle())) {
		  UserEntryXtra * uex = 
		    (UserEntryXtra *)u->FindEntry(G->entrytype_xtra);
		  XtraKeys * p = uex ? uex->FindKey("reject") : 0;

		  /* we're directly connected to the offending bot?! (shudder!) */
		  LOG_BOTS.Put(BNET->Lang(p ? 0x0C7 : 0x0C8, "ss",
					ptr->Nick(), p ? p->Data() : ""));
		  ptr->SendUnLinked(tb, p ? p->Data() : "",0);
		  ptr->SendBye(BNET->Lang(0x012, "s", p ? p->Data() : 0));
		  ptr->Close();
	       } else 
		 ptr->SendReject(G->botnetnick, u->Handle(), NULL);
	    }
	    context;
	 }
	 context;
	 u = u->Next();
	 
	 {
	    x++;
	    if (x > 1000) {
	       char *p = 0;
	       *p = 0;
	    }
	 }
	      
      }
      if (!autc) {
	 if ((cycle == 0) && !got_shared) {
	    cycle++;
	    u = G->userlist;
	 } else if ((cycle == 1) && !(got_shared || got_hub)) {
	    cycle++;
	    u = G->userlist;
	 }
      }
   }
   if (got_shared && (cycle == 0)) 
      autc = NULL;
   else if ((got_shared || got_hub) && (cycle == 1))
      autc = NULL;
   else if ((got_shared || got_hub || got_alt) && (cycle == 2))
      autc = NULL;
   if (autc)
     BotLink(NULL, NULL, autc->Handle());	/* try autoconnect */
   else if (!start && (G->dcc_total < G->max_dcc)) { /* check for +c bots */
      TandemBot * tb, *tb1;
      BotFork * bot;
      UserEntryBotaddr * bi = 0;
      
      for (tb = BNET->tandbot->Children(); tb; tb = tb->Peer())
	if (!tb->CrossLink()) 
	  for (tb1 = tb->Children(); tb1; tb1 = tb1->Peer()) 
	    if ((u = G->userlist->Find(tb1->Key()))) {
	       bfl.Set(u);
	       if (bfl.BotCross() &&  (bi = (UserEntryBotaddr *)
				       (u->FindEntry(BNET->entrytype_botaddr)))) {
		  bot = NEW ( BotFork ( u, bi, "",
				       (Bot *)G->socklist->Find(tb->Key(), DCT_BOT)));
		  G->LOG_MISC.Put(BNET->Lang(0x02E,"ssd",tb1->Key(),bi->Value(),bi->Telnet()));
		  if (bot->Scan(bot->RemotePort() + G->port_range))
		    bot->Eof();
		  return;
	       }
	    }
   }
}

/* link to another bot */
int ModuleEntryBotnet::BotLink (const char * linker, Socket * idx, const char * nick) {
   FUNCTION;
   UserEntryBotaddr * bi;
   UserRecord * u;
   Socket * ptr = G->socklist;
   int lang = 0;
   FlagRecord bfl;
   BotFork * bot;
   
   context;
   u = G->userlist->Find(nick);
   if (!u || !bfl.Set(u)->GlobBot())
     lang = 0x028;
   else if (!strcasecmp(nick, G->botnetnick)) 
     lang = 0x029;
   else if (BNET->tandbot->Find(nick) && (idx || linker)) 
     lang = 0x02A;
   else {
      while ((ptr = ptr->Next()))
	if ((ptr->User() == u) && ((ptr->Type() == BNET->BOT_FORK)
	    || (ptr->Type() == BNET->BOT_LINK)) 
	    || (ptr->Type() == BNET->BOT_PASS)) {
	   lang = 0x027;
	   break;
	}
      if (!ptr) {
	 /* address to connect to is in 'info' */
	 bi = (UserEntryBotaddr *)(u->FindEntry(BNET->entrytype_botaddr));
	 if (!bi) {
	    if (idx)
	      idx->Send(BNET->Lang(0x02B, "s", nick));
	    lang = 0x02D;
	 } else if (G->dcc_total == G->max_dcc) 
	      lang = 0x02C;
	 else {
	    context;
	    u->SetKey(nick);
	    bot = NEW ( BotFork ( u, bi, linker ) );
	    bot->SetNumver (idx || linker);
	    if (!idx && linker)
	      BNET->LOG_BOTS.Put(BNET->Lang(0x02E,"ssd",nick,bi->Value(),
					    bi->Telnet()));
	    else if (idx)
	      idx->Send(BNET->Lang(0x02E,"ssd",nick,bi->Value(),bi->Telnet()));
	    if (!bot->Scan(bot->RemotePort() + G->port_range))
	      return 1;
	    bot->Eof();
	 }
      }
   }
   if (idx && lang) 
     idx->Send(BNET->Lang(lang, "s", nick));     
   return 0;
}

int ModuleEntryBotnet::Unlink ( const char * key, Socket * idx, const char * reason ) {
   FUNCTION;
   Bot * ptr = (Bot *)G->socklist->Find( key, DCT_BOT );
   int lang = 0;
   char * t;
   
   if (ptr) {
      if (ptr->Type() == BOT_FORK)
	lang = 0x030;
      else if ((ptr->Type() == BOT_LOGIN) || (ptr->Type() == BOT_PASS)
	       || (ptr->Type() == BOT_LINK))
	lang = 0x031;
      else {
	 context;
	 if (!idx && !strncmp(ptr->Type(),"BOTS",4) && !share_unlinks)
	   return -1;
	 lang = 0x032;
	 ptr->SendBye2(reason);
	 ptr->SendUnLinked(tandbot->Find(key),
			   Lang((reason && reason[0])?0x033:0x034,
				"ss", ptr->Nick(), reason),0);
      }
      t = Lang(lang, "ssd", ptr->Nick(), ptr->Host(),ptr->RemotePort());
      LOG_BOTS.Put(t);
      if (idx) 
	idx->Send(t);
      ptr->Close();
      return 1;
   }
   return 0;
}

int ModuleEntryBotnet::BotnetFlags ( const char * fl ) {
   FUNCTION;
   int ret = 0;
   
   while (*fl) 
     switch (*fl++) {
       case 't':
	ret |= Bot::TRAFWATCH;
	break;
       case 's':
	ret |= Bot::SERVICE;
	break;
       case 'j':
	ret |= Bot::JUPE;
	break;
     }
   return ret;
}

const char * ModuleEntryBotnet::BotnetFlags ( int fl ) {
   FUNCTION;
   static char buf[5];
   int p = 0;
   
   if (fl & Bot::TRAFWATCH)
     buf[p++] = 't';
   if (fl & Bot::SERVICE)
     buf[p++] = 's';
   if (fl & Bot::JUPE)
     buf[p++] = 'j';
   if (!p)
     buf[p++] = '-';
   buf[p] = 0;
   return buf;
}
