/*
 * TandemBot.C - class members for TandemBot class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "botnet.h"
#include "TandemBot.h"
#include "BindTable.h"
#include "PartyLiner.h"
#include "Version.h"
#include "Bot.h"
#define CLASS "TandemBot"

int TandemBot::ExpMem ( void ) {
   FUNCTION;
   return sizeof(TandemBot) + (next ? Next()->ExpMem() : 0);
}

void * TandemBot::operator new ( size_t s, int x ) {
   FUNCTION;
   context;
   return BNET->memtype_botnet->Malloc(s,x);
}

void TandemBot::operator delete ( void * ptr ) {
   FUNCTION;
   BNET->memtype_botnet->Free(ptr);
}

TandemBot::TandemBot ( void ) : KeyedList<HANDLEN> ( ) {
   FUNCTION;
}

TandemBot::TandemBot ( TandemBot * up, const char * who, char f, int v, int b )
: KeyedList<HANDLEN> ( *BNET->tandbot, who, KeyedList<HANDLEN>::APPEND |
		      KeyedList<HANDLEN>::NODUPS ) {
   FUNCTION;
   share = f;
   ver = v;
   botnet_flags = b;
   uplink = (up == BNET->tandbot ? 0 : up);
   peer = up->child;
   up->child = this;
   BNET->tands++;
}

TandemBot::~TandemBot ( void ) {
   FUNCTION;
   struct massbot_call_funcs * t;
   void * args;
   
   context;
   for (t = BNET->masscall_list; t; t = t->next)
     t->rembot(this);
   args = (void *)this;
   BNET->tandbot->Remove(this);
   BNET->tands--;
   BNET->bt_disc->Check(key, 0, &args);
   while (partyliners)
     DELETE ( partyliners );
   while (child)
     DELETE ( child );
   if (!uplink)
     BNET->tandbot->RemoveChild(this);
   else
     uplink->RemoveChild(this);
}
		       
void TandemBot::RemoveChild ( TandemBot * one) {
   FUNCTION;
   TandemBot * ptr, * old;
   
   for (ptr = child, old = NULL; ptr; old = ptr, ptr = ptr->peer)
     if (ptr == one) {
	if (old)
	  old->peer = one->peer;
	else
	  child = one->peer;
	break;
     }
}
 
PartyLiner * TandemBot::FindParty ( int sock ) {
   FUNCTION;
   PartyLiner * ptr;
   
   for (ptr = partyliners; ptr; ptr = ptr->NextByBot())
     if (ptr->Sock() == sock)
       break;
   return ptr;
}

PartyLiner * TandemBot::FindParty ( char * name, int chan ) {
   FUNCTION;
   PartyLiner * ptr;
   
   for (ptr = partyliners; ptr; ptr = ptr->NextByBot())
     if ((ptr->Chan() == chan) && !strcasecmp(ptr->Key(), name))
       break;
   return ptr;
}

/* return index into dcc list of the bot that connects us to bot <x> */
Bot * TandemBot::NextBot ( void ) {
   FUNCTION;
   Bot * ptr = (Bot *)G->socklist;
   TandemBot * tb = this;
   
   if (tb) {
      while (tb->uplink)
	tb = tb->uplink;
      while ((ptr = ptr->Next()))
	if (!strcasecmp(ptr->Nick(), tb->Name()) && (ptr->TypeFlags() & DCT_BOT))
	  break;
   } else
     ptr = 0;
   return ptr;
}

void TandemBot::Update (Bot * sock, char s, int vernum, int flags) {
   FUNCTION;
   context;
   if (s) 
     share = s;
   if (vernum)
     ver = vernum;
   if (flags)
     botnet_flags = flags;
   sock->SendUpdate(this);
}

int TandemBot::Dump ( Socket * idx, char * work, int len, int lang ) {
   FUNCTION;
   char c, *s, buf[512];
   TandemBot * ptr;
   int tothops = len / 5, l = 0;

   if (botnet_flags & Bot::TRAFWATCH)
     l += G->SPrintF(buf + l, 511 - l, " (SPY)");
   if (botnet_flags & Bot::SERVICE)
     l += G->SPrintF(buf + l, 511 - l, " (SERVICE)");
   buf[l] = 0;
   if (!key[0]) 
     idx->Send(BNET->Lang(lang, "sdddds", G->botnetnick,G->Vers()->Major(),
			  G->Vers()->Minor(),G->Vers()->Fix(),G->Vers()->Patch(),
			  buf));
   else {
      if (ver) {
	 s = BNET->Lang(lang, "sdddds", key, ver/1000000, ver%1000000/10000,
			ver%10000/100, ver%100, buf);
	 c = share;
      } else {
	 c = '-';
	 s = key;
      }
      if (len)
	work[len - 2] = 0;
      idx->Printf("%s--%c%s\n", work, c, s);
   }
   if (len > 100)
      idx->Send(BNET->Lang(0x062, ""));
   else {
      if (child) {
	 if (len) {
	    if (work[len - 3] == '`') 
	      work[len - 3] = ' ';
	    if (crosslink)
	      strcpy(work + len - 2, "  | |");
	    else
	      strcpy(work + len - 2, "    |");
	 } else
	   strcpy(work, "  |");
      }
      for (ptr = child; ptr; ptr = ptr->peer) {
	 if (!ptr->peer) {
	    if (crosslink)
	      strcpy(work + len, "| `");
	    else
	      strcpy(work + len, "  `");
	 }
	 tothops += ptr->Dump(idx, work, len + 5, lang);
      }
      if (crosslink && (crosslink == peer) && len) {
	 work[len - 2] = 0;
	 idx->Printf("%s  |\n", work);
      }
   }
   return tothops;
}
