/*
 * UserEntryBotaddr.C - class members for UserEntryBotaddr class
 */
/*
 * This file is part of the eggdrop2 source code
 * copyright (c) 1998 Darrin Smith
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

#include "Bot.h"
#include "botnet.h"
#include "UserEntryBotaddr.h"
#include "Socket.h"
#include "FlagRecord.h"
#include "Bot.h"
#include "ScriptContext.h"
#include "ScriptArgument.h"
#include "UserEntryUnknown.h"
#define CLASS "UserEntryBotaddr"

UserEntryBotaddr::~UserEntryBotaddr ( void ) {
   FUNCTION;
}

void UserEntryBotaddr::Pack (UserEntry * e) {
   FUNCTION;
   UserEntryUnknown * ueu = 
     (UserEntryUnknown *)e->Owner()->GetEntry(G->entrytype_unknown);
   UserEntryBotaddr * ueb = (UserEntryBotaddr *)e;
   char work[512];
   int l; 
   
   l = G->SPrintF(work, 511, "%s:%d/%d", ueb->Value(), ueb->Telnet(), 
		  ueb->Relay());
   ueu->NewLine((char *)type, work);
}

UserEntry * UserEntryBotaddr::Create ( UserRecord * u ) {
   FUNCTION;
   return NEW ( UserEntryBotaddr ( u, this ) );
}

void UserEntryBotaddr::Set ( char * data ) {
   FUNCTION;
   char * q, *p;
   int x = G->noshare;
   FlagRecord bfl;
   
   G->noshare = 1;
   if (!(q = strchr(data,':'))) {
      UserEntryString::Set(data);
      telnet_port = 3333;
      relay_port = 3333;
   } else {
      *q = 0;
      UserEntryString::Set(data);
      *q = ':';
      p = q + 1;
      telnet_port = atoi(p);
      if (!(q = strchr(p, '/'))) 
	relay_port = telnet_port;
      else 
	relay_port = atoi(q+1);
   }
   G->noshare = x;
   if (!G->noshare && G->shareout && !(owner && bfl.Set(owner)->GlobUnshared()))
     G->shareout(NULL,"c%s %s %s %d %d\n", type->Name(),owner->Handle(),
		 string, telnet_port, relay_port);
}

void UserEntryBotaddr::GotShare (char * buf, Socket * idx) {
   FUNCTION;
   char *key;
   int i, x = G->noshare;
   
   G->noshare = 1;
   key = G->Split(&buf);
   UserEntryString::Set(key);
   key = G->Split(&buf);
   i = atoi(key);
   telnet_port = i ? i : 3333;
   i = atoi(buf);
   relay_port = i ? i : 3333;
   G->LOG_CMDS.Put(BNET->Lang(0x00B, "ss", idx->Nick(),owner->Handle()));
   G->noshare = x;
}

int UserEntryBotaddr::Write ( FILE * f, Socket * idx ) {
   FUNCTION;
   if (!idx)
     return (fprintf(f,"=%s %s:%d/%d\n", type->Name(), string, telnet_port,
		     relay_port) != EOF);
   return 1;
}

int UserEntryBotaddr::ScriptGet (ScriptContext * sc, int, ScriptArgument *) {
   FUNCTION;
   char number[512];
   
   context;
   G->SPrintF(number, 511, "%s %d %d", string, telnet_port, relay_port);
   sc->SetString(number);
   return SCRIPT_OK;
}

int UserEntryBotaddr::ScriptSet (ScriptContext * sc, int argc, 
				 ScriptArgument * argv) {
   FUNCTION;
   int x = G->noshare;
   char buf[512];
   
   FlagRecord bfl ( owner );
   BADARGS(4,6,BNET->Lang(0x00C, ""));
   if (bfl.GlobBot()) {
      G->noshare = 1;
      /* silently ignore for users */
      strncpy(buf, argv[3].AsString(),511);
      buf[511] = 0;
      UserEntryString::Set(buf);
      if (argc > 4) 
	telnet_port = argv[4].AsInteger();
      if (argc > 5)
	relay_port = argv[5].AsInteger();
      G->noshare = x;
   }
   return SCRIPT_OK;
}

int UserEntryBotaddr::ExpMem ( void ) {
   FUNCTION;
   return UserEntryString::ExpMem() + sizeof(UserEntryBotaddr)
     - sizeof(UserEntryString);
}

void UserEntryBotaddr::Display (Socket * idx) {
   FUNCTION;
   idx->Send(BNET->Lang(0x00D, "s", string));
   idx->Send(BNET->Lang(0x00E, "dd", telnet_port, relay_port));
}
